/**********************************************************************************************************************
 * @file            I3E.h
 *
 * @brief           The IEEE 2030.5 Client Application (I3E) establishes an internet connection to an IEEE 2030.5 
 *                  utility server. The client polls the server for relevant information, schedules and executes DER 
 *                  Control events, and provides the server updates in accordance with the IEEE 2030.5 (CSIP) 
 *                  specification. 
 *
 * @author          Vlad Bugayev
 * @date            02/20/2024
 * @copyright       2024 Lion Energy LLC.
 *                  All rights reserved
 *
 *********************************************************************************************************************/
/**
 * @addtogroup      I3E
 *                  IEEE 2030.5 Client Application Module
 *
 * @ref             I3E.h
 *
 *********************************************************************************************************************/

#include "COD_Compiler_Defines.h"

#ifndef LION_I3E_H__
#define LION_I3E_H__

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#include "lion/icc/ICC.h"

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifdef __cplusplus
extern "C"
{
#endif

/**********************************************************************************************************************
 * @cond                                              GLOBAL SCOPE                                             @endcond
 *********************************************************************************************************************/

/**
 * @brief   Initializes the I3E module.
 * 
 * @details This function initializes the I3E module and performs the intitial IEEE 2030.5 server resource discovery.
 * If this process fails, this function will exit before launching the application threads.
 * 
 */
GlobScop void I3E_Init();

/**
 * @brief   Helper function for getting lfdi and sfdi calculations based on loaded PKI certificates.
 * 
 * @details This function accesses the NSS module to get the lfdi and sfdi values based on the loaded PKI certificates.
 * This function is designed to be called from the CST_IAL 
 * 
 * @param sz_lfdi The LFDI value storage pointer (char).
 * @param sz_sfdi The SFDI value storage pointer (char).
 * 
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
 */
GlobScop int I3E_GetCertIndicators(char* sz_lfdi, char* sz_sfdi);

/**
 * @brief  Helper function for starting the I3E commissioning process.
 * 
 * @details This function is called from the CST_IAL module to start the I3E commissioning process. Included 
 * in this process is checking for existing PKI certificates and generating them if they do not exist by calling 
 * the Lion AWS PKI portal.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
 */
GlobScop int I3E_Commissioning(void);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifdef __cplusplus
}
#endif

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#endif /* LION_I3E_H__ */

