#include "lion/log/log.h"
#include "lion/I3E.h"
#include "lion/CST.h"

#include "I3E_NSS.hpp"
#include "I3E_CTS.hpp"
#include "I3E_SPP.hpp"
#include "I3E_SDA.hpp"
#include "I3E_DEM.hpp"
#include "I3E_DEH.hpp"
#include "I3E_EVQ.hpp"

LOG_MODULE_REGISTER(i3e, LOG_LEVEL_INF);

#define I3E_HEXCHAR_LEN 2
#define I3E_LFDI_LEN 40
#define I3E_SFDI_LEN 12

/**********************************************************************************************************************
 * @cond                                               FILE SCOPE                                              @endcond
/*********************************************************************************************************************/

FileScop pthread_t t_i3e_spp;
FileScop pthread_t t_i3e_dem;
FileScop pthread_t t_i3e_deh;
FileScop pthread_t t_i3e_sda;

/**
 * @brief   Thread launch point for the IEEE 2030.5 Poll Processor.
 *  
 * @details This function SHOULD NEVER RETURN. It contains the thread loop.
 */
FileScop void *i3e_PollProcessor(void*);  

/**
 * @brief   Thread launch point for the IEEE 2030.5 Event Manager.
 * 
 * @details This function SHOULD NEVER RETURN. It contains the thread loop.
 */
FileScop void *i3e_EventManager(void*);

/**
 * @brief   Thread launch point for the IEEE 2030.5 Event Handler.
 * 
 * @details This function SHOULD NEVER RETURN. It contains the thread loop.
 */
FileScop void *i3e_EventHandler(void*);

/**
 * @brief   Thread launch point for the IEEE 2030.5 SDA Reporter.
 * 
 * @details This function SHOULD NEVER RETURN. It contains the thread loop.
 */
FileScop void *i3e_SDAReporter(void*);

/**********************************************************************************************************************
 * @cond                                              GLOBAL SCOPE                                             @endcond
 *********************************************************************************************************************/

GlobScop void I3E_Init() 
{
    LOG_INF("IEEE 2030.5 CLIENT APPLICATION INITIALIZATION");

    StackScop bool bol_res = false;
    StackScop uint8_t u8_connAttempts = 0;

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Initialze timing resource protection
    if(I3E_CTS_Init() != 0)
    {
        LOG_ERR("Failed to initialize the I3E_CTS module - EXIT");
        return;
    }

    // Initialize the compliance standard translaation service  
    if(CST_Init() != 0)
    {
        LOG_ERR("Failed to initialize the CST module - EXIT");
        return;
    }

    // Initialize TLS Context
    if(I3E_NSS_Init() != 0)
    {
        LOG_ERR("Failed to initialize the I3E_NSS module - EXIT");
        return;
    }

    // Initialize HTTP 
    if(I3E_HTTP_Init() != 0)
    {
        LOG_ERR("Failed to initialize the I3E_HTTP module - EXIT");
        return;
    }

    // Initialize the DERC Event Manager
    if(I3E_DEM_Init() != 0)
    {
        LOG_ERR("Failed to initialize the I3E_DEM module - EXIT");
        return;
    };

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    LOG_INF("Attempting Connection to IEEE 2030.5 Server...");
    while(!bol_res && u8_connAttempts < 5 && I3E_NSS_GetCertValid()) {
        LOG_INF("Attempt %d", u8_connAttempts);
        bol_res = I3E_SPP_InitResourceDiscovery();
        u8_connAttempts ++;
    }

    if(!I3E_NSS_GetCertValid())
    {
        LOG_ERR("Server PKI certificate verification failure - EXIT");
        return;
    }

    // Check the outcome
    if (bol_res) {
        LOG_INF("Server connection successful");
        LOG_INF("Launching IEEE 2030.5 Client Application");
    } else {
        LOG_ERR("Server connection failed after 5 attempts - EXIT");
        return;
    }
    
    // Initial post of DERSettings and DERCapability
    I3E_SDA_InitResourcePost();
    
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    if(pthread_create(&t_i3e_spp, NULL, &i3e_PollProcessor, NULL) != 0)
    {
        LOG_ERR("Failed to create the I3E Poll Processor thread - EXIT");
        return;
    }

    if(pthread_create(&t_i3e_dem, NULL, &i3e_EventManager, NULL) != 0)
    {
        LOG_ERR("Failed to create the I3E Event Manager thread - EXIT");
        return;
    }

    if(pthread_create(&t_i3e_deh, NULL, &i3e_EventHandler, NULL) != 0)
    {
        LOG_ERR("Failed to create the I3E Event Handler thread - EXIT");
        return;
    }

    if(pthread_create(&t_i3e_sda, NULL, &i3e_SDAReporter, NULL) != 0)
    {
        LOG_ERR("Failed to create the I3E SDA Reporter thread - EXIT");
        return;
    }

}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop int I3E_GetCertIndicators(char* ptr_sz_lfdi, char* ptr_sz_sfdi) 
{
    
    // Create empty containers for the indicators & calculate the indicators
    StackScop SE_HexBinary160_t i3e_lFDI = {0};
    StackScop SE_SFDIType_t i3e_sFDI = 0;
    if (I3E_NSS_CalculateIndicators(&i3e_lFDI, &i3e_sFDI) != 0) 
    {
        LOG_ERR("Failed to calculate the PKI certificate indicators");
        return -EXIT_FAILURE;
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Convert i3e_lFDI (HexBinary160) to string
    StackScop std::string sz_lfdi;
    for (size_t i = 0; i < sizeof(i3e_lFDI); ++i) 
    {
        // 2 characters for hex + null terminator. Format as two-digit hex
        StackScop char buf[I3E_HEXCHAR_LEN + 1]; 
        snprintf(buf, sizeof(buf), "%02X", i3e_lFDI[i]);
        sz_lfdi += buf;
    }

    LOG_INF("LFDI String: %s\n", sz_lfdi.c_str());

    // Copy to ptr_sz_lfdi
    if (sz_lfdi.length() == I3E_LFDI_LEN) 
    {
        for (size_t i = 0; i < I3E_LFDI_LEN; ++i) {
            ptr_sz_lfdi[i] = sz_lfdi[i];
        }
        ptr_sz_lfdi[I3E_LFDI_LEN] = '\0'; 
    } 
    else 
    {
        LOG_ERR("LFDI string conversion failed due to unexpected length");
        return -EXIT_FAILURE;
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Convert i3e_sFDI (UInt40) to string
    StackScop std::string sz_sfdi = std::to_string(i3e_sFDI);

    // Copy to ptr_sz_sfdi
    if (sz_sfdi.length() == I3E_SFDI_LEN) 
    {
        for (size_t i = 0; i < I3E_SFDI_LEN + 1; ++i) {
            ptr_sz_sfdi[i] = sz_sfdi[i];
        }
        ptr_sz_sfdi[I3E_SFDI_LEN] = '\0'; 
    } 
    else 
    {
        LOG_ERR("SFDI string conversion failed due to unexpected length");
        return -EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
    
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop int I3E_Commissioning(void)
{

    // Check for existing PKI certificates
    StackScop int16_t s16_CheckPKISt = I3E_NSS_CheckPKI();
    if(s16_CheckPKISt != EXIT_SUCCESS)
    {
        if(s16_CheckPKISt == -EALREADY)
        {
            LOG_INF("I3E is already commissioned");
            return -EALREADY;
        }
        if(s16_CheckPKISt == -EXIT_FAILURE)
        {
            LOG_ERR("Failed to check PKI certificates - EXIT");
            return -EXIT_FAILURE;
        }
    }

    // Generate new PKI certificates
    StackScop int16_t s16_GenPKISt = I3E_NSS_GeneratePKI();
    if(s16_CheckPKISt != EXIT_SUCCESS)
    {
        LOG_ERR("Failed to generate PKI certificates - EXIT");
        return -EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


/**********************************************************************************************************************
 * @cond                                               FILE SCOPE                                              @endcond
/*********************************************************************************************************************/

FileScop void *i3e_PollProcessor(void*) 
{

    I3E_SPP_Exec();

    // THIS SHOULD NEVER BE REACHED
    return nullptr;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void *i3e_EventManager(void*) 
{

    I3E_DEM_Exec();

    // THIS SHOULD NEVER BE REACHED
    return nullptr;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void *i3e_EventHandler(void*) 
{

    I3E_DEH_Exec();

    // THIS SHOULD NEVER BE REACHED
    return nullptr;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void *i3e_SDAReporter(void*) 
{

    I3E_SDA_Exec();

    // THIS SHOULD NEVER BE REACHED
    return nullptr;
}