#include "lion/log/log.h"
#include "include/I3E_DAG.hpp"

LOG_MODULE_REGISTER(i3e_dag, LOG_LEVEL_INF);
/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop NodeData::NodeData() : u32_pollRate(0), time_pollNext(0), sz_resourceName(""),
                        sz_lastPollStr(""), bol_updated(false), var_rsrc() {}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop NodeData::NodeData(const std::string& sz_name, const RESOURCE_VARIANT& var_data) 
                        : sz_resourceName(sz_name), var_rsrc(var_data) {}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop bool NodeData::hasPollRate() const 
{
    return u32_pollRate != 0;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop uint32_t NodeData::getPollRate() const 
{
    return u32_pollRate;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop bool NodeData::isEmpty() const 
{
    return sz_resourceName.empty();
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop Graph::Graph() 
{
    ModScop pthread_mutexattr_t attr;
    pthread_mutexattr_init(&attr);
    pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
    pthread_mutex_init(&mtx_DAGMutex, &attr); // Initialize the pthread_mutex as recursive
    pthread_mutexattr_destroy(&attr); // Destroy the mutex attribute after initializing the mutex
}
    
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop Graph::~Graph() 
{
    pthread_mutex_destroy(&mtx_DAGMutex); // Initialize the pthread_mutex
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief Getter method to access the DAGMutex
 * 
 * @details This method returns a reference to the DAGMutex for shared access.
 * 
 * @returns Reference to the DAGMutex
 */
// Getter method to access the DAGMutex
ModScop pthread_mutex_t& Graph::I3E_DAG_GetDAGMutex() {
    return mtx_DAGMutex;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief Add a node to the graph
 * 
 * @details This method adds a node to the graph with the specified node name and NodeData.
 * 
 * @param sz_nodeName The name of the node to be added
 * @param str_data The NodeData associated with the node
 */
// Add a node to the graph 
ModScop void Graph::I3E_DAG_AddNode(const std::string& sz_nodeName, const NodeData& str_data) 
{
    pthread_mutex_lock(&mtx_DAGMutex);
    nodes[sz_nodeName] = str_data;
    pthread_mutex_unlock(&mtx_DAGMutex);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief Add an edge between two nodes in the graph
 * 
 * @details This method adds an edge between two nodes in the graph.
 * 
 * @param sz_fromNode The name of the node from which the edge originates
 * @param sz_toNode The name of the node to which the edge points
 */
ModScop void Graph::I3E_DAG_AddEdge(const std::string& sz_fromNode, const std::string& sz_toNode) 
{
    pthread_mutex_lock(&mtx_DAGMutex);
    adjacencyList[sz_fromNode].insert(sz_toNode);
    pthread_mutex_unlock(&mtx_DAGMutex);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief Print the details of a node in the graph
 * 
 * @details This method prints the details of a node in the graph, including its children nodes. This is 
 * a recursive function that prints the node name and then calls itself on each child node. This function is 
 * really just for debugging purposes
 * 
 * @param sz_nodeName The name of the node to print
 * @param depth The depth of the node in the graph
 */
ModScop void Graph::I3E_DAG_PrintNodes(const std::string& sz_nodeName, int depth = 0) 
{
    pthread_mutex_lock(&mtx_DAGMutex);

    // Pretty printing based on depth of the node 
    for (int i = 0; i < depth; ++i) 
    {
        std::cout << "  ";
    }
    std::cout << sz_nodeName << std::endl;

    // If the current node exists in the adjacency list, it has children nodes
    // proceed to loop through all children and recursively call I3E_DAG_PrintNodes
    if (adjacencyList.find(sz_nodeName) != adjacencyList.end()) 
    {
        for (const std::string& sz_child : adjacencyList[sz_nodeName]) 
        {
            I3E_DAG_PrintNodes(sz_child, depth + 1);
        }
    }
    pthread_mutex_unlock(&mtx_DAGMutex);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief Function to find a node in the graph and return its NodeData if it exists
 * 
 * @details This function finds a node in the graph with the specified node name and returns its NodeData 
 * if it exists.
 * 
 * @param sz_nodeName The name of the node to find
 * 
 * @returns The NodeData associated with the node
 */
ModScop NodeData Graph::I3E_DAG_GetNode(const std::string& sz_nodeName) 
{
    ModScop NodeData str_returnNode;
    pthread_mutex_lock(&mtx_DAGMutex);
    ModScop auto iter = nodes.find(sz_nodeName);
    if (iter != nodes.end()) 
    {
        pthread_mutex_unlock(&mtx_DAGMutex);
        str_returnNode = iter->second;
        return str_returnNode; // Return the NodeData associated with the node
    } 
    else 
    {
        // Node not found
        LOG_ERR("Node '%s' not found in the graph.", sz_nodeName.c_str());
        pthread_mutex_unlock(&mtx_DAGMutex);
        return str_returnNode;
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief Find resource nodes of a specific type in the graph
 * 
 * @details This method finds all nodes of a specific resource type in the graph and returns a vector of
 * their names for further processing.
 * 
 * @param enum_targetType The resource type to search for
 * 
 * @returns A vector of strings containing the names of nodes of the specified resource type
 */        
ModScop std::vector<std::string> Graph::I3E_DAG_FindResourceNodes(I3EresourceType enum_targetType) 
{
    pthread_mutex_lock(&mtx_DAGMutex);
    ModScop std::vector<std::string> vec_nodesOfType; // Vector to store nodes of the specified resource type
    for (const auto& node : nodes) 
    {
        if (node.second.enum_rsrcType == enum_targetType) {
            vec_nodesOfType.push_back(node.first); // Add node name to the vector if it matches the resource type
        }
    }
    pthread_mutex_unlock(&mtx_DAGMutex);
    return vec_nodesOfType;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief Update the content of a node in the graph
 * 
 * @details This method updates the content of a node in the graph with the specified node name and new content.
 * 
 * @param sz_nodeName The name of the node to update
 * @param str_newContent The new content to update the node with
 */
ModScop void Graph::I3E_DAG_UpdateNodeContent(const std::string& sz_nodeName, const NodeData& str_newContent) 
{
    pthread_mutex_lock(&mtx_DAGMutex);
    if (nodes.find(sz_nodeName) != nodes.end())
    {
        // Node found, update its content
        nodes[sz_nodeName] = str_newContent;
    } else {
        // Node not found in the graph
        LOG_ERR("Node '%s' not found in the graph.", sz_nodeName.c_str());
    }
    pthread_mutex_unlock(&mtx_DAGMutex);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief Find nodes of a specific resource type starting from a given node.
 * 
 * @details This method performs a breadth-first search (BFS) traversal starting from a given node to find all 
 * nodes of a specific resource type. Used in the resource rediscovery process.
 * 
 * @param sz_startNode The name of the node to start the search from
 * @param enum_targetType The resource type to search for
 */
ModScop std::vector<std::string> Graph::I3E_DAG_FindResourceNodesFromNode(const std::string& sz_startNode, 
                                I3EresourceType enum_targetType) 
{
    pthread_mutex_lock(&mtx_DAGMutex);

    ModScop std::vector<std::string> vec_nodesOfType; // Vector to store nodes of the specified resource type
    ModScop std::unordered_set<std::string> set_visited; // Set to track visited nodes during traversal
    ModScop std::queue<std::string> queue; // Queue for BFS traversal

    set_visited.insert(sz_startNode); // Mark the starting node as visited
    queue.push(sz_startNode); // Add the starting node to the queue for traversal

    while (!queue.empty()) 
    {
        ModScop std::string sz_currentNode = queue.front(); // Get the front node from the queue
        queue.pop(); // Remove the front node from the queue

        if (nodes.find(sz_currentNode) != nodes.end() && nodes[sz_currentNode].enum_rsrcType == enum_targetType) 
        {
            vec_nodesOfType.push_back(sz_currentNode); // Add node name to the vector if it matches the resource type
        }

        if (adjacencyList.find(sz_currentNode) != adjacencyList.end()) 
        {
            for (const std::string& neighbor : adjacencyList[sz_currentNode]) 
            {
                if (set_visited.find(neighbor) == set_visited.end()) {
                    set_visited.insert(neighbor); // Mark the neighbor as visited
                    queue.push(neighbor); // Add the neighbor to the queue for traversal
                }
            }
        }
    }

    pthread_mutex_unlock(&mtx_DAGMutex);
    return vec_nodesOfType;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief Check if a node exists in the graph
 * 
 * @details This method checks if a node with the specified name exists in the graph and returns a boolean 
 * value indicating its presence.
 * 
 * @param sz_nodeName The name of the node to check for
 * 
 * @returns A boolean value indicating whether the node exists in the graph
 */
ModScop bool Graph::I3E_DAG_DoesNodeExist(const std::string& sz_nodeName) 
{

    pthread_mutex_lock(&mtx_DAGMutex); 
    ModScop bool bol_nodeExists = nodes.find(sz_nodeName) != nodes.end();
    pthread_mutex_unlock(&mtx_DAGMutex); 
    return bol_nodeExists;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief Recursively scan all nodes in the graph for poll rates.
 * 
 * @details Utility functions used by I3E_DAG_ScanAllNodesWithPollRate to recursively scan all nodes in the graph
 * for nodes that contain a poll rate attribute. 
 * 
 * @param sz_currentNode The name of the current node being processed
 * @param set_visited A set to track visited nodes during traversal
 * @param vec_nodesWithPollRate A vector to store pointers to NodeData for nodes with poll rates
 */
ModScop void Graph::I3E_DAG_ScanAllNodesRecursiveUtil(const std::string& sz_currentNode, 
                                                std::unordered_set<std::string>& set_visited, 
                                                std::vector<NodeData*>& vec_nodesWithPollRate) 
{
    pthread_mutex_lock(&mtx_DAGMutex);
    set_visited.insert(sz_currentNode);
    ModScop NodeData& str_currentData = nodes[sz_currentNode];

    // Check if the current node has a poll rate attribute
    if (str_currentData.hasPollRate()) {
        vec_nodesWithPollRate.push_back(&str_currentData);
    }
    // Traverse neighbors recursively
    for (const auto& neighbor : adjacencyList[sz_currentNode]) 
    {
        if (set_visited.find(neighbor) == set_visited.end()) 
        {
            I3E_DAG_ScanAllNodesRecursiveUtil(neighbor, set_visited, vec_nodesWithPollRate);
        }
    }
    pthread_mutex_unlock(&mtx_DAGMutex);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief Scan all nodes in the graph for poll rates.
 * 
 * @details This method scans all nodes in the graph for poll rates and returns a vector of NodeData pointers
 * to the nodes that have a resource with a poll rate. Used by the SPP module to determine which resources 
 * need to be polled.
 * 
 * @returns A vector of NodeData pointers to nodes with poll rates
 */
ModScop std::vector<NodeData*> Graph::I3E_DAG_ScanAllNodesWithPollRate() 
{
    pthread_mutex_lock(&mtx_DAGMutex); 

    ModScop std::unordered_set<std::string> set_visited;
    ModScop std::vector<NodeData*> vec_nodesWithPollRate;

    for (const auto& node : nodes) {
        if (set_visited.find(node.first) == set_visited.end()) {
            I3E_DAG_ScanAllNodesRecursiveUtil(node.first, set_visited, vec_nodesWithPollRate);
        }
    }
    pthread_mutex_unlock(&mtx_DAGMutex);
    return vec_nodesWithPollRate;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// Function to remove a node and its dependencies from the graph
/**
 * @brief Remove a node and its dependencies from the graph
 * 
 * @details This method removes a node and all its dependencies from the graph. It performs a breadth-first
 * traversal starting from the specified node to remove all nodes that are dependent on it.
 * 
 * @param sz_nodeName The name of the node to remove
 */
ModScop void Graph::I3E_DAG_RemoveNodeAndDependencies(const std::string& sz_nodeName) 
{

    pthread_mutex_lock(&mtx_DAGMutex); 
    if (nodes.find(sz_nodeName) == nodes.end()) {
        // Node not found in the graph, no action needed
        return;
    }

    ModScop std::unordered_set<std::string> set_visited; // Set to track visited nodes during removal
    ModScop std::queue<std::string> queue; // Queue for BFS traversal to remove dependencies

    set_visited.insert(sz_nodeName); // Mark the initial node for removal as visited
    queue.push(sz_nodeName); // Add the initial node to the queue for processing

    // Continue processing nodes and their dependencies until the queue is empty
    while (!queue.empty()) {
        ModScop std::string sz_currentNode = queue.front(); // Get the front node from the queue
        queue.pop(); // Remove the front node from the queue

        if (nodes.find(sz_currentNode) != nodes.end()) {
            // Remove the current node from the graph
            nodes.erase(sz_currentNode);

            // Remove the current node from the adjacency list of its parent nodes
            for (auto& entry : adjacencyList) {
                entry.second.erase(sz_currentNode);
            }

            // Remove the current node from the adjacency list
            adjacencyList.erase(sz_currentNode);
        }

        // Explore neighbors of the current node for further removal
        if (nodes.find(sz_currentNode) != nodes.end() && adjacencyList.find(sz_currentNode) != adjacencyList.end()) {
            for (const std::string& neighbor : adjacencyList[sz_currentNode]) {
                if (set_visited.find(neighbor) == set_visited.end()) {
                    set_visited.insert(neighbor); // Mark the neighbor as visited
                    queue.push(neighbor); // Add the neighbor to the queue for removal
                }
            }
        }
    }
    pthread_mutex_unlock(&mtx_DAGMutex);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief Remove all children nodes from a specified node
 * 
 * @details This method removes all children nodes from a specified node in the graph. It recursively deletes
 * all children of the current node and then removes the current node from the graph.
 * 
 * @param sz_nodeName The name of the node to remove all children from
 */        
ModScop void Graph::I3E_DAG_DeleteAllChildren(const std::string& sz_nodeName)
{
    pthread_mutex_lock(&mtx_DAGMutex); 

    if (adjacencyList.find(sz_nodeName) != adjacencyList.end())
    {
        ModScop std::unordered_set<std::string> set_children = adjacencyList[sz_nodeName];
        for (const std::string& sz_child : set_children)
        {
            // Recursively delete children of the current child node
            I3E_DAG_DeleteAllChildren(sz_child);

            // Check if the child node still exists before deleting it
            if (nodes.find(sz_child) != nodes.end())
            {
                // Remove the child node from the graph
                nodes.erase(sz_child);

                // Remove the child node from the adjacency list of its parent
                for (auto& entry : adjacencyList)
                {
                    entry.second.erase(sz_child);
                }
            }
        }
        // Remove the children set from the adjacency list
        adjacencyList.erase(sz_nodeName);
    }
    pthread_mutex_unlock(&mtx_DAGMutex);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief Find the parent node of a given node
 * 
 * @details This method finds the parent node of a given node in the graph and returns its name. It searches
 * through the adjacency list to find the parent node of the specified node.
 * 
 * @param sz_nodeName The name of the node to find the parent of
 * 
 * @returns The name of the parent node of the specified node
 */
ModScop std::string Graph::I3E_DAG_FindParentNode(const std::string& sz_nodeName) 
{
    pthread_mutex_lock(&mtx_DAGMutex);
    for (const auto& entry : adjacencyList) 
    {
        if (entry.second.find(sz_nodeName) != entry.second.end()) 
        {
            pthread_mutex_unlock(&mtx_DAGMutex);
            return entry.first; // Found the parent node
        }
    }
    pthread_mutex_unlock(&mtx_DAGMutex);
    return ""; // Return empty string if parent node not found
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief Reset the updated flag of a node in the graph
 * 
 * @details This method resets the updated flag of a node in the graph with the specified node name. It sets
 * the updated flag to false for the specified node.
 * 
 * @param href The name of the node to reset the updated flag for
 */
ModScop void Graph::I3E_DAG_ResetUpdatedFlag(const std::string& href) 
{
    pthread_mutex_lock(&mtx_DAGMutex); // Acquire the mutex lock

    // Check if the node with the given href exists and update its updated flag
    if (nodes.find(href) != nodes.end()) {
        nodes[href].bol_updated = false; // Set the updated flag to false
    } else {
        LOG_ERR("Node with href %s not found.", href.c_str());
    }

    pthread_mutex_unlock(&mtx_DAGMutex); // Release the mutex lock
}