#include "include/I3E_XMLD.hpp"
#include "include/I3E_EVQ.hpp"
#include "lion/CST.h"

#include "lion/log/log.h"

LOG_MODULE_REGISTER(i3e_deh, LOG_LEVEL_INF);

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

/**
 * @brief DER Control Demultiplexer.
 * 
 * @details This function will scan through the DERControl event queue and determine which events have been flagged as 
 * active by the DEM. Then, based on the active events, the function will determine which DERControl functions to call.
 */
FileScop void i3e_deh_derControlDemux(void);

/**
 * @brief Default DER Control Demultiplexer.
 * 
 * @details This function will scan through the DERControl event queue and determine which default controls to activate
 * based on logic checks between the active DERControl events and the default controls.
 */
FileScop void i3e_deh_defaultDERControlDemux(void);

/**
 * @brief DER Curve Activation Function - opModFreqWatt
 * 
 * @details This function is responsible for activating the DER curve for the opModFreqWatt control.
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 */
FileScop void i3e_deh_opModFreqWatt(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, 
                                    bool bol_defCtrl); 

/**
 * @brief DER Curve Activation Function - opModHFRTMustTrip
 * 
 * @details This function is responsible for activating the DER curve for the opModHFRTMustTrip control.
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 */
FileScop void i3e_deh_opModHFRTMustTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, 
                                    bool bol_defCtrl); 

/**
 * @brief DER Curve Activation Function - opModHVRTMomentaryCessation
 * 
 * @details This function is responsible for activating the DER curve for the opModHVRTMomentaryCessation control.   
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 */
FileScop void i3e_deh_opModHVRTMomentaryCessation(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, 
                                    bool bol_defCtrl); 

/**
 * @brief DER Curve Activation Function - opModHVRTMustTrip
 * 
 * @details This function is responsible for activating the DER curve for the opModHVRTMustTrip control.       
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 */
FileScop void i3e_deh_opModHVRTMustTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, 
                                    bool bol_defCtrl); 

/**
 * @brief DER Curve Activation Function - opModLFRTMustTrip
 * 
 * @details This function is responsible for activating the DER curve for the opModLFRTMustTrip control.  
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 */
FileScop void i3e_deh_opModLFRTMustTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, 
                                    bool bol_defCtrl); 

/**
 * @brief DER Curve Activation Function - opModLVRTMomentaryCessation
 * 
 * @details This function is responsible for activating the DER curve for the opModLVRTMomentaryCessation control. 
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 */
FileScop void i3e_deh_opModLVRTMomentaryCessation(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, 
                                    bool bol_defCtrl); 

/**
 * @brief DER Curve Activation Function - opModLVRTMustTrip
 * 
 * @details This function is responsible for activating the DER curve for the opModLVRTMustTrip control. 
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 */
FileScop void i3e_deh_opModLVRTMustTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, 
                                    bool bol_defCtrl); 

/**
 * @brief DER Curve Activation Function - opModHFRTMayTrip
 * 
 * @details This function is responsible for activating the DER curve for the opModHFRTMayTrip control. 
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 */
FileScop void i3e_deh_opModHFRTMayTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, 
                                    bool bol_defCtrl); 

/**
 * @brief DER Curve Activation Function - opModHVRTMayTrip
 * 
 * @details This function is responsible for activating the DER curve for the opModHVRTMayTrip control. 
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 */
FileScop void i3e_deh_opModHVRTMayTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, 
                                    bool bol_defCtrl); 

/**
 * @brief DER Curve Activation Function - opModLFRTMayTrip
 * 
 * @details This function is responsible for activating the DER curve for the opModLFRTMayTrip control. 
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 */
FileScop void i3e_deh_opModLFRTMayTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, 
                                    bool bol_defCtrl); 

/**
 * @brief DER Curve Activation Function - opModLVRTMayTrip
 * 
 * @details This function is responsible for activating the DER curve for the opModLVRTMayTrip control.  
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 */
FileScop void i3e_deh_opModLVRTMayTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, 
                                    bool bol_defCtrl); 

/**
 * @brief DER Curve Activation Function - opModVoltVar
 * 
 * @details This function is responsible for activating the DER curve for the opModVoltVar control.  
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 */
FileScop void i3e_deh_opModVoltVar(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, 
                                    bool bol_defCtrl);  

/**
 * @brief DER Curve Activation Function - opModVoltWatt
 * 
 * @details This function is responsible for activating the DER curve for the opModVoltWatt control.  
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 */
FileScop void i3e_deh_opModVoltWatt(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, 
                                    bool bol_defCtrl); 

/**
 * @brief DER Curve Activation Function - opModWattPF
 * 
 * @details This function is responsible for activating the DER curve for the opModWattPF control.  
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 */
FileScop void i3e_deh_opModWattPF(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, 
                                    bool bol_defCtrl); 

/**
 * @brief DER Curve Activation Function - opModWattVar
 * 
 * @details This function is responsible for activating the DER curve for the opModWattVar control.  
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 */
FileScop void i3e_deh_opModWattVar(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, 
                                    bool bol_defCtrl); 

/**
 * @brief DER Control Activation Function - rampTms
 * 
 * @details This function is responsible for activating the rampTms control.
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_i3e_controlBase A new control base structure that is populated with only the controls that need to be set.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 * 
 * @returns A boolean indicating whether or not to write that control, based on logic checking within the function.
 */
FileScop bool i3e_deh_rampTms(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl); 

/**
 * @brief DER Control Activation Function - opModTargetW
 * 
 * @details This function is responsible for activating the opModTargetW control.
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_i3e_controlBase A new control base structure that is populated with only the controls that need to be set.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 * 
 * @returns A boolean indicating whether or not to write that control, based on logic checking within the function.
 */
FileScop bool i3e_deh_opModTargetW(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl); 

/**
 * @brief DER Control Activation Function - opModTargetVar
 * 
 * @details This function is responsible for activating the opModTargetVar control.
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_i3e_controlBase A new control base structure that is populated with only the controls that need to be set.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 * 
 * @returns A boolean indicating whether or not to write that control, based on logic checking within the function.
 */
FileScop bool i3e_deh_opModTargetVar(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl); 

/**
 * @brief DER Control Activation Function - opModMaxLimW
 * 
 * @details This function is responsible for activating the opModMaxLimW control.
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_i3e_controlBase A new control base structure that is populated with only the controls that need to be set.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 * 
 * @returns A boolean indicating whether or not to write that control, based on logic checking within the function.
 */
FileScop bool i3e_deh_opModMaxLimW(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl); 

/**
 * @brief DER Control Activation Function - opModFreqDroop
 * 
 * @details This function is responsible for activating the opModFreqDroop control.
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_i3e_controlBase A new control base structure that is populated with only the controls that need to be set.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 * 
 * @returns A boolean indicating whether or not to write that control, based on logic checking within the function.
 */
FileScop bool i3e_deh_opModFreqDroop(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl); 

/**
 * @brief DER Control Activation Function - opModFixedW
 * 
 * @details This function is responsible for activating the opModFixedW control.
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_i3e_controlBase A new control base structure that is populated with only the controls that need to be set.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 * 
 * @returns A boolean indicating whether or not to write that control, based on logic checking within the function.
 */
FileScop bool i3e_deh_opModFixedW(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl); 

/**
 * @brief DER Control Activation Function - opModFixedVar
 * 
 * @details This function is responsible for activating the opModFixedVar control.
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_i3e_controlBase A new control base structure that is populated with only the controls that need to be set.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 * 
 * @returns A boolean indicating whether or not to write that control, based on logic checking within the function.
 */
FileScop bool i3e_deh_opModFixedVar(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl); 

/**
 * @brief DER Control Activation Function - opModEnergize
 * 
 * @details This function is responsible for activating the opModEnergize control.
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_i3e_controlBase A new control base structure that is populated with only the controls that need to be set.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 * 
 * @returns A boolean indicating whether or not to write that control, based on logic checking within the function.
 */
FileScop bool i3e_deh_opModEnergize(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl); 

/**
 * @brief DER Control Activation Function - opModConnect
 * 
 * @details This function is responsible for activating the opModConnect control.
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_i3e_controlBase A new control base structure that is populated with only the controls that need to be set.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 * 
 * @returns A boolean indicating whether or not to write that control, based on logic checking within the function.
 */
FileScop bool i3e_deh_opModConnect(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl); 

/**
 * @brief DER Control Activation Function - opModFixedPFInjectW
 * 
 * @details This function is responsible for activating the opModFixedPFInjectW control.
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_i3e_controlBase A new control base structure that is populated with only the controls that need to be set.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 * 
 * @returns A boolean indicating whether or not to write that control, based on logic checking within the function.
 */
FileScop bool i3e_deh_opModFixedPFInjectW(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl); 

/**
 * @brief DER Control Activation Function - opModFixedPFAbsorbW
 * 
 * @details This function is responsible for activating the opModFixedPFAbsorbW control.
 * 
 * @param ptr_str_eventBlock The event block structure for the current control. Null if bol_defCtrl is true.
 * @param ptr_i3e_controlBase A new control base structure that is populated with only the controls that need to be set.
 * @param ptr_str_defBlock The defaul block structure for the current default control. Null if bol_defCtrl is false.
 * @param bol_defCtrl A boolean value indicating if the control is a default control.
 * 
 * @returns A boolean indicating whether or not to write that control, based on logic checking within the function.
 */
FileScop bool i3e_deh_opModFixedPFAbsorbW(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl); 

/**
 * @brief Binary to Hex String Conversion
 * 
 * @details This helper function will convert a binary MRID array to a hex string.
 * 
 * @param u8_byteArray Array of binary data to convert to hex string.
 * @param siz_length Length of the binary data array.
 * 
 * @returns A hex string representation of the binary data.
 */
FileScop std::string i3e_deh_binaryToHexString(const char* u8_byteArray, size_t siz_length);


/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

ModScop void I3E_DEH_Exec(void) 
{
    while(1) 
    {
        LOG_INF("Event Handler Thread Executing");

        // Run default controls 
        i3e_deh_derControlDemux();

        // Run default controls
        i3e_deh_defaultDERControlDemux();

        // Sleep before next queue check
        sleep(15);
    }
}

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop void i3e_deh_derControlDemux(void)
{

    StackScop int cst_err;

    // Scan through queue, check for which events are active
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop std::vector<std::string> vec_activeEventsHrefs = ptr_evq_dercEventQueue->I3E_EVQ_FindEventsWithActiveFlag();
    for(uint8_t i = 0; i < vec_activeEventsHrefs.size(); i++) 
    {
        StackScop EventBlock str_eventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent(vec_activeEventsHrefs.at(i));
        if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckEventFlag(str_eventBlock, EVQ_eventCancelled)) && 
            !(ptr_evq_dercEventQueue->I3E_EVQ_CheckEventFlag(str_eventBlock, EVQ_eventSuperseded)) && 
            !(ptr_evq_dercEventQueue->I3E_EVQ_CheckEventFlag(str_eventBlock, EVQ_eventStarted)))
        {
            
            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // opModFreqWatt control
            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModFreqWatt_exists)
            {
                i3e_deh_opModFreqWatt(&str_eventBlock, nullptr, false);
            }
            else{}

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // opModHFRTMustTrip control
            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModHFRTMustTrip_exists)
            {
                i3e_deh_opModHFRTMustTrip(&str_eventBlock, nullptr, false);
            }
            else{}

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // opModHVRTMomentaryCessation control
            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModHVRTMomentaryCessation_exists)
            {
                i3e_deh_opModHVRTMomentaryCessation(&str_eventBlock, nullptr, false);
            }
            else{}

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // opModHVRTMustTrip control
            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModHVRTMustTrip_exists)
            {
                i3e_deh_opModHVRTMustTrip(&str_eventBlock, nullptr, false);            
            }
            else{} 

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  
            // opModLFRTMustTrip control                 
            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModLFRTMustTrip_exists)
            {
                i3e_deh_opModLFRTMustTrip(&str_eventBlock, nullptr, false);        
            }
            else{}

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // opModLVRTMomentaryCessation control
            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModLVRTMomentaryCessation_exists)
            {
                i3e_deh_opModLVRTMomentaryCessation(&str_eventBlock, nullptr, false);
            }
            else{}

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // opModLVRTMustTrip control
            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModLVRTMustTrip_exists)
            {
                i3e_deh_opModLVRTMustTrip(&str_eventBlock, nullptr, false);
            }
            else{}

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // opModHFRTMayTrip control
            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModHFRTMayTrip_exists)
            {
                 i3e_deh_opModHFRTMayTrip(&str_eventBlock, nullptr, false);
            }
            else{}

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // opModHVRTMayTrip control
            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModHVRTMayTrip_exists)
            {
                i3e_deh_opModHVRTMayTrip(&str_eventBlock, nullptr, false);
            }
            else{}

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // opModLFRTMayTrip control
            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModLFRTMayTrip_exists)
            {
                i3e_deh_opModLFRTMayTrip(&str_eventBlock, nullptr, false);
            }
            else{}

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // opModLVRTMayTrip control
            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModLVRTMayTrip_exists)
            {
                i3e_deh_opModLVRTMayTrip(&str_eventBlock, nullptr, false);
            }
            else{}

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // opModVoltVar control
            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModVoltVar_exists)
            {
                i3e_deh_opModVoltVar(&str_eventBlock, nullptr, false);
            }
            else{}

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // opModVoltWatt control
            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModVoltWatt_exists)
            {
                i3e_deh_opModVoltWatt(&str_eventBlock, nullptr, false);
            }
            else{}

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // opModWattPF control
            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModWattPF_exists)
            {
                i3e_deh_opModWattPF(&str_eventBlock, nullptr, false);
            }
            else{}

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // opModWattVar control
            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModWattVar_exists)
            {
                i3e_deh_opModWattVar(&str_eventBlock, nullptr, false);
            }
            else{}
            
            // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

            if(str_eventBlock.i3e_derCtrl.DERControlBase._flags &  (   SE_rampTms_exists        | SE_opModTargetW_exists        
                                                                |  SE_opModTargetVar_exists | SE_opModMaxLimW_exists  
                                                                |  SE_opModFreqDroop_exists | SE_opModFixedW_exists  
                                                                |  SE_opModFixedVar_exists  | SE_opModEnergize_exists 
                                                                |  SE_opModConnect_exists   | SE_opModFixedPFInjectW_exists 
                                                                |  SE_opModFixedPFAbsorbW_exists))
            {

                StackScop SE_DERControlBase_t ptr_i3e_controlBase = {0};
                StackScop bool bol_writeControl = false;

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // rampTms control
                if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_rampTms_exists)
                {
                    bol_writeControl = i3e_deh_rampTms(&str_eventBlock, &ptr_i3e_controlBase, nullptr, false);
                }
                else{}

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModTargetW control
                if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModTargetW_exists)
                {
                    bol_writeControl = i3e_deh_opModTargetW(&str_eventBlock, &ptr_i3e_controlBase, nullptr, false);
                }
                else{}

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModTargetVar control
                if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModTargetVar_exists)
                {
                    bol_writeControl = i3e_deh_opModTargetVar(&str_eventBlock, &ptr_i3e_controlBase, nullptr, false);
                }
                else{}

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModMaxLimW control
                if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModMaxLimW_exists)
                {
                    bol_writeControl = i3e_deh_opModMaxLimW(&str_eventBlock, &ptr_i3e_controlBase, nullptr, false);
                }
                else{}

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModFreqDroop control
                if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModFreqDroop_exists)
                {
                    bol_writeControl = i3e_deh_opModFreqDroop(&str_eventBlock, &ptr_i3e_controlBase, nullptr, false);
                }
                else{}

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModFixedW control
                if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModFixedW_exists)
                {
                    bol_writeControl = i3e_deh_opModFixedW(&str_eventBlock, &ptr_i3e_controlBase, nullptr, false);
                }
                else{}

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModFixedVar control
                if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModFixedVar_exists)
                {
                    bol_writeControl = i3e_deh_opModFixedVar(&str_eventBlock, &ptr_i3e_controlBase, nullptr, false);
                }
                else{}

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModEnergize control
                if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModEnergize_exists)
                {
                    bol_writeControl = i3e_deh_opModEnergize(&str_eventBlock, &ptr_i3e_controlBase, nullptr, false);
                }
                else{}

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModConnect control
                if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModConnect_exists)
                {
                    bol_writeControl = i3e_deh_opModConnect(&str_eventBlock, &ptr_i3e_controlBase, nullptr, false);
                }
                else{}

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModFixedPFInjectW control
                if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModFixedPFInjectW_exists)
                {
                    bol_writeControl = i3e_deh_opModFixedPFInjectW(&str_eventBlock, &ptr_i3e_controlBase, nullptr, false);
                }
                else{}

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModFixedPFAbsorbW control
                if(str_eventBlock.i3e_derCtrl.DERControlBase._flags & SE_opModFixedPFAbsorbW_exists)
                {
                    bol_writeControl = i3e_deh_opModFixedPFAbsorbW(&str_eventBlock, &ptr_i3e_controlBase, nullptr, false);
                }
                else{}

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // Determined that there needs to be a control write to the inverters 
                if(bol_writeControl) 
                {
                    StackScop int cst_err = CST_SetControl(&(ptr_i3e_controlBase));
                    if(cst_err)
                    {
                        LOG_ERR("DER control write failed for %s - %d", str_eventBlock.i3e_derCtrl.href.c_str(), cst_err);
                    }
                    else{
                        LOG_INF("DER control write success for %s", str_eventBlock.i3e_derCtrl.href.c_str());
                    }

                    // CST_ResetControl(SE_opModEnergize_exists|SE_opModConnect_exists);
                }
            }

            // #TODO - This will set the event as "started" even when inverter writes failed
            ptr_evq_dercEventQueue->I3E_EVQ_SetEventFlags(str_eventBlock, EVQ_eventStarted);
            ptr_evq_dercEventQueue->I3E_EVQ_ReplaceEventBlockInQueue(str_eventBlock.i3e_derCtrl.href, str_eventBlock);
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// #TODO - these defaults may end up toggling, not sure if more than one default control will have these settings.
FileScop void i3e_deh_defaultDERControlDemux(void)
{

    CST_UtilModeStatus_t enm_utilModeStatus = CST_UtilCMDStatusGetter();
    if(enm_utilModeStatus == MODEL_AVAIL || enm_utilModeStatus == MODEL_ACTDER)
    {
        // Scan through queue, get all default hrefs
        StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
        StackScop std::vector<std::string> vec_defaultsQueuedHrefs = ptr_evq_dercEventQueue->I3E_EVQ_GetAllHrefsFromDefaultQueue();
        for(uint8_t i = 0; i < vec_defaultsQueuedHrefs.size(); i++) 
        {
            StackScop DefaultBlock str_defBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetDefaultBlockByHref(vec_defaultsQueuedHrefs.at(i));


            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

            // Determine if the control is a default control that isn't in the control base, if so then write those 
            if(str_defBlock.i3e_defDerCtrl._flags   &  (SE_setGradW_exists         | SE_setESDelay_exists   | SE_setESHighFreq_exists
                                                |   SE_setESHighVolt_exists    | SE_setESLowFreq_exists | SE_setESLowVolt_exists
                                                |   SE_setESRandomDelay_exists | SE_setSoftGradW_exists | SE_setESRampTms_exists)
                                                )
            {

                StackScop bool bol_writeControl = false;
                StackScop SE_DefaultDERControl_t i3e_defaultWrite = {0};

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // #TODO - these might need a flag set / clear / check logic like the other controls? they are however settings 
                // setGradW control
                if(str_defBlock.i3e_defDerCtrl._flags & SE_setGradW_exists)
                {
                    if(ptr_evq_dercEventQueue->I3E_EVQ_GetSettingsHrefAtIndex(SE_setGradW_hrefPos) != str_defBlock.i3e_defDerCtrl.href)
                    {
                        LOG_INF("DERC default control setGradW write for %s", str_defBlock.i3e_defDerCtrl.href.c_str());
                        i3e_defaultWrite.setGradW = str_defBlock.i3e_defDerCtrl.setGradW;
                        i3e_defaultWrite._flags |= SE_setGradW_exists;
                        std::string sz_mrid = i3e_deh_binaryToHexString((char *)str_defBlock.i3e_defDerCtrl.mRID, sizeof(str_defBlock.i3e_defDerCtrl.mRID));
                        ptr_evq_dercEventQueue->I3E_EVQ_AddSettingsHrefAtIndex(str_defBlock.i3e_defDerCtrl.href, sz_mrid, SE_setGradW_hrefPos);
                        ptr_evq_dercEventQueue->I3E_EVQ_SetSettings(SE_setGradW_exists);
                        bol_writeControl = true;
                    }
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // setESDelay control
                if(str_defBlock.i3e_defDerCtrl._flags & SE_setESDelay_exists)
                {
                    if(ptr_evq_dercEventQueue->I3E_EVQ_GetSettingsHrefAtIndex(SE_setESDelay_hrefPos) != str_defBlock.i3e_defDerCtrl.href)
                    {
                        LOG_INF("DERC default control setESDelay write for %s", str_defBlock.i3e_defDerCtrl.href.c_str()); 
                        i3e_defaultWrite.setESDelay = str_defBlock.i3e_defDerCtrl.setESDelay;
                        i3e_defaultWrite._flags |= SE_setESDelay_exists;
                        std::string sz_mrid = i3e_deh_binaryToHexString((char *)str_defBlock.i3e_defDerCtrl.mRID, sizeof(str_defBlock.i3e_defDerCtrl.mRID));
                        ptr_evq_dercEventQueue->I3E_EVQ_AddSettingsHrefAtIndex(str_defBlock.i3e_defDerCtrl.href, sz_mrid, SE_setESDelay_hrefPos);
                        ptr_evq_dercEventQueue->I3E_EVQ_SetSettings(SE_setESDelay_exists);
                        bol_writeControl = true;
                    }
                }
                else{}                        
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // setESHighFreq control
                if(str_defBlock.i3e_defDerCtrl._flags & SE_setESHighFreq_exists)
                {
                    if(ptr_evq_dercEventQueue->I3E_EVQ_GetSettingsHrefAtIndex(SE_setESHighFreq_hrefPos) != str_defBlock.i3e_defDerCtrl.href)
                    {
                        LOG_INF("DERC default control setESHighFreq write for %s", str_defBlock.i3e_defDerCtrl.href.c_str()); 
                        i3e_defaultWrite.setESHighFreq = str_defBlock.i3e_defDerCtrl.setESHighFreq;
                        i3e_defaultWrite._flags |= SE_setESHighFreq_exists;
                        std::string sz_mrid = i3e_deh_binaryToHexString((char *)str_defBlock.i3e_defDerCtrl.mRID, sizeof(str_defBlock.i3e_defDerCtrl.mRID));
                        ptr_evq_dercEventQueue->I3E_EVQ_AddSettingsHrefAtIndex(str_defBlock.i3e_defDerCtrl.href, sz_mrid, SE_setESHighFreq_hrefPos);
                        ptr_evq_dercEventQueue->I3E_EVQ_SetSettings(SE_setESHighFreq_exists);
                        bol_writeControl = true;
                    }
                }
                else{} 
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // setESHighVolt control
                if(str_defBlock.i3e_defDerCtrl._flags & SE_setESHighVolt_exists)
                {
                    if(ptr_evq_dercEventQueue->I3E_EVQ_GetSettingsHrefAtIndex(SE_setESHighVolt_hrefPos) != str_defBlock.i3e_defDerCtrl.href)
                    {
                        LOG_INF("DERC default control setESHighVolt write for %s", str_defBlock.i3e_defDerCtrl.href.c_str()); 
                        i3e_defaultWrite.setESHighVolt = str_defBlock.i3e_defDerCtrl.setESHighVolt;
                        i3e_defaultWrite._flags |= SE_setESHighVolt_exists;
                        std::string sz_mrid = i3e_deh_binaryToHexString((char *)str_defBlock.i3e_defDerCtrl.mRID, sizeof(str_defBlock.i3e_defDerCtrl.mRID));
                        ptr_evq_dercEventQueue->I3E_EVQ_AddSettingsHrefAtIndex(str_defBlock.i3e_defDerCtrl.href, sz_mrid, SE_setESHighVolt_hrefPos);
                        ptr_evq_dercEventQueue->I3E_EVQ_SetSettings(SE_setESHighVolt_exists);
                        bol_writeControl = true;
                    }
                }
                else{} 
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // setESLowFreq control
                if(str_defBlock.i3e_defDerCtrl._flags & SE_setESLowFreq_exists)
                {
                    if(ptr_evq_dercEventQueue->I3E_EVQ_GetSettingsHrefAtIndex(SE_setESLowFreq_hrefPos) != str_defBlock.i3e_defDerCtrl.href)
                    {
                        LOG_INF("DERC default control setESLowFreq write for %s", str_defBlock.i3e_defDerCtrl.href.c_str()); 
                        i3e_defaultWrite.setESLowFreq = str_defBlock.i3e_defDerCtrl.setESLowFreq;
                        i3e_defaultWrite._flags |= SE_setESLowFreq_exists;
                        std::string sz_mrid = i3e_deh_binaryToHexString((char *)str_defBlock.i3e_defDerCtrl.mRID, sizeof(str_defBlock.i3e_defDerCtrl.mRID));
                        ptr_evq_dercEventQueue->I3E_EVQ_AddSettingsHrefAtIndex(str_defBlock.i3e_defDerCtrl.href, sz_mrid, SE_setESLowFreq_hrefPos);
                        ptr_evq_dercEventQueue->I3E_EVQ_SetSettings(SE_setESLowFreq_exists);
                        bol_writeControl = true;
                    }
                }
                else{} 
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // setESLowVolt control
                if(str_defBlock.i3e_defDerCtrl._flags & SE_setESLowVolt_exists)
                {
                    if(ptr_evq_dercEventQueue->I3E_EVQ_GetSettingsHrefAtIndex(SE_setESLowVolt_hrefPos) != str_defBlock.i3e_defDerCtrl.href)
                    {
                        LOG_INF("DERC default control setESLowVolt write for %s", str_defBlock.i3e_defDerCtrl.href.c_str()); 
                        i3e_defaultWrite.setESLowVolt = str_defBlock.i3e_defDerCtrl.setESLowVolt;
                        i3e_defaultWrite._flags |= SE_setESLowVolt_exists; 
                        std::string sz_mrid = i3e_deh_binaryToHexString((char *)str_defBlock.i3e_defDerCtrl.mRID, sizeof(str_defBlock.i3e_defDerCtrl.mRID));
                        ptr_evq_dercEventQueue->I3E_EVQ_AddSettingsHrefAtIndex(str_defBlock.i3e_defDerCtrl.href, sz_mrid, SE_setESLowVolt_hrefPos);
                        ptr_evq_dercEventQueue->I3E_EVQ_SetSettings(SE_setESLowVolt_exists);
                        bol_writeControl = true;
                    }
                }
                else{} 
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // setESRandomDelay control
                if(str_defBlock.i3e_defDerCtrl._flags & SE_setESRandomDelay_exists)
                {
                    if(ptr_evq_dercEventQueue->I3E_EVQ_GetSettingsHrefAtIndex(SE_setESRandomDelay_hrefPos) != str_defBlock.i3e_defDerCtrl.href)
                    {
                        LOG_INF("DERC default control setESRandomDelay write for %s", str_defBlock.i3e_defDerCtrl.href.c_str());  
                        i3e_defaultWrite.setESRandomDelay = str_defBlock.i3e_defDerCtrl.setESRandomDelay;
                        i3e_defaultWrite._flags |= SE_setESRandomDelay_exists;
                        std::string sz_mrid = i3e_deh_binaryToHexString((char *)str_defBlock.i3e_defDerCtrl.mRID, sizeof(str_defBlock.i3e_defDerCtrl.mRID));
                        ptr_evq_dercEventQueue->I3E_EVQ_AddSettingsHrefAtIndex(str_defBlock.i3e_defDerCtrl.href, sz_mrid, SE_setESRandomDelay_hrefPos);
                        ptr_evq_dercEventQueue->I3E_EVQ_SetSettings(SE_setESRandomDelay_exists);
                        bol_writeControl = true;
                    }
                }
                else{} 
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // setSoftGradW control
                if(str_defBlock.i3e_defDerCtrl._flags & SE_setSoftGradW_exists)
                {
                    if(ptr_evq_dercEventQueue->I3E_EVQ_GetSettingsHrefAtIndex(SE_setSoftGradW_hrefPos) != str_defBlock.i3e_defDerCtrl.href)
                    {
                        LOG_INF("DERC default control setSoftGradW write for %s", str_defBlock.i3e_defDerCtrl.href.c_str()); 
                        i3e_defaultWrite.setSoftGradW = str_defBlock.i3e_defDerCtrl.setSoftGradW;
                        i3e_defaultWrite._flags |= SE_setSoftGradW_exists;
                        std::string sz_mrid = i3e_deh_binaryToHexString((char *)str_defBlock.i3e_defDerCtrl.mRID, sizeof(str_defBlock.i3e_defDerCtrl.mRID));
                        ptr_evq_dercEventQueue->I3E_EVQ_AddSettingsHrefAtIndex(str_defBlock.i3e_defDerCtrl.href, sz_mrid, SE_setSoftGradW_hrefPos);
                        ptr_evq_dercEventQueue->I3E_EVQ_SetSettings(SE_setSoftGradW_exists);
                        bol_writeControl = true;
                    }
                }
                else{} 
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // setESRampTms control
                if(str_defBlock.i3e_defDerCtrl._flags & SE_setESRampTms_exists)
                {
                    if(ptr_evq_dercEventQueue->I3E_EVQ_GetSettingsHrefAtIndex(SE_setESRampTms_hrefPos) != str_defBlock.i3e_defDerCtrl.href)
                    {
                        LOG_INF("DERC default control setESRampTms write for %s", str_defBlock.i3e_defDerCtrl.href.c_str()); 
                        i3e_defaultWrite.setESRampTms = str_defBlock.i3e_defDerCtrl.setESRampTms;
                        i3e_defaultWrite._flags |= SE_setESRampTms_exists;
                        std::string sz_mrid = i3e_deh_binaryToHexString((char *)str_defBlock.i3e_defDerCtrl.mRID, sizeof(str_defBlock.i3e_defDerCtrl.mRID));
                        ptr_evq_dercEventQueue->I3E_EVQ_AddSettingsHrefAtIndex(str_defBlock.i3e_defDerCtrl.href, sz_mrid, SE_setESRampTms_hrefPos);
                        ptr_evq_dercEventQueue->I3E_EVQ_SetSettings(SE_setESRampTms_exists);
                        bol_writeControl = true;
                    }
                }
                else{} 

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // Determined that there needs to be a control write to the inverters
                if(bol_writeControl)
                {
                    StackScop int cst_err = CST_SetDefaultControl(&i3e_defaultWrite);
                    if(cst_err)
                    {
                        LOG_ERR("DER default control write failed for %s - %d", str_defBlock.i3e_defDerCtrl.href.c_str(), cst_err);
                    }
                    else
                    {
                        LOG_INF("DERC control write success for %s", str_defBlock.i3e_defDerCtrl.href.c_str());
                    }
                }
            }      

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

            // Determine if the control is a curve, if so use the curve setting function in writing to inverter 
            if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & (  SE_opModFreqWatt_exists               | SE_opModHFRTMustTrip_exists  
                                                                | SE_opModHVRTMomentaryCessation_exists | SE_opModHVRTMustTrip_exists  
                                                                | SE_opModLFRTMustTrip_exists           | SE_opModLVRTMomentaryCessation_exists
                                                                | SE_opModLVRTMustTrip_exists           | SE_opModHFRTMayTrip_exists   
                                                                | SE_opModHVRTMayTrip_exists            | SE_opModLFRTMayTrip_exists   
                                                                | SE_opModLVRTMayTrip_exists            | SE_opModVoltVar_exists
                                                                | SE_opModVoltWatt_exists               | SE_opModWattPF_exists        
                                                                | SE_opModWattVar_exists)
                                                                )
            {   

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModFreqWatt control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModFreqWatt_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModFreqWatt_exists)) &&  
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModFreqWatt_exists)))
                    {
                        i3e_deh_opModFreqWatt(nullptr, &str_defBlock, true);
                    }
                    else{}
                }
                else{}

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModHFRTMustTrip control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModHFRTMustTrip_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModHFRTMustTrip_exists)) &&  
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModHFRTMustTrip_exists)))
                    {
                        i3e_deh_opModHFRTMustTrip(nullptr, &str_defBlock, true);
                    }
                    else{}
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModHVRTMomentaryCessation control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModHVRTMomentaryCessation_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModHVRTMomentaryCessation_exists)) && 
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModHVRTMomentaryCessation_exists)))
                    {
                        i3e_deh_opModHVRTMomentaryCessation(nullptr, &str_defBlock, true);
                    }
                    else{}
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModHVRTMustTrip control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModHVRTMustTrip_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModHVRTMustTrip_exists)) &&  
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModHVRTMustTrip_exists)))
                    {
                        i3e_deh_opModHVRTMustTrip(nullptr, &str_defBlock, true);
                    }
                    else{}
                }
                else{} 
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~   
                // opModLFRTMustTrip control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModLFRTMustTrip_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModLFRTMustTrip_exists)) &&  
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModLFRTMustTrip_exists)))
                    {
                        i3e_deh_opModLFRTMustTrip(nullptr, &str_defBlock, true);
                    }
                    else{}
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModLVRTMomentaryCessation control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModLVRTMomentaryCessation_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModLVRTMomentaryCessation_exists)) &&  
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModLVRTMomentaryCessation_exists)))
                    {
                        i3e_deh_opModLVRTMomentaryCessation(nullptr, &str_defBlock, true);
                    }
                    else{}
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModLVRTMustTrip control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModLVRTMustTrip_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModLVRTMustTrip_exists)) && 
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModLVRTMustTrip_exists)))
                    {
                        i3e_deh_opModLVRTMustTrip(nullptr, &str_defBlock, true);
                    }
                    else{}
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModHFRTMayTrip control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModHFRTMayTrip_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModHFRTMayTrip_exists)) &&  
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModHFRTMayTrip_exists)))
                    {
                        i3e_deh_opModHFRTMayTrip(nullptr, &str_defBlock, true);
                    }
                    else{}
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModHVRTMayTrip control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModHVRTMayTrip_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModHVRTMayTrip_exists)) && 
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModHVRTMayTrip_exists)))
                    {
                        i3e_deh_opModHVRTMayTrip(nullptr, &str_defBlock, true);
                    }
                    else{}
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModLFRTMayTrip control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModLFRTMayTrip_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModLFRTMayTrip_exists)) && 
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModLFRTMayTrip_exists)))
                    {
                        i3e_deh_opModLFRTMayTrip(nullptr, &str_defBlock, true);
                    }
                    else{}
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModLVRTMayTrip control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModLVRTMayTrip_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModLVRTMayTrip_exists)) && 
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModLVRTMayTrip_exists)))
                    {
                        i3e_deh_opModLVRTMayTrip(nullptr, &str_defBlock, true);
                    }
                    else{}
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModVoltVar control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModVoltVar_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModVoltVar_exists)) &&  
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModVoltVar_exists)))
                    {
                        i3e_deh_opModVoltVar(nullptr, &str_defBlock, true);
                    }
                    else{}
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModVoltWatt control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModVoltWatt_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModVoltWatt_exists)) && 
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModVoltWatt_exists)))
                    {
                        i3e_deh_opModVoltWatt(nullptr, &str_defBlock, true);
                    }
                    else{}
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModWattPF control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModWattPF_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModWattPF_exists)) &&  
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModWattPF_exists)))
                    {
                        i3e_deh_opModWattPF(nullptr, &str_defBlock, true);
                    }
                    else{}
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModWattVar control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModWattVar_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModWattVar_exists)) &&  
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModWattVar_exists)))
                    {
                        i3e_deh_opModWattVar(nullptr, &str_defBlock, true);
                    }
                    else{}
                }
                else{}
            }

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

            // Otherwise, it's a simple control within the control base. 
            if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & ( SE_opModTargetW_exists        | SE_opModTargetVar_exists     
                                                                | SE_opModMaxLimW_exists        | SE_opModFreqDroop_exists 
                                                                | SE_opModFixedW_exists         | SE_opModFixedVar_exists      
                                                                | SE_opModEnergize_exists       | SE_opModConnect_exists   
                                                                | SE_opModFixedPFInjectW_exists | SE_opModFixedPFAbsorbW_exists   
                                                                | SE_rampTms_exists)
                                                                )
            {   

                StackScop SE_DERControlBase_t ptr_i3e_controlBase = {0};
                StackScop bool bol_writeControl = false;

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // rampTms control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_rampTms_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_rampTms_exists)) &&  
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_rampTms_exists)))
                    {
                        bol_writeControl = i3e_deh_rampTms(nullptr, &ptr_i3e_controlBase, &str_defBlock, true);
                    }
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModTargetW control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModTargetW_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModTargetW_exists)) &&  
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModTargetW_exists)))
                    {
                        bol_writeControl = i3e_deh_opModTargetW(nullptr, &ptr_i3e_controlBase, &str_defBlock, true);
                    }
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModTargetVar control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModTargetVar_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModTargetVar_exists)) && 
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModTargetVar_exists)))
                    {
                        bol_writeControl = i3e_deh_opModTargetVar(nullptr, &ptr_i3e_controlBase, &str_defBlock, true);
                    }
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModMaxLimW control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModMaxLimW_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModMaxLimW_exists)) && 
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModMaxLimW_exists)))
                    {
                        bol_writeControl = i3e_deh_opModMaxLimW(nullptr, &ptr_i3e_controlBase, &str_defBlock, true);
                    }
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModFreqDroop control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModFreqDroop_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModFreqDroop_exists)) && 
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModFreqDroop_exists)))
                    {
                        bol_writeControl = i3e_deh_opModFreqDroop(nullptr, &ptr_i3e_controlBase, &str_defBlock, true);
                    }
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModFixedW control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModFixedW_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModFixedW_exists)) && 
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModFixedW_exists)))
                    {
                        bol_writeControl = i3e_deh_opModFixedW(nullptr, &ptr_i3e_controlBase, &str_defBlock, true);
                    }
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModFixedVar control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModFixedVar_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModFixedVar_exists)) &&
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModFixedVar_exists)))
                    {
                        bol_writeControl = i3e_deh_opModFixedVar(nullptr, &ptr_i3e_controlBase, &str_defBlock, true);
                    }
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModEnergize control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModEnergize_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModEnergize_exists)) &&  
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModEnergize_exists)))
                    {
                        bol_writeControl = i3e_deh_opModEnergize(nullptr, &ptr_i3e_controlBase, &str_defBlock, true);
                    }
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModConnect control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModConnect_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModConnect_exists)) &&  
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModConnect_exists)))
                    {
                        bol_writeControl = i3e_deh_opModConnect(nullptr, &ptr_i3e_controlBase, &str_defBlock, true);
                    } 
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModFixedPFInjectW control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModFixedPFInjectW_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModFixedPFInjectW_exists)) && 
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModFixedPFInjectW_exists)))
                    {                        
                        bol_writeControl = i3e_deh_opModFixedPFInjectW(nullptr, &ptr_i3e_controlBase, &str_defBlock, true);
                    }
                }
                else{}
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                // opModFixedPFAbsorbW control
                if(str_defBlock.i3e_defDerCtrl.DERControlBase._flags & SE_opModFixedPFAbsorbW_exists)
                {
                    if(!(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModFixedPFAbsorbW_exists)) &&  
                        !(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveDefault(SE_opModFixedPFAbsorbW_exists)))
                    {
                        bol_writeControl = i3e_deh_opModFixedPFAbsorbW(nullptr, &ptr_i3e_controlBase, &str_defBlock, true);
                    }
                }
                else{}

                // If there is a control to write, then write it
                if(bol_writeControl) 
                {
                    StackScop int cst_err = CST_SetControl(&(ptr_i3e_controlBase));
                    if(cst_err)
                    {
                        LOG_ERR("DER default control write failed for %s - %d", str_defBlock.i3e_defDerCtrl.href.c_str(), cst_err);
                    }
                    else{
                        LOG_INF("DER default control write success for %s", str_defBlock.i3e_defDerCtrl.href.c_str());
                    }
                } 
            }

        }

    }
    else
    {
        LOG_INF("DERC default control write skipped due to utility mode status");
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_deh_opModFreqWatt(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, bool bol_defCtrl) 
{
    StackScop int cst_err;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();
   
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModFreqWatt.href);    
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);

        // If the control parameter is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModFreqWatt_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModFreqWatt_hrefPos));
            // #TODO - these statements will prevent any control from taking over that is of the same primacy level.
            // #TODO - technically, the latest control (highest creation time) should take precedence, 
            // #TODO - but the QL test tool doesn't have any test cases fail currently... 
            // #TODO - if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModFreqWatt_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModFreqWatt_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModFreqWatt_hrefPos);
                cst_err = CST_SetCurve(&i3e_derCurveStored);
                if(cst_err)
                {
                    LOG_ERR("DER curve opModFreqWatt write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
                }
                else
                {
                    LOG_INF("DER curve opModFreqWatt write success for %s", ptr_str_eventBlock->sz_href.c_str());
                }
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModFreqWatt_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModFreqWatt_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModFreqWatt_hrefPos);
            cst_err = CST_SetCurve(&i3e_derCurveStored);
            if(cst_err)
            {
                LOG_ERR("DER curve opModFreqWatt write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
            }
            else
            {
                LOG_INF("DER curve opModFreqWatt write success for %s", ptr_str_eventBlock->sz_href.c_str());
            }
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModFreqWatt.href);    
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModFreqWatt_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModFreqWatt_hrefPos);
        cst_err = CST_SetCurve(&i3e_derCurveStored);
        if(cst_err)
        {
            LOG_ERR("DER default curve opModFreqWatt write failed for %s - %d", ptr_str_defBlock->i3e_defDerCtrl.href.c_str(), cst_err);
        }
        else
        {
            LOG_INF("DER default curve opModFreqWatt write success for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_deh_opModHFRTMustTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop int cst_err;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                            (ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModHFRTMustTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);

        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModHFRTMustTrip_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModHFRTMustTrip_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModHFRTMustTrip_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModHFRTMustTrip_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModHFRTMustTrip_hrefPos);
                cst_err = CST_SetCurve(&i3e_derCurveStored);
                if(cst_err)
                {
                    LOG_ERR("DER curve opModHFRTMustTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
                }
                else
                {
                    LOG_INF("DER curve opModHFRTMustTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
                }
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModHFRTMustTrip_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModHFRTMustTrip_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModHFRTMustTrip_hrefPos);
            cst_err = CST_SetCurve(&i3e_derCurveStored);
            if(cst_err)
            {
                LOG_ERR("DER curve opModHFRTMustTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
            }
            else
            {
                LOG_INF("DER curve opModHFRTMustTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
            }
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModHFRTMustTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModHFRTMustTrip_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModHFRTMustTrip_hrefPos);
        cst_err = CST_SetCurve(&i3e_derCurveStored);
        if(cst_err)
        {
            LOG_ERR("DER default curve opModHFRTMustTrip write failed for %s - %d", ptr_str_defBlock->i3e_defDerCtrl.href.c_str(), cst_err);
        }
        else
        {
            LOG_INF("DER default curve opModHFRTMustTrip write success for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_deh_opModHVRTMomentaryCessation(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop int cst_err; 
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModHVRTMomentaryCessation.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);

        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModHVRTMomentaryCessation_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModHVRTMomentaryCessation_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            { 
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModHVRTMomentaryCessation_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModHVRTMomentaryCessation_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModHVRTMomentaryCessation_hrefPos);
                cst_err = CST_SetCurve(&i3e_derCurveStored);
                if(cst_err)
                {
                    LOG_ERR("DER curve opModHVRTMomentaryCessation write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
                }
                else
                {
                    LOG_INF("DER curve opModHVRTMomentaryCessation write success for %s", ptr_str_eventBlock->sz_href.c_str());
                }
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModHVRTMomentaryCessation_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModHVRTMomentaryCessation_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModHVRTMomentaryCessation_hrefPos);
            cst_err = CST_SetCurve(&i3e_derCurveStored);
            if(cst_err)
            {
                LOG_ERR("DER curve opModHVRTMomentaryCessation write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
            }
            else
            {
                LOG_INF("DER curve opModHVRTMomentaryCessation write success for %s", ptr_str_eventBlock->sz_href.c_str());
            }
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    { 
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModHVRTMomentaryCessation.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModHVRTMomentaryCessation_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModHVRTMomentaryCessation_hrefPos);
        cst_err = CST_SetCurve(&i3e_derCurveStored);
        if(cst_err)
        {
            LOG_ERR("DER default curve opModHVRTMomentaryCessation write failed for %s - %d", ptr_str_defBlock->i3e_defDerCtrl.href.c_str(), cst_err);
        }
        else
        {
            LOG_INF("DER default curve opModHVRTMomentaryCessation write success for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_deh_opModHVRTMustTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop int cst_err;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModHVRTMustTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);

        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModHVRTMustTrip_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModHVRTMustTrip_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModHVRTMustTrip_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModHVRTMustTrip_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModHVRTMustTrip_hrefPos);
                cst_err = CST_SetCurve(&i3e_derCurveStored);
                if(cst_err)
                {
                    LOG_ERR("DER curve opModHVRTMustTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
                }
                else
                {
                    LOG_INF("DER curve opModHVRTMustTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
                }
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        { 
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModHVRTMustTrip_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModHVRTMustTrip_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModHVRTMustTrip_hrefPos);
            cst_err = CST_SetCurve(&i3e_derCurveStored);
            if(cst_err)
            {
                LOG_ERR("DER curve opModHVRTMustTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
            }
            else
            {
                LOG_INF("DER curve opModHVRTMustTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
            }
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModHVRTMustTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModHVRTMustTrip_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModHVRTMustTrip_hrefPos);
        cst_err = CST_SetCurve(&i3e_derCurveStored);
        if(cst_err)
        {
            LOG_ERR("DER default curve opModHVRTMustTrip write failed for %s - %d", ptr_str_defBlock->i3e_defDerCtrl.href.c_str(), cst_err);
        }
        else
        {
            LOG_INF("DER default curve opModHVRTMustTrip write success for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        }
    }
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_deh_opModLFRTMustTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop int cst_err;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModLFRTMustTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        
        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModLFRTMustTrip_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModLFRTMustTrip_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModLFRTMustTrip_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModLFRTMustTrip_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModLFRTMustTrip_hrefPos);
                cst_err = CST_SetCurve(&i3e_derCurveStored);
                if(cst_err)
                {
                    LOG_ERR("DER curve opModLFRTMustTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
                }
                else
                {
                    LOG_INF("DER curve opModLFRTMustTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
                }
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        { 
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModLFRTMustTrip_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModLFRTMustTrip_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModLFRTMustTrip_hrefPos);
            cst_err = CST_SetCurve(&i3e_derCurveStored);
            if(cst_err)
            {
                LOG_ERR("DER curve opModLFRTMustTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
            }
            else
            {
                LOG_INF("DER curve opModLFRTMustTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
            }
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModLFRTMustTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModLFRTMustTrip_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModLFRTMustTrip_hrefPos);
        cst_err = CST_SetCurve(&i3e_derCurveStored);
        if(cst_err)
        {
            LOG_ERR("DER default curve opModLFRTMustTrip write failed for %s - %d", ptr_str_defBlock->i3e_defDerCtrl.href.c_str(), cst_err);
        }
        else
        {
            LOG_INF("DER default curve opModLFRTMustTrip write success for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        }
    }
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_deh_opModLVRTMomentaryCessation(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop int cst_err;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModLVRTMomentaryCessation.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);

        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModLVRTMomentaryCessation_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModLVRTMomentaryCessation_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModLVRTMomentaryCessation_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModLVRTMomentaryCessation_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModLVRTMomentaryCessation_hrefPos);
                cst_err = CST_SetCurve(&i3e_derCurveStored);
                if(cst_err)
                {
                    LOG_ERR("DER curve opModLVRTMomentaryCessation write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
                }
                else
                {
                    LOG_INF("DER curve opModLVRTMomentaryCessation write success for %s", ptr_str_eventBlock->sz_href.c_str());
                }
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModLVRTMomentaryCessation_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModLVRTMomentaryCessation_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModLVRTMomentaryCessation_hrefPos);
            cst_err = CST_SetCurve(&i3e_derCurveStored);
            if(cst_err)
            {
                LOG_ERR("DER curve opModLVRTMomentaryCessation write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
            }
            else
            {
                LOG_INF("DER curve opModLVRTMomentaryCessation write success for %s", ptr_str_eventBlock->sz_href.c_str());
            }
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModLVRTMomentaryCessation.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModLVRTMomentaryCessation_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModLVRTMomentaryCessation_hrefPos);
        cst_err = CST_SetCurve(&i3e_derCurveStored);
        if(cst_err)
        {
            LOG_ERR("DER default curve opModLVRTMomentaryCessation write failed for %s - %d", ptr_str_defBlock->i3e_defDerCtrl.href.c_str(), cst_err);
        }
        else
        {
            LOG_INF("DER default curve opModLVRTMomentaryCessation write success for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        }
    }
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_deh_opModLVRTMustTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop int cst_err;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModLVRTMustTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);

        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModLVRTMustTrip_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModLVRTMustTrip_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModLVRTMustTrip_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModLVRTMustTrip_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModLVRTMustTrip_hrefPos);
                cst_err = CST_SetCurve(&i3e_derCurveStored);
                if(cst_err)
                {
                    LOG_ERR("DER curve opModLVRTMustTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
                }
                else
                {
                    LOG_INF("DER curve opModLVRTMustTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
                }
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {  
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModLVRTMustTrip_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModLVRTMustTrip_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModLVRTMustTrip_hrefPos);
            cst_err = CST_SetCurve(&i3e_derCurveStored);
            if(cst_err)
            {
                LOG_ERR("DER curve opModLVRTMustTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
            }
            else
            {
                LOG_INF("DER curve opModLVRTMustTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
            }
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    { 
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModLVRTMustTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModLVRTMustTrip_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModLVRTMustTrip_hrefPos);
        cst_err = CST_SetCurve(&i3e_derCurveStored);
        if(cst_err)
        {
            LOG_ERR("DER default curve opModLVRTMustTrip write failed for %s - %d", ptr_str_defBlock->i3e_defDerCtrl.href.c_str(), cst_err);
        }
        else
        {
            LOG_INF("DER default curve opModLVRTMustTrip write success for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_deh_opModHFRTMayTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop int cst_err;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModHFRTMayTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);

        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModHFRTMayTrip_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModHFRTMayTrip_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModHFRTMayTrip_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModHFRTMayTrip_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModHFRTMayTrip_hrefPos);
                cst_err = CST_SetCurve(&i3e_derCurveStored);
                if(cst_err)
                {
                    LOG_ERR("DER curve opModHFRTMayTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
                }
                else
                {
                    LOG_INF("DER curve opModHFRTMayTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
                }
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        { 
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModHFRTMayTrip_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModHFRTMayTrip_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModHFRTMayTrip_hrefPos);
            cst_err = CST_SetCurve(&i3e_derCurveStored);
            if(cst_err)
            {
                LOG_ERR("DER curve opModHFRTMayTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
            }
            else
            {
                LOG_INF("DER curve opModHFRTMayTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
            }
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModHFRTMayTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModHFRTMayTrip_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModHFRTMayTrip_hrefPos);
        cst_err = CST_SetCurve(&i3e_derCurveStored);
        if(cst_err)
        {
            LOG_ERR("DER default curve opModHFRTMayTrip write failed for %s - %d", ptr_str_defBlock->i3e_defDerCtrl.href.c_str(), cst_err);
        }
        else
        {
            LOG_INF("DER default curve opModHFRTMayTrip write success for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_deh_opModHVRTMayTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop int cst_err;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModHVRTMayTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);

        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModHVRTMayTrip_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModHVRTMayTrip_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            { 
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModHVRTMayTrip_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModHVRTMayTrip_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModHVRTMayTrip_hrefPos);
                cst_err = CST_SetCurve(&i3e_derCurveStored);
                if(cst_err)
                {
                    LOG_ERR("DER curve opModHVRTMayTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
                }
                else
                {
                    LOG_INF("DER curve opModHVRTMayTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
                }
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModHVRTMayTrip_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModHVRTMayTrip_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModHVRTMayTrip_hrefPos);
            cst_err = CST_SetCurve(&i3e_derCurveStored);
            if(cst_err)
            {
                LOG_ERR("DER curve opModHVRTMayTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
            }
            else
            {
                LOG_INF("DER curve opModHVRTMayTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
            }
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModHVRTMayTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModHVRTMayTrip_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModHVRTMayTrip_hrefPos);
        cst_err = CST_SetCurve(&i3e_derCurveStored);
        if(cst_err)
        {
            LOG_ERR("DER default curve opModHVRTMayTrip write failed for %s - %d", ptr_str_defBlock->i3e_defDerCtrl.href.c_str(), cst_err);
        }
        else
        {
            LOG_INF("DER default curve opModHVRTMayTrip write success for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_deh_opModLFRTMayTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop int cst_err;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModLFRTMayTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);

        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModLFRTMayTrip_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModLFRTMayTrip_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            { 
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModLFRTMayTrip_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModLFRTMayTrip_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModLFRTMayTrip_hrefPos);
                cst_err = CST_SetCurve(&i3e_derCurveStored);
                if(cst_err)
                {
                    LOG_ERR("DER curve opModLFRTMayTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
                }
                else
                {
                    LOG_INF("DER curve opModLFRTMayTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
                }
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModLFRTMayTrip_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModLFRTMayTrip_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModLFRTMayTrip_hrefPos);
            cst_err = CST_SetCurve(&i3e_derCurveStored);
            if(cst_err)
            {
                LOG_ERR("DER curve opModLFRTMayTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
            }
            else
            {
                LOG_INF("DER curve opModLFRTMayTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
            }
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModLFRTMayTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModLFRTMayTrip_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModLFRTMayTrip_hrefPos);
        cst_err = CST_SetCurve(&i3e_derCurveStored);
        if(cst_err)
        {
            LOG_ERR("DER default curve opModLFRTMayTrip write failed for %s - %d", ptr_str_defBlock->i3e_defDerCtrl.href.c_str(), cst_err);
        }
        else
        {
            LOG_INF("DER default curve opModLFRTMayTrip write success for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        }
    } 
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_deh_opModLVRTMayTrip(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, bool bol_defCtrl) 
{
    StackScop int cst_err;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default controlv
    if(!bol_defCtrl) 
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModLVRTMayTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);

        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModLVRTMayTrip_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModLVRTMayTrip_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModLVRTMayTrip_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModLVRTMayTrip_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModLVRTMayTrip_hrefPos);
                cst_err = CST_SetCurve(&i3e_derCurveStored);
                if(cst_err)
                {
                    LOG_ERR("DER curve opModLVRTMayTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
                }
                else
                {
                    LOG_INF("DER curve opModLVRTMayTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
                }
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        { 
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModLVRTMayTrip_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModLVRTMayTrip_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModLVRTMayTrip_hrefPos);
            cst_err = CST_SetCurve(&i3e_derCurveStored);
            if(cst_err)
            {
                LOG_ERR("DER curve opModLVRTMayTrip write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
            }
            else
            {
                LOG_INF("DER curve opModLVRTMayTrip write success for %s", ptr_str_eventBlock->sz_href.c_str());
            }
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModLVRTMayTrip.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModLVRTMayTrip_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModLVRTMayTrip_hrefPos);
        cst_err = CST_SetCurve(&i3e_derCurveStored);
        if(cst_err)
        {
            LOG_ERR("DER default curve opModLVRTMayTrip write failed for %s - %d", ptr_str_defBlock->i3e_defDerCtrl.href.c_str(), cst_err);
        }
        else
        {
            LOG_INF("DER default curve opModLVRTMayTrip write success for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        }
    } 
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_deh_opModVoltVar(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop int cst_err;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModVoltVar.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);

        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModVoltVar_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModVoltVar_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModVoltVar_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModVoltVar_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModVoltVar_hrefPos);
                cst_err = CST_SetCurve(&i3e_derCurveStored);
                if(cst_err)
                {
                    LOG_ERR("DER curve opModVoltVar write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
                }
                else
                {
                    LOG_INF("DER curve opModVoltVar write success for %s", ptr_str_eventBlock->sz_href.c_str());
                }
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModVoltVar_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModVoltVar_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModVoltVar_hrefPos);
            cst_err = CST_SetCurve(&i3e_derCurveStored);
            if(cst_err)
            {
                LOG_ERR("DER curve opModVoltVar write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
            }
            else
            {
                LOG_INF("DER curve opModVoltVar write success for %s", ptr_str_eventBlock->sz_href.c_str());
            }
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModVoltVar.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModVoltVar_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModVoltVar_hrefPos);
        cst_err = CST_SetCurve(&i3e_derCurveStored);
        if(cst_err)
        {
            LOG_ERR("DER default curve opModVoltVar write failed for %s - %d", ptr_str_defBlock->i3e_defDerCtrl.href.c_str(), cst_err);
        }
        else
        {
            LOG_INF("DER default curve opModVoltVar write success for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_deh_opModVoltWatt(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop int cst_err;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModVoltWatt.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);

        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModVoltWatt_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModVoltWatt_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModVoltWatt_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModVoltWatt_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModVoltWatt_hrefPos);
                cst_err = CST_SetCurve(&i3e_derCurveStored);
                if(cst_err)
                {
                    LOG_ERR("DER curve opModVoltWatt write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
                }
                else
                {
                    LOG_INF("DER curve opModVoltWatt write success for %s", ptr_str_eventBlock->sz_href.c_str());
                }
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModVoltWatt_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModVoltWatt_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModVoltWatt_hrefPos);
            cst_err = CST_SetCurve(&i3e_derCurveStored);
            if(cst_err)
            {
                LOG_ERR("DER curve opModVoltWatt write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
            }
            else
            {
                LOG_INF("DER curve opModVoltWatt write success for %s", ptr_str_eventBlock->sz_href.c_str());
            }
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModVoltWatt.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModVoltWatt_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModVoltWatt_hrefPos);
        cst_err = CST_SetCurve(&i3e_derCurveStored);
        if(cst_err)
        {
            LOG_ERR("DER default curve opModVoltWatt write failed for %s - %d", ptr_str_defBlock->i3e_defDerCtrl.href.c_str(), cst_err);
        }
        else
        {
            LOG_INF("DER default curve opModVoltWatt write success for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_deh_opModWattPF(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, bool bol_defCtrl) 
{
    StackScop int cst_err;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModWattPF.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);

        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModWattPF_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModWattPF_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            { 
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModWattPF_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModWattPF_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModWattPF_hrefPos);
                cst_err = CST_SetCurve(&i3e_derCurveStored);
                if(cst_err)
                {
                    LOG_ERR("DER curve opModWattPF write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
                }
                else
                {
                    LOG_INF("DER curve opModWattPF write success for %s", ptr_str_eventBlock->sz_href.c_str());
                }
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModWattPF_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModWattPF_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModWattPF_hrefPos);
            cst_err = CST_SetCurve(&i3e_derCurveStored);
            if(cst_err)
            {
                    LOG_ERR("DER curve opModWattPF write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
            }
            else
            {
                    LOG_INF("DER curve opModWattPF write success for %s", ptr_str_eventBlock->sz_href.c_str());
            }
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModWattPF.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModWattPF_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModWattPF_hrefPos);
        cst_err = CST_SetCurve(&i3e_derCurveStored);
        if(cst_err)
        {
            LOG_ERR("DER default curve opModWattPF write failed for %s - %d", ptr_str_defBlock->i3e_defDerCtrl.href.c_str(), cst_err);
        }
        else
        {
            LOG_INF("DER default curve opModWattPF write success for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_deh_opModWattVar(EventBlock *ptr_str_eventBlock, DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop int cst_err;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModWattVar.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);

        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModWattVar_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModWattVar_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModWattVar_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModWattVar_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModWattVar_hrefPos);
                cst_err = CST_SetCurve(&i3e_derCurveStored);
                if(cst_err)
                {
                    LOG_ERR("DER curve opModWattVar write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
                }
                else
                {
                    LOG_INF("DER curve opModWattVar write success for %s", ptr_str_eventBlock->sz_href.c_str());
                }
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModWattVar_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModWattVar_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModWattVar_hrefPos);
            cst_err = CST_SetCurve(&i3e_derCurveStored);
            if(cst_err)
            {
                    LOG_ERR("DER curve opModWattVar write failed for %s - %d", ptr_str_eventBlock->sz_href.c_str(), cst_err);
            }
            else
            {
                    LOG_INF("DER curve opModWattVar write success for %s", ptr_str_eventBlock->sz_href.c_str());
            }
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        StackScop NodeData str_curveNode = i3e_dagResources->I3E_DAG_GetNode
                        (ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModWattVar.href);
        StackScop SE_DERCurve_t& i3e_derCurveStored = std::get<SE_DERCurve_t>(str_curveNode.var_rsrc);
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModWattVar_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModWattVar_hrefPos);
        cst_err = CST_SetCurve(&i3e_derCurveStored);
        if(cst_err)
        {
            LOG_ERR("DER default curve opModWattVar write failed for %s - %d", ptr_str_defBlock->i3e_defDerCtrl.href.c_str(), cst_err);
        }
        else
        {
            LOG_INF("DER default curve opModWattVar write success for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop bool i3e_deh_rampTms(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                            DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop bool bol_write = false;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_rampTms_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_rampTms_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                LOG_INF("DER control rampTms write for %s", ptr_str_eventBlock->sz_href.c_str());
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_rampTms_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_rampTms_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_rampTms_hrefPos);
                ptr_i3e_controlBase->rampTms = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.rampTms;
                ptr_i3e_controlBase->_flags |= SE_rampTms_exists;
                bol_write = true;
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            LOG_INF("DER control rampTms write for %s", ptr_str_eventBlock->sz_href.c_str());
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_rampTms_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_rampTms_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_rampTms_hrefPos);
            ptr_i3e_controlBase->rampTms = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.rampTms;
            ptr_i3e_controlBase->_flags |= SE_rampTms_exists;
            bol_write = true;
        }
        
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        LOG_INF("DER default control rampTms write for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_rampTms_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_rampTms_hrefPos);
        ptr_i3e_controlBase->rampTms = ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.rampTms;
        ptr_i3e_controlBase->_flags |= SE_rampTms_exists;
        bol_write = true;
    }
    return bol_write;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop bool i3e_deh_opModTargetW(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase,
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop bool bol_write = false;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModTargetW_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModTargetW_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                LOG_INF("DER control opModTargetW write for %s", ptr_str_eventBlock->sz_href.c_str());
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModTargetW_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModTargetW_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModTargetW_hrefPos);
                ptr_i3e_controlBase->opModTargetW = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModTargetW;
                ptr_i3e_controlBase->_flags |= SE_opModTargetW_exists;
                bol_write = true;
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            LOG_INF("DER control opModTargetW write for %s", ptr_str_eventBlock->sz_href.c_str());
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModTargetW_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModTargetW_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModTargetW_hrefPos);
            ptr_i3e_controlBase->opModTargetW = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModTargetW;
            ptr_i3e_controlBase->_flags |= SE_opModTargetW_exists;
            bol_write = true;
        }  
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        LOG_INF("DER default control opModTargetW write for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModTargetW_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModTargetW_hrefPos);
        ptr_i3e_controlBase->opModTargetW = ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModTargetW;
        ptr_i3e_controlBase->_flags |= SE_opModTargetW_exists;
        bol_write = true;
    }
    return bol_write;                
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop bool i3e_deh_opModTargetVar(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop bool bol_write = false;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModTargetVar_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModTargetVar_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                LOG_INF("DER control opModTargetVar write for %s", ptr_str_eventBlock->sz_href.c_str());
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModTargetVar_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModTargetVar_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModTargetVar_hrefPos);
                ptr_i3e_controlBase->opModTargetVar = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModTargetVar;
                ptr_i3e_controlBase->_flags |= SE_opModTargetVar_exists;
                bol_write = true;
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            LOG_INF("DER control opModTargetVar write for %s", ptr_str_eventBlock->sz_href.c_str()); 
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModTargetVar_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModTargetVar_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModTargetVar_hrefPos);
            ptr_i3e_controlBase->opModTargetVar = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModTargetVar;
            ptr_i3e_controlBase->_flags |= SE_opModTargetVar_exists;
            bol_write = true;
        }   
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        LOG_INF("DER default control opModTargetVar write for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModTargetVar_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModTargetVar_hrefPos);
        ptr_i3e_controlBase->opModTargetVar = ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModTargetVar;
        ptr_i3e_controlBase->_flags |= SE_opModTargetVar_exists;
        bol_write = true;
    }
    return bol_write;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop bool i3e_deh_opModMaxLimW(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop bool bol_write = false;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();
    
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModMaxLimW_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModMaxLimW_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                LOG_INF("DER control opModMaxLimW write for %s", ptr_str_eventBlock->sz_href.c_str());
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModMaxLimW_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModMaxLimW_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModMaxLimW_hrefPos);
                ptr_i3e_controlBase->opModMaxLimW = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModMaxLimW;
                ptr_i3e_controlBase->_flags |= SE_opModMaxLimW_exists;
                bol_write = true;
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            LOG_INF("DER control opModMaxLimW write for %s", ptr_str_eventBlock->sz_href.c_str());
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModMaxLimW_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModMaxLimW_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModMaxLimW_hrefPos);
            ptr_i3e_controlBase->opModMaxLimW = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModMaxLimW;
            ptr_i3e_controlBase->_flags |= SE_opModMaxLimW_exists;
            bol_write = true;
        } 
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        LOG_INF("DER default control opModMaxLimW write for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModMaxLimW_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModMaxLimW_hrefPos);
        ptr_i3e_controlBase->opModMaxLimW = ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModMaxLimW;
        ptr_i3e_controlBase->_flags |= SE_opModMaxLimW_exists;
        bol_write = true;
    }
    return bol_write;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop bool i3e_deh_opModFreqDroop(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop bool bol_write = false;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModFreqDroop_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModFreqDroop_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                LOG_INF("DER control opModFreqDroop write for %s", ptr_str_eventBlock->sz_href.c_str());
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModFreqDroop_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModFreqDroop_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModFreqDroop_hrefPos);
                ptr_i3e_controlBase->opModFreqDroop = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModFreqDroop;
                ptr_i3e_controlBase->_flags |= SE_opModFreqDroop_exists;
                bol_write = true;
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            LOG_INF("DER control opModFreqDroop write for %s", ptr_str_eventBlock->sz_href.c_str());
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModFreqDroop_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModFreqDroop_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModFreqDroop_hrefPos);
            ptr_i3e_controlBase->opModFreqDroop = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModFreqDroop;
            ptr_i3e_controlBase->_flags |= SE_opModFreqDroop_exists;
            bol_write = true;
        }  
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        LOG_INF("DER default control opModFreqDroop write for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModFreqDroop_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModFreqDroop_hrefPos);
        ptr_i3e_controlBase->opModFreqDroop = ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModFreqDroop;
        ptr_i3e_controlBase->_flags |= SE_opModFreqDroop_exists;
        bol_write = true;
    }
    return bol_write;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop bool i3e_deh_opModFixedW(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop bool bol_write = false;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModFixedW_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModFixedW_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                LOG_INF("DER control opModFixedW write for %s", ptr_str_eventBlock->sz_href.c_str());
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModFixedW_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModFixedW_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModFixedW_hrefPos);
                ptr_i3e_controlBase->opModFixedW = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModFixedW;
                ptr_i3e_controlBase->_flags |= SE_opModFixedW_exists;
                bol_write = true;
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            LOG_INF("DER control opModFixedW write for %s", ptr_str_eventBlock->sz_href.c_str());
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModFixedW_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModFixedW_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModFixedW_hrefPos);
            ptr_i3e_controlBase->opModFixedW = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModFixedW;
            ptr_i3e_controlBase->_flags |= SE_opModFixedW_exists;
            bol_write = true;
        }   
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        LOG_INF("DER default control opModFixedW write for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModFixedW_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModFixedW_hrefPos);
        ptr_i3e_controlBase->opModFixedW = ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModFixedW;
        ptr_i3e_controlBase->_flags |= SE_opModFixedW_exists;
        bol_write = true;
    }
    return bol_write;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop bool i3e_deh_opModFixedVar(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop bool bol_write = false;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModFixedVar_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModFixedVar_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                LOG_INF("DER control opModFixedVar write for %s", ptr_str_eventBlock->sz_href.c_str()); 
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModFixedVar_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModFixedVar_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModFixedVar_hrefPos);
                ptr_i3e_controlBase->opModFixedVar = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModFixedVar;
                ptr_i3e_controlBase->_flags |= SE_opModFixedVar_exists;
                bol_write = true;
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            LOG_INF("DER control opModFixedVar write for %s", ptr_str_eventBlock->sz_href.c_str());
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModFixedVar_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModFixedVar_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModFixedVar_hrefPos);
            ptr_i3e_controlBase->opModFixedVar = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModFixedVar;
            ptr_i3e_controlBase->_flags |= SE_opModFixedVar_exists;
            bol_write = true;
        }  
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        LOG_INF("DER default control opModFixedVar write for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModFixedVar_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModFixedVar_hrefPos);
        ptr_i3e_controlBase->opModFixedVar = ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModFixedVar;
        ptr_i3e_controlBase->_flags |= SE_opModFixedVar_exists;
        bol_write = true;
    }
    return bol_write;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop bool i3e_deh_opModEnergize(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop bool bol_write = false;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModEnergize_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModEnergize_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                LOG_INF("DER control opModEnergize write for %s", ptr_str_eventBlock->sz_href.c_str());
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModEnergize_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModEnergize_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModEnergize_hrefPos);
                ptr_i3e_controlBase->opModEnergize = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModEnergize;
                ptr_i3e_controlBase->_flags |= SE_opModEnergize_exists;
                bol_write = true;
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            LOG_INF("DER control opModEnergize write for %s", ptr_str_eventBlock->sz_href.c_str());
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModEnergize_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModEnergize_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModEnergize_hrefPos);
            ptr_i3e_controlBase->opModEnergize = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModEnergize;
            ptr_i3e_controlBase->_flags |= SE_opModEnergize_exists;
            bol_write = true;
        }   
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        LOG_INF("DER default control opModEnergize write for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModEnergize_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModEnergize_hrefPos);
        ptr_i3e_controlBase->opModEnergize = ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModEnergize;
        ptr_i3e_controlBase->_flags |= SE_opModEnergize_exists;
        bol_write = true;
    }
    return bol_write;   
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop bool i3e_deh_opModConnect(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop bool bol_write = false;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModConnect_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModConnect_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                LOG_INF("DER control opModConnect write for %s", ptr_str_eventBlock->sz_href.c_str());
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModConnect_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModConnect_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModConnect_hrefPos);
                ptr_i3e_controlBase->opModConnect = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModConnect;
                ptr_i3e_controlBase->_flags |= SE_opModConnect_exists;
                bol_write = true;
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            LOG_INF("DER control opModConnect write for %s", ptr_str_eventBlock->sz_href.c_str());
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModConnect_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModConnect_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModConnect_hrefPos);
            ptr_i3e_controlBase->opModConnect = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModConnect;
            ptr_i3e_controlBase->_flags |= SE_opModConnect_exists;
            bol_write = true;
        }  
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        LOG_INF("DER default control opModConnect write for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModConnect_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModConnect_hrefPos);
        ptr_i3e_controlBase->opModConnect = ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModConnect;
        ptr_i3e_controlBase->_flags |= SE_opModConnect_exists;
        bol_write = true;
    }
    return bol_write;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop bool i3e_deh_opModFixedPFInjectW(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl) 
{
    StackScop bool bol_write = false;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {

        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModFixedPFInjectW_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModFixedPFInjectW_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                LOG_INF("DER control opModFixedPFInjectW write for %s", ptr_str_eventBlock->sz_href.c_str());
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModFixedPFInjectW_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModFixedPFInjectW_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModFixedPFInjectW_hrefPos);
                ptr_i3e_controlBase->opModFixedPFInjectW = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModFixedPFInjectW;
                ptr_i3e_controlBase->_flags |= SE_opModFixedPFInjectW_exists;
                bol_write = true;
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            LOG_INF("DER control opModFixedPFInjectW write for %s", ptr_str_eventBlock->sz_href.c_str());
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModFixedPFInjectW_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModFixedPFInjectW_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModFixedPFInjectW_hrefPos);
            ptr_i3e_controlBase->opModFixedPFInjectW = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModFixedPFInjectW;
            ptr_i3e_controlBase->_flags |= SE_opModFixedPFInjectW_exists;
            bol_write = true;
        } 
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        LOG_INF("DER default control opModFixedPFInjectW write for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModFixedPFInjectW_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModFixedPFInjectW_hrefPos);
        ptr_i3e_controlBase->opModFixedPFInjectW = ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModFixedPFInjectW;
        ptr_i3e_controlBase->_flags |= SE_opModFixedPFInjectW_exists;
        bol_write = true;
    }
    return bol_write;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop bool i3e_deh_opModFixedPFAbsorbW(EventBlock *ptr_str_eventBlock, SE_DERControlBase_t *ptr_i3e_controlBase, 
                                DefaultBlock *ptr_str_defBlock, bool bol_defCtrl)
{
    StackScop bool bol_write = false;
    StackScop EventQueue* ptr_evq_dercEventQueue = I3E_DEM_GetDERCEventQueue();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is not a default control
    if(!bol_defCtrl) 
    {
        // If the control is already active - primacy checking 
        if(ptr_evq_dercEventQueue->I3E_EVQ_CheckActiveControl(SE_opModFixedPFAbsorbW_exists))
        {
            StackScop EventBlock str_activeEventBlock = ptr_evq_dercEventQueue->I3E_EVQ_GetSpecificEvent
                            (ptr_evq_dercEventQueue->I3E_EVQ_GetHrefAtIndex(SE_opModFixedPFAbsorbW_hrefPos));
            if(ptr_str_eventBlock->primacy < str_activeEventBlock.primacy) 
            {
                LOG_INF("DER control opModFixedPFAbsorbW write for %s", ptr_str_eventBlock->sz_href.c_str());
                ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModFixedPFAbsorbW_exists);
                ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModFixedPFAbsorbW_exists);
                std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
                ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModFixedPFAbsorbW_hrefPos);
                ptr_i3e_controlBase->opModFixedPFAbsorbW = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModFixedPFAbsorbW;
                ptr_i3e_controlBase->_flags |= SE_opModFixedPFAbsorbW_exists;
                bol_write = true;
            }
            else {}
        }
        // If the control isn't active - simply set it 
        else
        {
            LOG_INF("DER control opModFixedPFAbsorbW write for %s", ptr_str_eventBlock->sz_href.c_str());
            ptr_evq_dercEventQueue->I3E_EVQ_SetActiveControl(SE_opModFixedPFAbsorbW_exists);
            ptr_evq_dercEventQueue->I3E_EVQ_ClearActiveDefault(SE_opModFixedPFAbsorbW_exists);
            std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_eventBlock->i3e_derCtrl.mRID, sizeof(ptr_str_eventBlock->i3e_derCtrl.mRID));
            ptr_evq_dercEventQueue->I3E_EVQ_AddHrefAtIndex(ptr_str_eventBlock->sz_href, sz_mrid, SE_opModFixedPFAbsorbW_hrefPos);
            ptr_i3e_controlBase->opModFixedPFAbsorbW = ptr_str_eventBlock->i3e_derCtrl.DERControlBase.opModFixedPFAbsorbW;
            ptr_i3e_controlBase->_flags |= SE_opModFixedPFAbsorbW_exists;
            bol_write = true;
        }  
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // If the control passed in is a default control
    else
    {
        LOG_INF("DER default control opModFixedPFAbsorbW write for %s", ptr_str_defBlock->i3e_defDerCtrl.href.c_str());
        ptr_evq_dercEventQueue->I3E_EVQ_SetActiveDefault(SE_opModFixedPFAbsorbW_exists);
        std::string sz_mrid = i3e_deh_binaryToHexString((char *)ptr_str_defBlock->i3e_defDerCtrl.mRID, sizeof(ptr_str_defBlock->i3e_defDerCtrl.mRID));
        ptr_evq_dercEventQueue->I3E_EVQ_AddDefaultHrefAtIndex(ptr_str_defBlock->sz_href, sz_mrid, SE_opModFixedPFAbsorbW_hrefPos);
        ptr_i3e_controlBase->opModFixedPFAbsorbW = ptr_str_defBlock->i3e_defDerCtrl.DERControlBase.opModFixedPFAbsorbW;
        ptr_i3e_controlBase->_flags |= SE_opModFixedPFAbsorbW_exists;
        bol_write = true;
    }
    return bol_write;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop std::string i3e_deh_binaryToHexString(const char* u8_byteArray, size_t siz_length) {
    
    StackScop std::ostringstream os_hexStream;

    for (size_t i = 0; i < siz_length; ++i) {
        // Convert each byte to a two-character hexadecimal string
        os_hexStream << std::hex << std::setw(2) << std::setfill('0') << static_cast<int>(u8_byteArray[i]);
    }
    return os_hexStream.str();
}