#include "include/I3E_XMLD.hpp"
#include "include/I3E_XMLS.hpp"
#include "lion/I3E.h"
#include "include/I3E_EVQ.hpp"
#include "include/I3E_CTS.hpp"
#include "include/I3E_SPP.hpp"
#include "include/I3E_NSS.hpp"
#include "lion/CST.h"

#include "lion/log/log.h"

LOG_MODULE_REGISTER(i3e_dem, LOG_LEVEL_INF);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define EVENTMASK ~SE_rampTms_exists

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/


/**
 * @brief Queue Monitoring Function
 * 
 * @details This function is responsible for the monitoring & maintenence of the DER control event queue. The events are
 * scheduled to run, checked for completion, checked for cancellation, and trimmed from the queue if they are no longer
 * needed. This function is called whenever semaphore times out in between semaphore posts from the server poll processor.
 * 
 * @param i3_serverTime The current server time
 * @param u8_loopcnt The loop count for the queue monitoring process (a few processes only need to run occasionally)
 */
FileScop void i3e_dem_queueMonitoring(SE_TimeType_t i3_serverTime, uint8_t u8_loopcnt);

/**
 * @brief Check Completed Events Function
 * 
 * @details This function checks the DER control event queue for events that have completed. If the event has expired in
 * duration, then the event is marked as completed and a message is sent to the server acknowledging the completion of
 * the event. This function also sets some flag parameters for the event to control the event lifecycle.
 * 
 * @param i3_serverTime The current server time
 */
FileScop void i3e_dem_queueCheckCompletedEvents(SE_TimeType_t i3_serverTime);

/**
 * @brief Queue Pre-Check Events Function
 * 
 * @details This function will check the DER control event queue for events that are starting in less than 60 seconds.
 * The event is checked via a HTTP GET method to the resource hosting the event. Upon checking the event, 
 * the function will determine if the event is still scheduled to run or has been cancelled. This function also sets 
 * some flag parameters for the event to control the event lifecycle.
 * 
 * @param i3_serverTime The current server time
 * @param sz_rxPayload The string to use for the HTTP GET response payload
 */
FileScop void i3e_dem_queuePreCheckEvents(SE_TimeType_t i3_serverTime, std::string sz_rxPayload);

/**
 * @brief Check Starting Events Function
 * 
 * @details This function will check the DER control event queue for events that are scheduled to start. The event is
 * checked against the current server time and if the event is scheduled to start, the event is marked as active and
 * a message is sent to the server that we have started this event. This function also sets some flag parameters for the
 * event to control the event lifecycle.
 * 
 * @param i3_serverTime The current server time
 */
FileScop void i3e_dem_queueCheckStartEvents(SE_TimeType_t i3_serverTime);

/**
 * @brief Check Active Events Function
 * 
 * @details This function will periodically check the DER control event queue for events that are currently active. If 
 * the event status has been changed by the server, the event is marked as cancelled and removed from the queue. This
 * function also sets some flag parameters for the event to control the event lifecycle.
 * 
 * @param i3_serverTime The current server time
 * @param u8_loopcnt The loop count for the queue monitoring process (a few processes only need to run occasionally)
 * @param sz_rxPayload The string to use for the HTTP GET response payload
 */
FileScop void i3e_dem_queueCheckActiveEvents(SE_TimeType_t i3_serverTime, uint8_t u8_loopcnt, std::string sz_rxPayload);

/**
 * @brief Check Cancelled Events Function
 * 
 * @details This function will periodically check the DER control event queue for events that have been cancelled by the
 * server. Technically, this is also done by i3e_dem_queueCheckActiveEvents, but the difference here is that sometimes
 * cancellations are not found by the application while the event is active, but rather the DAG will poll a control 
 * that indicates a cancellation and without this function there can be a case where the event cancellation message 
 * to the server is missed. 
 * 
 * @param i3_serverTime The current server time
 * @param u8_loopcnt The loop count for the queue monitoring process (a few processes only need to run occasionally)
 */
FileScop void i3e_dem_queueCheckCancelledEvents(SE_TimeType_t i3_serverTime, uint8_t u8_loopcnt);

/**
 * @brief Queue Event Trimmer
 * 
 * @details This function will clear out old DER Control events from the queue that are no longer needed. This function
 * is called occasionally to keep the queue size manageable and to prevent the queue from growing too large.
 * 
 * @param i3_serverTime The current server time
 */
FileScop void i3e_dem_queueEventTrimmer(SE_TimeType_t i3_serverTime);

/**
 * @brief Queue Scheuling Function
 * 
 * @details This function is responsible for the scheduling of DER Control events. Existing DER Control events and 
 * default controls are pulled in from the directed acyclic graph (DAG) and processed according to the IEEE 20330.5 
 * event rules. The trigger for this function to run is whenever the server poll processor (SPP) module posts a 
 * semaphore to the event manager indicating there is new information available in the DAG.
 * 
 * @param i3_serverTime The current server time
 */
FileScop void i3e_dem_queueScheduling(SE_TimeType_t i3_serverTime);

/**
 * @brief Update Events Function
 * 
 * @details This function will update existing events in the queue if there is a status change in the stored resource 
 * pulled in from the DAG. These changes need to be reflected in the event queue to keep the DAG events and the 
 * event queue events in sync. 
 * 
 * @param vec_derCtrlHrefs The vector of DER Control hrefs that are outdated. 
 */
FileScop void i3e_dem_updateEvents(std::vector<std::string> vec_derCtrlHrefs);

/**
 * @brief Update Default Events Function
 * 
 * @details This function will update existing default events in the queue if there is a status change in the stored 
 * resource pulled in from the DAG. These changes need to be reflected in the event queue to keep the DAG events and the 
 * event queue events in sync. 
 * 
 * @param vec_newCtrlHrefs The vector of new DER Default Control hrefs to process
 */
FileScop void i3e_dem_updateDefaults(std::vector<std::string> vec_defCtrlHrefs);

/**
 * @brief New Event Processing Function
 * 
 * @details This function will process new events that are found in the DAG but not in the event queue. These new events
 * are added to the event queue in start time sorted order and the event flags are set accordingly. This function performs
 * logical checks on the various events to ensure that only specific events are executed according to the IEEE 2030.5 
 * specification. 
 * 
 * @param vec_newCtrlHrefs The vector of new DER Control hrefs to process
 * @param i3_serverTime The current server time
 */
FileScop void i3e_dem_newEvents(std::vector<std::string> vec_newCtrlHrefs, SE_TimeType_t i3_serverTime);

/**
 * @brief New Default Processing Function 
 * 
 * @details This function will process new default events that are found in the DAG but not in the event queue. These 
 * default events are added to the event queue in primacy sorted order, so that the lowest primacy default events are 
 * first, then the next lowest, and so on. This allows the event handler to turn on a default control event if there is
 * no conflicting DERControl event and no lower primacy default control event is active.
 * 
 * @param vec_newDefCtrlHrefs The vector of new DER Default Control hrefs to process
 */
FileScop void i3e_dem_newDefaults(std::vector<std::string> vec_newDefCtrlHrefs);

/**
 * @brief Server Control Response Function
 * 
 * @details This function will send a response to the IEEE 2030.5 server regarding the status of the DER Control event.
 * It will package information into a DERControlResponse structure and send it to the server via a HTTP POST method.
 * 
 * @param i3_serverTime The current server time
 * @param str_currentEvent The current event block which contains the DER Control event information
 * @param u8_statusCode The status code to send to the server regarding the event status 
 */
FileScop void i3e_dem_controlResponse(SE_TimeType_t i3_serverTime, EventBlock str_currentEvent, uint8_t u8_statusCode);

/**
 * @brief Event Time Range Randomization Function
 * 
 * @details This function will perform event timing randomization according to the IEEE 2030.5 specification. The event
 * start time and duration are randomized within a specified range to prevent all events from starting at the same time.
 * 
 * @param i3e_hourRange The event time range to use for randomization.
 * 
 * @return The value to use for the randomized event time. 
 */
FileScop SE_OneHourRangeType_t i3e_dem_eventRandomization(SE_OneHourRangeType_t i3e_hourRange);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop EventQueue dercEventQueue;
FileScop sem_t eventSemaphore;

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop int I3E_DEM_Init(void) 
{
    StackScop int res = sem_init(&eventSemaphore, 0, 0);
    return res;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void I3E_DEM_PostSemaphore(void)
{
    sem_post(&eventSemaphore);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop EventQueue* I3E_DEM_GetDERCEventQueue(void)
{
    return &dercEventQueue;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void I3E_DEM_Exec(void) {

    StackScop struct timespec str_timeout;
    StackScop int result;
    StackScop uint8_t u8_loopcnt = 0;

    // Infinite DEM thread loop
    while(1) 
    {
        LOG_INF("DER Event Manager Thread Executing");
        
        clock_gettime(CLOCK_REALTIME, &str_timeout);
        str_timeout.tv_sec += 5; // Set str_timeout to 5 seconds
        result = sem_timedwait(&eventSemaphore, &str_timeout);

        StackScop SE_TimeType_t i3e_currentTime = I3E_CTS_getSyncTime();

        if(u8_loopcnt == 5)
        {
            u8_loopcnt = 0;
        }
        u8_loopcnt += 1;

        // Semaphore timed out condition - DER Control Event - queue monitoring 
        if (result == -1 && errno == ETIMEDOUT) 
        {
            i3e_dem_queueMonitoring(i3e_currentTime, u8_loopcnt);
        } 

        // Semaphore count increased condition - DER Control Event - queue scheduling
        else 
        {
            i3e_dem_queueScheduling(i3e_currentTime);
        }
    }
}
/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop void i3e_dem_queueScheduling(SE_TimeType_t i3_serverTime) 
{
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // Pull all DER Controls hrefs from the ResourcesGraph
    StackScop std::vector<std::string> vec_derCtrlHrefs = i3e_dagResources->I3E_DAG_FindResourceNodes(DERCTRL);

    // Update exisitng events with new data pulled from tree 
    i3e_dem_updateEvents(vec_derCtrlHrefs);

    // Determines which events are missing from the queue (aka, new events)
    StackScop std::vector<std::string> vec_newEventsHrefs = dercEventQueue.I3E_EVQ_FindMissingEventsInQueue(vec_derCtrlHrefs);

    // Process new events 
    i3e_dem_newEvents(vec_newEventsHrefs, i3_serverTime);

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Pull all DefaultDERControls from the Graph 
    StackScop std::vector<std::string> vec_defDerCtrlHrefs = i3e_dagResources->I3E_DAG_FindResourceNodes(DEFDERCTRL);

    // Trim Default Control events that don't exist anymore
    StackScop std::vector<std::string> vec_removedHrefs = dercEventQueue.I3E_EVQ_RemoveNonexistentDefaults(vec_defDerCtrlHrefs);
    for(uint8_t ix = 0; ix < vec_removedHrefs.size(); ix++)
    {
        uint32_t u32_ctrlsReset = dercEventQueue.I3E_EVQ_RemoveDefaultHrefAndClearFlags(vec_removedHrefs.at(ix));
        CST_ResetControl(u32_ctrlsReset);
        CST_ResetCurve(u32_ctrlsReset);
        uint32_t u32_defaultsReset = dercEventQueue.I3E_EVQ_RemoveSettingsHrefAndClearFlags(vec_removedHrefs.at(ix));
        CST_ResetDefaultControl(u32_defaultsReset);
    }

    // Update Default Control events with new data pulled from tree 
    i3e_dem_updateDefaults(vec_defDerCtrlHrefs);

    // Determines which defaults are missing from the queue (aka, new defaults)
    StackScop std::vector<std::string> vec_newDefaultsHrefs = 
                    dercEventQueue.I3E_EVQ_FindMissingEventsInDefaultQueue(vec_defDerCtrlHrefs);

    // Process new defaults
    i3e_dem_newDefaults(vec_newDefaultsHrefs);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_dem_queueMonitoring(SE_TimeType_t i3e_serverTime, uint8_t u8_loopcnt) 
{
    StackScop std::string sz_rxPayload;
    StackScop std::string sz_txPayload;

    // LOCK DERCEVENTQUEUE MUTEX
    pthread_mutex_lock(&dercEventQueue.I3E_EVQ_GetEventQueueMutex());   
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Pull Model state into I3E application 
    CST_UtilModeStatus_t enm_ModelState = CST_UtilCMDStatusGetter();

    // If the model has reported a conflict, we need to exit and disable controls gracefully
    if (enm_ModelState == MODEL_CONF || enm_ModelState == MODEL_UNKNOWN || enm_ModelState == MODEL_UNAVAIL)
    {

        // Process the model conflict in CST (clear locked registers, and set util_cmd_en to 0)
        CST_ModelConflict();

    }

    // #TODO - these cases may or may not be necessary here - kept in for now
    // If the model has returned to available, we need to permit controls 
    if (enm_ModelState == MODEL_AVAIL || enm_ModelState == MODEL_ACTDER)
    {

    }

    // Check for completed events (current time > event start + duration)
    i3e_dem_queueCheckCompletedEvents(i3e_serverTime);

    // Check for events starting in 60 seconds 
    i3e_dem_queuePreCheckEvents(i3e_serverTime, sz_rxPayload);

    // Check queue for events that need to start & start them (current time > event start time)
    i3e_dem_queueCheckStartEvents(i3e_serverTime);

    // Occasional check for cancelled DER Programs that are currently active - have to poll the DERControl and see if 
    // the server cancelled it
    i3e_dem_queueCheckActiveEvents(i3e_serverTime, u8_loopcnt, sz_rxPayload);

    // Occasional check for cancelled DER Programs 
    i3e_dem_queueCheckCancelledEvents(i3e_serverTime, u8_loopcnt);

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Occasional check for trimming old DERControls
    i3e_dem_queueEventTrimmer(i3e_serverTime);

    // Print all events in queue 
    // dercEventQueue.I3E_EVQ_PrintAllEventsInQueue();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    pthread_mutex_unlock(&dercEventQueue.I3E_EVQ_GetEventQueueMutex());

    // Periodically check if we can send util_cmd_en-0 to the EMS
    CST_UtilCMD(u8_loopcnt);
    
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_dem_queueCheckCompletedEvents(SE_TimeType_t i3_serverTime) 
{
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // Check for completed events (current time > event start + duration)
    StackScop std::vector<std::string> vec_eventsCompletedHrefs = dercEventQueue.I3E_EVQ_GetEventsCompleted(i3_serverTime);

    // Iterate through all events in the upcoming pipeline
    for(uint8_t i = 0; i < vec_eventsCompletedHrefs.size(); i++) 
    {
        StackScop EventBlock str_event = dercEventQueue.I3E_EVQ_GetSpecificEvent(vec_eventsCompletedHrefs.at(i));

        // Event hasn't been marked completed
        if(!(dercEventQueue.I3E_EVQ_CheckEventFlag(str_event, EVQ_eventCompleted)) && 
           !(dercEventQueue.I3E_EVQ_CheckEventFlag(str_event, EVQ_eventCancelled)) && 
           !(dercEventQueue.I3E_EVQ_CheckEventFlag(str_event, EVQ_eventOptOut)))
        {
            dercEventQueue.I3E_EVQ_ClearEventFlags(str_event, EVQ_eventActive);
            dercEventQueue.I3E_EVQ_SetEventFlags(str_event, EVQ_eventCompleted);
            uint32_t u32_ctrlsReset = dercEventQueue.I3E_EVQ_RemoveHrefAndClearFlags(str_event.sz_href);
            CST_ResetControl(u32_ctrlsReset);
            CST_ResetCurve(u32_ctrlsReset);

            // IF EVENT HAS A RESPONSE REQUIRED FIELD, WE WANT TO ACKNOEWLEDGE WE COMPLETED THE EVENT
            if (str_event.i3e_derCtrl.responseRequired == 7) 
            {
                dercEventQueue.I3E_EVQ_SetEventFlags(str_event, EVQ_eventCompletedMsg);
                i3e_dem_controlResponse(i3_serverTime, str_event, RSPSTAT_EVENT_COMPLETED);
            }
            dercEventQueue.I3E_EVQ_ReplaceEventBlockInQueue(vec_eventsCompletedHrefs.at(i), str_event);
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_dem_queuePreCheckEvents(SE_TimeType_t i3_serverTime, std::string sz_rxPayload)
{
    sz_rxPayload.clear();
    StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
    StackScop uint16_t u16_port = I3E_NSS_GetPORT();


    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // Catalog which events start in less than 60 seconds from now and check the resource once before activation
    StackScop std::vector<std::string> vec_eventPipelineHrefs = dercEventQueue.I3E_EVQ_GetEventsStartingIn60Seconds(i3_serverTime);

    // Iterate through all events in the upcoming pipeline
    for(uint8_t i = 0; i < vec_eventPipelineHrefs.size(); i++) 
    {

        StackScop EventBlock str_event = dercEventQueue.I3E_EVQ_GetSpecificEvent(vec_eventPipelineHrefs.at(i));

        // Event hasn't been checked 60 seconds prior 
        if(!(dercEventQueue.I3E_EVQ_CheckEventFlag(str_event, EVQ_eventPrecheck)) && 
           !(dercEventQueue.I3E_EVQ_CheckEventFlag(str_event, EVQ_eventCancelled)) &&
           !(dercEventQueue.I3E_EVQ_CheckEventFlag(str_event, EVQ_eventOptOut))) 
        {
            
            // Send HTTP GET Method to DERControl resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, (vec_eventPipelineHrefs.at(i) + MAX_RSRC_RET_SUFFIX), 
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // Finding this means the event was cancelled 
                StackScop std::size_t siz_found = sz_rxPayload.find("<currentStatus>2</currentStatus>");
                if (siz_found!=std::string::npos) 
                {
                    dercEventQueue.I3E_EVQ_ClearEventFlags(str_event, EVQ_eventActive);
                    dercEventQueue.I3E_EVQ_SetEventFlags(str_event, EVQ_eventCancelled);
                    uint32_t u32_ctrlsReset = dercEventQueue.I3E_EVQ_RemoveHrefAndClearFlags(str_event.sz_href);
                    CST_ResetControl(u32_ctrlsReset);
                    CST_ResetCurve(u32_ctrlsReset);

                    // IF EVENT HAS A RESPONSE REQUIRED FIELD, WE WANT TO ACKNOEWLEDGE WE CANCELLED THE EVENT
                    if (str_event.i3e_derCtrl.responseRequired == 7) 
                    {
                        dercEventQueue.I3E_EVQ_SetEventFlags(str_event, EVQ_eventCancelledMsg);
                        i3e_dem_controlResponse(i3_serverTime, str_event, RSPSTAT_EVENT_CANCELLED);
                    }
                    dercEventQueue.I3E_EVQ_ReplaceEventBlockInQueue(str_event.sz_href, str_event);
                }
            }
            dercEventQueue.I3E_EVQ_SetEventFlags(str_event, EVQ_eventPrecheck);
            dercEventQueue.I3E_EVQ_ReplaceEventBlockInQueue(vec_eventPipelineHrefs.at(i), str_event);
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_dem_queueCheckStartEvents(SE_TimeType_t i3_serverTime)
{
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // Check queue for events that need to start & start them (current time > event start time)
    StackScop std::vector<std::string> vec_eventsStartingHrefs = dercEventQueue.I3E_EVQ_GetEventsStarted(i3_serverTime);

    // Iterate through all events in the starting vector
    for(uint8_t i = 0; i < vec_eventsStartingHrefs.size(); i++) 
    {
        StackScop EventBlock str_event = dercEventQueue.I3E_EVQ_GetSpecificEvent(vec_eventsStartingHrefs.at(i));

        // Event hasn't been started yet
        if(!(dercEventQueue.I3E_EVQ_CheckEventFlag(str_event, EVQ_eventActive))     && 
           !(dercEventQueue.I3E_EVQ_CheckEventFlag(str_event, EVQ_eventSuperseded)) && 
           !(dercEventQueue.I3E_EVQ_CheckEventFlag(str_event, EVQ_eventCancelled))  && 
           !(dercEventQueue.I3E_EVQ_CheckEventFlag(str_event, EVQ_eventCompleted)) &&
           !(dercEventQueue.I3E_EVQ_CheckEventFlag(str_event, EVQ_eventOptOut)))
        {

            CST_UtilModeStatus_t enm_utilModeStat = CST_UtilCMDStatusGetter();    
            if(enm_utilModeStat == MODEL_UNKNOWN || enm_utilModeStat == MODEL_UNAVAIL || enm_utilModeStat == MODEL_CONF)
            {

                // #TODO - these resetting the control may or may not be necessary here - kept in for now
                // Mark new event as opted out
                dercEventQueue.I3E_EVQ_SetEventFlags(str_event, EVQ_eventOptOut);
                // uint32_t u32_ctrlsReset = dercEventQueue.I3E_EVQ_RemoveHrefAndClearFlags(str_event.sz_href);
                // CST_ResetControl(u32_ctrlsReset);
                // CST_ResetCurve(u32_ctrlsReset);

                // IF EVENT HAS A RESPONSE REQUIRED FIELD, WE WANT TO ACKNOEWLEDGE WE OPTED OUT
                if (str_event.i3e_derCtrl.responseRequired == 7) 
                {
                    dercEventQueue.I3E_EVQ_SetEventFlags(str_event, EVQ_eventOptOutMsg);
                    i3e_dem_controlResponse(i3_serverTime, str_event, RSPSTAT_USER_OPT_OUT);
                }
                dercEventQueue.I3E_EVQ_ReplaceEventBlockInQueue(str_event.sz_href, str_event);
            }
            else if(enm_utilModeStat == MODEL_AVAIL || enm_utilModeStat == MODEL_ACTDER)
            {
                // Check against any other events that are active in the queue with the same control type 
                StackScop std::vector<std::string> vec_eventsActiveHrefs = dercEventQueue.I3E_EVQ_FindEventsWithActiveFlag();
                for(uint8_t j = 0; j < vec_eventsActiveHrefs.size(); j++) 
                {

                    StackScop EventBlock str_activeEvent = dercEventQueue.I3E_EVQ_GetSpecificEvent(vec_eventsActiveHrefs.at(j));
                    if(str_activeEvent.i3e_derCtrl.DERControlBase._flags == str_event.i3e_derCtrl.DERControlBase._flags) 
                    {
                        // Found an event of same control type. We want to cancel that event we found - mark superseded
                        dercEventQueue.I3E_EVQ_SetEventFlags(str_activeEvent, (EVQ_eventSuperseded));
                        dercEventQueue.I3E_EVQ_ClearEventFlags(str_activeEvent, (EVQ_eventActive));
                        uint32_t u32_ctrlsReset = dercEventQueue.I3E_EVQ_RemoveHrefAndClearFlags(str_event.sz_href);
                        CST_ResetControl(u32_ctrlsReset);
                        CST_ResetCurve(u32_ctrlsReset);

                        // IF EVENT HAS A RESPONSE REQUIRED FIELD, WE WANT TO ACKNOEWLEDGE WE SUPERSEDED
                        if (str_activeEvent.i3e_derCtrl.responseRequired == 7) 
                        {
                            dercEventQueue.I3E_EVQ_SetEventFlags(str_activeEvent, (EVQ_eventSupersededMsg));
                            i3e_dem_controlResponse(i3_serverTime, str_activeEvent, RSPSTAT_EVENT_SUPERSEDED);
                        }
                        dercEventQueue.I3E_EVQ_ReplaceEventBlockInQueue(str_activeEvent.sz_href, str_activeEvent);
                    }
                }

                // Mark new event as active
                dercEventQueue.I3E_EVQ_SetEventFlags(str_event, EVQ_eventActive);

                // IF EVENT HAS A RESPONSE REQUIRED FIELD, WE WANT TO ACKNOEWLEDGE WE STARTED THE EVENT
                if (str_event.i3e_derCtrl.responseRequired == 7) 
                {
                    dercEventQueue.I3E_EVQ_SetEventFlags(str_event, EVQ_eventStartedMsg);
                    i3e_dem_controlResponse(i3_serverTime, str_event, RSPSTAT_EVENT_STARTED);
                }
                dercEventQueue.I3E_EVQ_ReplaceEventBlockInQueue(vec_eventsStartingHrefs.at(i), str_event);
            }
            else{}
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_dem_queueCheckActiveEvents(SE_TimeType_t i3_serverTime, uint8_t u8_loopcnt, std::string sz_rxPayload)
{
    sz_rxPayload.clear();
    StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
    StackScop uint16_t u16_port = I3E_NSS_GetPORT();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // Occasional check for cancelled DER Programs - have to poll the DERControl and see if the server cancelled it
    if(u8_loopcnt == 5) 
    {

        StackScop std::vector<std::string> vec_eventsStarted = dercEventQueue.I3E_EVQ_FindEventsWithActiveFlag();

        // Iterate through all events in the started vector
        for(uint8_t i = 0; i < vec_eventsStarted.size(); i++) 
        {
            StackScop EventBlock str_event = dercEventQueue.I3E_EVQ_GetSpecificEvent(vec_eventsStarted.at(i));

            // Send HTTP GET Method to DERControl resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, (vec_eventsStarted.at(i) + MAX_RSRC_RET_SUFFIX), 
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // Finding this means the event was cancelled 
                StackScop std::size_t siz_found = sz_rxPayload.find("<currentStatus>2</currentStatus>");
                if (siz_found!=std::string::npos) 
                {
                    dercEventQueue.I3E_EVQ_ClearEventFlags(str_event, EVQ_eventActive);
                    dercEventQueue.I3E_EVQ_SetEventFlags(str_event, EVQ_eventCancelled);
                    uint32_t u32_ctrlsReset = dercEventQueue.I3E_EVQ_RemoveHrefAndClearFlags(str_event.sz_href);
                    CST_ResetControl(u32_ctrlsReset);
                    CST_ResetCurve(u32_ctrlsReset);

                    // IF EVENT HAS A RESPONSE REQUIRED FIELD, WE WANT TO ACKNOEWLEDGE WE CANCELLED THE EVENT
                    if (str_event.i3e_derCtrl.responseRequired == 7) 
                    {
                        dercEventQueue.I3E_EVQ_SetEventFlags(str_event, EVQ_eventCancelledMsg);
                        i3e_dem_controlResponse(i3_serverTime, str_event, RSPSTAT_EVENT_CANCELLED);
                    }
                    dercEventQueue.I3E_EVQ_ReplaceEventBlockInQueue(vec_eventsStarted.at(i), str_event);
                }
            }
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_dem_queueCheckCancelledEvents(SE_TimeType_t i3_serverTime, uint8_t u8_loopcnt) 
{
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // Occasional check for cancelled DER Programs
    if(u8_loopcnt == 5) 
    {
        StackScop std::vector<std::string> vec_eventsCancelled = dercEventQueue.I3E_EVQ_FindCancelledEventsInQueue();

        // Iterate through all events in the cancelled vector
        for(uint8_t i = 0; i < vec_eventsCancelled.size(); i++) 
        {
            StackScop EventBlock str_event = dercEventQueue.I3E_EVQ_GetSpecificEvent(vec_eventsCancelled.at(i));

            if(!(dercEventQueue.I3E_EVQ_CheckEventFlag(str_event, EVQ_eventCancelled)))
            {
                dercEventQueue.I3E_EVQ_ClearEventFlags(str_event, EVQ_eventActive);
                dercEventQueue.I3E_EVQ_SetEventFlags(str_event, EVQ_eventCancelled);
                uint32_t u32_ctrlsReset = dercEventQueue.I3E_EVQ_RemoveHrefAndClearFlags(str_event.sz_href);
                CST_ResetControl(u32_ctrlsReset);
                CST_ResetCurve(u32_ctrlsReset);

                // IF EVENT HAS A RESPONSE REQUIRED FIELD, WE WANT TO ACKNOEWLEDGE WE CANCELLED THE EVENT
                if (str_event.i3e_derCtrl.responseRequired == 7) 
                {
                    dercEventQueue.I3E_EVQ_SetEventFlags(str_event, EVQ_eventCancelledMsg);
                    i3e_dem_controlResponse(i3_serverTime, str_event, RSPSTAT_EVENT_CANCELLED);
                }
                dercEventQueue.I3E_EVQ_ReplaceEventBlockInQueue(vec_eventsCancelled.at(i), str_event);
            }
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_dem_queueEventTrimmer(SE_TimeType_t i3_serverTime) 
{
    StackScop std::vector<std::string> vec_trimEventsHrefs = dercEventQueue.I3E_EVQ_FindEventsToTrim();
    for(uint8_t ix = 0; ix < vec_trimEventsHrefs.size(); ix++)
    {
        LOG_INF("Trimming Event --- %s", vec_trimEventsHrefs.at(ix).c_str());
    }

    if(!vec_trimEventsHrefs.empty())
    {
        dercEventQueue.I3E_EVQ_RemoveEventsByHref(vec_trimEventsHrefs);
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_dem_updateEvents(std::vector<std::string> vec_derCtrlHrefs)
{

    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // Scan through events and figure which ones have been updated - replace those DERControlResources in EventBlocks 
    for(uint8_t i = 0; i < vec_derCtrlHrefs.size(); i++) 
    {
        // Get the DERControl from node
        StackScop NodeData str_nodeDataDerCtrl = i3e_dagResources->I3E_DAG_GetNode(vec_derCtrlHrefs.at(i));
        
        // If node has been updated
        if(str_nodeDataDerCtrl.bol_updated) 
        {
            // Replace the DERControl structure in Event Block
            StackScop SE_DERControl_t& i3e_derControlStored = std::get<SE_DERControl_t>(str_nodeDataDerCtrl.var_rsrc);
            dercEventQueue.I3E_EVQ_ReplaceDerCtrlForEvent(vec_derCtrlHrefs.at(i), i3e_derControlStored);

            // Set the node in resource graph updated value to false
            i3e_dagResources->I3E_DAG_ResetUpdatedFlag(vec_derCtrlHrefs.at(i));
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_dem_newEvents(std::vector<std::string> vec_newCtrlHrefs, SE_TimeType_t i3_serverTime) 
{
    StackScop std::string sz_rxPayload;
    StackScop std::string sz_txPayload;
    
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Loop through all new controls
    for(uint8_t i = 0; i < vec_newCtrlHrefs.size(); i++) 
    {
        
        // Get the DERControl from graph node
        StackScop NodeData str_nodeDataDerCtrl = i3e_dagResources->I3E_DAG_GetNode(vec_newCtrlHrefs.at(i));
        StackScop SE_DERControl_t& i3e_derControlStored = std::get<SE_DERControl_t>(str_nodeDataDerCtrl.var_rsrc);

        // Create new event block 
        StackScop EventBlock str_newEvent;

        // Event randomization 
        str_newEvent.i3e_randStart = i3e_derControlStored.interval.start + 
                        i3e_dem_eventRandomization(i3e_derControlStored.randomizeStart); 
        str_newEvent.i3e_randDuration = i3e_derControlStored.interval.duration + 
                        i3e_dem_eventRandomization(i3e_derControlStored.randomizeDuration);
        str_newEvent.i3e_derCtrl = i3e_derControlStored;
        str_newEvent.sz_href = i3e_derControlStored.href;

        // Get DERProgram node data (primacy value) of added event 
        StackScop std::string sz_newEventDerProgramHref = 
                        i3e_dagResources->I3E_DAG_FindParentNode(i3e_dagResources->I3E_DAG_FindParentNode(str_newEvent.sz_href));
        StackScop NodeData str_newEventDerProgramNode = i3e_dagResources->I3E_DAG_GetNode(sz_newEventDerProgramHref);
        StackScop SE_DERProgram_t& i3e_newEventDerProgramStored = std::get<SE_DERProgram_t>(str_newEventDerProgramNode.var_rsrc);
        str_newEvent.primacy = i3e_newEventDerProgramStored.primacy;

        // Add event to queue (chronological insertion)
        dercEventQueue.I3E_EVQ_InsertEventByStartTime(str_newEvent);

        // IF EVENT HAS A RESPONSE REQUIRED FIELD, WE WANT TO ACKNOEWLEDGE WE GOT THE EVENT 
        if (str_newEvent.i3e_derCtrl.responseRequired == 7) 
        {   
            dercEventQueue.I3E_EVQ_SetEventFlags(str_newEvent, EVQ_eventReceievedMsg);
            i3e_dem_controlResponse(i3_serverTime, str_newEvent, RSPSTAT_EVENT_RECEIVED);
        }

        dercEventQueue.I3E_EVQ_ReplaceEventBlockInQueue(str_newEvent.sz_href, str_newEvent);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        // Check for time overlap between new event and other events in the queue
        StackScop std::vector<std::string> vec_overlappingEvents = dercEventQueue.I3E_EVQ_FindOverlappingEvents(str_newEvent.sz_href);

        // Compare the just inserted event against all other events in the queue that overlap in time 
        for (uint8_t j = 0; j < vec_overlappingEvents.size(); j++) 
        {

            // Grab the event block for the queued event we are comparing to
            StackScop EventBlock str_queuedEvent = dercEventQueue.I3E_EVQ_GetSpecificEvent(vec_overlappingEvents.at(j));

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

            // New event control parameter overlap? 
            // #TODO - this logic might need to change to compare flags differently, currently checking if the flags are
            // #TODO - the same, but we might need to check if the flags overlap control to control...
            // if(str_newEvent.i3e_derCtrl.DERControlBase._flags == str_queuedEvent.i3e_derCtrl.DERControlBase._flags)
            // if((str_newEvent.i3e_derCtrl.DERControlBase._flags & str_queuedEvent.i3e_derCtrl.DERControlBase._flags) != 0)
            if((str_newEvent.i3e_derCtrl.DERControlBase._flags & str_queuedEvent.i3e_derCtrl.DERControlBase._flags) 
                                == str_queuedEvent.i3e_derCtrl.DERControlBase._flags)
            { 

                // Get DERProgram node data (primacy value) of added event 
                StackScop std::string sz_newEventDerProgramHref = 
                                i3e_dagResources->I3E_DAG_FindParentNode(i3e_dagResources->I3E_DAG_FindParentNode(str_newEvent.sz_href));
                StackScop NodeData str_newEventDerProgramNode = 
                                i3e_dagResources->I3E_DAG_GetNode(sz_newEventDerProgramHref);
                StackScop SE_DERProgram_t& i3e_newEventDerProgramStored = 
                                std::get<SE_DERProgram_t>(str_newEventDerProgramNode.var_rsrc);

                // Get the overlapping DER Control's DERProgram node data 
                StackScop std::string sz_queuedEventDerProgramHref = 
                                i3e_dagResources->I3E_DAG_FindParentNode(i3e_dagResources->I3E_DAG_FindParentNode(str_queuedEvent.sz_href));
                StackScop NodeData str_queuedEventDerProgramNode = 
                                i3e_dagResources->I3E_DAG_GetNode(sz_queuedEventDerProgramHref);
                StackScop SE_DERProgram_t& i3e_queuedEventDerProgramStored = 
                                std::get<SE_DERProgram_t>(str_queuedEventDerProgramNode.var_rsrc);

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

                // Yes control parameter overlap: ---- (Check primacy) 

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

                // New event is lower primacy, is the event we are checking against currently active?
                if(i3e_newEventDerProgramStored.primacy < i3e_queuedEventDerProgramStored.primacy) 
                {
                    
                    // If the queued event is currently active and not completed, then we want to keep running that 
                    // until it is the time for lower primacy program
                    if(dercEventQueue.I3E_EVQ_CheckEventFlag(str_queuedEvent, EVQ_eventActive) && 
                        !(dercEventQueue.I3E_EVQ_CheckEventFlag(str_queuedEvent, EVQ_eventCompleted)))
                    {
                        
                    }
                    // If the queued event isn't currently active, we can mark that event as superseded and send the message
                    else{
                        dercEventQueue.I3E_EVQ_SetEventFlags(str_queuedEvent, (EVQ_eventSuperseded));
                        uint32_t u32_ctrlsReset = dercEventQueue.I3E_EVQ_RemoveHrefAndClearFlags(str_queuedEvent.sz_href);
                        CST_ResetControl(u32_ctrlsReset);
                        CST_ResetCurve(u32_ctrlsReset);

                        if (str_queuedEvent.i3e_derCtrl.responseRequired == 7) 
                        {
                            dercEventQueue.I3E_EVQ_SetEventFlags(str_queuedEvent, (EVQ_eventSupersededMsg));
                            i3e_dem_controlResponse(i3_serverTime, str_queuedEvent, RSPSTAT_EVENT_SUPERSEDED);
                        }
                        dercEventQueue.I3E_EVQ_ReplaceEventBlockInQueue(str_queuedEvent.sz_href, str_queuedEvent);
                    }
                }

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

                // New event is equal primacy, schedule the event that is newer (larger creation time)
                else if(i3e_newEventDerProgramStored.primacy == i3e_queuedEventDerProgramStored.primacy) 
                {
                    // If the queued event is currently active and not completed, then we want to keep running that 
                    //until it is the time for lower primacy program
                    if(dercEventQueue.I3E_EVQ_CheckEventFlag(str_queuedEvent, EVQ_eventActive) && 
                        !(dercEventQueue.I3E_EVQ_CheckEventFlag(str_queuedEvent, EVQ_eventCompleted)))
                    {

                    }
                    // If the queued event isn't currently active, we can mark that event as superseded and send the message
                    else{
                        dercEventQueue.I3E_EVQ_SetEventFlags(str_queuedEvent, (EVQ_eventSuperseded));
                        uint32_t u32_ctrlsReset = dercEventQueue.I3E_EVQ_RemoveHrefAndClearFlags(str_queuedEvent.sz_href);
                        CST_ResetControl(u32_ctrlsReset);
                        CST_ResetCurve(u32_ctrlsReset);

                        if (str_queuedEvent.i3e_derCtrl.responseRequired == 7) 
                        {   
                            dercEventQueue.I3E_EVQ_SetEventFlags(str_queuedEvent, (EVQ_eventSupersededMsg));
                            i3e_dem_controlResponse(i3_serverTime, str_queuedEvent, RSPSTAT_EVENT_SUPERSEDED);
                        }
                        dercEventQueue.I3E_EVQ_ReplaceEventBlockInQueue(str_queuedEvent.sz_href, str_queuedEvent);
                    }
                }

                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

                // New event is higher primacy, mark as superseded immediately and send the message it was superseded
                else if(i3e_newEventDerProgramStored.primacy > i3e_queuedEventDerProgramStored.primacy) 
                {
                    dercEventQueue.I3E_EVQ_SetEventFlags(str_newEvent, (EVQ_eventSuperseded));
                    uint32_t u32_ctrlsReset = dercEventQueue.I3E_EVQ_RemoveHrefAndClearFlags(str_queuedEvent.sz_href);
                    CST_ResetControl(u32_ctrlsReset);
                    CST_ResetCurve(u32_ctrlsReset);
                    
                    if (str_newEvent.i3e_derCtrl.responseRequired == 7) 
                    {
                        dercEventQueue.I3E_EVQ_SetEventFlags(str_queuedEvent, (EVQ_eventSupersededMsg));
                        i3e_dem_controlResponse(i3_serverTime, str_newEvent, RSPSTAT_EVENT_SUPERSEDED); 
                    }
                    dercEventQueue.I3E_EVQ_ReplaceEventBlockInQueue(str_newEvent.sz_href, str_newEvent);
                } 
                else{}

            }
            else{} // No control parameter overlap: ---- (both events can be scheduled to be active)

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        }

    }

    // Print all events in queue 
    // dercEventQueue.I3E_EVQ_PrintAllEventsInQueue();
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


FileScop void i3e_dem_controlResponse(SE_TimeType_t i3_serverTime, EventBlock str_currentEvent, uint8_t u8_statusCode) 
{

    StackScop std::string sz_rxPayload;
    StackScop std::string sz_txPayload;
    StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
    StackScop uint16_t u16_port = I3E_NSS_GetPORT();

    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    // Send received message for the events to the server 
    StackScop SE_DERControlResponse_t i3e_rspStruct = {0};

    // Find EDEV List for lFDI 
    StackScop std::string sz_edevHref = (i3e_dagResources->I3E_DAG_FindResourceNodes(EDEV)).at(0); 
    StackScop NodeData str_edevRsrc = i3e_dagResources->I3E_DAG_GetNode(sz_edevHref);
    SE_EndDevice_t& i3e_edevStored = std::get<SE_EndDevice_t>(str_edevRsrc.var_rsrc);

    // Populate DERControlResponse Struct
    i3e_rspStruct.createdDateTime = i3_serverTime; 

    i3e_rspStruct.status = (SE_UInt8_t)u8_statusCode;
    memcpy(i3e_rspStruct.subject, str_currentEvent.i3e_derCtrl.mRID, sizeof(i3e_rspStruct.subject));
    i3e_rspStruct.href = str_currentEvent.i3e_derCtrl.replyTo;
    memcpy(i3e_rspStruct.endDeviceLFDI, i3e_edevStored.lFDI, sizeof(i3e_rspStruct.endDeviceLFDI));

    // Serialize the struct data into payload 
    I3E_XML_SRLZ_DERControlResponse(sz_txPayload, &i3e_rspStruct);

    sz_rxPayload.clear();
    // std::cout << "TEST POST -- " <<std::endl;
    I3E_HTTP_POST(sz_ipv4Addr, u16_port, str_currentEvent.i3e_derCtrl.replyTo, sz_txPayload, sz_rxPayload);
    // std::cout << "END TEST POST -- " <<std::endl;
    sz_txPayload.clear();

}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop SE_OneHourRangeType_t i3e_dem_eventRandomization(SE_OneHourRangeType_t i3e_hourRange) {

    // Seed the random number generator
    srand(static_cast<unsigned int>(I3E_CTS_getSyncTime()));

    // Generate a random number between 0 and the absolute value of the i3e_hourRange
    StackScop SE_OneHourRangeType_t randomNum = rand() % (abs(i3e_hourRange) + 1);

    // Adjust the sign of the random number based on the original i3e_hourRange
    if (i3e_hourRange < 0) {
        randomNum *= -1; // Make the number negative if the i3e_hourRange is negative
    }

    return randomNum;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_dem_updateDefaults(std::vector<std::string> vec_defCtrlHrefs) 
{
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // Scan through existing defaults and figure out which ones have been updated - replace those DERControlResources
    for(uint8_t i = 0; i < vec_defCtrlHrefs.size(); i++) 
    {
        // Get the DefaultDERControl from node
        StackScop NodeData str_defDerCtrlData = i3e_dagResources->I3E_DAG_GetNode(vec_defCtrlHrefs.at(i));
        
        // If node has been updated
        if(str_defDerCtrlData.bol_updated) 
        {
            // Replace the DERControl structure in Event Block
            StackScop SE_DefaultDERControl_t& str_defDerControlStored = 
                            std::get<SE_DefaultDERControl_t>(str_defDerCtrlData.var_rsrc);
            dercEventQueue.I3E_EVQ_ReplaceDefDerCtrlForEvent(vec_defCtrlHrefs.at(i), str_defDerControlStored);

            // Set the node in resource graph updated value to false
            i3e_dagResources->I3E_DAG_ResetUpdatedFlag(vec_defCtrlHrefs.at(i));
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_dem_newDefaults(std::vector<std::string> vec_newDefCtrlHrefs) 
{

    StackScop std::string sz_rxPayload;
    StackScop std::string sz_txPayload;

    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Loop through all new default controls to compare against the ones in the queue
    for(uint8_t i = 0; i < vec_newDefCtrlHrefs.size(); i++) 
    {
        
        // Get the DefaultDERControl from node
        StackScop NodeData str_defDerCtrlData = i3e_dagResources->I3E_DAG_GetNode(vec_newDefCtrlHrefs.at(i));
        StackScop SE_DefaultDERControl_t& i3e_defDerCtrlStored = std::get<SE_DefaultDERControl_t>(str_defDerCtrlData.var_rsrc);

        StackScop DefaultBlock str_newDefault;
        str_newDefault.i3e_defDerCtrl = i3e_defDerCtrlStored;
        str_newDefault.sz_href = i3e_defDerCtrlStored.href;

        // Get DERProgram node data (primacy value) of added event 
        StackScop std::string sz_newDefaultHref = i3e_dagResources->I3E_DAG_FindParentNode(str_newDefault.sz_href);
        StackScop NodeData str_newDefaultProgramNode = i3e_dagResources->I3E_DAG_GetNode(sz_newDefaultHref);
        StackScop SE_DERProgram_t& i3e_newDefaultDerProgramStored = std::get<SE_DERProgram_t>(str_newDefaultProgramNode.var_rsrc);
        str_newDefault.primacy = i3e_newDefaultDerProgramStored.primacy;

        dercEventQueue.I3E_EVQ_AddDefaultToQueue(str_newDefault);
    }

    // Order queue by primacy values
    dercEventQueue.I3E_EVQ_OrderDefaultQueueByPrimacy();
}
