#include "lion/log/log.h"
#include "include/I3E_EVQ.hpp"

LOG_MODULE_REGISTER(i3e_evq, LOG_LEVEL_INF);
/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop EventQueue::EventQueue() 
{
    ModScop pthread_mutexattr_t attr;
    pthread_mutexattr_init(&attr);
    pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
    pthread_mutex_init(&mtx_eventQueueMutex, &attr); // Initialize the pthread_mutex as recursive
    pthread_mutexattr_destroy(&attr); // Destroy the mutex attribute after initializing the mutex
    vec_activeControlsHrefs.resize(SZ_ACTIVE_CONTROL_MAX);
    vec_activeDefaultsHrefs.resize(SZ_ACTIVE_CONTROL_MAX);
    vec_activeSettingsHrefs.resize(SZ_ACTIVE_CONTROL_MAX);
    vec_activeControlsMRIDs.resize(SZ_ACTIVE_CONTROL_MAX);
    vec_activeDefaultsMRIDs.resize(SZ_ACTIVE_CONTROL_MAX);
    vec_activeSettingsMRIDs.resize(SZ_ACTIVE_CONTROL_MAX);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
ModScop EventQueue::~EventQueue() 
{
    pthread_mutex_destroy(&mtx_eventQueueMutex); // Initialize the pthread_mutex
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop pthread_mutex_t& EventQueue::I3E_EVQ_GetEventQueueMutex() 
{
    return mtx_eventQueueMutex;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
ModScop uint32_t EventQueue::I3E_EVQ_GetActiveDERControlBitfield()
{
    return u32_activeControls;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~   
ModScop std::vector<std::string> EventQueue::I3E_EVQ_GetActiveDERControlMRIDs()
{
    return vec_activeControlsMRIDs;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
ModScop uint32_t EventQueue::I3E_EVQ_GetActiveDefaultsDERControlBitfield()
{
    return u32_activeDefaultControls;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~   
ModScop std::vector<std::string> EventQueue::I3E_EVQ_GetActiveDefaultsDERControlMRIDs()
{
    return vec_activeDefaultsMRIDs;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
ModScop uint32_t EventQueue::I3E_EVQ_GetSettingsBitfield()
{
    return u32_activeSettings;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~   
ModScop std::vector<std::string> EventQueue::I3E_EVQ_GetSettingsMRIDs()
{
    return vec_activeSettingsMRIDs;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void EventQueue::I3E_EVQ_AddEvent(const EventBlock& str_event)
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    q_eventQueue.push(str_event);
    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                   
ModScop void EventQueue::I3E_EVQ_SetActiveControl(uint32_t u32_flags) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    u32_activeControls |= u32_flags;
    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  
ModScop void EventQueue::I3E_EVQ_ClearActiveControl(uint32_t u32_flags) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    u32_activeControls &= ~u32_flags;
    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
ModScop bool EventQueue::I3E_EVQ_CheckActiveControl(uint32_t u32_flag) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    bool isSet = (u32_activeControls & u32_flag) != 0;
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return isSet;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void EventQueue::I3E_EVQ_SetActiveDefault(uint32_t u32_flags) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    u32_activeDefaultControls |= u32_flags;
    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        
ModScop void EventQueue::I3E_EVQ_ClearActiveDefault(uint32_t u32_flags) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    u32_activeDefaultControls &= ~u32_flags;
    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
       
ModScop bool EventQueue::I3E_EVQ_CheckActiveDefault(uint32_t u32_flag) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    bool isSet = (u32_activeDefaultControls & u32_flag) != 0;
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return isSet;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void EventQueue::I3E_EVQ_SetSettings(uint32_t u32_flags) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    u32_activeSettings |= u32_flags;
    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        
ModScop void EventQueue::I3E_EVQ_ClearSettings(uint32_t u32_flags) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    u32_activeSettings &= ~u32_flags;
    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
       
ModScop bool EventQueue::I3E_EVQ_CheckSettings(uint32_t u32_flag) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    bool isSet = (u32_activeSettings & u32_flag) != 0;
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return isSet;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        
ModScop void EventQueue::I3E_EVQ_SetEventFlags(EventBlock& str_event, uint32_t u32_flags) 
{
    str_event.u32_eventFlags |= u32_flags; // Set the specified flag
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void EventQueue::I3E_EVQ_ClearEventFlags(EventBlock& str_event, uint32_t u32_flags) 
{
    str_event.u32_eventFlags &= ~u32_flags; // Clear the specified flag
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop bool EventQueue::I3E_EVQ_CheckEventFlag(const EventBlock& str_event, uint32_t u32_flags) 
{
    return (str_event.u32_eventFlags & u32_flags) != 0; // Check if the specified flag is set
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      
ModScop std::vector<std::string> EventQueue::I3E_EVQ_FindMissingEventsInQueue(const std::vector<std::string>& vec_derCtrlHrefs)
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    ModScop std::vector<std::string> vec_missingEvents;
    for (const std::string& href : vec_derCtrlHrefs)
    {
        ModScop bool bol_found = false;
        ModScop std::queue<EventBlock> q_tempQueue = q_eventQueue; // Create a temporary queue for iteration

        while (!q_tempQueue.empty())
        {
            ModScop EventBlock str_currentEvent = q_tempQueue.front();
            q_tempQueue.pop();
            if (str_currentEvent.sz_href == href)
            {
                bol_found = true;
                break;
            }
        }
        if (!bol_found)
        {
            vec_missingEvents.push_back(href);
        }
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return vec_missingEvents;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void EventQueue::I3E_EVQ_PrintAllEventsInQueue()
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    
    ModScop Graph* i3e_dagResources = I3E_SPP_GetDAG();
    ModScop std::queue<EventBlock> q_tempQueue = q_eventQueue; // Create a temporary queue for iteration
    while (!q_tempQueue.empty())
    {
        ModScop EventBlock str_currentEvent = q_tempQueue.front();
        q_tempQueue.pop();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // Print out the details of the current event
        std::cout << "Event Details:" << str_currentEvent.i3e_derCtrl.description << std::endl;

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        std::cout << "Event ID: ";
        for (size_t i = 0; i < sizeof(str_currentEvent.i3e_derCtrl.mRID); i++) 
        {
            std::cout << std::hex << std::setw(2) << std::setfill('0') << 
                            static_cast<int>(str_currentEvent.i3e_derCtrl.mRID[i]) << " ";
        }
        std::cout << std::dec << std::endl;

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        std::cout << "DER Control Href: " << str_currentEvent.sz_href << std::endl;

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        ModScop std::string sz_parentNodeCtrlList = i3e_dagResources->
                        I3E_DAG_FindParentNode(str_currentEvent.i3e_derCtrl.href);
        ModScop std::string sz_parentNodeDERProgram = i3e_dagResources->
                        I3E_DAG_FindParentNode(sz_parentNodeCtrlList);
        std::cout << "DER Program Parent HREF: " << sz_parentNodeDERProgram << std::endl;

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        std::cout << "Event Block Flags: " << std::endl;
        
        std::cout 
        << ((str_currentEvent.u32_eventFlags & EVQ_eventReceievedMsg)   ? "1" : "0") << "-"
        << ((str_currentEvent.u32_eventFlags & EVQ_eventCompleted)      ? "1" : "0") << "-"
        << ((str_currentEvent.u32_eventFlags & EVQ_eventCompletedMsg)   ? "1" : "0") << "-"
        << ((str_currentEvent.u32_eventFlags & EVQ_eventStarted)        ? "1" : "0") << "-"
        << ((str_currentEvent.u32_eventFlags & EVQ_eventStartedMsg)     ? "1" : "0") << "-"
        << ((str_currentEvent.u32_eventFlags & EVQ_eventSuperseded)     ? "1" : "0") << "-"
        << ((str_currentEvent.u32_eventFlags & EVQ_eventSupersededMsg)  ? "1" : "0") << "-"
        << ((str_currentEvent.u32_eventFlags & EVQ_eventCancelled)      ? "1" : "0") << "-"
        << ((str_currentEvent.u32_eventFlags & EVQ_eventCancelledMsg)   ? "1" : "0") << "-"
        << ((str_currentEvent.u32_eventFlags & EVQ_eventActive)         ? "1" : "0") << "-"
        << ((str_currentEvent.u32_eventFlags & EVQ_eventClear)          ? "1" : "0") << "-"
        << ((str_currentEvent.u32_eventFlags & EVQ_eventPrecheck)       ? "1" : "0") << "-"
        << std::endl;

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        std::cout << "DER Control Status: " << static_cast<int>
                        (str_currentEvent.i3e_derCtrl.EventStatus.currentStatus) << std::endl;

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    }
    std::cout << "END OF QUEUE" << std::endl;
    pthread_mutex_unlock(&mtx_eventQueueMutex);
}


ModScop void EventQueue::I3E_EVQ_InsertEventByStartTime(const EventBlock& str_newEvent)
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    ModScop std::queue<EventBlock> q_updatedQueue; // Temporary queue to hold sorted events

    // Insert the new event in the correct position based on start time
    ModScop bool bol_newEventInserted = false;
    while (!q_eventQueue.empty())
    {
        ModScop EventBlock str_currentEvent = q_eventQueue.front();
        q_eventQueue.pop();

        if (!bol_newEventInserted && str_newEvent.i3e_randStart < str_currentEvent.i3e_randStart)
        {
            q_updatedQueue.push(str_newEvent); // Insert the new event
            bol_newEventInserted = true;
        }
        q_updatedQueue.push(str_currentEvent); // Push the current event
    }

    if (!bol_newEventInserted)
    {
        q_updatedQueue.push(str_newEvent); // Insert at the end if new event is the latest
    }

    // Update the eventQueue with the sorted events
    q_eventQueue = q_updatedQueue;
    pthread_mutex_unlock(&mtx_eventQueueMutex);
}   

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::vector<std::string> EventQueue::I3E_EVQ_FindOverlappingEvents(const std::string& sz_specificHref) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    
    ModScop std::vector<std::string> vec_overlappingEvents; // Vector to store hrefs of overlapping events
    
    // Get the start and duration times of the specific event
    ModScop SE_TimeType_t i31_specificStartTime;
    ModScop SE_UInt32_t i3e_specificDuration;
    
    // Find the specific event in the queue and extract its start time and duration
    ModScop EventBlock str_specificEvent;
    ModScop bool bol_foundSpecificEvent = false;
    ModScop std::queue<EventBlock> q_tempQueue = q_eventQueue; // Create a temporary queue for iteration
    
    while (!q_tempQueue.empty()) {
        ModScop EventBlock str_currentEvent = q_tempQueue.front();
        q_tempQueue.pop();
        
        if (str_currentEvent.sz_href == sz_specificHref) 
        {
            i31_specificStartTime = str_currentEvent.i3e_randStart;
            i3e_specificDuration = str_currentEvent.i3e_randDuration;
            str_specificEvent = str_currentEvent;
            bol_foundSpecificEvent = true;
            break;
        }
    }
    
    if (!bol_foundSpecificEvent) 
    {
        pthread_mutex_unlock(&mtx_eventQueueMutex);
        return vec_overlappingEvents; // Return empty vector if specific event not found
    }
    
    // Check for overlapping events
    q_tempQueue = q_eventQueue; // Reset the temporary queue for iteration
    
    while (!q_tempQueue.empty()) 
    {
        ModScop EventBlock str_currentEvent = q_tempQueue.front();
        q_tempQueue.pop();
        
        if (str_currentEvent.sz_href != sz_specificHref &&
            str_currentEvent.i3e_randStart < i31_specificStartTime + i3e_specificDuration &&
            i31_specificStartTime < str_currentEvent.i3e_randStart + str_currentEvent.i3e_randDuration) 
        {
            vec_overlappingEvents.push_back(str_currentEvent.sz_href);
        }
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex);

    return vec_overlappingEvents;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::vector<std::string> EventQueue::I3E_EVQ_FindEventsToTrim() 
{
    pthread_mutex_lock(&mtx_eventQueueMutex); // Lock the mutex to access the event queue safely

    ModScop Graph* ptr_dag_dagResources = I3E_SPP_GetDAG();
    ModScop std::vector<std::string> vec_eventsWithSpecificFlags; // Vector to store hrefs of events with specific flags
    ModScop std::queue<EventBlock> q_tempQueue = q_eventQueue; // Create a temporary queue for iteration

    while (!q_tempQueue.empty()) 
    {
        ModScop EventBlock str_currentEvent = q_tempQueue.front(); // Get the front event
        q_tempQueue.pop(); // Remove the front event from the temporary queue

        ModScop bool bol_hasCompletedPair = false;
        ModScop bool bol_hasSupersededPair = false;
        ModScop bool bol_hasCancelledPair = false;
        ModScop bool bol_hasOptOutPair = false;

        if(str_currentEvent.i3e_derCtrl.responseRequired == 7) 
        {
            bol_hasCompletedPair = (str_currentEvent.u32_eventFlags & EVQ_eventCompleted) 
                            && (str_currentEvent.u32_eventFlags & EVQ_eventCompletedMsg);
            bol_hasSupersededPair = (str_currentEvent.u32_eventFlags & EVQ_eventSuperseded) 
                            && (str_currentEvent.u32_eventFlags & EVQ_eventSupersededMsg);
            bol_hasCancelledPair = (str_currentEvent.u32_eventFlags & EVQ_eventCancelled) 
                            && (str_currentEvent.u32_eventFlags & EVQ_eventCancelledMsg);
            bol_hasOptOutPair = (str_currentEvent.u32_eventFlags & EVQ_eventOptOut) 
                            && (str_currentEvent.u32_eventFlags & EVQ_eventOptOutMsg);
        }
        else
        {
            bol_hasCompletedPair = (str_currentEvent.u32_eventFlags & EVQ_eventCompleted);
            bol_hasSupersededPair = (str_currentEvent.u32_eventFlags & EVQ_eventSuperseded);
            bol_hasCancelledPair = (str_currentEvent.u32_eventFlags & EVQ_eventCancelled);
            bol_hasOptOutPair = (str_currentEvent.u32_eventFlags & EVQ_eventOptOut);
        }

        if ((bol_hasCompletedPair || bol_hasSupersededPair || bol_hasCancelledPair || bol_hasOptOutPair) 
                        && !ptr_dag_dagResources->I3E_DAG_DoesNodeExist(str_currentEvent.i3e_derCtrl.href)) 
        {
            vec_eventsWithSpecificFlags.push_back(str_currentEvent.sz_href); // Add the href to the vector if any pair of flags is set
        }
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex); // Unlock the mutex after processing the event queue
    return vec_eventsWithSpecificFlags; // Return the vector of hrefs of events with specific flags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void EventQueue::I3E_EVQ_RemoveEventsByHref(const std::vector<std::string>& vec_hrefsToRemove) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex); // Lock the mutex to access the event queue safely

    ModScop std::queue<EventBlock> q_tempQueue; // Temporary queue to hold events that should remain

    while (!q_eventQueue.empty()) 
    {
        ModScop EventBlock str_currentEvent = q_eventQueue.front(); // Get the front event
        q_eventQueue.pop(); // Remove the front event from the original queue

        // Check if the current event's href is in the list of hrefs to remove
        if (std::find(vec_hrefsToRemove.begin(), vec_hrefsToRemove.end(), str_currentEvent.sz_href) 
                        == vec_hrefsToRemove.end()) 
        {
            q_tempQueue.push(str_currentEvent); // If not found, add the event to the temporary queue
        }
    }

    // Swap the original queue with the temporary queue
    std::swap(q_eventQueue, q_tempQueue);
    pthread_mutex_unlock(&mtx_eventQueueMutex); // Unlock the mutex after processing the event queue
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop EventBlock EventQueue::I3E_EVQ_GetSpecificEvent(const std::string& sz_targetHref) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    
    ModScop EventBlock str_specificEvent; // Store the specific event
    ModScop std::queue<EventBlock> q_tempQueue = q_eventQueue; // Create a temporary queue for iteration
    
    while (!q_tempQueue.empty()) {
        ModScop EventBlock str_currentEvent = q_tempQueue.front();
        
        // Check if the current event's href matches the targetHref
        if (str_currentEvent.sz_href == sz_targetHref) 
        {
            str_specificEvent = str_currentEvent;
            break; // Found the specific event, exit the loop
        }       
        q_tempQueue.pop();
    }  
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return str_specificEvent;
}  

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  

ModScop void EventQueue::I3E_EVQ_ReplaceEventBlockInQueue(const std::string& sz_targetHref, const EventBlock& str_newEventBlock) {
    pthread_mutex_lock(&mtx_eventQueueMutex);
    
    ModScop std::queue<EventBlock> q_tempQueue; // Create a temporary queue for modification
    
    while (!q_eventQueue.empty()) {
        ModScop EventBlock str_currentEvent = q_eventQueue.front();
        q_eventQueue.pop();
        
        if (str_currentEvent.sz_href == sz_targetHref) {
            q_tempQueue.push(str_newEventBlock); // Replace with the new EventBlock
        } else {
            q_tempQueue.push(str_currentEvent); // Keep the original EventBlock
        }
    }
    
    // Update the original queue with the modified elements
    q_eventQueue = q_tempQueue;
    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::vector<std::string> EventQueue::I3E_EVQ_GetEventsStartingIn60Seconds(SE_TimeType_t i3e_currentTime) {
    pthread_mutex_lock(&mtx_eventQueueMutex);
    
    ModScop std::vector<std::string> vec_eventsStartingIn60Seconds; // Store href strings of events starting within 60 seconds
    
    //#TODO - changed timing to be 50 seconds since test case timing was hyper specific for 60 
    ModScop SE_TimeType_t i3e_sixtySecondsLater = i3e_currentTime + 50; // Calculate time 60 seconds from now
    ModScop std::queue<EventBlock> q_tempQueue = q_eventQueue; // Create a temporary queue for iteration
    
    while (!q_tempQueue.empty()) {
        ModScop EventBlock& str_currentEvent = q_tempQueue.front();

        if (str_currentEvent.i3e_randStart > i3e_currentTime && str_currentEvent.i3e_randStart <= i3e_sixtySecondsLater
            // && !str_currentEvent.eventSuperseded) {
            ) 
        {
            vec_eventsStartingIn60Seconds.push_back(str_currentEvent.sz_href);
        }
        
        q_tempQueue.pop();
    }
    
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return vec_eventsStartingIn60Seconds;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::vector<std::string> EventQueue::I3E_EVQ_GetEventsStarted(SE_TimeType_t i3e_currentTime) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);     
    ModScop std::vector<std::string> vec_eventsStarted; // Store href strings of events that have started    
    ModScop std::queue<EventBlock> q_tempQueue = q_eventQueue; // Create a temporary queue for iteration
    while (!q_tempQueue.empty()) {
        ModScop EventBlock& str_currentEvent = q_tempQueue.front();
        
        if ((str_currentEvent.i3e_randStart <= i3e_currentTime) && 
                        !(str_currentEvent.u32_eventFlags & EVQ_eventSuperseded)) 
        {
            vec_eventsStarted.push_back(str_currentEvent.sz_href); // Push the href string into the vector
        }    
        q_tempQueue.pop();
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return vec_eventsStarted;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::vector<std::string> EventQueue::I3E_EVQ_GetEventsCompleted(SE_TimeType_t i3e_currentTime) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    
    ModScop std::vector<std::string> vec_eventsCompleted; // Store href strings of events that have completed
    ModScop std::queue<EventBlock> q_tempQueue = q_eventQueue; // Create a temporary queue for iteration
    while (!q_tempQueue.empty()) {

        ModScop EventBlock& str_currentEvent = q_tempQueue.front();
        
        if ((i3e_currentTime >= str_currentEvent.i3e_randStart + str_currentEvent.i3e_randDuration) 
                        && !(str_currentEvent.u32_eventFlags & EVQ_eventSuperseded)) 
        {
            vec_eventsCompleted.push_back(str_currentEvent.sz_href); // Push the href string into the vector
        }

        q_tempQueue.pop();
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return vec_eventsCompleted;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void EventQueue::I3E_EVQ_ReplaceDerCtrlForEvent(const std::string& sz_eventHref, const SE_DERControl_t& i3e_newDerCtrl) {
    pthread_mutex_lock(&mtx_eventQueueMutex);

    ModScop std::queue<EventBlock> q_updatedQueue;
    while (!q_eventQueue.empty()) {
        ModScop EventBlock str_currentEvent = q_eventQueue.front();
        q_eventQueue.pop();

        if (str_currentEvent.sz_href == sz_eventHref) {
            // Replace derCtrl for the specific event
            str_currentEvent.i3e_derCtrl = i3e_newDerCtrl;
        }
        q_updatedQueue.push(str_currentEvent);
    }
    q_eventQueue = q_updatedQueue;
    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::vector<std::string> EventQueue::I3E_EVQ_FindCancelledEventsInQueue() 
{
    pthread_mutex_lock(&mtx_eventQueueMutex); // Lock the mutex to access the event queue safely

    ModScop std::vector<std::string> vec_cancelledEvents; // Vector to store hrefs of canceled events
    ModScop std::queue<EventBlock> q_tempQueue = q_eventQueue; // Create a temporary queue for iteration

    while (!q_tempQueue.empty()) 
    {
        ModScop EventBlock str_currentEvent = q_tempQueue.front(); // Get the front event
        q_tempQueue.pop(); // Remove the front event from the temporary queue

        if (str_currentEvent.i3e_derCtrl.EventStatus.currentStatus == 2) {
            vec_cancelledEvents.push_back(str_currentEvent.sz_href); // Add the href to the vector if the currentStatus is 2
        }
    }

    pthread_mutex_unlock(&mtx_eventQueueMutex); // Unlock the mutex after processing the event queue
    return vec_cancelledEvents; // Return the vector of hrefs of canceled events
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::vector<std::string> EventQueue::I3E_EVQ_FindEventsWithActiveFlag() 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    ModScop std::vector<std::string> vec_eventsWithActiveFlag;
    ModScop std::queue<EventBlock> q_tempQueue = q_eventQueue;
    while (!q_tempQueue.empty()) {
        ModScop EventBlock str_currentEvent = q_tempQueue.front();
        q_tempQueue.pop();

        if (str_currentEvent.u32_eventFlags & EVQ_eventActive) 
        {
            vec_eventsWithActiveFlag.push_back(str_currentEvent.sz_href);
        }
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return vec_eventsWithActiveFlag;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void EventQueue::I3E_EVQ_AddDefaultToQueue(const DefaultBlock& str_newDefault) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    q_defaultQueue.push(str_newDefault);
    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop bool EventQueue::I3E_EVQ_RemoveDefaultFromQueue(const std::string& sz_href) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    ModScop std::queue<DefaultBlock> q_tempQueue;
    ModScop bool bol_found = false;
    while (!q_defaultQueue.empty()) 
    {
        ModScop DefaultBlock str_currentEvent = q_defaultQueue.front();
        q_defaultQueue.pop();
        if (str_currentEvent.sz_href == sz_href) 
        {
            bol_found = true;
        } else 
        {
            q_tempQueue.push(str_currentEvent);
        }
    }

    std::swap(q_defaultQueue, q_tempQueue);
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return bol_found;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void EventQueue::I3E_EVQ_ReplaceDefDerCtrlForEvent(const std::string& sz_eventHref, 
                const SE_DefaultDERControl_t& i3e_newDefDerCtrl) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    ModScop std::queue<DefaultBlock> q_updatedQueue;
    while (!q_defaultQueue.empty()) {
        ModScop DefaultBlock str_currentDefault = q_defaultQueue.front();
        q_defaultQueue.pop();

        if (str_currentDefault.sz_href == sz_eventHref) 
        {
            // Replace derCtrl for the specific event
            str_currentDefault.i3e_defDerCtrl = i3e_newDefDerCtrl;
        }
        q_updatedQueue.push(str_currentDefault);
    }
    q_defaultQueue = q_updatedQueue;
    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::vector<std::string> EventQueue::I3E_EVQ_FindMissingEventsInDefaultQueue(const std::vector<std::string>& vec_derCtrlHrefs)
{
    pthread_mutex_lock(&mtx_eventQueueMutex);
    ModScop std::vector<std::string> vec_missingEvents;
    for (const std::string& sz_href : vec_derCtrlHrefs)
    {
        ModScop bool bol_found = false;
        ModScop std::queue<DefaultBlock> q_tempQueue = q_defaultQueue; // Create a temporary queue for iteration
        while (!q_tempQueue.empty())
        {
            ModScop DefaultBlock str_currentEvent = q_tempQueue.front();
            q_tempQueue.pop();
            if (str_currentEvent.sz_href == sz_href)
            {
                bol_found = true;
                break;
            }
        }
        if (!bol_found)
        {
            vec_missingEvents.push_back(sz_href);
        }
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return vec_missingEvents;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::vector<std::string> EventQueue::I3E_EVQ_RemoveNonexistentDefaults(const std::vector<std::string>& vec_defCtrlHrefs) {
    pthread_mutex_lock(&mtx_eventQueueMutex);

    ModScop std::queue<DefaultBlock> q_tempQueue;
    ModScop std::vector<std::string> vec_removedHrefs;

    while (!q_defaultQueue.empty()) {
        DefaultBlock str_currentEvent = q_defaultQueue.front();
        q_defaultQueue.pop();

        // Check if str_currentEvent.href is in defCtrlHrefs
        if (std::find(vec_defCtrlHrefs.begin(), vec_defCtrlHrefs.end(), str_currentEvent.sz_href) != vec_defCtrlHrefs.end()) 
        {
            q_tempQueue.push(str_currentEvent); // If it exists, keep it in the q_tempQueue
        } else 
        {
            vec_removedHrefs.push_back(str_currentEvent.sz_href); // If it doesn't exist, add to removedHrefs
        }
    }

    // Swap the queues
    std::swap(q_defaultQueue, q_tempQueue);
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return vec_removedHrefs; // Returning the list of removed hrefs
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop SE_DefaultDERControl_t* EventQueue::I3E_EVQ_CompareControlTypeFlags(const SE_DefaultDERControl_t& i3e_newDefault) {
    pthread_mutex_lock(&mtx_eventQueueMutex);

    ModScop std::queue<DefaultBlock> q_tempQueue = q_defaultQueue; // Create a temporary queue for iteration
    ModScop SE_DefaultDERControl_t* i3e_overlappingDefault = nullptr;
    const uint32_t u32_exclusionMask = ~SE_rampTms_exists;

    while (!q_tempQueue.empty()) {
        ModScop DefaultBlock str_currentEvent = q_tempQueue.front();
        q_tempQueue.pop();

        // Check for overlap in control type flags, excluding the first flag
        if ((i3e_newDefault.DERControlBase._flags & u32_exclusionMask) & 
            (str_currentEvent.i3e_defDerCtrl.DERControlBase._flags & u32_exclusionMask)) {
            i3e_overlappingDefault = &str_currentEvent.i3e_defDerCtrl;
            break;
        }
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return i3e_overlappingDefault;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void EventQueue::I3E_EVQ_PrintDefaultsInQueue() 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    ModScop std::queue<DefaultBlock> q_tempQueue = q_defaultQueue; // Create a temporary queue for iteration

    while (!q_tempQueue.empty()) 
    {
        ModScop DefaultBlock str_currentEvent = q_tempQueue.front();
        q_tempQueue.pop();

        // Print details of the default block
        std::cout << "DefaultBlock href: " << str_currentEvent.sz_href << std::endl;
        // Print additional details if needed, but excluding event block flags and control status
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::vector<std::string> EventQueue::I3E_EVQ_GetAllHrefsFromDefaultQueue() 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    ModScop std::vector<std::string> vec_hrefs;
    ModScop std::queue<DefaultBlock> q_tempQueue = q_defaultQueue; // Create a temporary queue for iteration

    while (!q_tempQueue.empty()) 
    {
        ModScop DefaultBlock str_currentEvent = q_tempQueue.front();
        q_tempQueue.pop();
        vec_hrefs.push_back(str_currentEvent.sz_href); // Add the href to the vector
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return vec_hrefs;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop DefaultBlock EventQueue::I3E_EVQ_GetDefaultBlockByHref(const std::string& sz_href) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    ModScop std::queue<DefaultBlock> q_tempQueue = q_defaultQueue; // Create a temporary queue for iteration
    ModScop DefaultBlock str_foundBlock;
    ModScop bool bol_blockFound = false;

    while (!q_tempQueue.empty()) 
    {
        ModScop DefaultBlock str_currentEvent = q_tempQueue.front();
        q_tempQueue.pop();

        if (str_currentEvent.sz_href == sz_href) 
        {
            str_foundBlock = str_currentEvent;
            bol_blockFound = true;
            break;
        }
    }

    pthread_mutex_unlock(&mtx_eventQueueMutex);
    if (!bol_blockFound) 
    {
        throw std::runtime_error("DefaultBlock with the specified href not found");
    }
    return str_foundBlock;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void EventQueue::I3E_EVQ_OrderDefaultQueueByPrimacy() 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    // Extract elements from the queue into a vector
    ModScop std::vector<DefaultBlock> vec_tempVector;
    while (!q_defaultQueue.empty()) 
    {
        vec_tempVector.push_back(q_defaultQueue.front());
        q_defaultQueue.pop();
    }

    // Sort the vector by primacy (ascending order)
    std::sort(vec_tempVector.begin(), vec_tempVector.end(), [](const DefaultBlock& a, const DefaultBlock& b) 
    {
        return a.primacy < b.primacy;
    });

    // Reinsert the sorted elements back into the queue
    for (const auto& block : vec_tempVector) 
    {
        q_defaultQueue.push(block);
    }

    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void EventQueue::I3E_EVQ_AddHrefAtIndex(const std::string& sz_href, const std::string& sz_mrid, size_t siz_index) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    if (siz_index < vec_activeControlsHrefs.size()) 
    {
        vec_activeControlsHrefs[siz_index] = sz_href;
        vec_activeControlsMRIDs[siz_index] = sz_mrid;
    } else {
        std::cerr << "Index out of bounds" << std::endl;
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::string EventQueue::I3E_EVQ_GetHrefAtIndex(size_t siz_index) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    ModScop std::string sz_href;
    if (siz_index < vec_activeControlsHrefs.size()) 
    {
        sz_href = vec_activeControlsHrefs[siz_index];
    } else {
        std::cerr << "Index out of bounds" << std::endl;
    }

    pthread_mutex_unlock(&mtx_eventQueueMutex);

    return sz_href;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void EventQueue::I3E_EVQ_AddDefaultHrefAtIndex(const std::string& sz_href, const std::string& sz_mrid, size_t siz_index)  
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    if (siz_index < vec_activeDefaultsHrefs.size()) 
    {
        vec_activeDefaultsHrefs[siz_index] = sz_href;
        vec_activeDefaultsMRIDs[siz_index] = sz_mrid;
    } else {
        std::cerr << "Index out of bounds" << std::endl;
    }

    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::string EventQueue::I3E_EVQ_GetDefaultHrefAtIndex(size_t siz_index) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    ModScop std::string sz_href;
    if (siz_index < vec_activeDefaultsHrefs.size()) 
    {
        sz_href = vec_activeDefaultsHrefs[siz_index];
    } else {
        std::cerr << "Index out of bounds" << std::endl;
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return sz_href;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void EventQueue::I3E_EVQ_AddSettingsHrefAtIndex(const std::string& sz_href, const std::string& sz_mrid, size_t siz_index) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    if (siz_index < vec_activeSettingsHrefs.size()) 
    {
        vec_activeSettingsHrefs[siz_index] = sz_href;
        vec_activeSettingsMRIDs[siz_index] = sz_mrid;
    } else {
        std::cerr << "Index out of bounds" << std::endl;
    }

    pthread_mutex_unlock(&mtx_eventQueueMutex);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::string EventQueue::I3E_EVQ_GetSettingsHrefAtIndex(size_t siz_index) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    ModScop std::string sz_href;
    if (siz_index < vec_activeSettingsHrefs.size()) 
    {
        sz_href = vec_activeSettingsHrefs[siz_index];
    } else {
        std::cerr << "Index out of bounds" << std::endl;
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex);
    return sz_href;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop uint32_t EventQueue::I3E_EVQ_RemoveHrefAndClearFlags(const std::string& sz_href) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    uint32_t u32_ctrlsReset = 0;

    for (size_t i = 0; i < vec_activeControlsHrefs.size(); ++i) 
    {
        if (vec_activeControlsHrefs[i] == sz_href) {
            vec_activeControlsHrefs[i].clear();
            vec_activeControlsMRIDs[i].clear();

            switch (i) 
            {
                case SE_rampTms_hrefPos: 
                {
                    u32_activeControls &= ~SE_rampTms_exists; 
                    u32_ctrlsReset |= SE_rampTms_exists;
                    break;
                }
                case SE_opModWattVar_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModWattVar_exists; 
                    u32_ctrlsReset |= SE_opModWattVar_exists;
                    break;
                }
                case SE_opModWattPF_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModWattPF_exists; 
                    u32_ctrlsReset |= SE_opModWattPF_exists;
                    break;
                }
                case SE_opModVoltWatt_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModVoltWatt_exists; 
                    u32_ctrlsReset |= SE_opModVoltWatt_exists;
                    break;
                }
                case SE_opModVoltVar_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModVoltVar_exists; 
                    u32_ctrlsReset |= SE_opModVoltVar_exists;
                    break;
                }
                case SE_opModTargetW_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModTargetW_exists; 
                    u32_ctrlsReset |= SE_opModTargetW_exists;
                    break;
                }
                case SE_opModTargetVar_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModTargetVar_exists; 
                    u32_ctrlsReset |= SE_opModTargetVar_exists;
                    break;
                }
                case SE_opModMaxLimW_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModMaxLimW_exists; 
                    u32_ctrlsReset |= SE_opModMaxLimW_exists;
                    break;
                }
                case SE_opModLVRTMustTrip_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModLVRTMustTrip_exists; 
                    u32_ctrlsReset |= SE_opModLVRTMustTrip_exists;
                    break;
                }
                case SE_opModLVRTMomentaryCessation_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModLVRTMomentaryCessation_exists; 
                    u32_ctrlsReset |= SE_opModLVRTMomentaryCessation_exists;
                    break;
                }
                case SE_opModLFRTMustTrip_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModLFRTMustTrip_exists; 
                    u32_ctrlsReset |= SE_opModLFRTMustTrip_exists;
                    break;
                }
                case SE_opModHVRTMustTrip_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModHVRTMustTrip_exists; 
                    u32_ctrlsReset |= SE_opModHVRTMustTrip_exists;
                    break;
                }
                case SE_opModHVRTMomentaryCessation_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModHVRTMomentaryCessation_exists; 
                    u32_ctrlsReset |= SE_opModHVRTMomentaryCessation_exists;
                    break;
                }
                case SE_opModHFRTMustTrip_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModHFRTMustTrip_exists; 
                    u32_ctrlsReset |= SE_opModHFRTMustTrip_exists;
                    break;
                }
                case SE_opModFreqWatt_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModFreqWatt_exists; 
                    u32_ctrlsReset |= SE_opModFreqWatt_exists;
                    break;
                }
                case SE_opModFreqDroop_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModFreqDroop_exists; 
                    u32_ctrlsReset |= SE_opModFreqDroop_exists;
                    break;
                }
                case SE_opModFixedW_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModFixedW_exists; 
                    u32_ctrlsReset |= SE_opModFixedW_exists;
                    break;
                }
                case SE_opModFixedVar_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModFixedVar_exists; 
                    u32_ctrlsReset |= SE_opModFixedVar_exists;
                    break;
                }
                case SE_opModEnergize_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModEnergize_exists; 
                    u32_ctrlsReset |= SE_opModEnergize_exists;
                    break;
                }
                case SE_opModConnect_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModConnect_exists; 
                    u32_ctrlsReset |= SE_opModConnect_exists;
                    break;
                }
                case SE_opModHFRTMayTrip_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModHFRTMayTrip_exists; 
                    u32_ctrlsReset |= SE_opModHFRTMayTrip_exists;
                    break;
                }
                case SE_opModHVRTMayTrip_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModHVRTMayTrip_exists; 
                    u32_ctrlsReset |= SE_opModHVRTMayTrip_exists;
                    break;
                }
                case SE_opModLFRTMayTrip_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModLFRTMayTrip_exists; 
                    u32_ctrlsReset |= SE_opModLFRTMayTrip_exists;
                    break;
                }
                case SE_opModLVRTMayTrip_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModLVRTMayTrip_exists; 
                    u32_ctrlsReset |= SE_opModLVRTMayTrip_exists;
                    break;
                }
                case SE_opModFixedPFInjectW_hrefPos:
                { 
                    u32_activeControls &= ~SE_opModFixedPFInjectW_exists; 
                    u32_ctrlsReset |= SE_opModFixedPFInjectW_exists;
                    break;
                }
                case SE_opModFixedPFAbsorbW_hrefPos: 
                {
                    u32_activeControls &= ~SE_opModFixedPFAbsorbW_exists; 
                    u32_ctrlsReset |= SE_opModFixedPFAbsorbW_exists;
                    break;
                }
                default: break;
            }
        }
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex);

    return u32_ctrlsReset;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop uint32_t EventQueue::I3E_EVQ_RemoveDefaultHrefAndClearFlags(const std::string& sz_href) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    uint32_t u32_defaultCtrlsReset = 0;

    for (size_t i = 0; i < vec_activeDefaultsHrefs.size(); ++i) 
    {
        if (vec_activeDefaultsHrefs[i] == sz_href) {
            vec_activeDefaultsHrefs[i].clear();
            vec_activeDefaultsMRIDs[i].clear();

            switch (i) 
            {
                case SE_rampTms_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_rampTms_exists; 
                    u32_defaultCtrlsReset |= SE_rampTms_exists;
                    break;
                }
                case SE_opModWattVar_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModWattVar_exists; 
                    u32_defaultCtrlsReset |= SE_opModWattVar_exists;
                    break;
                }
                case SE_opModWattPF_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModWattPF_exists; 
                    u32_defaultCtrlsReset |= SE_opModWattPF_exists;
                    break;
                }
                case SE_opModVoltWatt_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModVoltWatt_exists; 
                    u32_defaultCtrlsReset |= SE_opModVoltWatt_exists;
                    break;
                }
                case SE_opModVoltVar_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModVoltVar_exists; 
                    u32_defaultCtrlsReset |= SE_opModVoltVar_exists;
                    break;
                }
                case SE_opModTargetW_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModTargetW_exists; 
                    u32_defaultCtrlsReset |= SE_opModTargetW_exists;
                    break;
                }
                case SE_opModTargetVar_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModTargetVar_exists; 
                    u32_defaultCtrlsReset |= SE_opModTargetVar_exists;
                    break;
                }
                case SE_opModMaxLimW_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModMaxLimW_exists; 
                    u32_defaultCtrlsReset |= SE_opModMaxLimW_exists;
                    break;
                }
                case SE_opModLVRTMustTrip_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModLVRTMustTrip_exists; 
                    u32_defaultCtrlsReset |= SE_opModLVRTMustTrip_exists;
                    break;
                }
                case SE_opModLVRTMomentaryCessation_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModLVRTMomentaryCessation_exists; 
                    u32_defaultCtrlsReset |= SE_opModLVRTMomentaryCessation_exists;
                    break;
                }
                case SE_opModLFRTMustTrip_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModLFRTMustTrip_exists; 
                    u32_defaultCtrlsReset |= SE_opModLFRTMustTrip_exists;
                    break;
                }
                case SE_opModHVRTMustTrip_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModHVRTMustTrip_exists; 
                    u32_defaultCtrlsReset |= SE_opModHVRTMustTrip_exists;
                    break;
                }
                case SE_opModHVRTMomentaryCessation_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModHVRTMomentaryCessation_exists; 
                    u32_defaultCtrlsReset |= SE_opModHVRTMomentaryCessation_exists;
                    break;
                }
                case SE_opModHFRTMustTrip_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModHFRTMustTrip_exists; 
                    u32_defaultCtrlsReset |= SE_opModHFRTMustTrip_exists;
                    break;
                }
                case SE_opModFreqWatt_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModFreqWatt_exists; 
                    u32_defaultCtrlsReset |= SE_opModFreqWatt_exists;
                    break;
                }
                case SE_opModFreqDroop_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModFreqDroop_exists; 
                    u32_defaultCtrlsReset |= SE_opModFreqDroop_exists;
                    break;
                }
                case SE_opModFixedW_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModFixedW_exists; 
                    u32_defaultCtrlsReset |= SE_opModFixedW_exists;
                    break;
                }
                case SE_opModFixedVar_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModFixedVar_exists; 
                    u32_defaultCtrlsReset |= SE_opModFixedVar_exists;
                    break;
                }
                case SE_opModEnergize_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModEnergize_exists; 
                    u32_defaultCtrlsReset |= SE_opModEnergize_exists;
                    break;
                }
                case SE_opModConnect_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModConnect_exists; 
                    u32_defaultCtrlsReset |= SE_opModConnect_exists;
                    break;
                }
                case SE_opModHFRTMayTrip_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModHFRTMayTrip_exists; 
                    u32_defaultCtrlsReset |= SE_opModHFRTMayTrip_exists;
                    break;
                }
                case SE_opModHVRTMayTrip_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModHVRTMayTrip_exists; 
                    u32_defaultCtrlsReset |= SE_opModHVRTMayTrip_exists;
                    break;
                }
                case SE_opModLFRTMayTrip_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModLFRTMayTrip_exists; 
                    u32_defaultCtrlsReset |= SE_opModLFRTMayTrip_exists;
                    break;
                }
                case SE_opModLVRTMayTrip_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModLVRTMayTrip_exists; 
                    u32_defaultCtrlsReset |= SE_opModLVRTMayTrip_exists;
                    break;
                }
                case SE_opModFixedPFInjectW_hrefPos:
                { 
                    u32_activeDefaultControls &= ~SE_opModFixedPFInjectW_exists; 
                    u32_defaultCtrlsReset |= SE_opModFixedPFInjectW_exists;
                    break;
                }
                case SE_opModFixedPFAbsorbW_hrefPos: 
                {
                    u32_activeDefaultControls &= ~SE_opModFixedPFAbsorbW_exists; 
                    u32_defaultCtrlsReset |= SE_opModFixedPFAbsorbW_exists;
                    break;
                }
                default: break;
            }
        }
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex);

    return u32_defaultCtrlsReset;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
ModScop uint32_t EventQueue::I3E_EVQ_RemoveSettingsHrefAndClearFlags(const std::string& sz_href) 
{
    pthread_mutex_lock(&mtx_eventQueueMutex);

    uint32_t u32_settingsReset = 0;

    for (size_t i = 0; i < vec_activeSettingsHrefs.size(); ++i) 
    {
        if (vec_activeSettingsHrefs[i] == sz_href) {
            vec_activeSettingsHrefs[i].clear();
            vec_activeSettingsMRIDs[i].clear();

            switch (i) 
            {
                case SE_setGradW_hrefPos: 
                {
                    u32_activeSettings &= ~SE_setGradW_hrefPos; 
                    u32_settingsReset |= SE_setGradW_exists;
                    break;
                }
                case SE_setESDelay_hrefPos: 
                {
                    u32_activeSettings &= ~SE_setESDelay_hrefPos; 
                    u32_settingsReset |= SE_setESDelay_exists;
                    break;
                }
                case SE_setESHighFreq_hrefPos: 
                {
                    u32_activeSettings &= ~SE_setESHighFreq_hrefPos; 
                    u32_settingsReset |= SE_setESHighFreq_exists;
                    break;
                }
                case SE_setESHighVolt_hrefPos: 
                {
                    u32_activeSettings &= ~SE_setESHighVolt_hrefPos; 
                    u32_settingsReset |= SE_setESHighVolt_exists;
                    break;
                }
                case SE_setESLowFreq_hrefPos: 
                {
                    u32_activeSettings &= ~SE_setESLowFreq_hrefPos; 
                    u32_settingsReset |= SE_setESLowFreq_exists;
                    break;
                }
                case SE_setESLowVolt_hrefPos: 
                {
                    u32_activeSettings &= ~SE_setESLowVolt_hrefPos; 
                    u32_settingsReset |= SE_setESLowVolt_exists;
                    break;
                }
                case SE_setESRandomDelay_hrefPos: 
                {
                    u32_activeSettings &= ~SE_setESRandomDelay_hrefPos; 
                    u32_settingsReset |= SE_setESRandomDelay_exists;
                    break;
                }
                case SE_setSoftGradW_hrefPos: 
                {
                    u32_activeSettings &= ~SE_setSoftGradW_hrefPos; 
                    u32_settingsReset |= SE_setSoftGradW_exists;
                    break;
                }
                case SE_setESRampTms_hrefPos: 
                {
                    u32_activeSettings &= ~SE_setESRampTms_hrefPos;
                    u32_settingsReset |= SE_setESRampTms_exists; 
                    break;
                }
                default: break;
            }
        }
    }
    pthread_mutex_unlock(&mtx_eventQueueMutex);

    return u32_settingsReset;
}