#include "lion/log/log.h"

#include "include/I3E_HTTP.hpp"
#include "include/I3E_NSS.hpp"

LOG_MODULE_REGISTER(i3e_http, LOG_LEVEL_INF);

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

/**
 * @brief libCurl write callback function.
 * 
 * @details Registered libCurl callback function for processing or storing data received in a response from a 
 * server during an HTTP request (GET).
 * 
 * @param ptr_contents The data received from the server.
 * @param siz_size The size of each element in the data.
 * @param siz_nmemb The number of elements in the data.
 * @param ptr_userdata The registered buffer to store the data for user application (set with curl_easy_setopt).
 * 
 * @returns Number of bytes that were successfully handled by the callback.
 */
FileScop size_t i3e_http_WriteCallback(void* ptr_contents, size_t siz_size, size_t siz_nmemb, void* ptr_userdata);

/**
 * @brief libCurl read callback function.
 * 
 * @details Registered libCurl callback function for providing data to be uploaded in a request to a server during an 
 * HTTP request (PUT or POST).
 * 
 * @param ptr_contents The data to be transmitted.
 * @param siz_size The size of each element in the data.
 * @param siz_nmemb The number of elements in the data.
 * @param ptr_userdata The registered buffer that contains the data to be uploaded (set with curl_easy_setopt).
 * 
 * @returns Number of bytes that were successfully handled by the callback.
 */
FileScop size_t i3e_http_ReadCallback(void* ptr_contents, size_t siz_size, size_t siz_nmemb, void* ptr_userdata);

/**
 * @brief libCurl header callback function.
 * 
 * @details Registered libCurl callback function for processing or storing header data received in a response from a
 * server during an HTTP request (GET, PUT, or POST).
 * 
 * @param ptr_contents The header data received from the server.
 * @param siz_size The size of each element in the data.
 * @param siz_nmemb The number of elements in the data.
 * @param ptr_userdata The registered buffer to store the header data for user application (set with curl_easy_setopt).
 * 
 * @returns Number of bytes that were successfully handled by the callback.
 */
FileScop size_t i3e_http_HeaderCallback(char* ptr_contents, size_t siz_size, size_t siz_nmemb, std::string* ptr_userdata);

/**
 * @brief Map the response code to an HTTPResponse enum.
 * 
 * @details This function maps the response code to an HTTPResponse enum.
 * 
 * @param code The response code.
 * 
 * @returns The HTTPResponse enum.
 */
FileScop HTTPResponse i3e_http_MapToHTTPResponse(long code);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop const std::unordered_map<long, HTTPResponse> HTTPResponseMap = {
    {200L, HTTPResponse::OK},
    {201L, HTTPResponse::CREATED},
    {202L, HTTPResponse::ACCEPTED},
    {204L, HTTPResponse::NO_CONTENT},
    {301L, HTTPResponse::MOVED_PERMANENTLY},
    {302L, HTTPResponse::FOUND},
    {304L, HTTPResponse::NOT_MODIFIED},
    {400L, HTTPResponse::BAD_REQUEST},
    {401L, HTTPResponse::UNAUTHORIZED},
    {403L, HTTPResponse::FORBIDDEN},
    {404L, HTTPResponse::NOT_FOUND},
    {405L, HTTPResponse::METHOD_NOT_ALLOWED},
    {500L, HTTPResponse::INTERNAL_SERVER_ERROR},
    {502L, HTTPResponse::BAD_GATEWAY},
    {503L, HTTPResponse::SERVICE_UNAVAILABLE},
    {504L, HTTPResponse::GATEWAY_TIMEOUT}
};

FileScop pthread_mutex_t mtx_SSLMutex;

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

// #TODO - could access http body and headers for extra control if we need it

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop int I3E_HTTP_Init(void) 
{
    StackScop int res = pthread_mutex_init(&mtx_SSLMutex, nullptr);
    return res;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop HTTPResponse I3E_HTTP_GET(std::string sz_pIPV4addr, uint16_t u16_portNum, 
                                  std::string sz_pURI, std::string& sz_rxPayload) 
{

    // Lock the mutex to protect cURL calls
    pthread_mutex_lock(&mtx_SSLMutex);

    // Variable to store the HTTP response code
    StackScop long http_code = 0;

    StackScop CURL* ptr_curlCtx = I3E_NSS_GetCURL();
    StackScop HTTPResponse enum_ret;

    std::string sz_sunspecCertC4 = I3E_NSS_GetCERT();
    std::string sz_sunspecPkey = I3E_NSS_GetPKEY();

    if (ptr_curlCtx) 
    {
        StackScop CURLcode enum_res;

        // Construct the full URL
        StackScop std::string sz_url = "https://" + sz_pIPV4addr + ":" + std::to_string(u16_portNum) + sz_pURI;  
        
        // Set the URL
        curl_easy_setopt(ptr_curlCtx, CURLOPT_URL, sz_url.c_str());

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLCERTTYPE, sz_CertType);

        // // Set the callback function to handle the response data
        // curl_easy_setopt(ptr_curlCtx, CURLOPT_WRITEFUNCTION, WriteCallback);

        // // Set the buffer to store the response data
        // curl_easy_setopt(ptr_curlCtx, CURLOPT_WRITEDATA, &readBuffer);

        // Set the callback for the SSL context
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_CTX_DATA, NULL);
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_CTX_FUNCTION, I3E_NSS_sslCtxCallback);

        // Set callback to inspect certificate chain
        curl_easy_setopt(ptr_curlCtx, CURLOPT_CERTINFO, 1L);

        // Enable verbose logging for debugging
        curl_easy_setopt(ptr_curlCtx, CURLOPT_VERBOSE, 1L);

        // Set HTTP version 1.1
        curl_easy_setopt(ptr_curlCtx, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);

        // Set TLS version 1.2
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLVERSION, CURL_SSLVERSION_TLSv1_2);

        // Set certificate verification to verify server
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_VERIFYPEER, 1L);
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_VERIFYHOST, 0L);

        // Enable mutual authentication
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLCERT, sz_sunspecCertC4.c_str());
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLKEY, sz_sunspecPkey.c_str());

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        // Ensure that the correct method is used by disabling others and enabling GET
        curl_easy_setopt(ptr_curlCtx, CURLOPT_NOBODY, 0L);   // Ensure HEAD is disabled
        curl_easy_setopt(ptr_curlCtx, CURLOPT_UPLOAD, 0L);   // Ensure PUT is disabled
        curl_easy_setopt(ptr_curlCtx, CURLOPT_POST, 0L);     // Ensure POST is disabled
        curl_easy_setopt(ptr_curlCtx, CURLOPT_HTTPGET, 1L);  // Enable GET

        // Set timeouts
        curl_easy_setopt(ptr_curlCtx, CURLOPT_TIMEOUT, 10L);             // Timeout for the entire request
        curl_easy_setopt(ptr_curlCtx, CURLOPT_CONNECTTIMEOUT, 5L);       // Timeout for the connection phase

        // Set headers
        StackScop struct curl_slist* ptr_headers = nullptr;
        ptr_headers = curl_slist_append(ptr_headers, HTTP_USERAGENT HTTP_USR);
        ptr_headers = curl_slist_append(ptr_headers, HTTP_SESSION HTTP_SES);
        ptr_headers = curl_slist_append(ptr_headers, HTTP_ACCEPT HTTP_SEPXML);
        ptr_headers = curl_slist_append(ptr_headers, HTTP_CONTENTTYPE HTTP_SEPXML);
        curl_easy_setopt(ptr_curlCtx, CURLOPT_HTTPHEADER, ptr_headers);

        // Set the callback function to handle the response data
        curl_easy_setopt(ptr_curlCtx, CURLOPT_WRITEFUNCTION, i3e_http_WriteCallback);

        // Set the buffer to store the response data
        curl_easy_setopt(ptr_curlCtx, CURLOPT_WRITEDATA, &sz_rxPayload);

        // Set the callback function to handle headers
        StackScop std::string redirectedUrl;
        curl_easy_setopt(ptr_curlCtx, CURLOPT_HEADERFUNCTION, i3e_http_HeaderCallback);
        curl_easy_setopt(ptr_curlCtx, CURLOPT_HEADERDATA, &redirectedUrl);

        // Perform the request
        enum_res = curl_easy_perform(ptr_curlCtx);

        // Check for errors
        if (enum_res != CURLE_OK) 
        {
            LOG_ERR("curl_easy_perform() failed: %s", curl_easy_strerror(enum_res));
            enum_ret = HTTPResponse::INTERNAL_SERVER_ERROR;
        } 
        else 
        {
            // Get the HTTP response code
            curl_easy_getinfo(ptr_curlCtx, CURLINFO_RESPONSE_CODE, &http_code);
            enum_ret = i3e_http_MapToHTTPResponse(http_code);

            if(enum_ret != HTTPResponse::OK) 
            {
                sz_rxPayload.clear();
                if ((enum_ret == HTTPResponse::MOVED_PERMANENTLY || 
                     enum_ret == HTTPResponse::FOUND) && 
                     !redirectedUrl.empty()) 
                {
                    LOG_INF("Redirected URL: %s", redirectedUrl);
                    sz_rxPayload = redirectedUrl;
                }
            }
        }
    }

    // Reset or clear options for the next transfer
    curl_easy_reset(ptr_curlCtx);

    // unlock the mutex to protect cURL calls
    pthread_mutex_unlock(&mtx_SSLMutex);

    return enum_ret;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop HTTPResponse I3E_HTTP_POST(std::string sz_pIPV4addr, uint16_t u16_portNum, 
                                   std::string sz_pURI, std::string& sz_txPayload, std::string& sz_rxPayload)
{

    // Lock the mutex to protect cURL calls
    pthread_mutex_lock(&mtx_SSLMutex);

    StackScop CURL* ptr_curlCtx = I3E_NSS_GetCURL();
    StackScop HTTPResponse enum_ret;

    // Variable to store the HTTP response code
    StackScop long http_code = 0;

    std::string sz_sunspecCertC4 = I3E_NSS_GetCERT();
    std::string sz_sunspecPkey = I3E_NSS_GetPKEY();

    if (ptr_curlCtx) 
    {

        StackScop CURLcode enum_res;

        // Construct the full URL
        StackScop std::string sz_url = "https://" + sz_pIPV4addr + ":" + std::to_string(u16_portNum) + sz_pURI;   
        
        // Set the URL
        curl_easy_setopt(ptr_curlCtx, CURLOPT_URL, sz_url.c_str());

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLCERTTYPE, sz_CertType);

        // // Set the callback function to handle the response data
        // curl_easy_setopt(ptr_curlCtx, CURLOPT_WRITEFUNCTION, WriteCallback);

        // // Set the buffer to store the response data
        // curl_easy_setopt(ptr_curlCtx, CURLOPT_WRITEDATA, &readBuffer);

        // Set the callback for the SSL context
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_CTX_DATA, NULL);
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_CTX_FUNCTION, I3E_NSS_sslCtxCallback);

        // Set callback to inspect certificate chain
        curl_easy_setopt(ptr_curlCtx, CURLOPT_CERTINFO, 1L);

        // Enable verbose logging for debugging
        curl_easy_setopt(ptr_curlCtx, CURLOPT_VERBOSE, 1L);

        // Set HTTP version 1.1
        curl_easy_setopt(ptr_curlCtx, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);

        // Set TLS version 1.2
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLVERSION, CURL_SSLVERSION_TLSv1_2);

        // Set certificate verification to verify server
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_VERIFYPEER, 1L);
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_VERIFYHOST, 0L);

        // Enable mutual authentication
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLCERT, sz_sunspecCertC4.c_str());
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLKEY, sz_sunspecPkey.c_str());

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        // Set the data to be posted
        curl_easy_setopt(ptr_curlCtx, CURLOPT_POSTFIELDS, sz_txPayload.c_str());

        // Ensure that the correct method is used by disabling others and enabling POST
        curl_easy_setopt(ptr_curlCtx, CURLOPT_NOBODY, 0L);   // Ensure HEAD is disabled
        curl_easy_setopt(ptr_curlCtx, CURLOPT_UPLOAD, 0L);   // Ensure PUT is disabled
        curl_easy_setopt(ptr_curlCtx, CURLOPT_HTTPGET, 0L);  // Ensure GET is disabled
        curl_easy_setopt(ptr_curlCtx, CURLOPT_POST, 1L);     // Enable POST

        // Set timeouts
        curl_easy_setopt(ptr_curlCtx, CURLOPT_TIMEOUT, 10L);             // Timeout for the entire request
        curl_easy_setopt(ptr_curlCtx, CURLOPT_CONNECTTIMEOUT, 5L);      // Timeout for the connection phase

        // Set headers
        StackScop struct curl_slist* ptr_headers = nullptr;
        ptr_headers = curl_slist_append(ptr_headers, HTTP_USERAGENT HTTP_USR);
        ptr_headers = curl_slist_append(ptr_headers, HTTP_SESSION HTTP_SES);
        ptr_headers = curl_slist_append(ptr_headers, HTTP_ACCEPT HTTP_SEPXML);
        ptr_headers = curl_slist_append(ptr_headers, HTTP_CONTENTTYPE HTTP_SEPXML);
        curl_easy_setopt(ptr_curlCtx, CURLOPT_HTTPHEADER, ptr_headers);

        StackScop std::string sz_dataRx;

        // Set the callback function to handle the response data
        curl_easy_setopt(ptr_curlCtx, CURLOPT_WRITEFUNCTION, i3e_http_WriteCallback);

        // Set the buffer to store the response data
        curl_easy_setopt(ptr_curlCtx, CURLOPT_WRITEDATA, &sz_dataRx);

        // Set the callback function to handle the header data
        curl_easy_setopt(ptr_curlCtx, CURLOPT_HEADERFUNCTION, i3e_http_HeaderCallback);
        curl_easy_setopt(ptr_curlCtx, CURLOPT_HEADERDATA, &sz_rxPayload);

        // Perform the request
        enum_res = curl_easy_perform(ptr_curlCtx);

        // Clean up headers
        if (ptr_headers) 
        {
            curl_slist_free_all(ptr_headers);
        }

        // Check for errors
        if (enum_res != CURLE_OK) 
        {
            LOG_ERR("curl_easy_perform() failed: %s", curl_easy_strerror(enum_res));
            enum_ret = HTTPResponse::INTERNAL_SERVER_ERROR;
        } 
        else 
        {
            // Get the HTTP response code
            curl_easy_getinfo(ptr_curlCtx, CURLINFO_RESPONSE_CODE, &http_code);
            enum_ret = i3e_http_MapToHTTPResponse(http_code);
        }
    }

    // Reset or clear options for the next transfer
    curl_easy_reset(ptr_curlCtx);

    // unlock the mutex to protect cURL calls
    pthread_mutex_unlock(&mtx_SSLMutex);

    return enum_ret;
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop HTTPResponse I3E_HTTP_PUT(std::string sz_pIPV4addr, uint16_t u16_portNum, 
                                  std::string sz_pURI, std::string& sz_txPayload, std::string& sz_rxPayload)
{

    // Lock the mutex to protect cURL calls
    pthread_mutex_lock(&mtx_SSLMutex);

    StackScop CURL* ptr_curlCtx = I3E_NSS_GetCURL();
    StackScop HTTPResponse enum_ret;

    // Variable to store the HTTP response code
    StackScop long http_code = 0;

    std::string sz_sunspecCertC4 = I3E_NSS_GetCERT();
    std::string sz_sunspecPkey = I3E_NSS_GetPKEY();

    if (ptr_curlCtx) 
    {

        StackScop CURLcode enum_res;

        // Construct the full URL
        StackScop std::string sz_url = "https://" + sz_pIPV4addr + ":" + std::to_string(u16_portNum) + sz_pURI;  

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLCERTTYPE, sz_CertType);

        // // Set the callback function to handle the response data
        // curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, WriteCallback);

        // // Set the buffer to store the response data
        // curl_easy_setopt(curl, CURLOPT_WRITEDATA, &readBuffer);

        // Set the callback for the SSL context
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_CTX_DATA, NULL);
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_CTX_FUNCTION, I3E_NSS_sslCtxCallback);

        // Set callback to inspect certificate chain
        curl_easy_setopt(ptr_curlCtx, CURLOPT_CERTINFO, 1L);

        // Enable verbose logging for debugging
        curl_easy_setopt(ptr_curlCtx, CURLOPT_VERBOSE, 1L);

        // Set HTTP version 1.1
        curl_easy_setopt(ptr_curlCtx, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);

        // Set TLS version 1.2
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLVERSION, CURL_SSLVERSION_TLSv1_2);

        // Set certificate verification to verify server
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_VERIFYPEER, 1L);
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_VERIFYHOST, 0L);

        // Enable mutual authentication
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLCERT, sz_sunspecCertC4.c_str());
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLKEY, sz_sunspecPkey.c_str());

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        // Set the URL
        curl_easy_setopt(ptr_curlCtx, CURLOPT_URL, sz_url.c_str());

        // Set the read callback function to provide the data
        curl_easy_setopt(ptr_curlCtx, CURLOPT_READFUNCTION, i3e_http_ReadCallback);

        // Set the buffer containing the data to be uploaded
        curl_easy_setopt(ptr_curlCtx, CURLOPT_READDATA, &sz_txPayload);

        // Set the size of the data to be uploaded
        curl_easy_setopt(ptr_curlCtx, CURLOPT_INFILESIZE_LARGE, (curl_off_t)sz_txPayload.size());

        // Ensure that the correct method is used by disabling others and enabling PUT
        curl_easy_setopt(ptr_curlCtx, CURLOPT_NOBODY, 0L);   // Ensure HEAD is disabled
        curl_easy_setopt(ptr_curlCtx, CURLOPT_HTTPGET, 0L);  // Ensure GET is disabled
        curl_easy_setopt(ptr_curlCtx, CURLOPT_POST, 0L);     // Ensure POST is disabled
        curl_easy_setopt(ptr_curlCtx, CURLOPT_UPLOAD, 1L);   // Enable PUT

        // Set timeouts
        curl_easy_setopt(ptr_curlCtx, CURLOPT_TIMEOUT, 10L);             // Timeout for the entire request
        curl_easy_setopt(ptr_curlCtx, CURLOPT_CONNECTTIMEOUT, 5L);       // Timeout for the connection phase

        // Set headers
        struct curl_slist* ptr_headers = nullptr;
        ptr_headers = curl_slist_append(ptr_headers, HTTP_USERAGENT HTTP_USR);
        ptr_headers = curl_slist_append(ptr_headers, HTTP_SESSION HTTP_SES);
        ptr_headers = curl_slist_append(ptr_headers, HTTP_ACCEPT HTTP_SEPXML);
        ptr_headers = curl_slist_append(ptr_headers, HTTP_CONTENTTYPE HTTP_SEPXML);

        curl_easy_setopt(ptr_curlCtx, CURLOPT_HTTPHEADER, ptr_headers);

        // Set the callback function to handle the response data
        curl_easy_setopt(ptr_curlCtx, CURLOPT_WRITEFUNCTION, i3e_http_WriteCallback);

        // Set the buffer to store the response data
        curl_easy_setopt(ptr_curlCtx, CURLOPT_WRITEDATA, &sz_rxPayload);
            
        // Perform the request
        enum_res = curl_easy_perform(ptr_curlCtx);

        // Clean up headers
        if (ptr_headers) 
        {
            curl_slist_free_all(ptr_headers);
        }

        // Check for errors
        if (enum_res != CURLE_OK) 
        {
            LOG_ERR("curl_easy_perform() failed: %s", curl_easy_strerror(enum_res));
            enum_ret = HTTPResponse::INTERNAL_SERVER_ERROR;
        } 
        else 
        {
            // Get the HTTP response code
            curl_easy_getinfo(ptr_curlCtx, CURLINFO_RESPONSE_CODE, &http_code);
            enum_ret = i3e_http_MapToHTTPResponse(http_code);
        }
    }

    // Reset or clear options for the next transfer
    curl_easy_reset(ptr_curlCtx);    

    // unlock the mutex to protect cURL calls
    pthread_mutex_unlock(&mtx_SSLMutex);

    return enum_ret;

}

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop HTTPResponse i3e_http_MapToHTTPResponse(long code) {
    StackScop auto it = HTTPResponseMap.find(code);
    if (it != HTTPResponseMap.end()) 
    {
        return it->second;
    } 
    else 
    {
        // Default to an error code if the input code is not found
        return HTTPResponse::BAD_REQUEST;
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop size_t i3e_http_WriteCallback(void* ptr_contents, size_t siz_size, size_t siz_nmemb, void* ptr_userdata)
{
    ((std::string*)ptr_userdata)->append((char*)ptr_contents, siz_size * siz_nmemb);
    return siz_size * siz_nmemb;
}
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop size_t i3e_http_ReadCallback(void* ptr_contents, size_t siz_size, size_t siz_nmemb, void* ptr_userdata)
{
    StackScop std::string* sz_uploadData = static_cast<std::string*>(ptr_userdata);
    StackScop size_t siz_maxCopySize = siz_size * siz_nmemb;
    if (sz_uploadData->size() < siz_maxCopySize) {
        siz_maxCopySize = sz_uploadData->size();
    }
    memcpy(ptr_contents, sz_uploadData->c_str(), siz_maxCopySize);
    sz_uploadData->erase(0, siz_maxCopySize);
    return siz_maxCopySize;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop size_t i3e_http_HeaderCallback(char* ptr_contents, size_t siz_size, size_t siz_nmemb, std::string* ptr_userdata)
{
    StackScop size_t siz_totalSize = siz_size * siz_nmemb;
    StackScop std::string sz_header(ptr_contents, siz_totalSize);
    
    // Look for the "Location" header
    StackScop std::string sz_locationHeader = HTTP_LOCATION;
    if (sz_header.find(sz_locationHeader) != std::string::npos) {
        *ptr_userdata = sz_header.substr(sz_locationHeader.size());
        // Remove any trailing CRLF
        ptr_userdata->erase(ptr_userdata->find_last_not_of("\r\n") + 1);
    }

    return siz_totalSize;
}


