#include "lion/log/log.h"
#include "include/I3E_NSS.hpp"

#include <openssl/ssl.h>
#include <iostream>
#include <ArduinoJson.hpp>
#include <filesystem>


LOG_MODULE_REGISTER(i3e_nss, LOG_LEVEL_INF);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define TLS_CHAIN_DEPTH 4

#define I3E_APPLICATION_DATA_DIR        "/data/var/lib/lion-i3e/"
#define I3E_CFG_FILE_PATH   I3E_APPLICATION_DATA_DIR "config/i3e_configuration.json"

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

/**
 * @brief Load the I3E configuration file.
 * 
 * @details This function loads the I3E configuration file and parses the contents into the relevant variables.
 * 
 * @returns An integer value indicating the status of the configuration file loading process.
 *    0: Configuration file loaded successfully.
 *    Non-zero: Configuration file loading failed.
 */
FileScop int i3e_nss_loadCfg(void);

/**
 * @brief Get the SHA256 fingerprint of a certificate file.
 * 
 * @details This function reads the certificate file at the specified path and calculates the SHA256 fingerprint.
 * 
 * @param sz_pemFilePath The path to the certificate file.
 * 
 * @returns A string of the SHA256 fingerprint of the certificate file.
 */
FileScop std::string i3e_nss_CertSHA256fp(const std::string& sz_pemFilePath);

/**
 * @brief Convert the SHA256 fingerprint to an lFDI and sFDI format for the IEEE 2030.5 protocol.
 * 
 * @details This function converts the previously calculated SHA256 fingerprint to an lFDI and sFDI format for enahnced
 * connection security. Refer to the IEEE 2030.5 specification for more information on formatting.
 * 
 * @param sz_fingerprint The SHA256 ref_fingerprint of the certificate file.
 */
FileScop void i3e_nss_CertlFDIsFDI(const std::string& sz_fingerprint, SE_HexBinary160_t* i3e_lFDI, SE_SFDIType_t* i3e_sFDI);

/**
 * @brief Custom certificate verification callback.
 * 
 * @details This callback verifies that the PKI certificate provided by the server is valid, trusted, and conforms 
 * to the IEEE 2030.5 specification.
 * 
 * @param ptr_ctx The X509 certificate store context pointer.
 * @param preverifyOk The pre-verification status of the certificate.
 * 
 * @returns An integer value indicating the status of the certificate verification process. This return value is 
 * handled internally by the openSSL library. 
 *  0 indicates a failure in the certificate verification process.
 *  1 indicates a successful certificate verification process.
 */
FileScop int i3e_nss_ChkCertCallback(int preverifyOk, X509_STORE_CTX *ptr_ctx);

/**
 * @brief Check the certificate name constraints.
 * 
 * @details This function checks the name constraints of the certificate.
 * 
 * @param ptr_cert The certificate to check.
 * 
 * @returns A boolean indicating the validity of the certificate name constraints.
 */
FileScop bool i3e_nss_ChkCertNameConstraints(X509 *ptr_cert);

/**
 * @brief Check the certificate policy mappings.
 * 
 * @details This function checks the policy mappings of the certificate.
 * 
 * @param ptr_cert The certificate to check.
 * 
 * @returns A boolean indicating the validity of the certificate policy mappings.
 */
FileScop bool i3e_nss_ChkCertPolicyMappings(X509 *ptr_cert);
    
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// Certificate & Context management 
FileScop CURL* ptr_curlCtx = nullptr; 
FileScop bool bol_certValid = true;

// End Device Registration Values 
FileScop std::string sz_sha256Fingerprint;
FileScop SE_SFDIType_t i3e_sFDIdevice;
FileScop SE_HexBinary160_t i3e_lFDIdevice;
FileScop SE_PINType_t i3e_pINdevice = 0; 
FileScop std::string sz_ipv4Addr = "";
FileScop uint16_t u16_portNum = 0;
FileScop std::string sz_URI;

FileScop std::string sz_SunspecRootCA;
FileScop std::string sz_SunspecCertC4;
FileScop std::string sz_SunspecPkey;

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

ModScop int I3E_NSS_Init(void)
{

    StackScop bol_certValid = true;
    StackScop CURLcode enum_curlcode;

    // Load the I3E configuration JSON file and parse out into the relevant variables
    if(i3e_nss_loadCfg() != 0)
    {
        LOG_ERR("Failed to load the I3E json configuration file");
        return -EXIT_FAILURE;
    }

    // Process the certificate through the SHA256 to get the fingerprint
    sz_sha256Fingerprint = i3e_nss_CertSHA256fp(sz_SunspecCertC4);
    LOG_INF("Client Cert SHA256 Fingerprint: %s", sz_sha256Fingerprint.c_str());
    SE_HexBinary160_t i3e_lFDI = {0};
    SE_SFDIType_t i3e_sFDI = 0;
    i3e_nss_CertlFDIsFDI(sz_sha256Fingerprint, &i3e_lFDI, &i3e_sFDI);
    memcpy(&i3e_lFDIdevice, &i3e_lFDI, sizeof(SE_HexBinary160_t));
    memcpy(&i3e_sFDIdevice, &i3e_sFDI, sizeof(SE_SFDIType_t));

    // Initialize CURL globally
    enum_curlcode = curl_global_init(CURL_GLOBAL_DEFAULT);
    if (enum_curlcode != CURLE_OK) {
        LOG_ERR("curl_global_init() failed: %s", curl_easy_strerror(enum_curlcode));
        return -EXIT_FAILURE;
    }
    
    // Initialize CURL easy handle
    ptr_curlCtx = curl_easy_init();
    if (!ptr_curlCtx) {
        LOG_ERR("curl_easy_init() failed");
        curl_global_cleanup();
        return -EXIT_FAILURE;
    }

    if (ptr_curlCtx)
    {
        // Set PEM filetype for certificates
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLCERTTYPE, sz_CertType);

        // Set the callback for the SSL context
        enum_curlcode = curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_CTX_DATA, NULL);
        if (enum_curlcode != CURLE_OK) {
            LOG_ERR("curl_easy_setopt(CURLOPT_SSL_CTX_DATA) failed: %s", curl_easy_strerror(enum_curlcode));
            curl_easy_cleanup(ptr_curlCtx);
            curl_global_cleanup();
            return -EXIT_FAILURE;
        }

        enum_curlcode = curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_CTX_FUNCTION, I3E_NSS_sslCtxCallback);
        if (enum_curlcode != CURLE_OK) {
            LOG_ERR("curl_easy_setopt(CURLOPT_SSL_CTX_FUNCTION) failed: %s", curl_easy_strerror(enum_curlcode));
            curl_easy_cleanup(ptr_curlCtx);
            curl_global_cleanup();
            return -EXIT_FAILURE;
        }

        // Set callback to inspect certificate chain
        curl_easy_setopt(ptr_curlCtx, CURLOPT_CERTINFO, 1L);

        // Enable verbose logging for debugging
        curl_easy_setopt(ptr_curlCtx, CURLOPT_VERBOSE, 1L);

        // Set HTTP version 1.1
        curl_easy_setopt(ptr_curlCtx, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);

        // Set TLS version 1.2
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLVERSION, CURL_SSLVERSION_TLSv1_2);

        // Set HTTP GET method explicitly (optional as it's the default)
        curl_easy_setopt(ptr_curlCtx, CURLOPT_HTTPGET, 1L);

        // Set certificate verification to verify server
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_VERIFYPEER, 1L);
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSL_VERIFYHOST, 0L);

        // Enable mutual authentication
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLCERT, sz_SunspecCertC4.c_str());
        curl_easy_setopt(ptr_curlCtx, CURLOPT_SSLKEY, sz_SunspecPkey.c_str());

    }

    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop int I3E_NSS_CalculateIndicators(SE_HexBinary160_t *i3e_lFDI, SE_SFDIType_t *i3e_sFDI)
{

    StackScop ArduinoJson::StaticJsonDocument<1024> jd_i3ecfg;

    // Retrieve the Translation Map
    StackScop std::fstream fs_i3ecfgFileStream(I3E_CFG_FILE_PATH, std::fstream::in);
    StackScop ArduinoJson::DeserializationError err_DeserializationSt = 
                                ArduinoJson::deserializeJson(jd_i3ecfg, fs_i3ecfgFileStream);

    // Attempt deserialization
    if (err_DeserializationSt != ArduinoJson::DeserializationError::Ok)
    {
        LOG_ERR("CST_Init() Failed due to DeserializationError: %s", err_DeserializationSt.c_str());
        return -EXIT_FAILURE;
    }

    // Get the certificate path from the configuration file
    StackScop std::string sz_cert = jd_i3ecfg["pki"]["certPath"].as<std::string>();
    if(sz_cert.empty())
    {
        LOG_ERR("I3E configuration file loading failure - please check for valid PKI and server settings");
        return -EXIT_FAILURE;
    }

    // Process the certificate through the SHA256 to get the fingerprint
    StackScop std::string sz_sha256fp = i3e_nss_CertSHA256fp(sz_cert);
    LOG_INF("Client Cert SHA256 Fingerprint: %s", sz_sha256fp.c_str());
    i3e_nss_CertlFDIsFDI(sz_sha256fp, i3e_lFDI, i3e_sFDI);
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop int I3E_NSS_CheckPKI(void)
{
    StackScop ArduinoJson::StaticJsonDocument<1024> jd_i3ecfg;

    // Retrieve the Translation Map
    StackScop std::fstream fs_i3ecfgFileStream(I3E_CFG_FILE_PATH, std::fstream::in);
    StackScop ArduinoJson::DeserializationError err_DeserializationSt = 
                                ArduinoJson::deserializeJson(jd_i3ecfg, fs_i3ecfgFileStream);

    // Attempt deserialization
    if (err_DeserializationSt != ArduinoJson::DeserializationError::Ok)
    {
        LOG_ERR("CST_Init() Failed due to DeserializationError: %s", err_DeserializationSt.c_str());
        return -EXIT_FAILURE;
    }

    // Get the certificate paths from the configuration file
    StackScop std::string sz_cert = jd_i3ecfg["pki"]["certPath"].as<std::string>();
    StackScop std::string sz_ca = jd_i3ecfg["pki"]["CAPath"].as<std::string>();
    StackScop std::string sz_pkey = jd_i3ecfg["pki"]["keyPath"].as<std::string>();

    // If all PKI paths have an entry, check for the files in the FS
    if(!sz_cert.empty() && !sz_ca.empty() && !sz_pkey.empty())
    {
        // Check if files exist
        if (std::filesystem::exists(sz_cert) && std::filesystem::exists(sz_ca) && std::filesystem::exists(sz_pkey)) 
        {
            LOG_INF("All PKI files exist.");
            return -EALREADY;
        } 
        else 
        {
            if (!std::filesystem::exists(sz_cert)) {
                LOG_ERR("Certificate file not found: %s", sz_cert.c_str());
            }
            if (!std::filesystem::exists(sz_ca)) {
                LOG_ERR("CA file not found: %s", sz_ca.c_str());
            }
            if (!std::filesystem::exists(sz_pkey)) {
                LOG_ERR("Private key file not found: %s", sz_pkey.c_str());
            }
            return -EXIT_FAILURE;
        }
    }

    // If no PKI paths are set, generate new PKI certificates
    if(sz_cert.empty() && sz_ca.empty() && sz_pkey.empty())
    {
        LOG_INF("PKI Paths are empty, generate new PKI certificates");
        return EXIT_SUCCESS;
    }

    // Shouldn't reach this but return failure if it does (would only happen if not all PKI paths are set)
    return EXIT_FAILURE;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop int I3E_NSS_GeneratePKI(void)
{
    // #TODO - Implement PKI generation
    // Call the Lion AWS PKI portal
    // Download PKI 
    // Parse and scan for PKI (CA, Cert, Key)
    // Store locations in config file
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop SE_SFDIType_t I3E_NSS_GetSFDI(void) 
{
    return i3e_sFDIdevice;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop SE_HexBinary160_t* I3E_NSS_GetLFDI(void) 
{
    return &i3e_lFDIdevice;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop SE_PINType_t I3E_NSS_GetPIN(void) 
{
    return i3e_pINdevice;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop CURL* I3E_NSS_GetCURL(void) 
{
    return ptr_curlCtx;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop bool I3E_NSS_GetCertValid(void) 
{
    return bol_certValid;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::string I3E_NSS_GetIPV4(void)
{
    return sz_ipv4Addr;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop uint16_t I3E_NSS_GetPORT(void)
{
    return u16_portNum;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::string I3E_NSS_GetPKEY(void)
{
    return sz_SunspecPkey;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::string I3E_NSS_GetCERT(void)
{
    return sz_SunspecCertC4;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::string I3E_NSS_GetCA(void)
{
    return sz_SunspecRootCA;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop std::string I3E_NSS_GetURI(void)
{
    return sz_URI;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void I3E_NSS_SetURI(std::string sz_uri)
{
    sz_URI = sz_uri;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop CURLcode I3E_NSS_sslCtxCallback(CURL* ptr_curl, void* ptr_sslctx, void* ptr_parm)
{

    StackScop SSL_CTX* ptr_ctx = static_cast<SSL_CTX*>(ptr_sslctx);

    // Set supported elliptic curves (secp256r1)
    SSL_CTX_set1_curves_list(ptr_ctx, sz_TLSElipCurve);

    // Load CA certificates
    if (!SSL_CTX_load_verify_locations(ptr_ctx, sz_SunspecRootCA.c_str(), NULL))
    {
        LOG_ERR("Error loading CA certificates.");
        return CURLE_SSL_CERTPROBLEM;
    }

    // Set verification depth
    SSL_CTX_set_verify_depth(ptr_ctx, TLS_CHAIN_DEPTH);

    // Load client certificate and chain
    if (SSL_CTX_use_certificate_chain_file(ptr_ctx, sz_SunspecCertC4.c_str()) <= 0)
    {
        LOG_ERR("Error loading client certificate chain.");
        return CURLE_SSL_CERTPROBLEM;
    }

    // Load client private key
    if (SSL_CTX_use_PrivateKey_file(ptr_ctx, sz_SunspecPkey.c_str(), SSL_FILETYPE_PEM) <= 0)
    {
        LOG_ERR("Error loading client private key.");
        return CURLE_SSL_CERTPROBLEM;
    }

    // Verify client private key
    if (!SSL_CTX_check_private_key(ptr_ctx))
    {
        LOG_ERR("Client private key verification failed.");
        return CURLE_SSL_CERTPROBLEM;
    }

    // Set custom cipher suites (optional)
    if (SSL_CTX_set_cipher_list(ptr_ctx, sz_TLSCipherFilter) <= 0)
    {
        LOG_ERR("Error setting cipher list.");
        return CURLE_SSL_CIPHER;
    }

    // Set the custom verification callback
    SSL_CTX_set_verify(ptr_ctx, SSL_VERIFY_PEER, i3e_nss_ChkCertCallback);

    return CURLE_OK;
}


/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop int i3e_nss_loadCfg(void)
{
    StackScop ArduinoJson::StaticJsonDocument<1024> jd_i3ecfg;

    // Retrieve the Translation Map
    StackScop std::fstream fs_i3ecfgFileStream(I3E_CFG_FILE_PATH, std::fstream::in);
    StackScop ArduinoJson::DeserializationError err_DeserializationSt = 
                                ArduinoJson::deserializeJson(jd_i3ecfg, fs_i3ecfgFileStream);
    if (err_DeserializationSt != ArduinoJson::DeserializationError::Ok)
    {
        LOG_ERR("CST_Init() Failed due to DeserializationError: %s", err_DeserializationSt.c_str());
        return -EXIT_FAILURE;
    }

    sz_ipv4Addr = jd_i3ecfg["server"]["ipv4"].as<std::string>();
    u16_portNum = jd_i3ecfg["server"]["port"].as<uint16_t>();
    sz_URI = jd_i3ecfg["server"]["uri"].as<std::string>();

    sz_SunspecRootCA = jd_i3ecfg["pki"]["CAPath"].as<std::string>();
    sz_SunspecCertC4 = jd_i3ecfg["pki"]["certPath"].as<std::string>();
    sz_SunspecPkey = jd_i3ecfg["pki"]["keyPath"].as<std::string>();

    i3e_pINdevice = jd_i3ecfg["security"]["pin"].as<SE_UInt32_t>();

    if(sz_ipv4Addr.empty() || u16_portNum == 0 || sz_URI.empty() || 
        sz_SunspecRootCA.empty() || sz_SunspecCertC4.empty() || sz_SunspecPkey.empty())
    {
        LOG_ERR("I3E configuration file loading failure - please check for valid PKI and server settings");
        return -EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// // Function to check if a critical extension is supported
// bool IsSupportedCriticalExtension(X509_EXTENSION *ext)
// {
//     // List of unsupported critical extensions
//     std::vector<std::string> unsupportedExtensions = {
//         "2.5.29.32", // Certificate Policies
//         // Add other unsupported critical extensions as needed
//     };

//     // Get the extension name
//     ASN1_OBJECT *obj = X509_EXTENSION_get_object(ext);
//     if (obj)
//     {
//         char ext_name[256];
//         OBJ_obj2txt(ext_name, sizeof(ext_name), obj, 1);

//         // Check if the extension is critical and if it's unsupported
//         if (X509_EXTENSION_get_critical(ext) &&
//             std::find(unsupportedExtensions.begin(), unsupportedExtensions.end(), ext_name) != unsupportedExtensions.end())
//         {
//             std::cout << "Unsupported critical extension detected: " << ext_name << ". Certificate must be rejected." << std::endl;
//             return false;
//         }
//     }

//     return true;
// }

// // Function to check if the certificate contains unsupported critical extensions
// bool CheckUnsupportedCriticalExtensions(X509 *cert)
// {
//     // Get the number of extensions
//     int ext_count = X509_get_ext_count(cert);
//     for (int i = 0; i < ext_count; ++i)
//     {
//         X509_EXTENSION *ext = X509_get_ext(cert, i);
//         if (ext && X509_EXTENSION_get_critical(ext))
//         {
//             // Check if the critical extension is supported
//             if (!IsSupportedCriticalExtension(ext))
//             {
//                 return false;
//             }
//         }
//     }
//     return true;
// }


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// Function to check if the certificate contains name-constraints
FileScop bool i3e_nss_ChkCertNameConstraints(X509 *ptr_cert)
{
    // Check if the certificate contains name-constraints extension
    StackScop int ext_nid = NID_name_constraints;
    StackScop int ext_index = X509_get_ext_by_NID(ptr_cert, ext_nid, -1);
    if (ext_index != -1)
    {
        LOG_ERR("Certificate contains name-constraints extension. Certificate must be rejected.");
        return true;
    }
    return false;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// Function to check if the certificate contains policy-mappings
FileScop bool i3e_nss_ChkCertPolicyMappings(X509 *ptr_cert)
{
    // Check if the certificate contains policy-mappings extension
    StackScop int ext_nid = NID_policy_mappings;
    StackScop int ext_index = X509_get_ext_by_NID(ptr_cert, ext_nid, -1);
    if (ext_index != -1)
    {
        LOG_ERR("Certificate contains policy-mappings extension. Certificate must be rejected.");
        return true;
    }
    return false;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// Custom verification function
static int i3e_nss_ChkCertCallback(int preverify_ok, X509_STORE_CTX *ptr_ctx)
{
    // Get the depth of the current certificate in the chain
    StackScop int depth = X509_STORE_CTX_get_error_depth(ptr_ctx);

    // Get the error code (if any)
    StackScop int err = X509_STORE_CTX_get_error(ptr_ctx);

    // If pre-verification failed, print the error message and return 0
    if (!preverify_ok)
    {
        LOG_ERR("Certificate verification error at depth %d: %s", depth, X509_verify_cert_error_string(err));
        bol_certValid = false;
        return 0;
    }

    // Get the current certificate in the chain
    StackScop X509 *ptr_cert = X509_STORE_CTX_get_current_cert(ptr_ctx);

    // Perform custom checks on the current certificate
    // For example, you can print certificate details
    StackScop char arr_u8_subjectName[256];
    X509_NAME_oneline(X509_get_subject_name(ptr_cert), arr_u8_subjectName, sizeof(arr_u8_subjectName));

    // Print critical extensions
    StackScop int ext_count = X509_get_ext_count(ptr_cert);
    for (int j = 0; j < ext_count; ++j)
    {
        StackScop X509_EXTENSION *ptr_ext = X509_get_ext(ptr_cert, j);
        if (X509_EXTENSION_get_critical(ptr_ext))
        {
            // Get the extension name
            StackScop char arr_u8_extName[256];
            OBJ_obj2txt(arr_u8_extName, sizeof(arr_u8_extName), X509_EXTENSION_get_object(ptr_ext), 1);

            // // Check for unsupported critical extensions
            // if (!CheckUnsupportedCriticalExtensions(cert))
            // {
            //     // Unsupported critical extensions detected, reject the certificate
            //     return 0;
            // }

            // Check for name-constraints
            if (i3e_nss_ChkCertNameConstraints(ptr_cert))
            {
                // Name-constraints detected, reject the certificate
                bol_certValid = false;
                return 0;
            }

            // Check for policy-mappings
            if (i3e_nss_ChkCertPolicyMappings(ptr_cert))
            {
                // Policy-mappings detected, reject the certificate
                bol_certValid = false;
                return 0;
            }
        }
    }
    return 1;
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop std::string i3e_nss_CertSHA256fp(const std::string& sz_pemFilePath) {

    StackScop std::string sz_fingerprint;

    // Read the .pem file into memory
    StackScop std::ifstream file(sz_pemFilePath, std::ios::binary);
    if (!file) {
        LOG_ERR("Error opening file: %s", sz_pemFilePath.c_str());
        return sz_fingerprint;
    }

    StackScop std::stringstream buffer;
    buffer << file.rdbuf();
    StackScop std::string sz_pemData = buffer.str();

    // Decode base64 data
    StackScop BIO* ptr_bio = BIO_new_mem_buf(sz_pemData.c_str(), sz_pemData.length());
    StackScop BIO* ptr_b64 = BIO_new(BIO_f_base64());
    ptr_bio = BIO_push(ptr_b64, ptr_bio);

    StackScop std::string sz_decodedData;
    StackScop char arr_u8_inbuf[512];
    StackScop int inlen;
    while ((inlen = BIO_read(ptr_bio, arr_u8_inbuf, sizeof(arr_u8_inbuf))) > 0) {
        sz_decodedData.append(arr_u8_inbuf, inlen);
    }

    BIO_free_all(ptr_bio);

    // Calculate SHA256 hash
    StackScop unsigned char arr_u8_hash[SHA256_DIGEST_LENGTH];
    SHA256((const unsigned char*)sz_decodedData.c_str(), sz_decodedData.length(), arr_u8_hash);

    // Convert hash to hexadecimal string (fingerprint)
    StackScop std::stringstream ss;
    for (int i = 0; i < SHA256_DIGEST_LENGTH; i++) {
        ss << std::hex << std::setw(2) << std::setfill('0') << (int)arr_u8_hash[i];
    }

    sz_fingerprint = ss.str();

    return sz_fingerprint;

}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_nss_CertlFDIsFDI(const std::string& sz_fingerprint, SE_HexBinary160_t* i3e_lFDI, SE_SFDIType_t* i3e_sFDI) 
{

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Convert to lFDI
    StackScop std::string sz_lFDIstr = sz_fingerprint.substr(0, 40);
    for (size_t i = 0; i < sz_lFDIstr.length(); i += 2) {
        StackScop std::string sz_byteString = sz_lFDIstr.substr(i, 2);
        (*i3e_lFDI)[i / 2] = static_cast<uint8_t>(std::stoul(sz_byteString, nullptr, 16));
    }

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Convert to sFDI
    StackScop std::string sz_sFDIstr = sz_fingerprint.substr(0, 9);
    StackScop uint8_t u8_tally = 0;
    StackScop uint64_t u64_tempsFDI = 0;
    StackScop std::string sz_tempsFDIstr;
    u64_tempsFDI = std::stoull(sz_sFDIstr, nullptr, 16);
    sz_tempsFDIstr = std::to_string(u64_tempsFDI);
    for (size_t i = 0; i < sz_tempsFDIstr.length(); i += 1) {   
        u8_tally += std::stoi(sz_tempsFDIstr.substr(i, 1), nullptr, 10);
    }

    // Get the CRC value and append to sFDI
    u8_tally = u8_tally % 10;
    StackScop uint8_t u8_crcval = 10 - u8_tally; 
    StackScop std::string sz_crcStr = std::to_string(u8_crcval);
    sz_tempsFDIstr = sz_tempsFDIstr + sz_crcStr;
    *i3e_sFDI = std::stoull(sz_tempsFDIstr, nullptr, 10);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

