#include "lion/log/log.h"
#include "include/I3E_SDA.hpp"
#include "include/I3E_DAG.hpp"
#include "include/I3E_CTS.hpp"
#include "include/I3E_XMLS.hpp"
#include "include/I3E_HTTP.hpp"
#include "include/I3E_NSS.hpp"
#include "lion/CST.h"

LOG_MODULE_REGISTER(I3E_ICC, LOG_LEVEL_INF);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define SDA_SLEEP 20
#define MUP_SERVCATKIND 0
#define MUP_STATUS 1 


/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

/**
 * @brief Hex string to byte array conversion.
 * 
 * @details This function converts a hex string to a byte array.
 * 
 * @param sz_hexString Hex string to convert.
 * @param ptr_u8_byteArray Byte array to store the converted hex string.
 * @param siz_byteArraySize Size of the byte array.
 */
FileScop void i3e_sda_hexStringToByteArray(const std::string& sz_hexString, uint8_t* ptr_u8_byteArray, 
                                           size_t siz_byteArraySize);

/**
 * @brief SDA Data reporting function.
 * 
 * @details This function is responsible for updating the IEEE 2030.5 server with metrology data that is periodically
 * collected from the DER. Values include frequency, voltage, reactive power and active power.  
 * 
 * @param ptr_u32_dataReportLoop Pointer to the data report loop counter.
 */
FileScop void i3e_sda_PeriodicDataReport(uint32_t *ptr_u32_dataReportLoop);

/**
 * @brief SDA Status reporting function.
 * 
 * @details This function is responsible for updating the IEEE 2030.5 server with status data that is periodically
 * collected from the DER.
 * 
 * @param ptr_u32_statusReportLoop Pointer to the status report loop counter.
 */
FileScop void i3e_sda_PeriodicStatusReport(uint32_t *ptr_u32_statusReportLoop);

/**
 * @brief SDA Alarm reporting function.
 * 
 * @details This function is responsible for updating the IEEE 2030.5 server with alarm data that is periodically
 * collected from the DER.
 * 
 * @param ptr_u32_alarmReportLoop Pointer to the alarm report loop counter.
 */
FileScop void i3e_sda_PeriodicAlarmReport(uint32_t *ptr_u32_alarmReportLoop);

/**
 * @brief DERSettings report function.
 * 
 * @details This function is responsible for updating the IEEE 2030.5 server with DERSettings data.
 * 
 * @param sz_txPayload Payload to send to the server.
 * @param sz_rxPayload Payload received from the server.
 */
FileScop void i3e_sda_derSettingsReport(std::string sz_txPayload, std::string sz_rxPayload);

/**
 * @brief DERCapability report function.
 * 
 * @details This function is responsible for updating the IEEE 2030.5 server with DERCapability data.
 * 
 * @param sz_txPayload Payload to send to the server.
 * @param sz_rxPayload Payload received from the server.
 */
FileScop void i3e_sda_derCapabilityReport(std::string sz_txPayload, std::string sz_rxPayload);

/**
 * @brief DERStatus report function.
 * 
 * @details This function is responsible for updating the IEEE 2030.5 server with DERStatus data.
 * 
 * @param sz_txPayload Payload to send to the server.
 * @param sz_rxPayload Payload received from the server.
 */
FileScop void i3e_sda_derStatusReport(std::string sz_txPayload, std::string sz_rxPayload);

/**
 * @brief DERAvailability report function.
 * 
 * @details This function is responsible for updating the IEEE 2030.5 server with DERAvailability data.
 * 
 * @param sz_txPayload Payload to send to the server.
 * @param sz_rxPayload Payload received from the server.
 */
FileScop void i3e_sda_derAvailabilityReport(std::string sz_txPayload, std::string sz_rxPayload);

/**
 * @brief Initialize the MirrorUsagePoint.
 * 
 * @details This function initializes the MirrorUsagePoint.
 * 
 * @param sz_txPayload Payload to send to the server.
 * @param sz_rxPayload Payload received from the server.
 */
FileScop void i3e_sda_derInitializeMup(std::string sz_txPayload, std::string sz_rxPayload);

/**
 * @brief Test function for log event. 
 * 
 * @details This function only runs a test POST for the log event functionality. Typically during runtime, we only 
 * want to report log events that actually occurred on the DER. This test case will report a "fake" log event.
 * 
 * @param sz_txPayload Payload to send to the server.
 * @param sz_rxPayload Payload received from the server.
 */
FileScop void i3e_sda_logEventTest(std::string sz_txPayload, std::string sz_rxPayload);

/**
 * @brief Populate the MirrorUsagePoint metadata.
 * 
 * @details This function populates the MirrorUsagePoint metadata.
 * 
 * @param i3e_mup Pointer to the MirrorUsagePoint.
 * @param i3e_mmr Pointer to the MirrorMeterReading.
 * @param bol_isInit Boolean flag indicating if the MUP is being initialized.
 */
FileScop void i3e_sda_mupMetaData(SE_MirrorUsagePoint_t *i3e_mup, SE_MirrorMeterReading_t *i3e_mmr, bool bol_isInit);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop std::string sz_AP_PA_mRID = "0000FFFFFFFFFFFFFFFFFFFF0000F0A0";          // Active Power: Phase A
FileScop std::string sz_AP_PB_mRID = "0001FFFFFFFFFFFFFFFFFFFF0000F0A0";          // Active Power: Phase B
FileScop std::string sz_AP_PC_mRID = "0002FFFFFFFFFFFFFFFFFFFF0000F0A0";          // Active Power: Phase C

FileScop std::string sz_RP_PA_mRID = "1000FFFFFFFFFFFFFFFFFFFF0000F0A0";          // Reactive Power: Phase A
FileScop std::string sz_RP_PB_mRID = "1001FFFFFFFFFFFFFFFFFFFF0000F0A0";          // Reactive Power: Phase B    
FileScop std::string sz_RP_PC_mRID = "1002FFFFFFFFFFFFFFFFFFFF0000F0A0";          // Reactive Power: Phase C    

FileScop std::string sz_V_PAN_mRID = "2000FFFFFFFFFFFFFFFFFFFF0000F0A0";          // Voltage: Phase AN    
FileScop std::string sz_V_PBN_mRID = "2001FFFFFFFFFFFFFFFFFFFF0000F0A0";          // Voltage: Phase BN    
FileScop std::string sz_V_PCN_mRID = "2002FFFFFFFFFFFFFFFFFFFF0000F0A0";          // Voltage: Phase CN    

FileScop std::string sz_F_mRID     = "3000FFFFFFFFFFFFFFFFFFFF0000F0A0";          // Frequency


FileScop std::string sz_MUP_mRID = "FFFFFFFFFFFFFFFFFFFFFFFF0000F0A0";            // MUP mRID                           
FileScop SE_String32_t sz_MUP_DESC = "DER: Lion EMS";                        // MUP Description
FileScop std::string sz_MUP_RoleFlags = "0041";                                   // MUP RoleFlags

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop uint32_t u32_DataPostRate   = 300;  // Init value
FileScop uint32_t u32_StatusPostRate = 900;  // Init value
FileScop uint32_t u32_AlarmsPostRate = 10;   // Init value 
FileScop bool bol_MupCreated = false;
FileScop std::string sz_MupLoc;

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/


// #TODO - This function contains the setup of the MUP, if the server is not available, the MUP will not be created
ModScop void I3E_SDA_InitResourcePost(void) 
{

    StackScop std::string sz_txPayload;
    StackScop std::string sz_rxPayload;

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Report DERSettings
    i3e_sda_derSettingsReport(sz_txPayload, sz_rxPayload);
    sz_rxPayload.clear();
    sz_txPayload.clear();

    // Report DERCapability
    i3e_sda_derCapabilityReport(sz_txPayload, sz_rxPayload);
    sz_rxPayload.clear();
    sz_txPayload.clear();

    // Report DERStatus
    i3e_sda_derStatusReport(sz_txPayload, sz_rxPayload);
    sz_rxPayload.clear();
    sz_txPayload.clear();

    //#TODO - The DERAvaialbility requires calculations for SOC and other parameters, need to develop this out later 
    // Report DERAvailability
    // i3e_sda_derAvailabilityReport(sz_txPayload, sz_rxPayload);
    // sz_rxPayload.clear();
    // sz_txPayload.clear();

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Initialize the MirrorUsagePoint with endpoints for reporting metrology data
    i3e_sda_derInitializeMup(sz_txPayload, sz_rxPayload);
    sz_rxPayload.clear();
    sz_txPayload.clear();
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void I3E_SDA_Exec(void) 
{

    // Looping counters for the various reporting functions.
    StackScop uint32_t u32_dataReportLoop = 0;
    StackScop uint32_t u32_statusReportLoop = 0;
    StackScop uint32_t u32_alarmReportLoop = 0; 

    StackScop std::string sz_txPayload;
    StackScop std::string sz_rxPayload;

    // Test function for log event
    //#TODO - in production, this function should not be called. 
    i3e_sda_logEventTest(sz_txPayload, sz_rxPayload);

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Infinite thread loop
    while(1) 
    {
        LOG_INF("SDA Reporter Thread Executing");

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // Data reporting
        if(bol_MupCreated)
        {
            // Data reporting 300s default, or defined post rate value in the MUP 
            i3e_sda_PeriodicDataReport(&u32_dataReportLoop);
        } 

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // Status reporting

        // Status reporting on pollRate defined in DERList:pollRate (avaialbility same)
        // DERCapability, DERSettings - on changes and startup
        i3e_sda_PeriodicStatusReport(&u32_statusReportLoop);

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // Alarms reporting

        // As they occur - polling rate
        i3e_sda_PeriodicAlarmReport(&u32_alarmReportLoop);
        
        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // Sleep SDA Thread
        sleep(SDA_SLEEP);
        u32_dataReportLoop    += SDA_SLEEP;
        u32_statusReportLoop  += SDA_SLEEP;
        u32_alarmReportLoop   += SDA_SLEEP; 
    }
}

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

// #TODO - this might be able to be combined with the intial MUP creation function
FileScop void i3e_sda_PeriodicDataReport(uint32_t *ptr_u32_dataReportLoop)
{
    
    if(*ptr_u32_dataReportLoop >= u32_DataPostRate)
    {
        StackScop std::string sz_rxPayload;
        StackScop std::string sz_txPayload;

        StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
        StackScop uint16_t u16_port = I3E_NSS_GetPORT();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        StackScop SE_MirrorUsagePoint_t i3e_mup = {0};
        StackScop SE_MirrorMeterReading_t i3e_mmr = {0};
        i3e_sda_mupMetaData(&i3e_mup, &i3e_mmr, false);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        // Serialize the MMRs and POST to the server
        for(uint8_t i = 0; i < i3e_mup.MirrorMeterReading.size(); i++) {

            I3E_XML_SRLZ_MirrorMeterReading(sz_txPayload, &i3e_mup.MirrorMeterReading.at(i));
            I3E_HTTP_POST(sz_ipv4Addr, u16_port, sz_MupLoc, sz_txPayload, sz_rxPayload); 

            sz_rxPayload.clear();
            sz_txPayload.clear();
        }

        *ptr_u32_dataReportLoop -= u32_DataPostRate;
    }
    else{}
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_sda_PeriodicStatusReport(uint32_t *ptr_u32_statusReportLoop)
{
    StackScop std::string sz_txPayload;
    StackScop std::string sz_rxPayload;

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    if(*ptr_u32_statusReportLoop >= u32_StatusPostRate)
    {

        i3e_sda_derStatusReport(sz_txPayload, sz_rxPayload);
        sz_rxPayload.clear();
        sz_txPayload.clear();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        // i3e_sda_derAvailabilityReport(sz_txPayload, sz_rxPayload);
        // sz_rxPayload.clear();
        // sz_txPayload.clear();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        *ptr_u32_statusReportLoop -= u32_StatusPostRate;
    }
    else{}
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_sda_PeriodicAlarmReport(uint32_t *ptr_u32_alarmReportLoop)
{

    StackScop std::string sz_txPayload;
    StackScop std::string sz_rxPayload;
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();

    StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
    StackScop uint16_t u16_port = I3E_NSS_GetPORT();

    if(*ptr_u32_alarmReportLoop >= u32_AlarmsPostRate)
    {
        // Acquire EndDevice node/s
        StackScop std::vector<std::string> vec_endDeviceHrefs = i3e_dagResources->I3E_DAG_FindResourceNodes(EDEV);

        // Loop through all endDevice instances 
        for (uint8_t i = 0; i < vec_endDeviceHrefs.size(); i++)
        {
            StackScop NodeData str_endDeviceNode = i3e_dagResources->I3E_DAG_GetNode(vec_endDeviceHrefs.at(i));
            StackScop SE_EndDevice_t& i3e_endDeviceStored = std::get<SE_EndDevice_t>(str_endDeviceNode.var_rsrc);

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

            StackScop SE_LogEvent_t i3e_logEvent = {0};   
            i3e_logEvent.logEventPEN = IEEE_LOGEV_PEN;
            StackScop CST_LogEventOccurred_t enum_logRet =  CST_HasLogEventOccurred(&i3e_logEvent);

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

            // Report the log events. This loop may run multiple times if multiple log events are detected.
            while (true) 
            {
                switch (enum_logRet) 
                {
                    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                    case CST_LOG_EVENT_NOT_OCCURRED: 
                    {
                        break;
                    }

                    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                    case CST_LOG_EVENT_OCCURRED: 
                    {
                        I3E_XML_SRLZ_LogEvent(sz_txPayload, &i3e_logEvent);
                        I3E_HTTP_POST(sz_ipv4Addr, u16_port, i3e_endDeviceStored.LogEventListLink.href, 
                                        sz_txPayload, sz_rxPayload);
                        sz_rxPayload.clear();
                        sz_txPayload.clear();
                        break;
                    }

                    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                    case CST_MULTIPLE_LOG_EVENTS_OCCURRED: 
                    {
                        // Rerun the function and process the log event
                        I3E_XML_SRLZ_LogEvent(sz_txPayload, &i3e_logEvent);
                        I3E_HTTP_POST(sz_ipv4Addr, u16_port, i3e_endDeviceStored.LogEventListLink.href, 
                                        sz_txPayload, sz_rxPayload);
                        sz_rxPayload.clear();
                        sz_txPayload.clear();
                        
                        // Rerun the function
                        i3e_logEvent = {0};
                        i3e_logEvent.logEventPEN = IEEE_LOGEV_PEN;
                        enum_logRet = CST_HasLogEventOccurred(&i3e_logEvent);
                        continue; // Restart the loop to check the new logRet value
                    }

                    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                    case CST_LOG_EVENT_ERROR: 
                    {
                        break;
                    }

                    default: 
                    {
                        break;
                    }
                }
                break; // Exit the loop if not CST_MULTIPLE_LOG_EVENTS_OCCURRED
            }

            // #TODO - not sure if this should be included
            // logEvent._flags |=  (
            //                     SE_details_exists
            //                     | SE_extendedData_exists
            //                     );

        }

       *ptr_u32_alarmReportLoop -= u32_AlarmsPostRate; 
    }
    else{}
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_sda_derSettingsReport(std::string sz_txPayload, std::string sz_rxPayload)
{
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();
    StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
    StackScop uint16_t u16_port = I3E_NSS_GetPORT();

    // Acquire DERSettings node/s
    StackScop std::vector<std::string> vec_derSettingsHrefs = i3e_dagResources->I3E_DAG_FindResourceNodes(DERSETT);

    // Loop through all DERSettings instances 
    for (uint8_t i = 0; i < vec_derSettingsHrefs.size(); i++)
    {
        StackScop NodeData str_derSettingsNode = i3e_dagResources->I3E_DAG_GetNode(vec_derSettingsHrefs.at(i));
        StackScop SE_DERSettings_t& i3e_derSettingsStored = std::get<SE_DERSettings_t>(str_derSettingsNode.var_rsrc);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        StackScop SE_DERSettings_t i3e_derSettingsReport = {0};

        // Read settings from inverter and report to server 
        CST_GetSettings(&i3e_derSettingsReport);       
        i3e_derSettingsReport.updatedTime = I3E_CTS_getSyncTime();
        // derSettings._flags |= (         
        //                         SE_setMaxW_exists                   //  REQ
        //                         | SE_setESRampTms_exists            //
        //                         | SE_setMaxV_exists                 //
        //                         | SE_setGradW_exists                //  REQ     
        //                         | SE_setVRef_exists                 //   
        //                         | SE_setMinPFNeg_exists             //   
        //                         | SE_setMinPF_exists                //   
        //                         | SE_setMaxWh_exists                //   
        //                         | SE_setMaxVarNeg_exists            //  
        //                         | SE_setMaxVar_exists               //   
        //                         | SE_setMaxVA_exists                //
        //                         | SE_setMaxDischargeRateW_exists    //
        //                         | SE_setMaxDischargeRateVA_exists   //
        //                         | SE_setMaxChargeRateW_exists       //
        //                         | SE_setMaxChargeRateVA_exists      //
        //                         | SE_setMaxAh_exists                //
        //                         | SE_setMaxA_exists                 //
        //                         | SE_modesEnabled_exists            //
        //                         | SE_setESDelay_exists              //
        //                         | SE_setESHighFreq_exists           //
        //                         | SE_setESHighVolt_exists           //
        //                         | SE_setESLowFreq_exists            //
        //                         | SE_setESLowVolt_exists            //
        //                         | SE_setESRandomDelay_exists        //
        //                         | SE_setSoftGradW_exists            //
        //                         | SE_setVRefOfs_exists              //
        //                         | SE_setMinPFOverExcited_exists     // 
        //                         | SE_setMinPFUnderExcited_exists    // 
        //                         | SE_setMinV_exists                 //
        //                         | SE_setVNom_exists                 //
        //                         );  

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        I3E_XML_SRLZ_DERSettings(sz_txPayload, &i3e_derSettingsReport);
        I3E_HTTP_PUT(sz_ipv4Addr, u16_port, i3e_derSettingsStored.href, sz_txPayload, sz_rxPayload);
    }
}
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_sda_derCapabilityReport(std::string sz_txPayload, std::string sz_rxPayload)
{
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();
    StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
    StackScop uint16_t u16_port = I3E_NSS_GetPORT();

    // Acquire DERCapability node/s
    StackScop std::vector<std::string> vec_derCapabilityHrefs = i3e_dagResources->I3E_DAG_FindResourceNodes(DERCAP);    

    // Loop through all der instances 
    for (uint8_t i = 0; i < vec_derCapabilityHrefs.size(); i++)
    {

        StackScop NodeData str_derCapabilityNode = i3e_dagResources->I3E_DAG_GetNode(vec_derCapabilityHrefs.at(i));
        StackScop SE_DERCapability_t& i3e_derCapabilityStored = std::get<SE_DERCapability_t>(str_derCapabilityNode.var_rsrc);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        StackScop SE_DERCapability_t i3e_derCapabilityReport = {0};   
        
        // Read Capability from inverter and report to server 
        CST_GetCapability(&i3e_derCapabilityReport);      

        // derCapability._flags |= (              
        //                         SE_rtgModesSupported_exists             // REQ
        //                         | SE_rtgAbnormalCategory_exists         //
        //                         | SE_rtgMaxA_exists                     //
        //                         | SE_rtgMaxAh_exists                    //
        //                         | SE_rtgMaxChargeRateVA_exists          //
        //                         | SE_rtgMaxChargeRateW_exists           //
        //                         | SE_rtgMaxDischargeRateVA_exists       //
        //                         | SE_rtgMaxDischargeRateW_exists        //
        //                         | SE_rtgMaxV_exists                     // 
        //                         | SE_rtgMaxVA_exists                    //
        //                         | SE_rtgMaxVar_exists                   //
        //                         | SE_rtgMaxVarNeg_exists                //
        //                         | SE_rtgMaxW_exists                     // REQ              
        //                         | SE_rtgMaxWh_exists                    //
        //                         | SE_rtgMinPFOverExcited_exists         //
        //                         | SE_rtgMinPFUnderExcited_exists        //
        //                         | SE_rtgMinV_exists                     //
        //                         | SE_rtgNormalCategory_exists           //
        //                         | SE_rtgOverExcitedPF_exists            //
        //                         | SE_rtgOverExcitedW_exists             //
        //                         | SE_rtgReactiveSusceptance_exists      //
        //                         | SE_rtgUnderExcitedPF_exists           //
        //                         | SE_rtgUnderExcitedW_exists            //
        //                         | SE_rtgVNom_exists                     //
        //                         | SE_type_exists                        // REQ
        //                         );     

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
       
        I3E_XML_SRLZ_DERCapability(sz_txPayload, &i3e_derCapabilityReport);
        I3E_HTTP_PUT(sz_ipv4Addr, u16_port, i3e_derCapabilityStored.href, sz_txPayload, sz_rxPayload); 
    }
}
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
FileScop void i3e_sda_derStatusReport(std::string sz_txPayload, std::string sz_rxPayload)
{
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();
    StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
    StackScop uint16_t u16_port = I3E_NSS_GetPORT();

    // Acquire DERStatus node/s
    StackScop std::vector<std::string> vec_derStatusHrefs = i3e_dagResources->I3E_DAG_FindResourceNodes(DERSTAT);    

    // Loop through all der instances 
    for (uint8_t i = 0; i < vec_derStatusHrefs.size(); i++)
    {
        StackScop NodeData str_derStatusNode = i3e_dagResources->I3E_DAG_GetNode(vec_derStatusHrefs.at(i));
        StackScop SE_DERStatus_t& i3e_derStatusStored = std::get<SE_DERStatus_t>(str_derStatusNode.var_rsrc);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        StackScop SE_DERStatus_t i3e_derStatusReport = {0}; 
        
        // Read Status from inverter and report to server 
        CST_GetStatus(&i3e_derStatusReport);
        i3e_derStatusReport.readingTime = I3E_CTS_getSyncTime();

                i3e_derStatusReport._flags |=  (               
        //                         SE_alarmStatus_exists                  //
        //                         | SE_genConnectStatus_exists           //
        //                         | SE_inverterStatus_exists             //
        //                         | SE_localControlModeStatus_exists     //
        //                         | SE_manufacturerStatus_exists         //
        //                         | SE_operationalModeStatus_exists      //
                                 SE_readingTime_exists                // REQ
        //                         | SE_stateOfChargeStatus_exists        //
        //                         | SE_storageModeStatus_exists          //
        //                         | SE_storConnectStatus_exists          //
                                );   
            
        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        I3E_XML_SRLZ_DERStatus(sz_txPayload, &i3e_derStatusReport);
        I3E_HTTP_PUT(sz_ipv4Addr, u16_port, i3e_derStatusStored.href, sz_txPayload, sz_rxPayload); 

    }
}
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//#TODO - unused function for now so commented out
FileScop void i3e_sda_derAvailabilityReport(std::string sz_txPayload, std::string sz_rxPayload)
{
    // // Acquire DERAvailability node/s
    // std::vector<std::string> derAvailabilityHrefs = i3e_dagResources->I3E_DAG_FindResourceNodes(DERAVAIL);    

    // // Loop through all der instances 
    // for (uint8_t i = 0; i < derAvailabilityHrefs.size(); i++)
    // {

    //     NodeData derAvailabilityNode = i3e_dagResources->getNode(derAvailabilityHrefs.at(i));
    //     SE_DERAvailability_t& derAvailabilityVariant = std::get<SE_DERAvailability_t>(derAvailabilityNode.rsrc);

    //     //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    //     SE_DERAvailability_t derAvailability = {0};   
    //     derAvailability.readingTime = I3E_CTS_getSyncTime();
    //     // derAvailability._flags |=  (              
    //     //                             SE_availabilityDuration_exists      //
    //     //                             | SE_maxChargeDuration_exists       //
    //     //                             | SE_readingTime_exists             // REQ
    //     //                             | SE_reserveChargePercent_exists    //
    //     //                             | SE_reservePercent_exists          //
    //     //                             | SE_statVarAvail_exists            //
    //     //                             | SE_statWAvail_exists              //
    //     //                             );       

    //     I3E_XML_SRLZ_DERAvailability(txPayload, &derAvailability);
    //     I3E_HTTP_PUT(HTTP_SRV_IPV4_ADDR, HTTP_SRV_TLS_PORT, derAvailabilityVariant.href, txPayload, rxPayload); 


    // }
}
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_sda_derInitializeMup(std::string sz_txPayload, std::string sz_rxPayload)
{
    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();
    StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
    StackScop uint16_t u16_port = I3E_NSS_GetPORT();

    // Acquire DCAP node/s
    StackScop std::vector<std::string> vec_deviceCapabilityHrefs = i3e_dagResources->I3E_DAG_FindResourceNodes(DCAP);

    // Loop through all DCAP instances 
    for (uint8_t i = 0; i < vec_deviceCapabilityHrefs.size(); i++)
    {
        StackScop NodeData str_deviceCapabilityNode = i3e_dagResources->I3E_DAG_GetNode(vec_deviceCapabilityHrefs.at(i));
        StackScop SE_DeviceCapability_t& i3e_deviceCapabilityStored = 
                                                    std::get<SE_DeviceCapability_t>(str_deviceCapabilityNode.var_rsrc);

        // If the MUP exists, create the endpoints for data reporting 
        if(i3e_deviceCapabilityStored.MirrorUsagePointListLink._flags &= SE_MirrorUsagePointListLink_exists)
        {

            StackScop SE_MirrorUsagePoint_t i3e_mup = {0};
            StackScop SE_MirrorMeterReading_t i3e_mmr = {0};
            i3e_sda_mupMetaData(&i3e_mup, &i3e_mmr, true);

            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

            I3E_XML_SRLZ_MirrorUsagePoint(sz_txPayload, &i3e_mup);
            StackScop HTTPResponse enum_postRes;
            enum_postRes = I3E_HTTP_POST(sz_ipv4Addr, u16_port, 
                                i3e_deviceCapabilityStored.MirrorUsagePointListLink.href, sz_txPayload, sz_rxPayload); 

            // If result of HTTP Post for MUP was successful, get the link 
            if(enum_postRes == HTTPResponse::CREATED)
            {

                // Store mup location
                sz_MupLoc = sz_rxPayload;

                sz_rxPayload.clear();
                sz_txPayload.clear();

                // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

                if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, sz_MupLoc, sz_rxPayload) == HTTPResponse::OK) 
                {
                    // RETRIEVE + STORE --- MUP
                    I3E_SPP_StoreMirrorUsagePoint(i3e_deviceCapabilityStored.MirrorUsagePointListLink.href, sz_rxPayload);
                    sz_rxPayload.clear();
                }
                else{}

                // Acquire MUP node/s to get postRate
                StackScop std::vector<std::string> vec_mupHrefs = i3e_dagResources->I3E_DAG_FindResourceNodes(MUP);
                StackScop NodeData str_mupNode = i3e_dagResources->I3E_DAG_GetNode(vec_mupHrefs.at(0));
                StackScop SE_MirrorUsagePoint_t& i3e_mupStored = std::get<SE_MirrorUsagePoint_t>(str_mupNode.var_rsrc);
                u32_DataPostRate = i3e_mupStored.postRate;
                bol_MupCreated = true;
            }
            sz_rxPayload.clear();
            sz_txPayload.clear();
        
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_sda_logEventTest(std::string sz_txPayload, std::string sz_rxPayload)
{

    StackScop Graph* i3e_dagResources = I3E_SPP_GetDAG();
    StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
    StackScop uint16_t u16_port = I3E_NSS_GetPORT();

    // Acquire EndDevice node/s
    StackScop std::vector<std::string> vec_endDeviceHrefs = i3e_dagResources->I3E_DAG_FindResourceNodes(EDEV);

    // Loop through all endDevice instances 
    for (uint8_t i = 0; i < vec_endDeviceHrefs.size(); i++)
    {
        StackScop NodeData str_endDeviceNode = i3e_dagResources->I3E_DAG_GetNode(vec_endDeviceHrefs.at(i));
        StackScop SE_EndDevice_t& i3e_endDeviceStored = std::get<SE_EndDevice_t>(str_endDeviceNode.var_rsrc);

        if(!(i3e_endDeviceStored.LogEventListLink.href.empty()))
        {
            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            StackScop SE_LogEvent_t i3e_logEvent = {0};   
            i3e_logEvent.logEventPEN = IEEE_LOGEV_PEN;
            i3e_logEvent.createdDateTime = I3E_CTS_getSyncTime();
            StackScop char sz_det[] = {"General Software Fault"};
            memcpy(i3e_logEvent.details, sz_det, strlen(sz_det));
            i3e_logEvent.logEventCode = 2;
            i3e_logEvent.logEventID = 1;
            i3e_logEvent.profileID = 2;

            I3E_XML_SRLZ_LogEvent(sz_txPayload, &i3e_logEvent);
            I3E_HTTP_POST(sz_ipv4Addr, u16_port, i3e_endDeviceStored.LogEventListLink.href, sz_txPayload, sz_rxPayload);
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_sda_mupMetaData(SE_MirrorUsagePoint_t *i3e_mup, SE_MirrorMeterReading_t *i3e_mmr, bool bol_isInit)
{
    StackScop size_t siz_mmr = CST_GetMirrorMeterReadingsSize();
    i3e_mup->MirrorMeterReading.resize(siz_mmr, *i3e_mmr); 
    CST_GetMirrorMeterReadings(&(i3e_mup->MirrorMeterReading), bol_isInit);
    StackScop SE_HexBinary160_t *i3e_ptr_lfdi =  I3E_NSS_GetLFDI();
    
    // Populate the MUP metadata
    i3e_sda_hexStringToByteArray(sz_MUP_mRID, i3e_mup->mRID, sizeof(i3e_mup->mRID));
    memcpy(i3e_mup->description, sz_MUP_DESC, sizeof(i3e_mup->description));
    i3e_sda_hexStringToByteArray(sz_MUP_RoleFlags, i3e_mup->roleFlags, sizeof(i3e_mup->roleFlags)); 
    i3e_mup->serviceCategoryKind = MUP_SERVCATKIND;
    i3e_mup->status = MUP_STATUS;
    memcpy(i3e_mup->deviceLFDI, i3e_ptr_lfdi, sizeof(i3e_mup->deviceLFDI));

    // Populate the MMR MRIDs
    i3e_sda_hexStringToByteArray(sz_AP_PA_mRID, i3e_mup->MirrorMeterReading.at(0).mRID, 
                                sizeof(i3e_mup->MirrorMeterReading.at(0).mRID));
    i3e_sda_hexStringToByteArray(sz_AP_PB_mRID, i3e_mup->MirrorMeterReading.at(1).mRID, 
                                sizeof(i3e_mup->MirrorMeterReading.at(1).mRID));
    i3e_sda_hexStringToByteArray(sz_AP_PC_mRID, i3e_mup->MirrorMeterReading.at(2).mRID, 
                                sizeof(i3e_mup->MirrorMeterReading.at(2).mRID));
    i3e_sda_hexStringToByteArray(sz_RP_PA_mRID, i3e_mup->MirrorMeterReading.at(3).mRID, 
                                sizeof(i3e_mup->MirrorMeterReading.at(3).mRID));
    i3e_sda_hexStringToByteArray(sz_RP_PB_mRID, i3e_mup->MirrorMeterReading.at(4).mRID, 
                                sizeof(i3e_mup->MirrorMeterReading.at(4).mRID));
    i3e_sda_hexStringToByteArray(sz_RP_PC_mRID, i3e_mup->MirrorMeterReading.at(5).mRID, 
                                sizeof(i3e_mup->MirrorMeterReading.at(5).mRID));
    i3e_sda_hexStringToByteArray(sz_V_PAN_mRID, i3e_mup->MirrorMeterReading.at(6).mRID, 
                                sizeof(i3e_mup->MirrorMeterReading.at(6).mRID));
    i3e_sda_hexStringToByteArray(sz_V_PBN_mRID, i3e_mup->MirrorMeterReading.at(7).mRID, 
                                sizeof(i3e_mup->MirrorMeterReading.at(7).mRID));
    i3e_sda_hexStringToByteArray(sz_V_PCN_mRID, i3e_mup->MirrorMeterReading.at(8).mRID, 
                                sizeof(i3e_mup->MirrorMeterReading.at(8).mRID));
    i3e_sda_hexStringToByteArray(sz_F_mRID, i3e_mup->MirrorMeterReading.at(9).mRID, 
                                sizeof(i3e_mup->MirrorMeterReading.at(9).mRID));

    // Set the required MUP flags 
    i3e_mup->_flags |=  (
                    SE_description_exists 
                    | SE_version_exists     
                    // | SE_ReadingType_exists 
                    // | SE_Reading_exists
                    );

}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_sda_hexStringToByteArray(const std::string& sz_hexString, uint8_t* ptr_u8_byteArray, size_t siz_byteArraySize) {
    // if (sz_hexString.length() != siz_byteArraySize * 2) {
    //     throw std::invalid_argument("Hex string length must be twice the byte array size.");
    // }

    for (size_t i = 0; i < siz_byteArraySize; ++i) {
        StackScop std::string sz_byteString = sz_hexString.substr(i * 2, 2);
        ptr_u8_byteArray[i] = static_cast<uint8_t>(std::stoi(sz_byteString, nullptr, 16));
    }
}
