#include "lion/log/log.h"
#include "include/I3E_SPP.hpp"
#include "include/I3E_NSS.hpp"
#include "include/I3E_XMLD.hpp"
#include "include/I3E_DAG.hpp"
#include "include/I3E_CTS.hpp"

#define POLL_PROCESSOR_SLEEP 10

LOG_MODULE_REGISTER(i3e_spp, LOG_LEVEL_INF);

/**********************************************************************************************************************
 * @cond                                               FILE SCOPE                                              @endcond
 *********************************************************************************************************************/
    
/**
 * @brief Store the Device Capability resource data.
 * 
 * @details This function will process the Device Capability resource data and either create a new graph node or 
 * overwrite an existing one with the new data.
 * 
 * @param sz_rxPayload The payload data from the HTTP GET response.
 */
FileScop void i3e_spp_storeDeviceCapability(std::string sz_rxPayload);

/**
 * @brief Store the Time resource data.
 * 
 * @details This function will process the Time resource data and either create a new graph node or 
 * overwrite an existing one with the new data.
 * 
 * @param sz_parentNodeHref The parent node href of the Time resource, which is used to link the child node to parent. 
 * @param sz_rxPayload The payload data from the HTTP GET response.
 */
FileScop void i3e_spp_storeTime(std::string sz_parentNodeHref, std::string sz_rxPayload);

/**
 * @brief Store the Registration resource data.
 * 
 * @details This function will process the Registration resource data and either create a new graph node or 
 * overwrite an existing one with the new data. This resource contains registration information that must be checked 
 * before the client accepts a connection to the server (lFDI, sFDI, PIN). 
 * 
 * @param sz_parentNodeHref The parent node href of the Registration resource, which is used to link the child node 
 * to parent. 
 * @param sz_rxPayload The payload data from the HTTP GET response.
 */
FileScop void i3e_spp_storeRegistration(std::string sz_parentNodeHref, std::string sz_rxPayload);

/**
 * @brief Store the End Device List resource data.
 * 
 * @details This function will process the End Device List resource data and either create a new graph node or 
 * overwrite an existing one with the new data. This resource contains a list of all end devices that are part of the 
 * client hardware system. Graph nodes are created for the list resource, as well as the individual end devices.
 * 
 * @param sz_parentNodeHref The parent node href of the End Device List resource, which is used to link the child node 
 * to parent. 
 * @param sz_rxPayload The payload data from the HTTP GET response.
 */
FileScop void i3e_spp_storeEndDeviceList(std::string sz_parentNodeHref, std::string sz_rxPayload);

/**
 * @brief Store the Function Set Assignment List resource data.
 * 
 * @details This function will process the Function Set Assignment List resource data and either create a new graph node 
 * or overwrite an existing one with the new data. This resource contains a list of all function set assignments that 
 * are assigned to this specific end device. Graph nodes are created for the list resource, as well as the individual 
 * function set assignments.
 * 
 * @param sz_parentNodeHref The parent node href of the Function Set Assignment List resource, which is used to link the
 * child node to parent. 
 * @param sz_rxPayload The payload data from the HTTP GET response.
 */
FileScop void i3e_spp_storeFunctionSetAssignmentList(std::string sz_parentNodeHref, std::string sz_rxPayload);

/**
 * @brief Store the DER Program List resource data.
 * 
 * @details This function will process the DER Program List resource data and either create a new graph node or 
 * overwrite an existing one with the new data. This resource contains a list of all DER programs that are assigned to 
 * this specific function set assignment. Graph nodes are created for the list resource, as well as the individual 
 * DER programs.
 * 
 * @param sz_parentNodeHref The parent node href of the DER Program List resource, which is used to link the child node 
 * to parent. 
 * @param sz_rxPayload The payload data from the HTTP GET response.
 */
FileScop void i3e_spp_storeDERProgramList(std::string sz_parentNodeHref, std::string sz_rxPayload);

/**
 * @brief Store the DER Control List resource data.
 * 
 * @details This function will process the DER Control List resource data and either create a new graph node or 
 * overwrite an existing one with the new data. This resource contains a list of all DER controls that are assigned to 
 * this specific DER program. Graph nodes are created for the list resource, as well as the individual DER controls.
 * 
 * @param sz_parentNodeHref The parent node href of the DER Control List resource, which is used to link the child node 
 * to parent. 
 * @param sz_rxPayload The payload data from the HTTP GET response.
 */
FileScop void i3e_spp_storeDERControlList(std::string sz_parentNodeHref, std::string sz_rxPayload); 

/**
 * @brief Store the Default DER Control resource data.
 * 
 * @details This function will process the Default DER Control resource data and either create a new graph node or 
 * overwrite an existing one with the new data. This resource contains the default DER control settings that are 
 * assigned to this specific DER program. This is not a list resource, so only one node is created.
 * 
 * @param sz_parentNodeHref The parent node href of the Default DER Control resource, which is used to link the child 
 * node to parent. 
 * @param sz_rxPayload The payload data from the HTTP GET response.
 */
FileScop void i3e_spp_storeDefaultDERControl(std::string sz_parentNodeHref, std::string sz_rxPayload);

/**
 * @brief Store the DER Curve List resource data.
 * 
 * @details This function will process the DER Curve List resource data and either create a new graph node or 
 * overwrite an existing one with the new data. This resource contains a list of all DER curves that are assigned to 
 * this specific DER program. Graph nodes are created for the list resource, as well as the individual DER curves.
 * 
 * @param sz_parentNodeHref The parent node href of the DER Curve List resource, which is used to link the child node 
 * to parent. 
 * @param sz_rxPayload The payload data from the HTTP GET response.
 */
FileScop void i3e_spp_storeDERCurveList(std::string sz_parentNodeHref, std::string sz_rxPayload); 

/**
 * @brief Store the DER List resource data.
 * 
 * @details This function will process the DER List resource data and either create a new graph node or 
 * overwrite an existing one with the new data. This resource contains a profile of DER configurations and status 
 * information that are assigned to this specific end device.
 * 
 * @param sz_parentNodeHref The parent node href of the DER List resource, which is used to link the child node 
 * to parent. 
 * @param sz_rxPayload The payload data from the HTTP GET response.
 */
FileScop void i3e_spp_storeDERList(std::string sz_parentNodeHref, std::string sz_rxPayload); 

/**
 * @brief Store the DER Availability resource data.
 * 
 * @details This function will process the DER Availability resource data and either create a new graph node or 
 * overwrite an existing one with the new data. This resource contains the availability status of the DER.
 * 
 * @param sz_parentNodeHref The parent node href of the DER Availability resource, which is used to link the child node 
 * to parent. 
 * @param sz_rxPayload The payload data from the HTTP GET response.
 */
FileScop void i3e_spp_storeDERAvailability(std::string sz_parentNodeHref, std::string sz_rxPayload);

/**
 * @brief Store the DER Capability resource data.
 * 
 * @details This function will process the DER Capability resource data and either create a new graph node or 
 * overwrite an existing one with the new data. This resource contains the capability settings of the DER.
 * 
 * @param sz_parentNodeHref The parent node href of the DER Capability resource, which is used to link the child node 
 * to parent. 
 * @param sz_rxPayload The payload data from the HTTP GET response.
 */
FileScop void i3e_spp_storeDERCapability(std::string sz_parentNodeHref, std::string sz_rxPayload);

/**
 * @brief Store the DER Settings resource data.
 * 
 * @details This function will process the DER Settings resource data and either create a new graph node or 
 * overwrite an existing one with the new data. This resource contains the settings of the DER.
 * 
 * @param sz_parentNodeHref The parent node href of the DER Settings resource, which is used to link the child node 
 * to parent. 
 * @param sz_rxPayload The payload data from the HTTP GET response.
 */
FileScop void i3e_spp_storeDERSettings(std::string sz_parentNodeHref, std::string sz_rxPayload);

/**
 * @brief Store the DER Status resource data.
 * 
 * @details This function will process the DER Status resource data and either create a new graph node or 
 * overwrite an existing one with the new data. This resource contains the status information of the DER.
 * 
 * @param sz_parentNodeHref The parent node href of the DER Status resource, which is used to link the child node 
 * to parent. 
 * @param sz_rxPayload The payload data from the HTTP GET response.
 */
FileScop void i3e_spp_storeDERStatus(std::string sz_parentNodeHref, std::string sz_rxPayload);

/**
 * @brief Resource rediscovery routine
 * 
 * @details Once the polling system launches and a difference is detected in a previously polled resource, this function
 * will be called to rediscover the resource and update the graph node with the new data. All of the children nodes will
 * be rebuilt as well, as there is no good way to determine if the children nodes have changed or moved to another 
 * location without a full rediscovery.
 * 
 * @param sz_rxPayload The payload data from the HTTP GET response on the resource that has changed.
 * @param ptr_str_rediscoverNode The specific node that has changed and needs to be rediscovered.
 */
FileScop void i3e_spp_resourceRediscovery(std::string sz_rxPayload, NodeData* ptr_str_rediscoverNode);

/**
 * @brief Resource rediscovery routine for the Device Capability resource
 * 
 * @details This function will refresh the data in the Device Capability node, as well as all children nodes by 
 * performing a rediscovery routine.
 * 
 * @param ptr_str_rediscoverNode The specific node that has changed and needs to be rediscovered.
 */
FileScop void i3e_spp_rediscoverDeviceCapability(NodeData* ptr_str_rediscoverNode);

/**
 * @brief Resource rediscovery routine for the Time resource
 * 
 * @details This function will refresh the data in the Time node, as well as all children nodes by performing a 
 * rediscovery routine.
 * 
 * @param ptr_str_rediscoverNode The specific node that has changed and needs to be rediscovered.
 */
FileScop void i3e_spp_rediscoverTime(NodeData* ptr_str_rediscoverNode);

/**
 * @brief Resource rediscovery routine for the Registration resource
 * 
 * @details This function will refresh the data in the Registration node, as well as all children nodes by performing a 
 * rediscovery routine.
 * 
 * @param ptr_str_rediscoverNode The specific node that has changed and needs to be rediscovered.
 */
FileScop void i3e_spp_rediscoverEndDeviceList(NodeData* ptr_str_rediscoverNode);

/**
 * @brief Resource rediscovery routine for the Function Set Assignment List resource
 * 
 * @details This function will refresh the data in the Function Set Assignment List node, as well as all children nodes 
 * by performing a rediscovery routine.
 * 
 * @param ptr_str_rediscoverNode The specific node that has changed and needs to be rediscovered.
 */
FileScop void i3e_spp_rediscoverFunctionSetAssignmentList(NodeData* ptr_str_rediscoverNode);

/**
 * @brief Resource rediscovery routine for the DER Program List resource
 * 
 * @details This function will refresh the data in the DER Program List node, as well as all children nodes by performing 
 * a rediscovery routine.
 * 
 * @param ptr_str_rediscoverNode The specific node that has changed and needs to be rediscovered.
 */
FileScop void i3e_spp_rediscoverDERProgramList(NodeData* ptr_str_rediscoverNode);

/**
 * @brief Process an HTTP GET response code. 
 * 
 * @details This function will process the HTTP GET response code and return a boolean value indicating the success or 
 * failure of the HTTP GET request. This is for feeback to the poll processor to determine if the resource discovery 
 * process has had any failures. This function will also do another GET request and store the resource if the 
 * response code is MOVED_PERMANENTLY or FOUND.
 * 
 * 
 * @param enum_rsrcType The type of resource that the HTTP GET request was made to.
 * @param enum_respStat The HTTP response code.
 * @param sz_rxPayload The payload data from the HTTP GET response.
 * @param sz_parentHref The parent node href of the resource that the HTTP GET request was made to.
 */
// #TODO - currently, only the DCAP resource uses this method, for passing test case ERR 001
FileScop bool i3e_spp_httpGetRsp(I3EresourceType enum_rsrcType, HTTPResponse enum_respStat, 
                                std::string &sz_rxPayload, std::string &sz_parentHref);
    
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// This is the resource graph that stores all of the data pulled from the server
Graph resourcesGraph;

/**********************************************************************************************************************
 * @cond                                              MODULE SCOPE                                             @endcond
 *********************************************************************************************************************/

ModScop Graph* I3E_SPP_GetDAG(void)
{
    return &resourcesGraph;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop bool I3E_SPP_InitResourceDiscovery(void) 
{

    StackScop std::string sz_rxPayload;
    StackScop std::string sz_txPayload;
    StackScop HTTPResponse enum_respStat; 
    StackScop bool bol_httpSuccess;
    StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
    StackScop uint16_t u16_port = I3E_NSS_GetPORT();
    StackScop std::string sz_uri = I3E_NSS_GetURI();


    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Send HTTP GET Method to DCAP resource 
    enum_respStat = I3E_HTTP_GET(sz_ipv4Addr, u16_port, sz_uri, sz_rxPayload);
    bol_httpSuccess = i3e_spp_httpGetRsp(DCAP, enum_respStat, sz_rxPayload, sz_txPayload);
    
    // Something went wrong
    if(!bol_httpSuccess)
    {
        return bol_httpSuccess;
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Acquire DCAP node
    StackScop std::vector<std::string> vec_dcapHrefs = resourcesGraph.I3E_DAG_FindResourceNodes(DCAP);
    StackScop NodeData str_dcapNode = resourcesGraph.I3E_DAG_GetNode(vec_dcapHrefs.at(0));
    StackScop SE_DeviceCapability_t& i3e_deviceCapabilityStored = std::get<SE_DeviceCapability_t>(str_dcapNode.var_rsrc);

    // Send HTTP GET Method to TM resource
    if(i3e_deviceCapabilityStored.TimeLink._flags & SE_TimeLink_exists) 
    {
        enum_respStat = I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                        (i3e_deviceCapabilityStored.TimeLink.href + MAX_RSRC_RET_SUFFIX), sz_rxPayload);
        i3e_spp_httpGetRsp(TM, enum_respStat, sz_rxPayload, i3e_deviceCapabilityStored.href);

        // Synchronize the server time to device time 
        I3E_CTS_SyncTime();
    }
    else{}

    // Send HTTP GET Method to EDEVLIST resource 
    if(i3e_deviceCapabilityStored.EndDeviceListLink._flags & SE_EndDeviceListLink_exists) 
    {
        enum_respStat = I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                        (i3e_deviceCapabilityStored.EndDeviceListLink.href + MAX_RSRC_RET_SUFFIX), sz_rxPayload);
        i3e_spp_httpGetRsp(EDEV_LIST, enum_respStat, sz_rxPayload, i3e_deviceCapabilityStored.href);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Acquire EDEV node/s
    StackScop std::vector<std::string> vec_edevHrefs = resourcesGraph.I3E_DAG_FindResourceNodes(EDEV);

    // Loop through all end device instances 
    for (uint8_t i = 0; i < vec_edevHrefs.size(); i++)
    {
        StackScop NodeData str_edevNode = resourcesGraph.I3E_DAG_GetNode(vec_edevHrefs.at(i));
        StackScop SE_EndDevice_t& i3e_endDeviceStored = std::get<SE_EndDevice_t>(str_edevNode.var_rsrc);

        // # TODO - going to have to add this REG functionaliy into the polling system somehow, not just on init
        // Verify current device 
        StackScop SE_SFDIType_t i3e_sFDIdevice = I3E_NSS_GetSFDI();
        StackScop SE_HexBinary160_t* i3e_lFDIdevice = I3E_NSS_GetLFDI();
        StackScop SE_PINType_t i3e_pinDevice = I3E_NSS_GetPIN();

        if(i3e_endDeviceStored.sFDI == i3e_sFDIdevice) 
        {
            if(memcmp(i3e_endDeviceStored.lFDI, i3e_lFDIdevice, sizeof(i3e_lFDIdevice)) == 0) 
            {
                // Send HTTP GET Method to REG resource 
                if(i3e_endDeviceStored.RegistrationLink._flags & SE_RegistrationLink_exists) 
                {
                    if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                                    i3e_endDeviceStored.RegistrationLink.href, sz_rxPayload) == HTTPResponse::OK) 
                    {
                        // RETRIEVE + STORE --- REGISTRATION
                        i3e_spp_storeRegistration(vec_edevHrefs.at(i), sz_rxPayload);
                        sz_rxPayload.clear();

                        // Get just added registration node to verify end device
                        StackScop NodeData str_regNode = resourcesGraph.I3E_DAG_GetNode(i3e_endDeviceStored.RegistrationLink.href);
                        StackScop SE_Registration_t& str_regStored = std::get<SE_Registration_t>(str_regNode.var_rsrc);

                        if(str_regStored.pIN == i3e_pinDevice) 
                        {
                            LOG_INF("End Device %s verified", vec_edevHrefs.at(i).c_str());
                        }
                        // Failed pIN check
                        else
                        {
                            resourcesGraph.I3E_DAG_RemoveNodeAndDependencies(vec_edevHrefs.at(i));
                            LOG_ERR("End Device %s verification failue - invalid PIN", vec_edevHrefs.at(i).c_str());
                        }
                    }
                    else{}
                }
                else{}
            }
            // Failed lFDI check
            else
            {
                resourcesGraph.I3E_DAG_RemoveNodeAndDependencies(vec_edevHrefs.at(i));
                LOG_ERR("End Device %s verification failue - invalid lFDI", vec_edevHrefs.at(i).c_str());
            }
        }
        // Failed sFDI check
        else
        {
            resourcesGraph.I3E_DAG_RemoveNodeAndDependencies(vec_edevHrefs.at(i));
            LOG_ERR("End Device %s verification failue - invalid sFDI", vec_edevHrefs.at(i).c_str());
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Acquire EDEV node/s
    vec_edevHrefs = resourcesGraph.I3E_DAG_FindResourceNodes(EDEV);

    // Loop through all end device instances 
    for (uint8_t i = 0; i < vec_edevHrefs.size(); i++)
    {
        StackScop NodeData str_edevNode = resourcesGraph.I3E_DAG_GetNode(vec_edevHrefs.at(i));
        StackScop SE_EndDevice_t& i3e_endDeviceStored = std::get<SE_EndDevice_t>(str_edevNode.var_rsrc);

        // Send HTTP GET Method to DERLISTLINK resource 
        if(i3e_endDeviceStored.DERListLink._flags & SE_DERListLink_exists) 
        {
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            i3e_endDeviceStored.DERListLink.href, sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERLIST
                i3e_spp_storeDERList(vec_edevHrefs.at(i), sz_rxPayload);
                sz_rxPayload.clear();
            }
            else{}
        }
        else{}
        
        // Send HTTP GET Method to FSALIST resource 
        if(i3e_endDeviceStored.FunctionSetAssignmentsListLink._flags & SE_FunctionSetAssignmentsListLink_exists) 
        {
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_endDeviceStored.FunctionSetAssignmentsListLink.href + MAX_RSRC_RET_SUFFIX),
                             sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE ---  FSAlist & FSAs
                i3e_spp_storeFunctionSetAssignmentList(vec_edevHrefs.at(i), sz_rxPayload);
                sz_rxPayload.clear();
            }
            else{}
        }
        else{}
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Acquire DER node/s
    StackScop std::vector<std::string> vec_derHrefs = resourcesGraph.I3E_DAG_FindResourceNodes(DER);

    // Loop through all der instances 
    for (uint8_t i = 0; i < vec_derHrefs.size(); i++)
    {
        StackScop NodeData str_derNode = resourcesGraph.I3E_DAG_GetNode(vec_derHrefs.at(i));
        StackScop SE_DER_t& i3e_derStored = std::get<SE_DER_t>(str_derNode.var_rsrc);

        if(i3e_derStored.DERAvailabilityLink._flags & SE_DERAvailabilityLink_exists)
        {
            // Send HTTP GET Method to DERAvailability resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            i3e_derStored.DERAvailabilityLink.href, sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERAvailability
                i3e_spp_storeDERAvailability(vec_derHrefs.at(i), sz_rxPayload);
                sz_txPayload = sz_rxPayload;
                sz_rxPayload.clear();
            }
            else{}
        }
        else{}

        //~~~~~~~~~~~~~~~~~
        if(i3e_derStored.DERCapabilityLink._flags & SE_DERCapabilityLink_exists)
        {
            // Send HTTP GET Method to DERCapability resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            i3e_derStored.DERCapabilityLink.href, sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERCapability
                sz_txPayload = sz_rxPayload;
                i3e_spp_storeDERCapability(vec_derHrefs.at(i), sz_rxPayload);
                
                sz_rxPayload.clear();
            }
            else{}
        }
        else{}

        if(i3e_derStored.DERSettingsLink._flags & SE_DERSettingsLink_exists)
        {
            // Send HTTP GET Method to DERSettings resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            i3e_derStored.DERSettingsLink.href, sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERSettings
                sz_txPayload = sz_rxPayload;
                i3e_spp_storeDERSettings(vec_derHrefs.at(i), sz_rxPayload);
                // sz_txPayload = sz_rxPayload;
                sz_rxPayload.clear();
            }
            else{}
        }
        else{}


        if(i3e_derStored.DERStatusLink._flags & SE_DERStatusLink_exists)
        {        
            // Send HTTP GET Method to DERStatus resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            i3e_derStored.DERStatusLink.href, sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERStatus 
                sz_txPayload = sz_rxPayload;
                i3e_spp_storeDERStatus(vec_derHrefs.at(i), sz_rxPayload);
                // sz_txPayload = sz_rxPayload;
                sz_rxPayload.clear();
            }
            else{}
        }
        else{}
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Acquire FSA node/s
    StackScop std::vector<std::string> vec_fsaHrefs = resourcesGraph.I3E_DAG_FindResourceNodes(FSA);

    // Loop through all function set assignment instances 
    for (uint8_t i = 0; i < vec_fsaHrefs.size(); i++)
    {
        StackScop NodeData str_fsaNode = resourcesGraph.I3E_DAG_GetNode(vec_fsaHrefs.at(i));
        StackScop SE_FunctionSetAssignments_t& i3e_functionSetAssignmentStored 
                        = std::get<SE_FunctionSetAssignments_t>(str_fsaNode.var_rsrc);

        if(i3e_functionSetAssignmentStored.DERProgramListLink._flags & SE_DERProgramListLink_exists)
        {   
            // Send HTTP GET Method to DERPROGLIST resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_functionSetAssignmentStored.DERProgramListLink.href + MAX_RSRC_RET_SUFFIX), 
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERProgramlist and DERPrograms
                i3e_spp_storeDERProgramList(vec_fsaHrefs.at(i), sz_rxPayload);
                sz_rxPayload.clear();
            }
            // HTTP GET FAILURE - exit 
            else{}
        }
        else{}
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Acquire DERPROGRAM node/s
    StackScop std::vector<std::string> vec_derprogHrefs = resourcesGraph.I3E_DAG_FindResourceNodes(DERPROG);

    // Loop through all derprogram instances 
    for (uint8_t i = 0; i < vec_derprogHrefs.size(); i++)
    {
        StackScop NodeData str_derprogNode = resourcesGraph.I3E_DAG_GetNode(vec_derprogHrefs.at(i));
        StackScop SE_DERProgram_t& i3e_derProgramStored = std::get<SE_DERProgram_t>(str_derprogNode.var_rsrc);

        if(i3e_derProgramStored.DefaultDERControlLink._flags & SE_DefaultDERControlLink_exists)
        {   
            // Send HTTP GET Method to DefaultDERControl resource
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_derProgramStored.DefaultDERControlLink.href + MAX_RSRC_RET_SUFFIX), 
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DefaultDERControlList and DefaultDERControls
                i3e_spp_storeDefaultDERControl(vec_derprogHrefs.at(i), sz_rxPayload);
                sz_rxPayload.clear();
            }
            // HTTP GET FAILURE - exit 
            else{}
        }
        else{}

        if(i3e_derProgramStored.DERControlListLink._flags & SE_DERControlListLink_exists)
        {   
            // Send HTTP GET Method to DERControlList resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_derProgramStored.DERControlListLink.href + MAX_RSRC_RET_SUFFIX), 
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERControlList and DERControls
                i3e_spp_storeDERControlList(vec_derprogHrefs.at(i), sz_rxPayload);
                sz_rxPayload.clear();
            }
            // HTTP GET FAILURE - exit 
            else{}
        }
        else{}

        if(i3e_derProgramStored.DERCurveListLink._flags & SE_DERCurveListLink_exists)
        {   
            // Send HTTP GET Method to DERCurveList resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_derProgramStored.DERCurveListLink.href + MAX_RSRC_RET_SUFFIX), 
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERCurveList and DERCurves
                i3e_spp_storeDERCurveList(vec_derprogHrefs.at(i), sz_rxPayload);
                sz_rxPayload.clear();
            }
            // HTTP GET FAILURE - exit 
            else{}
        }
        else{}
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // std::cout << "Printing Graph Storage -- " <<std::endl;
    // resourcesGraph.I3E_DAG_PrintNodes("/dcap");

    LOG_INF("Initial resource discovery success");

    // Initial semaphore post for the event manager to read the initialized data   
    I3E_DEM_PostSemaphore();

    return true;

}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void I3E_SPP_Exec(void) 
{

    StackScop SE_TimeType_t ire_currentTime = I3E_CTS_getSyncTime();
    StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
    StackScop uint16_t u16_port = I3E_NSS_GetPORT();
    StackScop std::string sz_uri = I3E_NSS_GetURI();

    // #TODO - this isn't compatible with mutex locking for resource protection - can't use pointer access like this 
    // #TODO - unless we use mutex protected getter/setters for pointer access - have to change this 
    // #TODO - for now, lock the entire polling process

    // Set up next polling times on all nodes that have poll rates 
    StackScop std::vector<NodeData*> vec_pollingResources = resourcesGraph.I3E_DAG_ScanAllNodesWithPollRate();  

    pthread_mutex_lock(&resourcesGraph.I3E_DAG_GetDAGMutex());

    // Set up each poll next value to be addition of current time + poll rate 
    for (int i = 0; i < vec_pollingResources.size(); i++) 
    {
        vec_pollingResources.at(i)->time_pollNext = ire_currentTime + vec_pollingResources.at(i)->u32_pollRate;
    }

    pthread_mutex_unlock(&resourcesGraph.I3E_DAG_GetDAGMutex());

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Endless polling loop
    while (1) 
    {
        LOG_INF("Poll Processor Thread Executing");

        // LOCK DAG MUTEX
        pthread_mutex_lock(&resourcesGraph.I3E_DAG_GetDAGMutex());

        // Run loop over the number of resources we need to poll 
        for(size_t i = 0; i < vec_pollingResources.size(); i++)
        {
            // If the current time exceeds the next polling time, poll that resource
            if (ire_currentTime >= vec_pollingResources.at(i)->time_pollNext) 
            {
                // Switch on type of resource for proper XML parsing method 
                switch (vec_pollingResources.at(i)->enum_rsrcType) 
                {
                    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                    case DCAP:
                    {
                        StackScop std::string sz_rxPayload;
                        StackScop NodeData *ptr_str_dcapNode = vec_pollingResources.at(i);
                        StackScop SE_DeviceCapability_t& i3e_dcapStored = 
                                        std::get<SE_DeviceCapability_t>(vec_pollingResources.at(i)->var_rsrc);

                        // Send HTTP GET Method to DCAP resource 
                        if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                                        sz_uri, sz_rxPayload) == HTTPResponse::OK) 
                        {
                            // If the two string (previous and current) are not equal, change in the resource. 
                            if(sz_rxPayload != ptr_str_dcapNode->sz_lastPollStr) 
                            {

                                // Run rediscovery
                                i3e_spp_resourceRediscovery(sz_rxPayload, ptr_str_dcapNode);
                                sz_rxPayload.clear();

                                // Updated graph - post to event manager semaphore   
                                I3E_DEM_PostSemaphore();

                                // Reset polling resources vector with new nodes and break the for loop
                                vec_pollingResources = resourcesGraph.I3E_DAG_ScanAllNodesWithPollRate();
                                i = vec_pollingResources.size() + 1;
                            }
                            else
                            {
                                vec_pollingResources.at(i)->time_pollNext = 
                                                ire_currentTime + vec_pollingResources.at(i)->u32_pollRate;
                            }
                        }
                        break;
                    }

                    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                    // #TODO this resource probably doesn't need to be checked, it will always be new, 
                    // #TODO and doesn't contain children resources? 
                    case TM:
                    {
                        StackScop std::string sz_rxPayload;
                        StackScop NodeData *ptr_str_timeNode = vec_pollingResources.at(i);
                        StackScop SE_Time_t& i3e_timeStored = std::get<SE_Time_t>(vec_pollingResources.at(i)->var_rsrc);

                        // Send HTTP GET Method to TM resource 
                        if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                                        (i3e_timeStored.href + MAX_RSRC_RET_SUFFIX),
                                        sz_rxPayload) == HTTPResponse::OK)
                        {
                            // If the two string (previous and current) are not equal, change in the resource. 
                            if(sz_rxPayload != ptr_str_timeNode->sz_lastPollStr) 
                            {

                                // Run rediscovery
                                i3e_spp_resourceRediscovery(sz_rxPayload, ptr_str_timeNode);
                                sz_rxPayload.clear();

                                I3E_CTS_SyncTime();

                                // Reset polling resources vector with new nodes and break the for loop
                                vec_pollingResources = resourcesGraph.I3E_DAG_ScanAllNodesWithPollRate();
                                i = vec_pollingResources.size() + 1;
                            }
                            else
                            {
                                vec_pollingResources.at(i)->time_pollNext = 
                                                ire_currentTime + vec_pollingResources.at(i)->u32_pollRate;
                            }
                        }
                        break;
                    }

                    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                    case EDEV_LIST:
                    {
                        StackScop std::string sz_rxPayload;
                        StackScop NodeData *ptr_str_edevlistNode = vec_pollingResources.at(i);
                        StackScop SE_EndDeviceList_t& i3e_edevlistStored = 
                                        std::get<SE_EndDeviceList_t>(vec_pollingResources.at(i)->var_rsrc);

                        // Send HTTP GET Method to EDEVLIST resource 
                        if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                                        (i3e_edevlistStored.href + MAX_RSRC_RET_SUFFIX),
                                        sz_rxPayload) == HTTPResponse::OK) 
                        {
                            // If the two string (previous and current) are not equal, change in the resource.  
                            if(sz_rxPayload != ptr_str_edevlistNode->sz_lastPollStr) 
                            {
                                // Run rediscovery
                                i3e_spp_resourceRediscovery(sz_rxPayload, ptr_str_edevlistNode);
                                sz_rxPayload.clear();
                                
                                // Updated graph - post to event manager semaphore   
                                I3E_DEM_PostSemaphore();                           

                                // Reset polling resources vector with new nodes and break the for loop
                                vec_pollingResources = resourcesGraph.I3E_DAG_ScanAllNodesWithPollRate();
                                i = vec_pollingResources.size() + 1;
                            }
                            else
                            {
                                vec_pollingResources.at(i)->time_pollNext = 
                                                ire_currentTime + vec_pollingResources.at(i)->u32_pollRate;
                            }
                        }
                        break;
                    }

                    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                    case FSA_LIST:
                    {
                        StackScop std::string sz_rxPayload;
                        StackScop NodeData *ptr_str_fsalistNode = vec_pollingResources.at(i);
                        StackScop SE_FunctionSetAssignmentsList_t& i3e_fsaListStored =
                                        std::get<SE_FunctionSetAssignmentsList_t>(vec_pollingResources.at(i)->var_rsrc);

                        // Send HTTP GET Method to FSALIST resource 
                        if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                                        (i3e_fsaListStored.href + MAX_RSRC_RET_SUFFIX),
                                        sz_rxPayload) == HTTPResponse::OK) 
                        {
                            // If the two string (previous and current) are not equal, change in the resource. 
                            if(sz_rxPayload != ptr_str_fsalistNode->sz_lastPollStr) 
                            {
                                // Run rediscovery
                                i3e_spp_resourceRediscovery(sz_rxPayload, ptr_str_fsalistNode);
                                sz_rxPayload.clear();

                                // Updated graph - post to event manager semaphore   
                                I3E_DEM_PostSemaphore();                           

                                // Reset polling resources vector with new nodes and break the for loop
                                vec_pollingResources = resourcesGraph.I3E_DAG_ScanAllNodesWithPollRate();
                                i = vec_pollingResources.size() + 1;
                            }
                            else
                            {
                                vec_pollingResources.at(i)->time_pollNext = 
                                                ire_currentTime + vec_pollingResources.at(i)->u32_pollRate;
                            }
                        }
                        break;
                    }

                    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                    case DERPROG_LIST:
                    {
                        StackScop std::string sz_rxPayload;
                        StackScop NodeData *ptr_str_derProgListNode = vec_pollingResources.at(i);
                        StackScop SE_DERProgramList_t& i3e_derProgListStored =
                                        std::get<SE_DERProgramList_t>(vec_pollingResources.at(i)->var_rsrc);

                        // Send HTTP GET Method to DERPROGLIST resource 
                        if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                                        (i3e_derProgListStored.href + MAX_RSRC_RET_SUFFIX),
                                        sz_rxPayload) == HTTPResponse::OK) 
                        {

                            // If the two string (previous and current) are not equal, change in the resource.  
                            if(sz_rxPayload != ptr_str_derProgListNode->sz_lastPollStr) 
                            {
                                // Run rediscovery
                                i3e_spp_resourceRediscovery(sz_rxPayload, ptr_str_derProgListNode);
                                sz_rxPayload.clear();

                                // Updated graph - post to event manager semaphore   
                                I3E_DEM_PostSemaphore();                               

                                // Reset polling resources vector with new nodes and break the for loop
                                vec_pollingResources = resourcesGraph.I3E_DAG_ScanAllNodesWithPollRate();
                                i = vec_pollingResources.size() + 1;
                            }
                            else
                            {
                                vec_pollingResources.at(i)->time_pollNext = 
                                                ire_currentTime + vec_pollingResources.at(i)->u32_pollRate;
                            }
                        }
                        else{}
                        break;
                    }

                }
                // std::cout << "Printing Graph Storage on poll -- " <<std::endl;
                // resourcesGraph.I3E_DAG_PrintNodes("/dcap");
            }
        }

        // Check if any next poll times are 0 (means the nodes were recreated in rediscovery), and adjust them for new loop
        for (int i = 0; i < vec_pollingResources.size(); i++) 
        {
            if(vec_pollingResources.at(i)->time_pollNext == 0) {
                vec_pollingResources.at(i)->time_pollNext =
                                ire_currentTime + vec_pollingResources.at(i)->u32_pollRate;
            }
        }

        // UNLOCK DAG MUTEX
        pthread_mutex_unlock(&resourcesGraph.I3E_DAG_GetDAGMutex());
        sleep(POLL_PROCESSOR_SLEEP);
        ire_currentTime = I3E_CTS_getSyncTime();
    }
}

/**********************************************************************************************************************
 * @cond                                               FILE SCOPE                                              @endcond
 *********************************************************************************************************************/

FileScop void i3e_spp_storeDeviceCapability(std::string sz_rxPayload) 
{
    // Create struct for payload storage
    StackScop SE_DeviceCapability_t i3e_deviceCapability = {0};
    i3e_deviceCapability.pollRate = 900;
    
    // Deserialize and store the DCAP payload
    I3E_XML_PayloadDeserializeMux(sz_rxPayload, DCAP, &i3e_deviceCapability);
    
    StackScop NodeData stR_deviceCapabilityNode(i3e_deviceCapability.href, i3e_deviceCapability);
    stR_deviceCapabilityNode.enum_rsrcType = DCAP;
    stR_deviceCapabilityNode.u32_pollRate = i3e_deviceCapability.pollRate;
    stR_deviceCapabilityNode.sz_lastPollStr = sz_rxPayload;
    stR_deviceCapabilityNode.bol_updated = true;

    // If the node already exists, replace the contents and delete all children nodes for rediscovery
    if(resourcesGraph.I3E_DAG_DoesNodeExist(i3e_deviceCapability.href)) 
    {
        resourcesGraph.I3E_DAG_UpdateNodeContent(i3e_deviceCapability.href, stR_deviceCapabilityNode);
        resourcesGraph.I3E_DAG_DeleteAllChildren(i3e_deviceCapability.href);
    }
    // If the node doesn't exist, add it to the graph normally and link the graph edges
    else
    {
        resourcesGraph.I3E_DAG_AddNode(i3e_deviceCapability.href, stR_deviceCapabilityNode);
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void I3E_SPP_StoreMirrorUsagePoint(std::string sz_parentNodeHref, std::string sz_rxPayload)
{
    
    // Create struct for payload storage
    StackScop SE_MirrorUsagePoint_t i3e_mirrorUsagePoint = {0};

    // Deserialize and store the MUP payload
    I3E_XML_PayloadDeserializeMux(sz_rxPayload, MUP, &i3e_mirrorUsagePoint);
    
    StackScop NodeData str_mirrorUsagePointNode(i3e_mirrorUsagePoint.href, i3e_mirrorUsagePoint);
    str_mirrorUsagePointNode.enum_rsrcType = MUP;
    str_mirrorUsagePointNode.u32_pollRate = 0;
    str_mirrorUsagePointNode.sz_lastPollStr = sz_rxPayload;
    str_mirrorUsagePointNode.bol_updated = true;

    // If the node already exists, replace the contents and delete all children nodes for rediscovery
    if(resourcesGraph.I3E_DAG_DoesNodeExist(i3e_mirrorUsagePoint.href)) 
    {
        resourcesGraph.I3E_DAG_UpdateNodeContent(i3e_mirrorUsagePoint.href, str_mirrorUsagePointNode);
        resourcesGraph.I3E_DAG_DeleteAllChildren(i3e_mirrorUsagePoint.href);
    }
    // If the node doesn't exist, add it to the graph normally and link the graph edges
    else
    {
        resourcesGraph.I3E_DAG_AddNode(i3e_mirrorUsagePoint.href, str_mirrorUsagePointNode);
        resourcesGraph.I3E_DAG_AddEdge(sz_parentNodeHref, i3e_mirrorUsagePoint.href);
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_spp_storeTime(std::string sz_parentNodeHref, std::string sz_rxPayload)
{

    // Create struct for payload storage
    StackScop SE_Time_t i3e_time = {0};
    i3e_time.pollRate = 900;

    // Deserialize and store the TM payload
    I3E_XML_PayloadDeserializeMux(sz_rxPayload, TM, &i3e_time);
   
    StackScop NodeData str_timeNode(i3e_time.href, i3e_time);
    str_timeNode.enum_rsrcType = TM;
    str_timeNode.u32_pollRate = i3e_time.pollRate;
    str_timeNode.sz_lastPollStr = sz_rxPayload;
    str_timeNode.bol_updated = true;

    // If the node already exists, replace the contents and delete all children nodes for rediscovery
    if(resourcesGraph.I3E_DAG_DoesNodeExist(i3e_time.href)) 
    {
        resourcesGraph.I3E_DAG_UpdateNodeContent(i3e_time.href, str_timeNode);
        resourcesGraph.I3E_DAG_DeleteAllChildren(i3e_time.href);
    }
    // If the node doesn't exist, add it to the graph normally and link the graph edges
    else
    {
        resourcesGraph.I3E_DAG_AddNode(i3e_time.href, str_timeNode);
        resourcesGraph.I3E_DAG_AddEdge(sz_parentNodeHref, i3e_time.href);
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_spp_storeEndDeviceList(std::string sz_parentNodeHref, std::string sz_rxPayload)
{
    // Create struct for payload storage   
    StackScop SE_EndDeviceList_t i3e_endDeviceList = {0};
    i3e_endDeviceList.pollRate = 900;
    
    // Deserialize and store the EDEV_LIST payload
    I3E_XML_PayloadDeserializeMux(sz_rxPayload, EDEV_LIST, &i3e_endDeviceList);
    StackScop NodeData str_endDeviceListNode(i3e_endDeviceList.href, i3e_endDeviceList);
    str_endDeviceListNode.u32_pollRate = i3e_endDeviceList.pollRate;
    str_endDeviceListNode.enum_rsrcType = EDEV_LIST;
    str_endDeviceListNode.sz_lastPollStr = sz_rxPayload;
    str_endDeviceListNode.bol_updated = true;

    // If the node already exists, replace the contents and delete all children nodes for rediscovery
    if(resourcesGraph.I3E_DAG_DoesNodeExist(i3e_endDeviceList.href)) 
    {
        resourcesGraph.I3E_DAG_UpdateNodeContent(i3e_endDeviceList.href, str_endDeviceListNode);
        resourcesGraph.I3E_DAG_DeleteAllChildren(i3e_endDeviceList.href);
    }
    // If the node doesn't exist, add it to the graph normally and link the graph edges
    else
    {
        resourcesGraph.I3E_DAG_AddNode(i3e_endDeviceList.href, str_endDeviceListNode);
        resourcesGraph.I3E_DAG_AddEdge(sz_parentNodeHref, i3e_endDeviceList.href);
    }

    // Iterate through the list of end devices and create graph nodes and edges for each 
    for(int i = 0; i < i3e_endDeviceList.EndDevice.size(); i++) 
    {
        StackScop NodeData str_endDeviceNode(i3e_endDeviceList.EndDevice.at(i).href, i3e_endDeviceList.EndDevice.at(i));
        str_endDeviceNode.enum_rsrcType = EDEV;
        str_endDeviceNode.bol_updated = true;
        resourcesGraph.I3E_DAG_AddNode(i3e_endDeviceList.EndDevice.at(i).href, str_endDeviceNode);
        resourcesGraph.I3E_DAG_AddEdge(i3e_endDeviceList.href, i3e_endDeviceList.EndDevice.at(i).href);
    }
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_spp_storeRegistration(std::string sz_parentNodeHref, std::string sz_rxPayload)
{
    // Create struct for payload storage   
    StackScop SE_Registration_t i3e_registration = {0};
    
    // Deserialize and store the REG payload
    I3E_XML_PayloadDeserializeMux(sz_rxPayload, REG, &i3e_registration);
    
    StackScop NodeData str_registrationNode(i3e_registration.href, i3e_registration);
    str_registrationNode.enum_rsrcType = REG;
    str_registrationNode.u32_pollRate = i3e_registration.pollRate;
    str_registrationNode.sz_lastPollStr = sz_rxPayload;
    str_registrationNode.bol_updated = true;

    // If the node already exists, replace the contents and delete all children nodes for rediscovery
    if(resourcesGraph.I3E_DAG_DoesNodeExist(i3e_registration.href)) 
    {
        resourcesGraph.I3E_DAG_UpdateNodeContent(i3e_registration.href, str_registrationNode);
        resourcesGraph.I3E_DAG_DeleteAllChildren(i3e_registration.href);
    }
    // If the node doesn't exist, add it to the graph normally and link the graph edges
    else
    {
        resourcesGraph.I3E_DAG_AddNode(i3e_registration.href, str_registrationNode);
        resourcesGraph.I3E_DAG_AddEdge(sz_parentNodeHref, i3e_registration.href);
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_spp_storeDERList(std::string sz_parentNodeHref, std::string sz_rxPayload)
{

    // Create struct for payload storage   
    StackScop SE_DERList_t i3e_derList = {0};
    i3e_derList.pollRate = 900;

    // Deserialize and store the DERLIST payload
    I3E_XML_PayloadDeserializeMux(sz_rxPayload, DER_LIST, &i3e_derList);

    StackScop NodeData str_derListNode(i3e_derList.href, i3e_derList);
    str_derListNode.u32_pollRate = i3e_derList.pollRate;
    str_derListNode.enum_rsrcType = DER_LIST;
    str_derListNode.sz_lastPollStr = sz_rxPayload;
    str_derListNode.bol_updated = true;

    // If the node already exists, replace the contents and delete all children nodes for rediscovery
    if(resourcesGraph.I3E_DAG_DoesNodeExist(i3e_derList.href)) 
    {
        resourcesGraph.I3E_DAG_UpdateNodeContent(i3e_derList.href, str_derListNode);
        resourcesGraph.I3E_DAG_DeleteAllChildren(i3e_derList.href);
    }
    // If the node doesn't exist, add it to the graph normally and link the graph edges
    else
    {
        resourcesGraph.I3E_DAG_AddNode(i3e_derList.href, str_derListNode);
        resourcesGraph.I3E_DAG_AddEdge(sz_parentNodeHref, i3e_derList.href);
    }

    // Iterate through the list of ders and create graph nodes and edges for each 
    for(int i = 0; i < i3e_derList.DER.size(); i++) 
    {
        StackScop NodeData str_derNode(i3e_derList.DER.at(i).href, i3e_derList.DER.at(i));
        str_derNode.enum_rsrcType = DER;
        str_derNode.bol_updated = true;
        resourcesGraph.I3E_DAG_AddNode(i3e_derList.DER.at(i).href, str_derNode);
        resourcesGraph.I3E_DAG_AddEdge(i3e_derList.href, i3e_derList.DER.at(i).href);
        
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_spp_storeDERAvailability(std::string sz_parentNodeHref, std::string sz_rxPayload)
{
    // Create struct for payload storage   
    StackScop SE_DERAvailability_t i3e_derAvail = {0};
    
    // Deserialize and store the DERAVAIL payload
    I3E_XML_PayloadDeserializeMux(sz_rxPayload, DERAVAIL, &i3e_derAvail);

    StackScop NodeData str_derAvailNode(i3e_derAvail.href, i3e_derAvail);
    str_derAvailNode.enum_rsrcType = DERAVAIL;
    str_derAvailNode.sz_lastPollStr = sz_rxPayload;
    str_derAvailNode.bol_updated = true;

    // If the node already exists, replace the contents and delete all children nodes for rediscovery
    if(resourcesGraph.I3E_DAG_DoesNodeExist(i3e_derAvail.href)) 
    {
        resourcesGraph.I3E_DAG_UpdateNodeContent(i3e_derAvail.href, str_derAvailNode);
        resourcesGraph.I3E_DAG_DeleteAllChildren(i3e_derAvail.href);
    }
    // If the node doesn't exist, add it to the graph normally and link the graph edges
    else
    {
        resourcesGraph.I3E_DAG_AddNode(i3e_derAvail.href, str_derAvailNode);
        resourcesGraph.I3E_DAG_AddEdge(sz_parentNodeHref, i3e_derAvail.href);
    }

}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_spp_storeDERCapability(std::string sz_parentNodeHref, std::string sz_rxPayload)
{
    // Create struct for payload storage   
    StackScop SE_DERCapability_t i3e_derCap = {0};
    
    // Deserialize and store the DERCAP payload
    I3E_XML_PayloadDeserializeMux(sz_rxPayload, DERCAP, &i3e_derCap);
    
    StackScop NodeData str_derCapNode(i3e_derCap.href, i3e_derCap);
    str_derCapNode.enum_rsrcType = DERCAP;
    str_derCapNode.sz_lastPollStr = sz_rxPayload;
    str_derCapNode.bol_updated = true;

    // If the node already exists, replace the contents and delete all children nodes for rediscovery
    if(resourcesGraph.I3E_DAG_DoesNodeExist(i3e_derCap.href)) 
    {
        resourcesGraph.I3E_DAG_UpdateNodeContent(i3e_derCap.href, str_derCapNode);
        resourcesGraph.I3E_DAG_DeleteAllChildren(i3e_derCap.href);
    }
    // If the node doesn't exist, add it to the graph normally and link the graph edges
    else
    {
        resourcesGraph.I3E_DAG_AddNode(i3e_derCap.href, str_derCapNode);
        resourcesGraph.I3E_DAG_AddEdge(sz_parentNodeHref, i3e_derCap.href);
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_spp_storeDERSettings(std::string sz_parentNodeHref, std::string sz_rxPayload)
{
    // Create struct for payload storage   
    StackScop SE_DERSettings_t i3e_derSett = {0};
    
    // Deserialize and store the DERSETT payload
    I3E_XML_PayloadDeserializeMux(sz_rxPayload, DERSETT, &i3e_derSett);
    
    StackScop NodeData str_derSettNode(i3e_derSett.href, i3e_derSett);
    str_derSettNode.enum_rsrcType = DERSETT;
    str_derSettNode.sz_lastPollStr = sz_rxPayload;
    str_derSettNode.bol_updated = true;

    // If the node already exists, replace the contents and delete all children nodes for rediscovery
    if(resourcesGraph.I3E_DAG_DoesNodeExist(i3e_derSett.href)) 
    {
        resourcesGraph.I3E_DAG_UpdateNodeContent(i3e_derSett.href, str_derSettNode);
        resourcesGraph.I3E_DAG_DeleteAllChildren(i3e_derSett.href);
    }
    // If the node doesn't exist, add it to the graph normally and link the graph edges
    else
    {
        resourcesGraph.I3E_DAG_AddNode(i3e_derSett.href, str_derSettNode);
        resourcesGraph.I3E_DAG_AddEdge(sz_parentNodeHref, i3e_derSett.href);
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_spp_storeDERStatus(std::string sz_parentNodeHref, std::string sz_rxPayload)
{
    // Create struct for payload storage   
    StackScop SE_DERStatus_t i3e_derStat = {0};
    
    // Deserialize and store the DERSTAT payload
    I3E_XML_PayloadDeserializeMux(sz_rxPayload, DERSTAT, &i3e_derStat);
    
    StackScop NodeData str_derStatNode(i3e_derStat.href, i3e_derStat);
    str_derStatNode.enum_rsrcType = DERSTAT;
    str_derStatNode.sz_lastPollStr = sz_rxPayload;
    str_derStatNode.bol_updated = true;

    // If the node already exists, replace the contents and delete all children nodes for rediscovery
    if(resourcesGraph.I3E_DAG_DoesNodeExist(i3e_derStat.href)) 
    {
        resourcesGraph.I3E_DAG_UpdateNodeContent(i3e_derStat.href, str_derStatNode);
        resourcesGraph.I3E_DAG_DeleteAllChildren(i3e_derStat.href);
    }
    // If the node doesn't exist, add it to the graph normally and link the graph edges
    else
    {
        resourcesGraph.I3E_DAG_AddNode(i3e_derStat.href, str_derStatNode);
        resourcesGraph.I3E_DAG_AddEdge(sz_parentNodeHref, i3e_derStat.href);
    }

}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_spp_storeFunctionSetAssignmentList(std::string sz_parentNodeHref, std::string sz_rxPayload)
{

    // Create struct for payload storage   
    StackScop SE_FunctionSetAssignmentsList_t i3e_functionSetAssignmentsList = {0};
    i3e_functionSetAssignmentsList.pollRate = 900;

    // Deserialize and store the FSALIST payload
    I3E_XML_PayloadDeserializeMux(sz_rxPayload, FSA_LIST, &i3e_functionSetAssignmentsList);
    
    StackScop NodeData str_functionSetAssignmentListNode(i3e_functionSetAssignmentsList.href,
                     i3e_functionSetAssignmentsList);
    str_functionSetAssignmentListNode.u32_pollRate = i3e_functionSetAssignmentsList.pollRate;
    str_functionSetAssignmentListNode.enum_rsrcType = FSA_LIST;
    str_functionSetAssignmentListNode.sz_lastPollStr = sz_rxPayload;
    str_functionSetAssignmentListNode.bol_updated = true;

    // If the node already exists, replace the contents and delete all children nodes for rediscovery
    if(resourcesGraph.I3E_DAG_DoesNodeExist(i3e_functionSetAssignmentsList.href)) 
    {
        resourcesGraph.I3E_DAG_UpdateNodeContent(i3e_functionSetAssignmentsList.href, str_functionSetAssignmentListNode);
        resourcesGraph.I3E_DAG_DeleteAllChildren(i3e_functionSetAssignmentsList.href);
    }
    // If the node doesn't exist, add it to the graph normally and link the graph edges
    else
    {
        resourcesGraph.I3E_DAG_AddNode(i3e_functionSetAssignmentsList.href, str_functionSetAssignmentListNode);
        resourcesGraph.I3E_DAG_AddEdge(sz_parentNodeHref, i3e_functionSetAssignmentsList.href);
    }

    // Iterate through the list of fsas and create graph nodes and edges for each 
    for(int i = 0; i < i3e_functionSetAssignmentsList.FunctionSetAssignments.size(); i++) 
    {
        StackScop NodeData str_functionSetAssignmentsNode
                        (i3e_functionSetAssignmentsList.FunctionSetAssignments.at(i).href, 
                         i3e_functionSetAssignmentsList.FunctionSetAssignments.at(i));
        str_functionSetAssignmentsNode.enum_rsrcType = FSA;
        str_functionSetAssignmentsNode.bol_updated = true;
        resourcesGraph.I3E_DAG_AddNode(i3e_functionSetAssignmentsList.FunctionSetAssignments.at(i).href,
                        str_functionSetAssignmentsNode);
        resourcesGraph.I3E_DAG_AddEdge(i3e_functionSetAssignmentsList.href, 
                        i3e_functionSetAssignmentsList.FunctionSetAssignments.at(i).href);
        
    }
   
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_spp_storeDERProgramList(std::string sz_parentNodeHref, std::string sz_rxPayload)
{

    // Create struct for payload storage 
    StackScop SE_DERProgramList_t i3e_derProgramsList = {0};
    i3e_derProgramsList.pollRate = 900;

    // Deserialize and store the DERPROGLIST payload
    I3E_XML_PayloadDeserializeMux(sz_rxPayload, DERPROG_LIST, &i3e_derProgramsList);

    StackScop NodeData str_derProgramlistNode(i3e_derProgramsList.href, i3e_derProgramsList);
    str_derProgramlistNode.u32_pollRate = i3e_derProgramsList.pollRate;
    str_derProgramlistNode.enum_rsrcType = DERPROG_LIST;
    str_derProgramlistNode.sz_lastPollStr = sz_rxPayload;
    str_derProgramlistNode.bol_updated = true;

    // If the node already exists, replace the contents and delete all children nodes for rediscovery
    if(resourcesGraph.I3E_DAG_DoesNodeExist(i3e_derProgramsList.href)) 
    {
        resourcesGraph.I3E_DAG_UpdateNodeContent(i3e_derProgramsList.href, str_derProgramlistNode);
        resourcesGraph.I3E_DAG_DeleteAllChildren(i3e_derProgramsList.href);
    }
    // If the node doesn't exist, add it to the graph normally and link the graph edges
    else
    {
        resourcesGraph.I3E_DAG_AddNode(i3e_derProgramsList.href, str_derProgramlistNode);
        resourcesGraph.I3E_DAG_AddEdge(sz_parentNodeHref, i3e_derProgramsList.href);
    }

    // Iterate through the list of DERPrograms and create graph nodes and edges for each 
    for(int i = 0; i < i3e_derProgramsList.DERProgram.size(); i++) 
    {
        StackScop NodeData str_derProgramNode(i3e_derProgramsList.DERProgram.at(i).href,
                        i3e_derProgramsList.DERProgram.at(i));
        str_derProgramNode.enum_rsrcType = DERPROG;
        str_derProgramNode.bol_updated = true;
        resourcesGraph.I3E_DAG_AddNode(i3e_derProgramsList.DERProgram.at(i).href, str_derProgramNode);
        resourcesGraph.I3E_DAG_AddEdge(i3e_derProgramsList.href, i3e_derProgramsList.DERProgram.at(i).href);
    }
}
  
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_spp_storeDERControlList(std::string sz_parentNodeHref, std::string sz_rxPayload)
{
    // Create struct for payload storage 
    StackScop SE_DERControlList_t i3e_derControlList = {0};

    // Deserialize and store the DERP_LIST payload
    I3E_XML_PayloadDeserializeMux(sz_rxPayload, DERCTRL_LIST, &i3e_derControlList);
    
    StackScop NodeData str_derControlListNode(i3e_derControlList.href, i3e_derControlList);
    str_derControlListNode.enum_rsrcType = DERCTRL_LIST;
    str_derControlListNode.sz_lastPollStr = sz_rxPayload;
    str_derControlListNode.bol_updated = true;

    // If the node already exists, replace the contents and delete all children nodes for rediscovery
    if(resourcesGraph.I3E_DAG_DoesNodeExist(i3e_derControlList.href)) 
    {
        resourcesGraph.I3E_DAG_UpdateNodeContent(i3e_derControlList.href, str_derControlListNode);
        resourcesGraph.I3E_DAG_DeleteAllChildren(i3e_derControlList.href);
    }
    // If the node doesn't exist, add it to the graph normally and link the graph edges
    else
    {
        resourcesGraph.I3E_DAG_AddNode(i3e_derControlList.href, str_derControlListNode);
        resourcesGraph.I3E_DAG_AddEdge(sz_parentNodeHref, i3e_derControlList.href);
    }

    // Iterate through the list of DERControls and create graph nodes and edges for each 
    for(int k = 0; k < i3e_derControlList.DERControl.size(); k++) 
    {
        StackScop NodeData str_derControlNode(i3e_derControlList.DERControl.at(k).href,
                        i3e_derControlList.DERControl.at(k));
        str_derControlNode.enum_rsrcType = DERCTRL;
        str_derControlNode.bol_updated = true;
        resourcesGraph.I3E_DAG_AddNode(i3e_derControlList.DERControl.at(k).href, str_derControlNode);
        resourcesGraph.I3E_DAG_AddEdge(i3e_derControlList.href, i3e_derControlList.DERControl.at(k).href);
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_spp_storeDefaultDERControl(std::string sz_parentNodeHref, std::string sz_rxPayload) {
    
    // Create struct for payload storage   
    StackScop SE_DefaultDERControl_t i3e_defDerCtrl = {0};
    
    // Deserialize and store the DEFCON payload
    I3E_XML_PayloadDeserializeMux(sz_rxPayload, DEFDERCTRL, &i3e_defDerCtrl);

    StackScop NodeData str_defaultDerCtrlNode(i3e_defDerCtrl.href, i3e_defDerCtrl);
    str_defaultDerCtrlNode.enum_rsrcType = DEFDERCTRL;
    str_defaultDerCtrlNode.u32_pollRate = 0;
    str_defaultDerCtrlNode.sz_lastPollStr = sz_rxPayload;
    str_defaultDerCtrlNode.bol_updated = true;

    // If the node already exists, replace the contents and delete all children nodes for rediscovery
    if(resourcesGraph.I3E_DAG_DoesNodeExist(i3e_defDerCtrl.href)) 
    {
        resourcesGraph.I3E_DAG_UpdateNodeContent(i3e_defDerCtrl.href, str_defaultDerCtrlNode);
        resourcesGraph.I3E_DAG_DeleteAllChildren(i3e_defDerCtrl.href);
    }
    // If the node doesn't exist, add it to the graph normally and link the graph edges
    else
    {
        resourcesGraph.I3E_DAG_AddNode(i3e_defDerCtrl.href, str_defaultDerCtrlNode);
        resourcesGraph.I3E_DAG_AddEdge(sz_parentNodeHref, i3e_defDerCtrl.href);
    }
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_spp_storeDERCurveList(std::string sz_parentNodeHref, std::string sz_rxPayload)
{
    // Create struct for payload storage 
    StackScop SE_DERCurveList_t i3e_derCurveList = {0};

    // Deserialize and store the DERCURVE_LIST payload
    I3E_XML_PayloadDeserializeMux(sz_rxPayload, DERCURVE_LIST, &i3e_derCurveList);
    
    StackScop NodeData str_derCurveListNode(i3e_derCurveList.href, i3e_derCurveList);
    str_derCurveListNode.enum_rsrcType = DERCURVE_LIST;
    str_derCurveListNode.sz_lastPollStr = sz_rxPayload;
    str_derCurveListNode.bol_updated = true;

    // If the node already exists, replace the contents and delete all children nodes for rediscovery
    if(resourcesGraph.I3E_DAG_DoesNodeExist(i3e_derCurveList.href)) 
    {
        resourcesGraph.I3E_DAG_UpdateNodeContent(i3e_derCurveList.href, str_derCurveListNode);
        resourcesGraph.I3E_DAG_DeleteAllChildren(i3e_derCurveList.href);
    }
    // If the node doesn't exist, add it to the graph normally and link the graph edges
    else
    {
        resourcesGraph.I3E_DAG_AddNode(i3e_derCurveList.href, str_derCurveListNode);
        resourcesGraph.I3E_DAG_AddEdge(sz_parentNodeHref, i3e_derCurveList.href);
    }

    // Iterate through the list of DERCurves and create graph nodes and edges for each 
    for(int k = 0; k < i3e_derCurveList.DERCurve.size(); k++) 
    {
        StackScop NodeData str_derCurveNode(i3e_derCurveList.DERCurve.at(k).href, i3e_derCurveList.DERCurve.at(k));
        str_derCurveNode.enum_rsrcType = DERCURVE;
        str_derCurveNode.bol_updated = true;
        resourcesGraph.I3E_DAG_AddNode(i3e_derCurveList.DERCurve.at(k).href, str_derCurveNode);
        resourcesGraph.I3E_DAG_AddEdge(i3e_derCurveList.href, i3e_derCurveList.DERCurve.at(k).href);   
    }
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_spp_resourceRediscovery(std::string sz_rxPayload, NodeData* ptr_str_rediscoverNode)
{

    // Switch on type of resource for running a specific section of rediscovery, based on the depth of the node
    switch (ptr_str_rediscoverNode->enum_rsrcType) 
    {
        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        case DCAP:
        {
            // RETRIEVE + STORE --- DCAP 
            i3e_spp_storeDeviceCapability(sz_rxPayload);
            sz_rxPayload.clear();
            // resourcesGraph.I3E_DAG_PrintNodes("/dcap");

            // RE-DISCOVERY of children nodes
            i3e_spp_rediscoverDeviceCapability(ptr_str_rediscoverNode);
            break;
        }

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        case TM:
        {
            // RETRIEVE + STORE --- TM
            i3e_spp_storeTime("", sz_rxPayload);
            sz_rxPayload.clear();

            // RE-DISCOVERY of children nodes
            // i3e_spp_rediscoverTime(ptr_str_rediscoverNode);
            break;
        }

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        case EDEV_LIST:
        {
            // RETRIEVE + STORE --- EDEVlist & EDEVs
            i3e_spp_storeEndDeviceList("", sz_rxPayload);
            sz_rxPayload.clear();
            // resourcesGraph.I3E_DAG_PrintNodes("/dcap");

            // RE-DISCOVERY of children nodes
            i3e_spp_rediscoverEndDeviceList(ptr_str_rediscoverNode);
            break;
        }

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        case DER_LIST:
        {
            // // RETRIEVE + STORE --- DERlist & DERs
            // i3e_spp_storeEndDeviceList("", sz_rxPayload);
            // sz_rxPayload.clear();

            // std::cout << "EDEVLIST REDISCOVERY  " <<std::endl;
            // resourcesGraph.I3E_DAG_PrintNodes("/dcap");

            // // RE-DISCOVERY of children nodes
            // i3e_spp_rediscoverEndDeviceList(ptr_str_rediscoverNode);
            break;
        }

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        case FSA_LIST:
        {
            // RETRIEVE + STORE ---  FSAlist & FSAs
            i3e_spp_storeFunctionSetAssignmentList("", sz_rxPayload);
            sz_rxPayload.clear();
            // resourcesGraph.I3E_DAG_PrintNodes("/dcap");

            // RE-DISCOVERY of children nodes
            i3e_spp_rediscoverFunctionSetAssignmentList(ptr_str_rediscoverNode);
            break;
        }

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        case DERPROG_LIST:
        {
            // RETRIEVE + STORE --- DERProgramlist and DERPrograms
            i3e_spp_storeDERProgramList("", sz_rxPayload);
            sz_rxPayload.clear();
            // resourcesGraph.I3E_DAG_PrintNodes("/dcap");

            // RE-DISCOVERY of children nodes
            i3e_spp_rediscoverDERProgramList(ptr_str_rediscoverNode);
            break;
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//#TODO - consolidate later into one function for all rediscovery?
FileScop void i3e_spp_rediscoverDeviceCapability(NodeData* ptr_str_rediscoverNode)
{
    // std::cout << "REDISCOVERY ROUTINE - DCAP/CHILDREN NODES -- " <<std::endl;

}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void i3e_spp_rediscoverTime(NodeData* ptr_str_rediscoverNode)
{
    // std::cout << "REDISCOVERY ROUTINE - TIME/CHILDREN NODES -- " <<std::endl;

}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


FileScop void i3e_spp_rediscoverEndDeviceList(NodeData* ptr_str_rediscoverNode)
{
    StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
    StackScop uint16_t u16_port = I3E_NSS_GetPORT();
    StackScop std::string sz_rxPayload;
    StackScop SE_EndDeviceList_t& i3e_edevListVariant = std::get<SE_EndDeviceList_t>(ptr_str_rediscoverNode->var_rsrc);

    // Loop through all end devices in edev list 
    for (uint8_t i = 0; i < i3e_edevListVariant.EndDevice.size(); i++)
    {
        if(!(i3e_edevListVariant.EndDevice.at(i).RegistrationLink.href.empty()))
        {   
            // Send HTTP GET Method to REG resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            i3e_edevListVariant.EndDevice.at(i).RegistrationLink.href,
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- REGISTRATION
                i3e_spp_storeRegistration(i3e_edevListVariant.EndDevice.at(i).href, sz_rxPayload);
                sz_rxPayload.clear();
            }
            else{}
        }
        else{}

        // #TODO - this is where we will have to check the registrations of the end devices to match sFDI, lFDI and pin #
        // #TODO - If registration matches, pull the FSAs 
        if(!(i3e_edevListVariant.EndDevice.at(i).FunctionSetAssignmentsListLink.href.empty()))
        {           
            // Send HTTP GET Method to FSALIST resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_edevListVariant.EndDevice.at(i).FunctionSetAssignmentsListLink.href + MAX_RSRC_RET_SUFFIX),
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE ---  FSAlist & FSAs
                i3e_spp_storeFunctionSetAssignmentList(i3e_edevListVariant.EndDevice.at(i).href, sz_rxPayload);
                sz_rxPayload.clear();
            }
            else{}
        }
        else{}               
    }

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Have to collect all the new FSA resources added and rediscover those 
    StackScop std::vector<std::string> vec_fsaHrefs = 
                    resourcesGraph.I3E_DAG_FindResourceNodesFromNode(i3e_edevListVariant.href, FSA);

    // Loop through all functionsetassignments in fsa list 
    for (uint8_t i = 0; i < vec_fsaHrefs.size(); i++)
    {
        StackScop NodeData str_fsaNode = resourcesGraph.I3E_DAG_GetNode(vec_fsaHrefs.at(i));
        StackScop SE_FunctionSetAssignments_t& i3e_fsaStored = std::get<SE_FunctionSetAssignments_t>(str_fsaNode.var_rsrc);

        if(!(i3e_fsaStored.DERProgramListLink.href.empty()))
        {   
            // Send HTTP GET Method to DERPROGLIST resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_fsaStored.DERProgramListLink.href + MAX_RSRC_RET_SUFFIX),
                             sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERProgramlist and DERPrograms
                i3e_spp_storeDERProgramList(i3e_fsaStored.href, sz_rxPayload);
                sz_rxPayload.clear();
            }
            // HTTP GET FAILURE - exit 
            else{}
        }
        else{}
    }

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Have to collect all the new DERPROG resources added and rediscover those 
    StackScop std::vector<std::string> vec_derprogHrefs = 
                    resourcesGraph.I3E_DAG_FindResourceNodesFromNode(i3e_edevListVariant.href, DERPROG);

    // Loop through all derprograms in derprogram list 
    for (uint8_t i = 0; i < vec_derprogHrefs.size(); i++)
    {
        StackScop NodeData str_derprogNode = resourcesGraph.I3E_DAG_GetNode(vec_derprogHrefs.at(i));
        StackScop SE_DERProgram_t& i3e_derProgramStored = std::get<SE_DERProgram_t>(str_derprogNode.var_rsrc);

        if(!(i3e_derProgramStored.DefaultDERControlLink.href.empty()))
        {   
            // Send HTTP GET Method to DefaultDERControl resource
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_derProgramStored.DefaultDERControlLink.href + MAX_RSRC_RET_SUFFIX),
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DefaultDERControlList and DefaultDERControls
                i3e_spp_storeDefaultDERControl(i3e_derProgramStored.href, sz_rxPayload);
                sz_rxPayload.clear();
            }
            // HTTP GET FAILURE - exit 
            else{}
        }
        else{}     
 
        if(!(i3e_derProgramStored.DERControlListLink.href.empty()))
        {   
            // Send HTTP GET Method to DERControlList resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_derProgramStored.DERControlListLink.href + MAX_RSRC_RET_SUFFIX),
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERControlList and DERControls
                i3e_spp_storeDERControlList(i3e_derProgramStored.href, sz_rxPayload);
                sz_rxPayload.clear();
            }
            // HTTP GET FAILURE - exit 
            else{}
        }
        else{}       

        if(!(i3e_derProgramStored.DERCurveListLink.href.empty()))
        {   
            // Send HTTP GET Method to DERCurveList resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_derProgramStored.DERCurveListLink.href + MAX_RSRC_RET_SUFFIX), 
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERCurveList and DERCurves
                i3e_spp_storeDERCurveList(i3e_derProgramStored.href, sz_rxPayload);
                sz_rxPayload.clear();
            }
            // HTTP GET FAILURE - exit 
            else{}
        }
        else{}
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//#TODO - consolidate later into one function for all rediscovery?
FileScop void i3e_spp_rediscoverFunctionSetAssignmentList(NodeData* ptr_str_rediscoverNode)
{
    StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
    StackScop uint16_t u16_port = I3E_NSS_GetPORT();
    StackScop std::string sz_rxPayload;
    StackScop SE_FunctionSetAssignmentsList_t& i3e_fsaListStored =
                    std::get<SE_FunctionSetAssignmentsList_t>(ptr_str_rediscoverNode->var_rsrc);

    // Loop through all functionsetassignments in fsa list 
    for (uint8_t i = 0; i < i3e_fsaListStored.FunctionSetAssignments.size(); i++)
    {
        if(!(i3e_fsaListStored.FunctionSetAssignments.at(i).DERProgramListLink.href.empty()))
        {   
            // Send HTTP GET Method to DERPROGLIST resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_fsaListStored.FunctionSetAssignments.at(i).DERProgramListLink.href + MAX_RSRC_RET_SUFFIX),
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERProgramlist and DERPrograms
                i3e_spp_storeDERProgramList(i3e_fsaListStored.FunctionSetAssignments.at(i).href, sz_rxPayload);
                sz_rxPayload.clear();
            }
            // HTTP GET FAILURE - exit 
            else{}
        }
        else{}
    }
    
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Have to collect all the new DERPROG resources added and rediscover those 
    StackScop std::vector<std::string> vec_derprogHrefs =
                    resourcesGraph.I3E_DAG_FindResourceNodesFromNode(i3e_fsaListStored.href, DERPROG);

    // Loop through all derprograms in derprogram list 
    for (uint8_t i = 0; i < vec_derprogHrefs.size(); i++)
    {

        StackScop NodeData str_derprogNode = resourcesGraph.I3E_DAG_GetNode(vec_derprogHrefs.at(i));
        StackScop SE_DERProgram_t& i3e_derProgramStored = std::get<SE_DERProgram_t>(str_derprogNode.var_rsrc);

        if(!(i3e_derProgramStored.DefaultDERControlLink.href.empty()))
        {        
            // Send HTTP GET Method to DefaultDERControl resource
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_derProgramStored.DefaultDERControlLink.href + MAX_RSRC_RET_SUFFIX),
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DefaultDERControlList and DefaultDERControls
                i3e_spp_storeDefaultDERControl(i3e_derProgramStored.href, sz_rxPayload);
                sz_rxPayload.clear();
            }
            // HTTP GET FAILURE - exit 
            else{}
        }
        else{}  

        if(!(i3e_derProgramStored.DERControlListLink.href.empty()))
        {   
            // Send HTTP GET Method to DERControlList resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_derProgramStored.DERControlListLink.href + MAX_RSRC_RET_SUFFIX),
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERControlList and DERControls
                i3e_spp_storeDERControlList(i3e_derProgramStored.href, sz_rxPayload);
                sz_rxPayload.clear();
            }
            // HTTP GET FAILURE - exit 
            else{}
        }
        else{}

        if(!(i3e_derProgramStored.DERCurveListLink.href.empty()))
        {   
            // Send HTTP GET Method to DERCurveList resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_derProgramStored.DERCurveListLink.href + MAX_RSRC_RET_SUFFIX), 
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERCurveList and DERCurves
                i3e_spp_storeDERCurveList(i3e_derProgramStored.href, sz_rxPayload);
                sz_rxPayload.clear();
            }
            // HTTP GET FAILURE - exit 
            else{}
        }
        else{}
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//#TODO - consolidate later into one function for all rediscovery?
FileScop void i3e_spp_rediscoverDERProgramList(NodeData* ptr_str_rediscoverNode)
{
    StackScop std::string sz_rxPayload;
    StackScop SE_DERProgramList_t& i3e_derProgListStored = std::get<SE_DERProgramList_t>(ptr_str_rediscoverNode->var_rsrc);
    StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
    StackScop uint16_t u16_port = I3E_NSS_GetPORT();

    // Loop through all derprograms in derprogram list 
    for (uint8_t i = 0; i < i3e_derProgListStored.DERProgram.size(); i++)
    {
        if(!(i3e_derProgListStored.DERProgram.at(i).DefaultDERControlLink.href.empty()))
        {           
            // Send HTTP GET Method to DefaultDERControl resource
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_derProgListStored.DERProgram.at(i).DefaultDERControlLink.href + MAX_RSRC_RET_SUFFIX),
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DefaultDERControlList and DefaultDERControls
                i3e_spp_storeDefaultDERControl(i3e_derProgListStored.DERProgram.at(i).href, sz_rxPayload);
                sz_rxPayload.clear();
            }
            // HTTP GET FAILURE - exit 
            else{}
        }
        else{}

        if(!(i3e_derProgListStored.DERProgram.at(i).DERControlListLink.href.empty()))
        {   
            // Send HTTP GET Method to DERControlList resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_derProgListStored.DERProgram.at(i).DERControlListLink.href + MAX_RSRC_RET_SUFFIX), 
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERControlList and DERControls
                i3e_spp_storeDERControlList(i3e_derProgListStored.DERProgram.at(i).href, sz_rxPayload);
                sz_rxPayload.clear();
            }
            // HTTP GET FAILURE - exit 
            else{}
        }
        else{}

        if(!(i3e_derProgListStored.DERProgram.at(i).DERCurveListLink.href.empty()))
        {  
            // Send HTTP GET Method to DERCurveList resource 
            if(I3E_HTTP_GET(sz_ipv4Addr, u16_port, 
                            (i3e_derProgListStored.DERProgram.at(i).DERCurveListLink.href + MAX_RSRC_RET_SUFFIX),
                            sz_rxPayload) == HTTPResponse::OK) 
            {
                // RETRIEVE + STORE --- DERCurveList and DERCurves
                i3e_spp_storeDERCurveList(i3e_derProgListStored.DERProgram.at(i).href, sz_rxPayload);
                sz_rxPayload.clear();
            }
            // HTTP GET FAILURE - exit 
            else{}
        }
        else{}
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void I3E_SPP_RsrcStorageSwitch(I3EresourceType enum_rsrcType, std::string &sz_rxPayload, std::string &parentHref) 
{

    // REG,
    // EDEV,
    // FSA_LIST,
    // FSA,
    // DERPROG_LIST,
    // DERPROG, 
    // DERCTRL_LIST,
    // DERCTRL,
    // DERCURVE_LIST,
    // DERCURVE,
    // DEFDERCTRL,
    // DEFDERCTRL_LIST,
    // DER_LIST,
    // DER,
    // DERSTAT,
    // DERCAP,
    // DERSETT,
    // DERAVAIL,
    // MMR,
    // LOGEVENT

    switch(enum_rsrcType) {
        case DCAP:
        {
            // RETRIEVE + STORE --- DCAP 
            i3e_spp_storeDeviceCapability(sz_rxPayload);
            sz_rxPayload.clear();
            break;
        }
        case TM:
        {
            // RETRIEVE + STORE --- TM
            i3e_spp_storeTime(parentHref, sz_rxPayload);
            sz_rxPayload.clear();
            break;
        }
        case EDEV_LIST:
        {
            // RETRIEVE + STORE --- TM
            i3e_spp_storeEndDeviceList(parentHref, sz_rxPayload);
            sz_rxPayload.clear();
            break;
        }        
        case MUP:
        {
            // RETRIEVE + STORE --- MUP
            I3E_SPP_StoreMirrorUsagePoint(parentHref, sz_rxPayload);
            sz_rxPayload.clear();
            break;
        }
        default:
        {
            break;
        }
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop bool i3e_spp_httpGetRsp(I3EresourceType enum_rsrcType, HTTPResponse enum_respStat, 
                                std::string &sz_rxPayload, std::string &sz_parentHref)
{

    StackScop std::string sz_ipv4Addr = I3E_NSS_GetIPV4();
    StackScop uint16_t u16_port = I3E_NSS_GetPORT();
   
    // OK Response - process the data normally 
    if(enum_respStat == HTTPResponse::OK) 
    {
        I3E_SPP_RsrcStorageSwitch(enum_rsrcType, sz_rxPayload, sz_parentHref);
        return true;
    }

    // Moved or Found Response - get new URI location 
    else if (enum_respStat == HTTPResponse::MOVED_PERMANENTLY || enum_respStat == HTTPResponse::FOUND) 
    {
        StackScop std::string sz_path;

        // Find the position of the third '/' which will be the start of the path
        StackScop std::size_t siz_pathStart = sz_rxPayload.find('/', sz_rxPayload.find("//") + 2);
        if (siz_pathStart != std::string::npos) 
        {
            sz_path = sz_rxPayload.substr(siz_pathStart);
            sz_rxPayload.clear();
            enum_respStat = I3E_HTTP_GET(sz_ipv4Addr, u16_port, sz_path, sz_rxPayload);

            // OK Response - process the data normally 
            if(enum_respStat == HTTPResponse::OK) 
            {
                I3E_NSS_SetURI(sz_path);
                I3E_SPP_RsrcStorageSwitch(enum_rsrcType, sz_rxPayload, sz_parentHref);
            }
            // NOT OK - error
            else 
            {
                return false;
            }
        } 
        // Not found - error case 
        else 
        {
            return false;
        }
    }

    // Internal server error, failed discovery 
    else if (enum_respStat == HTTPResponse::INTERNAL_SERVER_ERROR) 
    {
        return false;
    }

    return false;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~