#include "include/I3E_XMLS.hpp"
#include "include/I3E_CTS.hpp"

#include "Poco/XML/XMLWriter.h"
#include "Poco/DOM/Text.h"
#include "Poco/DOM/DOMWriter.h"

#include "lion/log/log.h"

LOG_MODULE_REGISTER(i3e_xmls, LOG_LEVEL_INF);

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

/**
 * @brief Pad a number with leading zeros to a fixed size
 * 
 * @details This function pads a number with leading zeros to a fixed size to make sure specific fields are always 
 * the same size, as required by the IEEE 2030.5 standard.
 * 
 * @param numToPad The number to pad with leading zeros.
 * 
 * @returns A string containing the padded number.
 */
FileScop std::string i3e_xmls_toPaddedSz(int numToPad);

/**
 * @brief Convert a byte array to a hex string
 * 
 * @details This function converts a byte array to a hex string. There are a few fields in the IEEE 2030.5 XML schema
 * that are stored as byte arrays, and this function is used to convert them to hex strings for serialization.
 * 
 * @param ptr_u8_byteArray The byte array to convert to a hex string.
 * @param siz_byteArraySize The size of the byte array.
 * 
 * @returns A string containing the hex representation of the byte array.
 */
FileScop std::string i3e_xmls_byteArrToHexSz(const uint8_t* ptr_u8_byteArray, size_t siz_byteArraySize);

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

ModScop void I3E_XML_SRLZ_DERControlResponse(std::string& sz_txPayload, SE_DERControlResponse_t* i3e_ptr_derCtrlRsp) 
{
    StackScop std::ostringstream oss_xmlStream;

    // Build a DOM document 
    StackScop Poco::AutoPtr<Poco::XML::Document> ptr_pdoc_xmlDoc = new Poco::XML::Document;
    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_xmlRoot = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCRSP);
    ptr_pdoc_xmlDoc->appendChild(ptr_pel_xmlRoot);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XMLNS, XSD_TAG_NS);

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    StackScop std::stringstream ss_convertLFDI;
    ss_convertLFDI << std::hex << std::setfill('0');
    for (int a = 0; a < sizeof(i3e_ptr_derCtrlRsp->endDeviceLFDI); a++) {
        ss_convertLFDI << std::hex << std::setw(2) << static_cast<int>(i3e_ptr_derCtrlRsp->endDeviceLFDI[a]);
    }
    StackScop std::string sz_stringLFDI = ss_convertLFDI.str();


    StackScop std::stringstream ss_convertSubject;
    ss_convertSubject << std::hex << std::setfill('0');
    for (int a = 0; a < sizeof(i3e_ptr_derCtrlRsp->subject); a++) {
        ss_convertSubject << std::hex << std::setw(2) << static_cast<int>(i3e_ptr_derCtrlRsp->subject[a]);
    }
    StackScop std::string sz_stringSubject = ss_convertSubject.str();


    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                                ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCRSP_CRDT);
    StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                                ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derCtrlRsp->createdDateTime));
    ptr_pel_child1->appendChild(ptr_ptex_ctex1);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child1);

    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child2 = 
                                ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCRSP_EDEVLFDI);
    StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex2 = 
                                ptr_pdoc_xmlDoc->createTextNode(sz_stringLFDI); 
    ptr_pel_child2->appendChild(ptr_ptex_ctex2);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child2);

    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child3 = 
                                ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCRSP_STAT);
    StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex3 = 
                                ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derCtrlRsp->status));
    ptr_pel_child3->appendChild(ptr_ptex_ctex3);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child3);

    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child4 = 
                                ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCRSP_SUBJECT);
    StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex4 = 
                                ptr_pdoc_xmlDoc->createTextNode(sz_stringSubject); 
    ptr_pel_child4->appendChild(ptr_ptex_ctex4);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child4);

    StackScop Poco::XML::DOMWriter pdw_xmlWriter;
    pdw_xmlWriter.setNewLine("\n");
    // pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::PRETTY_PRINT);  // set pretty print for nice debugger view
    pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::CANONICAL_XML);
    pdw_xmlWriter.writeNode(oss_xmlStream, ptr_pdoc_xmlDoc);
    sz_txPayload = oss_xmlStream.str();
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void I3E_XML_SRLZ_DERSettings(std::string& sz_txPayload, SE_DERSettings_t* i3e_ptr_derSettings)
{
    StackScop std::ostringstream oss_xmlStream;

    // Build a DOM document 
    StackScop Poco::AutoPtr<Poco::XML::Document> ptr_pdoc_xmlDoc = new Poco::XML::Document;
    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_xmlRoot = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT);
    ptr_pdoc_xmlDoc->appendChild(ptr_pel_xmlRoot);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XMLNS, XSD_TAG_NS);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XSI, XSD_TAG_XSIURL);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XSISL, XSD_TAG_XSD);

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // ModesEnabled attribute
    if(i3e_ptr_derSettings->_flags & SE_modesEnabled_exists) 
    {
        StackScop std::stringstream ss_convertModesEn;
        ss_convertModesEn << std::hex << std::setfill('0');
        for (int a = 0; a < sizeof(i3e_ptr_derSettings->modesEnabled); a++) {
            ss_convertModesEn << std::hex << std::setw(2) << static_cast<int>(i3e_ptr_derSettings->modesEnabled[a]);
        }
        StackScop std::string sz_stringModesEn = ss_convertModesEn.str();

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_MODESEN);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                            ptr_pdoc_xmlDoc->createTextNode(sz_stringModesEn);
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setESDelay attribute
    if(i3e_ptr_derSettings->_flags & SE_setESDelay_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETESDEL);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                            ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derSettings->setESDelay));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setESHighFreq attribute
    if(i3e_ptr_derSettings->_flags & SE_setESHighFreq_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETESHF);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                            ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derSettings->setESHighFreq));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}                
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setESHighVolt attribute
    if(i3e_ptr_derSettings->_flags & SE_setESHighVolt_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETESHV);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                            ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derSettings->setESHighVolt));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setESLowFreq attribute   
    if(i3e_ptr_derSettings->_flags & SE_setESLowFreq_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETESLF);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                            ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derSettings->setESLowFreq));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}        
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setESLowVolt attribute
    if(i3e_ptr_derSettings->_flags & SE_setESLowVolt_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETESLV);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                            ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derSettings->setESLowVolt));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setESRampTms attribute   
    if(i3e_ptr_derSettings->_flags & SE_setESRampTms_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETESRT);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                            ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derSettings->setESRampTms));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setESRandomDelay attribute  
    if(i3e_ptr_derSettings->_flags & SE_setESRandomDelay_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETESRD);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                            ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derSettings->setESRandomDelay));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}        

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // SetGradW attribute
    if(i3e_ptr_derSettings->_flags & SE_setGradW_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETGRADW);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                            ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derSettings->setGradW));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setMaxA attribute
    if(i3e_ptr_derSettings->_flags & SE_setMaxA_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETMAXA);
        
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxA.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxA.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1); 
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setMaxAh attribute
    if(i3e_ptr_derSettings->_flags & SE_setMaxAh_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETMAXAH);
        
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxAh.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxAh.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1); 
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setMaxChargeRateVA attribute
    if(i3e_ptr_derSettings->_flags & SE_setMaxChargeRateVA_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETMAXCHRGRVA);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxDischargeRateVA.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxDischargeRateVA.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1); 
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setMaxChargeRateW attribute
    if(i3e_ptr_derSettings->_flags & SE_setMaxChargeRateW_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETMAXCHRGRW);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxChargeRateW.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxChargeRateW.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1); 
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setMaxDischargeRateVA attribute
    if(i3e_ptr_derSettings->_flags & SE_setMaxDischargeRateVA_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETMAXDCHRGRVA);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxDischargeRateVA.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxDischargeRateVA.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1); 
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setMaxDischargeRateW attribute
    if(i3e_ptr_derSettings->_flags & SE_setMaxDischargeRateW_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETMAXDCHRGRW);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxDischargeRateW.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxDischargeRateW.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1); 
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setMaxV attribute
    if(i3e_ptr_derSettings->_flags & SE_setMaxV_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETMAXV);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxV.multiplier)));
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxV.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1); 
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setMaxVA attribute
    if(i3e_ptr_derSettings->_flags & SE_setMaxVA_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETMAXVA);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxVA.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxVA.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1); 
    }
    else{}        
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setMaxVar attribute
    if(i3e_ptr_derSettings->_flags & SE_setMaxVar_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETMAXVAR);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxVar.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxVar.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1); 
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setMaxVarNeg attribute
    if(i3e_ptr_derSettings->_flags & SE_setMaxVarNeg_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETMAXVARNEG);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxVarNeg.multiplier)));
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxVarNeg.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1); 
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // SetMaxW attribute
    if(i3e_ptr_derSettings->_flags & SE_setMaxW_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETMAXW);
        
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxW.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxW.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setMaxWh attribute
    if(i3e_ptr_derSettings->_flags & SE_setMaxWh_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETMAXWH);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxWh.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMaxWh.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setMinPFOverExcited attribute 
    if(i3e_ptr_derSettings->_flags & SE_setMinPFOverExcited_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETMINPFOE);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DISPLACEMENT);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMinPFOverExcited.displacement))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMinPFOverExcited.multiplier)));
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setMinPFUnderExcited attribute
    if(i3e_ptr_derSettings->_flags & SE_setMinPFUnderExcited_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETMINPFUE);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DISPLACEMENT);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMinPFUnderExcited.displacement))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMinPFUnderExcited.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setMinV attribute   
    if(i3e_ptr_derSettings->_flags & SE_setMinV_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETMINV);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMinV.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value =
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setMinV.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setSoftGradW attribute
    if(i3e_ptr_derSettings->_flags & SE_setSoftGradW_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETSOFTGRADW);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                            ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setSoftGradW));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setVNom attribute   
    if(i3e_ptr_derSettings->_flags & SE_setVNom_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETVNOM);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setVNom.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setVNom.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setVRef attribute
    if(i3e_ptr_derSettings->_flags & SE_setVRef_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETVREF);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setVRef.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setVRef.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // setVRefOfs attribute   
    if(i3e_ptr_derSettings->_flags & SE_setVRefOfs_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_SETVREFOFS);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setVRefOfs.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derSettings->setVRefOfs.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // UpdatedTime attribute
    // SE_TimeType_t currentTime = I3E_CTS_getSyncTime();

    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                        ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSETT_UPDATEDTM);
    StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                        ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derSettings->updatedTime));
    ptr_pel_child1->appendChild(ptr_ptex_ctex1);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child1);

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~      
    // Consolidate XML Object

    StackScop Poco::XML::DOMWriter pdw_xmlWriter;
    pdw_xmlWriter.setNewLine("\n");
    // pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::PRETTY_PRINT);  // set pretty print for nice debugger view
    pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::CANONICAL_XML);
    pdw_xmlWriter.writeNode(oss_xmlStream, ptr_pdoc_xmlDoc);
    sz_txPayload = oss_xmlStream.str();
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void I3E_XML_SRLZ_DERCapability(std::string& sz_txPayload, SE_DERCapability_t* i3e_ptr_derCapability)
{
    StackScop std::ostringstream oss_xmlStream;

    // Build a DOM document 
    StackScop Poco::AutoPtr<Poco::XML::Document> ptr_pdoc_xmlDoc = new Poco::XML::Document;

    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_xmlRoot = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP);
    ptr_pdoc_xmlDoc->appendChild(ptr_pel_xmlRoot);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XMLNS, XSD_TAG_NS);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XSI, XSD_TAG_XSIURL);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XSISL, XSD_TAG_XSD);

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // modesSupported attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgModesSupported_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 
                            = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_MODESUP);

        StackScop std::stringstream ss_convertModesSupp;
        ss_convertModesSupp << std::hex << std::setfill('0');
        for (int a = 0; a < sizeof(i3e_ptr_derCapability->modesSupported); a++) {
            ss_convertModesSupp << std::hex << std::setw(2) << static_cast<int>(i3e_ptr_derCapability->modesSupported[a]);
        }
        StackScop std::string sz_stringModesSupp = ss_convertModesSupp.str();

        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                            ptr_pdoc_xmlDoc->createTextNode(sz_stringModesSupp);
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgAbnormalCategory attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgAbnormalCategory_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_ABCTG);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                            ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derCapability->rtgAbnormalCategory));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgMaxA attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgMaxA_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_MAXA);
        
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxA.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxA.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgMaxAh attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgMaxAh_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_MAXAH);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxAh.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxAh.value)));
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgMaxChargeRateVA attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgMaxChargeRateVA_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_MAXCHRGRVA);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxChargeRateVA.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxChargeRateVA.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgMaxChargeRateW attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgMaxChargeRateW_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_MAXCHRGRW);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxChargeRateW.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxChargeRateW.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgMaxDischargeRateVA attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgMaxDischargeRateVA_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_MAXDCHRGRVA);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxDischargeRateVA.multiplier)));
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxDischargeRateVA.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgMaxDischargeRateW attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgMaxDischargeRateW_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_MAXDCHRGRW);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxDischargeRateW.multiplier)));
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxDischargeRateW.value)));
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgMaxV attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgMaxV_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_MAXV);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxV.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxV.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgMaxVA attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgMaxVA_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_MAXVA);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxVA.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxVA.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgMaxVar attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgMaxVar_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_MAXVAR);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxVar.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxVar.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgMaxVarNeg attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgMaxVarNeg_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_MAXVARNEG);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxVarNeg.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxVarNeg.value)));
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgMaxW attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgMaxW_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_MAXW);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxW.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxW.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgMaxWh attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgMaxWh_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_MAXWH);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxWh.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMaxWh.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgMinPFOverExcited attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgMinPFOverExcited_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_MINPFOE);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DISPLACEMENT);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMinPFOverExcited.displacement))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMinPFOverExcited.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgMinPFUnderExcited attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgMinPFUnderExcited_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_MINPRUE);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DISPLACEMENT);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMinPFUnderExcited.displacement))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMinPFUnderExcited.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgMinV attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgMinV_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_MINV);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMinV.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgMinV.value)));
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgNormalCategory attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgNormalCategory_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_NORMCAT);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                            ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derCapability->rtgNormalCategory));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgOverExcitedPF attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgOverExcitedPF_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_OEPF);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DISPLACEMENT);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgOverExcitedPF.displacement))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgOverExcitedPF.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgOverExcitedW attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgOverExcitedW_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_OEW);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgOverExcitedW.multiplier)));
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgOverExcitedW.value)));
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgReactiveSusceptance attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgReactiveSusceptance_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_REACSUCEP);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgReactiveSusceptance.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgReactiveSusceptance.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgUnderExcitedPF attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgUnderExcitedPF_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_UEPF);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DISPLACEMENT);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgUnderExcitedPF.displacement))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgUnderExcitedPF.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgUnderExcitedW attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgUnderExcitedW_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_UEW);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgUnderExcitedW.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgUnderExcitedW.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // rtgVNom attribute
    if(i3e_ptr_derCapability->_flags & SE_rtgVNom_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_RTG_VNOM);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgVNom.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derCapability->rtgVNom.value))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // type attribute
    if(i3e_ptr_derCapability->_flags & SE_type_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERCAP_TYPE);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                            ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derCapability->type));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~      
    // Consolidate XML Object

    StackScop Poco::XML::DOMWriter pdw_xmlWriter;
    pdw_xmlWriter.setNewLine("\n");
    // pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::PRETTY_PRINT);  // set pretty print for nice debugger view
    pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::CANONICAL_XML);
    pdw_xmlWriter.writeNode(oss_xmlStream, ptr_pdoc_xmlDoc);

    sz_txPayload = oss_xmlStream.str();
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void I3E_XML_SRLZ_DERStatus(std::string& sz_txPayload, SE_DERStatus_t* i3e_ptr_derStatus)
{
    StackScop std::ostringstream oss_xmlStream;

    // Build a DOM document 
    StackScop Poco::AutoPtr<Poco::XML::Document> ptr_pdoc_xmlDoc = new Poco::XML::Document;
    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_xmlRoot = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSTAT);
    ptr_pdoc_xmlDoc->appendChild(ptr_pel_xmlRoot);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XMLNS, XSD_TAG_NS);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XSI, XSD_TAG_XSIURL);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XSISL, XSD_TAG_XSD);

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // alarmStatus attribute
    if(i3e_ptr_derStatus->_flags & SE_alarmStatus_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSTAT_ALARMSTAT);

        StackScop std::stringstream ss_convertalarmStatus;
        ss_convertalarmStatus << std::hex << std::setfill('0');
        for (int a = 0; a < sizeof(i3e_ptr_derStatus->alarmStatus); a++) {
            ss_convertalarmStatus << std::hex << std::setw(2) << static_cast<int>(i3e_ptr_derStatus->alarmStatus[a]);
        }
        StackScop std::string sz_stringalarmStatus = ss_convertalarmStatus.str();

        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(sz_stringalarmStatus);
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // genConnectStatus attribute
    if(i3e_ptr_derStatus->_flags & SE_genConnectStatus_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSTAT_GENCONSTAT);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DATETIME);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derStatus->readingTime))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (i3e_xmls_toPaddedSz(i3e_ptr_derStatus->genConnectStatus.value))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // inverterStatus attribute
    if(i3e_ptr_derStatus->_flags & SE_inverterStatus_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSTAT_INVERTSTAT);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DATETIME);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derStatus->readingTime))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derStatus->inverterStatus.value))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // localControlModeStatus attribute
    if(i3e_ptr_derStatus->_flags & SE_localControlModeStatus_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSTAT_LOCALCTRLMODESTAT);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DATETIME);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derStatus->readingTime))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derStatus->localControlModeStatus.value))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // manufacturerStatus attribute
    if(i3e_ptr_derStatus->_flags & SE_manufacturerStatus_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSTAT_MANUFCSTAT);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DATETIME);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derStatus->readingTime))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            ((i3e_ptr_derStatus->manufacturerStatus.value))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // operationalModeStatus attribute
    if(i3e_ptr_derStatus->_flags & SE_operationalModeStatus_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSTAT_OPMODESTAT);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DATETIME);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derStatus->readingTime))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derStatus->operationalModeStatus.value))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // readingTime attribute
    if(i3e_ptr_derStatus->_flags & SE_readingTime_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSTAT_READTIME);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                            ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derStatus->readingTime));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // stateOfChargeStatus attribute
    if(i3e_ptr_derStatus->_flags & SE_stateOfChargeStatus_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSTAT_SOCSTAT);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DATETIME);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derStatus->readingTime))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derStatus->stateOfChargeStatus.value))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // storageModeStatus attribute
    if(i3e_ptr_derStatus->_flags & SE_storageModeStatus_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSTAT_STORAGEMODESTAT);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DATETIME);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derStatus->readingTime))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derStatus->storageModeStatus.value))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // storConnectStatus attribute
    if(i3e_ptr_derStatus->_flags & SE_storConnectStatus_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                            ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERSTAT_STORCONNSTAT);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DATETIME);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_derStatus->readingTime))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                            (i3e_xmls_toPaddedSz(i3e_ptr_derStatus->storConnectStatus.value))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~      
    // Consolidate XML Object

    StackScop Poco::XML::DOMWriter pdw_xmlWriter;
    pdw_xmlWriter.setNewLine("\n");
    // pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::PRETTY_PRINT);  // set pretty print for nice debugger view
    pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::CANONICAL_XML);
    pdw_xmlWriter.writeNode(oss_xmlStream, ptr_pdoc_xmlDoc);

    sz_txPayload = oss_xmlStream.str();
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void I3E_XML_SRLZ_DERAvailability(std::string& sz_txPayload, SE_DERAvailability_t* i3e_ptr_derAvailability)
{
    StackScop std::ostringstream oss_xmlStream;

    // Build a DOM document 
    StackScop Poco::AutoPtr<Poco::XML::Document> ptr_pdoc_xmlDoc = new Poco::XML::Document;
    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_xmlRoot = ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERAVAIL);
    ptr_pdoc_xmlDoc->appendChild(ptr_pel_xmlRoot);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XMLNS, XSD_TAG_NS);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XSI, XSD_TAG_XSIURL);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XSISL, XSD_TAG_XSD);

    
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // availabilityDuration attribute
    if(i3e_ptr_derAvailability->_flags & SE_availabilityDuration_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                        ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERAVAIL_AVAILDUR);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                        ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derAvailability->availabilityDuration));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // maxChargeDuration attribute
    if(i3e_ptr_derAvailability->_flags & SE_maxChargeDuration_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                        ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERAVAIL_MAXCHRGDUR);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                        ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derAvailability->maxChargeDuration));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // readingTime attribute
    if(i3e_ptr_derAvailability->_flags & SE_readingTime_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                        ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERAVAIL_READTIME);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                        ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derAvailability->readingTime));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // reserveChargePercent attribute
    if(i3e_ptr_derAvailability->_flags & SE_reserveChargePercent_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                        ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERAVAIL_RESCHRGPERC);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                        ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derAvailability->reserveChargePercent));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // reservePercent attribute
    if(i3e_ptr_derAvailability->_flags & SE_reservePercent_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                        ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERAVAIL_RESPERC);
        StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1 = 
                        ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_derAvailability->reservePercent));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // statVarAvail attribute
    if(i3e_ptr_derAvailability->_flags & SE_statVarAvail_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                        ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERAVAIL_STATVARAVAIL);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                        (std::to_string(i3e_ptr_derAvailability->statVarAvail.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                        (std::to_string(i3e_ptr_derAvailability->statVarAvail.value))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // statWAvail attribute
    if(i3e_ptr_derAvailability->_flags & SE_statWAvail_exists) 
    {
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1 = 
                        ptr_pdoc_xmlDoc->createElement(XSD_TAG_DERAVAIL_STATWAAVAIL);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_value = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MULT);
        ptr_pel_value->appendChild(ptr_pdoc_xmlDoc->createTextNode
                        (std::to_string(i3e_ptr_derAvailability->statWAvail.multiplier))); 
        ptr_pel_child1->appendChild(ptr_pel_value);

        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_multiplier = ptr_pdoc_xmlDoc->createElement(XSD_TAG_VAL);
        ptr_pel_multiplier->appendChild(ptr_pdoc_xmlDoc->createTextNode
                        (std::to_string(i3e_ptr_derAvailability->statWAvail.value))); 
        ptr_pel_child1->appendChild(ptr_pel_multiplier);

        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~      
    // Consolidate XML Object

    StackScop Poco::XML::DOMWriter pdw_xmlWriter;
    pdw_xmlWriter.setNewLine("\n");
    // pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::PRETTY_PRINT);  // set pretty print for nice debugger view
    pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::CANONICAL_XML);
    pdw_xmlWriter.writeNode(oss_xmlStream, ptr_pdoc_xmlDoc);

    sz_txPayload = oss_xmlStream.str();
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void I3E_XML_SRLZ_LogEvent(std::string& sz_txPayload, SE_LogEvent_t* i3e_ptr_logEvent)
{
    StackScop std::ostringstream oss_xmlStream;

    // Build a DOM document 
    StackScop Poco::AutoPtr<Poco::XML::Document> ptr_pdoc_xmlDoc = new Poco::XML::Document;
    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_xmlRoot = ptr_pdoc_xmlDoc->createElement(XSD_TAG_LOGEVENT);
    ptr_pdoc_xmlDoc->appendChild(ptr_pel_xmlRoot);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XMLNS, XSD_TAG_NS);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XSI, XSD_TAG_XSIURL);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XSISL, XSD_TAG_XSD);
    
    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1;
    StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1;

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // createdDateTime attribute
    
    ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_LOGEVENT_CRTDATETIME);
    ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_logEvent->createdDateTime));
    ptr_pel_child1->appendChild(ptr_ptex_ctex1);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // details attribute
    if(i3e_ptr_logEvent->_flags & SE_details_exists) 
    {
        ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_LOGEVENT_DETAILS);
        ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(i3e_ptr_logEvent->details);
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // extendedData attribute
    if(i3e_ptr_logEvent->_flags & SE_extendedData_exists) 
    {
        ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_LOGEVENT_EXTDATA);
        ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_logEvent->extendedData));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // functionSet attribute
    ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_LOGEVENT_FNCSET);
    ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_logEvent->functionSet));
    ptr_pel_child1->appendChild(ptr_ptex_ctex1);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child1);


    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // logEventCode attribute
    ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_LOGEVENT_LOGEVCODE);
    ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_logEvent->logEventCode));
    ptr_pel_child1->appendChild(ptr_ptex_ctex1);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child1);


    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // logEventID attribute
    ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_LOGEVENT_LOGEVID);
    ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_logEvent->logEventID));
    ptr_pel_child1->appendChild(ptr_ptex_ctex1);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child1);


    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // logEventPEN attribute
    ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_LOGEVENT_LOGEVPEN);
    ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_logEvent->logEventPEN));
    ptr_pel_child1->appendChild(ptr_ptex_ctex1);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child1);

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // profileID attribute
    ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_LOGEVENT_PROFILEID);
    ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_logEvent->profileID));
    ptr_pel_child1->appendChild(ptr_ptex_ctex1);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
                    
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~      
    // Consolidate XML Object

    StackScop Poco::XML::DOMWriter pdw_xmlWriter;
    pdw_xmlWriter.setNewLine("\n");
    // pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::PRETTY_PRINT);  // set pretty print for nice debugger view
    pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::CANONICAL_XML);
    pdw_xmlWriter.writeNode(oss_xmlStream, ptr_pdoc_xmlDoc);

    sz_txPayload = oss_xmlStream.str();
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void I3E_XML_SRLZ_MirrorUsagePoint(std::string& sz_txPayload, SE_MirrorUsagePoint_t* i3e_ptr_mup)
{
    StackScop std::ostringstream oss_xmlStream;

    // Build a DOM document 
    StackScop Poco::AutoPtr<Poco::XML::Document> ptr_pdoc_xmlDoc = new Poco::XML::Document;
    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_xmlRoot = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MUP);
    ptr_pdoc_xmlDoc->appendChild(ptr_pel_xmlRoot);

    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XMLNS, XSD_TAG_NS);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XSI, XSD_TAG_XSIURL);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XSISL, XSD_TAG_XSD);
    
    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1;
    StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1;
    
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // mRID attribute
    ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MUP_MRID);
    ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode
                    (i3e_xmls_byteArrToHexSz(i3e_ptr_mup->mRID, sizeof(i3e_ptr_mup->mRID)));
    ptr_pel_child1->appendChild(ptr_ptex_ctex1);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child1);

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // description attribute
    if(i3e_ptr_mup->_flags & SE_description_exists) 
    {
        ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MUP_DESC);
        ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(i3e_ptr_mup->description);
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // roleFlags attribute
    ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MUP_ROLEFLAGS);
    ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode
                    (i3e_xmls_byteArrToHexSz(i3e_ptr_mup->roleFlags, sizeof(i3e_ptr_mup->roleFlags)));
    ptr_pel_child1->appendChild(ptr_ptex_ctex1);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child1);

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // serviceCategoryKind attribute
    ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MUP_SERVCATKIND);
    ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_mup->serviceCategoryKind));
    ptr_pel_child1->appendChild(ptr_ptex_ctex1);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // status attribute
    ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MUP_STAT);
    ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_mup->status));
    ptr_pel_child1->appendChild(ptr_ptex_ctex1);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child1);

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // deviceLFDI attribute
    ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MUP_DEVLFDI);
    ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode
                    (i3e_xmls_byteArrToHexSz(i3e_ptr_mup->deviceLFDI, sizeof(i3e_ptr_mup->deviceLFDI)));
    ptr_pel_child1->appendChild(ptr_ptex_ctex1);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child1);

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  
    // Create MirrorMeterReading entries for each reading type 
    StackScop std::string sz_mmrPayload;
    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_mmr;
    for (uint8_t i = 0; i < i3e_ptr_mup->MirrorMeterReading.size(); i++) 
    {
        I3E_XML_SRLZ_MirrorMeterReadingInit(sz_mmrPayload, &(i3e_ptr_mup->MirrorMeterReading.at(i)), 
                        ptr_pdoc_xmlDoc, ptr_pel_xmlRoot);
    }
    
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~      
    // Consolidate XML Object

    StackScop Poco::XML::DOMWriter pdw_xmlWriter;
    pdw_xmlWriter.setNewLine("\n");
    // pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::PRETTY_PRINT);  // set pretty print for nice debugger view
    pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::CANONICAL_XML);
    pdw_xmlWriter.writeNode(oss_xmlStream, ptr_pdoc_xmlDoc);

    sz_txPayload = oss_xmlStream.str();
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void I3E_XML_SRLZ_MirrorMeterReadingInit(std::string& sz_txPayload, SE_MirrorMeterReading_t* i3e_ptr_mmr, 
                                                 Poco::AutoPtr<Poco::XML::Document> &ptr_pdoc_xmlDoc, 
                                                 Poco::AutoPtr<Poco::XML::Element> &poco_pel_xmlRoot)
{
    // MirrorMeterReading attribute and its children
    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_mmr = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MMR);
    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1;
    StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1;

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // mRID attribute
    ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MMR_MRID);
    ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(i3e_xmls_byteArrToHexSz(i3e_ptr_mmr->mRID, sizeof(i3e_ptr_mmr->mRID)));
    ptr_pel_child1->appendChild(ptr_ptex_ctex1);
    ptr_pel_mmr->appendChild(ptr_pel_child1);

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // description attribute
    if(i3e_ptr_mmr->_flags & SE_description_exists) 
    {
        ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MMR_DESC);
        ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(i3e_ptr_mmr->description);
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_mmr->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // ReadingType
    if(i3e_ptr_mmr->_flags & SE_ReadingType_exists)
    {

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // ReadingType attribute and its children
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_readType = ptr_pdoc_xmlDoc->createElement(XSD_TAG_READTYPE);

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // accumulationBehaviour attribute
        if(i3e_ptr_mmr->ReadingType._flags & SE_accumulationBehaviour_exists) 
        {
            ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_READTYPE_ACCUMBEH);
            ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode
                            (std::to_string(i3e_ptr_mmr->ReadingType.accumulationBehaviour));
            ptr_pel_child1->appendChild(ptr_ptex_ctex1);
            ptr_pel_readType->appendChild(ptr_pel_child1);
        }
        else{}

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // commodity attribute
        if(i3e_ptr_mmr->ReadingType._flags & SE_commodity_exists) 
        {
            ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_READTYPE_COMMODITY);
            ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_mmr->ReadingType.commodity));
            ptr_pel_child1->appendChild(ptr_ptex_ctex1);
            ptr_pel_readType->appendChild(ptr_pel_child1);
        }
        else{}

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // dataQualifier attribute
        if(i3e_ptr_mmr->ReadingType._flags & SE_dataQualifier_exists) 
        {
            ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_READTYPE_DATAQUAL);
            ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_mmr->ReadingType.dataQualifier));
            ptr_pel_child1->appendChild(ptr_ptex_ctex1);
            ptr_pel_readType->appendChild(ptr_pel_child1);
        }
        else{}

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // flowDirection attribute
        if(i3e_ptr_mmr->ReadingType._flags & SE_flowDirection_exists) 
        {
            ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_READTYPE_FLOWDIR);
            ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_mmr->ReadingType.flowDirection));
            ptr_pel_child1->appendChild(ptr_ptex_ctex1);
            ptr_pel_readType->appendChild(ptr_pel_child1);
        }
        else{}

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // kind attribute
        if(i3e_ptr_mmr->ReadingType._flags & SE_kind_exists) 
        {
            ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_READTYPE_KIND);
            ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_mmr->ReadingType.kind));
            ptr_pel_child1->appendChild(ptr_ptex_ctex1);
            ptr_pel_readType->appendChild(ptr_pel_child1);
        }
        else{}

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // phase attribute
        if(i3e_ptr_mmr->ReadingType._flags & SE_phase_exists) 
        {
            ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_READTYPE_PHASE);
            ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_mmr->ReadingType.phase));
            ptr_pel_child1->appendChild(ptr_ptex_ctex1);
            ptr_pel_readType->appendChild(ptr_pel_child1);
        }
        else{}

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // powerOfTenMultiplier attribute
        if(i3e_ptr_mmr->ReadingType._flags & SE_powerOfTenMultiplier_exists) 
        {
            ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_READTYPE_POTMULT);
            ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_mmr->ReadingType.powerOfTenMultiplier));
            ptr_pel_child1->appendChild(ptr_ptex_ctex1);
            ptr_pel_readType->appendChild(ptr_pel_child1);
        }
        else{}

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // uom attribute
        if(i3e_ptr_mmr->ReadingType._flags & SE_uom_exists) 
        {
            ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_READTYPE_UOM);
            ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(i3e_ptr_mmr->ReadingType.uom));
            ptr_pel_child1->appendChild(ptr_ptex_ctex1);
            ptr_pel_readType->appendChild(ptr_pel_child1);
        }
        else{}

        //~~~~~~~~~~~~~~~~~~~~~~~~~~

        ptr_pel_mmr->appendChild(ptr_pel_readType);

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        poco_pel_xmlRoot->appendChild(ptr_pel_mmr);
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop void I3E_XML_SRLZ_MirrorMeterReading(std::string& sz_txPayload, SE_MirrorMeterReading_t* i3e_ptr_mmr)
{
    StackScop std::ostringstream oss_xmlStream;

    // Build a DOM document 
    StackScop Poco::AutoPtr<Poco::XML::Document> ptr_pdoc_xmlDoc = new Poco::XML::Document;
    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_xmlRoot = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MMR);
    ptr_pdoc_xmlDoc->appendChild(ptr_pel_xmlRoot);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XMLNS, XSD_TAG_NS);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XSI, XSD_TAG_XSIURL);
    ptr_pel_xmlRoot->setAttribute(XSD_TAG_XSISL, XSD_TAG_XSD);
    
    StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_child1;
    StackScop Poco::AutoPtr<Poco::XML::Text> ptr_ptex_ctex1;

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // mRID attribute
    ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MMR_MRID);
    ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(i3e_xmls_byteArrToHexSz(i3e_ptr_mmr->mRID, sizeof(i3e_ptr_mmr->mRID)));
    ptr_pel_child1->appendChild(ptr_ptex_ctex1);
    ptr_pel_xmlRoot->appendChild(ptr_pel_child1);

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // description attribute
    if(i3e_ptr_mmr->_flags & SE_description_exists) 
    {
        ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_MMR_DESC);
        ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(i3e_ptr_mmr->description);
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_xmlRoot->appendChild(ptr_pel_child1);
    }
    else{}

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // Reading
    if(i3e_ptr_mmr->_flags & SE_Reading_exists)
    {
        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // Reading attribute and its children
        StackScop Poco::AutoPtr<Poco::XML::Element> ptr_pel_read = ptr_pdoc_xmlDoc->createElement(XSD_TAG_READING);

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        // value attribute
        ptr_pel_child1 = ptr_pdoc_xmlDoc->createElement(XSD_TAG_READING_VAL);
        ptr_ptex_ctex1 = ptr_pdoc_xmlDoc->createTextNode(std::to_string(static_cast<int16_t>(i3e_ptr_mmr->Reading.value & 0x00FFFFFFFFFFFF)));
        ptr_pel_child1->appendChild(ptr_ptex_ctex1);
        ptr_pel_read->appendChild(ptr_pel_child1);

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        ptr_pel_xmlRoot->appendChild(ptr_pel_read);
    }
                
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~      
    // Consolidate XML Object

    StackScop Poco::XML::DOMWriter pdw_xmlWriter;
    pdw_xmlWriter.setNewLine("\n");
    // pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::PRETTY_PRINT);  // set pretty print for nice debugger view
    pdw_xmlWriter.setOptions(Poco::XML::XMLWriter::CANONICAL_XML);
    pdw_xmlWriter.writeNode(oss_xmlStream, ptr_pdoc_xmlDoc);

    sz_txPayload = oss_xmlStream.str();
}

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop std::string i3e_xmls_toPaddedSz(int numToPad) {
    StackScop std::stringstream ss;
    ss << std::setw(2) << std::setfill('0') << numToPad;
    return ss.str();
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop std::string i3e_xmls_byteArrToHexSz(const uint8_t* ptr_u8_byteArray, size_t siz_byteArraySize) {
    StackScop std::stringstream ss;
    for (size_t i = 0; i < siz_byteArraySize; ++i) {
        ss << std::hex << std::setw(2) << std::setfill('0') << static_cast<int>(ptr_u8_byteArray[i]);
    }
    return ss.str();
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~