/**********************************************************************************************************************
 * @file            I3E_DAG.hpp
 *
 * @brief           The Directed Acyclic Graph (DAG) Class contains the functional definition for the data structure
 *                  responseible for storing IEEE 2030.5 resources in memory. Included are functions that allow for
 *                  manipulation of the data nodes, searching methods, etc. 
 *
 * @author          Vlad Bugayev
 * @date            03/10/2024
 * @copyright       2024 Lion Energy LLC.
 *                  All rights reserved
 *
 *********************************************************************************************************************/
/**
 * @addtogroup      DAG
 *                  Directed Acyclic Graph
 *
 * @ref             I3E_DAG.hpp
 *
 *********************************************************************************************************************/

#ifndef LION_I3E_DAG_H__
#define LION_I3E_DAG_H__

#include "lion/sch/I3E_Schema.hpp"
#include "I3E_SPP.hpp"

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

// Define a variant that can hold instances of I3E schema types - used for storing unique structs within the graph nodes
#define RESOURCE_VARIANT    std::variant<SE_DeviceCapability_t, SE_Time_t, SE_EndDeviceList_t, SE_EndDevice_t,      \
                                         SE_FunctionSetAssignmentsList_t,  SE_FunctionSetAssignments_t,             \
                                         SE_DERProgramList_t, SE_DERProgram_t, SE_DERControlList_t,                 \
                                         SE_DERControl_t, SE_DERCurveList_t, SE_DefaultDERControl_t, SE_DERList_t,  \
                                         SE_DER_t, SE_DERStatus_t, SE_DERCapability_t, SE_DERSettings_t,            \
                                         SE_DERAvailability_t, SE_MirrorUsagePoint_t, SE_MirrorMeterReading_t,      \
                                         SE_LogEvent_t, SE_DERCurve_t, SE_Registration_t>

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// Struct definition for data contained within the DAG graph nodes 
ModScop struct NodeData 
{

    public:

        // Node metadata
        ModScop uint32_t u32_pollRate = 0;
        ModScop time_t time_pollNext = 0;
        ModScop std::string sz_resourceName = "";
        ModScop std::string sz_lastPollStr = ""; 
        ModScop bool bol_updated = false; 
        ModScop I3EresourceType enum_rsrcType;

        // Node contents
        ModScop RESOURCE_VARIANT var_rsrc;


        /**
         * @brief NodeData Constructor
         * 
         * @details This constructor initializes the NodeData struct with default values.
         */
        NodeData();

        /**
         * @brief NodeData Constructor
         * 
         * @details This constructor initializes the NodeData struct with the specified resource name and data.
         * 
         * @param sz_name Node name 
         * @param var_data Node data
         */
        NodeData(const std::string& sz_name, const RESOURCE_VARIANT& var_data);

        /**
         * @brief Determine if the node has a poll rate
         * 
         * @details This method checks if the node has a poll rate attribute.
         * 
         * @returns A boolean value indicating whether the node has a poll rate
         */
        bool hasPollRate() const;

        /**
         * @brief Get the node poll rate 
         * 
         * @details This method returns the poll rate of the node.
         * 
         * @returns The poll rate of the node
         */
        uint32_t getPollRate() const;

        /**
         * @brief Check if node is empty
         * 
         * @details This method checks if the node is empty.
         * 
         * @returns A boolean value indicating whether the node is empty
         */
        bool isEmpty() const;
};

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// Directed Acyclic Graph --- Class definition 
ModScop class Graph 
{
    private:

        // Initialize the DAG conditional variable and mutex used for shared access
        ModScop pthread_mutex_t mtx_DAGMutex;

        // Key -- HREF, value -- NodeData 
        ModScop std::unordered_map<std::string, NodeData> nodes;

        // Key -- HREF, value -- Unordered set of strings (other HREFs)
        ModScop std::unordered_map<std::string, std::unordered_set<std::string>> adjacencyList;
    
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    public:

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief DAG Contstructor
         * 
         * @details This constructor initializes the DAG graph and the mutex used for shared access.
         * 
         */
        ModScop Graph();
            
        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief DAG Destructor
         * 
         * @details This destructor destroys the DAG graph and the mutex used for shared access.
         * 
         */
        ModScop ~Graph();

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Getter method to access the DAGMutex
         * 
         * @details This method returns a reference to the DAGMutex for shared access.
         * 
         * @returns Reference to the DAGMutex
         */
        // Getter method to access the DAGMutex
        ModScop pthread_mutex_t& I3E_DAG_GetDAGMutex();

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Add a node to the graph
         * 
         * @details This method adds a node to the graph with the specified node name and NodeData.
         * 
         * @param sz_nodeName The name of the node to be added
         * @param str_data The NodeData associated with the node
         */
        // Add a node to the graph 
        ModScop void I3E_DAG_AddNode(const std::string& sz_nodeName, const NodeData& str_data);

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Add an edge between two nodes in the graph
         * 
         * @details This method adds an edge between two nodes in the graph.
         * 
         * @param sz_fromNode The name of the node from which the edge originates
         * @param sz_toNode The name of the node to which the edge points
         */
        ModScop void I3E_DAG_AddEdge(const std::string& sz_fromNode, const std::string& sz_toNode);

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Print the details of a node in the graph
         * 
         * @details This method prints the details of a node in the graph, including its children nodes. This is 
         * a recursive function that prints the node name and then calls itself on each child node. This function is 
         * really just for debugging purposes
         * 
         * @param sz_nodeName The name of the node to print
         * @param depth The depth of the node in the graph
         */
        ModScop void I3E_DAG_PrintNodes(const std::string& sz_nodeName, int depth);

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Function to find a node in the graph and return its NodeData if it exists
         * 
         * @details This function finds a node in the graph with the specified node name and returns its NodeData 
         * if it exists.
         * 
         * @param sz_nodeName The name of the node to find
         * 
         * @returns The NodeData associated with the node
         */
        ModScop NodeData I3E_DAG_GetNode(const std::string& sz_nodeName);

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        
        /**
         * @brief Find resource nodes of a specific type in the graph
         * 
         * @details This method finds all nodes of a specific resource type in the graph and returns a vector of
         * their names for further processing.
         * 
         * @param enum_targetType The resource type to search for
         * 
         * @returns A vector of strings containing the names of nodes of the specified resource type
         */        
        ModScop std::vector<std::string> I3E_DAG_FindResourceNodes(I3EresourceType enum_targetType);

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Update the content of a node in the graph
         * 
         * @details This method updates the content of a node in the graph with the specified node name and new content.
         * 
         * @param sz_nodeName The name of the node to update
         * @param str_newContent The new content to update the node with
         */
        ModScop void I3E_DAG_UpdateNodeContent(const std::string& sz_nodeName, const NodeData& str_newContent);
        
        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Find nodes of a specific resource type starting from a given node.
         * 
         * @details This method performs a breadth-first search (BFS) traversal starting from a given node to find all 
         * nodes of a specific resource type. Used in the resource rediscovery process.
         * 
         * @param sz_startNode The name of the node to start the search from
         * @param enum_targetType The resource type to search for
         */
        ModScop std::vector<std::string> I3E_DAG_FindResourceNodesFromNode(const std::string& sz_startNode, 
                                        I3EresourceType enum_targetType);
      
        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Check if a node exists in the graph
         * 
         * @details This method checks if a node with the specified name exists in the graph and returns a boolean 
         * value indicating its presence.
         * 
         * @param sz_nodeName The name of the node to check for
         * 
         * @returns A boolean value indicating whether the node exists in the graph
         */
        ModScop bool I3E_DAG_DoesNodeExist(const std::string& sz_nodeName);

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Recursively scan all nodes in the graph for poll rates.
         * 
         * @details Utility functions used by I3E_DAG_ScanAllNodesWithPollRate to recursively scan all nodes in the graph
         * for nodes that contain a poll rate attribute. 
         * 
         * @param sz_currentNode The name of the current node being processed
         * @param set_visited A set to track visited nodes during traversal
         * @param vec_nodesWithPollRate A vector to store pointers to NodeData for nodes with poll rates
         */
        ModScop void I3E_DAG_ScanAllNodesRecursiveUtil(const std::string& sz_currentNode, 
                                                        std::unordered_set<std::string>& set_visited, 
                                                        std::vector<NodeData*>& vec_nodesWithPollRate);

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Scan all nodes in the graph for poll rates.
         * 
         * @details This method scans all nodes in the graph for poll rates and returns a vector of NodeData pointers
         * to the nodes that have a resource with a poll rate. Used by the SPP module to determine which resources 
         * need to be polled.
         * 
         * @returns A vector of NodeData pointers to nodes with poll rates
         */
        ModScop std::vector<NodeData*> I3E_DAG_ScanAllNodesWithPollRate();

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        // Function to remove a node and its dependencies from the graph
        /**
         * @brief Remove a node and its dependencies from the graph
         * 
         * @details This method removes a node and all its dependencies from the graph. It performs a breadth-first
         * traversal starting from the specified node to remove all nodes that are dependent on it.
         * 
         * @param sz_nodeName The name of the node to remove
         */
        ModScop void I3E_DAG_RemoveNodeAndDependencies(const std::string& sz_nodeName);

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Remove all children nodes from a specified node
         * 
         * @details This method removes all children nodes from a specified node in the graph. It recursively deletes
         * all children of the current node and then removes the current node from the graph.
         * 
         * @param sz_nodeName The name of the node to remove all children from
         */        
        ModScop void I3E_DAG_DeleteAllChildren(const std::string& sz_nodeName);

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Find the parent node of a given node
         * 
         * @details This method finds the parent node of a given node in the graph and returns its name. It searches
         * through the adjacency list to find the parent node of the specified node.
         * 
         * @param sz_nodeName The name of the node to find the parent of
         * 
         * @returns The name of the parent node of the specified node
         */
        ModScop std::string I3E_DAG_FindParentNode(const std::string& sz_nodeName);

        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Reset the updated flag of a node in the graph
         * 
         * @details This method resets the updated flag of a node in the graph with the specified node name. It sets
         * the updated flag to false for the specified node.
         * 
         * @param href The name of the node to reset the updated flag for
         */
        ModScop void I3E_DAG_ResetUpdatedFlag(const std::string& href);
};

#endif