/**********************************************************************************************************************
 * @file            I3E_DEM.hpp
 *
 * @brief           The DERC Event Manager (DEM) module draws existing data from stored IEEE 2030.5 resources from the
 *                  DAG module and enqueues the DERC events in a time-sorted queue. This module performs logical checks
 *                  on the events to ensure only specific events are executed, and respones to the IEEE 2030.5 server
 *                  regarding event status information are executed at the correct times. 
 *    
 *
 * @author          Vlad Bugayev
 * @date            04/10/2024
 * @copyright       2024 Lion Energy LLC.
 *                  All rights reserved
 *
 *********************************************************************************************************************/
/**
 * @addtogroup      DEM
 *                  DERC Event Manager
 *
 * @ref             I3E_DEM.hpp
 *
 *********************************************************************************************************************/

#include "COD_Compiler_Defines.h"

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifndef LION_I3E_DEM_H__
#define LION_I3E_DEM_H__

#include "lion/sch/I3E_Schema.hpp"

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define RSPSTAT_EVENT_RECEIVED                      1   // Event received
#define RSPSTAT_EVENT_STARTED                       2   // Event started
#define RSPSTAT_EVENT_COMPLETED                     3   // Event completed
#define RSPSTAT_USER_OPT_OUT                        4   // User has chosen to opt-out
#define RSPSTAT_USER_OPT_IN                         5   // User has chosen to opt-in
#define RSPSTAT_EVENT_CANCELLED                     6   // The event has been cancelled
#define RSPSTAT_EVENT_SUPERSEDED                    7   // The event has been superseded
#define RSPSTAT_PARTIAL_COMPLETED_OPT_OUT           8   // Event partially completed with user opt-out
#define RSPSTAT_PARTIAL_COMPLETED_OPT_IN            9   // Event partially completed due to user opt-in
#define RSPSTAT_EVENT_COMPLETED_NO_PARTICIPATION    10  // Event completed, no user participation (previous opt-out)
#define RSPSTAT_USER_ACKNOWLEDGED                   11  // User has acknowledged the event
#define RSPSTAT_CANNOT_BE_DISPLAYED                 12  // Cannot be displayed
#define RSPSTAT_EVENT_ABORTED_ALTERNATE_SERVER      13  // Event aborted due to alternate server event
#define RSPSTAT_EVENT_ABORTED_ALTERNATE_PROGRAM     14  // Event aborted due to alternate program event
#define RSPSTAT_REJECTED_CONTROL_NOT_APPLICABLE     252 // Rejected: control parameter not applicable
#define RSPSTAT_REJECTED_INVALID_EVENT              253 // Rejected: invalid event
#define RSPSTAT_REJECTED_EVENT_EXPIRED              254 // Rejected: event was received after it expired

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// Forward declaration Event Queue Structures 
struct EventBlock;
struct EventQueue;
struct DefaultBlock;

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

/**
 * @brief Initializes the DERC Event Manager module.
 * 
 * @details This function initializes the event semaphore for the DERC Event Manager module.
 * 
 * @returns 0 on success, non-zero on failure
 */
ModScop int I3E_DEM_Init(void);

/**
 * @brief Thread execution function for the DERC Event Manager module.
 *
 * @details This function is the main execution function for the DEM module. It is responsible for management of all 
 * DERC events and the execution of the event sequence. There is a separate logic chain for if the semaphore 
 * is posted by the SPP module (new event to process), or if the timer runes out (event queue management checks).
 * 
 * This function should never return (infinite looping runtime behavior). 
 */
ModScop void I3E_DEM_Exec(void);

/**
 * @brief DERC Event Queue Getter
 * 
 * @details This function returns a pointer to the DERC Event Queue for the DEM module.
 * 
 * @returns Pointer to the DERC Event Queue object. 
 */
ModScop EventQueue* I3E_DEM_GetDERCEventQueue(void);

/**
 * @brief DERC Event Manager Semaphore post function
 * 
 * @details This function allows other modules to post the DERC Event Manager semaphore, signaling that there is a new 
 * event to process (specifically, the SPP module)
 * 
 */
ModScop void I3E_DEM_PostSemaphore(void);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#endif /* LION_I3E_HTTP_H__ */