/**********************************************************************************************************************
 * @file            I3E_EVQ.hpp
 *
 * @brief           The Event Queue (EVQ) class contains the logic for storing IEEE 2030.5 DERC events in a queued 
 *                  format. The EVQ class contains functions that allow for manipulation, searching, and event flagging
 *                  for proper DERC event execution. 
 *    
 *
 * @author          Vlad Bugayev
 * @date            05/05/2024
 * @copyright       2024 Lion Energy LLC.
 *                  All rights reserved
 *
 *********************************************************************************************************************/
/**
 * @addtogroup      EVQ
 *                  Event Qeuue
 *
 * @ref             I3E_EVQ.hpp
 *
 *********************************************************************************************************************/

#include "COD_Compiler_Defines.h"

#ifndef LION_I3E_EVQ_H__
#define LION_I3E_EVQ_H__

#include "lion/sch/I3E_Schema.hpp"
#include "I3E_DAG.hpp"

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define EVQ_eventReceievedMsg   (1 << 0)
#define EVQ_eventCompleted      (1 << 1)
#define EVQ_eventCompletedMsg   (1 << 2)
#define EVQ_eventStarted        (1 << 3)
#define EVQ_eventStartedMsg     (1 << 4)
#define EVQ_eventSuperseded     (1 << 5)
#define EVQ_eventSupersededMsg  (1 << 6)
#define EVQ_eventCancelled      (1 << 7)
#define EVQ_eventCancelledMsg   (1 << 8)
#define EVQ_eventActive         (1 << 9)
#define EVQ_eventClear          (1 << 10)
#define EVQ_eventPrecheck       (1 << 11)
#define EVQ_eventOptOut         (1 << 12)
#define EVQ_eventOptOutMsg      (1 << 13)

#define SE_rampTms_hrefPos                       0 
#define SE_opModWattVar_hrefPos                  1 
#define SE_opModWattPF_hrefPos                   2 
#define SE_opModVoltWatt_hrefPos                 3 
#define SE_opModVoltVar_hrefPos                  4 
#define SE_opModTargetW_hrefPos                  5 
#define SE_opModTargetVar_hrefPos                6 
#define SE_opModMaxLimW_hrefPos                  7 
#define SE_opModLVRTMustTrip_hrefPos             8 
#define SE_opModLVRTMomentaryCessation_hrefPos   9 
#define SE_opModLFRTMustTrip_hrefPos            10
#define SE_opModHVRTMustTrip_hrefPos            11
#define SE_opModHVRTMomentaryCessation_hrefPos  12
#define SE_opModHFRTMustTrip_hrefPos            13
#define SE_opModFreqWatt_hrefPos                14
#define SE_opModFreqDroop_hrefPos               15
#define SE_opModFixedW_hrefPos                  16
#define SE_opModFixedVar_hrefPos                17
#define SE_opModEnergize_hrefPos                20
#define SE_opModConnect_hrefPos                 22
#define SE_opModHFRTMayTrip_hrefPos             23 
#define SE_opModHVRTMayTrip_hrefPos             24 
#define SE_opModLFRTMayTrip_hrefPos             25 
#define SE_opModLVRTMayTrip_hrefPos             26 
#define SE_opModFixedPFInjectW_hrefPos          27 
#define SE_opModFixedPFAbsorbW_hrefPos          28 

#define SE_setGradW_hrefPos             5
#define SE_setESDelay_hrefPos           20
#define SE_setESHighFreq_hrefPos        21
#define SE_setESHighVolt_hrefPos        22
#define SE_setESLowFreq_hrefPos         23
#define SE_setESLowVolt_hrefPos         24
#define SE_setESRandomDelay_hrefPos     25
#define SE_setSoftGradW_hrefPos         26
#define SE_setESRampTms_hrefPos         3

#define SZ_ACTIVE_CONTROL_MAX 32

/**********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

// Struct definition for the EventBlock structures that are the building blocks of the EventQueue
ModScop struct EventBlock
{
    ModScop std::string sz_href; 
    ModScop int primacy; 
    ModScop SE_UInt32_t i3e_randDuration = 0;
    ModScop SE_TimeType_t i3e_randStart = 0;
    ModScop uint32_t u32_eventFlags = 0; 
    ModScop SE_DERControl_t i3e_derCtrl = {0};
};

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// Struct definition for the DefaultBlock structures that are the building blocks of the Default Queue
ModScop struct DefaultBlock
{
    ModScop std::string sz_href; 
    ModScop int primacy;
    ModScop SE_DefaultDERControl_t i3e_defDerCtrl = {0};
};

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ModScop class EventQueue
{

    private:

        //#TODO - may have to switch to using mRID values instead of hrefs at some point

        ModScop uint32_t u32_activeControls = 0;
        ModScop std::vector<std::string> vec_activeControlsHrefs;
        ModScop std::vector<std::string> vec_activeControlsMRIDs;

        ModScop uint32_t u32_activeDefaultControls = 0;
        ModScop std::vector<std::string> vec_activeDefaultsHrefs;
        ModScop std::vector<std::string> vec_activeDefaultsMRIDs;

        ModScop uint32_t u32_activeSettings = 0;
        ModScop std::vector<std::string> vec_activeSettingsHrefs;
        ModScop std::vector<std::string> vec_activeSettingsMRIDs;

        // Queue of EventBlocks
        ModScop std::queue<EventBlock> q_eventQueue;

        // Queue of DefaultBlocks
        ModScop std::queue<DefaultBlock> q_defaultQueue;
        
        // Mutex protection on queue push and pop
        ModScop pthread_mutex_t mtx_eventQueueMutex;

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    public:

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Event Queue constructor.
         * 
         * @details Initializes the Event Queue object and the pthread_mutex for thread safety. Also 
         * initializes the vectors for activeControls and activeDefaults to be a specific size
         * 
         */    
        ModScop EventQueue();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Event Queue destructor.
         * 
         * @details Destroys the Event Queue object and the pthread_mutex.
         * 
         */    
        ModScop ~EventQueue();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Getter for the Event Queue Mutex.
         * 
         * @details Getter method to access the EVQ Mutex
         * 
         * @returns Event Queue Mutex
         */         
        ModScop pthread_mutex_t& I3E_EVQ_GetEventQueueMutex();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Getter for the active DER control bitfield.
         * 
         * @returns The active DER control bitfield
         */         
        ModScop uint32_t I3E_EVQ_GetActiveDERControlBitfield();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Getter for the active DER control MRIDs
         * 
         * @returns The active DER control MRIDs
         */         
        ModScop std::vector<std::string> I3E_EVQ_GetActiveDERControlMRIDs();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /**
         * @brief Getter for the active default DER control bitfield.
         * 
         * @returns The active default DER control bitfield
         */           
        ModScop uint32_t I3E_EVQ_GetActiveDefaultsDERControlBitfield();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~   
        /**
         * @brief Getter for the active default DER control MRIDs
         * 
         * @returns The active default DER control MRIDs
        */     
        ModScop std::vector<std::string> I3E_EVQ_GetActiveDefaultsDERControlMRIDs();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /**
         * @brief Getter for the active settings bitfield.
         * 
         * @returns The active settings control bitfield
        */    
        ModScop uint32_t I3E_EVQ_GetSettingsBitfield();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~   
        /**
         * @brief Getter for the active settings MRIDs
         * 
         * @returns The active settings MRIDs
        */     
        ModScop std::vector<std::string> I3E_EVQ_GetSettingsMRIDs();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~   
        /**
         * @brief Add EventBlock to the queue
         * 
         * @details Adds an EventBlock to the queue
         * 
         * @param str_event The EventBlock to be added to the queue
         */          
        ModScop void I3E_EVQ_AddEvent(const EventBlock& str_event);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        
        /**
         * @brief Set active controls
         * 
         * @details Adds to the active control flags parameter to keep track of what types of controls are active. 
         * Only one control of each kind can be active at a time. Used for determining which controls to execute.
         * 
         * @param u32_flags The flags to set for active controls
         */                      
        ModScop void I3E_EVQ_SetActiveControl(uint32_t u32_flags);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Clears active controls 
         * 
         * @details Clears the specific active control flags parameter bits for controls that are no longer active. 
         * This allows other controls of the same type to activate.
         * 
         * @param u32_flags The flags to clear for active controls
         */                
        ModScop void I3E_EVQ_ClearActiveControl(uint32_t u32_flags);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Check active controls 
         * 
         * @details Checks if the specific active control flags parameter bits are set for active controls.
         * 
         * @param u32_flags The flags to check for active controls
         */               
        ModScop bool I3E_EVQ_CheckActiveControl(uint32_t u32_flag);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Set active default controls
         * 
         * @details Adds to the active default control flags parameter to keep track of what types of default controls 
         * are active. Only one control of each kind can be active at a time. Used for determining which default controls
         * to execute.
         * 
         * @param u32_flags The flags to set for active default controls
         */     
        ModScop void I3E_EVQ_SetActiveDefault(uint32_t u32_flags);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Clears active default controls 
         * 
         * @details Clears the specific active default control flags parameter bits for default controls that are no 
         * longer active. This allows other default controls of the same type to activate.
         * 
         * @param u32_flags The flags to clear for active default controls
         */               
        ModScop void I3E_EVQ_ClearActiveDefault(uint32_t u32_flags);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Check active default controls 
         * 
         * @details Checks if the specific active default control flags parameter bits are set for active default 
         * controls.
         * 
         * @param u32_flags The flags to check for active default controls
         */          
        ModScop bool I3E_EVQ_CheckActiveDefault(uint32_t u32_flag);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /**
         * @brief Set settings
         * 
         * @details Adds to the active settings flags parameter to keep track of what types of settings 
         * are active. Only one settings of each kind can be active at a time. Used for determining which settings
         * to execute.
         * 
         * @param u32_flags The flags to set for settings
        */    
        ModScop void I3E_EVQ_SetSettings(uint32_t u32_flags);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /**
         * @brief Clears settings
         * 
         * @details Clears the specific settings flags parameter bits for settings that are no 
         * longer active. This allows other settings of the same type to activate.
         * 
         * @param u32_flags The flags to clear for settings
        */            
        ModScop void I3E_EVQ_ClearSettings(uint32_t u32_flags);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /**
         * @brief Check settings 
         * 
         * @details Checks if the specific settings flags parameter bits are set for settings.
         * 
         * @param u32_flags The flags to check for settings
        */          
        ModScop bool I3E_EVQ_CheckSettings(uint32_t u32_flag);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Sets the event status flags for the specified event.
         * 
         * @details This function will update an EventBlock with a new set of event status flags that details what
         * state the event is in (Cancelled, Superseded, Completed, Active, etc.). Also sets some flags that 
         * indicate if a server repsone message has been sent for the event. The DEM module uses these flags to 
         * determine what to do with the event.
         * 
         * @param str_event The EventBlock to update
         * @param u32_flags The flags to set for the event 
         */          
        ModScop void I3E_EVQ_SetEventFlags(EventBlock& str_event, uint32_t u32_flags);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Clears the event status flags for the specified event.
         * 
         * @details This function will clear the specified flags of an EventBlock which assists the DEM module in 
         * dictating what to do with these events.
         * 
         * @param str_event The EventBlock to update
         * @param u32_flags The flags to clear for the event 
         */  
        ModScop void I3E_EVQ_ClearEventFlags(EventBlock& str_event, uint32_t u32_flags);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Checks the event status flags for the specified event.
         * 
         * @details This function will check if the specified flags are set for an EventBlock. This is used by the DEM
         * module to determine what state the event is in and what to do with the event.
         * 
         * @param str_event The EventBlock to update
         * @param u32_flags The flags to check for the event 
         */  
        ModScop bool I3E_EVQ_CheckEventFlag(const EventBlock& str_event, uint32_t u32_flags);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Find which events are missing in the queue
         * 
         * @details This function will compare the hrefs of the DERControl objects in the queue with the hrefs of the
         * DERControl objects in the vector passed in. If an href is not found in the queue, it will be returned in a 
         * vector of missing events for the calling function to add to the queue.
         * 
         * @param vec_derCtrlHrefs The vector of hrefs to compare with the queue
         * 
         * @returns A vector of hrefs that are missing in the queue
         */          
        ModScop std::vector<std::string> I3E_EVQ_FindMissingEventsInQueue(const std::vector<std::string>& vec_derCtrlHrefs);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Print events in the queue 
         * 
         * @details This function will print out the details of the events that are currently in the queue. This is
         * used for debugging purposes to ensure that the events are being added and removed from the queue correctly.
         */
        ModScop void I3E_EVQ_PrintAllEventsInQueue();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Insert new event by start time
         * 
         * @details This function will insert a new event into the queue based on the start time of the event. The
         * queue will be sorted with the earliest event at the front of the queue, and latest will be last. 
         * 
         * @param str_newEvent The new event to insert into the queue
         */
        ModScop void I3E_EVQ_InsertEventByStartTime(const EventBlock& str_newEvent);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Event timing overlap check
         * 
         * @details This function will find all the events in the queue that overlap with the specific event based on
         * the start time and duration of the specific event. The function will return a vector of hrefs of the
         * overlapping events to do more checks on the events to determine which to execute.
         * 
         * @param sz_specificHref The href of the specific event to check for overlaps with existing events in the queue.
         * 
         * @returns A vector of hrefs of events that overlap with the given event. 
         */
        ModScop std::vector<std::string> I3E_EVQ_FindOverlappingEvents(const std::string& sz_specificHref);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Trim events from queue
         * 
         * @details This function will find all the events in the queue that have specific flags set which indicate
         * that the event is no longer needed in the queue.
         * 
         * @returns A vector of hrefs of events that have specific flags set and should be removed from the queue.
         */
        ModScop std::vector<std::string> I3E_EVQ_FindEventsToTrim();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Remove events from queue
         * 
         * @details This function will remove events from the queue based on the hrefs of the events that are passed in
         * as a vector. The function will iterate through the queue and remove the events that have hrefs that match
         * the hrefs in the vector.
         * 
         * @param vec_hrefsToRemove Vector of hrefs of events to remove from the queue.
         */
        ModScop void I3E_EVQ_RemoveEventsByHref(const std::vector<std::string>& vec_hrefsToRemove);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Get specific event from queue
         * 
         * @details This function will look for a specific event in the queue based on the href of the event. 
         * The function will iterate through the queue and return the EventBlock that matches the href that is passed in.
         * 
         * @param sz_targetHref Target event to find in the queue
         * 
         * @returns The ventBlock of the event that matches the href
         */
        ModScop EventBlock I3E_EVQ_GetSpecificEvent(const std::string& sz_targetHref);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  

        /**
         * @brief Replace event block in queue
         * 
         * @details This function will search for a target event and replace the existing event block with new data.
         * 
         * @param sz_targetHref The href of the event to replace
         * @param str_newEventBlock The new EventBlock to replace the existing event block
         */
        ModScop void I3E_EVQ_ReplaceEventBlockInQueue(const std::string& sz_targetHref, const EventBlock& str_newEventBlock);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Find events that are starting in 60 seconds 
         * 
         * @details This function will find all the events in the queue that are starting within 60 seconds from the
         * current time. The function will return a vector of hrefs of the events that are starting within 60 seconds.
         * 
         * @param i3e_currentTime The current time to compare with the start time of the events.
         * 
         * @returns A vector of hrefs of events that are starting within 60 seconds.
         */
        ModScop std::vector<std::string> I3E_EVQ_GetEventsStartingIn60Seconds(SE_TimeType_t i3e_currentTime);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Find started events
         * 
         * @details This function will find all the events in the queue that have started based on the current time. The
         * function will return a vector of hrefs of the events that have started.
         * 
         * @param i3e_currentTime  The current time to compare with the start time of the events.
         * 
         * @returns A vector of hrefs of events that have started.
         */
        ModScop std::vector<std::string> I3E_EVQ_GetEventsStarted(SE_TimeType_t i3e_currentTime);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Find completed events
         * 
         * @details This function will find all the events in the queue that have completed based on the current time. 
         * The function will return a vector of hrefs of the events that have completed.
         * 
         * @param i3e_currentTime The current time to compare with the start time and duration of the events.
         * 
         * @returns A vector of hrefs of events that have completed.
         */
        ModScop std::vector<std::string> I3E_EVQ_GetEventsCompleted(SE_TimeType_t i3e_currentTime);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Replace DERControl for event
         * 
         * @details This function will replace the DERControl for a specific event in the queue. The function will iterate
         * through the queue and replace the DERControl for the event that matches the href that is passed in.
         * 
         * @param sz_eventHref The href of the event to replace the DERControl for
         * @param i3e_newDerCtrl The new DERControl to replace the existing DERControl
         */
        ModScop void I3E_EVQ_ReplaceDerCtrlForEvent(const std::string& sz_eventHref, const SE_DERControl_t& i3e_newDerCtrl);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Find cancelled events in the queue
         * 
         * @details This function will find all the events in the queue that have been cancelled. The function will return
         * a vector of hrefs of the events that have been cancelled.
         * 
         * @returns A vector of hrefs of events that have been cancelled.
         */
        ModScop std::vector<std::string> I3E_EVQ_FindCancelledEventsInQueue();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Find events with active flag
         * 
         * @details This function will search for events in the queue that have the active flag set. The function will
         * return a vector of hrefs of the events that have the active flag set.
         * 
         * @returns A vector of hrefs of events that have the active flag set.
         */
        ModScop std::vector<std::string> I3E_EVQ_FindEventsWithActiveFlag();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Add default event to queue
         * 
         * @details This function will add a new default event to the queue. The function will push the new default event
         * to the back of the queue.
         * 
         * @param str_newDefault The new default event to add to the queue
         */
        ModScop void I3E_EVQ_AddDefaultToQueue(const DefaultBlock& str_newDefault);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Remove default from queue
         * 
         * @details This function will remove a default event from the queue based on the href of the event. The function
         * will iterate through the queue and remove the default event that matches the href that is passed in.
         * 
         * @param sz_href The href of the default event to remove from the queue
         * 
         * @returns True if the default event was added to the queue, false otherwise
         */
        ModScop bool I3E_EVQ_RemoveDefaultFromQueue(const std::string& sz_href);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /**
         * @brief Replace default der control for event
         * 
         * @details This function will replace the DefaultDERControl for a specific event in the queue. The function will
         * iterate through the queue and replace the DefaultDERControl for the event that matches the href that is passed in.
         * 
         * @param sz_eventHref  The href of the event to replace the DefaultDERControl for
         * @param i3e_newDefDerCtrl The new DefaultDERControl to replace the existing DefaultDERControl
         */
        ModScop void I3E_EVQ_ReplaceDefDerCtrlForEvent(const std::string& sz_eventHref, 
                        const SE_DefaultDERControl_t& i3e_newDefDerCtrl);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Find missing events in default queue
         * 
         * @details This function will compare the hrefs of the DERControl objects in the default queue with the hrefs of the
         * DERControl objects in the vector passed in. If an href is not found in the default queue, it will be returned in a
         * vector of missing events for the calling function to add to the queue.
         * 
         * @param vec_derCtrlHrefs The vector of hrefs to compare with the default queue
         * 
         * @returns A vector of hrefs that are missing in the default queue
         */
        ModScop std::vector<std::string> I3E_EVQ_FindMissingEventsInDefaultQueue(const std::vector<std::string>& vec_derCtrlHrefs);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Remove non existant defaults
         * 
         * @details This function will remove default events from the queue that do not exist in the vector of hrefs
         * passed in. The function will iterate through the queue and remove the default events that do not exist in the
         * vector of hrefs.
         * 
         * @param vec_derCtrlHrefs The vector of hrefs to compare with the default queue
         * 
         * @returns A vector of hrefs that were removed from the default queue
         */
        ModScop std::vector<std::string> I3E_EVQ_RemoveNonexistentDefaults(const std::vector<std::string>& vec_defCtrlHrefs);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Compare control flag types 
         * 
         * @details This function will compare the control type flags of the new default with the control type flags of the
         * existing defaults in the queue. The function will return a pointer to the overlapping default if there is an
         * overlap in the control type flags.
         * 
         * @param i3e_newDefault The new default to compare with the existing defaults in the queue
         * 
         * @returns A pointer to the overlapping default if there is an overlap in the control type flags, nullptr
         */
        ModScop SE_DefaultDERControl_t* I3E_EVQ_CompareControlTypeFlags(const SE_DefaultDERControl_t& i3e_newDefault);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Print defaults in queue
         * 
         * @details This function will print the details of the default events in the queue. The function will iterate
         * through the queue and print the details of each default event.
         */
        ModScop void I3E_EVQ_PrintDefaultsInQueue();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Get all hrefs from default queue.
         * 
         * @details This function will return a vector of hrefs of all the default events in the queue.
         * 
         * @returns A vector of hrefs of all the default events in the queue
         */
        ModScop std::vector<std::string> I3E_EVQ_GetAllHrefsFromDefaultQueue();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Get default block by href
         * 
         * @details This function will find a default event in the queue based on the href of the event. The function will
         * iterate through the queue and return the DefaultBlock that matches the href that is passed in.
         * 
         * @param sz_href The href of the default event to find in the queue
         * 
         * @returns The DefaultBlock of the event that matches the href
         */
        ModScop DefaultBlock I3E_EVQ_GetDefaultBlockByHref(const std::string& sz_href);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Order default queue by primacy of program
         * 
         * @details This function will order the default queue based on the primacy of the program. The function will extract
         * the elements from the queue into a vector, sort the vector by primacy, and reinsert the sorted elements back into
         * the queue.
         */
        ModScop void I3E_EVQ_OrderDefaultQueueByPrimacy();

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Add href at index
         * 
         * @details This function will add an href at a specific index in the active controls hrefs vector. The function will
         * lock the mutex, check if the index is within bounds, and add the href at the specified index.
         * 
         * @param sz_href  The href to add
         * @param sz_mrid  The mRID to add
         * @param siz_index The index to add the href at
         */
        ModScop void I3E_EVQ_AddHrefAtIndex(const std::string& sz_href, const std::string& sz_mrid, size_t siz_index);
        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Get href at index
         * 
         * @details This function will get an href at a specific index in the active controls hrefs vector. The function will
         * lock the mutex, check if the index is within bounds, and return the href at the specified index.
         * 
         * @param siz_index The index to get the href from
         * 
         * @returns The href at the specified index
         */
        ModScop std::string I3E_EVQ_GetHrefAtIndex(size_t siz_index);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Add default to href at index
         * 
         * @details This function will add a default href at a specific index in the active defaults hrefs vector. The function
         * will lock the mutex, check if the index is within bounds, and add the default href at the specified index.
         * 
         * @param sz_href  The href to add
         * @param sz_mrid  The mRID to add
         * @param siz_index The index to add the href at
         */
        ModScop void I3E_EVQ_AddDefaultHrefAtIndex(const std::string& sz_href, const std::string& sz_mrid, size_t siz_index);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Get default href at index
         * 
         * @details This function will get a default href at a specific index in the active defaults hrefs vector. The function
         * will lock the mutex, check if the index is within bounds, and return the default href at the specified index.
         * 
         * @param siz_index The index to get the default href from
         * 
         * @returns The default href at the specified index
         */
        ModScop std::string I3E_EVQ_GetDefaultHrefAtIndex(size_t siz_index);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /**
         * @brief Add settings to href at index
         * 
         * @details This function will add a settings href at a specific index in the active settings hrefs vector. The function
         * will lock the mutex, check if the index is within bounds, and add the settings href at the specified index.
         * 
         * @param sz_href  The href to add
         * @param sz_mrid  The mRID to add
         * @param siz_index The index to add the href at
         */
        ModScop void I3E_EVQ_AddSettingsHrefAtIndex(const std::string& sz_href, const std::string& sz_mrid, size_t siz_index);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        /**
         * @brief Get settings href at index
         * 
         * @details This function will get a settings href at a specific index in the active settings hrefs vector. The function
         * will lock the mutex, check if the index is within bounds, and return the settings href at the specified index.
         * 
         * @param siz_index The index to get the settings href from
         * 
         * @returns The settings href at the specified index
         */
        ModScop std::string I3E_EVQ_GetSettingsHrefAtIndex(size_t siz_index);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Remove href and clear flags 
         * 
         * @details This function will remove an href from the active controls hrefs vector and clear the flags associated
         * with the href. The function will lock the mutex, iterate through the vector, clear the href, and clear the flags
         * 
         * @param sz_href The href to remove and clear flags for
         * 
         * @returns Bitfield indicating which control types were removed (controls to reset to defaults)
         */
        ModScop uint32_t I3E_EVQ_RemoveHrefAndClearFlags(const std::string& sz_href);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Remove default href and clear flags 
         * 
         * @details This function will remove a default href from the active default controls hrefs vector and clear the
         * flags associated with the href. The function will lock the mutex, iterate through the vector, clear the href, 
         * and clear the flags
         * 
         * @param sz_href The href to remove and clear flags for
         * 
         * @returns Bitfield indicating which control types were removed
         */
        ModScop uint32_t I3E_EVQ_RemoveDefaultHrefAndClearFlags(const std::string& sz_href);

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        /**
         * @brief Remove settings href and clear flags 
         * 
         * @details This function will remove a settings href from the active settings controls hrefs vector and clear the
         * flags associated with the href. The function will lock the mutex, iterate through the vector, clear the href, 
         * and clear the flags
         * 
         * @param sz_href The href to remove and clear flags for
         * 
         * @returns Bitfield indicating which control types were removed
         */

        ModScop uint32_t I3E_EVQ_RemoveSettingsHrefAndClearFlags(const std::string& sz_href);

};


#endif