/**********************************************************************************************************************
 * @file            I3E_HTTP.hpp
 *
 * @brief           The HTTP module contains the functions for performing HTTP transfers with the IEEE 2030.5 server. 
 *                  The module utilizes the libCurl library for structuring HTTP requests and the necessary context
 *                  configuration for receiving response codes and response bodies. 
 *    
 *
 * @author          Vlad Bugayev
 * @date            03/27/2024
 * @copyright       2024 Lion Energy LLC.
 *                  All rights reserved
 *
 *********************************************************************************************************************/
/**
 * @addtogroup      HTTP
 *                  
 *
 * @ref             I3E_HTTP.hpp
 *
 *********************************************************************************************************************/

#include "COD_Compiler_Defines.h"

#ifndef LION_I3E_HTTP_H__
#define LION_I3E_HTTP_H__

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#include "lion/sch/I3E_Schema.hpp"
#include <string>
#include <bits/stdc++.h>

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define HTTP_VER "HTTP/1.1"
#define HTTP_USR "Lion EMS"
#define HTTP_SES "Lion HTTPSClientSession"
#define HTTP_ACCEPT "Accept: "
#define HTTP_SEPXML "application/sep+xml"

#define HTTP_CONTENTTYPE "Content-Type: "
#define HTTP_USERAGENT "User-Agent: "
#define HTTP_SESSION "Session: "
#define HTTP_LOCATION "Location: "

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

enum class HTTPResponse {
    OK = 200,
    CREATED = 201,
    ACCEPTED = 202,
    NO_CONTENT = 204,
    MOVED_PERMANENTLY = 301,
    FOUND = 302,
    NOT_MODIFIED = 304,
    BAD_REQUEST = 400,
    UNAUTHORIZED = 401,
    FORBIDDEN = 403,
    NOT_FOUND = 404,
    METHOD_NOT_ALLOWED = 405,
    INTERNAL_SERVER_ERROR = 500,
    BAD_GATEWAY = 502,
    SERVICE_UNAVAILABLE = 503,
    GATEWAY_TIMEOUT = 504
};

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

/**
 * @brief Initialize the HTTP module.
 * 
 * @details Initializes the HTTP module and performs any necessary setup. This simply initializes an SSL Mutex for 
 * read/write protection.
 * 
 * @returns An integer value indicating the status of the initialization process.
 *    0: Initialization successful.
 *    Non-zero: Initialization failed.
 */
ModScop int I3E_HTTP_Init(void);

/**
 * @brief Perform an HTTP GET request.
 * 
 * @details This function performs an HTTP GET request to the specified server and URI. The function will return the 
 * response code and the response body.
 * 
 * @param sz_pIPV4addr The IPv4 address of the server.
 * @param u16_portNum The port number of the server.
 * @param sz_pURI The URI requested.
 * @param sz_rxPayload The response body.
 * 
 * @returns The response code.
 */
ModScop HTTPResponse I3E_HTTP_GET(std::string sz_pIPV4addr, uint16_t u16_portNum, 
                                  std::string sz_pURI, std::string& sz_rxPayload);

/**
 * @brief Perform an HTTP POST request.
 * 
 * @details This function performs an HTTP POST request to the specified server and URI. The function will return the 
 * response code and the response body.
 * 
 * @param sz_pIPV4addr The IPv4 address of the server.
 * @param u16_portNum The port number of the server.
 * @param sz_pURI The URI requested.
 * @param sz_txPayload The request body.
 * @param sz_rxPayload The response body.
 * 
 * @returns The response code.
 */
ModScop HTTPResponse I3E_HTTP_POST(std::string sz_pIPV4addr, uint16_t u16_portNum, 
                                   std::string sz_pURI, std::string& sz_txPayload, std::string& sz_rxPayload);

/**
 * @brief Perform an HTTP PUT request.
 * 
 * @details This function performs an HTTP PUT request to the specified server and URI. The function will return the 
 * response code and the response body.
 * 
 * @param sz_pIPV4addr The IPv4 address of the server.
 * @param u16_portNum The port number of the server.
 * @param sz_pURI The URI requested.
 * @param sz_txPayload The request body.
 * @param sz_rxPayload The response body.
 * 
 * @returns The response code.
 */
ModScop HTTPResponse I3E_HTTP_PUT(std::string sz_pIPV4addr, uint16_t u16_portNum, 
                                  std::string sz_pURI, std::string& sz_txPayload, std::string& sz_rxPayload);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#endif /* LION_I3E_HTTP_H__ */

