/**********************************************************************************************************************
 * @file            I3E_NSS.hpp
 *
 * @brief           The IEEE 2030.5 Network Security Service (SRS) manages & initializes the TLS context for client
 *                  and server communication. The service also performs various PKI certificate verification, 
 *                  accorinding to the IEEE 2030.5 specification. 
 *    
 *
 * @author          Vlad Bugayev
 * @date            03/18/2024
 * @copyright       2024 Lion Energy LLC.
 *                  All rights reserved
 *
 *********************************************************************************************************************/
/**
 * @addtogroup      NSS
 *                  Network Security Service    
 *
 * @ref             I3E_NSS.hpp
 *
 *********************************************************************************************************************/

#include "COD_Compiler_Defines.h"

#ifndef LION_I3E_NSS_H__
#define LION_I3E_NSS_H__

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#include "lion/sch/I3E_Schema.hpp"
#include "lion/icc/ICC.h"

#include <string>
#include <fstream>
#include <iomanip>
#include <openssl/bio.h>
#include <curl/curl.h>

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
 *********************************************************************************************************************/

ModScop const char sz_TLSCipherFilter[] = {"ALL:!ADH:!LOW:!EXP:!MD5:@STRENGTH"};
ModScop const char sz_TLSElipCurve[] = {"secp256r1"};
ModScop const char sz_CertType[] = {"PEM"};

/**
 * @brief Initialize the NSS (Network Security Service).
 * 
 * @details Initializes the NSS module and performs any necessary setup. This includes setting up the curl context,
 * and some extra TLS parameters.
 * 
 * @returns An integer value indicating the status of the initialization process.
 *    0: Initialization successful.
 *    Non-zero: Initialization failed.
 */
ModScop int I3E_NSS_Init(void);

/**
 * @brief Custom PKI cert indicator calculation.
 * 
 * @details This function will load the PKI certificate and calculate the sfdi and lfdi values and store them in the
 * provided pointers.
 * 
 * @param i3e_lFDI The LFDI value storage pointer (I3E schema type).
 * @param i3e_sFDI The SFDI value storage pointer (I3E schema type). 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
 */
ModScop int I3E_NSS_CalculateIndicators(SE_HexBinary160_t *i3e_lFDI, SE_SFDIType_t *i3e_sFDI);

/**
 * @brief Check for the PKI certificates.
 * 
 * @details Using the i3e configuration file, this function will use the PKI directory locations to check for existing 
 * PKI certificates.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
 */
ModScop int I3E_NSS_CheckPKI(void);

/**
 * @brief Generate new PKI certificates.
 * 
 * @details This function will generate new PKI certificates by calling the Lion AWS PKI portal. The certificates will be
 * downloaded and stored in the appropriate locations (config file will be updated to have the locations of the PKI).
 * 
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
 */
ModScop int I3E_NSS_GeneratePKI(void);

/**
 * @brief Get the SFDI of the certificate.
 * 
 * @details This function returns the SFDI of the certificate.
 * 
 * @returns The SFDI of the certificate.
 */
ModScop SE_SFDIType_t I3E_NSS_GetSFDI(void);

/**
 * @brief Get the LFDI of the certificate.
 * 
 * @details This function returns the LFDI of the certificate.
 * 
 * @returns The LFDI of the certificate.
 */
ModScop SE_HexBinary160_t* I3E_NSS_GetLFDI(void);

/**
 * @brief Get the PIN of the certificate.
 * 
 * @details This function returns the PIN of the certificate.
 * 
 * @returns The PIN of the certificate.
 */
ModScop SE_PINType_t I3E_NSS_GetPIN(void);

/**
 * @brief Get the IPv4 address.
 * 
 * @details This function returns the IPv4 address.
 * 
 * @returns A string representing the IPv4 address.
 */
ModScop std::string I3E_NSS_GetIPV4(void);

/**
 * @brief Get the port number.
 * 
 * @details This function returns the port number for the TLS connection.
 * 
 * @returns A 32 bit unsigned integer representing the port number.
 */
ModScop uint16_t I3E_NSS_GetPORT(void);

/**
 * @brief Get the server certificate.
 * 
 * @details This function returns the server certificate.
 * 
 * @returns A string representing the server certificate.
 */
ModScop std::string I3E_NSS_GetPKEY(void);

/**
 * @brief Get the client certificate.
 * 
 * @details This function returns the client certificate.
 * 
 * @returns A string representing the client certificate.
 */
ModScop std::string I3E_NSS_GetCERT(void);

/**
 * @brief Get the certificate authority.
 * 
 * @details This function returns the certificate authority.
 * 
 * @returns A string representing the certificate authority.
 */
ModScop std::string I3E_NSS_GetCA(void);

/**
 * @brief Get the URI.
 * 
 * @details This function returns the DCAP URI.
 * 
 * @returns A string representing the DCAP URI.
 */
ModScop std::string I3E_NSS_GetURI(void);

/**
 * @brief Set the URI.  
 * 
 * @details This function sets the URI for the DCAP HTTP request.
 */
ModScop void I3E_NSS_SetURI(std::string sz_uri);

/**
 * @brief Get the configured CURL context.
 * 
 * @details This function returns a pointer to the configured CURL context.
 * 
 * @returns A pointer to the configured CURL context.
 */
ModScop CURL* I3E_NSS_GetCURL(void);

/**
 * @brief Get the certificate status.
 * 
 * @details This function returns the validity status of the certificate.
 * 
 * @returns A boolean representing the validity status of the certificate.
 */
ModScop bool I3E_NSS_GetCertValid(void);

/**
 * @brief Custom SSL context callback.
 * 
 * @details This callback is used whenever an HTTP request is triggered by the client. This is the function that will 
 * load all the certificates and keys into the SSL context and then perform the custom server certificate verification.
 * This is automatically triggered by the CURL library.
 * 
 * @param ptr_curl The CURL context.
 * @param ptr_sslctx The SSL context.
 * @param ptr_parm Optional additional argument (not used)
 * 
 * @returns Return description.
 */    
ModScop CURLcode I3E_NSS_sslCtxCallback(CURL* ptr_curl, void* ptr_sslctx, void* ptr_parm);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#endif /* LION_I3E_NSS_H__ */



