/**********************************************************************************************************************
 * @file            I3E_SPP.hpp
 *
 * @brief           The Server Poll Processor (SPP) module is responsible for periodically polling the IEEE 2030.5 
 *                  server and discovering resources at the defined rates within the resources (pollRate). This module
 *                  will determine if resources have changed since the last poll, and run a rediscovery routine from 
 *                  that graph node and store the acquired data in the DAG module. 
 *    
 *
 * @author          Vlad Bugayev
 * @date            03/18/2024
 * @copyright       2024 Lion Energy LLC.
 *                  All rights reserved
 *
 *********************************************************************************************************************/
/**
 * @addtogroup      SPP
 *                  Server Poll Processor 
 *
 * @ref             I3E_SPP.hpp
 *
 *********************************************************************************************************************/

#include "COD_Compiler_Defines.h"

#ifndef LION_I3E_SRS_H__
#define LION_I3E_SRS_H__

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

typedef enum {
    DCAP,
    TM,
    REG,
    EDEV_LIST, 
    EDEV,
    FSA_LIST,
    FSA,
    DERPROG_LIST,
    DERPROG, 
    DERCTRL_LIST,
    DERCTRL,
    DERCURVE_LIST,
    DERCURVE,
    DEFDERCTRL,
    DEFDERCTRL_LIST,
    DER_LIST,
    DER,
    DERSTAT,
    DERCAP,
    DERSETT,
    DERAVAIL,
    MUP,
    MMR,
    LOGEVENT
} I3EresourceType;

struct Graph;


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#include "I3E_HTTP.hpp"
#include "I3E_DEM.hpp"
#include "I3E_DAG.hpp"

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define MAX_RSRC_RET_SUFFIX  "?l=255"

/**********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

/**
 * @brief Perform the initial resource discovery on the IEEE 2030.5 server.
 * 
 * @details This function will run through a discovery process on the IEEE 2030.5 server to acquire the initial 
 * resources that are available. The function will store the acquired data in the DAG. Included in this process are
 * a few checks to make sure the server is hosting the correct resources for the specified client device. 
 * This is done with the LFDI, SFDI and PIN values. 
 * 
 * @returns A boolean value indicating the success of the resource discovery process.
 */
// #TODO - this function currently doesn't do HTTP code checking on ALL of the HTTP GET requests, just the first few
ModScop bool I3E_SPP_InitResourceDiscovery(void);

/**
 * @brief Thread execution function for the SPP module.
 *
 * @details This function is the main execution function for the SPP module. It is responsible for periodically
 * polling the IEEE 2030.5 server for all resource updates on resources that contain a pollRate. If a resource has
 * changed since the last poll, the function will run a rediscovery routine from that graph node and store the
 * acquired data in the DAG module.
 * 
 * This function should never return (infinite looping runtime behavior). 
 */
ModScop void I3E_SPP_Exec(void);

/**
 * @brief Get the DAG object for the SPP module.
 * 
 * @details This function returns a pointer to the DAG object for the SPP module.
 * 
 * @returns Pointer to the DAG object.
 */
ModScop Graph* I3E_SPP_GetDAG(void);

/**
 * @brief Store the MirrorUsagePoint resource in the DAG.
 * 
 * @details This function is called by the SDA module which handles the initial resource posting for the MUP resource. 
 * The function will store the MirrorUsagePoint resource in the DAG.
 * 
 * @param sz_parentNodeHref The parent node href of the MirrorUsagePoint resource.
 * @param sz_rxPayload The response body of the HTTP GET request for the created MUP resource.
 */
ModScop void I3E_SPP_StoreMirrorUsagePoint(std::string sz_parentNodeHref, std::string sz_rxPayload);


#endif /* LION_I3E_SRS_H__ */

















