/**********************************************************************************************************************
 * @file            I3E_XMLD.hpp
 *
 * @brief           The XML Deserializer module contains utility functions for taking incoming XML payloads from the 
 *                  IEEE 2030.5 server and parsing the information out into C structures defined in the I3E_Schema.h
 *                  folder. It currently utilizes the Poco C++ DOMParser to traverse XML payloads. 
 *    
 *
 * @author          Vlad Bugayev
 * @date            03/27/2024
 * @copyright       2024 Lion Energy LLC.
 *                  All rights reserved
 *
 *********************************************************************************************************************/
/**
 * @addtogroup      XMLD
 *                  XML Deserializer
 *
 * @ref             I3E_XMLD.hpp
 *
 *********************************************************************************************************************/

#include "COD_Compiler_Defines.h"

#ifndef LION_I3E_XMLD_H__
#define LION_I3E_XMLD_H__

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#include "lion/sch/I3E_Schema.hpp"
#include "Poco/DOM/Element.h"
#include "I3E_SPP.hpp"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// GENERAL TAGS

#define XSD_TAG_NS                 "urn:ieee:std:2030.5:ns"
#define XSD_TAG_HREF               "href"
#define XSD_TAG_ALL                "all"
#define XSD_TAG_RESULTS            "results"
#define XSD_TAG_SUBSCRIBABLE       "subscribable"
#define XSD_TAG_REPLYTO            "replyTo"
#define XSD_TAG_RESPONSEREQUIRED   "responseRequired"
#define XSD_TAG_POLLRATE           "pollRate"
#define XSD_TAG_TRUE               "true"
#define XSD_TAG_FALSE              "false"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DEVICE CAPABILITY TAGS

#define XSD_TAG_DCAP_DEVICECAPABILITY      "DeviceCapability"
#define XSD_TAG_DCAP_RESPONSESET_LL        "ResponseSetListLink"
#define XSD_TAG_DCAP_TIMELINK              "TimeLink"
#define XSD_TAG_DCAP_USAGEPOINT_LL         "UsagePointListLink"
#define XSD_TAG_DCAP_ENDDEVICE_LL          "EndDeviceListLink"
#define XSD_TAG_DCAP_MIRRORUSAGEPOINT_LL   "MirrorUsagePointListLink"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// END DEVICE TAGS

#define XSD_TAG_EDEV_ENDDEVICELIST     "EndDeviceList"
#define XSD_TAG_EDEV_ENDDEVICE         "EndDevice"
#define XSD_TAG_EDEV_LFDI              "lFDI"
#define XSD_TAG_EDEV_LOGEVENT_LL       "LogEventListLink"
#define XSD_TAG_EDEV_SFDI              "sFDI"
#define XSD_TAG_EDEV_CHANGEDTIME       "changedTime"
#define XSD_TAG_EDEV_ENABLED           "enabled"
#define XSD_TAG_EDEV_REGISTRATIONLINK  "RegistrationLink"
#define XSD_TAG_EDEV_SUBSCRIPTION_LL   "SubscriptionListLink"
#define XSD_TAG_EDEV_DER_LL            "DERListLink"
#define XSD_TAG_EDEV_FSA_LL            "FunctionSetAssignmentsListLink"
#define XSD_TAG_EDEV_DEVICECATEGORY    "deviceCategory"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MIRROR USAGE POINT TAGS

#define XSD_TAG_MUP_MUP                "MirrorUsagePoint"
#define XSD_TAG_MUP_MRID               "mRID"
#define XSD_TAG_MUP_DESC               "description"
#define XSD_TAG_MUP_ROLEFLAGS          "roleFlags"
#define XSD_TAG_MUP_SERVCATKIND        "serviceCategoryKind"
#define XSD_TAG_MUP_STAT               "status"
#define XSD_TAG_MUP_DEVLFDI            "deviceLFDI"
#define XSD_TAG_MUP_POSTRATE           "postRate"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// TIME TAGS

#define XSD_TAG_TM_TIME          "Time"
#define XSD_TAG_TM_CURRENTTIME   "currentTime"
#define XSD_TAG_TM_DSTENDTIME    "dstEndTime"
#define XSD_TAG_TM_DSTOFFSET     "dstOffset"
#define XSD_TAG_TM_DSTSTARTTIME  "dstStartTime"
#define XSD_TAG_TM_LOCALTIME     "localTime"
#define XSD_TAG_TM_QUALITY       "quality"
#define XSD_TAG_TM_TZOFFSET      "tzOffset"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// REGISTRATION TAGS

#define XSD_TAG_REG_REGISTRATION    "Registration"
#define XSD_TAG_REG_REGDATETIME     "dateTimeRegistered"
#define XSD_TAG_REG_PIN             "pIN"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DER TAGS

#define XSD_TAG_DER_DERLIST          "DERList"
#define XSD_TAG_DER_DER              "DER"
#define XSD_TAG_DER_DERASSOC_LL      "AssociatedDERProgramListLink"
#define XSD_TAG_DER_DERCURRPROGLINK  "CurrentDERProgramLink"                              
#define XSD_TAG_DER_DERAVAIL_LL      "DERAvailabilityLink"
#define XSD_TAG_DER_DERCAP_LL        "DERCapabilityLink"
#define XSD_TAG_DER_DERSET_LL        "DERSettingsLink"
#define XSD_TAG_DER_DERSTAT_LL       "DERStatusLink"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DER AVAILABILITY TAGS

#define XSD_TAG_DERAVAIL_READTIME       "readingTime"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DER CAPABILITY TAGS

#define XSD_TAG_DERCAP_MODESSUPPORTED   "modesSupported"
#define XSD_TAG_DERCAP_RTGMAXW          "rtgMaxW"
#define XSD_TAG_DERCAP_MULT             "multiplier"
#define XSD_TAG_DERCAP_VAL              "value"    
#define XSD_TAG_DERCAP_TYPE             "type"    

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DER SETTINGS TAGS

#define XSD_TAG_DERSETT_SETGRADW        "setGradW"         
#define XSD_TAG_DERSETT_SETMAXW         "setMaxW"         
#define XSD_TAG_DERSETT_MULT            "multiplier"         
#define XSD_TAG_DERSETT_VAL             "value"         
#define XSD_TAG_DERSETT_UPDATEDTIME     "updatedTime"                 

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DER STATUS TAGS

#define XSD_TAG_DERSTAT_READTIME        "readingTime"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// FUNCTION SET ASSIGNMENT TAGS

#define XSD_TAG_FSA_FSALIST      "FunctionSetAssignmentsList"
#define XSD_TAG_FSA_FSA          "FunctionSetAssignments"
#define XSD_TAG_FSA_MRID         "mRID"
#define XSD_TAG_FSA_DESCRIPTION  "description"
#define XSD_TAG_FSA_DERPROG_LL   "DERProgramListLink" 
#define XSD_TAG_FSA_TIMELINK     "TimeLink"
#define XSD_TAG_FSA_RSPSET_LL    "ResponseSetListLink"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DER PROGRAM TAGS

#define XSD_TAG_DERP_DERPROGLIST             "DERProgramList"
#define XSD_TAG_DERP_DERPROG                 "DERProgram"
#define XSD_TAG_DERP_MRID                    "mRID"
#define XSD_TAG_DERP_DESCRIPTION             "description"
#define XSD_TAG_DERP_ACTIVEDERCONTROL_LL     "ActiveDERControlListLink"
#define XSD_TAG_DERP_DERCONTROL_LL           "DERControlListLink"
#define XSD_TAG_DERP_DERCURVE_LL             "DERCurveListLink"
#define XSD_TAG_DERP_PRIMACY                 "primacy"
#define XSD_TAG_DERP_DEFAULTDERCONTROL_LL    "DefaultDERControlLink"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DEFAULT DER CONTROL TAGS
 
#define XSD_TAG_DEFCON_DEFCTRL                  "DefaultDERControl"
#define XSD_TAG_DEFCON_MRID                     "mRID"
#define XSD_TAG_DEFCON_DESCRIPTION              "description"
#define XSD_TAG_DEFCON_VERSION                  "version"
#define XSD_TAG_DEFCON_SETESDEL                 "setESDelay"
#define XSD_TAG_DEFCON_SETESHF                  "setESHighFreq"
#define XSD_TAG_DEFCON_SETESHV                  "setESHighVolt"
#define XSD_TAG_DEFCON_SETESLF                  "setESLowFreq"
#define XSD_TAG_DEFCON_SETESLV                  "setESLowVolt"
#define XSD_TAG_DEFCON_SETESRAMPTMS             "setESRampTms"
#define XSD_TAG_DEFCON_SETESRANDDEL             "setESRandomDelay"
#define XSD_TAG_DEFCON_SETGRADW                 "setGradW"
#define XSD_TAG_DEFCON_SETSOFTGRADW             "setSoftGradW"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DER CONTROL TAGS

#define XSD_TAG_DERCON_DERCCTRLLIST            "DERControlList"           
#define XSD_TAG_DERCON_DERCTRL                 "DERControl"
#define XSD_TAG_DERCON_MRID                    "mRID"
#define XSD_TAG_DERCON_DESCRIPTION             "description"
#define XSD_TAG_DERCON_VERSION                 "version"
#define XSD_TAG_DERCON_CREATETIME              "creationTime"
#define XSD_TAG_DERCON_EVENTSTATUS             "EventStatus"
#define XSD_TAG_DERCON_CURRENTSTATUS           "currentStatus"            
#define XSD_TAG_DERCON_DATETIME                "dateTime"
#define XSD_TAG_DERCON_POTSUPERSEDE            "potentiallySuperseded"    
#define XSD_TAG_DERCON_POTSUPERSEDETIME        "potentiallySupersededTime"            
#define XSD_TAG_DERCON_REASON                  "reason"        
#define XSD_TAG_DERCON_INTERVAL                "interval"            
#define XSD_TAG_DERCON_DURATION                "duration"        
#define XSD_TAG_DERCON_START                   "start"            
#define XSD_TAG_DERCON_RANDOMDURATION          "randomizeDuration"        
#define XSD_TAG_DERCON_RANDOMSTART             "randomizeStart"        
#define XSD_TAG_DERCON_DERCTRLBASE             "DERControlBase"    

#define XSD_TAG_DERCON_OPMODCONNECT            "opModConnect"
#define XSD_TAG_DERCON_OPMODENERGIZE           "opModEnergize"
#define XSD_TAG_DERCON_OPMODHVRTMOMCESS        "opModHVRTMomentaryCessation"
#define XSD_TAG_DERCON_OPMODLVRTMOMCESS        "opModLVRTMomentaryCessation"
#define XSD_TAG_DERCON_OPMODHVRTMUSTTRIP       "opModHVRTMustTrip"
#define XSD_TAG_DERCON_OPMODLVRTMUSTTRIP       "opModLVRTMustTrip"
#define XSD_TAG_DERCON_OPMODHFRTMUSTTRIP       "opModHFRTMustTrip"
#define XSD_TAG_DERCON_OPMODLFRTMUSTTRIP       "opModLFRTMustTrip"
#define XSD_TAG_DERCON_OPMODHVRTMAYTRIP        "opModHVRTMayTrip"
#define XSD_TAG_DERCON_OPMODLVRTMAYTRIP        "opModLVRTMayTrip"
#define XSD_TAG_DERCON_OPMODHFRTMAYTRIP        "opModHFRTMayTrip"
#define XSD_TAG_DERCON_OPMODLFRTMAYTRIP        "opModLFRTMayTrip"
#define XSD_TAG_DERCON_OPMODMAXLIMW            "opModMaxLimW"
#define XSD_TAG_DERCON_OPMODVOLTWATT           "opModVoltWatt"
#define XSD_TAG_DERCON_OPMODVOLTVAR            "opModVoltVar"
#define XSD_TAG_DERCON_OPMODWATTPF             "opModWattPF" 
#define XSD_TAG_DERCON_OPMODFREQWATT           "opModFreqWatt"    
#define XSD_TAG_DERCON_OPMODFREQDROOP          "opModFreqDroop"
#define XSD_TAG_DERCON_DBOF                    "dBOF"    
#define XSD_TAG_DERCON_DBUF                    "dBUF"    
#define XSD_TAG_DERCON_KOF                     "kOF"    
#define XSD_TAG_DERCON_KUF                     "kUF"    
#define XSD_TAG_DERCON_OPENLOOPTMS             "openLoopTms"    

#define XSD_TAG_DERCON_OPMODFIXEDW             "opModFixedW"    
#define XSD_TAG_DERCON_OPMODTARGETW            "opModTargetW"

#define XSD_TAG_DERCON_OPMODFIXPFINJECTW       "opModFixedPFInjectW"
#define XSD_TAG_DERCON_OPMODFIXPFABSORBW       "opModFixedPFAbsorbW"  
#define XSD_TAG_DERCON_DISPLACEMENT            "displacement"            
#define XSD_TAG_DERCON_EXCITATION              "excitation"    
#define XSD_TAG_DERCON_MULTIPLIER              "multiplier"        
#define XSD_TAG_DERCON_VALUE                   "value"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DER CURVE TAGS

#define XSD_TAG_DERCUR_DERCURVELIST          "DERCurveList"
#define XSD_TAG_DERCUR_DERCURVE              "DERCurve"
#define XSD_TAG_DERCUR_MRID                  "mRID"
#define XSD_TAG_DERCUR_DESCRIPTION           "description"
#define XSD_TAG_DERCUR_AUTOVREFEN            "autonomousVRefEnable"
#define XSD_TAG_DERCUR_AUTOVREFTC            "autonomousVRefTimeConstant"
#define XSD_TAG_DERCUR_VREF                  "vRef"
#define XSD_TAG_DERCUR_CREATETIME            "creationTime"
#define XSD_TAG_DERCUR_CURVEDATA             "CurveData"
#define XSD_TAG_DERCUR_XVAL                  "xvalue"
#define XSD_TAG_DERCUR_YVAL                  "yvalue"
#define XSD_TAG_DERCUR_CURVETYPE             "curveType"
#define XSD_TAG_DERCUR_RAMPDECTMS            "rampDecTms"
#define XSD_TAG_DERCUR_RAMPINCTMS            "rampIncTms"
#define XSD_TAG_DERCUR_RAMPPT1TMS            "rampPT1Tms"
#define XSD_TAG_DERCUR_OPENLOOPTMS           "openLoopTms"
#define XSD_TAG_DERCUR_XMULT                 "xMultiplier"
#define XSD_TAG_DERCUR_YMULT                 "yMultiplier"
#define XSD_TAG_DERCUR_YREFTYPE              "yRefType"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

/**
 * @brief Deserialize an incoming XML payload.
 * 
 * @details This function multiplexes the incoming XML payload to the appropriate deserialization function based on the
 * resource type. The function will then call the appropriate deserialization function to parse the XML payload into the
 * appropriate C structure.
 * 
 * @param sz_rxPayload The incoming XML payload.
 * @param enum_rSrcType The type of resource being deserialized.
 * @param ptr_rSrc Abstracted pointer that is typecasted to the appropriate type within the deserialization function.
 */
ModScop void I3E_XML_PayloadDeserializeMux(std::string& sz_rxPayload, I3EresourceType enum_rSrcType, void* ptr_rSrc);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#endif /* LION_I3E_XMLD_H__ */

