/**********************************************************************************************************************
 * @file            I3E_XMLS.hpp
 *
 * @brief           The XML Serializer module contains utility functions for packaging outgoing XML payloads utilizing
 *                  the fields within the C style structs containing relevent IEEE 2030.5 information. Currently utilizes
 *                  the Poco C++ XML DOMParser to achieve this. 
 *    
 *
 * @author          Vlad Bugayev
 * @date            03/27/2024
 * @copyright       2024 Lion Energy LLC.
 *                  All rights reserved
 *
 *********************************************************************************************************************/
/**
 * @addtogroup      XMLS
 *                  XML Seerializer
 *
 * @ref             I3E_XMLS.hpp
 *
 *********************************************************************************************************************/

#include "COD_Compiler_Defines.h"

#ifndef LION_I3E_XMLS_H__
#define LION_I3E_XMLS_H__

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#include <bits/stdc++.h>
#include "lion/sch/I3E_Schema.hpp"

#include "Poco/DOM/Document.h"
#include "Poco/DOM/AutoPtr.h"

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// GENERAL TAGS

#define XSD_TAG_NS                 "urn:ieee:std:2030.5:ns"
#define XSD_TAG_XMLNS              "xmlns"
#define XSD_TAG_XSI                "xmlns:xsi" 
#define XSD_TAG_XSIURL             "http://www.w3.org/2001/XMLSchema-instance"
#define XSD_TAG_XSISL              "xsi:schemaLocation" 
#define XSD_TAG_XSD                "urn:ieee:std:2030.5:ns sep.xsd"

#define XSD_TAG_MULT               "multiplier"
#define XSD_TAG_VAL                "value"
#define XSD_TAG_DATETIME           "dateTime"
#define XSD_TAG_DISPLACEMENT       "displacement"

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DER CONTROL RESPONSE TAGS

#define XSD_TAG_DERCRSP            "DERControlResponse"
#define XSD_TAG_DERCRSP_CRDT       "createdDateTime"
#define XSD_TAG_DERCRSP_EDEVLFDI   "endDeviceLFDI"
#define XSD_TAG_DERCRSP_STAT       "status"
#define XSD_TAG_DERCRSP_SUBJECT    "subject"

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DER SETTINGS TAGS

#define XSD_TAG_DERSETT                 "DERSettings"
#define XSD_TAG_DERSETT_MODESEN         "modesEnabled"
#define XSD_TAG_DERSETT_SETESDEL        "setESDelay"
#define XSD_TAG_DERSETT_SETESHF         "setESHighFreq"
#define XSD_TAG_DERSETT_SETESHV         "setESHighVolt"
#define XSD_TAG_DERSETT_SETESLF         "setESLowFreq"
#define XSD_TAG_DERSETT_SETESLV         "setESLowVolt"
#define XSD_TAG_DERSETT_SETESRT         "setESRampTms"
#define XSD_TAG_DERSETT_SETESRD         "setESRandomDelay"
#define XSD_TAG_DERSETT_SETGRADW        "setGradW"
#define XSD_TAG_DERSETT_SETMAXA         "setMaxA"
#define XSD_TAG_DERSETT_SETMAXAH        "setMaxAh"
#define XSD_TAG_DERSETT_SETMAXCHRGRVA   "setMaxChargeRateVA"
#define XSD_TAG_DERSETT_SETMAXCHRGRW    "setMaxChargeRateW"
#define XSD_TAG_DERSETT_SETMAXDCHRGRVA  "setMaxDischargeRateVA"
#define XSD_TAG_DERSETT_SETMAXDCHRGRW   "setMaxDischargeRateW"
#define XSD_TAG_DERSETT_SETMAXV         "setMaxV"
#define XSD_TAG_DERSETT_SETMAXVA        "setMaxVA"
#define XSD_TAG_DERSETT_SETMAXVAR       "setMaxVar"
#define XSD_TAG_DERSETT_SETMAXVARNEG    "setMaxVarNeg"
#define XSD_TAG_DERSETT_SETMAXW         "setMaxW"
#define XSD_TAG_DERSETT_SETMAXWH        "setMaxWh"
#define XSD_TAG_DERSETT_SETMINPFOE      "setMinPFOverExcited"
#define XSD_TAG_DERSETT_SETMINPFUE      "setMinPFUnderExcited"
#define XSD_TAG_DERSETT_SETMINV         "setMinV"
#define XSD_TAG_DERSETT_SETSOFTGRADW    "setSoftGradW"
#define XSD_TAG_DERSETT_SETVNOM         "setVNom"
#define XSD_TAG_DERSETT_SETVREF         "setVRef"
#define XSD_TAG_DERSETT_SETVREFOFS      "setVRefOfs"
#define XSD_TAG_DERSETT_UPDATEDTM       "updatedTime"


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DER CAPABILITY TAGS

#define XSD_TAG_DERCAP                  "DERCapability"
#define XSD_TAG_DERCAP_MODESUP          "modesSupported"
#define XSD_TAG_DERCAP_RTG_ABCTG        "rtgAbnormalCategory"
#define XSD_TAG_DERCAP_RTG_MAXA         "rtgMaxA"
#define XSD_TAG_DERCAP_RTG_MAXAH        "rtgMaxAh"
#define XSD_TAG_DERCAP_RTG_MAXCHRGRVA   "rtgMaxChargeRateVA"
#define XSD_TAG_DERCAP_RTG_MAXCHRGRW    "rtgMaxChargeRateW"
#define XSD_TAG_DERCAP_RTG_MAXDCHRGRVA  "rtgMaxDischargeRateVA"
#define XSD_TAG_DERCAP_RTG_MAXDCHRGRW   "rtgMaxDischargeRateW"
#define XSD_TAG_DERCAP_RTG_MAXV         "rtgMaxV"
#define XSD_TAG_DERCAP_RTG_MAXVA        "rtgMaxVA"
#define XSD_TAG_DERCAP_RTG_MAXVAR       "rtgMaxVar"
#define XSD_TAG_DERCAP_RTG_MAXVARNEG    "rtgMaxVarNeg"
#define XSD_TAG_DERCAP_RTG_MAXW         "rtgMaxW"
#define XSD_TAG_DERCAP_RTG_MAXWH        "rtgMaxWh"
#define XSD_TAG_DERCAP_RTG_MINPFOE      "rtgMinPFOverExcited"
#define XSD_TAG_DERCAP_RTG_MINPRUE      "rtgMinPFUnderExcited"
#define XSD_TAG_DERCAP_RTG_MINV         "rtgMinV"
#define XSD_TAG_DERCAP_RTG_NORMCAT      "rtgNormalCategory"
#define XSD_TAG_DERCAP_RTG_OEPF         "rtgOverExcitedPF"
#define XSD_TAG_DERCAP_RTG_OEW          "rtgOverExcitedW"
#define XSD_TAG_DERCAP_RTG_REACSUCEP    "rtgReactiveSusceptance"
#define XSD_TAG_DERCAP_RTG_UEPF         "rtgUnderExcitedPF"
#define XSD_TAG_DERCAP_RTG_UEW          "rtgUnderExcitedW"
#define XSD_TAG_DERCAP_RTG_VNOM         "rtgVNom"
#define XSD_TAG_DERCAP_TYPE             "type"

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DER STATUS TAGS

#define XSD_TAG_DERSTAT                        "DERStatus" 
#define XSD_TAG_DERSTAT_ALARMSTAT              "alarmStatus" 
#define XSD_TAG_DERSTAT_GENCONSTAT             "genConnectStatus" 
#define XSD_TAG_DERSTAT_INVERTSTAT             "inverterStatus" 
#define XSD_TAG_DERSTAT_LOCALCTRLMODESTAT      "localControlModeStatus" 
#define XSD_TAG_DERSTAT_MANUFCSTAT             "manufacturerStatus" 
#define XSD_TAG_DERSTAT_OPMODESTAT             "operationalModeStatus" 
#define XSD_TAG_DERSTAT_READTIME               "readingTime" 
#define XSD_TAG_DERSTAT_SOCSTAT                "stateOfChargeStatus" 
#define XSD_TAG_DERSTAT_STORAGEMODESTAT        "storageModeStatus" 
#define XSD_TAG_DERSTAT_STORCONNSTAT           "storConnectStatus"     

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DER AVAILABILITY TAGS

#define XSD_TAG_DERAVAIL                "DERAvailability"
#define XSD_TAG_DERAVAIL_AVAILDUR       "availabilityDuration"
#define XSD_TAG_DERAVAIL_MAXCHRGDUR     "maxChargeDuration"
#define XSD_TAG_DERAVAIL_READTIME       "readingTime"
#define XSD_TAG_DERAVAIL_RESCHRGPERC    "reserveChargePercent"
#define XSD_TAG_DERAVAIL_RESPERC        "reservePercent"
#define XSD_TAG_DERAVAIL_STATVARAVAIL   "statVarAvail"
#define XSD_TAG_DERAVAIL_STATWAAVAIL    "statWAvail"

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// LOG EVENT TAGS

#define XSD_TAG_LOGEVENT                    "LogEvent"
#define XSD_TAG_LOGEVENT_CRTDATETIME        "createdDateTime"
#define XSD_TAG_LOGEVENT_DETAILS            "details"
#define XSD_TAG_LOGEVENT_EXTDATA            "extendedData"
#define XSD_TAG_LOGEVENT_FNCSET             "functionSet"
#define XSD_TAG_LOGEVENT_LOGEVCODE          "logEventCode"
#define XSD_TAG_LOGEVENT_LOGEVID            "logEventID"
#define XSD_TAG_LOGEVENT_LOGEVPEN           "logEventPEN"
#define XSD_TAG_LOGEVENT_PROFILEID          "profileID"

// LOG EVENT CODES 
#define DER_FAULT_OVER_CURRENT              0
#define DER_FAULT_OVER_CURRENT_RTN          1
#define DER_FAULT_OVER_VOLTAGE              2
#define DER_FAULT_OVER_VOLTAGE_RTN          3
#define DER_FAULT_UNDER_VOLTAGE             4
#define DER_FAULT_UNDER_VOLTAGE_RTN         5
#define DER_FAULT_OVER_FREQUENCY            6
#define DER_FAULT_OVER_FREQUENCY_RTN        7
#define DER_FAULT_UNDER_FREQUENCY           8
#define DER_FAULT_UNDER_FREQUENCY_RTN       9
#define DER_FAULT_VOLTAGE_IMBALANCE         10
#define DER_FAULT_VOLTAGE_IMBALANCE_RTN     11
#define DER_FAULT_CURRENT_IMBALANCE         12
#define DER_FAULT_CURRENT_IMBALANCE_RTN     13
#define DER_FAULT_EMERGENCY_LOCAL           14
#define DER_FAULT_EMERGENCY_LOCAL_RTN       15
#define DER_FAULT_EMERGENCY_REMOTE          16
#define DER_FAULT_EMERGENCY_REMOTE_RTN      17
#define DER_FAULT_LOW_POWER_INPUT           18
#define DER_FAULT_LOW_POWER_INPUT_RTN       19
#define DER_FAULT_PHASE_ROTATION            20
#define DER_FAULT_PHASE_ROTATION_RTN        21

#define IEEE_LOGEV_PEN        40732

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MIRROR USAGE POINT TAGS

#define LION_PEN 0xF0A0 //61600 decimal

#define XSD_TAG_MUP                 "MirrorUsagePoint"
#define XSD_TAG_MUP_MRID            "mRID"
#define XSD_TAG_MUP_DESC            "description"
#define XSD_TAG_MUP_ROLEFLAGS       "roleFlags"
#define XSD_TAG_MUP_SERVCATKIND     "serviceCategoryKind"
#define XSD_TAG_MUP_STAT            "status"
#define XSD_TAG_MUP_DEVLFDI         "deviceLFDI"

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MIRROR METER READING TAGS

#define XSD_TAG_MMR         "MirrorMeterReading"         
#define XSD_TAG_MMR_MRID    "mRID"
#define XSD_TAG_MMR_DESC    "description"

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// READING TYPE TAGS

#define XSD_TAG_READTYPE                "ReadingType"
#define XSD_TAG_READTYPE_ACCUMBEH       "accumulationBehaviour"
#define XSD_TAG_READTYPE_COMMODITY      "commodity"
#define XSD_TAG_READTYPE_DATAQUAL       "dataQualifier"
#define XSD_TAG_READTYPE_FLOWDIR        "flowDirection"
#define XSD_TAG_READTYPE_KIND           "kind"
#define XSD_TAG_READTYPE_PHASE          "phase"
#define XSD_TAG_READTYPE_POTMULT        "powerOfTenMultiplier"
#define XSD_TAG_READTYPE_UOM            "uom"

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// READING TAGS

#define XSD_TAG_READING                "Reading"
#define XSD_TAG_READING_VAL            "value"

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

/**
 * @brief Serialize the DERControlResponse struct into an XML payload.
 * 
 * @details This function serializes the DERControlResponse struct into an XML payload. The XML payload is then stored
 * in the txPayload string.
 * 
 * @param sz_txPayload The string to store the XML payload.
 * @param i3e_ptr_derCtrlRsp The DERControlResponse struct to serialize.
 */
ModScop void I3E_XML_SRLZ_DERControlResponse(std::string& sz_txPayload, SE_DERControlResponse_t* i3e_ptr_derCtrlRsp);       

/**
 * @brief Serialize the DERSettings struct into an XML payload.
 * 
 * @details This function serializes the DERSettings struct into an XML payload. The XML payload is then stored
 * in the txPayload string.
 * 
 * @param sz_txPayload The string to store the XML payload.
 * @param i3e_ptr_derSettings The DERSettings struct to serialize.
 */
ModScop void I3E_XML_SRLZ_DERSettings(std::string& sz_txPayload, SE_DERSettings_t* i3e_ptr_derSettings);

/**
 * @brief Serialize the DERCapability struct into an XML payload.
 * 
 * @details This function serializes the DERCapability struct into an XML payload. The XML payload is then stored
 * in the txPayload string.
 * 
 * @param sz_txPayload The string to store the XML payload.
 * @param i3e_ptr_derCapability The DERCapability struct to serialize.
 */
ModScop void I3E_XML_SRLZ_DERCapability(std::string& sz_txPayload, SE_DERCapability_t* i3e_ptr_derCapability);

/**
 * @brief Serialize the DERStatus struct into an XML payload.
 * 
 * @details This function serializes the DERStatus struct into an XML payload. The XML payload is then stored
 * in the txPayload string.
 * 
 * @param sz_txPayload The string to store the XML payload.
 * @param i3e_ptr_derStatus The DERStatus struct to serialize.
 */
ModScop void I3E_XML_SRLZ_DERStatus(std::string& sz_txPayload, SE_DERStatus_t* i3e_ptr_derStatus);

/**
 * @brief Serialize the DERAvailability struct into an XML payload.
 * 
 * @details This function serializes the DERAvailability struct into an XML payload. The XML payload is then stored
 * in the txPayload string.
 * 
 * @param sz_txPayload The string to store the XML payload.
 * @param i3e_ptr_derAvailability The DERAvailability struct to serialize.
 */
ModScop void I3E_XML_SRLZ_DERAvailability(std::string& sz_txPayload, SE_DERAvailability_t* i3e_ptr_derAvailability);

/**
 * @brief Serialize the LogEvent struct into an XML payload.
 * 
 * @details This function serializes the LogEvent struct into an XML payload. The XML payload is then stored
 * in the txPayload string.
 * 
 * @param sz_txPayload The string to store the XML payload.
 * @param i3e_ptr_logEvent The LogEvent struct to serialize.
 */
ModScop void I3E_XML_SRLZ_LogEvent(std::string& sz_txPayload, SE_LogEvent_t* i3e_ptr_logEvent);

/**
 * @brief Serialize the MirrorUsagePoint struct into an XML payload.
 * 
 * @details This function serializes the MirrorUsagePoint struct into an XML payload. The XML payload is then stored
 * in the txPayload string. This function will call the I3E_XML_SRLZ_MirrorMeterReadingInit function to serialize the
 * child MirrorMeterReading struct.
 * 
 * @param sz_txPayload The string to store the XML payload.
 * @param i3e_ptr_mup The MirrorUsagePoint struct to serialize.
 */
ModScop void I3E_XML_SRLZ_MirrorUsagePoint(std::string& sz_txPayload, SE_MirrorUsagePoint_t* i3e_ptr_mup);

/**
 * @brief Serialize the MirrorMeterReading struct into an XML payload. 
 * 
 * @details This function serializes the MirrorMeterReading struct into an XML payload. The XML payload is then stored
 * in the txPayload string. This function is called by the I3E_XML_SRLZ_MirrorUsagePoint function to setup the initial
 * MUP payload.
 * 
 * @param sz_txPayload The string to store the XML payload.
 * @param i3e_ptr_mmr The MirrorMeterReading struct to serialize.
 * @param ptr_pdoc_xmlDoc The Poco XML Document object of the parent MUP structure.
 * @param poco_pel_xmlRoot The Poco XML Element root object of the parent MUP structure.
 */
ModScop void I3E_XML_SRLZ_MirrorMeterReadingInit(std::string& sz_txPayload, SE_MirrorMeterReading_t* i3e_ptr_mmr, 
                                                Poco::AutoPtr<Poco::XML::Document> &ptr_pdoc_xmlDoc, 
                                                Poco::AutoPtr<Poco::XML::Element> &poco_pel_xmlRoot);

/**
 * @brief Serialize the MirrorMeterReading struct into an XML payload.
 * 
 * @details This function serializes the MirrorMeterReading struct into an XML payload. The XML payload is then stored
 * in the txPayload string.
 * 
 * @param sz_txPayload The string to store the XML payload.
 * @param i3e_ptr_mmr The MirrorMeterReading struct to serialize.
 */
ModScop void I3E_XML_SRLZ_MirrorMeterReading(std::string& sz_txPayload, SE_MirrorMeterReading_t* i3e_ptr_mmr);


#endif /* LION_I3E_XMLS_H__ */