/**********************************************************************************************************************
 * @file            ICC_Function.h
 *
 * @brief           The function subsystem of the ICC allows for the submission and processing of request and response
 *                  packets to mimic a "function call". Function "calls" generate a request packet with a function
 *                  defined payload, submit it to the bus and wait for a response packet. Likewise, function handlers
 *                  process requests and generate a response packet with a function defined payload and submit them to
 *                  the bus.
 *
 * @author          Abram Early
 * @date            05/05/2023
 * @copyright       2023 Lion Energy LLC.
 *                  All rights reserved
 *
 *********************************************************************************************************************/
/**
 * @addtogroup      ICC
 *                  Inter-Controller Communication Module
 *
 * @ref             ICC.h
 *
 *********************************************************************************************************************/

#ifndef LION_ICC_FUNCTIONS_H__
#define LION_ICC_FUNCTIONS_H__

#include "COD_Compiler_Defines.h"
#include "lion/rtos/RTOS_Timeout.h"
#include "lion/utils/dlist.h"
#include <stddef.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C"
{
#endif

    /**********************************************************************************************************************
     * @cond                                              GLOBAL SCOPE @endcond
     *********************************************************************************************************************/

#define ICC_Function_ANY_ID -1

    typedef struct ICC_Context_t ICC_Context_t;

    /**
     * @brief A function called to process an incoming function call request
     * @param u32Id The function ID called for this request
     * @param ptrData Pointer to the function-defined request payload
     * @param u32DataLength Number of bytes in the request payload
     * @param ptrResponse Pointer to a buffer the response payload should be written to
     * @param u32ResponseLength Initialized to the available response payload buffer size. Should be set to the actual
     * payload size before returning.
     * @param ptrUserData The user-defined pointer passed to @ref ICC_Function_InitHandler
     * @returns A return code that will be passed verbatim in the response packet.
     */
    typedef int (*ICC_Function_HandlerCallback_t)(uint32_t u32Id, const void *ptrData, uint32_t u32DataLength,
                                                  void *ptrResponse, uint32_t *u32ResponseLength, void *ptrUserData);

    /**
     * @brief A function called to process an incoming function call response
     * @param u32FunctionId The function ID called for this response
     * @param intResult The return code returned by the function handler.
     * @param ptrResponse Pointer to the function-defined response payload
     * @param u32ResponseLength Number of bytes in the response payload
     * @param ptrUserData The user-defined pointer passed to @ref ICC_Function_CallAsync
     */
    typedef void (*ICC_Function_Callback_t)(uint32_t u32FunctionId, int intResult, const void *ptrResponse,
                                            uint32_t u32ResponseLength, void *ptrUserData);
    typedef struct
    {
        sys_dlist_t                    node;
        uint32_t                       u32FunctionId;
        uint32_t                       u32FunctionIdMask;
        ICC_Function_HandlerCallback_t fnCallback;
        void                          *ptrUserData;
    } ICC_Function_Handler_t;

    struct ICC_FunctionCall_t
    {
        ICC_Context_t          *ptr_ICCContext;
        sys_dnode_t             strNode;
        uint32_t                u32FunctionId;
        uint16_t                u16_RequestCRC;
        ICC_Function_Callback_t fnCallback;
        void                   *ptrUserData;
    };
    typedef struct ICC_FunctionCall_t ICC_FunctionCall_t;

    /**
     * @brief Initialize an ICC_Function_Handler_t
     * @param handler A reference an ICC_Function_Handler_t structure allocated by the user
     * @param functionId The functionId this handler will be called for
     * @param functionIdMask A mask that is first applied to the functionId before checking for a match
     * @param callback The function that will be called when the event occurs
     * @param userData An arbitrary pointer that will be passed to the callback function
     */
    GlobScop void ICC_Function_InitHandler(ICC_Function_Handler_t *handler, uint32_t functionId,
                                           uint32_t functionIdMask, ICC_Function_HandlerCallback_t callback,
                                           void *userData);

    /**
     * @brief Registers a handler with the function subsystem
     * @warning handler must remain valid until ICC_Function_UnregisterHandler is called
     * @param handler An initialized ICC_Function_Handler_t structure
     * @returns 0 on Success
     * @returns -EALREADY if handler is already registered, or the ID + mask combo conflicts with another handler
     */
    GlobScop int ICC_Function_RegisterHandler(ICC_Context_t *context, ICC_Function_Handler_t *handler);

    /**
     * @brief Unregisters a handler from the function subsystem
     * @param handler An initialized ICC_Function_Handler_t structure
     * @returns 0 on Success
     * @returns -EALREADY if handler is already unregistered
     * @warning Currently Unsafe to Call: Mutex doesn't properly lock the list during packet processing.
     */
    GlobScop int ICC_Function_UnregisterHandler(ICC_Context_t *context, ICC_Function_Handler_t *handler);

    /**
     * @brief Dispatches a function to the first encountered function handler. (Local First)
     * @param ptrContext The ICC Context
     * @param u32FunctionId The function ID
     * @param ptrData A pointer to the data to be passed to the function handler
     * @param u32DataLength The length in bytes of data
     * @param ptrResponse Pointer to copy response payload to. May be NULL
     * @param u32ResponseLength On call, the size of the buffer at @p ptrResponse . On return, the length of the
     * response payload. May be NULL if ptrResponse is NULL
     * @param strTimeout Duration to wait for a response
     * @returns 0 on Success
     * @returns -EINVAL If an illegal event ID is used
     * @returns -EAGAIN If timeout expires waiting for queue to unblock, or queue is full when called from ISR
     * @returns -ENOMEM If u32DataLength is too large or unable to allocate packet
     */
    GlobScop int ICC_Function_Call(ICC_Context_t *ptrContext, uint32_t u32FunctionId, const void *ptrData,
                                   uint32_t u32DataLength, void *ptrResponse, uint32_t *u32ResponseLength,
                                   RTOS_Timeout_t strTimeout);

    /**
     * @brief Dispatches a function to the first encountered function handler asynchronously. (Local First)
     * @param ptrContext The ICC Context
     * @param ptrFunctionCall An initialized function call context
     * @param u32FunctionId The function ID
     * @param ptrData A pointer to the data to be passed to the function handler
     * @param u32DataLength The length in bytes of data
     * @param fnCallback The callback implementing @ref ICC_Function_Callback_t to call when a response is received
     * @param ptrUserData An arbitrary user-defined pointer to pass along to the callback
     * @returns 0 on Success
     * @returns -EINVAL If an illegal event ID is used
     * @returns -ENOMEM If u32DataLength is too large or unable to allocate packet
     */
    GlobScop int ICC_Function_CallAsync(ICC_Context_t *ptrContext, ICC_FunctionCall_t *ptrFunctionCall,
                                        uint32_t u32FunctionId, const void *ptrData, uint32_t u32DataLength,
                                        ICC_Function_Callback_t fnCallback, void *ptrUserData);

    /**
     * @brief Initializes a @ref ICC_FunctionCall_t
     * @note I'm genuinely not sure why I added this instead of just making @ref ICC_Function_CallAsync initialize it.
     * @param ptrFunctionCall An uninitialized function call context
     */
    GlobScop void ICC_Function_InitCall(ICC_FunctionCall_t *ptrFunctionCall);

    /**
     * @brief @b Safely destroys the provided function call context.
     * @param ptrFunctionCall An initialized function call context
     */
    GlobScop void ICC_Function_DestroyCall(ICC_FunctionCall_t *ptrFunctionCall);

#ifdef __cplusplus
}
#endif

#endif /* LION_ICC_FUNCTIONS_H__ */