#include "lion/icc/ICC.h"
#include "ICC_Internal.h"
#include "lion/log/log.h"
#include <assert.h>
#include <errno.h>
#ifndef __ZEPHYR__
    #include <stdlib.h>
#endif

LOG_MODULE_REGISTER(icc, CONFIG_ICC_LOG_LEVEL);

#ifndef __ZEPHYR__
ICC_Context_t *ICC_Allocate() { return malloc(sizeof(ICC_Context_t)); }
void           ICC_Free(ICC_Context_t *ptr_Context) { free(ptr_Context); }
#endif

void ICC_Init(ICC_Context_t *context)
{
    context->u32PacketUsage = 0U;
    RTOS_SemInit(&context->strXmitSemaphore, CONFIG_ICC_u32XMIT_BUFFER_COUNT, CONFIG_ICC_u32XMIT_BUFFER_COUNT);
    RTOS_SemInit(&context->strRecvSemaphore, CONFIG_ICC_u32RECV_BUFFER_COUNT, CONFIG_ICC_u32RECV_BUFFER_COUNT);
    RTOS_MutexInit(&context->strHandlerLock);
    RTOS_MutexInit(&context->mutex);
    sys_dlist_init(&context->events);
    sys_dlist_init(&context->functions);
    RTOS_MutexInit(&context->strTransportLock);
    sys_dlist_init(&context->transports);
    sys_dlist_init(&context->strFunctionCalls);
    sys_dlist_init(&context->packetQueue);
    RTOS_MutexInit(&context->mtx_QueueLock);
    RTOS_CondVarInit(&context->cnd_QueueVar);
    context->u8_Transaction = 0;
}

GlobScop void ICC_Process(ICC_Context_t *ptr_Context)
{
    while (ICC_ProcessPacket(ptr_Context, RTOS_NO_WAIT))
    {
    }
    ICC_ProcessIO(ptr_Context);
}

GlobScop void ICC_ProcessIO(ICC_Context_t *ptr_Context)
{
    StackScop ICC_Transport_t *ptr_Transport;
    StackScop ICC_Transport_t *ptr_Next;
    RTOS_MutexLock(&ptr_Context->strTransportLock, RTOS_FOREVER);
    /* cppcheck-suppress misra-c2012-18.4 */
    SYS_DLIST_FOR_EACH_CONTAINER_SAFE(&ptr_Context->transports, ptr_Transport, ptr_Next, node)
    {
        ptr_Transport->process(ptr_Context, ptr_Transport->userData);
    }
    RTOS_MutexUnlock(&ptr_Context->strTransportLock);
}

GlobScop bool ICC_ProcessPacket(ICC_Context_t *ptr_Context, RTOS_Timeout_t tim_Timeout)
{
    StackScop RTOS_Timepoint_t tim_Timepoint = RTOS_TimepointCalc(tim_Timeout);
    StackScop RTOS_Timeout_t   tim_ActiveTimeout;
    StackScop ICC_Packet_t    *ptr_Packet           = NULL;
    StackScop bool             bol_DidProcessPacket = false;
    do
    {
        tim_ActiveTimeout = RTOS_TimepointToTimeout(tim_Timepoint);
        RTOS_MutexLock(&ptr_Context->mtx_QueueLock, RTOS_FOREVER);
        if (sys_dlist_is_empty(&ptr_Context->packetQueue))
        {
            if (!RTOS_TIMEOUT_EQ(tim_ActiveTimeout, RTOS_NO_WAIT))
            {
                LOG_INF("Waiting For Packets...");
            }
            RTOS_CondVarWait(&ptr_Context->cnd_QueueVar, &ptr_Context->mtx_QueueLock, tim_ActiveTimeout);
        }
        StackScop sys_dlist_t *ptr_Node = sys_dlist_get(&ptr_Context->packetQueue);
        RTOS_MutexUnlock(&ptr_Context->mtx_QueueLock);
        ptr_Packet = SYS_DLIST_CONTAINER(ptr_Node, ptr_Packet, node);
    } /* So long as the active timeout isn't RTOS_NO_WAIT, keep waiting. */
    while (ptr_Packet == NULL && !RTOS_TIMEOUT_EQ(tim_ActiveTimeout, RTOS_NO_WAIT));

    if (ptr_Packet != NULL)
    {
        LOG_INF("Got one! (%p)", ptr_Packet);
        /* The following lines utilize the short-circuit behavior of the || operator:
         * If bolIsConsumed is true, then the function call on the right will not be called. */
        bool bolIsConsumed = false;
        bolIsConsumed      = bolIsConsumed || icc_Event_ProcessPacket(ptr_Context, ptr_Packet);
        bolIsConsumed      = bolIsConsumed || icc_Function_ProcessFunctionPacket(ptr_Context, ptr_Packet);
        bolIsConsumed      = bolIsConsumed || icc_Function_ProcessReturnPacket(ptr_Context, ptr_Packet);

        if (!bolIsConsumed)
        {
            icc_transport_ForwardTXPacket(ptr_Context, NULL, ptr_Packet);
        }
        else
        {
            icc_ReleasePacket(ptr_Context, ptr_Packet);
        }

        bol_DidProcessPacket = true;
    }
    return bol_DidProcessPacket;
}

void icc_EnqueuePacket(ICC_Context_t *iccContext, ICC_Packet_t *packet)
{
    RTOS_MutexLock(&iccContext->mtx_QueueLock, RTOS_FOREVER);
    sys_dlist_append(&iccContext->packetQueue, &packet->node);
    RTOS_CondVarSignal(&iccContext->cnd_QueueVar);
    RTOS_MutexUnlock(&iccContext->mtx_QueueLock);
}

ICC_Packet_t *icc_AllocateTXPacket(ICC_Context_t *iccContext, uint16_t length)
{
    ICC_Packet_t *buffer = NULL;
    if (length > icc_u32XMIT_BUFFER_SIZE)
    {
        LOG_ERR("%p: Requested packet length too large! (%u > %u)", iccContext, length, icc_u32XMIT_BUFFER_SIZE);
    }
    else if (RTOS_SemTake(&iccContext->strXmitSemaphore, RTOS_NO_WAIT) >= 0)
    {
        RTOS_MutexLock(&iccContext->mutex, RTOS_FOREVER);
        for (uint8_t u8Index = 0; u8Index < CONFIG_ICC_u32XMIT_BUFFER_COUNT; u8Index++)
        {
            if (!(icc_u32XMIT_BIT(u8Index) & iccContext->u32PacketUsage))
            {
                iccContext->u32PacketUsage |= icc_u32BIT(icc_u32XMIT_INDEX(u8Index));
                buffer          = &iccContext->arrPackets[icc_u32XMIT_INDEX(u8Index)];
                buffer->u8Index = icc_u32XMIT_INDEX(u8Index);
                sys_dnode_init(&buffer->node);
                buffer->ptrSource = NULL;
                buffer->ptrTarget = NULL;
                buffer->u16Length = icc_u32XMIT_BUFFER_SIZE;
                buffer->arrBuffer = iccContext->arrXmitBuffers[u8Index];
                LOG_INF("%p: Allocated TX packet %p with %u bytes", iccContext, buffer, length);
                break;
            }
        }
        RTOS_MutexUnlock(&iccContext->mutex);
        assert(buffer != NULL);
    }
    else
    {
        /* This isn't really an error, and causes more noise than helpful feedback. If errors are needed, put in caller.
         */
        LOG_DBG("%p: Unable to allocate TX packet", iccContext);
    }
    return buffer;
}

ICC_Packet_t *icc_AllocateRXPacket(ICC_Context_t *iccContext, uint16_t length)
{
    ICC_Packet_t *buffer = NULL;
    if (length > icc_u32RECV_BUFFER_SIZE)
    {
        LOG_ERR("%p: Requested packet length too large! (%u > %u)", iccContext, length, icc_u32RECV_BUFFER_SIZE);
    }
    else if (RTOS_SemTake(&iccContext->strRecvSemaphore, RTOS_NO_WAIT) >= 0)
    {
        RTOS_MutexLock(&iccContext->mutex, RTOS_FOREVER);
        for (uint8_t u8Index = 0; u8Index < CONFIG_ICC_u32RECV_BUFFER_COUNT; u8Index++)
        {
            if (!(icc_u32RECV_BIT(u8Index) & iccContext->u32PacketUsage))
            {
                iccContext->u32PacketUsage |= icc_u32BIT(icc_u32RECV_INDEX(u8Index));
                buffer          = &iccContext->arrPackets[icc_u32RECV_INDEX(u8Index)];
                buffer->u8Index = icc_u32RECV_INDEX(u8Index);
                sys_dnode_init(&buffer->node);
                buffer->ptrSource = NULL;
                buffer->ptrTarget = NULL;
                buffer->u16Length = icc_u32RECV_BUFFER_SIZE;
                buffer->arrBuffer = iccContext->arrRecvBuffers[u8Index];
                LOG_INF("%p: Allocated RX packet %p with %u bytes", iccContext, buffer, length);
                break;
            }
        }
        RTOS_MutexUnlock(&iccContext->mutex);
        assert(buffer != NULL);
    }
    else
    {
        /* This isn't really an error, and causes more noise than helpful feedback. If errors are needed, put in caller.
         */
        LOG_DBG("%p: Unable to allocate RX packet", iccContext);
    }
    return buffer;
}

void icc_ReleasePacket(ICC_Context_t *iccContext, ICC_Packet_t *packet)
{
    LOG_INF("%p: Releasing packet %p", iccContext, packet);
    RTOS_MutexLock(&iccContext->mutex, RTOS_FOREVER);
    if ((icc_u32BIT(packet->u8Index) & icc_u32RECV_MASK) != 0U)
    {
        iccContext->u32PacketUsage &= ~icc_u32BIT(packet->u8Index);
        RTOS_SemGive(&iccContext->strRecvSemaphore);
    }
    else if ((icc_u32BIT(packet->u8Index) & icc_u32XMIT_MASK) != 0U)
    {
        iccContext->u32PacketUsage &= ~icc_u32BIT(packet->u8Index);
        RTOS_SemGive(&iccContext->strXmitSemaphore);
    }
    else
    {
    }
    RTOS_MutexUnlock(&iccContext->mutex);
}