#include "ICC_Internal.h"
#include "lion/log/log.h"
#include <errno.h>

LOG_MODULE_DECLARE(icc, CONFIG_ICC_LOG_LEVEL);

/**********************************************************************************************************************
 * @cond                                              MODULE SCOPE @endcond
 *********************************************************************************************************************/

ModScop void icc_transport_ForwardTXPacket(ICC_Context_t *icc_Context, ICC_Transport_t *ptr_Transport,
                                           ICC_Packet_t *ptr_Packet)
{
    assert(icc_Context != NULL); /* icc_Context MUST NOT be NULL */
    assert(ptr_Packet != NULL);  /* ptr_Packet MUST NOT be NULL */
    assert((ptr_Transport == NULL) ||
           (ptr_Transport->context ==
            icc_Context)); /* ptr_Transport MUST be NULL or it MUST be added to the provided ICC_Context_t */
    assert(!sys_dnode_is_linked(&ptr_Packet->node)); /* Packet MUST NOT be in a list */

    /* Transport list must be locked while accessing next transport. */
    RTOS_MutexLock(&icc_Context->strTransportLock, RTOS_FOREVER);
    StackScop sys_dnode_t     *ptr_CurrentNode = &ptr_Transport->node;
    StackScop ICC_Transport_t *ptr_Destination = NULL;
    SYS_DLIST_ITERATE_FROM_NODE(&icc_Context->transports, ptr_CurrentNode)
    {
        StackScop ICC_Transport_t *ptr_CurrentTransport = CONTAINER_OF(ptr_CurrentNode, ICC_Transport_t, node);

        if (ptr_Packet->ptrTarget != NULL)
        {
            if (ptr_Packet->ptrTarget != ptr_CurrentTransport)
            {
                LOG_INF("%p: Packet %p Skipped Transport %p: %s", icc_Context, ptr_Packet, ptr_CurrentTransport,
                        "Not Target");
                continue;
            }
        }
        else if (ptr_Packet->ptrSource == ptr_CurrentTransport)
        {
            LOG_INF("%p: Packet %p Skipped Transport %p: %s", icc_Context, ptr_Packet, ptr_CurrentTransport,
                    "Is Source");
            continue;
        }

        LOG_INF("%p: Packet %p Enqueued to Transport %p", icc_Context, ptr_Packet, ptr_CurrentTransport);
        sys_dlist_append(&(ptr_CurrentTransport->lst_TXPackets), &(ptr_Packet->node));
        ptr_Destination = ptr_CurrentTransport;
        break;
    }
    RTOS_MutexUnlock(&icc_Context->strTransportLock);

    if (ptr_Destination == NULL)
    {
        icc_ReleasePacket(icc_Context, ptr_Packet);
    }
}

/**********************************************************************************************************************
 * @cond                                              GLOBAL SCOPE @endcond
 *********************************************************************************************************************/

GlobScop ICC_Packet_t *ICC_Transport_PopTXPacket(ICC_Transport_t *ptr_Transport)
{
    assert(ptr_Transport != NULL);          /* ptr_Transport MUST NOT be NULL */
    assert(ptr_Transport->context != NULL); /* MUST be added to an ICC_Context_t */

    StackScop ICC_Context_t *icc_Context = ptr_Transport->context;
    RTOS_MutexLock(&icc_Context->strTransportLock, RTOS_FOREVER);
    StackScop sys_dnode_t  *ptr_Node   = sys_dlist_get(&(ptr_Transport->lst_TXPackets));
    StackScop ICC_Packet_t *ptr_Packet = SYS_DLIST_CONTAINER(ptr_Node, ptr_Packet, node);
    RTOS_MutexUnlock(&icc_Context->strTransportLock);
    return ptr_Packet;
}

GlobScop void ICC_Transport_ReleaseTXPacket(ICC_Transport_t *ptr_Transport, ICC_Packet_t *ptr_Packet)
{
    assert(ptr_Transport != NULL);          /* ptr_Transport MUST NOT be NULL */
    assert(ptr_Transport->context != NULL); /* MUST be added to an ICC_Context_t */

    StackScop ICC_Context_t *icc_Context = ptr_Transport->context;
    icc_transport_ForwardTXPacket(icc_Context, ptr_Transport, ptr_Packet);
}

GlobScop void ICC_Transport_Init(ICC_Transport_t *transport, ICC_Transport_Process_t process,
                                 ICC_Transport_Reset_t reset, void *userData)
{
    sys_dnode_init(&transport->node);
    transport->context  = NULL;
    transport->process  = process;
    transport->userData = userData;
    transport->reset    = reset;
    sys_dlist_init(&transport->lst_TXPackets);
}

GlobScop int ICC_AddTransport(ICC_Context_t *context, ICC_Transport_t *transport)
{
    int errcode = 0;
    if (sys_dnode_is_linked(&transport->node))
    {
        errcode = -EALREADY;
    }
    else
    {
        RTOS_MutexLock(&context->strTransportLock, RTOS_FOREVER);
        transport->context = context;
        sys_dlist_append(&context->transports, &transport->node);
        RTOS_MutexUnlock(&context->strTransportLock);
    }
    return errcode;
}

GlobScop int ICC_RemoveTransport(ICC_Context_t *context, ICC_Transport_t *transport)
{
    LOG_INF("%p: Reset Transport %p", context, transport);
    RTOS_MutexLock(&context->strTransportLock, RTOS_FOREVER);
    StackScop ICC_Packet_t *ptr_Packet;
    while ((ptr_Packet = ICC_Transport_PopTXPacket(transport)) != NULL)
    {
        ICC_Transport_ReleaseTXPacket(transport, ptr_Packet);
    }
    if (transport->reset != NULL)
    {
        transport->reset(context, transport->userData);
    }
    sys_dlist_remove(&transport->node);
    RTOS_MutexUnlock(&context->strTransportLock);
    return 0;
}