#include <lion/VRS.h>
#include <lion/log/log.h>
#include <lion/rtos/RTOS_Mutex.h>
#include <lion/rtos/RTOS_Sleep.h>
#include <lion/ial/ems/IAL_Data.h>
#include <lion/ial/IAL.h>

#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include "app_version.h"

LOG_MODULE_REGISTER(VRS, LOG_LEVEL_INF);

/**********************************************************************************************************************
 * @cond                                              FILE SCOPE                                               @endcond
 **********************************************************************************************************************/

#define vrs_siz_MAX_KEY_SIZE             64
#define vrs_siz_MAX_VERSION_SIZE         32
#define vrs_siz_MAX_VERSIONS             16
#define vrs_sz_APPLICATION_KEY           PROJECT_NAME
#define vrs_u64_BROADCAST_INTERVAL       5000
#define vrs_u64_IDLE_INTERVAL            1000


typedef struct
{
    VRS_VersionPullFunction_t ptr_VersionPullFunction;
} vrs_Context_t;

typedef enum 
{
    VRS_IDLE,
    VRS_BROADCAST
} vrs_State_t;

/** 
 * @brief Context for the VRS module
 * 
 * @note This is a singleton
*/
FileScop vrs_Context_t vrs_str_Context;

/**
 * @brief Boolean indicating whether vrs should be broadcasting
 */
FileScop volatile bool bol_BradcastRequest;

/**
 * @brief Boolean indicating whether the VRS thread should keep running
 */
FileScop bool vrs_bol_KeepRunning;

/**
 * @brief Thread for VRS
 * 
 * @details This thread will pull the version information and broadcast it to the ICC on startup and request
 */
FileScop pthread_t vrs_str_Thread;

/**
 * @brief Mutex for the VRS thread to keep running
 */
FileScop RTOS_Mutex_t vrs_str_ThreadMutex;

/**
 * @brief Mutex for the broadcast request boolean
 */
FileScop RTOS_Mutex_t vrs_str_BroadcastMutex;

/**
 * @brief Event handler for the version request
 * 
 * @details This event handler will be registered with the ICC to handle version requests
 */
FileScop ICC_Event_Handler_t vrs_str_VersionRequestHandler;

/**
 * @brief Pulls the cmake defined version information
 * 
 * @details If VRS is initialized with a NULL version pull function, this function will be used to pull the version.
 *          Should be used as a template for custom version pull functions.
 * 
 * @param arr_sz_Keys Array of keys to store the version keys
 * @param siz_MaxKey Maximum size of the key in array
 * @param arr_sz_Versions Array of versions to store the version values
 * @param siz_MaxVersion Maximum size of the version in array
 * @param siz_MaxVersions Maximum number of versions to pull
 * @param ptr_siz_Versions Pointer to store the number of versions pulled
 * 
 * @returns EXIT_SUCCESS or 0 on success, standard positive error code on failure.
 */
FileScop int vrs_DefaultVersionPull(char **arr_sz_Keys, size_t siz_MaxKey, char **arr_sz_Versions, 
    size_t siz_MaxVersion, size_t siz_MaxVersions, size_t* ptr_siz_Versions);

/**
 * @brief Getter for the broadcast request boolean
 */
FileScop bool vrs_GetBroadcastRequest(void);

/**
 * @brief Setter for the broadcast request boolean
 */
FileScop void vrs_SetBroadcastRequest(bool bol_Request);

/**
 * @brief Pulls versions using context function pointer, and broadcasts them to the ICC
 * 
 * @returns EXIT_SUCCESS or 0 on success, standard positive error code on failure.
 */
FileScop int vrs_PullVersionAndBroadcast(void);

/**
 * @brief Version request callback
 * 
 * @details This function will be called when a version request is received from the ICC. Sets the broadcast request
 *          boolean to true.
 * 
 * @param int_Id ICC ID
 * @param vptr_Data Data from ICC
 * @param u32_DataLength Length of data
 * @param vptr_Response Response to ICC
 * 
 * @returns EXIT_SUCCESS or 0 on success, standard positive error code on failure.
 */
FileScop void vrs_VersionRequestCB(int int_Id, const void *vptr_Data, uint32_t u32_DataLength, void *vptr_Response);

/**
 * @brief Thread entry function for VRS
 * 
 * @param arg Unused
 */
FileScop void * vrs_ThreadEntry(void *arg);

/**********************************************************************************************************************
 * @cond                                              GLOBAL SCOPE                                             @endcond
 **********************************************************************************************************************/

GlobScop int VRS_Init(VRS_VersionPullFunction_t ptr_VersionPullFunction)
{
    if (ptr_VersionPullFunction == NULL)
    {
        vrs_str_Context.ptr_VersionPullFunction = vrs_DefaultVersionPull;
    }
    else 
    {
        vrs_str_Context.ptr_VersionPullFunction = ptr_VersionPullFunction;
    }

    if (IAL_RegisterEventCallback(IAL_VERSION_INFO_ID, &vrs_str_VersionRequestHandler, vrs_VersionRequestCB) != EXIT_SUCCESS)
    {
        return -EXIT_FAILURE;
    }

    bol_BradcastRequest = true;

    RTOS_MutexInit(&vrs_str_ThreadMutex);
    RTOS_MutexInit(&vrs_str_BroadcastMutex);
    return EXIT_SUCCESS;
}

GlobScop int VRS_Start(void)
{
    StackScop int int_Rc;
    RTOS_MutexLock(&vrs_str_ThreadMutex, RTOS_FOREVER);
    if (vrs_bol_KeepRunning) {
        int_Rc = -EALREADY;
    } else {
        vrs_bol_KeepRunning = true;
        int_Rc = pthread_create(&vrs_str_Thread, NULL, vrs_ThreadEntry, NULL);
    }
    RTOS_MutexUnlock(&vrs_str_ThreadMutex);

    return int_Rc;
}

GlobScop int VRS_Stop(void)
{
    StackScop int int_Rc;

    RTOS_MutexLock(&vrs_str_ThreadMutex, RTOS_FOREVER);
    if (vrs_bol_KeepRunning) {
        vrs_bol_KeepRunning = false;
        int_Rc = pthread_join(vrs_str_Thread, NULL);
    } else {
        int_Rc = -EALREADY;
    }
    RTOS_MutexUnlock(&vrs_str_ThreadMutex);

    return int_Rc;
}

/**********************************************************************************************************************
 * @cond                                              FILE SCOPE                                               @endcond
 **********************************************************************************************************************/

FileScop int vrs_DefaultVersionPull(char **arr_sz_Keys, size_t siz_MaxKey, char **arr_sz_Versions, 
    size_t siz_MaxVersion, size_t siz_MaxVersions, size_t* ptr_siz_Versions)
{
    // Validate parameters
    if (arr_sz_Keys == NULL || siz_MaxKey == 0 || arr_sz_Versions == NULL || siz_MaxVersion == 0 
        || siz_MaxVersions == 0 || ptr_siz_Versions == NULL)
    {
        LOG_ERR("Invalid parameters");
        return -EXIT_FAILURE;
    }

    // Pull the application version
    const char * sz_ApplicationKey = vrs_sz_APPLICATION_KEY;
    size_t siz_ApplicationKey = strnlen(sz_ApplicationKey, vrs_siz_MAX_KEY_SIZE);
    if (siz_ApplicationKey >= siz_MaxKey)
    {
        return -EXIT_FAILURE;
    }

    const char * sz_ApplicationVersion = APP_VERSION_STRING;
    size_t siz_ApplicationVersion = strnlen(sz_ApplicationVersion, vrs_siz_MAX_VERSION_SIZE);
    if (siz_ApplicationVersion >= siz_MaxVersion)
    {
        return -EXIT_FAILURE;
    }

    strncpy(arr_sz_Keys[0], sz_ApplicationKey, siz_ApplicationKey + 1);
    arr_sz_Keys[0][siz_ApplicationKey] = '\0'; // Ensure null termination

    strncpy(arr_sz_Versions[0], sz_ApplicationVersion, siz_ApplicationVersion + 1);
    arr_sz_Versions[0][siz_ApplicationVersion] = '\0'; // Ensure null termination

    *ptr_siz_Versions = 1;
    
    return EXIT_SUCCESS;
}

FileScop void vrs_VersionRequestHandler(int int_Id, const void *vptr_Data, uint32_t u32_DataLength, 
                                                void *vptr_Response)
{
    vrs_SetBroadcastRequest(true);
}

FileScop bool vrs_GetBroadcastRequest(void)
{
    bool bol_Request;
    RTOS_MutexLock(&vrs_str_BroadcastMutex, RTOS_FOREVER);
    bol_Request = bol_BradcastRequest;
    RTOS_MutexUnlock(&vrs_str_BroadcastMutex);
    return bol_Request;
}

FileScop void vrs_SetBroadcastRequest(bool bol_Request)
{
    RTOS_MutexLock(&vrs_str_BroadcastMutex, RTOS_FOREVER);
    bol_BradcastRequest = bol_Request;
    RTOS_MutexUnlock(&vrs_str_BroadcastMutex);
    return;
}

FileScop int vrs_PullVersionAndBroadcast(void)
{
    // Allocate space for the version information
    char arr_sz_Keys[vrs_siz_MAX_VERSIONS][vrs_siz_MAX_KEY_SIZE];
    char arr_sz_Versions[vrs_siz_MAX_VERSIONS][vrs_siz_MAX_VERSION_SIZE];

    // Convert 2d array to array of pointers
    char *arr_ptr_sz_Keys[vrs_siz_MAX_VERSIONS];
    char *arr_ptr_sz_Versions[vrs_siz_MAX_VERSIONS];
    for (size_t i = 0; i < vrs_siz_MAX_VERSIONS; i++)
    {
        arr_ptr_sz_Keys[i] = arr_sz_Keys[i];
        arr_ptr_sz_Versions[i] = arr_sz_Versions[i];
    }

    // Pull the version information
    size_t siz_Versions;
    int int_RC = vrs_str_Context.ptr_VersionPullFunction(arr_ptr_sz_Keys, vrs_siz_MAX_KEY_SIZE, arr_ptr_sz_Versions, 
        vrs_siz_MAX_VERSION_SIZE, vrs_siz_MAX_VERSIONS, &siz_Versions);
    if (int_RC != EXIT_SUCCESS)
    {
        LOG_ERR("Failed to pull version");
        return int_RC;
    }

    // Broadcast the version information
    return IAL_Data_PublishVersionInfo(IAL_TARGET_COMP, siz_Versions, (const char **)arr_ptr_sz_Keys, 
        (const char **)arr_ptr_sz_Versions);
}

FileScop void vrs_VersionRequestCB(int int_Id, const void *vptr_Data, uint32_t u32_DataLength, void *vptr_Response)
{
    vrs_SetBroadcastRequest(true);
}

FileScop void* vrs_ThreadEntry(void *arg)
{
    LOG_DBG("VRS Thread Running");
    FuncScop vrs_State_t enm_State = VRS_IDLE;

    while (vrs_bol_KeepRunning)
    {
        switch(enm_State)
        {
            case VRS_IDLE:
            {
                if (vrs_GetBroadcastRequest())
                {
                    enm_State = VRS_BROADCAST;
                }
                RTOS_Sleep(vrs_u64_IDLE_INTERVAL);
                break;

            }
            case VRS_BROADCAST:
            {

                // While the broadcast request is set, pull the version and broadcast
                if (vrs_GetBroadcastRequest())
                {
                    if (vrs_PullVersionAndBroadcast() == EXIT_SUCCESS)
                    {
                        vrs_SetBroadcastRequest(false);
                    }
                    else
                    {
                    RTOS_Sleep(vrs_u64_BROADCAST_INTERVAL);
                    }
                }
                else
                {
                    // If the broadcast request is cleared, go back to idle
                    enm_State = VRS_IDLE;
                }
                break;
            }   
        }
    }
    return NULL;
}
