/**********************************************************************************************************************
 * @file            CST.h
 *
 * @brief           The Compliance Standard Translation Service (CST) will act as a translation layer between the 
 *                  IEEE 2030.5 Application, and the inverter. It will support reading and writing of all IEEE 2030.5 
 *                  attributes as defined at: https://lionenergy.atlassian.net/wiki/spaces/ES/pages/825819188
 *
 * @authors         Vlad Bugayev
 *                  Jake Jones
 * @date            10/23/2024
 * @copyright       2024 Lion Energy LLC.
 *                  All rights reserved
 *
 *********************************************************************************************************************/
/**
 * @addtogroup      CST
 *                  Compliance Standard Translation Service
 *
 * @ref             CST.h
 *
 *********************************************************************************************************************/

#ifndef LION_CST_H_
#define LION_CST_H_

#include "COD_Compiler_Defines.h"
#include "lion/sch/I3E_Schema.hpp"
#include "lion/icc/ICC.h"
#include "pthread.h"
#include <unordered_set>

#ifdef __cplusplus
extern "C"
{
#endif

/**********************************************************************************************************************
 * @cond                                              GLOBAL SCOPE                                             @endcond
 *********************************************************************************************************************/
typedef enum 
{ 
    CST_LOG_EVENT_NOT_OCCURRED = 0,
    CST_LOG_EVENT_OCCURRED = 1, 
    CST_MULTIPLE_LOG_EVENTS_OCCURRED = 2,
    CST_LOG_EVENT_ERROR = 3
} CST_LogEventOccurred_t;

typedef enum {
    MODEL_UNAVAIL,
    MODEL_AVAIL,
    MODEL_ACTDER,
    MODEL_CONF,
    MODEL_UNKNOWN
} CST_UtilModeStatus_t;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Initializes the Compliance Standard Translation Service (CST) module.
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
*/
GlobScop int CST_Init(void);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Set the utility command mode status in the CST module.
 * @param   enm_ModelStatus 
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
*/
GlobScop void CST_UtilCMDStatusSetter(CST_UtilModeStatus_t enm_ModelStatus);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Get the utility command mode status in the CST module.
 * @param   enm_ModelStatus 
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
*/
GlobScop CST_UtilModeStatus_t CST_UtilCMDStatusGetter(void);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Periodic check to see if the CST can send a util_cmd_en-0 to the EMS (relinquish control).
 * @param   u8_loopCnt A counter variable to control periodic behavior of the function. 
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
*/
GlobScop int CST_UtilCMD(uint8_t u8_loopCnt);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Function that will clear the locked registers, reset values to default and set util_cmd_en to 0 
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
*/
GlobScop int CST_ModelConflict(void);


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Populates the referenced DERCapability structure with the current capabilities of the inverter.
 * @param   ptr_capability  A pointer to the DERCapability structure to be populated.
 * 
 */
GlobScop int CST_GetCapability(SE_DERCapability_t *ptr_capability);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Populates the referenced DERSettings structure with the current settings of the inverter.
 * @param   ptr_settings    A pointer to the DERSettings structure to be populated.
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
 */
GlobScop int CST_GetSettings(SE_DERSettings_t *ptr_settings);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Populates the referenced DERStatus structure with the current status of the inverter.
 * @param   ptr_status  A pointer to the DERStatus structure to be populated.
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
 */
GlobScop int CST_GetStatus(SE_DERStatus_t *ptr_status);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Populates the referenced vector of Mirror Meter Readings with the current readings of the inverter.
 * @param   ptr_meterReadings    A pointer to the vector of Mirror Meter Readings to be populated.
 * @param   bol_isInitalReading  A boolean value indicating if the readings are the initial readings.
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
 */
GlobScop int CST_GetMirrorMeterReadings(std::vector<SE_MirrorMeterReading_t> *ptr_meterReadings, 
                                        bool bol_isInitalReading);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Helper function used to get the size of the Mirror Meter Readings.
 * @returns The size of the Mirror Meter Readings.
 */
GlobScop size_t CST_GetMirrorMeterReadingsSize(void);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Determines if a new log event has occurred and populates the referenced LogEvent structure with the current
 *          log events of the inverter.
 * @param   ptr_logEvent    A pointer to the LogEvent structure to be populated.
 * @returns CST_LOG_EVENT_NOT_OCCURRED, CST_LOG_EVENT_OCCURRED, CST_LOG_EVENT_ERROR
 * @returns CST_MULTIPLE_LOG_EVENTS_OCCURRED indicating that the function should be called again
 */
GlobScop CST_LogEventOccurred_t CST_HasLogEventOccurred(SE_LogEvent_t *ptr_logEvent);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Sets the current control settings of the inverter to the referenced DERControl structure.
 * @param   ptr_control A pointer to the DERControl structure to be populated.
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
 */
GlobScop int CST_SetControl(SE_DERControlBase_t *ptr_control);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Resets the current control settings of the inverter to the default values defined in the mapping.
 * @param   u32_rstCtrlFlags A bitmask of controls to reset.
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
 */
GlobScop int CST_ResetControl(uint32_t u32_rstCtrlFlags);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Sets the default control settings of the inverter to the referenced DERControl structure.
 * @param   ptr_defaultControl  A pointer to the DefaultDERControl structure to be populated.
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
 */
GlobScop int CST_SetDefaultControl(SE_DefaultDERControl_t *ptr_defaultControl);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Resets the default control settings of the inverter to the to the default values defined in the mapping.
 * @param   u32_rstCtrlFlags  A bitmask of controls to reset.
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
 */
GlobScop int CST_ResetDefaultControl(uint32_t u32_rstCtrlFlags);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Sets the current curve settings of the inverter to the referenced DERCurve structure.
 * @param   ptr_curve   A pointer to the DERCurve structure to be populated.
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
 */
GlobScop int CST_SetCurve(SE_DERCurve_t *ptr_curve);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * @brief   Resets the current curve settings of the inverter to the default values defined in the mapping.
 * @param   u32_rstCtrlFlags   A bitmask of controls to reset.
 * @returns EXIT_SUCCESS or 0 upon success, otherwise the POSIX error code is returned.
 */
GlobScop int CST_ResetCurve(uint32_t u32_rstCtrlFlags);

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifdef __cplusplus
}
#endif

#endif // LION_CST_H_