#include <zcbor_decode.h>
#include <zcbor_encode.h>
#include <stdio.h> 
#include <cstdlib>

#include "lion/ial/IAL.h"
#include "lion/ial/ems/IAL_Data.h"
#include "lion/ial/ems/IAL_Control.h"
#include "lion/ial/ems/IAL_Modbus.h"
#include "lion/log/log.h"
#include "include/CST_IAL.hpp"
#include "lion/CST.h"
#include "lion/I3E.h"

LOG_MODULE_REGISTER(cst_ial, LOG_LEVEL_INF);


#define SZ_MODEL_UNAVAIL "unavailable"
#define SZ_MODEL_AVAIL   "available"
#define SZ_MODEL_ACTDER  "active_der"
#define SZ_MODEL_CONF    "conflict"
#define SZ_MODEL_UNKNOWN "UNKNOWN"

FileScop const char *sz_arr_UtilModeStatus[] = {
    SZ_MODEL_UNAVAIL,
    SZ_MODEL_AVAIL,
    SZ_MODEL_ACTDER,
    SZ_MODEL_CONF,
    SZ_MODEL_UNKNOWN
};

#define CST_IAL_CERT_LFDI_LEN 40
#define CST_IAL_CERT_SFDI_LEN 12

#define CST_SYSCTL_RESTART "systemctl restart lion-ems-cm-i3e"

/**
 * @brief Enumerated type for the different unit types in the system.
 */
typedef enum { PCS, PDU, PMU, DCDC, GW, HVAC, EMS, I3E, __UNIT_TYPE_MAX } unit_type_t;

#define CST_sz_UNIT_TYPE_I3E     COD_sz_UNIT_TYPE_I3E

/**********************************************************************************************************************
 * @cond                                               FILE SCOPE                                              @endcond
/*********************************************************************************************************************/
FileScop ICC_Function_Handler_t str_i3eUtilCmdModeStatusHandler;
FileScop ICC_Function_Handler_t str_i3eCommissioningHandler;
FileScop ICC_Function_Handler_t str_i3eIndicatorsHandler;
FileScop ICC_Event_Handler_t str_i3eConnectionSettingsModifiedHandler;

/**
 * @brief Registers all callback functions with the ICC.
 */
FileScop int cst_ial_RegisterCallbackHandlers(void);

/**
 * @brief Unregisters all callback functions with the ICC.
 */
FileScop void cst_ial_UnregisterCallbackHandlers(void);

/**
 * @brief The ICC callback for receiving the utility command mode status enum.
 *
 * @param u32_ID ICC Function ID
 * @param vptr_Data Pointer to CBOR data
 * @param u32_DataLength Length of CBOR data
 * @param vptr_Response Pointer to response buffer
 * @param ptr_u32_responseLength Pointer to response length
 * @param vptr_UserData Pointer to user data
 *
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
FileScop int cst_ial_i3eUtilCmdModeStatCallback(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength, 
                                        void *vptr_Response, uint32_t *ptr_u32_responseLength, void *vptr_UserData);

/**
 * @brief The ICC callback for receiving the comissioning request from the Lion Support Tool.
 *
 * @param u32_ID ICC Function ID
 * @param vptr_Data Pointer to CBOR data
 * @param u32_DataLength Length of CBOR data
 * @param vptr_Response Pointer to response buffer
 * @param ptr_u32_responseLength Pointer to response length
 * @param vptr_UserData Pointer to user data
 *
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
FileScop int cst_ial_i3eCommissioningCallback(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength, 
                                        void *vptr_Response, uint32_t *ptr_u32_responseLength, void *vptr_UserData);

/**
 * @brief The ICC callback for receiving the PKI certificates lfdi/sfdi request from the Lion Support Tool.
 *
 * @param u32_ID ICC Function ID
 * @param vptr_Data Pointer to CBOR data
 * @param u32_DataLength Length of CBOR data
 * @param vptr_Response Pointer to response buffer
 * @param ptr_u32_responseLength Pointer to response length
 * @param vptr_UserData Pointer to user data
 *
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
FileScop int cst_ial_i3eIndicatorsCallback(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength, 
                                        void *vptr_Response, uint32_t *ptr_u32_responseLength, void *vptr_UserData);

/**
 * @brief The ICC callback for processing a connection settings modified event (changes to IP/Port of utility, 
 * thereby triggering a connection request).
 *
 * @param int32_t ICC Event ID
 * @param vptr_Data Pointer to CBOR data
 * @param siz_Data Length of CBOR data
 * @param vptr_UsrData Pointer to user data
 */
FileScop void cst_ial_i3eConnectionSettingsModifiedCallback(int32_t s32_EventId, const void *vptr_Data, size_t siz_Data,
                                                void *vptr_UsrData);

/**
 * @brief Validates the unit type string and returns the corresponding unit type.
 *
 * @param sz_UnitType Unit type to validate
 * @param ptr_UnitType Pointer to the unit type
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
FileScop int cst_ial_ValidateUnitType(const char *sz_UnitType, unit_type_t *ptr_UnitType);

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

ModScop void CST_IAL_Init() 
{
    cst_ial_RegisterCallbackHandlers();
}

/**********************************************************************************************************************
 * @cond                                               FILE SCOPE                                              @endcond
/*********************************************************************************************************************/

FileScop int cst_ial_RegisterCallbackHandlers(void)
{
    uint32_t u32_FunctionID = 0;
    
    // Register Utility Command Mode Status Callback
    u32_FunctionID = IAL_Int_BuildTargetedFunctionID(IAL_TARGET_COMP, IAL_UTIL_CMD_MD_STATUS_ID);
    if(IAL_RegisterCallback(u32_FunctionID, &str_i3eUtilCmdModeStatusHandler, 
                                    cst_ial_i3eUtilCmdModeStatCallback) != EXIT_SUCCESS)
    {
        LOG_ERR("Failed to register Read Registers Callback");
        return EXIT_FAILURE;
    }

    // Register I3E Comissioning Callback
    u32_FunctionID = IAL_Int_BuildTargetedFunctionID(IAL_TARGET_COMP, IAL_START_I3E_COMMISSION_ID);
    if(IAL_RegisterCallback(u32_FunctionID, &str_i3eCommissioningHandler, 
                                    cst_ial_i3eCommissioningCallback) != EXIT_SUCCESS)
    {
        LOG_ERR("Failed to register Read Registers Callback");
        return EXIT_FAILURE;
    }

    // Register LFDI/SFDI Request Callback
    u32_FunctionID = IAL_Int_BuildTargetedFunctionID(IAL_TARGET_COMP, IAL_GET_I3E_INDICATORS_ID);
    if(IAL_RegisterCallback(u32_FunctionID, &str_i3eIndicatorsHandler, 
                                    cst_ial_i3eIndicatorsCallback) != EXIT_SUCCESS)
    {
        LOG_ERR("Failed to register Read Registers Callback");
        return EXIT_FAILURE;
    }

    // Register Connection Settings Callback
    if(IAL_RegisterEventCallback(IAL_SETTINGS_MODIFIED_ID, &str_i3eConnectionSettingsModifiedHandler, 
                                    cst_ial_i3eConnectionSettingsModifiedCallback) != EXIT_SUCCESS)
    {
        LOG_ERR("Failed to register Read Registers Callback");
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void cst_ial_UnregisterCallbackHandlers(void)
{
    IAL_UnregisterCallback(&str_i3eUtilCmdModeStatusHandler);
    IAL_UnregisterCallback(&str_i3eCommissioningHandler);
    IAL_UnregisterCallback(&str_i3eIndicatorsHandler);
    IAL_UnregisterEventCallback(&str_i3eConnectionSettingsModifiedHandler);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop int cst_ial_i3eUtilCmdModeStatCallback(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength, 
                                        void *vptr_Response, uint32_t *ptr_u32_responseLength, void *vptr_UserData)
{

    // Decode the string from the CBOR data
    ZCBOR_STATE_D(CBORDecodeState, 1, static_cast<const uint8_t*>(vptr_Data), u32_DataLength, 1, 0);

    StackScop struct zcbor_string str_ModelEnumState;
    if (zcbor_tstr_decode(CBORDecodeState, &str_ModelEnumState) == 0)
    {
        return -EINVAL;
    }

    StackScop int int_rc = str_ModelEnumState.len;   
    if (int_rc < 0)
    {
        return -EINVAL;
    }

    // Null-terminate the string
    StackScop char sz_ModelEnumStatus[int_rc + 1] = {0};
    memcpy(sz_ModelEnumStatus, str_ModelEnumState.value, int_rc);

    // Log the received utility command mode status
    LOG_INF("Utility Command Mode Status: %s", sz_ModelEnumStatus);

    // Use setter function to update the utility command mode status in the CST module (if matched)
    for (size_t i = 0; i < sizeof(sz_arr_UtilModeStatus) / sizeof(sz_arr_UtilModeStatus[i]); i++) 
    {
        // String match
        if (strcmp(sz_ModelEnumStatus, sz_arr_UtilModeStatus[i]) == 0) 
        {
            CST_UtilCMDStatusSetter(CST_UtilModeStatus_t(i));
            break;
        }
    }

    // Set returns for the ICC callback
    ptr_u32_responseLength = 0;
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop int cst_ial_i3eCommissioningCallback(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength, 
                                        void *vptr_Response, uint32_t *ptr_u32_responseLength, void *vptr_UserData)
{
    // Start the I3E Commissioning process
    StackScop int16_t s16_CommissioningSt = I3E_Commissioning();

    // Error checking
    if(s16_CommissioningSt != EXIT_SUCCESS)
    {
        LOG_ERR("Failed to commission I3E");

        // If the I3E is already commissioned, return a denied response
        if(s16_CommissioningSt == -EALREADY)
        {
            LOG_ERR("I3E is already commissioned");
            if(IAL_Ctl_EncodeResponse_StartI3eCommissioning(IAL_sz_I3E_COMMISSION_ST_DENIED, vptr_Response, ptr_u32_responseLength) != EXIT_SUCCESS)
            {
                LOG_ERR("Failed to encode I3E Commissioning Response");
                return -EXIT_FAILURE;
            }
        }
        
        // If the I3E commissioning process failed, return a failure response
        if(s16_CommissioningSt == -EXIT_FAILURE)
        {
            LOG_ERR("Failed to commission I3E");
            if(IAL_Ctl_EncodeResponse_StartI3eCommissioning(IAL_sz_I3E_COMMISSION_ST_FAILURE, vptr_Response, ptr_u32_responseLength) != EXIT_SUCCESS)
            {
                LOG_ERR("Failed to encode I3E Commissioning Response");
                return -EXIT_FAILURE;
            }
        }

    }

    else
    {
        // If we reached this point, commissioning was successful - encode the successful response
        if(IAL_Ctl_EncodeResponse_StartI3eCommissioning(IAL_sz_I3E_COMMISSION_ST_SUCCESS, vptr_Response, ptr_u32_responseLength) != EXIT_SUCCESS)
        {
            LOG_INF("Failed to encode I3E Commissioning Response");
            return -EXIT_FAILURE;
        }
    }
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop int cst_ial_i3eIndicatorsCallback(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength, 
                                        void *vptr_Response, uint32_t *ptr_u32_responseLength, void *vptr_UserData)
{

    // Get the I3E Indicators from the I3E application
    StackScop char sz_lfdi[CST_IAL_CERT_LFDI_LEN + 1] = {0};
    StackScop char sz_sfdi[CST_IAL_CERT_SFDI_LEN + 1] = {0};
    if(I3E_GetCertIndicators(sz_lfdi, sz_sfdi) != EXIT_SUCCESS)
    {
        LOG_ERR("Failed to get I3E Indicators");
        return -EXIT_FAILURE;
    }
    
    // Encode the response with the I3E Indicators
    if(IAL_EncodeResponse_GetI3EIndicators(sz_lfdi, sz_sfdi, vptr_Response, ptr_u32_responseLength) != EXIT_SUCCESS)
    {
        LOG_ERR("Failed to encode I3E Indicators");
        return -EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop void cst_ial_i3eConnectionSettingsModifiedCallback(int32_t s32_EventId, const void *vptr_Data, size_t siz_Data,
                                                void *vptr_UsrData)
{
    LOG_INF("Callback for Settings Modified");

    // Decode keys/values from the CBOR encoded data
    ZCBOR_STATE_D(CBORDecodeState, 1, (uint8_t*)vptr_Data, siz_Data, 1, 0);
    if (zcbor_list_start_decode(CBORDecodeState))
    {
        StackScop struct zcbor_string sz_zcbor_UnitType;
        for (uint32_t i = 0; zcbor_tstr_decode(CBORDecodeState, &sz_zcbor_UnitType); i++)
        {
            unit_type_t enm_UnitType = __UNIT_TYPE_MAX;
            if (cst_ial_ValidateUnitType((char *)sz_zcbor_UnitType.value, &enm_UnitType) == EXIT_SUCCESS)
            {
                LOG_INF("%s Config file has been modified externally, restarting I3E application", 
                                                (char *)sz_zcbor_UnitType.value);

                // #TODO - this may or may not be a good solution, but it's very easy and lets systemd do the thinking 
                // Call systemctl to restart the service
                int ret = std::system(CST_SYSCTL_RESTART);
            }
        }
        zcbor_list_end_decode(CBORDecodeState);
    }
    else
    {
        LOG_INF("Decoding CBOR Error");
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop int cst_ial_ValidateUnitType(const char *sz_UnitType, unit_type_t *ptr_UnitType)
{
    if (strncmp(sz_UnitType, CST_sz_UNIT_TYPE_I3E, strlen(CST_sz_UNIT_TYPE_I3E)) == EXIT_SUCCESS)
    {
        *ptr_UnitType = I3E;
    }
    else
    {
        LOG_ERR("Error: Unknown device type: %s", sz_UnitType);
        return -ENODEV;
    }
    return EXIT_SUCCESS;
}