/**********************************************************************************************************************
* @file            IAL.h
*
* @brief           ICC Abstraction Library. This service acts as an abstraction layer for ICC function calls 
                   and callback registration for applications.
*
* @author          Devan Hone
* @date            07/31/2024
* @copyright       2024 Lion Energy LLC.
*                  All rights reserved
*
*********************************************************************************************************************/
/**
* @addtogroup      IAL
*                  Identifier Description for the ICC Abstraction Library (IAL) Module  
* 
* @ref             IAL.h 
*                  
*********************************************************************************************************************/
#include "COD_Compiler_Defines.h"

#ifndef LION_IAL_H__
#define LION_IAL_H__

/*********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

#include <lion/icc/ICC.h>
#include <stdlib.h>

#ifdef __cplusplus
extern "C"
{
#endif

//---------------------------------------//
// Section: Defines/Enums/Typedefs       //
//---------------------------------------//

#define IAL_u16_ICC_REQUEST_BUFFER_SIZE     2048
#define IAL_u16_ICC_RESPONSE_BUFFER_SIZE    1024

#define IAL_u8_ICC_CALL_TIMEOUT_S           10
#define IAL_u8_ICC_CONFIG_CALL_TIMEOUT_S    30
#define IAL_u8_ICC_EVENT_TIMEOUT_S          1

//-----EMS Pro Function IDs-----//
#define IAL_HEARTBEAT_ID                0x01000000
#define IAL_SOS_HEARTBEAT_ID            0x01100000
#define IAL_GET_SYNC_LOCAL_TIME_ID      0x01000001
#define IAL_GET_QUANTIFIABLE_DATA_ID    0x02000100
#define IAL_SET_QUANTIFIABLE_DATA_ID    0x02000101
#define IAL_GET_UNIT_DATA_ID            0x02000102
#define IAL_SET_UNIT_DATA_ID            0x02000103
#define IAL_GET_UNIT_CODES_ID           0x02000104
#define IAL_GET_BUFFERED_DATA           0x02000105
#define IAL_READ_REGISTERS_ID           0x02000106
#define IAL_WRITE_REGISTERS_ID          0x02000107
#define IAL_READ_MULTIPLE_ID            0x02000108
#define IAL_WRITE_MULTIPLE_ID           0x02000109
#define IAL_LOCK_REGISTERS_ID           0x0200010A
#define IAL_UNLOCK_REGISTERS_ID         0x0200010B
#define IAL_RESET_LOCK_REGISTRY_ID      0x0200010C
#define IAL_GET_EMS_CONFIG_ID           0x02000200
#define IAL_SET_EMS_CONFIG_ID           0x02000201
#define IAL_GET_PMU_CONFIG_ID           0x03000200
#define IAL_SET_PMU_CONFIG_ID           0x03000201
#define IAL_GET_UNIT_CONFIG_ID          0x02000202
#define IAL_SET_UNIT_CONFIG_ID          0x02000203
#define IAL_GET_TOU_ID                  0x02000204
#define IAL_SET_TOU_ID                  0x02000205
#define IAL_DCU_RESTART_ID              0x02FF0000
#define IAL_PMU_RESTART_ID              0x03FF0000
#define IAL_DCU_REINIT_COM_ID           0x02000300
#define IAL_ALERT_STATUS_ID             0x01000310
#define IAL_VERSION_INFO_ID             0x01000301
#define IAL_SELF_PROVISION_START_ID     0x01000302
#define IAL_SELF_PROVISION_STATE_ID     0x01000303
#define IAL_SETTINGS_MODIFIED_ID        0x01000320
#define IAL_UTIL_CMD_MD_STATUS_ID       0x01000321
#define IAL_START_I3E_COMMISSION_ID     0x01000322
#define IAL_GET_I3E_INDICATORS_ID       0x01000323

#define IAL_RESOURE_RELEASE_ID          0x04000330
#define IAL_START_UPDATE_ID             0x04000331
#define IAL_UPDATE_STATUS_ID            0x04000332
#define IAL_CLEAR_UPDATE_STATUS_ID      0x04000333

#define IAL_U32_TARGET_MASK 0xFF000000
#define IAL_U8_TARGET_SHIFT 24

//-----EMS Pro Unit Types-----//
#if CONFIG_EMS_SCH
#define IAL_sz_UNIT_TYPE_ATS     COD_sz_UNIT_TYPE_ATS
#define IAL_sz_UNIT_TYPE_DCDC    COD_sz_UNIT_TYPE_DCDC
#define IAL_sz_UNIT_TYPE_DCU     COD_sz_UNIT_TYPE_DCU
#define IAL_sz_UNIT_TYPE_EMS     COD_sz_UNIT_TYPE_EMS
#define IAL_sz_UNIT_TYPE_GW      COD_sz_UNIT_TYPE_GW
#define IAL_sz_UNIT_TYPE_HVAC    COD_sz_UNIT_TYPE_HVAC
#define IAL_sz_UNIT_TYPE_PCS     COD_sz_UNIT_TYPE_PCS
#define IAL_sz_UNIT_TYPE_PDU     COD_sz_UNIT_TYPE_PDU
#define IAL_sz_UNIT_TYPE_PMU     COD_sz_UNIT_TYPE_PMU
#else
#define IAL_sz_UNIT_TYPE_DCU     "INV" //DCU is not supported
#define IAL_sz_UNIT_TYPE_PMU     "INV" //PMU is not supported
#endif

#define IAL_sz_TYPE_TOU          "tou"

typedef enum {
    IAL_TARGET__INVALID = 0,
    IAL_TARGET_CCU, //EMS Pro Cloud Communication Unit
    IAL_TARGET_DCU, //EMS Pro Domain Controller Unit
    IAL_TARGET_PMU, //EMS Pro Power Metering Unit
    IAL_TARGET_COMP, //EMS Compact
} IAL_TargetDevice_t;

//---------------------------------------//
// Section: Function Prototypes          //
//---------------------------------------//

/**
 * @brief Initialize the ICC Abstraction Library (IAL).
 * 
 * @param ptr_IccCtx A pointer to the ICC context.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 */
GlobScop int IAL_Init(ICC_Context_t *ptr_IccCtx);

/**
 * @brief  Registers a callback function to be called when the specified function ID is received from the ICC.
 * 
 * @param u32_FunctionID  The function ID the handler will be registered to.
 * @param ptr_str_Handler Pointer to the function handler to be registered.
 * @param fn_Callback     Callback function to be called when an alert status update is received.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_RegisterCallback(uint32_t u32_FunctionID, ICC_Function_Handler_t *ptr_str_Handler,
                                    ICC_Function_HandlerCallback_t fn_Callback);

/**
 * @brief  Unregisters a callback function from the ICC.
 * 
 * @param ptr_str_Handler Pointer to the function handler to be unregistered.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 */
GlobScop int IAL_UnregisterCallback(ICC_Function_Handler_t *ptr_str_Handler);

/**
 * @brief Registers a callback function to be called when the specified event ID is received from the ICC.
 * 
 * @param u32_EventID     The event ID the handler will be registered to.
 * @param ptr_str_Handler Pointer to the event handler to be registered.
 * @param fn_Callback     Callback function to be called when an event is received.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 */
GlobScop int IAL_RegisterEventCallback(uint32_t u32_EventID, ICC_Event_Handler_t *ptr_str_Handler,
                                        ICC_Event_Callback_t fn_Callback);

/**
 * @brief Unregisters an event callback function from the ICC.
 * 
 * @param ptr_str_Handler Pointer to the event handler to be unregistered.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 */
GlobScop int IAL_UnregisterEventCallback(ICC_Event_Handler_t *ptr_str_Handler);

/**
 * @brief  Helper function to build a targeted function ID for the specified target device.
 * 
 * @details Currently, the ICC does not have a target parameter in the function calls. The target is instead embedded
 *          in the MSB of the function ID. This function clears the MSB of the 32-bit function ID and sets it to the
 *          specified target device.
 * 
 * @param enm_Target The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param u32_FunctionID The function ID of the call to be made.
 * 
 * @returns The built targeted function ID.
 */
GlobScop uint32_t IAL_Int_BuildTargetedFunctionID(IAL_TargetDevice_t enm_Target, uint32_t u32_FunctionID);


#ifdef __cplusplus
}
#endif

#endif // LION_IAL_H__
//EOF
