/**********************************************************************************************************************
* @file            IAL_Control.h
*
* @brief           Provides APIs for control related ICC functions and events for the EMS.
*                  
*
* @author          Devan Hone
* @date            07/25/2024
* @copyright       2024 Lion Energy LLC.
*                  All rights reserved
*
*********************************************************************************************************************/
/**
* @addtogroup      IAL
*                  Identifier Description for the ICC Abstraction Library (IAL) Module  
* 
* @ref             IAL.h 
*                  
*********************************************************************************************************************/
#include "COD_Compiler_Defines.h"

#ifndef LION_IAL_CONTROL_H__
#define LION_IAL_CONTROL_H__

/*********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

#include <lion/ial/IAL.h>

#define IAL_sz_MAX_PROV_ST_LEN      (16)      //Max length of a provisioning state string, including null terminator
#define IAL_sz_PROV_ST_NOT_STARTED  "nostart" //AWS self-provisioning state - Not started
#define IAL_sz_PROV_ST_ACTIVE       "active"  //AWS self-provisioning state - Active
#define IAL_sz_PROV_ST_SUCCESS      "succ"    //AWS self-provisioning state - Success
#define IAL_sz_PROV_ST_FAILURE      "fail"    //AWS self-provisioning state - Failure

#define IAL_sz_MAX_I3E_COMMISSION_ST_LEN    (10) //Max length of an I3E commissioning state string, including null terminator
#define IAL_sz_I3E_COMMISSION_ST_SUCCESS    "succ" //I3E commissioning state - Successfully started the commissioning process
#define IAL_sz_I3E_COMMISSION_ST_FAILURE    "fail" //I3E commissioning state - Failed to start the commissioning process
#define IAL_sz_I3E_COMMISSION_ST_DENIED     "denied" //I3E commissioning state - Denied the commissioning request

#define IAL_sz_MAX_I3E_UTIL_CMD_MD_STATUS_LEN     (12) //Max length of an I3E utility command mode status string, including null terminator
#define IAL_sz_I3E_UTIL_CMD_MD_STATUS_UNAVAILABLE "unavailable" //I3E utility command mode status - Unavailable
#define IAL_sz_I3E_UTIL_CMD_MD_STATUS_AVAILABLE   "available"   //I3E utility command mode status - Available
#define IAL_sz_I3E_UTIL_CMD_MD_STATUS_ACTIVE_DER  "active_der"  //I3E utility command mode status - Active DER
#define IAL_sz_I3E_UTIL_CMD_MD_STATUS_CONFLICT    "conflict"    //I3E utility command mode status - Utility command mode
                                                                //has been aborted due to a conflict (e.g. a fault occurred)

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @brief  Send a restart request to the two boards of the EMS using the Inter-Controller Communication (ICC) Service. 
 * 
 * @param enm_Target The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param sz_UnitType Unit type to restart, ie. "dcu", "pmu"
 * @param bol_Force true to force the restart, false to allow the DCU model to enter OTA (standby) mode before restarting.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, EXIT_FAILURE upon failure
 */
GlobScop int IAL_Ctl_Restart(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, bool bol_Force);

/**
 * @brief  Send a request to reinitialize communication services to the DCU (Domain Controller Unit) using the 
 *         Inter-Controller Communication (ICC) Service. 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, EXIT_FAILURE upon failure
 */
GlobScop int IAL_Ctl_ReinitComDCU(void);

/**
 * @brief  Send a request to all units in the system to publish their current version information.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Ctl_RequestVersionInfo(void);

/**
 * @brief  Starts the AWS self-provisioning process.
 * 
 * @param enm_Target The target device for the ICC call, eg. IAL_TARGET_CCU, IAL_TARGET_COMP
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Ctl_StartAwsSelfProvisioning(IAL_TargetDevice_t enm_Target);

/**
 * @brief  Get the current state of the AWS self-provisioning process.
 * 
 * @param enm_Target The target device for the ICC call, eg. IAL_TARGET_CCU, IAL_TARGET_COMP
 * @param sz_State Pointer to a buffer to store the state string.
 * @param siz_State  Size of the buffer to store the state string.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Ctl_GetSelfProvisioningState(IAL_TargetDevice_t enm_Target, char *sz_State, size_t siz_State);

/**
 * @brief A helper function to encode the response for calls to get the AWS self-provisioning state.
 * 
 * @details This function is intended to be called inside a callback function for the @ref IAL_SELF_PROVISION_STATE_ID
 *          function ID. It takes a string (the self-provisioning state) and encodes it into the response buffer.
 * 
 * @param sz_State The self-provisioning state string to encode.
 * @param siz_State The size of the self-provisioning state string.
 * @param vptr_Response The buffer to encode the response into.
 * @param ptr_siz_Response A pointer to store the size of the encoded response.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Ctl_EncodeResponse_SelfProvisioningState(char *sz_State, size_t siz_State, void *vptr_Response,
                                                            size_t *ptr_siz_Response);

/**
 * @brief Broadcasts an event that notifies the system of component settings that have been modified.
 * 
 * @param arr_sz_UnitsModified Array of unit types that have been modified.
 * @param siz_UnitsModified Size of the array of modified units.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Ctl_SettingsModified(const char **arr_sz_UnitsModified, size_t siz_UnitsModified);

/**
 * @brief Request the I3E application to start the commissioning process.
 * 
 * @details This function sends a request to the I3E application to start the commissioning process. The result of the
 *          request is stored in the sz_Result buffer. The I3E application can reject the request if it has already been
 *          commissioned or if the commissioning process is already in progress.
 * 
 * @param enm_Target The target device for the ICC call, eg. IAL_TARGET_CCU, IAL_TARGET_COMP.
 * @param sz_Result Pointer to a buffer to store the result of the commissioning request.
 *                  Expected return values are "succ", "fail", or "denied".
 * @param siz_Result Size of the buffer to store the result of the commissioning request.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 */
GlobScop int IAL_Ctl_StartI3eCommissioning(IAL_TargetDevice_t enm_Target, char *sz_Result, size_t siz_Result);

/**
 * @brief A helper function to encode the response for calls to start the I3E commissioning process.
 * 
 * @details This function is intended to be called inside a callback function for the @ref IAL_I3E_START_COMMISSION_ID
 *          function ID. It takes a string and encodes it into the response buffer.
 * 
 * @param sz_Result The I3E commissioning start result string to encode.
 *                  Expected return values are "succ", "fail", or "denied".
 * @param vptr_Response The buffer to encode the response into.
 * @param ptr_siz_Response A pointer to store the size of the encoded response.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 * 
 * @example
 *      //StartI3eCommissioning callback function
 *      //This function will need to be registered with the IAL/ICC using IAL_RegisterCallback().
 *      //startI3eCommissioning() is a placeholder function that would be implemented in the I3E application.
 *      int StartI3eCommissioning_CB(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength, void *vptr_Response, 
 *                                      uint32_t *ptr_u32_ResponseLength, void *vptr_UserData)
 *      {
 *         UNUSED(u32_ID);
 *         UNUSED(vptr_Data);
 *         UNUSED(u32_DataLength);
 *         char result[IAL_sz_MAX_I3E_COMMISSION_ST_LEN] = startI3eCommissioning();
 *         return IAL_Ctl_EncodeResponse_StartI3eCommissioning((const char*)result, vptr_Response, ptr_u32_ResponseLength);
 *      }
 */
GlobScop int IAL_Ctl_EncodeResponse_StartI3eCommissioning(const char *sz_Result, void *vptr_Response,
                                                            size_t *ptr_siz_Response);

/**
 * @brief Sends a message to the I3E application to report the status of utility command mode of the EMS.
 * 
 * @param enm_Target The target device for the ICC call, eg. IAL_TARGET_CCU, IAL_TARGET_COMP.
 * @param sz_Status The status of the utility command mode. Expected values are "unavailable", "available", "active_der",
 *                  or "conflict".
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 */
GlobScop int IAL_Ctl_ReportUtilCmdStatus(IAL_TargetDevice_t enm_Target, const char *sz_Status);

/**
 * @brief A helper function to decode the payload of the I3E utility command mode status message.
 * 
 * @details This function is intended to be called inside a callback function for the @ref IAL_I3E_UTIL_CMD_STATUS_ID
 * 
 * @param vptr_Data Pointer to the CBOR encoded data.
 * @param u32_DataLength Length of the CBOR encoded data.
 * @param sz_Status Pointer to a buffer to store the utility command mode status.
 * @param siz_Status Size of the buffer to store the utility command mode status.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 * 
 * @example
 *      //ReportUtilCmdStatus callback function
 *      //This function will need to be registered with the IAL/ICC using IAL_RegisterCallback().
 *      int UtilCmdStatusCallback(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength, void *vptr_Response,
 *                                  uint32_t *ptr_u32_ResponseLength, void *vptr_UserData)
        {
            char sz_Status[IAL_sz_MAX_I3E_UTIL_CMD_MD_STATUS_LEN] = "";
            int int_RC = IAL_Ctl_DecodeRequest_UtilCmdStatus(vptr_Data, u32_DataLength, sz_Status, sizeof(sz_Status));
            LOG_INF("Received UtilCmdStatus: %s with err %d", sz_Status, int_RC);

            //Handle the status here

            *ptr_u32_ResponseLength = 0; //No response data needed, simply acknowledge the message
            return int_RC;
        }
 */
GlobScop int IAL_Ctl_DecodeRequest_UtilCmdStatus(const void *vptr_Data, uint32_t u32_DataLength, char *sz_Status,
                                                    size_t siz_Status);

/**
 * @brief Broadcasts a heartbeat message to all units in the system.
 * 
 * @param enm_Device The device ID of the broadcasting unit.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 */
GlobScop int IAL_Ctl_HeartbeatBroadcast(IAL_TargetDevice_t enm_Device);

/**
 * @brief Send a request to all units in the system to release a resource
 * 
 * @param sz_Resource The resource to release (specified on a per implementation basis)
 * @param u8_Urgency The urgency of the resource release
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Ctl_ResourceRelease(const char * sz_Resource, uint8_t u8_Urgency);

/**
 * @brief A helper function to decode the payload of the resource release request message.
 * 
 * @details This function is intended to be called inside a callback function for the @ref IAL_RESOURCE_RELEASE_ID
 * 
 * @param vptr_Data Pointer to the CBOR encoded data.
 * @param u32_DataLength Length of the CBOR encoded data.
 * @param sz_Resource Pointer to a buffer to store the resource to release.
 * @param siz_Resource Size of the buffer to store the resource to release.
 * @param ptr_u8_Urgency Pointer to store the urgency of the resource release.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 * 
 * @example
 *     //ResourceReleaseRequest callback function
 *     //This function will need to be registered with the IAL/ICC using IAL_RegisterCallback().
 *     int ResourceReleaseRequest_CB(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength, void *vptr_Response,
 *                                   uint32_t *ptr_u32_ResponseLength, void *vptr_UserData)
 *    {
 *      UNUSED(u32_ID);
 *      UNUSED(vptr_Response);
 *      UNUSED(ptr_u32_ResponseLength);
 *      char sz_Resource[IAL_sz_MAX_RESOURCE_LEN] = "";
 *      uint8_t u8_Urgency = 0; 
 *      int int_RC = IAL_Ctl_DecodeRequest_ResourceRelease(vptr_Data, u32_DataLength, sz_Resource, sizeof(sz_Resource),
 *                                                        &u8_Urgency);
 *      // Handle the resource release here
 * 
 *      return EXIT_SUCCESS;
 *   }
 */
GlobScop int IAL_Ctl_DecodeRequest_ResourceRelease(const void *vptr_Data, uint32_t u32_DataLength, char *sz_Resource,
                                                    size_t siz_Resource, uint8_t *ptr_u8_Urgency);
/**
 * @brief Send a request to the UMA application to update a specific component.
 * 
 * @param enm_Target The target device for the ICC call, eg. IAL_TARGET_CCU, IAL_TARGET_COMP
 * @param sz_SourceKey A unique key to identify the source of the update request
 * @param u32_Type The type of update request
 * @param sz_ArtifactPath The path to the artifact to update
 * @param arr_sz_Keys Array of keys for the update request
 * @param arr_sz_Values Array of values for the update request
 * @param siz_Pairs The number of key-value pairs in the update request
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Ctl_StartUpdate(IAL_TargetDevice_t enm_Target, const char * sz_SourceKey, uint32_t u32_Type, 
    const char * sz_ArtifactPath, const char ** arr_sz_Keys, const char ** arr_sz_Values, size_t siz_Pairs);

/**
 * @brief A helper function to decode the payload of the start update message.
 *  
 * @details This function is intended to be called inside a callback function for the @ref IAL_START_UPDATE_ID
 * 
 * @param vptr_Data Pointer to the CBOR encoded data.
 * @param u32_DataLength Length of the CBOR encoded data.
 * @param sz_SourceKey Pointer to a buffer to store the source key.
 * @param siz_SourceKey Size of the buffer to store the source key.
 * @param u32_Type Pointer to store the type of update request.
 * @param sz_ArtifactPath Pointer to a buffer to store the path to the artifact to update.
 * @param siz_ArtifactPath Size of the buffer to store the path to the artifact to update.
 * @param arr_sz_Keys Pointer to an array of keys for the update request.
 * @param siz_Key Size of each key buffer in the array.
 * @param arr_sz_Values Pointer to an array of values for the update request.
 * @param siz_Value Size of each value buffer in the array.
 * @param siz_MaxPairs The maximum number of key-value pairs that can be decoded.
 * @param siz_Pairs The number of key-value pairs buffers in the update request.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 * 
 * @example
 *    //StartUpdate callback function
 *    //This function will need to be registered with the IAL/ICC using IAL_RegisterCallback().
 *    int StartUpdate_CB(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength, void *vptr_Response,
 *                        uint32_t *ptr_u32_ResponseLength, void *vptr_UserData)
 *    {
 *      UNUSED(u32_ID);
 *      UNUSED(vptr_Response);
 *      UNUSED(ptr_u32_ResponseLength);
 *      char sz_ArtifactPath[IAL_sz_MAX_ARTIFACT_PATH_LEN] = "";
 *      char arr_sz_Keys[IAL_sz_MAX_KEY_LEN][IAL_sz_MAX_PAIRS];
 *      char arr_sz_Values[IAL_sz_MAX_VALUE_LEN][IAL_sz_MAX_PAIRS];
 *      
 *      char *arr_ptr_sz_Keys[IAL_sz_MAX_PAIRS];
 *      char *arr_ptr_sz_Values[IAL_sz_MAX_PAIRS];
 *      for (size_t i = 0; i < IAL_sz_MAX_PAIRS; i++)
 *      {
 *          arr_ptr_sz_Keys[i] = arr_sz_Keys[i];
 *          arr_ptr_sz_Values[i] = arr_sz_Values[i];
 *      }
 * 
 *      size_t siz_Pairs = 0;
 *      
 *      int int_RC = IAL_Ctl_DecodeRequest_StartUpdate(vptr_Data, u32_DataLength, sz_SourceKey, sizeof(sz_SourceKey),
 *                                                     u32_Type, sz_ArtifactPath, sizeof(sz_ArtifactPath), arr_ptr_sz_Keys,
 *                                                     sizeof(arr_sz_Keys[0]), arr_ptr_sz_Values, sizeof(arr_sz_Values[0]),
 *                                                     MAX_PAIRS, &siz_Pairs);
 *      // Handle the update request here
 * 
 *      return EXIT_SUCCESS;
 *   }
 */
GlobScop int IAL_Ctl_DecodeRequest_StartUpdate(const void* vptr_Data, uint32_t u32_DataLength, char * sz_SourceKey, 
                                                size_t siz_MaxSourceKey, uint32_t * ptr_u32_Type, char * sz_ArtifactPath, 
                                                size_t siz_MaxArtifactPath, char ** arr_sz_Keys, size_t siz_Key, 
                                                char ** arr_sz_Values, size_t siz_Value, size_t siz_MaxPairs, 
                                                size_t *ptr_siz_Pairs);

#ifdef __cplusplus
}
#endif

#endif /* LION_IAL_CONTROL_H__ */
//EOF
