/**********************************************************************************************************************
* @file            IAL_Data.h
*
* @brief           Provides APIs for data related ICC functions and events for the EMS.
*                  
*
* @author          Devan Hone
* @date            07/25/2024
* @copyright       2024 Lion Energy LLC.
*                  All rights reserved
*
*********************************************************************************************************************/
/**
* @addtogroup      IAL
*                  Identifier Description for the ICC Abstraction Library (IAL) Module  
* 
* @ref             IAL.h 
*                  
*********************************************************************************************************************/
#include "COD_Compiler_Defines.h"

#ifndef LION_IAL_DATA_H__
#define LION_IAL_DATA_H__

/*********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

#include <lion/ial/IAL.h>

#ifdef __cplusplus
extern "C"
{
#endif

#define IAL_DATA_u8_ALERT_BUFFER_SIZE   128
#define IAL_DATA_sz_TOU_KEY             "tou_prf"

/**
 * @brief  Retreieves data on the EMS from the Domain Controller Unit (DCU) using the Inter-Controller Communication
 *         (ICC) Service. Expects to only receive keys supported by the Data Source Type, and to only request for one
 *         data source or device per request. 
 * 
 * @param enm_Target        The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param sz_UnitType       Unit type to retrieve data from, eg. "pcs", "pdu", "dcdc", "pmu", "hvac", "gw"
 * @param u8_UnitIndex      Unit index to retrieve data from, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param u8_NumberOfKeys   Size of the array of requested Keys
 * @param arr_sz_Keys       Requested keys as an array of strings
 * @param arr_s32_Responses Empty array of size (u8_NumberOfKeys) to be filled with retrieved values corresponding 
 *                          the arr_sz_Keys 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_GetQuantifiableData(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                            uint8_t u8_NumberOfKeys, const char **arr_sz_Keys, int32_t *arr_s32_Responses);

/**
 * @brief  Sets parameters of the EMS from the Domain Controller Unit (DCU) using the Inter-Controller Communication
 *         (ICC) Service. Expects to only receive keys supported by the Data Source Type, and to only request for one
 *         data source or device per request. 
 * 
 * @param enm_Target        The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param sz_UnitType       Unit type data to set for, eg. "pcs", "pdu", "dcdc", "pmu", "hvac", "gw"
 * @param u8_UnitIndex      Unit index to set data for, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param u8_NumberOfKeys   Size of the array of requested Keys
 * @param arr_sz_Keys       Requested keys or settings to be written as an array of strings
 * @param s32_arr_Values    Requested values of settings to be written as an array of strings
 * @param arr_int_Statuses  Empty array of size (u8_NumberOfKeys) to be filled with the status to writing the requested
 *                          value to the device corresponding to the arr_sz_Keys (0 upon success)
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_SetQuantifiableData(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex, 
                                            uint8_t u8_NumberOfKeys, const char **arr_sz_Keys, int32_t *s32_arr_Values,
                                            int *arr_int_Statuses);
/**
 * @brief  Retreieves device information on the EMS from the Domain Controller Unit (DCU) using the Inter-Controller 
 *         Communication (ICC) Service. Expects to only receive keys supported by the Data Source Type, and to only 
 *         request for one data source or device per request. 
 * 
 * @param enm_Target        The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param sz_UnitType       Unit type to retrieve data from, eg. "pcs", "pdu", "dcdc", "pmu", "hvac", "gw"
 * @param u8_UnitIndex      Unit index to retrieve data from, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param u8_NumberOfKeys   Size of the array of requested Keys
 * @param arr_sz_Keys       Requested keys as an array of strings
 * @param arr_sz_Responses  Empty array of size (u8_NumberOfKeys) to be filled with retrieved values corresponding 
 *                          the arr_sz_Keys 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_GetUnitInfo(IAL_TargetDevice_t enm_Target, const char* sz_UnitType, uint8_t u8_UnitIndex,
                                    uint8_t u8_NumberOfKeys, const char** arr_sz_Keys, char** arr_sz_Responses);

/**
 * @brief  Retreieves Fault, Alarm, and Status "Code" data on the EMS from the Domain Controller Unit (DCU) using the 
 *         Inter-Controller Communication (ICC) Service. Expects to only receive keys supported by the Data Source Type,
 *         and to only request for one data source or device per request. 
 * 
 * @param enm_Target        The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param sz_UnitType       Unit type to retrieve data from, eg. "pcs", "pdu", "dcdc", "pmu", "hvac", "gw"
 * @param u8_UnitIndex      Unit index to retrieve data from, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param sz_CodeType       "fault", "alarm", or "status"
 * @param arr_sz_Responses  Empty array of size (u8_NumberOfKeys) to be filled with retrieved values 
 *                          corresponding the sz_arr_Keys 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_GetUnitCodes(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex, 
                                    const char *sz_CodeType, char **arr_sz_Responses);

/**
 * @brief Retrieves bufferred component data from the DCU (Domain Controller Unit) 
 *        using the Inter-Controller Communication (ICC) Service.
 * 
 * @param enm_Target        The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param sz_UnitType       Unit type to retrieve data from, eg. "pcs", "pdu", "dcdc", "pmu", "hvac", "gw"
 * @param u8_UnitIndex      Unit index to retrieve data from, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param vptr_Buff         Destination buffer to store the data
 * @param ptr_u32_BuffSize  Size of the destination buffer. Also, when the function returns 
 *                          this parameter will be set to the size of the data retrieved.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
*/
GlobScop int IAL_Data_GetBufferedData(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex, 
                                        void *vptr_Buff, uint32_t *ptr_u32_BuffSize);

/**
 * @brief  Retreieves Configuration Settings on the EMS from the Domain Controller Unit (DCU) using the Inter-Controller
 *         Communication (ICC) Service.
 * 
 * @param enm_Target        The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param sz_UnitType       Unit type to retrieve data from, eg. "pcs", "pdu", "dcdc", "pmu", "hvac", "gw"
 * @param u8_UnitIndex      Unit index to retrieve data from, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param u8_NumberOfKeys   Size of the array of requested Keys
 * @param arr_sz_Keys       Requested keys as an array of strings
 * @param arr_sz_Responses  Empty array of size (u8_NumberOfKeys) to be filled with retrieved values corresponding 
 *                          the arr_sz_Keys 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_GetEMSConfiguration(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                            uint8_t u8_NumberOfKeys, const char** arr_sz_Keys, char ** arr_sz_Responses);

/**
 * @brief  Sets Configuration Settings of the EMS from the Domain Controller Unit (DCU) using the Inter-Controller
 *         Communication (ICC) Service. 
 * 
 * @param enm_Target        The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param sz_UnitType       Configuration Type ("pcs", "pdu", "dcdc", "pmu", "hvac", "gw")
 * @param u8_UnitIndex      Unit ID (when applicable, otherwise send COD_u8_NULL_UNIT_ID)
 * @param u8_NumberOfKeys   Size of the array of requested Keys
 * @param arr_sz_Keys       Requested keys or settings to be written as an array of strings
 * @param arr_sz_Values     Requested values of settings to be written as an array of strings
 * @param arr_int_Statuses  Empty array of size (u8_NumberOfKeys) to be filled with the status to writing the requested
 *                          value to the device corresponding to the arr_sz_Keys (0 upon success)
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_SetEMSConfiguration(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                            uint8_t u8_NumberOfKeys, const char **arr_sz_Keys,
                                            const char **arr_sz_Values, int *arr_int_Statuses);

/**
 * @brief  Retreieves Configuration Settings of the various Units/Components from the Domain Controller Unit (DCU)
 *         using the Inter-Controller Communication (ICC) Service.
 * 
 * @param enm_Target        The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param sz_UnitType       Unit type to retrieve data from, eg. "pcs", "pdu", "dcdc", "pmu", "hvac", "gw"
 * @param u8_UnitIndex      Unit index to retrieve data from, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param u8_NumberOfKeys   Size of the array of requested Keys
 * @param arr_sz_Keys       Requested keys as an array of strings
 * @param arr_sz_Responses  Empty array of size (u8_NumberOfKeys) to be filled with retrieved values corresponding 
 *                          the arr_sz_Keys 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_GetUnitConfiguration(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                            uint8_t u8_NumberOfKeys, const char** arr_sz_Keys, char ** arr_sz_Responses);

/**
 * @brief  Sets Configuration Settings of the various Units/Components from the Domain Controller Unit (DCU) using the 
 *         Inter-Controller Communication (ICC) Service. 
 * 
 * @param enm_Target        The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param sz_UnitType       Unit type data to set for, eg. "pcs", "pdu", "dcdc", "pmu", "hvac", "gw"
 * @param u8_UnitIndex      Unit index to set data for, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param u8_NumberOfKeys   Size of the array of requested Keys
 * @param arr_sz_Keys       Requested keys or settings to be written as an array of strings
 * @param arr_sz_Values     Requested values of settings to be written as an array of strings
 * @param arr_int_Statuses  Empty array of size (u8_NumberOfKeys) to be filled with the status to writing the requested
 *                          value to the device corresponding to the arr_sz_Keys (0 upon success)
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_SetUnitConfiguration(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                            uint8_t u8_NumberOfKeys, const char **arr_sz_Keys,
                                            const char **arr_sz_Values, int *arr_int_Statuses);

/**
 * @brief  Retreieves Time of Use Settings on the EMS from the Domain Controller Unit (DCU) using the Inter-Controller 
 *          Communication (ICC) Service.
 * 
 * @param enm_Target            The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param u8_ProfileID          Profile ID
 * @param sz_SettingPath        Name of the desired setting to get
 * @param u8_SettingDepth       Depth of Setting Path (0 indexed)
 * @param sz_Response           Retrieved setting value (either a value or an serialized JSON Array)
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_GetTouSettings(IAL_TargetDevice_t enm_Target, uint8_t u8_ProfileID, const char **sz_SettingPath,
                                        uint8_t u8_SettingDepth, char *sz_Response);

/**
 * @brief  Sets Time of Use Settings of the EMS from the Domain Controller Unit (DCU) using the Inter-Controller
 *         Communication (ICC) Service. 
 * @param enm_Target            The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param u8_ProfileID          Profile ID
 * @param sz_SerializedSettings Serailized JSON Array of settings to be configured
 * @param sz_ArrayName          Name of the array serialized in the previous argument
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_SetTouSettings(IAL_TargetDevice_t enm_Target, uint8_t u8_ProfileID,
                                        const char *sz_SerializedSettings, const char *sz_ArrayName);

/**
 * @brief Test function used to validate the send alert status function via ICC commands
 * 
 * @param enm_Target        The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param sz_UnitType Unit type
 * @param u8_UnitIndex Unit index
 *
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_TestAlertStatusCallback(IAL_TargetDevice_t enm_Target, const char *sz_UnitType,
                                                uint8_t u8_UnitIndex);

/**
 * @brief  Gets the I3E Indicators from the I3E application.
 * 
 * @param enm_Target The target device for the ICC call, eg. IAL_TARGET_CCU, IAL_TARGET_COMP.
 * @param sz_LongForm Pointer to a buffer to store the long-form of the I3E Indicators.
 * @param siz_LongForm Size of the long-form buffer.
 * @param sz_ShortForm Pointer to a buffer to store the short-form of the I3E Indicators.
 * @param siz_ShortForm Size of the short-form buffer.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 */
GlobScop int IAL_Data_GetI3EIndicators(IAL_TargetDevice_t enm_Target, char *sz_LongForm, size_t siz_LongForm,
                                        char *sz_ShortForm, size_t siz_ShortForm);

/**
 * @brief A helper function to encode the response for calls to get the I3E Indicators.
 * 
 * @details This function is intended to be called inside a callback function for the @ref IAL_GET_I3E_INDICATORS_ID
 *          function ID. It takes a long-form and short-form string of I3E Indicators and encodes them into the
 *          response buffer.
 * 
 * @param sz_LongForm The long-form string of I3E Indicators to encode.
 * @param sz_ShortForm The short-form string of I3E Indicators to encode.
 * 
 * @param vptr_Response The buffer to encode the response into.
 * @param ptr_siz_Response A pointer to store the size of the encoded response.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 * 
 * @example
 *      //Get Indicators callback function
 *      //This function will need to be registered with the IAL/ICC using IAL_RegisterCallback().
 *      //calculateIndicators() is a placeholder function that would be implemented in the I3E application.
 *      int GetIndicators_CB(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength, void *vptr_Response, 
 *                              uint32_t *ptr_u32_ResponseLength, void *vptr_UserData)
 *      {
 *          UNUSED(u32_ID);
 *          UNUSED(vptr_Data);
 *          UNUSED(u32_DataLength);
 *          char sz_LongForm[64];
 *          char sz_ShortForm[32];
 *          calculateIndicators(sz_LongForm, sz_ShortForm);
 *          return IAL_EncodeResponse_GetI3EIndicators(sz_LongForm, sz_ShortForm, vptr_Response, ptr_u32_ResponseLength);
 *      }
 */
GlobScop int IAL_EncodeResponse_GetI3EIndicators(const char *sz_LongForm, const char *sz_ShortForm,
                                                    void *vptr_Response, size_t *ptr_siz_Response);

/**
 * @brief  Reports the status of any curreny update
 *
 * @param sz_SourceKey The destination key of the update status.
 * @param u32_Type The type of update.
 * @param u32_State The state of the update.
 * @param arr_sz_Keys An array of meta-data keys.
 * @param arr_sz_Values An array of meta-data values.
 * @param siz_Pairs The number of key-value pairs in the update status.
 * @param u16_Id The ID to identify the update status.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 */
GlobScop int IAL_Data_UpdateStatus(const char * sz_SourceKey, uint32_t u32_Type, uint32_t u32_State, 
                                    const char ** arr_sz_Keys, const char ** arr_sz_Values, 
                                    size_t siz_Pairs, uint16_t u16_Id);

/**
 * @brief A helper function to decode the payload of the update status message.
 * 
 * @details This function is intended to be called inside a callback function for the @ref IAL_UPDATE_STATUS_ID
 * 
 * @param vptr_Data Pointer to the CBOR encoded data.
 * @param u32_DataLength Length of the CBOR encoded data.
 * @param sz_SourceKey Pointer to a buffer to store the source key.
 * @param siz_MaxSourceKey Size of the buffer to store the source key.
 * @param ptr_u32_Type Pointer to store the type of update status.
 * @param ptr_u32_State Pointer to store the state of the update status.
 * @param arr_sz_Keys Pointer to an array of keys for the update status.
 * @param siz_MaxKey Size of each key buffer in the array.
 * @param arr_sz_Values Pointer to an array of values for the update status.
 * @param siz_MaxValue Size of each value buffer in the array.
 * @param siz_MaxPairs The maximum number of key-value pairs that can be decoded.
 * @param ptr_siz_Pairs Pointer to store the number of key-value pairs in the update status.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 * 
 * @example
 *     //UpdateStatus callback function
 *     //This function will need to be registered with the IAL/ICC using IAL_RegisterCallback().
 *     int UpdateStatus_CB(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength, void *vptr_Response,
 *                         uint32_t *ptr_u32_ResponseLength, void *vptr_UserData)
 *     {
 *       UNUSED(u32_ID);
 *       UNUSED(vptr_Response);
 *       UNUSED(ptr_u32_ResponseLength);
 *       char sz_SerializedData[TEM_siz_MAX_STATUS];
 *       size_t siz_Pairs = 0;
 *       uint16_t u16_MessageId;  
 *       char sz_SourceKey[TEM_siz_MAX_KEY_LEN];
 *       uint32_t u32_Type;
 *       uint32_t u32_State;
 *       char arr_sz_Keys[TEM_siz_MAX_PAIRS][TEM_siz_MAX_KEY_LEN];
 *       char arr_sz_Values[TEM_siz_MAX_PAIRS][TEM_siz_MAX_VALUE_LEN];
 *       for (size_t i = 0; i < TEM_siz_MAX_PAIRS; i++) 
 *       {
 *          arr_ptr_sz_Keys[i] = arr_sz_Keys[i];
 *          arr_ptr_sz_Values[i] = arr_sz_Values[i];
 *       }
 * 
 *       int int_RC = IAL_Data_DecodeRequest_UpdateStatus(vptr_Data, u32_DataLength, sz_SourceKey, sizeof(sz_SourceKey),
 *                                                  &u32_Type, &u32_State, arr_ptr_sz_Keys, sizeof(arr_sz_Keys[0]), arr_ptr_sz_Values,                         
 *                                                 sizeof(arr_sz_Values[0]), TEM_siz_MAX_PAIRS, &siz_Pairs,
 *                                                &u16_MessageId);
 * 
 *       // Handle the status here
 * 
 *       IAL_ClearUpdateStatus(IAL_TARGET_COMP, u16_Id);
 * 
 *       return EXIT_SUCCESS;
 *     }
 */
GlobScop int IAL_Data_DecodeRequest_UpdateStatus(const uint8_t * ptr_u8_Data, uint32_t u32_DataLength, 
                                                    char * sz_SourceKey, size_t siz_MaxSourceKey,
                                                    uint32_t * ptr_u32_Type, uint32_t * ptr_u32_State, 
                                                    char ** arr_sz_Keys, size_t siz_MaxKey, char ** arr_sz_Values, 
                                                    size_t siz_MaxValue, size_t siz_MaxPairs, size_t * ptr_siz_Pairs, 
                                                    uint16_t * ptr_u16_Id);

/**
 * @brief  Clears the update status message of the specified ID from the UMA application.
 * 
 * @param enm_Target The target device for the ICC call, eg. IAL_TARGET_CCU, IAL_TARGET_COMP.
 * @param u16_Id The ID of the update status to clear
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 */
GlobScop int IAL_Data_ClearUpdateStatus(IAL_TargetDevice_t enm_Target, uint16_t u16_Id);

/**
 * @brief  A helper function to decode the payload of the clear update status message.
 *
 * @details This function is intended to be called inside a callback function for the @ref IAL_CLEAR_UPDATE_STATUS_ID
 * 
 * @param vptr_Data Pointer to the CBOR encoded data.
 * @param u32_DataLength Length of the CBOR encoded data.
 * @param u16_Id Pointer to store the ID of the update status. 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 * 
 * @example
 *    //ClearUpdateStatus callback function
 *    //This function will need to be registered with the IAL/ICC using IAL_RegisterCallback().
 *    //clearRequest() is a placeholder function that would be implemented in the UMA application.
 *    int ClearUpdateStatus_CB(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength, void *vptr_Response,
 *                            uint32_t *ptr_u32_ResponseLength, void *vptr_UserData)
 *    {
 *      UNUSED(u32_ID);
 *      UNUSED(vptr_Response);
 *      UNUSED(ptr_u32_ResponseLength);
 *      uint16_t u16_Id;
 *      int int_RC = IAL_Data_DecodeRequest_ClearUpdateStatus(vptr_Data, u32_DataLength, &u16_Id);
 *      
 *      clearRequest(u16_Id);
 *      
 *      return EXIT_SUCCESS;
 *    }
 */
GlobScop int IAL_Data_DecodeRequest_ClearUpdateStatus(const uint8_t* ptr_u8_Data, uint32_t u32_DataLength, 
                                                        uint16_t * ptr_u16_Id);

/**
 * @brief  Sends the version information to the desired unit using the Inter-Controller Communication (ICC) Service.
 * 
 * @param enm_Target The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU.
 * @param u8_NumKeys The number of keys to set.
 * @param arr_sz_Keys An array of keys to set.
 * @param arr_sz_Values An array of values to set.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 * 
 * @example
 *   //getVersion() is a placeholder for a function that would get the version information.
 *   int SetVersionInfo()
 *   {
 *      const char * sz_Key = "tem_app";
 *      char sz_Version[TEM_siz_MAX_VERSION];
 *      getVersion(sz_Version);
 *      return IAL_DATA_PublishVersionInfo(IAL_TARGET_COMP, 1, &sz_Key, &sz_Version);
 *   }
 */
GlobScop int IAL_Data_PublishVersionInfo(IAL_TargetDevice_t enm_Target, uint8_t u8_NumKeys, const char **arr_sz_Keys, 
                                            const char **arr_sz_Values);
/**
 * @brief  A helper function to decode the payload of the version information message.
 * 
 * @details This function is intended to be called inside a callback function for the @ref IAL_VERSION_INFO_ID. The 
 *          parsed values stored in the provided arrays reference their location within the CBOR encoded data. This
 *          datas scope is limited to the scope of CBOR encoded data.
 * 
 * @param ptr_u8_Data Pointer to the CBOR encoded data.
 * @param u32_DataLength Length of the CBOR encoded data.
 * @param arr_sz_Keys An array of keys to store the parsed keys.
 * @param arr_sz_Values An array of values to store the parsed values.
 * @param u8_MaxKeys The maximum number of keys that can be stored in the provided arrays.
 * @param ptr_u8_NumKeys Pointer to store the number of keys parsed.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 * 
 * @example
 *   //VersionInfo_CB() is an example callback function
 *   //This function will need to be registered with the IAL/ICC using IAL_RegisterCallback().
 *   //versionUpdate() is a placeholder function to handle the version update
 *   int VersionInfo_CB(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength, void *vptr_Response,
 *   {
 *      UNUSED(u32_ID);
 *      UNUSED(vptr_Response);
 *      UNUSED(u32_DataLength);
 * 
 *      const char * arr_sz_Keys[TEM_siz_MAX_PAIRS];
 *      const char * arr_sz_Values[TEM_siz_MAX_PAIRS];
 *      uint8_t u8_NumKeys;
 *      int int_RC = IAL_DATA_DecodeResponse_PublishVersionInfo(vptr_Data, u32_DataLength, arr_sz_Keys, arr_sz_Values,
 *                                                             TEM_siz_MAX_PAIRS, &u8_NumKeys);
 *     if (int_RC == EXIT_SUCCESS)
 *     {
 *        versionUpdate(arr_sz_Keys, arr_sz_Values, u8_NumKeys);
 *     }
 *     else
 *     {
 *       //Handle error
 *       return -EXIT_FAILURE;
 *     }
 * 
 *     return EXIT_SUCCESS;
 *   }
 */
GlobScop int IAL_Data_DecodeRequest_PublishVersionInfo(const uint8_t* ptr_u8_Data, uint32_t u32_DataLength, 
                                                        const char **arr_sz_Keys, const char **arr_sz_Values, 
                                                        uint8_t u8_MaxKeys, uint8_t* ptr_u8_NumKeys);

#ifdef __cplusplus
}
#endif

#endif /* LION_IAL_DATA_H__ */
//EOF
