/**********************************************************************************************************************
* @file            IAL_Modbus.h
*
* @brief           Provides APIs for Modbus direct related ICC functions and events for the EMS.
*                  
*
* @author          Devan Hone
* @date            07/25/2024
* @copyright       2024 Lion Energy LLC.
*                  All rights reserved
*
*********************************************************************************************************************/
/**
* @addtogroup      IAL
*                  Identifier Description for the ICC Abstraction Library (IAL) Module  
* 
* @ref             IAL.h 
*                  
*********************************************************************************************************************/
#include "COD_Compiler_Defines.h"

#ifndef LION_IAL_MODBUS_H__
#define LION_IAL_MODBUS_H__

/*********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

#include <lion/ial/IAL.h>

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @brief   Reads the specified array of register addresses from the modbus device across the ICC.
 * @param   enm_Target         The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param   sz_UnitType        Unit type to read from, eg. "pcs", "dcdc"
 * @param   u8_UnitIndex       Unit index to read from, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param   arr_u16_Addresses  An array of register addresses to be read.
 * @param   arr_u16_Values     An array of register values to be populated.
 * @param   siz_Request        The number of register addresses to be read.
 * @returns EXIT_SUCCESS or 0 upon success, standard posix error code on failure.
*/
GlobScop int IAL_Mbs_ReadRegisters(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                     uint16_t *arr_u16_Addresses, uint16_t *arr_u16_Values, size_t siz_Request);

/**
 * @brief   Reads a batch of sequential registers from the start address from the modbus device across the ICC for the specified 
 *          size of the request.
 * @param   enm_Target         The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param   sz_UnitType        Unit type to read from, eg. "pcs", "dcdc"
 * @param   u8_UnitIndex       Unit index to read from, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param   u16_startAddress   The starting register address to be read.
 * @param   arr_u16_Values     An array of register values to be populated.
 * @param   siz_Request        The number of register addresses to be read.
 * @returns EXIT_SUCCESS or 0 upon success, standard posix error code on failure.
*/
GlobScop int IAL_Mbs_ReadMultipleRegisters(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                             uint16_t u16_startAddress, uint16_t *arr_u16_Values, size_t siz_Request);

/**
 * @brief   Writes the array of values to the specified array of register addresses to the modbus device across the ICC.
 * @param   enm_Target         The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param   sz_UnitType        Unit type to write to, eg. "pcs", "dcdc"
 * @param   u8_UnitIndex       Unit index to write to, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param   arr_u16_Addresses  An array of register addresses to be written.
 * @param   arr_u16_Values     An array of register values to be written.
 * @param   siz_Request        The number of register addresses to be written.
 * @returns EXIT_SUCCESS or 0 upon success, standard posix error code on failure.
*/
GlobScop int IAL_Mbs_WriteRegisters(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                    uint16_t *arr_u16_Addresses, uint16_t *arr_u16_Values, int32_t *arr_s32_results,
                                    size_t siz_Request);

/**
 * @brief   Writes the array of values to the specified array of register addresses to the modbus device across the 
 *          ICC from a specified source.
 * 
 * @details This function is similar to IAL_Mbs_WriteRegisters, but it allows specifying a source for the write
 *          operation. This is useful for tracking the origin of the write request, especially for I3E operations when
 *          Modbus registers are locked for writing.
 * 
 * @param   enm_Target         The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param   sz_Source          The source of the write operation (acronym of the write module, e.g., "CCA" or "CST").
 * @param   sz_UnitType        Unit type to write to, eg. "pcs", "dcdc"
 * @param   u8_UnitIndex       Unit index to write to, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param   arr_u16_Addresses  An array of register addresses to be written.
 * @param   arr_u16_Values     An array of register values to be written.
 * @param   arr_s32_results    An array to store the results of the write operation for each register.
 * @param   siz_Request        The number of register addresses to be written.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix error code on failure.
 */
GlobScop int IAL_Mbs_WriteRegisters_Sourced(IAL_TargetDevice_t enm_Target, const char *sz_Source, const char *sz_UnitType,
                                            uint8_t u8_UnitIndex, uint16_t *arr_u16_Addresses, uint16_t *arr_u16_Values,
                                            int32_t *arr_s32_results, size_t siz_Request);

/**
 * @brief   Writes a batch of sequential registers from the start address to the modbus device across the ICC for the specified
 *          size of the request.
 * @param   enm_Target         The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param   sz_UnitType        Unit type to write to, eg. "pcs", "dcdc"
 * @param   u8_UnitIndex       Unit index to write to, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param   u16_startAddress   The starting register address to be written.
 * @param   arr_u16_Values     An array of register values to be written.
 * @param   siz_Request        The number of register addresses to be written.
 * @returns EXIT_SUCCESS or 0 upon success, standard posix error code on failure.

*/
GlobScop int IAL_Mbs_WriteMultipleRegisters(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                            uint16_t u16_startAddress, uint16_t *arr_u16_Values, size_t siz_Request);

/**
 * @brief   Writes a batch of sequential registers from the start address to the modbus device across the ICC for
 *          the specified size of the request from a specified source.
 * 
 * @details This function is similar to IAL_Mbs_WriteMultipleRegisters, but it allows specifying a source for the write
 *          operation. This is useful for tracking the origin of the write request, especially for I3E operations when
 *          Modbus registers are locked for writing.
 * 
 * @param   enm_Target         The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param   sz_Source          The source of the write operation (acronym of the write module, e.g., "CCA" or "CST").
 * @param   sz_UnitType        Unit type to write to, eg. "pcs", "dcdc"
 * @param   u8_UnitIndex       Unit index to write to, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param   u16_startAddress   The starting register address to be written.
 * @param   arr_u16_Values     An array of register values to be written.
 * @param   siz_Request        The number of register addresses to be written.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix error code on failure.
 */
GlobScop int IAL_Mbs_WriteMultipleRegisters_Sourced(IAL_TargetDevice_t enm_Target, const char *sz_Source,
                                                    const char *sz_UnitType, uint8_t u8_UnitIndex,
                                                    uint16_t u16_startAddress, uint16_t *arr_u16_Values, size_t siz_Request);

/**
 * @brief Lock out writes to a list of registers for the specified modbus device. Response contains a list
 *        of results of the lock operation for each register.
 * 
 * @param enm_Target         The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param sz_UnitType        Unit type to lock registers for, eg. "pcs", "dcdc"
 * @param u8_UnitIndex       Unit index to lock registers for, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param arr_u16_Addresses  An array of register addresses to be locked.
 * @param arr_s32_results    An array to store the results of the lock operation for each register.
 * @param siz_Request        The number of register addresses to be locked.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix error code on failure.
 */
GlobScop int IAL_Mbs_LockRegisters(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                    uint16_t *arr_u16_Addresses, int32_t *arr_s32_results, size_t siz_Request);

/**
 * @brief Unlock writing to a list of registers for the specified modbus device. Response contains a list 
 *        of results of the unlock operation for each register.
 * 
 * @param enm_Target         The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param sz_UnitType        Unit type to unlock registers for, eg. "pcs", "dcdc"
 * @param u8_UnitIndex       Unit index to unlock registers for, [0, UNIT_MAX_INDEX], 255 (Null)
 * @param arr_u16_Addresses  An array of register addresses to be unlocked.
 * @param arr_s32_results    An array to store the results of the unlock operation for each register.
 * @param siz_Request        The number of register addresses to be unlocked.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix error code on failure.
 */
GlobScop int IAL_Mbs_UnlockRegisters(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                        uint16_t *arr_u16_Addresses, int32_t *arr_s32_results, size_t siz_Request);

/**
 * @brief Reset the lock registry for the specified modbus device.
 * 
 * @details In the event of a reset on the locking device, it is necessary to reset the entire lock registry as there is
 *          no way to determine which registers were locked previously.
 * 
 * @param enm_Target         The target device for the ICC call, eg. IAL_TARGET_COMP, IAL_TARGET_DCU
 * @param sz_UnitType        Unit type to reset the lock registry for, eg. "pcs", "dcdc"
 * @param u8_UnitIndex       Unit index to reset the lock registry for, [0, UNIT_MAX_INDEX], 255 (Null)
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix error code on failure.
 */
GlobScop int IAL_Mbs_ResetLockRegistry(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex);

#ifdef __cplusplus
}
#endif

#endif /* LION_IAL_MODBUS_H__ */
//EOF
