#include <lion/ial/ems/IAL_Data.h>
#include <lion/log/log.h>
#if CONFIG_EMS_SCH
#include <lion/sch/SCH_AtsSchema.h>
#include <lion/sch/SCH_DcdcSchema.h>
#include <lion/sch/SCH_EmsSchema.h>
#include <lion/sch/SCH_GwSchema.h>
#include <lion/sch/SCH_HvacSchema.h>
#include <lion/sch/SCH_PcsSchema.h>
#include <lion/sch/SCH_PduSchema.h>
#include <lion/sch/SCH_PmuSchema.h>
#endif
#include <ArduinoJson.h>
#include <zcbor_decode.h>
#include <zcbor_encode.h>

LOG_MODULE_REGISTER(ial_data, LOG_LEVEL_INF);

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

#define ial_sz_I3E_LF_IND_KEY "long" //IEEE 2030.5 long-form indicator key
#define ial_sz_I3E_SF_IND_KEY "short" //IEEE 2030.5 short-form indicator key

/**
 * @brief Gives the size of the buffer required to store data structs of the given unit type.
 * 
 * @param[in] sz_UnitType The unit type to get the size of the data struct for.
 * 
 * @return The size of the data struct for the given unit type.
*/
FileScop size_t ial_data_BufferedDataSize(const char *sz_UnitType);


/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

ModScop extern ICC_Context_t *ial_ptr_IccCtx; //Defined in IAL.c

/*********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

GlobScop int IAL_Data_GetQuantifiableData(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                            uint8_t u8_NumberOfKeys, const char **arr_sz_Keys, int32_t *arr_s32_Responses)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_GET_QUANTIFIABLE_DATA_ID);

    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    zcbor_list_start_encode(CBORState, u8_NumberOfKeys);
    for (uint8_t i = 0; i < u8_NumberOfKeys; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Keys[i], strlen(arr_sz_Keys[i]) + 1);
    }
    zcbor_list_end_encode(CBORState, u8_NumberOfKeys);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    uint8_t arr_s32_EncodedResponses[2 + u8_NumberOfKeys * 5] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\", Attempting Retry", int_Rc, strerror(-int_Rc));
        // Try again
        u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
        int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
        if (int_Rc < 0) {
            LOG_INF("ICC_Function_Call Retry Failed with Error %d \"%s\"", int_Rc, strerror(-int_Rc));
            return int_Rc;
        }
    }
    

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, arr_s32_EncodedResponses, u32_ResponseLength, 3, 0);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        for (uint32_t i = 0; i < u8_NumberOfKeys; i++)
        {
            zcbor_int32_decode(CBORDecodeState, &arr_s32_Responses[i]);
        }
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_INF("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_SetQuantifiableData(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex, 
                                            uint8_t u8_NumberOfKeys, const char **arr_sz_Keys, int32_t *s32_arr_Values,
                                            int *arr_int_Statuses)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_SET_QUANTIFIABLE_DATA_ID);

    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    const uint8_t u8_MAP_WIDTH = 2; // Keys | Values
    zcbor_map_start_encode(CBORState, u8_NumberOfKeys*u8_MAP_WIDTH);
    for (uint8_t i = 0; i < u8_NumberOfKeys; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Keys[i], strlen(arr_sz_Keys[i]) + 1);
        zcbor_int32_put(CBORState, s32_arr_Values[i]);
    }
    zcbor_map_end_encode(CBORState, u8_NumberOfKeys*u8_MAP_WIDTH);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    FileScop uint8_t u8_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(u8_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    u8_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, u8_EncodedResponses, u32_ResponseLength, 3, 0);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        for (uint32_t i = 0; i < u8_NumberOfKeys; i++)
        {
            zcbor_int32_decode(CBORDecodeState, &arr_int_Statuses[i]);
        }
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_INF("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_GetUnitInfo(IAL_TargetDevice_t enm_Target, const char* sz_UnitType, uint8_t u8_UnitIndex,
                                    uint8_t u8_NumberOfKeys, const char** arr_sz_Keys, char** arr_sz_Responses)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_GET_UNIT_DATA_ID);

    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    zcbor_list_start_encode(CBORState, u8_NumberOfKeys);
    for (uint8_t i = 0; i < u8_NumberOfKeys; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Keys[i], strlen(arr_sz_Keys[i]) + 1);
    }
    zcbor_list_end_encode(CBORState, u8_NumberOfKeys);
    
    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    uint8_t u8_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(u8_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    u8_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, u8_EncodedResponses, u32_ResponseLength, 3, 0);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        struct zcbor_string sz_zcborKey; 
        for (uint32_t i = 0; zcbor_bstr_decode(CBORDecodeState, &sz_zcborKey); i++)
        {
            arr_sz_Responses[i] = (char *)sz_zcborKey.value;
        }
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_INF("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_GetUnitCodes(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex, 
                                    const char *sz_CodeType, char **arr_sz_Responses)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_GET_UNIT_CODES_ID);

    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    zcbor_tstr_encode_ptr(CBORState, sz_CodeType, strlen(sz_CodeType)+1);
    
    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    FileScop uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, arr_s32_EncodedResponses, u32_ResponseLength, 3, 0);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding EMS Codes CBOR Error");
    }
    else 
    {
        struct zcbor_string sz_zcborKey; 
        for (uint32_t i = 0; zcbor_tstr_decode(CBORDecodeState, &sz_zcborKey); i++)
        {
            arr_sz_Responses[i] = (char *)sz_zcborKey.value;
        }
        zcbor_list_end_decode(CBORDecodeState);
    }

    LOG_INF("Call for %s %d %s codes: %d, %u bytes", sz_UnitType, u8_UnitIndex, sz_CodeType, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_GetBufferedData(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex, 
                                        void *vptr_Buff, uint32_t *ptr_u32_BuffSize)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_GET_BUFFERED_DATA);

    //Ensure buffer can hold the data structure
    if (*ptr_u32_BuffSize < ial_data_BufferedDataSize(sz_UnitType)) {
        LOG_ERR("Destination buffer is too small for %s data", sz_UnitType);
        return -ENOMEM;
    }

    //Build the request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    
    //Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    vptr_Buff, ptr_u32_BuffSize, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    //No need to decode response as the data is just the serialized data structure for the unit.
    
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }
    return 0;
}

GlobScop int IAL_Data_GetEMSConfiguration(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                            uint8_t u8_NumberOfKeys, const char** arr_sz_Keys, char ** arr_sz_Responses)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_GET_EMS_CONFIG_ID);

    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));     
    }
    zcbor_list_start_encode(CBORState, u8_NumberOfKeys);
    for (uint8_t i = 0; i < u8_NumberOfKeys; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Keys[i], strlen(arr_sz_Keys[i]) + 1);
    }
    zcbor_list_end_encode(CBORState, u8_NumberOfKeys);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    uint8_t u8_EncodedResponses[2 + u8_NumberOfKeys * 17] = {0};
    uint32_t u32_ResponseLength = sizeof(u8_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    u8_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CONFIG_CALL_TIMEOUT_S));
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\", Attempting Retry", int_Rc, strerror(-int_Rc));
        // Try again
        u32_ResponseLength = sizeof(u8_EncodedResponses);
        int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    u8_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CONFIG_CALL_TIMEOUT_S));
        if (int_Rc < 0) {
            LOG_INF("ICC_Function_Call Retry Failed with Error %d \"%s\"", int_Rc, strerror(-int_Rc));
            return int_Rc;
        }
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, u8_EncodedResponses, u32_ResponseLength, 3, 0);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        struct zcbor_string sz_zcborKey; 
        for (uint32_t i = 0; zcbor_tstr_decode(CBORDecodeState, &sz_zcborKey); i++)
        {
            arr_sz_Responses[i] = (char *)sz_zcborKey.value;
        }
        zcbor_list_end_decode(CBORDecodeState);
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_INF("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_SetEMSConfiguration(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                            uint8_t u8_NumberOfKeys, const char **arr_sz_Keys,
                                            const char **arr_sz_Values, int *arr_int_Statuses) 
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_SET_EMS_CONFIG_ID);

    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    const uint8_t u8_MAP_WIDTH = 2; // Keys | Values
    zcbor_map_start_encode(CBORState, u8_NumberOfKeys*u8_MAP_WIDTH);
    for (uint8_t i = 0; i < u8_NumberOfKeys; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Keys[i], strlen(arr_sz_Keys[i]) + 1);
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Values[i], strlen(arr_sz_Values[i]) + 1);
    }
    zcbor_map_end_encode(CBORState, u8_NumberOfKeys*u8_MAP_WIDTH);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    FileScop uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CONFIG_CALL_TIMEOUT_S));
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, arr_s32_EncodedResponses, u32_ResponseLength, 3, 0);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        for (uint32_t i = 0; i < u8_NumberOfKeys; i++)
        {
            zcbor_int32_decode(CBORDecodeState, &arr_int_Statuses[i]);
        }
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_INF("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_GetUnitConfiguration(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                            uint8_t u8_NumberOfKeys, const char** arr_sz_Keys, char ** arr_sz_Responses)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_GET_UNIT_CONFIG_ID);

    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));     
    }
    zcbor_list_start_encode(CBORState, u8_NumberOfKeys);
    for (uint8_t i = 0; i < u8_NumberOfKeys; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Keys[i], strlen(arr_sz_Keys[i]) + 1);
    }
    zcbor_list_end_encode(CBORState, u8_NumberOfKeys);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    uint8_t u8_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(u8_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    u8_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, u8_EncodedResponses, u32_ResponseLength, 3, 0);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        struct zcbor_string sz_zcborKey; 
        for (uint32_t i = 0; zcbor_tstr_decode(CBORDecodeState, &sz_zcborKey); i++)
        {
            arr_sz_Responses[i] = (char *)sz_zcborKey.value;
        }
        zcbor_list_end_decode(CBORDecodeState);
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_INF("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_SetUnitConfiguration(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                            uint8_t u8_NumberOfKeys, const char **arr_sz_Keys,
                                            const char **arr_sz_Values, int *arr_int_Statuses)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_SET_UNIT_CONFIG_ID);

    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    const uint8_t u8_MAP_WIDTH = 2; // Keys | Values
    zcbor_map_start_encode(CBORState, u8_NumberOfKeys*u8_MAP_WIDTH);
    for (uint8_t i = 0; i < u8_NumberOfKeys; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Keys[i], strlen(arr_sz_Keys[i]) + 1);
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Values[i], strlen(arr_sz_Values[i]) + 1);
    }
    zcbor_map_end_encode(CBORState, u8_NumberOfKeys*u8_MAP_WIDTH);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    FileScop uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, arr_s32_EncodedResponses, u32_ResponseLength, 3, 0);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        for (uint32_t i = 0; i < u8_NumberOfKeys; i++)
        {
            zcbor_int32_decode(CBORDecodeState, &arr_int_Statuses[i]);
        }
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_INF("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_GetTouSettings(IAL_TargetDevice_t enm_Target, uint8_t u8_ProfileID, const char **sz_SettingPath,
                                        uint8_t u8_SettingDepth, char *sz_Response)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_GET_TOU_ID);

    const char *sz_PathPrefix = IAL_DATA_sz_TOU_KEY;
    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_list_start_encode(CBORState, 0);                                              // Start List
    zcbor_tstr_encode_ptr(CBORState, sz_PathPrefix, strlen(sz_PathPrefix)+1);               // Path Prefix
    if (u8_ProfileID != COD_u8_NULL_UNIT_ID)
    {
        zcbor_uint32_put(CBORState, u8_ProfileID);                                          // Profile ID
    }
    if (u8_SettingDepth > 0) 
    {
        for (uint8_t i = 0; i < u8_SettingDepth; i++)
        {
            zcbor_tstr_encode_ptr(CBORState, sz_SettingPath[i], strlen(sz_SettingPath[i])); // Setting/Extended Path Name
        }
        
    }
    zcbor_list_end_encode(CBORState, 0);                                                // End List
    zcbor_uint32_put(CBORState, 0);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    uint8_t u8_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(u8_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    u8_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_Rc < 0) 
    {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        u32_ResponseLength = sizeof(u8_EncodedResponses);
        int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    u8_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
        if (int_Rc < 0) 
        {
            LOG_INF("ICC_Function_Call Retry Error %d \"%s\"", int_Rc, strerror(-int_Rc));
            return int_Rc;
        }
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, u8_EncodedResponses, u32_ResponseLength, 3, 0);
    if (zcbor_map_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        ArduinoJson::StaticJsonDocument<IAL_u16_ICC_RESPONSE_BUFFER_SIZE> jd_response;
        struct zcbor_string sz_zcborKey; 
        for (uint32_t i = 0; zcbor_tstr_decode(CBORDecodeState, &sz_zcborKey); i++)
        {
            
            // Store those key value pairs in the Response JSON
            const char *sz_key = (char *)sz_zcborKey.value;
            struct zcbor_string sz_zcborValue;
            zcbor_tstr_decode(CBORDecodeState, &sz_zcborValue);
            const char *sz_value = (char *)sz_zcborValue.value;
            jd_response[sz_key] = sz_value;
        }
        zcbor_map_end_decode(CBORDecodeState);
        char sz_buffer[1024];
        ArduinoJson::serializeJson(jd_response, sz_buffer);
        strcpy(sz_Response, sz_buffer);
        LOG_INF("JSON Response: %s", sz_Response);
    }

    LOG_INF("Call for ToU Get: %s %d data: %d, %u bytes", sz_PathPrefix, u8_ProfileID, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_SetTouSettings(IAL_TargetDevice_t enm_Target, uint8_t u8_ProfileID,
                                        const char *sz_SerializedSettings, const char *sz_ArrayName)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_SET_TOU_ID);

    const char *sz_PathPrefix = IAL_DATA_sz_TOU_KEY;
    // Convert the serialized array of settings to a Json Doc, then extract the array
    ArduinoJson::StaticJsonDocument<IAL_u16_ICC_RESPONSE_BUFFER_SIZE> jd_SettingsDoc;
    ArduinoJson::deserializeJson(jd_SettingsDoc, sz_SerializedSettings);
    if (!jd_SettingsDoc.containsKey(sz_ArrayName))
    {
        LOG_ERR("Failed to set ToU settings due to invalid arrayName: %s", sz_ArrayName);
        return EINVAL;
    }
    ArduinoJson::JsonArray ja_Settings = jd_SettingsDoc[sz_ArrayName];
    // Start building the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_list_start_encode(CBORState, 0);                                      // Start List
    zcbor_tstr_encode_ptr(CBORState, sz_PathPrefix, strlen(sz_PathPrefix));         // Path Prefix
    if (u8_ProfileID != COD_u8_NULL_UNIT_ID)
    {
        zcbor_uint32_put(CBORState, u8_ProfileID);                                  // Profile ID
    }
    zcbor_list_end_encode(CBORState, 0);                                        // End List
    if (!ja_Settings.isNull())
    {
        // Build the Request
        zcbor_map_start_encode(CBORState, 0);                                   // Start Map
        zcbor_tstr_encode_ptr(CBORState, sz_ArrayName, strlen(sz_ArrayName));       // Array/List Name
        zcbor_list_start_encode(CBORState, 0);                                      // Start List
        for (uint8_t i = 0; i < ja_Settings.size(); i++)    // Iterate through List
        {
            ArduinoJson::JsonObject jo_arrayElement = ja_Settings[i];
            zcbor_map_start_encode(CBORState, 0);                                       // Start Map
            for (ArduinoJson::JsonPair keyValue : jo_arrayElement)  // Iterate Through Map
            {  
                const char *sz_Key = keyValue.key().c_str();
                ArduinoJson::JsonVariant jv_value = keyValue.value();
                
                char sz_Value[64] = "";
                if (!jv_value.is<const char*>())
                {
                    //If the value is an integer, convert it to a string
                    sprintf(sz_Value, "%i", jv_value.as<int32_t>());
                }
                else
                {
                    strcpy(sz_Value, jv_value.as<const char*>());
                }
                zcbor_tstr_encode_ptr(CBORState, sz_Key, strlen(sz_Key));                   // Key
                zcbor_tstr_encode_ptr(CBORState, sz_Value, strlen(sz_Value));
            }
            zcbor_map_end_encode(CBORState, 0);                                         // End Map
        }
        zcbor_list_end_encode(CBORState, 0);                                        // End List
        zcbor_map_end_encode(CBORState, 0);                                     // End Map
    }
    else // Non Array type requests
    {
        zcbor_map_start_encode(CBORState, 0);                                   // Start Map
        for (ArduinoJson::JsonPair jp_keyValue : jd_SettingsDoc.as<ArduinoJson::JsonObject>()) 
        {  
            const char *sz_key = jp_keyValue.key().c_str();
            ArduinoJson::JsonVariant jv_value = jp_keyValue.value();
            char sz_value[64];
            sprintf(sz_value, "%i", jv_value.as<int32_t>());
            zcbor_tstr_encode_ptr(CBORState, sz_key, strlen(sz_key));
            zcbor_tstr_encode_ptr(CBORState, sz_value, strlen(sz_value));
        }
        zcbor_map_end_encode(CBORState, 0);                                     // End Map
    }
    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    FileScop uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_Rc < 0) 
    {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
        int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
        if (int_Rc < 0) 
        {
            LOG_INF("ICC_Function_Call Retry Error %d \"%s\"", int_Rc, strerror(-int_Rc));
            return int_Rc;
        }
    }
    LOG_INF("Call for ToU Settings: %s %d data: %d, %u bytes", sz_PathPrefix, u8_ProfileID, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_TestAlertStatusCallback(IAL_TargetDevice_t enm_Target, const char *sz_UnitType,
                                                uint8_t u8_UnitIndex)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_ALERT_STATUS_ID);

    uint8_t u8_NewCodeQty = 2;
    const char *arr_sz_NewCodes[u8_NewCodeQty] = { "A1_2", "F2_7" };
    uint8_t u8_ClearedCodeQty = 2;
    const char *arr_sz_ClearedCodes[u8_ClearedCodeQty] = { "A1_3", "F2_8" };
    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    // Encode Unit Type
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    // Encode Unit ID
    zcbor_uint32_put(CBORState, (u8_UnitIndex));
    // Encode New codes
    zcbor_list_start_encode(CBORState, u8_NewCodeQty);
    for (uint8_t i = 0; i < u8_NewCodeQty; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_NewCodes[i], strlen(arr_sz_NewCodes[i]) + 1);
    }
    zcbor_list_end_encode(CBORState, u8_NewCodeQty);
    // Encode Cleared codes
    zcbor_list_start_encode(CBORState, u8_ClearedCodeQty);
    for (uint8_t i = 0; i < u8_ClearedCodeQty; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_ClearedCodes[i], strlen(arr_sz_ClearedCodes[i]) + 1);
    }
    zcbor_list_end_encode(CBORState, u8_ClearedCodeQty);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    FileScop uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }
    LOG_INF("AWS SEND ALERT TEST RESULT %d \"%s\"", int_Rc, strerror(-int_Rc));
    return int_Rc;
}

GlobScop int IAL_Data_GetI3EIndicators(IAL_TargetDevice_t enm_Target, char *sz_LongForm, size_t siz_LongForm,
                                        char *sz_ShortForm, size_t siz_ShortForm)
{
    //Verify the buffers
    if (sz_LongForm == NULL || siz_LongForm == 0 || sz_ShortForm == NULL || siz_ShortForm == 0)
    {
        LOG_DBG("IAL_Data_GetI3EIndicators Error - Invalid buffer(s)");
        return -EINVAL;
    }

    //Send the request (no body)
    StackScop uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_GET_I3E_INDICATORS_ID);
    StackScop uint8_t u8_ResponseBuffer[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    StackScop uint32_t u32_ResponseLength = sizeof(u8_ResponseBuffer);
    StackScop int int_RC = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, NULL, 0, u8_ResponseBuffer,
                                                &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));                               
    if (int_RC != EXIT_SUCCESS) {
        LOG_DBG("ICC call to get I3E Indicators from target[%u] failed with error %d", (uint32_t)enm_Target, int_RC);
        return int_RC;
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, u8_ResponseBuffer, u32_ResponseLength, 1, 0);
    StackScop struct zcbor_string sz_zcbor_Key;
    StackScop struct zcbor_string sz_zcbor_Value;
    if (zcbor_map_start_decode(CBORDecodeState) == 0)
    {
        LOG_DBG("IAL_Data_GetI3EIndicators Error - Failed to decode CBOR map");
        LOG_HEXDUMP_DBG(u8_ResponseBuffer, u32_ResponseLength, "CBOR Data");
        return -EBADMSG;
    }
    zcbor_tstr_decode(CBORDecodeState, &sz_zcbor_Key);
    zcbor_tstr_decode(CBORDecodeState, &sz_zcbor_Value);
    if ((strcmp((const char*)sz_zcbor_Key.value, ial_sz_I3E_LF_IND_KEY) == 0) && sz_zcbor_Key.len <= siz_LongForm)
    {
        strncpy(sz_LongForm, (const char*)sz_zcbor_Value.value, sz_zcbor_Value.len);
    }
    else
    {
        int_RC = -ENOMEM; //We could not copy the long form indicator
    }
    if (int_RC == EXIT_SUCCESS)
    {
        zcbor_tstr_decode(CBORDecodeState, &sz_zcbor_Key);
        zcbor_tstr_decode(CBORDecodeState, &sz_zcbor_Value);
        if ((strcmp((const char*)sz_zcbor_Key.value, ial_sz_I3E_SF_IND_KEY) == 0) && sz_zcbor_Key.len <= siz_ShortForm)
        {
            strncpy(sz_ShortForm, (const char*)sz_zcbor_Value.value, sz_zcbor_Value.len);
        }
        else
        {
            int_RC = -ENOMEM; //We could not copy the short form indicator
        }
    }
    zcbor_map_end_decode(CBORDecodeState);
    return int_RC;
}

GlobScop int IAL_EncodeResponse_GetI3EIndicators(const char *sz_LongForm, const char *sz_ShortForm,
                                                    void *vptr_Response, size_t *ptr_siz_Response)
{
    //Verify the buffers are valid
    if (sz_LongForm == NULL || sz_ShortForm == NULL || vptr_Response == NULL || ptr_siz_Response == NULL)
    {
        LOG_DBG("IAL_EncodeResponse_GetI3EIndicators Error - Invalid parameter(s)");
        return -EINVAL;
    }

    //Encode the response
    ZCBOR_STATE_E(CBORState, 1, (uint8_t *)vptr_Response, *ptr_siz_Response, 1);
    zcbor_map_start_encode(CBORState, 2);
    zcbor_tstr_encode_ptr(CBORState, ial_sz_I3E_LF_IND_KEY, strlen(ial_sz_I3E_LF_IND_KEY) + 1);
    zcbor_tstr_encode_ptr(CBORState, sz_LongForm, strlen(sz_LongForm) + 1);
    zcbor_tstr_encode_ptr(CBORState, ial_sz_I3E_SF_IND_KEY, strlen(ial_sz_I3E_SF_IND_KEY) + 1);
    zcbor_tstr_encode_ptr(CBORState, sz_ShortForm, strlen(sz_ShortForm) + 1);
    zcbor_map_end_encode(CBORState, 2);
    *ptr_siz_Response = (CBORState[0].payload - (uint8_t *)vptr_Response);

    return EXIT_SUCCESS;
}

GlobScop int IAL_Data_UpdateStatus(const char * sz_SourceKey, uint32_t u32_Type, uint32_t u32_State, 
                                    const char ** arr_sz_Keys, const char ** arr_sz_Values, 
                                    size_t siz_Pairs, uint16_t u16_Id)
{
    if (sz_SourceKey == NULL || arr_sz_Keys == NULL || arr_sz_Values == NULL) // Ensure the buffers are valid
    {
        LOG_ERR("IAL_Data_DecodeRequest_UpdateStatus Error - Invalid parameter(s)");
        return -EINVAL;
    }

    // Build the Request
    StackScop uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);

    if (zcbor_tstr_encode_ptr(CBORState, sz_SourceKey, strlen(sz_SourceKey) + 1) == 0) // size + 1 for "\0"
    {
        LOG_ERR("IAL_Data_UpdateStatus Error - Failed to encode Source Key");
        return -ENOMEM;
    }

    if (zcbor_uint_encode(CBORState, &u32_Type, sizeof(u32_Type)) == 0)
    {
        LOG_ERR("IAL_Data_UpdateStatus Error - Failed to encode Type");
        return -ENOMEM;
    }

    if (zcbor_uint_encode(CBORState, &u32_State, sizeof(u32_State)) == 0)
    {
        LOG_ERR("IAL_Data_UpdateStatus Error - Failed to encode State");
        return -ENOMEM;
    }

    if (siz_Pairs == 0)
    {
        zcbor_nil_put(CBORState, NULL);
    }
    else
    {
        // Encode all key value pairs
        StackScop const uint8_t u8_MAP_WIDTH = 2; // Keys | Values
        zcbor_map_start_encode(CBORState, siz_Pairs * u8_MAP_WIDTH);
        for (size_t siz_Index = 0; siz_Index < siz_Pairs; siz_Index++)
        {
            if (arr_sz_Keys[siz_Index] == NULL || arr_sz_Values[siz_Index] == NULL)
            {
                LOG_ERR("IAL_Data_UpdateStatus Error - Invalid parameter(s)");
                return -EINVAL;
            }

            if (zcbor_tstr_encode_ptr(CBORState, arr_sz_Keys[siz_Index], strlen(arr_sz_Keys[siz_Index]) + 1) == 0)
            {
                LOG_ERR("IAL_Data_UpdateStatus Error - Failed to encode Key");
                return -ENOMEM;
            }
            if (zcbor_tstr_encode_ptr(CBORState, arr_sz_Values[siz_Index], strlen(arr_sz_Values[siz_Index]) + 1) == 0)
            {
                LOG_ERR("IAL_Data_UpdateStatus Error - Failed to encode Value");
                return -ENOMEM;
            }
        }
        zcbor_map_end_encode(CBORState, siz_Pairs);
    }

    if (zcbor_uint_encode(CBORState, &u16_Id, sizeof(u16_Id)) == 0)
    {
        LOG_ERR("IAL_Data_UpdateStatus Error - Failed to encode ID");
        return -ENOMEM;
    }

    // Send the Status
    StackScop uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    StackScop int int_Rc = ICC_Event_Dispatch(ial_ptr_IccCtx, IAL_UPDATE_STATUS_ID, u8_RequestBuffer, u32_RequestLength, 
                                        RTOS_SECONDS(IAL_u8_ICC_EVENT_TIMEOUT_S));
    
    if (int_Rc < 0) 
    {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }
    return int_Rc;
}

GlobScop int IAL_Data_DecodeRequest_UpdateStatus(const uint8_t * ptr_u8_Data, uint32_t u32_DataLength, 
                                                    char * sz_SourceKey, size_t siz_MaxSourceKey,
                                                    uint32_t * ptr_u32_Type, uint32_t * ptr_u32_State, 
                                                    char ** arr_sz_Keys, size_t siz_MaxKey, char ** arr_sz_Values, 
                                                    size_t siz_MaxValue, size_t siz_MaxPairs, size_t * ptr_siz_Pairs, 
                                                    uint16_t * ptr_u16_Id)
{
    if (ptr_u8_Data == NULL || sz_SourceKey == NULL || ptr_u32_Type == NULL || ptr_u32_State == NULL || 
        arr_sz_Keys == NULL || arr_sz_Values == NULL || ptr_siz_Pairs == NULL || ptr_u16_Id == NULL) // Ensure the buffers are valid
    {
        LOG_ERR("IAL_Data_DecodeRequest_UpdateStatus Error - Invalid parameter(s)");
        return -EINVAL;
    }

    // Decode the response
    ZCBOR_STATE_D(CBORState_Decode, 1, ptr_u8_Data, u32_DataLength, 5, 0);
    StackScop struct zcbor_string sz_zcbor_SourceKey;
    if (zcbor_tstr_decode(CBORState_Decode, &sz_zcbor_SourceKey) == 0)
    {
        LOG_DBG("IAL_Data_DecodeRequest_UpdateStatus Error - Failed to decode Source Key");
        LOG_HEXDUMP_DBG(ptr_u8_Data, u32_DataLength, "CBOR Data");
        return -EBADMSG;
    }

    // Validate Source Key Length
    if (sz_zcbor_SourceKey.len > siz_MaxSourceKey)
    {
        LOG_DBG("IAL_Data_DecodeRequest_UpdateStatus Error - Source Key too long");
        return -ENOMEM;
    }
    else
    {
        strncpy(sz_SourceKey, (const char *)sz_zcbor_SourceKey.value, sz_zcbor_SourceKey.len);
        sz_SourceKey[sz_zcbor_SourceKey.len] = '\0'; // Ensure Null Termination
    }

    if (zcbor_uint_decode(CBORState_Decode, ptr_u32_Type, sizeof(*ptr_u32_Type)) == 0)
    {
        LOG_DBG("IAL_Data_DecodeRequest_UpdateStatus Error - Failed to decode Type");
        LOG_HEXDUMP_DBG(ptr_u8_Data, u32_DataLength, "CBOR Data");
        return -EBADMSG;
    }

    if (zcbor_uint_decode(CBORState_Decode, ptr_u32_State, sizeof(*ptr_u32_State)) == 0)
    {
        LOG_DBG("IAL_Data_DecodeRequest_UpdateStatus Error - Failed to decode State");
        LOG_HEXDUMP_DBG(ptr_u8_Data, u32_DataLength, "CBOR Data");
        return -EBADMSG;
    }

    *ptr_siz_Pairs = 0;

    if (zcbor_nil_expect(CBORState_Decode, NULL) == 0)
    {
        // Parse Key Value Pairs
        StackScop struct zcbor_string sz_zcbor_Key;
        StackScop struct zcbor_string sz_zcbor_Value;
        if (zcbor_map_start_decode(CBORState_Decode) == 0)
        {
            LOG_DBG("IAL_Data_DecodeRequest_UpdateStatus Error - Failed to decode CBOR map");
            LOG_HEXDUMP_DBG(ptr_u8_Data, u32_DataLength, "CBOR Data");
            return -EBADMSG;
        }

        for (size_t siz_Index = 0; zcbor_tstr_decode(CBORState_Decode, &sz_zcbor_Key); siz_Index++)
        {
            if (siz_Index >= siz_MaxPairs)
            {
                LOG_DBG("IAL_Data_DecodeRequest_UpdateStatus Error - Too many key-value pairs");
                return -ENOMEM;
            }

            if (zcbor_tstr_decode(CBORState_Decode, &sz_zcbor_Value) == 0)
            {
                LOG_DBG("IAL_Data_DecodeRequest_UpdateStatus Error - Failed to decode Value");
                LOG_HEXDUMP_DBG(ptr_u8_Data, u32_DataLength, "CBOR Data");
                return -EBADMSG;
            }

            // Validate Key Length
            if (sz_zcbor_Key.len > siz_MaxKey || sz_zcbor_Value.len > siz_MaxValue)
            {
                LOG_DBG("IAL_Data_DecodeRequest_UpdateStatus Error - Key or Value too long");
                return -ENOMEM;
            }
            if (arr_sz_Keys[siz_Index] != NULL && arr_sz_Values[siz_Index] != NULL)
            {
                strncpy(arr_sz_Keys[siz_Index], (const char *)sz_zcbor_Key.value, sz_zcbor_Key.len);
                arr_sz_Keys[siz_Index][sz_zcbor_Key.len - 1] = '\0'; // Ensure Null Termination
                strncpy(arr_sz_Values[siz_Index], (const char *)sz_zcbor_Value.value, sz_zcbor_Value.len);
                arr_sz_Values[siz_Index][sz_zcbor_Value.len - 1] = '\0'; // Ensure Null Termination
            }
            else
            {
                LOG_DBG("IAL_Data_DecodeRequest_UpdateStatus Error - Invalid buffer(s)");
                return -EINVAL;
            }
            (*ptr_siz_Pairs)++;
        }
        zcbor_map_end_decode(CBORState_Decode);
    }

    if (zcbor_uint_decode(CBORState_Decode, ptr_u16_Id, sizeof(*ptr_u16_Id)) == 0)
    {
        LOG_DBG("IAL_Data_DecodeRequest_UpdateStatus Error - Failed to decode Id");
        LOG_HEXDUMP_DBG(ptr_u8_Data, u32_DataLength, "CBOR Data");
        return -EBADMSG;
    }
    else
    {
        return EXIT_SUCCESS;
    }
}

GlobScop int IAL_Data_ClearUpdateStatus(IAL_TargetDevice_t enm_Target, uint16_t u16_Id)
{
    // Build the Function ID
    StackScop uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_CLEAR_UPDATE_STATUS_ID);

    // Build the Request
    StackScop uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE >> 1];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    if (zcbor_uint_encode(CBORState, &u16_Id, sizeof(u16_Id)) == 0)
    {
        LOG_ERR("IAL_Data_ClearUpdateStatus Error - Failed to encode Id");
        return -ENOMEM;
    }

    // Send the Status
    StackScop uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    NULL, 0, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    
    if (int_Rc < 0) 
    {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }
    return int_Rc;
}

GlobScop int IAL_Data_DecodeRequest_ClearUpdateStatus(const uint8_t* ptr_u8_Data, uint32_t u32_DataLength, 
                                                        uint16_t * ptr_u16_Id)
{
    if (ptr_u16_Id == NULL)
    {
        LOG_ERR("IAL_Data_DecodeRequest_ClearDecodeRequest Error - Invalid parameter(s)");
        return -EINVAL;
    }

    // Decode the response
    ZCBOR_STATE_D(CBORState_Decode, 1, ptr_u8_Data, u32_DataLength, 1, 0);

    if (zcbor_uint_decode(CBORState_Decode, ptr_u16_Id, sizeof(*ptr_u16_Id)) == 0)
    {
        LOG_DBG("IAL_Data_DecodeRequest_ClearDecodeRequest Error - Failed to decode Id");
        LOG_HEXDUMP_DBG(ptr_u8_Data, u32_DataLength, "CBOR Data");
        return -EBADMSG;
    }
    else
    {
        return EXIT_SUCCESS;
    }
}

GlobScop int IAL_Data_PublishVersionInfo(IAL_TargetDevice_t enm_Target, uint8_t u8_NumKeys, const char **arr_sz_Keys, 
                                            const char **arr_sz_Values)
{
    //Build the request ID
    StackScop uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_VERSION_INFO_ID);

    //Build the request
    StackScop uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORStateE, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    StackScop const uint8_t u8_MAP_WIDTH = 2; //Key and value pair
    if (u8_NumKeys == 0 || arr_sz_Keys == NULL || arr_sz_Values == NULL)
    {
        LOG_ERR("IAL_Data_PublishVersionInfo Error - Invalid parameter(s)");
        return -EINVAL;
    }
    zcbor_map_start_encode(CBORStateE, u8_NumKeys*u8_MAP_WIDTH);
    for (uint8_t u8_Index = 0; u8_Index < u8_NumKeys; u8_Index++)
    {
        if (arr_sz_Keys[u8_Index] == NULL || arr_sz_Values[u8_Index] == NULL)
        {
            LOG_ERR("IAL_Data_PublishVersionInfo Error - Invalid parameter(s)");
            return -EINVAL;
        }
        zcbor_tstr_encode_ptr(CBORStateE, arr_sz_Keys[u8_Index], strlen(arr_sz_Keys[u8_Index]) + 1); //+1 for null terminator
        zcbor_tstr_encode_ptr(CBORStateE, arr_sz_Values[u8_Index], strlen(arr_sz_Values[u8_Index]) + 1); //+1 for null terminator
    }
    zcbor_map_end_encode(CBORStateE, u8_NumKeys*u8_MAP_WIDTH);

    //Send the request
    //Get the length of the request by subtracting the start of the buffer from the current position
    StackScop uint32_t u32_ReqLen = (CBORStateE[0].payload - u8_RequestBuffer); 
    StackScop int int_RC = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_ReqLen, NULL, NULL, 
                                    RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));

    // Check for errors
    if (int_RC < 0)
    {
        LOG_ERR("IAL_Data_PublishVersionInfo Error - ICC_Function_Call failed with error %d", int_RC);
        return int_RC;
    }
    return int_RC;
}


GlobScop int IAL_Data_DecodeRequest_PublishVersionInfo(const uint8_t* ptr_u8_Data, uint32_t u32_DataLength, 
                                                        const char **arr_sz_Keys, const char **arr_sz_Values, 
                                                        uint8_t u8_MaxKeys, uint8_t* ptr_u8_NumKeys)
{
    // Verify input parameters are valid
    if (arr_sz_Keys == NULL || arr_sz_Values == NULL || ptr_u8_NumKeys == NULL || u8_MaxKeys == 0)
    {
        LOG_ERR("IAL_Data_DecodeRequest_PublishVersionInfo Error - Invalid parameter(s)");
        return -EINVAL;
    }

    //Decode keys/values from the CBOR encoded data
    ZCBOR_STATE_D(CBORDecodeState, 1, ptr_u8_Data, u32_DataLength, 1, 0);
    if (zcbor_map_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
        return -EINVAL;
    }

    // Decode the key value pairs pointes and store them in the provided arrays
    *ptr_u8_NumKeys = 0;
    StackScop struct zcbor_string sz_zcborKey;
    StackScop struct zcbor_string sz_zcborValue;
    for (uint8_t u8_Index = 0; zcbor_tstr_decode(CBORDecodeState, &sz_zcborKey); u8_Index++)
    {
        if (u8_Index >= u8_MaxKeys)
        {
            LOG_ERR("IAL_Data_DecodeRequest_PublishVersionInfo Error - Max pairs exceeded");
            return -ENOMEM;
        }
        zcbor_tstr_decode(CBORDecodeState, &sz_zcborValue);

        // Set key and value pointes to the decoded values pointers (limited to scope of ptr_u8_Data)
        arr_sz_Keys[u8_Index] = (const char *)sz_zcborKey.value;
        arr_sz_Values[u8_Index] = (const char *)sz_zcborValue.value;
        (*ptr_u8_NumKeys)++;
    }
    zcbor_map_end_decode(CBORDecodeState);
    return EXIT_SUCCESS;
}


/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop size_t ial_data_BufferedDataSize(const char *sz_UnitType)
{
    #if CONFIG_EMS_SCH
    if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_ATS) == 0)
    {
        return sizeof(SCH_AtsParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_DCDC) == 0)
    {
        return sizeof(SCH_DcdcParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_EMS) == 0)
    {
        return sizeof(SCH_EmsParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_GW) == 0)
    {
        return sizeof(SCH_GwParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_HVAC) == 0)
    {
        return sizeof(SCH_HvacParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_PCS) == 0)
    {
        return sizeof(SCH_PcsParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_PDU) == 0)
    {
        return sizeof(SCH_PduParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_PMU) == 0)
    {
        return sizeof(SCH_PmuParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_TYPE_TOU) == 0)
    {
        return sizeof(SCH_EmsTouSettings_t);
    }
    else
    {
        return INT_MAX; //Invalid Unit Type, return very large number
    }
    #else
    return 0;
    #endif
}

//EOF
