#include <lion/ial/ems/IAL_Modbus.h>
#include <lion/log/log.h>
#include <zcbor_decode.h>
#include <zcbor_encode.h>

LOG_MODULE_REGISTER(ial_modbus, LOG_LEVEL_INF);

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

/**
 * @brief Checks the ICC return value for invalid data and handles it accordingly.
 * @param s32_IccReturnVal The return value from the ICC function call.
 * @param ptr_u16_ValueDest The destination for the value to be stored.
 */
FileScop inline void ial_mbs_InvalidDataHandler(int32_t s32_IccReturnVal, uint16_t *ptr_u16_ValueDest);

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

ModScop extern ICC_Context_t *ial_ptr_IccCtx; //Defined in IAL.c

/*********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

GlobScop int IAL_Mbs_ReadRegisters(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                     uint16_t *arr_u16_Addresses, uint16_t *arr_u16_Values, size_t siz_Request)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_READ_REGISTERS_ID);

    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    else
    {
        zcbor_nil_put(CBORState, NULL);
    }
    zcbor_list_start_encode(CBORState, siz_Request);
    for (uint16_t i = 0; i < siz_Request; i++)
    {
        zcbor_uint_encode(CBORState, &arr_u16_Addresses[i], sizeof(uint16_t));
    }
    zcbor_list_end_encode(CBORState, siz_Request);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_rc < 0) {
        LOG_DBG("ICC_Function_Call Error %d \"%s\", Attempting Retry", int_rc, strerror(-int_rc));
        // Try again
        u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
        int_rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
        if (int_rc < 0) {
            LOG_ERR("ICC_Function_Call Retry Failed with Error %d \"%s\"", int_rc, strerror(-int_rc));
            return int_rc;
        }
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, arr_s32_EncodedResponses, u32_ResponseLength, 3, 0);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_ERR("Decoding CBOR Error");
    }
    else 
    {
        int32_t s32_DecodedValue;
        for (uint16_t i = 0; i < siz_Request; i++)
        {
            zcbor_int32_decode(CBORDecodeState, &s32_DecodedValue);
            ial_mbs_InvalidDataHandler(s32_DecodedValue, &arr_u16_Values[i]);
        }
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_DBG("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_rc, u32_ResponseLength);
    return int_rc;
}

GlobScop int IAL_Mbs_ReadMultipleRegisters(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                             uint16_t u16_startAddress, uint16_t *arr_u16_Values, size_t siz_Request)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_READ_MULTIPLE_ID);

    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    else
    {
        zcbor_nil_put(CBORState, NULL);
    }
    zcbor_uint_encode(CBORState, &u16_startAddress, sizeof(uint16_t));
    zcbor_uint_encode(CBORState, &siz_Request, sizeof(uint16_t));

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_rc < 0) {
        LOG_DBG("ICC_Function_Call Error %d \"%s\", Attempting Retry", int_rc, strerror(-int_rc));
        // Try again
        u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
        int_rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
        if (int_rc < 0) {
            LOG_ERR("ICC_Function_Call Retry Failed with Error %d \"%s\"", int_rc, strerror(-int_rc));
            return int_rc;
        }
    }
    
    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, arr_s32_EncodedResponses, u32_ResponseLength, 3, 0);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_ERR("Decoding CBOR Error");
    }
    else 
    {
        for (uint16_t i = 0; i < siz_Request; i++)
        {
            zcbor_int_decode(CBORDecodeState, &arr_u16_Values[i], sizeof(uint16_t));
        }
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_DBG("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_rc, u32_ResponseLength);
    return int_rc;
}

GlobScop int IAL_Mbs_WriteRegisters(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                    uint16_t *arr_u16_Addresses, uint16_t *arr_u16_Values, int32_t *arr_s32_results,
                                    size_t siz_Request)
{
    return IAL_Mbs_WriteRegisters_Sourced(enm_Target, NULL, sz_UnitType, u8_UnitIndex, arr_u16_Addresses, arr_u16_Values,
                                            arr_s32_results, siz_Request);
}

GlobScop int IAL_Mbs_WriteRegisters_Sourced(IAL_TargetDevice_t enm_Target, const char *sz_Source, const char *sz_UnitType,
                                            uint8_t u8_UnitIndex, uint16_t *arr_u16_Addresses, uint16_t *arr_u16_Values,
                                            int32_t *arr_s32_results, size_t siz_Request)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_WRITE_REGISTERS_ID);

    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    else
    {
        zcbor_nil_put(CBORState, NULL);
    }
    if (sz_Source != NULL) {
        zcbor_tstr_encode_ptr(CBORState, sz_Source, strlen(sz_Source)+1);
    }
    else
    {
        zcbor_nil_put(CBORState, NULL);
    }
    zcbor_map_start_encode(CBORState, siz_Request);
    for (uint16_t i = 0; i < siz_Request; i++)
    {
        zcbor_uint_encode(CBORState, &arr_u16_Addresses[i], sizeof(uint16_t));
        zcbor_uint_encode(CBORState, &arr_u16_Values[i], sizeof(uint16_t));
    }
    zcbor_map_end_encode(CBORState, siz_Request);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_rc < 0) {
        LOG_DBG("ICC_Function_Call Error %d \"%s\", Attempting Retry", int_rc, strerror(-int_rc));
        // Try again
        u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
        int_rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
        if (int_rc < 0) {
            LOG_ERR("ICC_Function_Call Retry Failed with Error %d \"%s\"", int_rc, strerror(-int_rc));
            return int_rc;
        }
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, arr_s32_EncodedResponses, u32_ResponseLength, 3, 0);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_ERR("Decoding CBOR Error");
    }
    else 
    {
        for (uint16_t i = 0; i < siz_Request; i++)
        {
            zcbor_int32_decode(CBORDecodeState, &arr_s32_results[i]);
        }
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_DBG("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_rc, u32_ResponseLength);
    return int_rc;
}

GlobScop int IAL_Mbs_WriteMultipleRegisters(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                            uint16_t u16_startAddress, uint16_t *arr_u16_Values, size_t siz_Request)
{
    return IAL_Mbs_WriteMultipleRegisters_Sourced(enm_Target, NULL, sz_UnitType, u8_UnitIndex, u16_startAddress,
                                                    arr_u16_Values, siz_Request);
}

GlobScop int IAL_Mbs_WriteMultipleRegisters_Sourced(IAL_TargetDevice_t enm_Target, const char *sz_Source,
                                                    const char *sz_UnitType, uint8_t u8_UnitIndex,
                                                    uint16_t u16_startAddress, uint16_t *arr_u16_Values, size_t siz_Request)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_WRITE_MULTIPLE_ID);

    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    else
    {
        zcbor_nil_put(CBORState, NULL);
    }
    if (sz_Source != NULL) {
        zcbor_tstr_encode_ptr(CBORState, sz_Source, strlen(sz_Source)+1);
    }
    else
    {
        zcbor_nil_put(CBORState, NULL);
    }
    zcbor_uint_encode(CBORState, &u16_startAddress, sizeof(uint16_t));
    zcbor_uint_encode(CBORState, &siz_Request, sizeof(uint16_t));
    zcbor_list_start_encode(CBORState, siz_Request);
    for (uint16_t i = 0; i < siz_Request; i++)
    {
        zcbor_uint_encode(CBORState, &arr_u16_Values[i], sizeof(uint16_t));
    }
    zcbor_list_end_encode(CBORState, siz_Request);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_rc < 0) {
        LOG_DBG("ICC_Function_Call Error %d \"%s\", Attempting Retry", int_rc, strerror(-int_rc));
        // Try again
        u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
        int_rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
        if (int_rc < 0) {
            LOG_ERR("ICC_Function_Call Retry Failed with Error %d \"%s\"", int_rc, strerror(-int_rc));
            return int_rc;
        }
    }

    //No response expected, so no decoding required

    LOG_DBG("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_rc, u32_ResponseLength);
    return int_rc;
}

GlobScop int IAL_Mbs_LockRegisters(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                    uint16_t *arr_u16_Addresses, int32_t *arr_s32_results, size_t siz_Request)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_LOCK_REGISTERS_ID);

    //Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBOREncodeState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 3);
    zcbor_tstr_encode_ptr(CBOREncodeState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBOREncodeState, (u8_UnitIndex));
    }
    else
    {
        zcbor_nil_put(CBOREncodeState, NULL);
    }
    zcbor_list_start_encode(CBOREncodeState, siz_Request);
    for (uint16_t i = 0; i < siz_Request; i++)
    {
        zcbor_uint_encode(CBOREncodeState, &arr_u16_Addresses[i], sizeof(uint16_t));
    }
    zcbor_list_end_encode(CBOREncodeState, siz_Request);

    //Send the Request
    uint32_t u32_RequestLength = (CBOREncodeState[0].payload - u8_RequestBuffer);
    uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_RC = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_RC < 0) {
        LOG_DBG("ICC_Function_Call Error %d \"%s\", Attempting Retry", int_RC, strerror(-int_RC));
        // Try again
        u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
        int_RC = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
        if (int_RC < 0) {
            LOG_ERR("ICC_Function_Call Retry Failed with Error %d \"%s\"", int_RC, strerror(-int_RC));
            return int_RC;
        }
    }

    //Decode the response
    ZCBOR_STATE_D(CBORDecodeState, 1, arr_s32_EncodedResponses, u32_ResponseLength, 1, 0);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_ERR("Decoding CBOR Error");
    }
    else 
    {
        for (uint16_t i = 0; i < siz_Request; i++)
        {
            zcbor_int32_decode(CBORDecodeState, &arr_s32_results[i]);
        }
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_DBG("Call for lock %s %d registers: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_RC, u32_ResponseLength);
    return int_RC;
}

GlobScop int IAL_Mbs_UnlockRegisters(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex,
                                        uint16_t *arr_u16_Addresses, int32_t *arr_s32_results, size_t siz_Request)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_UNLOCK_REGISTERS_ID);

    //Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBOREncodeState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 3);
    zcbor_tstr_encode_ptr(CBOREncodeState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBOREncodeState, (u8_UnitIndex));
    }
    else
    {
        zcbor_nil_put(CBOREncodeState, NULL);
    }
    zcbor_list_start_encode(CBOREncodeState, siz_Request);
    for (uint16_t i = 0; i < siz_Request; i++)
    {
        zcbor_uint_encode(CBOREncodeState, &arr_u16_Addresses[i], sizeof(uint16_t));
    }
    zcbor_list_end_encode(CBOREncodeState, siz_Request);

    //Send the Request
    uint32_t u32_RequestLength = (CBOREncodeState[0].payload - u8_RequestBuffer);
    uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_RC = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_RC < 0) {
        LOG_DBG("ICC_Function_Call Error %d \"%s\", Attempting Retry", int_RC, strerror(-int_RC));
        // Try again
        u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
        int_RC = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
        if (int_RC < 0) {
            LOG_ERR("ICC_Function_Call Retry Failed with Error %d \"%s\"", int_RC, strerror(-int_RC));
            return int_RC;
        }
    }

    //Decode the response
    ZCBOR_STATE_D(CBORDecodeState, 1, arr_s32_EncodedResponses, u32_ResponseLength, 1, 0);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_ERR("Decoding CBOR Error");
    }
    else 
    {
        for (uint16_t i = 0; i < siz_Request; i++)
        {
            zcbor_int32_decode(CBORDecodeState, &arr_s32_results[i]);
        }
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_DBG("Call for unlock %s %d registers: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_RC, u32_ResponseLength);
    return int_RC;
}

GlobScop int IAL_Mbs_ResetLockRegistry(IAL_TargetDevice_t enm_Target, const char *sz_UnitType, uint8_t u8_UnitIndex)
{
    uint32_t u32_FunctionID = IAL_Int_BuildTargetedFunctionID(enm_Target, IAL_RESET_LOCK_REGISTRY_ID);

    //Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBOREncodeState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 2);
    zcbor_tstr_encode_ptr(CBOREncodeState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBOREncodeState, (u8_UnitIndex));
    }
    else
    {
        zcbor_nil_put(CBOREncodeState, NULL);
    }

    //Send the Request
    uint32_t u32_RequestLength = (CBOREncodeState[0].payload - u8_RequestBuffer);
    int int_RC = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    NULL, NULL, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_RC < 0) {
        LOG_DBG("ICC_Function_Call Error %d \"%s\", Attempting Retry", int_RC, strerror(-int_RC));
        // Try again
        int_RC = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionID, u8_RequestBuffer, u32_RequestLength, 
                                    NULL, NULL, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
        if (int_RC < 0) {
            LOG_ERR("ICC_Function_Call Retry Failed with Error %d \"%s\"", int_RC, strerror(-int_RC));
            return int_RC;
        }
    }

    LOG_DBG("Call for reset %s %d lock registry: %d", sz_UnitType, u8_UnitIndex, int_RC);
    return int_RC;
}

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop inline void ial_mbs_InvalidDataHandler(int32_t s32_IccReturnVal, uint16_t *ptr_u16_ValueDest)
{
    if (s32_IccReturnVal == COD_s32_INVALID_DATA)
    {
        //How we handle invalid data is subject to change. For now, set the return value to zero.
        *ptr_u16_ValueDest = 0; 
    }
    else
    {
        *ptr_u16_ValueDest = s32_IccReturnVal;
    }
}

//EOF
