#include "lion/CST.h"
#include "lion/ial/ems/IAL_Modbus.h"
#include "lion/ial/ems/IAL_Data.h"
#include "lion/ial/ems/IAL_Control.h"
#include "lion/log/log.h"
#include "CST_IAL.hpp"
#include <stdio.h>
#include <string.h>
#include <fstream>
#include <math.h>
#include <ArduinoJson.hpp>
#include <unordered_set>
#include <unistd.h>

LOG_MODULE_REGISTER(CST, LOG_LEVEL_INF);

#define CST_INVERTER_MAP_SIZE           (14*1024)
#define CST_APPLICATION_DATA_DIR        "/data/var/lib/lion-i3e/"
#define CST_TRANSLATION_MAP_FILE_PATH   CST_APPLICATION_DATA_DIR "config/cst_megarevo_map.json"
#define CST_CALCULATION_REQUIRED        420
#define CST_SINGLE_REGISTER_SIZE        16
#define CST_HEX_BASE                    16
#define CST_BYTE_SIZE                   8
#define CST_MAX_READ_SIZE               64
#define CST_MAX_DEFAULT_CONTROL_SIZE    16
#define CST_MAX_CURVE_SIZE              32
// #define CST_MIN_CONTROL_TYPE_SIZE       2
// #define CST_PDU_INDEX                   0 
#define CST_PARAMETER_TYPE_SIZE         32
// #define CST_PERCENTAGE_MULTIPLIER       0.01
#define CST_MAX_PERCENT                 100
#define CST_MAX_RATIO                   1.0
#define CST_CURVEPOINT_MASK             0xFFFFE1FF
#define CST_LSB_MASK                    0x000000FF
#define CST_HEX_PTR_OFS                 2


#define CST_MOD_ACRONYM                 "CST"
#define CST_MOD_NAME                    "Compliance Standard Translation Service"

//#TODO - opModTargetVar, opModTargetW, opModFixedW seem to be unsupported (could not find registers to map with)
//#TODO - for now, these are set as supported but not enabled with no control mapping (so controls coming in for these
//#TODO - will not work and they will be reported as OFF always)

//#TODO - a couple of settings values were set to the ratings (setMaxVA, setMaxVar)

/**********************************************************************************************************************
 * @cond                                               FILE SCOPE                                              @endcond
 *********************************************************************************************************************/
typedef struct 
{
    ArduinoJson::JsonVariant jv_modbusAddress;
    ArduinoJson::JsonVariant jv_bitStartIndex;
    ArduinoJson::JsonVariant jv_bitLength;
    ArduinoJson::JsonVariant jv_scalingFactor;
    ArduinoJson::JsonVariant jv_offset;
    ArduinoJson::JsonVariant jv_isSigned;
    ArduinoJson::JsonVariant jv_canWrite;
    ArduinoJson::JsonVariant jv_addressOffset;
    ArduinoJson::JsonVariant jv_stride;
    ArduinoJson::JsonVariant jv_negate;
    ArduinoJson::JsonVariant jv_absoluteValue;
    ArduinoJson::JsonVariant jv_batchStart;
} cst_ParameterAttributes_t;
typedef enum { 
    CST_CAPABILITY, CST_SETTING, CST_STATUS, CST_METER_READING, CST_LOG_EVENT,
    CST_CONTROL, CST_CURVE, CST_DEFAULT_CONTROL
} cst_ParameterType_t;
typedef enum 
{
    REF_NONE = 0,
    REF_PC_SET_MAX_W = 1,
    REF_PC_SET_MAX_VAR = 2,
    REF_PC_STAT_VAR_AVAIL = 3,
    REF_PC_SET_EFFECTIVE_V = 4,
    REF_PC_SET_MAX_CHARGE_RATE_W = 5,
    REF_PC_SET_MAX_DISCHARGE_RATE_W = 6,
    REF_PC_STAT_W_AVAIL = 7
} cst_der_unit_ref_t;
typedef enum 
{
    CST_RESET_REG = 0,
    CST_SET_REG = 1
} cst_writeType_t;

FileScop ArduinoJson::DynamicJsonDocument jd_translationMap(CST_INVERTER_MAP_SIZE);
FileScop uint32_t cst_lastAlarmStatus = 0;

FileScop pthread_mutex_t cst_mtx_lockedRegisters;
FileScop std::unordered_set<uint16_t> cst_uset_lockedRegisters;
FileScop pthread_mutex_t cst_mtx_modelStatus;
FileScop pthread_mutex_t cst_mtx_utilCmdState;

FileScop const char *CST_sz_key_utilCmdEN[] = {"util_cmd_en"};
FileScop const char *CST_sz_val_utilCmdEnON[] = {"1"};
FileScop const char *CST_sz_val_utilCmdEnOFF[] = {"0"};
FileScop uint8_t CST_u8_previousStateUtilCmdEN = 0;

FileScop CST_UtilModeStatus_t cst_enm_ModelStatus = MODEL_AVAIL;


/**
 * @brief   Retrieves the string representation of the given parameter type (for pulling the correct JSON object from the
 *          translation map).
 * @param   enm_type    Parameter type to retrieve the string for
 * @param   sz_type     String to populate with the parameter type
 * @returns Return EXIT_SUCCESS or -EXIT_FAILURE
*/
FileScop int cst_GetParameterTypeString(cst_ParameterType_t enm_type, char *sz_type);

/**
 * @brief   Retrieves all of the single register read values for the given Json Object type from the inverter, using the
 *          translation map.
 * @param   jo_type JSON Object of the parameter type to retrieve the values for
 * @param   arr_u16_singleReadAddresses Array of single register addresses to read
 * @param   arr_u16_singleReadValues Array of single register values to populate
 * @param   ptr_singleReads Reference to the size of the single read addresses and values arrays
 * @returns Return EXIT_SUCCESS or -EXIT_FAILURE
 */
FileScop int cst_GetSingleReadValuesFromObject(ArduinoJson::JsonObject jo_type, uint16_t *arr_u16_singleReadAddresses, 
                                                uint16_t *arr_u16_singleReadValues, size_t *ptr_singleReads);

/**
 * @brief   Populates attributes struct for a given parameter attributes set to default values if not found in 
 *          the parameter map
 * @param   jo_parameter    Parameter to retrieve attributes for
 * @param   ptr_attributes  Reference to the structure to populate with the attributes
 * @returns Return EXIT_SUCCESS, -EXIT_FAILURE, or ICS_CALCULATION_REQUIRED
 */
FileScop int cst_GetAttributes(ArduinoJson::JsonObject jo_parameter, cst_ParameterAttributes_t *ptr_attributes);

/**
 * @brief   Retrieves the attributes for the given parameter and populates the given DERControlType
 * @param   arr_u8_controlType   Pointer to the DERControlType to populate
 * @param   enm_type        Parameter type to retrieve the value for
 * @param   sz_key          Key of the parameter to retrieve the value for
 * @returns Return EXIT_SUCCESS or -EXIT_FAILURE
*/
FileScop int cst_GetDERControlType(uint8_t *arr_u8_controlType, cst_ParameterType_t enm_type, const char *sz_key);

/**
 * @brief   Retrieves the attributes for the given parameter and populates the given CurrentRMS
 * @param   ptr_CurrentRMS  Pointer to the CurrentRMS to populate
 * @param   enm_type        Parameter type to retrieve the value for
 * @param   sz_key          Key of the parameter to retrieve the value for
 * @returns Return EXIT_SUCCESS or -EXIT_FAILURE
*/
FileScop int cst_GetCurrentRMS(SE_CurrentRMS_t *ptr_CurrentRMS, cst_ParameterType_t enm_type, const char *sz_key);

/**
 * @brief  Retrieves and formats the parameter value for generic parameters.
 * @param  enm_type                 Parameter type to retrieve the value for
 * @param  sz_key                   Key of the parameter to retrieve the value for
 * @param  arr_u16_readAddresses    Array of addresses read
 * @param  arr_u16_readValues       Array of values read
 * @param  siz_reads                Size of the read arrays
 * @param  ptr_value                Pointer to the value to populate
 * @returns Return EXIT_SUCCESS or -Posix error
*/
FileScop int cst_GetParameterValue(cst_ParameterType_t enm_type, const char *sz_key, uint16_t *arr_u16_readAddresses, 
                                    uint16_t *arr_u16_readValues, size_t siz_reads, int32_t *ptr_value);

/**
 * @brief   Retrieves the multiplier for the given parameter
 * @param   enm_type        Parameter type to retrieve the multiplier for
 * @param   sz_key          Key of the parameter to retrieve the multiplier for
 * @param   ptr_multiplier  Pointer to the multiplier to populate
 * @returns Return EXIT_SUCCESS or -EXIT_FAILURE
*/
FileScop int cst_GetParameterMultiplier(cst_ParameterType_t enm_type, const char *sz_key, 
                                        SE_PowerOfTenMultiplierType_t *ptr_multiplier);
    
/**
 * @brief   Retrieves the calculated parameter value for the given parameter
 * @param   jo_parameter    JsonObject containing the parameter to retrieve the value for
 * @param   sz_key          Key of the parameter to retrieve the value for
 * @param   ptr_value       Pointer to the value to populate
 * @returns Return EXIT_SUCCESS or -EXIT_FAILURE
*/                          
FileScop int cst_GetCalculatedParameter(ArduinoJson::JsonObject jo_parameter, const char *sz_key, int32_t *ptr_value);

/**
 * @brief   Determines if the give mode is enabled *Only supports modees that can be determined from the inverter
 * @param   jo_mode                     JsonObject of the mode to determine if it is enabled
 * @param   sz_key                      Key of the mode to determine if it is enabled
 * @param   arr_u16_readAddresses Array of single register addresses to read
 * @param   arr_u16_readValues    Array of single register values to populate
 * @param   siz_reads             Size of the single read arrays
 */
FileScop bool cst_IsModeEnabled(ArduinoJson::JsonObject jo_mode, const char *sz_key, uint16_t *arr_u16_readAddresses, 
                                uint16_t *arr_u16_readValues, size_t siz_reads);

/**
 * @brief   Retrieves all of the single register read values for the given Json Array type from the inverter, using the
 *          translation map.
 * @param   ja_type JSON Array of the parameter type to retrieve the values for
 * @param   arr_u16_singleReadAddresses Array of single register addresses to read
 * @param   arr_u16_singleReadValues Array of single register values to populate
 * @param   ptr_singleReads Reference to the size of the single read addresses and values arrays
 * @returns Return EXIT_SUCCESS or -EXIT_FAILURE
 */
FileScop int cst_GetSingleReadValuesFromArray(ArduinoJson::JsonArray ja_type, uint16_t *arr_u16_singleReadAddresses, 
                                                uint16_t *arr_u16_singleReadValues, size_t *ptr_singleReads);


GlobScop int cst_CurveBounds(ArduinoJson::JsonObject jo_curve, uint16_t *arr_u16_writeVal, size_t siz_request);

// FileScop int cst_SetCalculatedParameter(ArduinoJson::JsonObject jo_parameter, const char *sz_key, int32_t s32_value);
// FileScop int cst_ResetCalculatedParameter(ArduinoJson::JsonObject jo_parameter, const char *sz_key);





// FileScop int cst_FormatWriteValues(cst_ParameterType_t enm_type, const char *sz_key, int32_t s32_rawValue, 
//                                     uint16_t* arr_u16_writeAddr, uint16_t* arr_u16_writeVal, uint8_t u8_position);
// FileScop int cst_FormatResetWriteValues(cst_ParameterType_t enm_type, const char *sz_key, uint16_t* arr_u16_writeAddr,
// uint16_t* arr_u16_writeVal, uint8_t u8_position);

FileScop uint8_t cst_getUtilCmdState(void);
FileScop void cst_setUtilCmdState(uint8_t u8_state);

FileScop int cst_WriteControl(cst_writeType_t cst_enum_writeType, size_t siz_request, 
                                        uint16_t* arr_u16_writeAddr, uint16_t* arr_u16_writeVal);

FileScop std::vector<std::pair<float, float>> CST_piecewiseLinearInterpolation(const std::vector<float>& vec_f_xRaw, 
                                        const std::vector<float>& vec_f_yRaw, uint8_t u8_numPoints);

FileScop float cst_interpolateY(float f_xNew, float f_x1, float f_y1, float f_x2, float f_y2); 

FileScop void cst_LogAddrVal(const uint16_t* arr_u16_addresses, const uint16_t* arr_u16_values, size_t size);



/**********************************************************************************************************************
 * @cond                                              GLOBAL SCOPE                                             @endcond
 *********************************************************************************************************************/
GlobScop int CST_Init()
{

    StackScop int res;
    res = pthread_mutex_init(&cst_mtx_lockedRegisters, NULL);
    res = pthread_mutex_init(&cst_mtx_modelStatus, NULL);
    res = pthread_mutex_init(&cst_mtx_utilCmdState, NULL);

    // Retrieve the Translation Map
    StackScop std::fstream fs_translationMapFileStream(CST_TRANSLATION_MAP_FILE_PATH, std::fstream::in);
    StackScop ArduinoJson::DeserializationError err_DeserializationSt = 
                                ArduinoJson::deserializeJson(jd_translationMap, fs_translationMapFileStream);
    if (err_DeserializationSt != ArduinoJson::DeserializationError::Ok)
    {
        LOG_ERR("CST_Init() Failed due to DeserializationError: %s", err_DeserializationSt.c_str());
        return -EXIT_FAILURE;
    }

    CST_IAL_Init();

    // // Reset lock registry on reinitialization 
    // if(IAL_Mbs_ResetLockRegistry(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID) != EXIT_SUCCESS)
    // {
    //     LOG_ERR("CST_Init() Failed to reset locks to INV");
    //     return -EXIT_FAILURE;
    // }

    StackScop size_t siz_request = 5;
    StackScop uint16_t arr_u16_values[siz_request] = {0};
    StackScop uint16_t arr_u16_addrs[siz_request] = {0x3443, 0x3145, 0x221D, 0x3445, 0x340F};
    StackScop int32_t arr_u16_res[siz_request] = {0};
    // 0x3443 Operating Mode --- 0：disable, 1；sell first, 2：limit grid consumption, 3：zero export
    // 0x3145 BATT SOC --- 0.1 % resolution
    // 0x221D Grid Reconnect Time --- 0.1s resolution
    // 0x3445 Grid function --- Bit 0: Ban fast check of grid voltage; Bit 1: 10min over voltage; Bit 2: Grid Charge Enable; Bit 3: Test; Command 1; Bit 4: Time of Use enable; Bit 5: Test Command 2; Bit 6: Battery Priority; Bit 7: Charge solar only
    // 0x340F Battery Type --- 0: DC, 1: Lead Acid, 2: Lithium
    if(IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_addrs, 
                                        arr_u16_values, siz_request) != EXIT_SUCCESS)
    {
        LOG_ERR("CST_Init() Failed to read from INV");
        // return -EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop int CST_GetCapability(SE_DERCapability_t *ptr_capability)
{
    // Initialize the flags indicating which attribtues are supported
    ptr_capability->_flags = 0;

    // Set the Parameter Type for the function
    StackScop cst_ParameterType_t enm_type = CST_CAPABILITY;
    
    // Retrieve the capability mapping from the translation map
    StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
    cst_GetParameterTypeString(enm_type, sz_type);
    StackScop ArduinoJson::JsonObject jo_capability = jd_translationMap[sz_type];
    if (jo_capability.isNull())
    {
        LOG_ERR("CST_GetCapability() Failed to retrieve the capability from the translation map");
        return -EXIT_FAILURE;
    }

    // Read all of the single registers required
    StackScop uint16_t arr_u16_singleReadAddresses[CST_MAX_READ_SIZE] = {0};
    StackScop uint16_t arr_u16_singleReadValues[CST_MAX_READ_SIZE] = {0};
    StackScop size_t siz_singleReads = 0;
    cst_GetSingleReadValuesFromObject(jo_capability, arr_u16_singleReadAddresses, arr_u16_singleReadValues, &siz_singleReads);

    // Retrieve the modesSupported
    if (cst_GetDERControlType((uint8_t*)&ptr_capability->modesSupported, enm_type, "modesSupported") == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgModesSupported_exists;
    }
    LOG_INF("modesSupported: %#x %#x %#x %#x", ptr_capability->modesSupported[3], ptr_capability->modesSupported[2],
            ptr_capability->modesSupported[1], ptr_capability->modesSupported[0]);

    // Retrieve the rtgAbnormalCategory
    if (cst_GetParameterValue(enm_type, "rtgAbnormalCategory", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgAbnormalCategory) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgAbnormalCategory_exists;
    }
    LOG_INF("rtgAbnormalCategory: %d", ptr_capability->rtgAbnormalCategory);

    // Retrieve the rtgMaxA
    if (cst_GetCurrentRMS(&ptr_capability->rtgMaxA, enm_type, "rtgMaxA") == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgMaxA_exists;
    }
    LOG_INF("rtgMaxA: %d %d", ptr_capability->rtgMaxA.value, ptr_capability->rtgMaxA.multiplier);
    /** @todo remove cst_GetCurrentRMS() and make into generic function */

    // Retrieve the rtgMaxAh
    if (cst_GetParameterValue(enm_type, "rtgMaxAh", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgMaxAh.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgMaxAh_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgMaxAh", &ptr_capability->rtgMaxAh.multiplier);
    LOG_INF("rtgMaxAh: %d %d", ptr_capability->rtgMaxAh.value, ptr_capability->rtgMaxAh.multiplier);

    // Retrieve the rtgMaxChargeRateVA
    if (cst_GetParameterValue(enm_type, "rtgMaxChargeRateVA", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgMaxChargeRateVA.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgMaxChargeRateVA_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgMaxChargeRateVA", &ptr_capability->rtgMaxChargeRateVA.multiplier);
    LOG_INF("rtgMaxChargeRateVA: %d %d",    ptr_capability->rtgMaxChargeRateVA.value, 
                                            ptr_capability->rtgMaxChargeRateVA.multiplier);

    // Retrieve the rtgMaxChargeRateW
    if (cst_GetParameterValue(enm_type, "rtgMaxChargeRateW", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgMaxChargeRateW.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgMaxChargeRateW_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgMaxChargeRateW", &ptr_capability->rtgMaxChargeRateW.multiplier);
    LOG_INF("rtgMaxChargeRateW: %d %d", ptr_capability->rtgMaxChargeRateW.value, 
                                        ptr_capability->rtgMaxChargeRateW.multiplier);

    // Retrieve the rtgMaxDischargeRateVA
    if (cst_GetParameterValue(enm_type, "rtgMaxDischargeRateVA", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgMaxDischargeRateVA.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgMaxDischargeRateVA_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgMaxDischargeRateVA", &ptr_capability->rtgMaxDischargeRateVA.multiplier);
    LOG_INF("rtgMaxDischargeRateVA: %d %d", ptr_capability->rtgMaxDischargeRateVA.value, 
                                            ptr_capability->rtgMaxDischargeRateVA.multiplier);

    // Retrieve the rtgMaxDischargeRateW
    if (cst_GetParameterValue(enm_type, "rtgMaxDischargeRateW", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgMaxDischargeRateW.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgMaxDischargeRateW_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgMaxDischargeRateW", &ptr_capability->rtgMaxDischargeRateW.multiplier);
    LOG_INF("rtgMaxDischargeRateW: %d %d", ptr_capability->rtgMaxDischargeRateW.value, 
                                            ptr_capability->rtgMaxDischargeRateW.multiplier);

    // Retrieve the rtgMaxV
    if (cst_GetParameterValue(enm_type, "rtgMaxV", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgMaxV.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgMinV_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgMaxV", &ptr_capability->rtgMaxV.multiplier);
    LOG_INF("rtgMaxV: %d %d", ptr_capability->rtgMaxV.value, ptr_capability->rtgMaxV.multiplier);

    // Retrieve the rtgMaxVA
    if (cst_GetParameterValue(enm_type, "rtgMaxVA", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgMaxVA.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgMaxVA_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgMaxVA", &ptr_capability->rtgMaxVA.multiplier);
    LOG_INF("rtgMaxVA: %d %d", ptr_capability->rtgMaxVA.value, ptr_capability->rtgMaxVA.multiplier);

    // Retrieve the rtgMaxVar
    if (cst_GetParameterValue(enm_type, "rtgMaxVar", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgMaxVar.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgMaxVar_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgMaxVar", &ptr_capability->rtgMaxVar.multiplier);
    LOG_INF("rtgMaxVar: %d %d", ptr_capability->rtgMaxVar.value, ptr_capability->rtgMaxVar.multiplier);

    // Retrieve the rtgMaxVarNeg
    if (cst_GetParameterValue(enm_type, "rtgMaxVarNeg", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgMaxVarNeg.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgMaxVarNeg_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgMaxVarNeg", &ptr_capability->rtgMaxVarNeg.multiplier);
    LOG_INF("rtgMaxVarNeg: %d %d", ptr_capability->rtgMaxVarNeg.value, ptr_capability->rtgMaxVarNeg.multiplier);

    // Retrieve the rtgMaxW
    if (cst_GetParameterValue(enm_type, "rtgMaxW", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgMaxW.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgMaxW_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgMaxW", &ptr_capability->rtgMaxW.multiplier);
    LOG_INF("rtgMaxW: %d %d", ptr_capability->rtgMaxW.value, ptr_capability->rtgMaxW.multiplier);

    // Retrieve the rtgMaxWh
    if (cst_GetParameterValue(enm_type, "rtgMaxWh", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgMaxWh.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgMaxWh_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgMaxWh", &ptr_capability->rtgMaxWh.multiplier);
    LOG_INF("rtgMaxWh: %d %d", ptr_capability->rtgMaxWh.value, ptr_capability->rtgMaxWh.multiplier);

    // Retrieve the rtgMinPFOverExcited
    if (cst_GetParameterValue(enm_type, "rtgMinPFOverExcited", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                        siz_singleReads, (int32_t*)&ptr_capability->rtgMinPFOverExcited.displacement) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgMinPFOverExcited_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgMinPFOverExcited", &ptr_capability->rtgMinPFOverExcited.multiplier);
    LOG_INF("rtgMinPFOverExcited: %d %d", ptr_capability->rtgMinPFOverExcited.displacement, 
                                            ptr_capability->rtgMinPFOverExcited.multiplier);

    // Retrieve the rtgMinPFUnderExcited
    if (cst_GetParameterValue(enm_type, "rtgMinPFUnderExcited", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                        siz_singleReads, (int32_t*)&ptr_capability->rtgMinPFUnderExcited.displacement) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgMinPFUnderExcited_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgMinPFUnderExcited", &ptr_capability->rtgMinPFUnderExcited.multiplier);
    LOG_INF("rtgMinPFUnderExcited: %d %d", ptr_capability->rtgMinPFUnderExcited.displacement, 
                                            ptr_capability->rtgMinPFUnderExcited.multiplier);

    // Retrieve the rtgMinV
    if (cst_GetParameterValue(enm_type, "rtgMinV", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgMinV.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgMinV_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgMinV", &ptr_capability->rtgMinV.multiplier);
    LOG_INF("rtgMinV: %d %d", ptr_capability->rtgMinV.value, ptr_capability->rtgMinV.multiplier);
    
    // Retrieve the rtgNormalCategory
    if (cst_GetParameterValue(enm_type, "rtgNormalCategory", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgNormalCategory) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgNormalCategory_exists;
    }
    LOG_INF("rtgNormalCategory: %d", ptr_capability->rtgNormalCategory);

    // Retrieve the rtgOverExcitedPF
    if (cst_GetParameterValue(enm_type, "rtgOverExcitedPF", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgOverExcitedPF.displacement) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgOverExcitedPF_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgOverExcitedPF", &ptr_capability->rtgOverExcitedPF.multiplier);
    LOG_INF("rtgOverExcitedPF: %d %d", ptr_capability->rtgOverExcitedPF.displacement, 
                                        ptr_capability->rtgOverExcitedPF.multiplier);
    
    // Retrieve the rtgOverExcitedW
    if (cst_GetParameterValue(enm_type, "rtgOverExcitedW", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgOverExcitedW.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgOverExcitedW_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgOverExcitedW", &ptr_capability->rtgOverExcitedW.multiplier);
    LOG_INF("rtgOverExcitedW: %d %d", ptr_capability->rtgOverExcitedW.value, ptr_capability->rtgOverExcitedW.multiplier);

    // Retrieve the rtgReactiveSusceptance
    if (cst_GetParameterValue(enm_type, "rtgReactiveSusceptance", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgReactiveSusceptance.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgReactiveSusceptance_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgReactiveSusceptance", &ptr_capability->rtgReactiveSusceptance.multiplier);
    LOG_INF("rtgReactiveSusceptance: %d %d", ptr_capability->rtgReactiveSusceptance.value, 
                                            ptr_capability->rtgReactiveSusceptance.multiplier);

    // Retrieve the rtgUnderExcitedPF
    if (cst_GetParameterValue(enm_type, "rtgUnderExcitedPF", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgUnderExcitedPF.displacement) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgUnderExcitedPF_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgUnderExcitedPF", &ptr_capability->rtgUnderExcitedPF.multiplier);
    LOG_INF("rtgUnderExcitedPF: %d %d", ptr_capability->rtgUnderExcitedPF.displacement, 
                                        ptr_capability->rtgUnderExcitedPF.multiplier);

    // Retrieve the rtgUnderExcitedW
    if (cst_GetParameterValue(enm_type, "rtgUnderExcitedW", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgUnderExcitedW.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgUnderExcitedW_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgUnderExcitedW", &ptr_capability->rtgUnderExcitedW.multiplier);
    LOG_INF("rtgUnderExcitedW: %d %d", ptr_capability->rtgUnderExcitedW.value, ptr_capability->rtgUnderExcitedW.multiplier);

    // Retrieve the rtgVNom
    if (cst_GetParameterValue(enm_type, "rtgVNom", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->rtgVNom.value) == EXIT_SUCCESS)
    {
        ptr_capability->_flags |= SE_rtgVNom_exists;
    }
    cst_GetParameterMultiplier(enm_type, "rtgVNom", &ptr_capability->rtgVNom.multiplier);
    LOG_INF("rtgVNom: %d %d", ptr_capability->rtgVNom.value, ptr_capability->rtgVNom.multiplier);
    
    // Retrieve the type
    if (cst_GetParameterValue(enm_type, "type", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_capability->type) == -EINVAL)
    {
        ptr_capability->_flags |= SE_type_exists;
    }
    LOG_INF("type: %d", ptr_capability->type);

    LOG_INF("Capability Flags: %#x", ptr_capability->_flags);
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop int CST_GetSettings(SE_DERSettings_t *ptr_settings)
{
    // Initialize the flags indicating which attribtues are supported
    ptr_settings->_flags = 0;

    // Set the Parameter Type for the function
    StackScop cst_ParameterType_t enm_type = CST_SETTING;
    
    // Retrieve the settings mapping from the translation map
    StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
    cst_GetParameterTypeString(enm_type, sz_type);
    StackScop ArduinoJson::JsonObject jo_settings = jd_translationMap[sz_type];
    if (jo_settings.isNull())
    {
        LOG_ERR("CST_GetSettings() Failed to retrieve the settings from the translation map");
        return -EXIT_FAILURE;
    }

    // Read all of the single registers required
    StackScop uint16_t arr_u16_singleReadAddresses[CST_MAX_READ_SIZE] = {0};
    StackScop uint16_t arr_u16_singleReadValues[CST_MAX_READ_SIZE] = {0};
    StackScop size_t siz_singleReads = 0;
    cst_GetSingleReadValuesFromObject(jo_settings, arr_u16_singleReadAddresses, arr_u16_singleReadValues, &siz_singleReads);

    // Retrieve the modesEnabled
    if (cst_GetDERControlType((uint8_t*)&ptr_settings->modesEnabled, enm_type, "modesEnabled") == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_modesEnabled_exists;
    }
    LOG_INF("modesEnabled: %#x %#x %#x %#x", ptr_settings->modesEnabled[3], ptr_settings->modesEnabled[2],
            ptr_settings->modesEnabled[1], ptr_settings->modesEnabled[0]);
    /** @todo Fix Setting DERControlType */

    // Retrieve the setESDelay
    if (cst_GetParameterValue(enm_type, "setESDelay", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setESDelay) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setESDelay_exists;
    }
    LOG_INF("setESDelay: %d", ptr_settings->setESDelay);
    /** @todo Check if this struct needs to be made into a uint32 instead of uint16 */

    // Retrieve the setESHighFreq
    if (cst_GetParameterValue(enm_type, "setESHighFreq", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setESHighFreq) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setESHighFreq_exists;
    }
    LOG_INF("setESHighFreq: %d", ptr_settings->setESHighFreq);

    // Retrieve the setESRampTms
    if (cst_GetParameterValue(enm_type, "setESRampTms", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setESRampTms) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setESRampTms_exists;
    }
    LOG_INF("setESRampTms: %d", ptr_settings->setESRampTms);

    // Retrieve the setMaxV
    if (cst_GetParameterValue(enm_type, "setMaxV", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMaxV.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMaxV_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMaxV", &ptr_settings->setMaxV.multiplier);
    LOG_INF("setMaxV: %d %d", ptr_settings->setMaxV.value, ptr_settings->setMaxV.multiplier);

    // Retrieve the setMinPFOverExcited
    if (cst_GetParameterValue(enm_type, "setMinPFOverExcited", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMinPFOverExcited.displacement) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMinPFOverExcited_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMinPFOverExcited", &ptr_settings->setMinPFOverExcited.multiplier);
    LOG_INF("setMinPFOverExcited: %d %d", ptr_settings->setMinPFOverExcited.displacement, 
                                        ptr_settings->setMinPFOverExcited.multiplier);
    
    // Retrieve the setMinPFUnderExcited
    if (cst_GetParameterValue(enm_type, "setMinPFUnderExcited", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMinPFUnderExcited.displacement) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMinPFUnderExcited_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMinPFUnderExcited", &ptr_settings->setMinPFUnderExcited.multiplier);
    LOG_INF("setMinPFUnderExcited: %d %d", ptr_settings->setMinPFUnderExcited.displacement, 
                                        ptr_settings->setMinPFUnderExcited.multiplier);
    
    // Retrieve the setMinV
    if (cst_GetParameterValue(enm_type, "setMinV", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMinV.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMinV_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMinV", &ptr_settings->setMinV.multiplier);
    LOG_INF("setMinV: %d %d", ptr_settings->setMinV.value, ptr_settings->setMinV.multiplier);

    // Retrieve the setVNom
    if (cst_GetParameterValue(enm_type, "setVNom", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setVNom.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setVNom_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setVNom", &ptr_settings->setVNom.multiplier);
    LOG_INF("setVNom: %d %d", ptr_settings->setVNom.value, ptr_settings->setVNom.multiplier);

    // Retrieve the setESHighVolt
    if (cst_GetParameterValue(enm_type, "setESHighVolt", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setESHighVolt) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setESHighVolt_exists;
    }
    LOG_INF("setESHighVolt: %d", ptr_settings->setESHighVolt);

    // Retrieve the setESLowFreq
    if (cst_GetParameterValue(enm_type, "setESLowFreq", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setESLowFreq) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setESLowFreq_exists;
    }
    LOG_INF("setESLowFreq: %d", ptr_settings->setESLowFreq);

    // Retrieve the setESLowVolt
    if (cst_GetParameterValue(enm_type, "setESLowVolt", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setESLowVolt) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setESLowVolt_exists;
    }
    LOG_INF("setESLowVolt: %d", ptr_settings->setESLowVolt);

    // Retrieve the setESRandomDelay
    if (cst_GetParameterValue(enm_type, "setESRandomDelay", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setESRandomDelay) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setESRandomDelay_exists;
    }
    LOG_INF("setESRandomDelay: %d", ptr_settings->setESRandomDelay);

    // Retrieve the setGradW
    if (cst_GetParameterValue(enm_type, "setGradW", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setGradW) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setGradW_exists;
    }
    LOG_INF("setGradW: %d", ptr_settings->setGradW);

    // Retrieve the setMaxA
    if (cst_GetParameterValue(enm_type, "setMaxA", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMaxA.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMaxA_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMaxA", &ptr_settings->setMaxA.multiplier);
    LOG_INF("setMaxA: %d %d", ptr_settings->setMaxA.value, ptr_settings->setMaxA.multiplier);

    // Retrieve the setMaxAh
    if (cst_GetParameterValue(enm_type, "setMaxAh", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMaxAh.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMaxAh_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMaxAh", &ptr_settings->setMaxAh.multiplier);
    LOG_INF("setMaxAh: %d %d", ptr_settings->setMaxAh.value, ptr_settings->setMaxAh.multiplier);

    // Retrieve the setMaxChargeRateVA
    if (cst_GetParameterValue(enm_type, "setMaxChargeRateVA", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMaxChargeRateVA.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMaxChargeRateVA_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMaxChargeRateVA", &ptr_settings->setMaxChargeRateVA.multiplier);
    LOG_INF("setMaxChargeRateVA: %d %d", ptr_settings->setMaxChargeRateVA.value, 
                                            ptr_settings->setMaxChargeRateVA.multiplier);

    // Retrieve the setMaxChargeRateW
    if (cst_GetParameterValue(enm_type, "setMaxChargeRateW", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMaxChargeRateW.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMaxChargeRateW_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMaxChargeRateW", &ptr_settings->setMaxChargeRateW.multiplier);
    LOG_INF("setMaxChargeRateW: %d %d", ptr_settings->setMaxChargeRateW.value, 
                                        ptr_settings->setMaxChargeRateW.multiplier);

    // Retrieve the setMaxDischargeRateVA
    if (cst_GetParameterValue(enm_type, "setMaxDischargeRateVA", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMaxDischargeRateVA.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMaxDischargeRateVA_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMaxDischargeRateVA", &ptr_settings->setMaxDischargeRateVA.multiplier);
    LOG_INF("setMaxDischargeRateVA: %d %d", ptr_settings->setMaxDischargeRateVA.value, 
                                            ptr_settings->setMaxDischargeRateVA.multiplier);
    
    // Retrieve the setMaxDischargeRateW
    if (cst_GetParameterValue(enm_type, "setMaxDischargeRateW", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMaxDischargeRateW.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMaxDischargeRateW_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMaxDischargeRateW", &ptr_settings->setMaxDischargeRateW.multiplier);
    LOG_INF("setMaxDischargeRateW: %d %d", ptr_settings->setMaxDischargeRateW.value, 
                                            ptr_settings->setMaxDischargeRateW.multiplier);

    // Retrieve the setMaxVA
    if (cst_GetParameterValue(enm_type, "setMaxVA", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMaxVA.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMaxVA_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMaxVA", &ptr_settings->setMaxVA.multiplier);
    LOG_INF("setMaxVA: %d %d", ptr_settings->setMaxVA.value, ptr_settings->setMaxVA.multiplier);

    // Retrieve the setMaxVar
    if (cst_GetParameterValue(enm_type, "setMaxVar", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMaxVar.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMaxVar_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMaxVar", &ptr_settings->setMaxVar.multiplier);
    LOG_INF("setMaxVar: %d %d", ptr_settings->setMaxVar.value, ptr_settings->setMaxVar.multiplier);

    // Retrieve the setMaxVarNeg
    if (cst_GetParameterValue(enm_type, "setMaxVarNeg", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMaxVarNeg.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMaxVarNeg_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMaxVarNeg", &ptr_settings->setMaxVarNeg.multiplier);
    LOG_INF("setMaxVarNeg: %d %d", ptr_settings->setMaxVarNeg.value, ptr_settings->setMaxVarNeg.multiplier);

    // Retrieve the setMaxW
    if (cst_GetParameterValue(enm_type, "setMaxW", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMaxW.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMaxW_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMaxW", &ptr_settings->setMaxW.multiplier);
    LOG_INF("setMaxW: %d %d", ptr_settings->setMaxW.value, ptr_settings->setMaxW.multiplier);

    // Retrieve the setMaxWh
    if (cst_GetParameterValue(enm_type, "setMaxWh", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMaxWh.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMaxWh_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMaxWh", &ptr_settings->setMaxWh.multiplier);
    LOG_INF("setMaxWh: %d %d", ptr_settings->setMaxWh.value, ptr_settings->setMaxWh.multiplier);

    // Retrieve the setMinPF
    if (cst_GetParameterValue(enm_type, "setMinPF", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMinPF.displacement) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMinPF_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMinPF", &ptr_settings->setMinPF.multiplier);
    LOG_INF("setMinPF: %d %d", ptr_settings->setMinPF.displacement, ptr_settings->setMinPF.multiplier);

    // Retrieve the setMinPFNeg
    if (cst_GetParameterValue(enm_type, "setMinPFNeg", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setMinPFNeg.displacement) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setMinPFNeg_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setMinPFNeg", &ptr_settings->setMinPFNeg.multiplier);
    LOG_INF("setMinPFNeg: %d %d", ptr_settings->setMinPFNeg.displacement, ptr_settings->setMinPFNeg.multiplier);

    // Retrieve the setSoftGradW
    if (cst_GetParameterValue(enm_type, "setSoftGradW", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setSoftGradW) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setSoftGradW_exists;
    }
    LOG_INF("setSoftGradW: %d", ptr_settings->setSoftGradW);
    
    // Retrieve the setVRef
    if (cst_GetParameterValue(enm_type, "setVRef", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setVRef.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setVRef_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setVRef", &ptr_settings->setVRef.multiplier);
    LOG_INF("setVRef: %d %d", ptr_settings->setVRef.value, ptr_settings->setVRef.multiplier);
    
    
    // Retrieve the setVRefOfs
    if (cst_GetParameterValue(enm_type, "setVRefOfs", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_settings->setVRefOfs.value) == EXIT_SUCCESS)
    {
        ptr_settings->_flags |= SE_setVRefOfs_exists;
    }
    cst_GetParameterMultiplier(enm_type, "setVRefOfs", &ptr_settings->setVRefOfs.multiplier);
    LOG_INF("setVRefOfs: %d %d", ptr_settings->setVRefOfs.value, ptr_settings->setVRefOfs.multiplier);

    LOG_INF("Settings Flags: %#x", ptr_settings->_flags);
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop int CST_GetStatus(SE_DERStatus_t *ptr_status)
{
    // Initialize the flags indicating which attribtues are supported
    ptr_status->_flags = 0;

    // Set the Parameter Type for the function
    StackScop cst_ParameterType_t enm_type = CST_STATUS;
    
    // Retrieve the status mapping from the translation map
    StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
    cst_GetParameterTypeString(enm_type, sz_type);
    StackScop ArduinoJson::JsonObject jo_status = jd_translationMap[sz_type];
    if (jo_status.isNull())
    {
        LOG_ERR("CST_GetStatus() Failed to retrieve the status from the translation map");
        return -EXIT_FAILURE;
    }

    // Read all of the single registers required
    StackScop uint16_t arr_u16_singleReadAddresses[CST_MAX_READ_SIZE] = {0};
    StackScop uint16_t arr_u16_singleReadValues[CST_MAX_READ_SIZE] = {0};
    StackScop size_t siz_singleReads = 0;
    cst_GetSingleReadValuesFromObject(jo_status, arr_u16_singleReadAddresses, arr_u16_singleReadValues, &siz_singleReads);

    // Retrieve the genConnectStatus
    if (cst_GetParameterValue(enm_type, "genConnectStatus", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_status->genConnectStatus.value) == EXIT_SUCCESS)
    {
        ptr_status->_flags |= SE_genConnectStatus_exists;
        LOG_INF("genConnectStatus: %d", ptr_status->genConnectStatus.value);
    }
    else
    {
        LOG_INF("genConnectStatus: Not Supported");
    }
    

    // Retrieve the invertrStatus
    if (cst_GetParameterValue(enm_type, "inverterStatus", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_status->inverterStatus.value) == EXIT_SUCCESS)
    {
        ptr_status->_flags |= SE_inverterStatus_exists;
        LOG_INF("inverterStatus: %d", ptr_status->inverterStatus.value);
    }
    else
    {
        LOG_INF("inverterStatus: Not Supported");
        /** @todo Implement calculation to support this later if required by i3e application */
    }
    

    // Retrieve the localControlModeStatus
    if (cst_GetParameterValue(enm_type, "localControlModeStatus", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_status->localControlModeStatus.value) == EXIT_SUCCESS)
    {
        ptr_status->_flags |= SE_localControlModeStatus_exists;
        LOG_INF("localControlModeStatus: %d", ptr_status->localControlModeStatus.value);
    }
    else
    {
        LOG_INF("localControlModeStatus: Not Supported");
    }

    // Retrieve the manufacturerStatus
    if (cst_GetParameterValue(enm_type, "manufacturerStatus", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_status->manufacturerStatus.value) == EXIT_SUCCESS)
    {
        ptr_status->_flags |= SE_manufacturerStatus_exists;
        LOG_INF("manufacturerStatus: %d", ptr_status->manufacturerStatus.value);
    }
    else
    {
        LOG_INF("manufacturerStatus: Not Supported");
    }

    // Retrieve the operationalModeStatus
    if (cst_GetParameterValue(enm_type, "operationalModeStatus", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_status->operationalModeStatus.value) == EXIT_SUCCESS)
    {
        ptr_status->_flags |= SE_operationalModeStatus_exists;
        LOG_INF("operationalModeStatus: %d", ptr_status->operationalModeStatus.value);
    }
    else
    {
        LOG_INF("operationalModeStatus: Not Supported");
    }

    // Retrieve the stateOfChargeStatus
    if (cst_GetParameterValue(enm_type, "stateOfChargeStatus", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_status->stateOfChargeStatus.value) == EXIT_SUCCESS)
    {
        ptr_status->_flags |= SE_stateOfChargeStatus_exists;
        LOG_INF("stateOfChargeStatus: %d", ptr_status->stateOfChargeStatus.value);
    }
    else
    {
        LOG_INF("stateOfChargeStatus: Not Supported");
    }

    // Retrieve the storageModeStatus
    if (cst_GetParameterValue(enm_type, "storageModeStatus", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_status->storageModeStatus.value) == EXIT_SUCCESS)
    {
        ptr_status->_flags |= SE_storageModeStatus_exists;
        LOG_INF("storageModeStatus: %d", ptr_status->storageModeStatus.value);
    }
    else
    {
        LOG_INF("storageModeStatus: Not Supported");
    }

    // Retrieve the storConnectStatus
    if (cst_GetParameterValue(enm_type, "storConnectStatus", arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                            siz_singleReads, (int32_t*)&ptr_status->storConnectStatus.value) == EXIT_SUCCESS)
    {
        ptr_status->_flags |= SE_storConnectStatus_exists;
        LOG_INF("storConnectStatus: %d", ptr_status->storConnectStatus.value);
    }
    else
    {
        LOG_INF("storConnectStatus: Not Supported");
    }

    LOG_INF("Status Flags: %#x", ptr_status->_flags);
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop int CST_GetMirrorMeterReadings(std::vector<SE_MirrorMeterReading_t> *ptr_meterReadings, 
                                        bool bol_isInitalReading)
{
    // Set the Parameter Type for the function
    StackScop cst_ParameterType_t enm_type = CST_METER_READING;
    
    // Retrieve the settings mapping from the translation map
    StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
    cst_GetParameterTypeString(enm_type, sz_type);
    StackScop ArduinoJson::JsonArray ja_meterReadings = jd_translationMap[sz_type];
    if (ja_meterReadings.isNull())
    {
        LOG_ERR("CST_GetMirrorMeterReadings() Failed to retrieve the Mirror Meter Readings from the translation map");
        return -EXIT_FAILURE;
    }
    
    StackScop uint16_t arr_u16_singleReadAddresses[CST_MAX_READ_SIZE] = {0};
    StackScop uint16_t arr_u16_singleReadValues[CST_MAX_READ_SIZE] = {0};
    StackScop size_t siz_singleReads = 0;
    if (!bol_isInitalReading)
    {
        // Read all of the single registers required
        cst_GetSingleReadValuesFromArray(ja_meterReadings, arr_u16_singleReadAddresses, arr_u16_singleReadValues, &siz_singleReads);
    }

    StackScop uint8_t u8_vectorIndex = 0;
    for (ArduinoJson::JsonObject jo_readingMap : ja_meterReadings)
    {
        StackScop SE_MirrorMeterReading_t str_meterReading;
        // Initialize the flags indicating which attribtues are supported
        str_meterReading._flags = 0;

        // Retrive the data for the meter reading from the map
        if (!jo_readingMap["desc"].isNull())
        {
            str_meterReading._flags |= SE_description_exists;
            strcpy(str_meterReading.description, jo_readingMap["desc"].as<const char*>());
        }
        
        if (!jo_readingMap["type"].isNull())
        {
            str_meterReading._flags |= SE_ReadingType_exists;
            StackScop ArduinoJson::JsonObject jo_readingType = jo_readingMap["type"];
            str_meterReading.ReadingType._flags = 0;

            if (bol_isInitalReading)
            {
                // Retrieve the ReadingType data from the map
                if (!jo_readingType["acc_beh"].isNull())
                {
                    str_meterReading.ReadingType._flags |= SE_accumulationBehaviour_exists;
                    str_meterReading.ReadingType.accumulationBehaviour = jo_readingType["acc_beh"].as<uint8_t>();
                }
                if (!jo_readingType["commodity"].isNull())
                {
                    str_meterReading.ReadingType._flags |= SE_commodity_exists;
                    str_meterReading.ReadingType.commodity = jo_readingType["commodity"].as<uint8_t>();
                }
                if (!jo_readingType["data_qual"].isNull())
                {
                    str_meterReading.ReadingType._flags |= SE_dataQualifier_exists;
                    str_meterReading.ReadingType.dataQualifier = jo_readingType["data_qual"].as<uint8_t>();
                }
                if (!jo_readingType["kind"].isNull())
                {
                    str_meterReading.ReadingType._flags |= SE_kind_exists;
                    str_meterReading.ReadingType.kind = jo_readingType["kind"].as<uint8_t>();
                }
                if (!jo_readingType["phase"].isNull())
                {
                    str_meterReading.ReadingType._flags |= SE_phase_exists;
                    str_meterReading.ReadingType.phase = jo_readingType["phase"].as<uint8_t>();
                }
                if (!jo_readingType["mult"].isNull())
                {
                    str_meterReading.ReadingType._flags |= SE_powerOfTenMultiplier_exists;
                    str_meterReading.ReadingType.powerOfTenMultiplier = jo_readingType["mult"].as<int8_t>();
                }
                if (!jo_readingType["uom"].isNull())
                {
                    str_meterReading.ReadingType._flags |= SE_uom_exists;
                    str_meterReading.ReadingType.uom = jo_readingType["uom"].as<uint8_t>();
                }
            }

            // Retrieve the value and flow direction of the reading 
            str_meterReading.Reading._flags = 0;
            if (!bol_isInitalReading) 
            {
                str_meterReading._flags |= SE_Reading_exists;
                StackScop int32_t s32_tempReading = 0;
                if (cst_GetParameterValue(enm_type, str_meterReading.description, arr_u16_singleReadAddresses, 
                                        arr_u16_singleReadValues, siz_singleReads, &s32_tempReading) == EXIT_SUCCESS)
                {
                    str_meterReading.Reading.value = static_cast<int16_t>(s32_tempReading);
                    str_meterReading.Reading._flags |= SE_value_exists;
                    if (str_meterReading.Reading.value > 0)
                    {
                        str_meterReading.ReadingType.flowDirection = jo_readingType["rev_flow"].as<uint8_t>();
                    }
                    else
                    {
                        str_meterReading.ReadingType.flowDirection = jo_readingType["fwd_flow"].as<uint8_t>();
                    }
                    str_meterReading.ReadingType._flags |= SE_flowDirection_exists;
                }
            }
        }
        // Add the meter reading to the vector
        ptr_meterReadings->at(u8_vectorIndex) = str_meterReading;
        u8_vectorIndex++;
    }
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop size_t CST_GetMirrorMeterReadingsSize(void)
{
    // Set the Parameter Type for the function
    StackScop cst_ParameterType_t enm_type = CST_METER_READING;
    
    // Retrieve the settings mapping from the translation map
    StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
    cst_GetParameterTypeString(enm_type, sz_type);
    StackScop ArduinoJson::JsonArray ja_meterReadings = jd_translationMap[sz_type];
    return ja_meterReadings.size();
    return 0;

}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop CST_LogEventOccurred_t CST_HasLogEventOccurred(SE_LogEvent_t *ptr_logEvent)
{
//     // Initialize the flags indicating which attribtues are supported
//     ptr_logEvent->_flags = 0;

//     // Set the Parameter Type for the function
//     StackScop cst_ParameterType_t enm_type = CST_LOG_EVENT;
    
//     // Retrieve the log evenet mapping from the translation map
//     StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
//     cst_GetParameterTypeString(enm_type, sz_type);
//     StackScop ArduinoJson::JsonObject jo_alarmStatus = jd_translationMap[sz_type]["alarmStatus"];
//     if (jo_alarmStatus.isNull())
//     {
//         LOG_ERR("CST_HasLogEventOccurred() Failed to retrieve the alarmStatus from the translation map");
//         return CST_LOG_EVENT_ERROR;
//     }

//     // Read the alarmStatus from the inverter
//     StackScop size_t siz_request = jo_alarmStatus["b_len"].as<uint8_t>() / CST_SINGLE_REGISTER_SIZE;
//     StackScop uint16_t arr_u16_values[siz_request] = {0};
//     if (IAL_Mbs_ReadMultipleRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, jo_alarmStatus["addr"].as<uint16_t>(), 
//                                         arr_u16_values, siz_request) != EXIT_SUCCESS)
//     {
//         LOG_ERR("CST_HasLogEventOccurred() Failed to read the alarmStatus from the inverter");
//         return CST_LOG_EVENT_ERROR;
//     }
//     StackScop uint32_t u32_alarmStatus = (arr_u16_values[1] << 16) | arr_u16_values[0];

//     // Compare the alarmStatus read to the previously read value
//     StackScop bool bol_LogEventOccurred = false;
//     for (uint8_t i = 0; i <= jo_alarmStatus["b_range"].as<uint8_t>(); i++)
//     {
//         // Isolate the alarm bit
//         StackScop uint32_t u32_bitMask = 0b1 << i;
//         StackScop bool bol_alarm = (u32_alarmStatus & u32_bitMask) >> i;
//         StackScop bool bol_lastAlarm = (cst_lastAlarmStatus & u32_bitMask) >> i;
//         if ((bol_alarm != bol_lastAlarm) && bol_LogEventOccurred)
//         {
//             // Multiple alarms have occurred!
//             return CST_MULTIPLE_LOG_EVENTS_OCCURRED;
            
//         }
//         else if ((bol_alarm && !bol_lastAlarm) && !bol_LogEventOccurred)
//         {
//             // A new alarm has occurred
//             bol_LogEventOccurred = true;
//             ptr_logEvent->logEventCode = (2*i);

//             // Update the last alarm status (set the bit high)
//             cst_lastAlarmStatus |= u32_bitMask;
            
//         }
//         else if ((!bol_alarm && bol_lastAlarm) && !bol_LogEventOccurred)
//         {
//             // An alarm has been cleared
//             bol_LogEventOccurred = true;
//             ptr_logEvent->logEventCode = (2*i)+1;

//             // Update the last alarm status (set the bit low)
//             cst_lastAlarmStatus &= ~u32_bitMask;
//         }
//         else 
//         {
//             // No change in the alarm status
//         }
//     }

//     // Return the result
//     if (bol_LogEventOccurred)
//     {
//         return CST_LOG_EVENT_OCCURRED;
//     }
//     else 
//     {
//         return CST_LOG_EVENT_NOT_OCCURRED;
//     }
    return CST_LOG_EVENT_NOT_OCCURRED;

}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop int CST_SetControl(SE_DERControlBase_t *ptr_control)
{
    bool bol_validControl = false;

    // Set the Parameter Type for the function
    StackScop cst_ParameterType_t enm_type = CST_CONTROL;
    
    // Retrieve the status mapping from the translation map
    StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
    cst_GetParameterTypeString(enm_type, sz_type);
    StackScop ArduinoJson::JsonObject jo_controls = jd_translationMap[sz_type];
    if (jo_controls.isNull())
    {
        LOG_ERR("CST_SetControl() Failed to retrieve the controls from the translation map");
        return -EXIT_FAILURE;
    }

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // #TODO - Temporary exit on controls that are currently unimplemented - written here for easy recall later
    if ((ptr_control->_flags & SE_opModEnergize_exists)     || 
        (ptr_control->_flags & SE_opModConnect_exists)      ||
        (ptr_control->_flags & SE_opModFixedVar_exists)     ||
        (ptr_control->_flags & SE_opModFreqDroop_exists)    ||
        (ptr_control->_flags & SE_opModMaxLimW_exists)      ||
        (ptr_control->_flags & SE_opModTargetVar_exists)    ||
        (ptr_control->_flags & SE_rampTms_exists))
    {
        return EXIT_FAILURE;
    }
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    

    // // Control Event for opModEnergize
    // if (ptr_control->_flags & SE_opModEnergize_exists)
    // {
    //     bol_validControl = true;

    //     // Get initial input data required to perform control
    //     StackScop ArduinoJson::JsonObject jo_opModEnergize = jo_controls["opModEnergize"];

    //     // Set the Energize control mode accordingly
    //     StackScop uint16_t u16_setEnergize = (uint16_t)ptr_control->opModEnergize;
    //     StackScop uint16_t u16_setEnergizeAddr = jo_opModEnergize["addr"].as<uint16_t>(); 
    //     StackScop size_t siz_request = 1;

    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_SET_REG, siz_request, &u16_setEnergizeAddr, &u16_setEnergize);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_SetControl() failed to execute opModEnergize");
    //         return -EXIT_FAILURE;
    //     }
    // }

    // // Control Event for opModConnect
    // if (ptr_control->_flags & SE_opModConnect_exists)
    // {
    //     bol_validControl = true;

    //     // Get initial input data required to perform control
    //     StackScop ArduinoJson::JsonObject jo_opModConnect = jo_controls["opModConnect"];

    //     // Set the Connect control mode accordingly
    //     StackScop uint16_t u16_setEnergize = 0;
    //     if(ptr_control->opModConnect != true)
    //     {
    //         u16_setEnergize = 1;
    //     }
    //     StackScop uint16_t u16_setEnergizeAddr = jo_opModConnect["addr"].as<uint16_t>(); 
    //     StackScop size_t siz_request = 1;

    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_SET_REG, siz_request, &u16_setEnergizeAddr, &u16_setEnergize);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_SetControl() failed to execute opModConnect");
    //         return -EXIT_FAILURE;
    //     }
    // }

    // Control Event for opModFixedPFInjectW
    if (ptr_control->_flags & SE_opModFixedPFInjectW_exists)
    {
        bol_validControl = true;

        // Get initial input data required to perfrom control
        StackScop ArduinoJson::JsonObject jo_opModFixedPFInjectW = jo_controls["opModFixedPFInjectW"];
        
        StackScop int8_t s8_setPfMult = ptr_control->opModFixedPFInjectW.multiplier - 
                            jo_opModFixedPFInjectW["setpoint"]["mult"].as<int8_t>();
        StackScop float f_setPF = ptr_control->opModFixedPFInjectW.displacement * pow(10, s8_setPfMult);
        StackScop int16_t s16_setPF = ((int16_t)f_setPF);
        if(jo_opModFixedPFInjectW["setpoint"]["neg"].as<bool>())
        {
            s16_setPF = (-1)*(f_setPF);
        }
        if(s16_setPF < jo_opModFixedPFInjectW["setpoint"]["min"].as<int16_t>())
        {
            s16_setPF = jo_opModFixedPFInjectW["setpoint"]["min"].as<int16_t>();
        }
        if(s16_setPF > jo_opModFixedPFInjectW["setpoint"]["max"].as<int16_t>())
        {
            s16_setPF = jo_opModFixedPFInjectW["setpoint"]["max"].as<int16_t>();
        }

        // Set the target PF setpoint and enable the Fixed PF control mode
        StackScop uint16_t arr_u16_addresses[] =  {   
                                            (uint16_t)strtol(jo_opModFixedPFInjectW["setpoint"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE),
                                            (uint16_t)strtol(jo_opModFixedPFInjectW["en"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop uint16_t arr_u16_values[] =     {   
                                            (uint16_t)s16_setPF,
                                            jo_opModFixedPFInjectW["en"]["val"].as<uint16_t>()
        };
        StackScop size_t siz_request = sizeof(arr_u16_addresses) / sizeof(arr_u16_addresses[0]);

        cst_LogAddrVal(arr_u16_addresses, arr_u16_values, siz_request);

        StackScop int16_t s16_WriteSt = cst_WriteControl(CST_SET_REG, siz_request, arr_u16_addresses, arr_u16_values);
        if (s16_WriteSt != EXIT_SUCCESS)
        {
            LOG_ERR("CST_SetControl() failed to execute opModFixedPFInjectW");
            return -EXIT_FAILURE;
        }
    }

    // Control Event for opModFixedPFAbsorbW
    if (ptr_control->_flags & SE_opModFixedPFAbsorbW_exists)
    {
        bol_validControl = true;

        // Get initial input data required to perfrom control
        StackScop ArduinoJson::JsonObject jo_opModFixedPFAbsorbW = jo_controls["opModFixedPFAbsorbW"];
        
        StackScop int8_t s8_setPfMult = ptr_control->opModFixedPFAbsorbW.multiplier - 
                            jo_opModFixedPFAbsorbW["setpoint"]["mult"].as<int8_t>();
        StackScop float f_setPF = ptr_control->opModFixedPFAbsorbW.displacement * pow(10, s8_setPfMult);
        StackScop int16_t s16_setPF = ((int16_t)f_setPF);
        if(jo_opModFixedPFAbsorbW["setpoint"]["neg"].as<bool>())
        {
            s16_setPF = (-1)*(f_setPF);
        }
        if(s16_setPF < jo_opModFixedPFAbsorbW["setpoint"]["min"].as<int16_t>())
        {
            s16_setPF = jo_opModFixedPFAbsorbW["setpoint"]["min"].as<int16_t>();
        }
        if(s16_setPF > jo_opModFixedPFAbsorbW["setpoint"]["max"].as<int16_t>())
        {
            s16_setPF = jo_opModFixedPFAbsorbW["setpoint"]["max"].as<int16_t>();
        }

        // Set the target PF setpoint and enable the Fixed PF control mode
        StackScop uint16_t arr_u16_addresses[] =  {   
                                            (uint16_t)strtol(jo_opModFixedPFAbsorbW["setpoint"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE), 
                                            (uint16_t)strtol(jo_opModFixedPFAbsorbW["en"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop uint16_t arr_u16_values[] =     {   
                                            (uint16_t)s16_setPF,
                                            jo_opModFixedPFAbsorbW["en"]["val"].as<uint16_t>()
        };
        StackScop size_t siz_request = sizeof(arr_u16_addresses) / sizeof(arr_u16_addresses[0]);

        cst_LogAddrVal(arr_u16_addresses, arr_u16_values, siz_request);

        StackScop int16_t s16_WriteSt = cst_WriteControl(CST_SET_REG, siz_request, arr_u16_addresses, arr_u16_values);
        if (s16_WriteSt != EXIT_SUCCESS)
        {
            LOG_ERR("CST_SetControl() failed to execute opModFixedPFAbsorbW");
            return -EXIT_FAILURE;
        }
    }

    // // Control Event for opModFixedVar
    // if (ptr_control->_flags & SE_opModFixedVar_exists)
    // {
    //     bol_validControl = true;
        
    //     // Get initial input data required to perfrom control
    //     StackScop ArduinoJson::JsonObject jo_opModFixedVar = jo_controls["opModFixedVar"];
    //     // Retrieve setMaxVar = rtgMaxVar
    //     StackScop uint16_t u16_setMaxVarAddr = jo_opModFixedVar["setMaxVar"]["addr"].as<uint16_t>();
    //     StackScop uint16_t u16_setMaxVar = 0;
    //     if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, &u16_setMaxVarAddr, &u16_setMaxVar, 1) != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_SetControl() failed to read setMaxVar from the inverter");
    //     }
    //     // Determine what the Var value should be set to based on the yreftype
    //     StackScop int32_t s32_setVar = 0;
    //     if(ptr_control->opModFixedVar.refType == REF_PC_SET_MAX_W)
    //     {
    //         s32_setVar = u16_setMaxVar * (ptr_control->opModFixedVar.value * CST_PERCENTAGE_MULTIPLIER);
    //         /** @todo Figure out what to do here - not sure how to convert from %W to Var... */
    //     }
    //     else if (ptr_control->opModFixedVar.refType == REF_PC_SET_MAX_VAR)
    //     {
    //         s32_setVar = u16_setMaxVar * (ptr_control->opModFixedVar.value * CST_PERCENTAGE_MULTIPLIER);
    //     }
    //     else if (ptr_control->opModFixedVar.refType == REF_PC_STAT_VAR_AVAIL)
    //     {
    //         s32_setVar = u16_setMaxVar * (ptr_control->opModFixedVar.value * CST_PERCENTAGE_MULTIPLIER);
    //         /** @todo Figure out how to get statVaravail */
    //     }
    //     else 
    //     {
    //         LOG_ERR("CST_SetControl() received invalid yreftype: %u", ptr_control->opModFixedVar.refType);
    //         return -EINVAL;
    //     }

    //     // Set the target var setpoint and enable the Fixed reactive power control mode
    //     StackScop uint16_t arr_u16_addresses[] =  {   
    //                                         jo_opModFixedVar["setpoint"]["addr"].as<uint16_t>(), 
    //                                         jo_opModFixedVar["en"]["addr"].as<uint16_t>()
    //     };
    //     StackScop uint16_t arr_u16_values[] =     {   
    //                                         (uint16_t)s32_setVar,
    //                                         jo_opModFixedVar["en"]["val"].as<uint16_t>()
    //     };
    //     StackScop size_t siz_request = sizeof(arr_u16_addresses) / sizeof(arr_u16_addresses[0]);

    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_SET_REG, siz_request, arr_u16_addresses, arr_u16_values);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_SetControl() failed to execute opModFixedVar");
    //         return -EXIT_FAILURE;
    //     }
    // }

    // Control Event for opModFixedW
    if (ptr_control->_flags & SE_opModFixedW_exists)
    {
        bol_validControl = true;

        StackScop uint16_t u16_wattSetting = 0;
        StackScop uint16_t u16_wattSettingAddr = 0;

        // Discharge setpoint (grid)
        if(ptr_control->opModFixedW > 0)
        {
            u16_wattSetting = (uint16_t)((float)(ptr_control->opModFixedW)*(jo_controls["opModFixedW"]["discharge"]["sc"].as<float>()));
            u16_wattSettingAddr = (uint16_t)strtol(jo_controls["opModFixedW"]["discharge"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE);
        }
        // Charge setpoint (battery)
        else if (ptr_control->opModFixedW < 0)
        {

        }
        else{}

        if(u16_wattSetting < jo_controls["opModFixedW"]["discharge"]["min"].as<uint16_t>())
        {
            u16_wattSetting = jo_controls["opModFixedW"]["discharge"]["min"].as<uint16_t>();
        }
        if(u16_wattSetting > jo_controls["opModFixedW"]["discharge"]["max"].as<uint16_t>())
        {
            u16_wattSetting = jo_controls["opModFixedW"]["discharge"]["max"].as<uint16_t>();
        }

    //     // Determine what the W value should be set to based on the yreftype
    //     /** @todo Type is not supported by the schema currently */
    //     StackScop int32_t s32_setW = u16_rtgMaxW  * (u16_setMaxPc * CST_PERCENTAGE_MULTIPLIER) 
    //                                     * (ptr_control->opModFixedW * CST_PERCENTAGE_MULTIPLIER * CST_PERCENTAGE_MULTIPLIER);

        // Set the target W setpoint and enable the Fixed active power control mode
        StackScop uint16_t arr_u16_addresses[] =  {   
                                            u16_wattSettingAddr
        };
        StackScop uint16_t arr_u16_values[] =     {   
                                            u16_wattSetting
        };
        StackScop size_t siz_request = sizeof(arr_u16_addresses) / sizeof(arr_u16_addresses[0]);

        cst_LogAddrVal(arr_u16_addresses, arr_u16_values, siz_request);

        StackScop int16_t s16_WriteSt = cst_WriteControl(CST_SET_REG, siz_request, arr_u16_addresses, arr_u16_values);
        if (s16_WriteSt != EXIT_SUCCESS)
        {
            LOG_ERR("CST_SetControl() failed to execute opModFixedW");
            return -EXIT_FAILURE;
        }
    }
    
    // // Control Event for opModFreqDroop
    // if (ptr_control->_flags & SE_opModFreqDroop_exists)
    // {
    //     bol_validControl = true;

    //     // Get initial input data required to perfrom control
    //     StackScop ArduinoJson::JsonObject jo_opModFreqDroop = jo_controls["opModFreqDroop"];
        
    //     // Extract write values and addresses required to perform control 
    //     StackScop uint16_t arr_u16_writeAddr[] =  {
    //                                         jo_opModFreqDroop["dBOF"]["addr"].as<uint16_t>(),
    //                                         jo_opModFreqDroop["dBUF"]["addr"].as<uint16_t>(),
    //                                         jo_opModFreqDroop["kOF"]["addr"].as<uint16_t>(),
    //                                         jo_opModFreqDroop["kUF"]["addr"].as<uint16_t>(),
    //                                         jo_opModFreqDroop["openLoopTms"]["addr"].as<uint16_t>(),
    //                                         jo_opModFreqDroop["en"]["addr"].as<uint16_t>()
    //     };
    //     StackScop uint16_t arr_u16_writeVal[] =   {
    //                                         (uint16_t)ptr_control->opModFreqDroop.dBOF,
    //                                         (uint16_t)ptr_control->opModFreqDroop.dBUF,
    //                                         (uint16_t)ptr_control->opModFreqDroop.kOF,
    //                                         (uint16_t)ptr_control->opModFreqDroop.kUF,
    //                                         (uint16_t)ptr_control->opModFreqDroop.openLoopTms,
    //                                         jo_opModFreqDroop["en"]["val"].as<uint16_t>()
    //     };

    //     // Write the control to the inverter 
    //     StackScop size_t siz_request = sizeof(arr_u16_writeAddr) / sizeof(arr_u16_writeAddr[0]);

    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_SET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_SetControl() failed to execute opModFreqDroop");
    //         return -EXIT_FAILURE;
    //     }
    // }
    
    // // Control Event for opModMaxLimW
    // if (ptr_control->_flags & SE_opModMaxLimW_exists)
    // {
    //     bol_validControl = true;

    //     // Get initial input data required to perfrom control
    //     StackScop ArduinoJson::JsonObject jo_opModMaxLimW = jo_controls["opModMaxLimW"];
    //     StackScop uint16_t u16_opModMaxLimWAddr = jo_opModMaxLimW["addr"].as<uint16_t>();
    //     StackScop uint16_t u16_opModMaxLimW = ptr_control->opModMaxLimW * jo_opModMaxLimW["sc"].as<float>();

    //     // Write the control to the inverter 
    //     StackScop size_t siz_request = 1;

    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_SET_REG, siz_request, &u16_opModMaxLimWAddr, &u16_opModMaxLimW);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_SetControl() failed to execute opModMaxLimW");
    //         return -EXIT_FAILURE;
    //     }
    // }
    
    // // Control Event for opModTargetVar
    // if (ptr_control->_flags & SE_opModTargetVar_exists)
    // {
    //     bol_validControl = true;

    //     // Get initial input data required to perfrom control
    //     StackScop ArduinoJson::JsonObject jo_opModTargetVar = jo_controls["opModTargetVar"];
    //     StackScop float f_multiplier = pow(10, ptr_control->opModTargetVar.multiplier);
    //     f_multiplier *= jo_opModTargetVar["setpoint"]["sc"].as<float>();
    //     StackScop int16_t s16_targetVar = ptr_control->opModTargetVar.value * f_multiplier;
        
    //     // Extract write values and addresses required to perform control 
    //     StackScop uint16_t arr_u16_writeAddr[] =  {
    //                                         jo_opModTargetVar["setpoint"]["addr"].as<uint16_t>(),
    //                                         jo_opModTargetVar["en"]["addr"].as<uint16_t>()
    //     };
    //     StackScop uint16_t arr_u16_writeVal[] =   {
    //                                         (uint16_t)s16_targetVar,
    //                                         jo_opModTargetVar["en"]["val"].as<uint16_t>()
    //     };

    //     // Write the control to the inverter 
    //     StackScop size_t siz_request = sizeof(arr_u16_writeAddr) / sizeof(arr_u16_writeAddr[0]);
        
    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_SET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_SetControl() failed to execute opModTargetVar");
    //         return -EXIT_FAILURE;
    //     }
    // }

    // Control Event for opModTargetW
    if (ptr_control->_flags & SE_opModTargetW_exists)
    {
        bol_validControl = true;

        // Read all of the single registers required
        StackScop uint16_t arr_u16_singleReadAddresses[] = {
            (uint16_t)strtol(jo_controls["opModTargetW"]["rating"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop size_t siz_singleReads = sizeof(arr_u16_singleReadAddresses) / sizeof(arr_u16_singleReadAddresses[0]);
        StackScop uint16_t arr_u16_singleReadValues[siz_singleReads] = {0};
        if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                siz_singleReads) != EXIT_SUCCESS)
        {
            LOG_ERR("cst_GetSingleReadValuesFromObject() Failed to read the single registers required");
            return -EXIT_FAILURE;
        }

        // Get initial input data required to perfrom control
        StackScop ArduinoJson::JsonObject jo_opModTargetW = jo_controls["opModTargetW"];
        StackScop float f_multiplier = pow(10, ptr_control->opModTargetW.multiplier);
        // f_multiplier *= jo_opModTargetW["setpoint"]["sc"].as<float>();
        StackScop int16_t s16_targetW = (ptr_control->opModTargetW.value * f_multiplier);
        StackScop uint16_t u16_wattPercent = (uint16_t)(((float)s16_targetW / (float)arr_u16_singleReadValues[0])*CST_MAX_PERCENT);

        if(u16_wattPercent < jo_controls["opModTargetW"]["setpoint"]["min"].as<uint16_t>())
        {
            u16_wattPercent = jo_controls["opModTargetW"]["setpoint"]["min"].as<uint16_t>();
        }
        if(u16_wattPercent > jo_controls["opModTargetW"]["setpoint"]["max"].as<uint16_t>())
        {
            u16_wattPercent = jo_controls["opModTargetW"]["setpoint"]["max"].as<uint16_t>();
        }

        // Extract write values and addresses required to perform control 
        StackScop uint16_t arr_u16_writeAddr[] =  {
                                            (uint16_t)strtol(jo_opModTargetW["setpoint"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop uint16_t arr_u16_writeVal[] =   {
                                            (uint16_t)u16_wattPercent
        };

        // Write the control to the inverter 
        StackScop size_t siz_request = sizeof(arr_u16_writeAddr) / sizeof(arr_u16_writeAddr[0]);

        cst_LogAddrVal(arr_u16_writeAddr, arr_u16_writeVal, siz_request);

        StackScop int16_t s16_WriteSt = cst_WriteControl(CST_SET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
        if (s16_WriteSt != EXIT_SUCCESS)
        {
            LOG_ERR("CST_SetControl() failed to execute opModTargetW");
            return -EXIT_FAILURE;
        }
    }

    // // Control Event for rampTms
    // if (ptr_control->_flags & SE_rampTms_exists)
    // {
    //     bol_validControl = true;

    //     // Get initial input data required to perfrom control
    //     StackScop ArduinoJson::JsonObject jo_rampTms = jo_controls["rampTms"];
    //     StackScop uint16_t u16_rampTms = ptr_control->rampTms;

    //     // Convert the time into a ramp rate (s -> %/s)
    //     StackScop float f_rampTms_s = ptr_control->rampTms * 0.01;
    //     StackScop float f_setGradW = (CST_MAX_PERCENT * jo_rampTms["setGradW"]["sc"].as<float>()) / f_rampTms_s;
    //     StackScop uint16_t u16_setGradW =  (uint16_t)round(f_setGradW);
        
    //     // Extract write values and addresses required to perform control 
    //     StackScop uint16_t arr_u16_writeAddr[] =  { jo_rampTms["setGradW"]["addr"].as<uint16_t>() };
    //     StackScop uint16_t arr_u16_writeVal[] =   { u16_setGradW };

    //     // Write the control to the inverter 
    //     StackScop size_t siz_request = sizeof(arr_u16_writeAddr) / sizeof(arr_u16_writeAddr[0]);

    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_SET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_SetControl() failed to execute rampTms");
    //         return -EXIT_FAILURE;
    //     }
    // }

    if(!bol_validControl)
    {
        LOG_ERR("CST_SetControl() received invalid control type");
        return -EINVAL;
    }
    else
    {
        return EXIT_SUCCESS;
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop int CST_ResetControl(uint32_t u32_rstCtrlFlags)
{
    // Set the Parameter Type for the function
    StackScop cst_ParameterType_t enm_type = CST_CONTROL;
    
    // // Retrieve the status mapping from the translation map    
    // #TODO - potential for buffer overflow when we don't check if we're exceeding this size (more than one place in this file)

    StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
    cst_GetParameterTypeString(enm_type, sz_type);
    StackScop ArduinoJson::JsonObject jo_controls = jd_translationMap[sz_type];
    if (jo_controls.isNull())
    {
        LOG_ERR("CST_SetControl() Failed to retrieve the controls from the translation map");
        return -EXIT_FAILURE;
    }

    // // Control Reset Event for opModEnergize
    // if (u32_rstCtrlFlags & SE_opModEnergize_exists)
    // {
    //     // Get initial input data required to perform control
    //     StackScop ArduinoJson::JsonObject jo_opModEnergize = jo_controls["opModEnergize"];

    //     // Set the Energize control mode to default value
    //     StackScop uint16_t u16_setEnergize = (uint16_t)jo_opModEnergize["def"].as<uint16_t>();
    //     StackScop uint16_t u16_setEnergizeAddr = jo_opModEnergize["addr"].as<uint16_t>(); 
    //     StackScop size_t siz_request = 1;

    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, &u16_setEnergizeAddr, &u16_setEnergize);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_ResetControl() failed to execute opModEnergize");
    //         return -EXIT_FAILURE;
    //     }
    // }

    // // Control Reset Event for opModConnect
    // if (u32_rstCtrlFlags & SE_opModConnect_exists)
    // {
    //     // Get initial input data required to perform control
    //     StackScop ArduinoJson::JsonObject jo_opModConnect = jo_controls["opModConnect"];

    //     // Set the Connect control mode to default value
    //     StackScop uint16_t u16_setEnergize = (uint16_t)jo_opModConnect["def"].as<uint16_t>();
    //     StackScop uint16_t u16_setEnergizeAddr = jo_opModConnect["addr"].as<uint16_t>(); 
    //     StackScop size_t siz_request = 1;

    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, &u16_setEnergizeAddr, &u16_setEnergize);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_ResetControl() failed to execute opModConnect");
    //         return -EXIT_FAILURE;
    //     }
    // }

    // Control Reset Event for opModFixedPFInjectW
    if (u32_rstCtrlFlags & SE_opModFixedPFInjectW_exists)
    {
        // Get initial input data required to perform control
        StackScop ArduinoJson::JsonObject jo_opModFixedPFInjectW = jo_controls["opModFixedPFInjectW"];

        // Set the target PF setpoint an the Fixed PF control mode to default values
        StackScop uint16_t arr_u16_addresses[] =  {   
                                            (uint16_t)strtol(jo_opModFixedPFInjectW["setpoint"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE),    
        };
        StackScop uint16_t arr_u16_values[] =     {
                                            jo_opModFixedPFInjectW["setpoint"]["def"].as<uint16_t>()
        };

        StackScop size_t siz_request = sizeof(arr_u16_addresses) / sizeof(arr_u16_addresses[0]);

        cst_LogAddrVal(arr_u16_addresses, arr_u16_values, siz_request);
        
        StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_addresses, arr_u16_values);
        if (s16_WriteSt != EXIT_SUCCESS)
        {
            LOG_ERR("CST_ResetControl() failed to execute opModFixedPFInjectW");
            return -EXIT_FAILURE;
        }
    }

    // Control Reset Event for opModFixedPFAbsorbW
    if (u32_rstCtrlFlags & SE_opModFixedPFAbsorbW_exists)
    {
        // Get initial input data required to perform control
        StackScop ArduinoJson::JsonObject jo_opModFixedPFAbsorbW = jo_controls["opModFixedPFAbsorbW"];

        // Set the target PF setpoint an the Fixed PF control mode to default values
        StackScop uint16_t arr_u16_addresses[] =  {   
                                            (uint16_t)strtol(jo_opModFixedPFAbsorbW["setpoint"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE)    
        };
        StackScop uint16_t arr_u16_values[] =     {   
                                            jo_opModFixedPFAbsorbW["setpoint"]["def"].as<uint16_t>()
        };

        StackScop size_t siz_request = sizeof(arr_u16_addresses) / sizeof(arr_u16_addresses[0]);

        cst_LogAddrVal(arr_u16_addresses, arr_u16_values, siz_request);
        
        StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_addresses, arr_u16_values);
        if (s16_WriteSt != EXIT_SUCCESS)
        {
            LOG_ERR("CST_ResetControl() failed to execute opModFixedPFInjectW");
            return -EXIT_FAILURE;
        }
    }

    // // Control Reset Event for opModFixedVar
    // if (u32_rstCtrlFlags & SE_opModFixedVar_exists)
    // {
    //     // Get initial input data required to perform control
    //     StackScop ArduinoJson::JsonObject jo_opModFixedVar = jo_controls["opModFixedVar"];

    //     // Set the target var setpoint and the Fixed reactive power control mode to default values
    //     StackScop uint16_t arr_u16_addresses[] =  {   
    //                                         jo_opModFixedVar["setpoint"]["addr"].as<uint16_t>(), 
    //                                         jo_opModFixedVar["en"]["addr"].as<uint16_t>(),
    //                                         jo_opModFixedVar["setMaxVar"]["addr"].as<uint16_t>()
    //     };
    //     StackScop uint16_t arr_u16_values[] =     {   
    //                                         jo_opModFixedVar["setpoint"]["def"].as<uint16_t>(),
    //                                         jo_opModFixedVar["en"]["def"].as<uint16_t>(),
    //                                         jo_opModFixedVar["setMaxVar"]["def"].as<uint16_t>()                           
    //     };
    //     StackScop size_t siz_request = sizeof(arr_u16_addresses) / sizeof(arr_u16_addresses[0]);

    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_addresses, arr_u16_values);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_ResetControl() failed to execute opModFixedVar");
    //         return -EXIT_FAILURE;
    //     }
    // }

    // Control Reset Event for opModFixedW
    if (u32_rstCtrlFlags & SE_opModFixedW_exists)
    {
        // Get initial input data required to perform control
        StackScop ArduinoJson::JsonObject jo_opModFixedW = jo_controls["opModFixedW"];

        // Set the target W setpoint and the Fixed active power control mode to default values
        StackScop uint16_t arr_u16_addresses[] =  {   
                                            (uint16_t)strtol(jo_opModFixedW["discharge"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop uint16_t arr_u16_values[] =     {    
                                            jo_opModFixedW["discharge"]["default"].as<uint16_t>()                             
        };
        StackScop size_t siz_request = sizeof(arr_u16_addresses) / sizeof(arr_u16_addresses[0]);

        cst_LogAddrVal(arr_u16_addresses, arr_u16_values, siz_request);

        StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_addresses, arr_u16_values);
        if (s16_WriteSt != EXIT_SUCCESS)
        {
            LOG_ERR("CST_ResetControl() failed to execute opModFixedW");
            return -EXIT_FAILURE;
        }
    }
    
    // // Control Reset Event for opModFreqDroop
    // if (u32_rstCtrlFlags & SE_opModFreqDroop_exists)
    // {
    //     // Get initial input data required to perform control
    //     StackScop ArduinoJson::JsonObject jo_opModFreqDroop = jo_controls["opModFreqDroop"];

    //     // Extract write values and addresses required to perform control 
    //     StackScop uint16_t arr_u16_writeAddr[] =  {
    //                                         jo_opModFreqDroop["dBOF"]["addr"].as<uint16_t>(),
    //                                         jo_opModFreqDroop["dBUF"]["addr"].as<uint16_t>(),
    //                                         jo_opModFreqDroop["kOF"]["addr"].as<uint16_t>(),
    //                                         jo_opModFreqDroop["kUF"]["addr"].as<uint16_t>(),
    //                                         jo_opModFreqDroop["openLoopTms"]["addr"].as<uint16_t>(),
    //                                         jo_opModFreqDroop["en"]["addr"].as<uint16_t>()
    //     };
    //     StackScop uint16_t arr_u16_writeVal[] =   {
    //                                         jo_opModFreqDroop["dBOF"]["def"].as<uint16_t>(),
    //                                         jo_opModFreqDroop["dBUF"]["def"].as<uint16_t>(),
    //                                         jo_opModFreqDroop["kOF"]["def"].as<uint16_t>(),
    //                                         jo_opModFreqDroop["kUF"]["def"].as<uint16_t>(),
    //                                         jo_opModFreqDroop["openLoopTms"]["def"].as<uint16_t>(),
    //                                         jo_opModFreqDroop["en"]["def"].as<uint16_t>()
    //     };

    //     // Write the control to the inverter 
    //     StackScop size_t siz_request = sizeof(arr_u16_writeAddr) / sizeof(arr_u16_writeAddr[0]);

    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_ResetControl() failed to execute opModFreqDroop");
    //         return -EXIT_FAILURE;
    //     }
    // }
    
    // // Control Reset Event for opModMaxLimW
    // if (u32_rstCtrlFlags & SE_opModMaxLimW_exists)
    // {
    //     // Get initial input data required to perform control
    //     StackScop ArduinoJson::JsonObject jo_opModMaxLimW = jo_controls["opModMaxLimW"];
    //     StackScop uint16_t u16_opModMaxLimWAddr = jo_opModMaxLimW["addr"].as<uint16_t>();
    //     StackScop uint16_t u16_opModMaxLimW = jo_opModMaxLimW["def"].as<uint16_t>();

    //     // Write the default control to the inverter 
    //     StackScop size_t siz_request = 1;

    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, &u16_opModMaxLimWAddr, &u16_opModMaxLimW);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_ResetControl() failed to execute opModMaxLimW");
    //         return -EXIT_FAILURE;
    //     }
    // }
    
    // // Control Reset Event for opModTargetVar
    // if (u32_rstCtrlFlags & SE_opModTargetVar_exists)
    // {
    //     // Get initial input data required to perform control
    //     StackScop ArduinoJson::JsonObject jo_opModTargetVar = jo_controls["opModTargetVar"];

    // // Extract write values and addresses required to perform control 
    //     StackScop uint16_t arr_u16_writeAddr[] =  {
    //                                         jo_opModTargetVar["setpoint"]["addr"].as<uint16_t>(),
    //                                         jo_opModTargetVar["en"]["addr"].as<uint16_t>()
    //     };
    //     StackScop uint16_t arr_u16_writeVal[] =   {
    //                                         jo_opModTargetVar["setpoint"]["def"].as<uint16_t>(),
    //                                         jo_opModTargetVar["en"]["def"].as<uint16_t>()
    //     };

    //     // Write the control to the inverter 
    //     StackScop size_t siz_request = sizeof(arr_u16_writeAddr) / sizeof(arr_u16_writeAddr[0]);

    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_ResetControl() failed to execute opModTargetVar");
    //         return -EXIT_FAILURE;
    //     }
    // }

    // Control Reset Event for opModTargetW
    if (u32_rstCtrlFlags & SE_opModTargetW_exists)
    {
        // Get initial input data required to perform control
        StackScop ArduinoJson::JsonObject jo_opModTargetW = jo_controls["opModTargetW"];

        // Extract write values and addresses required to perform control 
        StackScop uint16_t arr_u16_writeAddr[] =  {
                                            (uint16_t)strtol(jo_opModTargetW["setpoint"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE)
                                           
        };
        StackScop uint16_t arr_u16_writeVal[] =   {
                                            jo_opModTargetW["setpoint"]["def"].as<uint16_t>()
                                          
        };

        // Write the control to the inverter 
        StackScop size_t siz_request = sizeof(arr_u16_writeAddr) / sizeof(arr_u16_writeAddr[0]);

        cst_LogAddrVal(arr_u16_writeAddr, arr_u16_writeVal, siz_request);

        StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
        if (s16_WriteSt != EXIT_SUCCESS)
        {
            LOG_ERR("CST_ResetControl() failed to execute opModTargetW");
            return -EXIT_FAILURE;
        }
    }

    // // Control Reset Event for rampTms
    // if (u32_rstCtrlFlags & SE_rampTms_exists)
    // {
    //     // Get initial input data required to perform control
    //     StackScop ArduinoJson::JsonObject jo_rampTms = jo_controls["rampTms"];

    //     // Extract write values and addresses required to perform control 
    //     StackScop uint16_t arr_u16_writeAddr[] =  { jo_rampTms["setGradW"]["addr"].as<uint16_t>() };
    //     StackScop uint16_t arr_u16_writeVal[] =   { jo_rampTms["setGradW"]["def"].as<uint16_t>() };

    //     // Write the control to the inverter 
    //     StackScop size_t siz_request = sizeof(arr_u16_writeAddr) / sizeof(arr_u16_writeAddr[0]);

    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_ResetControl() failed to execute rampTms");
    //         return -EXIT_FAILURE;
    //     }
    // }
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop int CST_SetDefaultControl(SE_DefaultDERControl_t *ptr_defaultControl)
{
    // // Set the Parameter Type for the function
    // StackScop cst_ParameterType_t enm_type = CST_DEFAULT_CONTROL;
    
    // // Retrieve the status mapping from the translation map
    // StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
    // cst_GetParameterTypeString(enm_type, sz_type);
    // StackScop ArduinoJson::JsonObject jo_defaultControls = jd_translationMap[sz_type];
    // if (jo_defaultControls.isNull())
    // {
    //     LOG_ERR("CST_SetDefaultControl() Failed to retrieve the settings from the translation map");
    //     return -EXIT_FAILURE;
    // }
    // // Prepare arrays of addresses and values to write to the inverter
    // StackScop uint16_t arr_u16_writeAddr[CST_MAX_DEFAULT_CONTROL_SIZE] = {0};
    // StackScop uint16_t arr_u16_writeVal[CST_MAX_DEFAULT_CONTROL_SIZE] = {0};
    // StackScop size_t siz_request = 0;

    // ptr_defaultControl->setESRampTms = 1000;
    // ptr_defaultControl->_flags |= SE_setESRampTms_exists;

    // // Default Control for setESDelay
    // if (ptr_defaultControl->_flags & SE_setESDelay_exists) 
    // {
    //     if (cst_FormatWriteValues(enm_type, "setESDelay", (int32_t)ptr_defaultControl->setESDelay, 
    //                             arr_u16_writeAddr, arr_u16_writeVal, (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_SetDefaultControl() setESDelay is not supported by the inverter");
    //     }
    // }
    
    // // Default Control for setESHighFreq
    // if (ptr_defaultControl->_flags & SE_setESHighFreq_exists) 
    // {
    //     if (cst_FormatWriteValues(enm_type, "setESHighFreq", (int32_t)ptr_defaultControl->setESHighFreq, 
    //                             arr_u16_writeAddr, arr_u16_writeVal, (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_SetDefaultControl() setESHighFreq is not supported by the inverter");
    //     }
    // }
    
    // // Default Control for setESHighVolt
    // if (ptr_defaultControl->_flags & SE_setESHighVolt_exists) 
    // {
    //     if (cst_FormatWriteValues(enm_type, "setESHighVolt", (int32_t)ptr_defaultControl->setESHighVolt, 
    //                             arr_u16_writeAddr, arr_u16_writeVal, (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_SetDefaultControl() setESHighVolt is not supported by the inverter");
    //     }
    // }
    
    // // Default Control for setESLowFreq
    // if (ptr_defaultControl->_flags & SE_setESLowFreq_exists) 
    // {
    //     if (cst_FormatWriteValues(enm_type, "setESLowFreq", (int32_t)ptr_defaultControl->setESLowFreq, 
    //                             arr_u16_writeAddr, arr_u16_writeVal, (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_SetDefaultControl() setESLowFreq is not supported by the inverter");
    //     }
    // }
    
    // // Default Control for setESLowVolt
    // if (ptr_defaultControl->_flags & SE_setESLowVolt_exists) 
    // {
    //     if (cst_FormatWriteValues(enm_type, "setESLowVolt", (int32_t)ptr_defaultControl->setESLowVolt, 
    //                             arr_u16_writeAddr, arr_u16_writeVal, (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_SetDefaultControl() setESLowVolt is not supported by the inverter");
    //     }
    // }

    // // Default Control for setESRampTms
    // if (ptr_defaultControl->_flags & SE_setESRampTms_exists) 
    // {
    //     if (cst_FormatWriteValues(enm_type, "setESRampTms", (int32_t)ptr_defaultControl->setESRampTms, 
    //                             arr_u16_writeAddr, arr_u16_writeVal, (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_SetDefaultControl() setESRampTms is not supported by the inverter");
    //     }

    //     CST_ResetDefaultControl(SE_setESRampTms_exists);

    // }
    
    // // Default Control for setESRandomDelay
    // if (ptr_defaultControl->_flags & SE_setESRandomDelay_exists) 
    // {
    //     if (cst_FormatWriteValues(enm_type, "setESRandomDelay", (int32_t)ptr_defaultControl->setESRandomDelay, 
    //                             arr_u16_writeAddr, arr_u16_writeVal, (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_SetDefaultControl() setESRandomDelay is not supported by the inverter");
    //     }
    // }
    
    // // Default Control for setGradW
    // if (ptr_defaultControl->_flags & SE_setGradW_exists) 
    // {
    //     if (cst_FormatWriteValues(enm_type, "setGradW", (int32_t)ptr_defaultControl->setGradW, 
    //                             arr_u16_writeAddr, arr_u16_writeVal, (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_SetDefaultControl() setGradW is not supported by the inverter");
    //     }
    // }
    
    // // Default Control for setSoftGradW
    // if (ptr_defaultControl->_flags & SE_setSoftGradW_exists) 
    // {
    //     if (cst_FormatWriteValues(enm_type, "setSoftGradW", (int32_t)ptr_defaultControl->setSoftGradW, 
    //                             arr_u16_writeAddr, arr_u16_writeVal, (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_SetDefaultControl() setSoftGradW is not supported by the inverter");
    //     }
    // }

    // // Write the defaults controls to the inverter

    // StackScop int16_t s16_WriteSt = cst_WriteControl(CST_SET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    // if (s16_WriteSt != EXIT_SUCCESS)
    // {
    //     LOG_ERR("CST_SetDefaultControl() failed to execute");
    //     return -EXIT_FAILURE;
    // }

    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop int CST_ResetDefaultControl(uint32_t u32_rstCtrlFlags)
{
    // // Set the Parameter Type for the function
    // StackScop cst_ParameterType_t enm_type = CST_DEFAULT_CONTROL;
    
    // // Retrieve the status mapping from the translation map
    // StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
    // cst_GetParameterTypeString(enm_type, sz_type);
    // StackScop ArduinoJson::JsonObject jo_defaultControls = jd_translationMap[sz_type];
    // if (jo_defaultControls.isNull())
    // {
    //     LOG_ERR("CST_ResetDefaultControl() Failed to retrieve the settings from the translation map");
    //     return -EXIT_FAILURE;
    // }
    // // Prepare arrays of addresses and values to write to the inverter
    // StackScop uint16_t arr_u16_writeAddr[CST_MAX_DEFAULT_CONTROL_SIZE] = {0};
    // StackScop uint16_t arr_u16_writeVal[CST_MAX_DEFAULT_CONTROL_SIZE] = {0};
    // StackScop size_t siz_request = 0;

    // u32_rstCtrlFlags |= SE_setESRampTms_exists;

    // // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // // Default Control for setESDelay
    // if (u32_rstCtrlFlags & SE_setESDelay_exists) 
    // {
    //     if (cst_FormatResetWriteValues(enm_type, "setESDelay", arr_u16_writeAddr, arr_u16_writeVal, 
    //                             (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_ResetDefaultControl() setESDelay is not supported by the inverter");
    //     }
    // }
    
    // // Default Control for setESHighFreq
    // if (u32_rstCtrlFlags & SE_setESHighFreq_exists) 
    // {
    //     if (cst_FormatResetWriteValues(enm_type, "setESHighFreq", arr_u16_writeAddr, arr_u16_writeVal, 
    //                             (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_ResetDefaultControl() setESHighFreq is not supported by the inverter");
    //     }
    // }
    
    // // Default Control for setESHighVolt
    // if (u32_rstCtrlFlags & SE_setESHighVolt_exists) 
    // {
    //     if (cst_FormatResetWriteValues(enm_type, "setESHighVolt", arr_u16_writeAddr, arr_u16_writeVal, 
    //                             (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_ResetDefaultControl() setESHighVolt is not supported by the inverter");
    //     }
    // }
    
    // // Default Control for setESLowFreq
    // if (u32_rstCtrlFlags & SE_setESLowFreq_exists) 
    // {
    //     if (cst_FormatResetWriteValues(enm_type, "setESLowFreq", arr_u16_writeAddr, arr_u16_writeVal, 
    //                             (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_ResetDefaultControl() setESLowFreq is not supported by the inverter");
    //     }
    // }
    
    // // Default Control for setESLowVolt
    // if (u32_rstCtrlFlags & SE_setESLowVolt_exists) 
    // {
    //     if (cst_FormatResetWriteValues(enm_type, "setESLowVolt", arr_u16_writeAddr, arr_u16_writeVal, 
    //                             (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_ResetDefaultControl() setESLowVolt is not supported by the inverter");
    //     }
    // }

    // // Default Control for setESRampTms
    // if (u32_rstCtrlFlags & SE_setESRampTms_exists) 
    // {
    //     if (cst_FormatResetWriteValues(enm_type, "setESRampTms",  arr_u16_writeAddr, arr_u16_writeVal, 
    //                             (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_ResetDefaultControl() setESRampTms is not supported by the inverter");
    //     }
    // }
    
    // // Default Control for setESRandomDelay
    // if (u32_rstCtrlFlags & SE_setESRandomDelay_exists) 
    // {
    //     if (cst_FormatResetWriteValues(enm_type, "setESRandomDelay", arr_u16_writeAddr, arr_u16_writeVal, 
    //                             (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_ResetDefaultControl() setESRandomDelay is not supported by the inverter");
    //     }
    // }
    
    // // Default Control for setGradW
    // if (u32_rstCtrlFlags & SE_setGradW_exists) 
    // {
    //     if (cst_FormatResetWriteValues(enm_type, "setGradW", arr_u16_writeAddr, arr_u16_writeVal, 
    //                             (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_ResetDefaultControl() setGradW is not supported by the inverter");
    //     }
    // }
    
    // // Default Control for setSoftGradW
    // if (u32_rstCtrlFlags & SE_setSoftGradW_exists) 
    // {
    //     if (cst_FormatResetWriteValues(enm_type, "setSoftGradW", arr_u16_writeAddr, arr_u16_writeVal, 
    //                             (uint8_t)siz_request) == EXIT_SUCCESS)
    //     {
    //         siz_request++;
    //     }
    //     else
    //     {
    //         LOG_ERR("CST_ResetDefaultControl() setSoftGradW is not supported by the inverter");
    //     }
    // }

    // // Write the defaults controls to the inverter 
    // StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    // if (s16_WriteSt != EXIT_SUCCESS)
    // {
    //     LOG_ERR("CST_ResetDefaultControl() failed to execute");
    //     return -EXIT_FAILURE;
    // }

    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// #TODO - This function should be modified to call helper functions to set the curve data for each curve type
// #TODO - which will allow the function to be more modular and easier to maintain (min/max ranges per register, custom
// #TODO - calculations, etc.)
GlobScop int CST_SetCurve(SE_DERCurve_t *ptr_curve)
{
    // Set the Parameter Type for the function
    StackScop cst_ParameterType_t enm_type = CST_CURVE;
    
    // Retrieve the status mapping from the translation map
    StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
    cst_GetParameterTypeString(enm_type, sz_type);
    StackScop ArduinoJson::JsonObject jo_curves = jd_translationMap[sz_type];
    if (jo_curves.isNull())
    {
        LOG_ERR("CST_SetCurve() Failed to retrieve the curves from the translation map");
        return -EXIT_FAILURE;
    }

    // Prepare arrays of addresses and values to write to the inverter
    StackScop uint16_t arr_u16_writeAddr[CST_MAX_CURVE_SIZE] = {0};
    StackScop uint16_t arr_u16_writeVal[CST_MAX_CURVE_SIZE] = {0};
    StackScop size_t siz_request = 0;

    // Check the curve type to determine if the curve is supported
    StackScop char sz_curveType[CST_PARAMETER_TYPE_SIZE] = {""};
    sprintf(sz_curveType, "%u", ptr_curve->curveType);
    StackScop ArduinoJson::JsonObject jo_curve = jo_curves[sz_curveType];
    if (jo_curve.isNull() || jo_curve["x_addr"].isNull() || jo_curve["y_addr"].isNull())
    {
        LOG_ERR("CST_SetCurve() curveType %u (%s) is not supported by the inverter", ptr_curve->curveType, ptr_curve->description);
        return -EXIT_FAILURE;
    }

    // Extract the curve data from the curve object
    StackScop size_t siz_curveData = jo_curve["x_addr"].size();
    StackScop ArduinoJson::JsonArray ja_xAddr = jo_curve["x_addr"];
    StackScop ArduinoJson::JsonArray ja_yAddr = jo_curve["y_addr"];

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // #TODO - Temporary exit
    if(    ptr_curve->curveType == 0 || ptr_curve->curveType == 1 || ptr_curve->curveType == 2 || ptr_curve->curveType == 3 
        || ptr_curve->curveType == 4 || ptr_curve->curveType == 5 || ptr_curve->curveType == 6 || ptr_curve->curveType == 7
        || ptr_curve->curveType == 8 || ptr_curve->curveType == 9 || ptr_curve->curveType == 10 
        || ptr_curve->curveType == 13 || ptr_curve->curveType == 14 || ptr_curve->curveType == 15)
    {
        // LOG_ERR("exit");
        return EXIT_SUCCESS;
    }
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Curve interpolation if the curve data does not match the number of available registers
    if (siz_curveData != SE_CurveData_count(&(ptr_curve->_flags)))
    {
        LOG_INF("CST_SetCurve() curveType %u (%s) does not match the number of points: %d", ptr_curve->curveType, 
                                                                            ptr_curve->description, siz_curveData);
        LOG_INF("Performing curve interpolation");

        uint8_t u8_curvePointCount = SE_CurveData_count(&(ptr_curve->_flags));
        std::vector <float> vec_f_xRaw;
        std::vector <float> vec_f_yRaw;   

        // Extract the raw curve data and convert to float for higher granularity
        for (uint8_t i = 0; i < u8_curvePointCount; i++)
        {
            vec_f_xRaw.push_back(ptr_curve->CurveData[i].xvalue);
            vec_f_yRaw.push_back(ptr_curve->CurveData[i].yvalue);
        }

        // Interpolate the curve data   
        std::vector<std::pair<float, float>> vec_interpolatedData = CST_piecewiseLinearInterpolation(vec_f_xRaw, vec_f_yRaw, siz_curveData);

        // Clear the curve data and update the curve data count
        memset(ptr_curve->CurveData, 0, sizeof(ptr_curve->CurveData));
        ptr_curve->_flags &= CST_CURVEPOINT_MASK;
        SE_set_CurveData_count(&(ptr_curve->_flags), siz_curveData);

        // Update the curve data with the interpolated values
        for (uint8_t i = 0; i < siz_curveData; i++)
        {
            ptr_curve->CurveData[i].xvalue = static_cast<SE_Int32_t>(vec_interpolatedData[i].first);
            ptr_curve->CurveData[i].yvalue = static_cast<SE_Int32_t>(vec_interpolatedData[i].second);
        }
    }

    // //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // Read any registers required to perform data manipulation
    StackScop float f_varRatio = 0;
    // StackScop float f_wRatio = 0;
    StackScop char sz_readType[CST_PARAMETER_TYPE_SIZE] = {""};
    cst_GetParameterTypeString(CST_SETTING, sz_readType);
    StackScop ArduinoJson::JsonObject jo_settings = jd_translationMap[sz_readType];
    if (ptr_curve->yRefType == REF_PC_SET_MAX_VAR)
    {
        // Retrieve setMaxVar = rtgMaxVar, therefore the ratio is 1
        f_varRatio = CST_MAX_RATIO;
    }
    else if (ptr_curve->yRefType == REF_PC_STAT_VAR_AVAIL)
    {
        // No reading required
    }
    else if (ptr_curve->yRefType == REF_PC_SET_MAX_W)
    {
        // Retrieve setMaxW (acitve power ratio)
        // StackScop uint16_t u16_setMaxWAddr = jo_settings["setMaxW"]["maxLimitRatioW"].as<uint16_t>();
        // StackScop uint16_t u16_setMaxWAddr[3] = {0x3420, 0x3423, 0x3422};
        // StackScop uint16_t u16_readVals[3] = {0, 0, 0};
        
        // StackScop uint16_t u16_setMaxW = 0;

        // if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, u16_setMaxWAddr, u16_readVals, 1) 
        //     != EXIT_SUCCESS)
        // {
        //     LOG_ERR("CST_SetCurve() failed to read setMaxW from the inverter");
        //     return -EXIT_FAILURE;
        // }

        // if(u16_readVals[0] * u16_readVals[1] > u16_readVals[0] * u16_readVals[2])
        // {
        //     u16_setMaxW = u16_readVals[0] * u16_readVals[1];
        // }
        // else
        // {
        //     u16_setMaxW = u16_readVals[0] * u16_readVals[2];
        // }

        // u16_setMaxW = 700;
        // // If the active power ratio is zero, there is no limit, so set it 100%
        // if (u16_setMaxW == 0)
        // {
        //     f_wRatio = CST_MAX_RATIO;
        // }
        // else 
        // {
        //     f_wRatio = u16_setMaxW;
        //     f_wRatio = f_wRatio * pow(10, jo_settings["setMaxW"]["mult"].as<int8_t>());
        // }
    }
    else if (ptr_curve->yRefType == REF_PC_STAT_W_AVAIL)
    {
        // No reading required
    }
    else if (ptr_curve->yRefType == REF_NONE)
    {
        // No reading required
    }
    else if (ptr_curve->yRefType == REF_PC_SET_EFFECTIVE_V)
    {
        // No reading required
    }
    else 
    {
        LOG_ERR("CST_SetCurve() received invalid yreftype: %u", ptr_curve->yRefType);
        return -EINVAL;
    }

    // Determine true x and y multipliers required for writing the curve data
    StackScop float f_xMultiplier = pow(10, (ptr_curve->xMultiplier - jo_curve["x_mult"].as<int8_t>()));
    StackScop float f_yMultiplier = pow(10, (ptr_curve->yMultiplier - jo_curve["y_mult"].as<int8_t>()));

    // Format the curve data values as needed
    for (uint8_t i = 0; i < siz_curveData; i++)
    {
        // Perform manipulations on x data value
        StackScop int32_t s32_xWriteVal = ptr_curve->CurveData[i].xvalue * f_xMultiplier;

        // Perform manipulations on y data value
        StackScop int32_t s32_yWriteVal = 0;
        if (ptr_curve->yRefType == REF_PC_SET_MAX_VAR)
        {
            s32_yWriteVal = ptr_curve->CurveData[i].yvalue * (f_yMultiplier * f_varRatio);
        }
        else if (ptr_curve->yRefType == REF_PC_STAT_VAR_AVAIL)
        {
            s32_yWriteVal = ptr_curve->CurveData[i].yvalue * f_yMultiplier;
        }
        else if (ptr_curve->yRefType == REF_PC_SET_MAX_W)
        {
            // s32_yWriteVal = ptr_curve->CurveData[i].yvalue * (f_yMultiplier * f_wRatio);
            s32_yWriteVal = ptr_curve->CurveData[i].yvalue * (f_yMultiplier);
        }
        else if (ptr_curve->yRefType == REF_PC_STAT_W_AVAIL)
        {
            s32_yWriteVal = ptr_curve->CurveData[i].yvalue * f_yMultiplier;
        }
        else if (ptr_curve->yRefType == REF_NONE)
        {
            s32_yWriteVal = ptr_curve->CurveData[i].yvalue * f_yMultiplier;
        }
        else if (ptr_curve->yRefType == REF_PC_SET_EFFECTIVE_V)
        {
            s32_yWriteVal = ptr_curve->CurveData[i].yvalue * f_yMultiplier;
        }
        else
        {
            LOG_ERR("CST_SetCurve() received invalid yreftype: %u", ptr_curve->yRefType);
            return -EINVAL;
        }

        // Store the values in the write arrays
        arr_u16_writeAddr[siz_request] = (uint16_t)strtol(ja_xAddr[i].as<const char*>(), nullptr, CST_HEX_BASE);
        arr_u16_writeVal[siz_request] = (uint16_t)s32_xWriteVal;
        siz_request++;

        arr_u16_writeAddr[siz_request] = (uint16_t)strtol(ja_yAddr[i].as<const char*>(), nullptr, CST_HEX_BASE);
        arr_u16_writeVal[siz_request] = (uint16_t)s32_yWriteVal;
        siz_request++;
    }

    cst_CurveBounds(jo_curve, arr_u16_writeVal, siz_request);

    // Determine if other curve parameters need to be set
    if (jo_curve.containsKey("openLoopTms") && (ptr_curve->_flags & SE_openLoopTms_exists))
    {
        arr_u16_writeAddr[siz_request] = (uint16_t)strtol(jo_curve["openLoopTms"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE);
        if(ptr_curve->openLoopTms < jo_curve["openLoopTms"]["min"].as<int16_t>())
        {
            arr_u16_writeVal[siz_request] = (uint16_t)jo_curve["openLoopTms"]["min"].as<int16_t>();
        }
        else if(ptr_curve->openLoopTms > jo_curve["openLoopTms"]["max"].as<int16_t>())
        {
            arr_u16_writeVal[siz_request] = (uint16_t)jo_curve["openLoopTms"]["max"].as<int16_t>();
        }
        else
        {
            arr_u16_writeVal[siz_request] = ptr_curve->openLoopTms;
        }
        siz_request++;
    }

    // if (jo_curve.containsKey("vRef") && (ptr_curve->_flags & SE_vRef_exists))
    // {
    //     arr_u16_writeAddr[siz_request] = (uint16_t)strtol(jo_curve["vRef"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE);
    //     arr_u16_writeVal[siz_request] = ptr_curve->vRef;
    //     siz_request++;
    // }

    if (jo_curve.containsKey("autonomousVRefEnable") && (ptr_curve->_flags & SE_autonomousVRefEnable_exists))
    {
        arr_u16_writeAddr[siz_request] = (uint16_t)strtol(jo_curve["autonomousVRefEnable"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE);

        // Read all of the single registers required
        StackScop uint16_t arr_u16_singleReadAddresses[] = {
            (uint16_t)strtol(jo_curve["autonomousVRefEnable"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop size_t siz_singleReads = sizeof(arr_u16_singleReadAddresses) / sizeof(arr_u16_singleReadAddresses[0]);
        StackScop uint16_t arr_u16_singleReadValues[siz_singleReads] = {0};
        if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                siz_singleReads) != EXIT_SUCCESS)
        {
            LOG_ERR("cst_GetSingleReadValuesFromObject() Failed to read the single registers required");
            return -EXIT_FAILURE;
        }

        // Create the mask based on b_st and b_len
        StackScop uint16_t u16_mask = ((1 << jo_curve["autonomousVRefEnable"]["b_len"].as<uint8_t>()) - 1) 
                        << jo_curve["autonomousVRefEnable"]["b_st"].as<uint8_t>();

        if(ptr_curve->autonomousVRefEnable == true)
        {
            arr_u16_singleReadValues[0] = arr_u16_singleReadValues[0] | u16_mask;
        }
        else
        {
            arr_u16_singleReadValues[0] = arr_u16_singleReadValues[0] & ~u16_mask;
        }
        arr_u16_writeVal[siz_request] = arr_u16_singleReadValues[0];
        siz_request++;
    }

    if (jo_curve.containsKey("autonomousVRefTimeConstant") && (ptr_curve->_flags & SE_autonomousVRefTimeConstant_exists))
    {
        arr_u16_writeAddr[siz_request] = (uint16_t)strtol(jo_curve["autonomousVRefTimeConstant"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE);
        arr_u16_writeVal[siz_request] = ptr_curve->autonomousVRefTimeConstant * jo_curve["autonomousVRefTimeConstant"]["sc"].as<uint16_t>();

        // Min/max checking
        if(arr_u16_writeVal[siz_request] < jo_curve["autonomousVRefTimeConstant"]["min"].as<uint16_t>())
        {
            arr_u16_writeVal[siz_request] = jo_curve["autonomousVRefTimeConstant"]["min"].as<uint16_t>();
        }
        else if(arr_u16_writeVal[siz_request] > jo_curve["autonomousVRefTimeConstant"]["max"].as<uint16_t>())
        {
            arr_u16_writeVal[siz_request] = jo_curve["autonomousVRefTimeConstant"]["max"].as<uint16_t>();
        }   
        siz_request++;
    }

    if (jo_curve.containsKey("en"))
    {

        arr_u16_writeAddr[siz_request] = (uint16_t)strtol(jo_curve["en"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE);
        arr_u16_writeVal[siz_request] = jo_curve["en"]["val"].as<uint16_t>();
        siz_request++;
    }

    // Validate the size of the request
    if (siz_request > CST_MAX_CURVE_SIZE)
    {
        LOG_ERR("CST_SetCurve() Request size exceeds the maximum curve size");
        return -EINVAL;
    }

    cst_LogAddrVal(arr_u16_writeAddr, arr_u16_writeVal, siz_request);

    // Write the curve data to the inverter
    StackScop int16_t s16_WriteSt = cst_WriteControl(CST_SET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    // s16_WriteSt = cst_WriteControl(CST_SET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    if (s16_WriteSt != EXIT_SUCCESS)
    {
        LOG_ERR("CST_SetCurve() failed to execute");
        return -EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop int CST_ResetCurve(uint32_t u32_rstCtrlFlags) 
{
    // Set the Parameter Type for the function
    StackScop cst_ParameterType_t enm_type = CST_CURVE;
    
    // Retrieve the status mapping from the translation map
    StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
    cst_GetParameterTypeString(enm_type, sz_type);
    StackScop ArduinoJson::JsonObject jo_curves = jd_translationMap[sz_type];
    if (jo_curves.isNull())
    {
        LOG_ERR("CST_SetCurve() Failed to retrieve the curves from the translation map");
        return -EXIT_FAILURE;
    }

    // Prepare arrays of addresses and values to write to the inverter
    StackScop uint16_t arr_u16_writeAddr[CST_MAX_CURVE_SIZE] = {0};
    StackScop uint16_t arr_u16_writeVal[CST_MAX_CURVE_SIZE] = {0};
    StackScop size_t siz_request = 0;

    // //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    // if(u32_rstCtrlFlags & SE_opModHFRTMustTrip_exists) 
    // {
    //     StackScop ArduinoJson::JsonObject jo_curve = jo_curves["2"];
    //     StackScop ArduinoJson::JsonArray ja_xAddr = jo_curve["x_addr"];
    //     StackScop ArduinoJson::JsonArray ja_yAddr = jo_curve["y_addr"];
    //     StackScop ArduinoJson::JsonArray ja_xDef = jo_curve["x_def"];
    //     StackScop ArduinoJson::JsonArray ja_yDef = jo_curve["y_def"];

    //     for(uint8_t i = 0; i < ja_xAddr.size(); i++)
    //     {
    //         // Store the values in the write arrays
    //         arr_u16_writeAddr[siz_request] = ja_xAddr[i].as<uint16_t>();
    //         arr_u16_writeVal[siz_request] = ja_xDef[i].as<uint16_t>();
    //         siz_request++;
    //         arr_u16_writeAddr[siz_request] = ja_yAddr[i].as<uint16_t>();
    //         arr_u16_writeVal[siz_request] = ja_yDef[i].as<uint16_t>();
    //         siz_request++;
    //     }

    //     // Write the curve data to the inverter
    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_ResetCurve() failed to execute");
    //         return -EXIT_FAILURE;
    //     }

    //     memset(arr_u16_writeAddr, 0, sizeof(arr_u16_writeAddr));
    //     memset(arr_u16_writeVal, 0, sizeof(arr_u16_writeVal));
    //     siz_request = 0;
    // }

    // //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // if(u32_rstCtrlFlags & SE_opModHVRTMomentaryCessation_exists) 
    // {
    //     StackScop ArduinoJson::JsonObject jo_curve = jo_curves["4"];
    //     StackScop ArduinoJson::JsonArray ja_xAddr = jo_curve["x_addr"];
    //     StackScop ArduinoJson::JsonArray ja_yAddr = jo_curve["y_addr"];
    //     StackScop ArduinoJson::JsonArray ja_xDef = jo_curve["x_def"];
    //     StackScop ArduinoJson::JsonArray ja_yDef = jo_curve["y_def"];

    //     for(uint8_t i = 0; i < ja_xAddr.size(); i++)
    //     {
    //         // Store the values in the write arrays
    //         arr_u16_writeAddr[siz_request] = ja_xAddr[i].as<uint16_t>();
    //         arr_u16_writeVal[siz_request] = ja_xDef[i].as<uint16_t>();
    //         siz_request++;
    //         arr_u16_writeAddr[siz_request] = ja_yAddr[i].as<uint16_t>();
    //         arr_u16_writeVal[siz_request] = ja_yDef[i].as<uint16_t>();
    //         siz_request++;
    //     }

    //     // Write the curve data to the inverter
    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_ResetCurve() failed to execute");
    //         return -EXIT_FAILURE;
    //     }

    //     memset(arr_u16_writeAddr, 0, sizeof(arr_u16_writeAddr));
    //     memset(arr_u16_writeVal, 0, sizeof(arr_u16_writeVal));
    //     siz_request = 0;
    // }

    // //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // if(u32_rstCtrlFlags & SE_opModHVRTMustTrip_exists) 
    // {
    //     StackScop ArduinoJson::JsonObject jo_curve = jo_curves["5"];
    //     StackScop ArduinoJson::JsonArray ja_xAddr = jo_curve["x_addr"];
    //     StackScop ArduinoJson::JsonArray ja_yAddr = jo_curve["y_addr"];
    //     StackScop ArduinoJson::JsonArray ja_xDef = jo_curve["x_def"];
    //     StackScop ArduinoJson::JsonArray ja_yDef = jo_curve["y_def"];

    //     for(uint8_t i = 0; i < ja_xAddr.size(); i++)
    //     {
    //         // Store the values in the write arrays
    //         arr_u16_writeAddr[siz_request] = ja_xAddr[i].as<uint16_t>();
    //         arr_u16_writeVal[siz_request] = ja_xDef[i].as<uint16_t>();
    //         siz_request++;
    //         arr_u16_writeAddr[siz_request] = ja_yAddr[i].as<uint16_t>();
    //         arr_u16_writeVal[siz_request] = ja_yDef[i].as<uint16_t>();
    //         siz_request++;
    //     }

    //     // Write the curve data to the inverter
    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_ResetCurve() failed to execute");
    //         return -EXIT_FAILURE;
    //     }

    //     memset(arr_u16_writeAddr, 0, sizeof(arr_u16_writeAddr));
    //     memset(arr_u16_writeVal, 0, sizeof(arr_u16_writeVal));
    //     siz_request = 0;
    // }

    // //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // if(u32_rstCtrlFlags & SE_opModLFRTMustTrip_exists) 
    // {
    //     StackScop ArduinoJson::JsonObject jo_curve = jo_curves["7"];
    //     StackScop ArduinoJson::JsonArray ja_xAddr = jo_curve["x_addr"];
    //     StackScop ArduinoJson::JsonArray ja_yAddr = jo_curve["y_addr"];
    //     StackScop ArduinoJson::JsonArray ja_xDef = jo_curve["x_def"];
    //     StackScop ArduinoJson::JsonArray ja_yDef = jo_curve["y_def"];

    //     for(uint8_t i = 0; i < ja_xAddr.size(); i++)
    //     {
    //         // Store the values in the write arrays
    //         arr_u16_writeAddr[siz_request] = ja_xAddr[i].as<uint16_t>();
    //         arr_u16_writeVal[siz_request] = ja_xDef[i].as<uint16_t>();
    //         siz_request++;
    //         arr_u16_writeAddr[siz_request] = ja_yAddr[i].as<uint16_t>();
    //         arr_u16_writeVal[siz_request] = ja_yDef[i].as<uint16_t>();
    //         siz_request++;
    //     }

    //     // Write the curve data to the inverter
    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_ResetCurve() failed to execute");
    //         return -EXIT_FAILURE;
    //     }

    //     memset(arr_u16_writeAddr, 0, sizeof(arr_u16_writeAddr));
    //     memset(arr_u16_writeVal, 0, sizeof(arr_u16_writeVal));
    //     siz_request = 0;
    // }

    // //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // if(u32_rstCtrlFlags & SE_opModLVRTMustTrip_exists) 
    // {
    //     StackScop ArduinoJson::JsonObject jo_curve = jo_curves["10"];
    //     StackScop ArduinoJson::JsonArray ja_xAddr = jo_curve["x_addr"];
    //     StackScop ArduinoJson::JsonArray ja_yAddr = jo_curve["y_addr"];
    //     StackScop ArduinoJson::JsonArray ja_xDef = jo_curve["x_def"];
    //     StackScop ArduinoJson::JsonArray ja_yDef = jo_curve["y_def"];

    //     for(uint8_t i = 0; i < ja_xAddr.size(); i++)
    //     {
    //         // Store the values in the write arrays
    //         arr_u16_writeAddr[siz_request] = ja_xAddr[i].as<uint16_t>();
    //         arr_u16_writeVal[siz_request] = ja_xDef[i].as<uint16_t>();
    //         siz_request++;
    //         arr_u16_writeAddr[siz_request] = ja_yAddr[i].as<uint16_t>();
    //         arr_u16_writeVal[siz_request] = ja_yDef[i].as<uint16_t>();
    //         siz_request++;
    //     }

    //     // Write the curve data to the inverter
    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_ResetCurve() failed to execute");
    //         return -EXIT_FAILURE;
    //     }

    //     memset(arr_u16_writeAddr, 0, sizeof(arr_u16_writeAddr));
    //     memset(arr_u16_writeVal, 0, sizeof(arr_u16_writeVal));
    //     siz_request = 0;
    // }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    if(u32_rstCtrlFlags & SE_opModVoltVar_exists) 
    {
        StackScop ArduinoJson::JsonObject jo_curve = jo_curves["11"];
        StackScop ArduinoJson::JsonArray ja_xAddr = jo_curve["x_addr"];
        StackScop ArduinoJson::JsonArray ja_yAddr = jo_curve["y_addr"];
        StackScop ArduinoJson::JsonArray ja_xDef = jo_curve["x_def"];
        StackScop ArduinoJson::JsonArray ja_yDef = jo_curve["y_def"];

        for(uint8_t i = 0; i < ja_xAddr.size(); i++)
        {
            // Store the values in the write arrays
            arr_u16_writeAddr[siz_request] = (uint16_t)strtol(ja_xAddr[i].as<const char*>(), nullptr, CST_HEX_BASE);
            arr_u16_writeVal[siz_request] = ja_xDef[i].as<uint16_t>();
            siz_request++;
            arr_u16_writeAddr[siz_request] = (uint16_t)strtol(ja_yAddr[i].as<const char*>(), nullptr, CST_HEX_BASE);
            arr_u16_writeVal[siz_request] = ja_yDef[i].as<uint16_t>();
            siz_request++;
        }

        // Read all of the single registers required
        StackScop uint16_t arr_u16_singleReadAddresses[] = {
            (uint16_t)strtol(jo_curve["autonomousVRefEnable"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop size_t siz_singleReads = sizeof(arr_u16_singleReadAddresses) / sizeof(arr_u16_singleReadAddresses[0]);
        StackScop uint16_t arr_u16_singleReadValues[siz_singleReads] = {0};
        if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                siz_singleReads) != EXIT_SUCCESS)
        {
            LOG_ERR("cst_GetSingleReadValuesFromObject() Failed to read the single registers required");
            return -EXIT_FAILURE;
        }

        StackScop uint16_t u16_bitMask = ((1 << jo_curve["autonomousVRefEnable"]["b_len"].as<uint16_t>()) - 1) 
                                            << jo_curve["autonomousVRefEnable"]["b_st"].as<uint16_t>();
        StackScop uint16_t u16_enVal = arr_u16_singleReadValues[0];
        
        // Set the specified bits to 1
        if (jo_curve["autonomousVRefEnable"]["def"].as<uint16_t>() == 1) 
        {
            u16_enVal |= u16_bitMask;
        } 
        // Clear the specified bits (set to 0)
        else 
        {
            u16_enVal &= ~u16_bitMask;
        }


        arr_u16_writeAddr[siz_request] = (uint16_t)strtol(jo_curve["openLoopTms"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE);
        arr_u16_writeVal[siz_request] = jo_curve["openLoopTms"]["def"].as<uint16_t>();
        siz_request++;
    
        arr_u16_writeAddr[siz_request] = (uint16_t)strtol(jo_curve["vRef"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE);
        arr_u16_writeVal[siz_request] = jo_curve["vRef"]["def"].as<uint16_t>();
        siz_request++;
        
        arr_u16_writeAddr[siz_request] = (uint16_t)strtol(jo_curve["autonomousVRefEnable"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE);
        arr_u16_writeVal[siz_request] = u16_enVal;
        siz_request++;

        arr_u16_writeAddr[siz_request] = (uint16_t)strtol(jo_curve["autonomousVRefTimeConstant"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE);
        arr_u16_writeVal[siz_request] = jo_curve["autonomousVRefTimeConstant"]["def"].as<uint16_t>();
        siz_request++;

        cst_LogAddrVal(arr_u16_writeAddr, arr_u16_writeVal, siz_request);
        
        // Write the curve data to the inverter
        StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
        if (s16_WriteSt != EXIT_SUCCESS)
        {
            LOG_ERR("CST_ResetCurve() failed to execute");
            return -EXIT_FAILURE;
        }

        memset(arr_u16_writeAddr, 0, sizeof(arr_u16_writeAddr));
        memset(arr_u16_writeVal, 0, sizeof(arr_u16_writeVal));
        siz_request = 0;
    }

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    if(u32_rstCtrlFlags & SE_opModVoltWatt_exists) 
    {
        StackScop ArduinoJson::JsonObject jo_curve = jo_curves["12"];
        StackScop ArduinoJson::JsonArray ja_xAddr = jo_curve["x_addr"];
        StackScop ArduinoJson::JsonArray ja_yAddr = jo_curve["y_addr"];
        StackScop ArduinoJson::JsonArray ja_xDef = jo_curve["x_def"];
        StackScop ArduinoJson::JsonArray ja_yDef = jo_curve["y_def"];

        for(uint8_t i = 0; i < ja_xAddr.size(); i++)
        {
            // Store the values in the write arrays
            arr_u16_writeAddr[siz_request] = (uint16_t)strtol(ja_xAddr[i].as<const char*>(), nullptr, CST_HEX_BASE);
            arr_u16_writeVal[siz_request] = ja_xDef[i].as<uint16_t>();
            siz_request++;
            arr_u16_writeAddr[siz_request] = (uint16_t)strtol(ja_yAddr[i].as<const char*>(), nullptr, CST_HEX_BASE);
            arr_u16_writeVal[siz_request] = ja_yDef[i].as<uint16_t>();
            siz_request++;
        }

        arr_u16_writeAddr[siz_request] = (uint16_t)strtol(jo_curve["openLoopTms"]["addr"].as<const char*>(), nullptr, CST_HEX_BASE);
        arr_u16_writeVal[siz_request] = jo_curve["openLoopTms"]["def"].as<uint16_t>();
        siz_request++;

        cst_LogAddrVal(arr_u16_writeAddr, arr_u16_writeVal, siz_request);

        // Write the curve data to the inverter
        StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
        if (s16_WriteSt != EXIT_SUCCESS)
        {
            LOG_ERR("CST_ResetCurve() failed to execute");
            return -EXIT_FAILURE;
        }

        memset(arr_u16_writeAddr, 0, sizeof(arr_u16_writeAddr));
        memset(arr_u16_writeVal, 0, sizeof(arr_u16_writeVal));
        siz_request = 0;
    }

    // //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // if(u32_rstCtrlFlags & SE_opModWattVar_exists) 
    // {
    //     StackScop ArduinoJson::JsonObject jo_curve = jo_curves["14"];
    //     StackScop ArduinoJson::JsonArray ja_xAddr = jo_curve["x_addr"];
    //     StackScop ArduinoJson::JsonArray ja_yAddr = jo_curve["y_addr"];
    //     StackScop ArduinoJson::JsonArray ja_xDef = jo_curve["x_def"];
    //     StackScop ArduinoJson::JsonArray ja_yDef = jo_curve["y_def"];

    //     for(uint8_t i = 0; i < ja_xAddr.size(); i++)
    //     {
    //         // Store the values in the write arrays
    //         arr_u16_writeAddr[siz_request] = ja_xAddr[i].as<uint16_t>();
    //         arr_u16_writeVal[siz_request] = ja_xDef[i].as<uint16_t>();
    //         siz_request++;
    //         arr_u16_writeAddr[siz_request] = ja_yAddr[i].as<uint16_t>();
    //         arr_u16_writeVal[siz_request] = ja_yDef[i].as<uint16_t>();
    //         siz_request++;
    //     }

    //     arr_u16_writeAddr[siz_request] = jo_curve["en"]["addr"].as<uint16_t>();
    //     arr_u16_writeVal[siz_request] = jo_curve["en"]["def"].as<uint16_t>();
    //     siz_request++;

    //     // Write the curve data to the inverter
    //     StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
    //     if (s16_WriteSt != EXIT_SUCCESS)
    //     {
    //         LOG_ERR("CST_ResetCurve() failed to execute");
    //         return -EXIT_FAILURE;
    //     }

    //     memset(arr_u16_writeAddr, 0, sizeof(arr_u16_writeAddr));
    //     memset(arr_u16_writeVal, 0, sizeof(arr_u16_writeVal));
    //     siz_request = 0;
    // }

    // //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop int CST_UtilCMD(uint8_t u8_loopCnt)
{
    if(u8_loopCnt == 3) 
    {
    //     pthread_mutex_lock(&cst_mtx_lockedRegisters);
    //     bool bol_conflictFound = false;
    //     std::vector <uint16_t> vec_ConflictReg;

    //     // Scan through conflict register list to see if any are locked
    //     StackScop ArduinoJson::JsonArray jo_conflictReg = jd_translationMap["conflict_registers"];
    //     if (!jo_conflictReg.isNull()) 
    //     {
    //         for (ArduinoJson::JsonVariant value : jo_conflictReg) 
    //         {
    //             uint16_t u16_exclusiveRegAddr = value.as<uint16_t>(); 

    //             // If one of the registers in the conflict list is locked, set bol_conflictFound to true
    //             if (cst_uset_lockedRegisters.find(u16_exclusiveRegAddr) != cst_uset_lockedRegisters.end()) 
    //             {
    //                 vec_ConflictReg.push_back(u16_exclusiveRegAddr);
    //                 bol_conflictFound = true;
    //             }
    //         }
    //     } 
    //     pthread_mutex_unlock(&cst_mtx_lockedRegisters);

    //     // Print out all the exclusive registers that are locked
    //     for(int i = 0; i < vec_ConflictReg.size(); i++)
    //     {
    //         LOG_INF("CST_UtilCMD() Exclusive register is currently locked: %d", vec_ConflictReg[i]);
    //     }

    //     // If there was no conflict register locked, send util_cmd_en-0
    //     if(bol_conflictFound == false) 
    //     {
    //         LOG_INF("CST_UtilCMD() No register conflict found, sending util_cmd_en-0");
    //         int int_arr_Statuses[1] = {0};
    //         IAL_Data_SetEMSConfiguration(IAL_TARGET_COMP, IAL_sz_UNIT_TYPE_EMS, COD_u8_NULL_UNIT_ID, 1, 
    //                                         CST_sz_key_utilCmdEN, CST_sz_val_utilCmdEnOFF, int_arr_Statuses);

    //         // Send settings modified event 
    //         StackScop int16_t s16_WriteEventSt = IAL_Ctl_SettingsModified(CST_sz_unitTypesPCS, 1);
    //         if (s16_WriteEventSt != EXIT_SUCCESS)
    //         {
    //             LOG_ERR("CST_SetControl() failed to send settings updated event");
    //             return -EXIT_FAILURE;
    //         }

    //         return EXIT_SUCCESS;
    //     }
    //     else
    //     {
    //         return -EXIT_FAILURE;
    //     }
    }
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop void CST_UtilCMDStatusSetter(CST_UtilModeStatus_t enm_ModelStatus)
{
    pthread_mutex_lock(&cst_mtx_modelStatus);
    cst_enm_ModelStatus = enm_ModelStatus;
    pthread_mutex_unlock(&cst_mtx_modelStatus);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop CST_UtilModeStatus_t CST_UtilCMDStatusGetter(void)
{
    CST_UtilModeStatus_t enm_tempStatus;
    pthread_mutex_lock(&cst_mtx_modelStatus);
    enm_tempStatus = cst_enm_ModelStatus;
    pthread_mutex_unlock(&cst_mtx_modelStatus);
    return enm_tempStatus;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop int CST_ModelConflict(void)
{
    // Reset the locked registers (both locally and ICC function to DCU)
    pthread_mutex_lock(&cst_mtx_lockedRegisters);
    cst_uset_lockedRegisters.clear();
    pthread_mutex_unlock(&cst_mtx_lockedRegisters);

    // #TODO - should we handle if this reports a failure?
    // StackScop int16_t s16_ResetLocksSt = IAL_Mbs_ResetLockRegistry(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, 
    //                                                                 COD_u8_NULL_UNIT_ID);

    uint8_t u8_utilCmdState = cst_getUtilCmdState();
    if(u8_utilCmdState == 1)
    {
        // // Send util_cmd_en-0
        // LOG_INF("CST_UtilCMD() reset of all locked registers, sending util_cmd_en-0");
        // int int_arr_Statuses[1] = {0};
        // IAL_Data_SetEMSConfiguration(IAL_TARGET_COMP, IAL_sz_UNIT_TYPE_EMS, COD_u8_NULL_UNIT_ID, 1, 
        //                                 CST_sz_key_utilCmdEN, CST_sz_val_utilCmdEnOFF, int_arr_Statuses);

        // // Send settings modified event 
        // StackScop int16_t s16_WriteEventSt = IAL_Ctl_SettingsModified(CST_sz_unitTypesPCS, 1);
        // if (s16_WriteEventSt != EXIT_SUCCESS)
        // {
        //     LOG_ERR("CST_SetControl() failed to send settings updated event");
        //     return -EXIT_FAILURE;
        // }
        cst_setUtilCmdState(0);
    }

    return EXIT_SUCCESS;
}

/**********************************************************************************************************************
 * @cond                                               FILE SCOPE                                              @endcond
 *********************************************************************************************************************/

FileScop int cst_GetParameterTypeString(cst_ParameterType_t enm_type, char *sz_type)
{
    switch (enm_type)
    {
        case CST_CAPABILITY:
            strcpy(sz_type, "capability");
            break;
        case CST_SETTING:
            strcpy(sz_type, "settings");
            break;
        case CST_STATUS:
            strcpy(sz_type, "status");
            break;
        case CST_METER_READING:
            strcpy(sz_type, "mirror_meter_readings");
            break;
        case CST_LOG_EVENT:
            strcpy(sz_type, "log_events");
            break;
        case CST_CONTROL:
            strcpy(sz_type, "controls");
            break;
        case CST_CURVE:
            strcpy(sz_type, "curves");
            break;
        case CST_DEFAULT_CONTROL:
            strcpy(sz_type, "settings");    // Note that the default control are stored in the settings
            break;
        default:
            LOG_ERR("cst_GetParameterTypeString() Failed to find the parameter type for the given key");
            return -EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


FileScop int cst_GetSingleReadValuesFromObject(ArduinoJson::JsonObject jo_type, uint16_t *arr_u16_singleReadAddresses, 
                                                uint16_t *arr_u16_singleReadValues, size_t *ptr_singleReads)
{
    // Determine what registers need to be read
    for (ArduinoJson::JsonPair jp_pair : jo_type)
    {
        StackScop const char *sz_key = jp_pair.key().c_str();
        StackScop ArduinoJson::JsonObject jo_parameter = jo_type[sz_key];
        StackScop cst_ParameterAttributes_t str_attributes;

        // Retreive the parameter's attributes
        if (cst_GetAttributes(jo_parameter, &str_attributes) == EXIT_SUCCESS)
        {
            // Ensure the parameter has a Modbus address
            if (!str_attributes.jv_modbusAddress.isNull())
            {
                // Determine if the value can be read in a single register
                if(str_attributes.jv_bitLength.as<uint16_t>() <= CST_SINGLE_REGISTER_SIZE)
                {
                    StackScop const char* ptr_sz_addr = str_attributes.jv_modbusAddress.as<const char*>();
                    StackScop uint16_t u16_addr = strtol(ptr_sz_addr, nullptr, CST_HEX_BASE);
                    arr_u16_singleReadAddresses[(*ptr_singleReads)++] = u16_addr;
                }
            }
        }
    }

    // Read all of the single registers required
    if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
            *ptr_singleReads) != EXIT_SUCCESS)
    {
        LOG_ERR("cst_GetSingleReadValuesFromObject() Failed to read the single registers required");
        return -EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop int cst_GetAttributes(ArduinoJson::JsonObject jo_parameter, cst_ParameterAttributes_t *ptr_attributes)
{
    StackScop ArduinoJson::JsonObject jo_DefaultConfiguration = jd_translationMap["default_attribute"];

    // // Retrieve the Modbus Address from the map
    // const char* ptr_sz_addr = jo_parameter["addr"].as<const char*>();
    // if(ptr_sz_addr != nullptr)
    // {
    //     int u16_addr = strtol(ptr_sz_addr, nullptr, CST_HEX_BASE);
    //     jo_parameter["addr"] = u16_addr;
    //     ptr_attributes->jv_modbusAddress = jo_parameter["addr"].as<uint16_t>();
    //     // ptr_attributes->jv_modbusAddress = u16_addr;
    // }

    ptr_attributes->jv_modbusAddress = jo_parameter["addr"];
    if (ptr_attributes->jv_modbusAddress.isNull())
    {
        // Check if the parameter needs to be calculated
        if (jo_parameter["calc"].as<bool>())
        {
            return CST_CALCULATION_REQUIRED;
        }
        return -EXIT_FAILURE; // Key did not contain a Modbus address
    }

    // Retrieve all other needed configurations
    ptr_attributes->jv_bitStartIndex = jo_parameter["b_st"];
    if (ptr_attributes->jv_bitStartIndex.isNull())
    {
        ptr_attributes->jv_bitStartIndex = jo_DefaultConfiguration["b_st"];
    }

    ptr_attributes->jv_bitLength = jo_parameter["b_len"];
    if (ptr_attributes->jv_bitLength.isNull())
    {
        ptr_attributes->jv_bitLength = jo_DefaultConfiguration["b_len"];
    }

    ptr_attributes->jv_scalingFactor = jo_parameter["sc"];
    if (ptr_attributes->jv_scalingFactor.isNull())
    {
        ptr_attributes->jv_scalingFactor = jo_DefaultConfiguration["sc"];
    }

    ptr_attributes->jv_offset = jo_parameter["ofst"];
    if (ptr_attributes->jv_offset.isNull())
    {
        ptr_attributes->jv_offset = jo_DefaultConfiguration["ofst"];
    }

    ptr_attributes->jv_isSigned = jo_parameter["s"];
    if (ptr_attributes->jv_isSigned.isNull())
    {
        ptr_attributes->jv_isSigned = jo_DefaultConfiguration["s"];
    }

    ptr_attributes->jv_canWrite = jo_parameter["w"];
    if (ptr_attributes->jv_canWrite.isNull())
    {
        ptr_attributes->jv_canWrite = jo_DefaultConfiguration["w"];
    }

    ptr_attributes->jv_addressOffset = jo_parameter["a_ofst"];
    if (ptr_attributes->jv_addressOffset.isNull())
    {
        ptr_attributes->jv_addressOffset = jo_DefaultConfiguration["a_ofst"];
    }

    ptr_attributes->jv_stride = jo_parameter["a_str"];
    if (ptr_attributes->jv_stride.isNull())
    {
        ptr_attributes->jv_stride = jo_DefaultConfiguration["a_str"];
    }

    ptr_attributes->jv_negate = jo_parameter["neg"];
    if (ptr_attributes->jv_negate.isNull())
    {
        ptr_attributes->jv_negate = jo_DefaultConfiguration["neg"];
    }

    ptr_attributes->jv_absoluteValue = jo_parameter["abs"];
    if (ptr_attributes->jv_absoluteValue.isNull())
    {
        ptr_attributes->jv_absoluteValue = jo_DefaultConfiguration["abs"];
    }

    // ptr_attributes->jv_batchStart = jo_parameter["batch_st"];
    // if (ptr_attributes->jv_batchStart.isNull())
    // {
    //     // Default Batch Start Address is the same as the Modbus Address
    //     ptr_attributes->jv_batchStart = ptr_attributes->jv_modbusAddress;  
    // }
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


FileScop int cst_GetSingleReadValuesFromArray(ArduinoJson::JsonArray ja_type, uint16_t *arr_u16_singleReadAddresses, 
                                                uint16_t *arr_u16_singleReadValues, size_t *ptr_singleReads)
{
    // Determine what registers need to be read
    for (ArduinoJson::JsonObject jo_parameter : ja_type)
    {
        StackScop cst_ParameterAttributes_t str_attributes;
        // Retreive the parameter's attributes
        if (cst_GetAttributes(jo_parameter, &str_attributes) == EXIT_SUCCESS)
        {
            // Ensure the parameter has a Modbus address
            if (!str_attributes.jv_modbusAddress.isNull())
            {
                // Determine if the value can be read in a single register
                if(str_attributes.jv_bitLength.as<uint16_t>() <= CST_SINGLE_REGISTER_SIZE)
                {
                    StackScop const char* ptr_sz_addr = str_attributes.jv_modbusAddress.as<const char*>();
                    StackScop uint16_t u16_addr = strtol(ptr_sz_addr, nullptr, CST_HEX_BASE);
                    arr_u16_singleReadAddresses[(*ptr_singleReads)++] = u16_addr;
                }
            }
        }
    }

    // Read all of the single registers required
    if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
            *ptr_singleReads) != EXIT_SUCCESS)
    {
        LOG_ERR("cst_GetSingleReadValuesFromObject() Failed to read the single registers required");
        return -EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
}


// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop int cst_GetDERControlType(uint8_t *arr_u8_controlType, cst_ParameterType_t enm_type, const char *sz_key)
{
    // Retrieve the parameter from the translation map
    StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
    StackScop cst_GetParameterTypeString(enm_type, sz_type);
    StackScop ArduinoJson::JsonObject jo_parameter = jd_translationMap[sz_type][sz_key];

    // Retrieve the attributes for the given parameter
    StackScop cst_ParameterAttributes_t str_attributes;
    cst_GetAttributes(jo_parameter, &str_attributes);

    // Modes supported parameter is not stored in the inverter correctly, so the value is simply retrieved from the JSON
    if (strcmp(sz_key, "modesSupported") == EXIT_SUCCESS)
    {
        // Read the bitfield from the JSON document
        StackScop const char* ptr_sz_value = jo_parameter["val"];
        ptr_sz_value+=CST_HEX_PTR_OFS;

        // Convert modesSupported hex string to a 32bit integer
        StackScop uint32_t u32_value = strtoul(ptr_sz_value, NULL, CST_HEX_BASE);

        // Transfer the 32bit integer to the control type array
        arr_u8_controlType[3] = (uint8_t)(u32_value & CST_LSB_MASK);
        arr_u8_controlType[2] = (uint8_t)((u32_value >> CST_BYTE_SIZE) & CST_LSB_MASK);
        arr_u8_controlType[1] = (uint8_t)((u32_value >> CST_BYTE_SIZE*2) & CST_LSB_MASK);
        arr_u8_controlType[0] = (uint8_t)((u32_value >> CST_BYTE_SIZE*3) & CST_LSB_MASK);
    }
    else if (strcmp(sz_key, "modesEnabled") == EXIT_SUCCESS)
    {

        StackScop uint32_t u32_modesEnabled = 0x00000000;

        // Read the BAT Power from the inverter (sign convention determines overall charge/discharge state)
        // #TODO - magic number 314A
        StackScop uint16_t arr_u16_readAddr[] =   { 
                                    (uint16_t)strtol("0x314A", nullptr, CST_HEX_BASE)
        };
        StackScop size_t siz_readReq = sizeof(arr_u16_readAddr) / sizeof(arr_u16_readAddr[0]);
        StackScop uint16_t arr_u16_readVal[siz_readReq] = { 0 }; 
        if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_readAddr, 
                                arr_u16_readVal, siz_readReq) != EXIT_SUCCESS)
        {
            LOG_ERR("CST_GetCalculatedParameter() failed to read charge/discharge status from the inverter");
        }

        StackScop const uint32_t u32_chargeBit = 0;    // (+) Charge mode (bit position 0)
        StackScop const uint32_t u32_dischargeBit = 1; // (-) Discharge mode (bit position 1)

        // Set the Charge mode bit and clear the Discharge mode bit
        if ((int16_t)arr_u16_readVal[0] > 0) 
        {
            u32_modesEnabled |= (1 << u32_chargeBit);         
            u32_modesEnabled &= ~(1 << u32_dischargeBit);     
        }
        // Set the Discharge mode bit and clear the Charge mode bit
        else if ((int16_t)arr_u16_readVal[0] < 0) 
        {
            u32_modesEnabled |= (1 << u32_dischargeBit);       
            u32_modesEnabled &= ~(1 << u32_chargeBit);          
        }
        // Clear both Charge and Discharge mode bits
        else 
        {
            u32_modesEnabled &= ~(1 << u32_chargeBit);          
            u32_modesEnabled &= ~(1 << u32_dischargeBit);   
        }

        // Read the data from the inverter required for the calculation
        StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
        cst_GetParameterTypeString(enm_type, sz_type);
        StackScop ArduinoJson::JsonObject jo_modes = jd_translationMap[sz_type][sz_key]["modes"];
        StackScop uint16_t arr_u16_singleReadAddresses[CST_MAX_READ_SIZE] = {0};
        StackScop uint16_t arr_u16_singleReadValues[CST_MAX_READ_SIZE] = {0};
        StackScop size_t siz_singleReads = 0;
        cst_GetSingleReadValuesFromObject(jo_modes, arr_u16_singleReadAddresses, arr_u16_singleReadValues, &siz_singleReads);

        // 2 - opModConnect -> 26 - opModWattVar (Watt-Var Mode)
        StackScop uint32_t u32_mask = 0b100;  // Start at bit 2 (0-indexed)
        for (ArduinoJson::JsonPair jp_pair : jo_modes)
        {
            const char *sz_key = jp_pair.key().c_str();
            if (cst_IsModeEnabled(jo_modes[sz_key], sz_key, arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                                    siz_singleReads))
            {
                u32_modesEnabled |= u32_mask;
            }
            else
            {
                u32_modesEnabled &= (~u32_mask);
            }
            u32_mask <<= 1;
        }

        // Set the control type
        arr_u8_controlType[0] = (uint8_t)(u32_modesEnabled & 0x000000FF);
        arr_u8_controlType[1] = (uint8_t)((u32_modesEnabled & 0x0000FF00) >> CST_BYTE_SIZE);
        arr_u8_controlType[2] = (uint8_t)((u32_modesEnabled & 0x00FF0000) >> (CST_BYTE_SIZE * 2));
        arr_u8_controlType[3] = (uint8_t)((u32_modesEnabled & 0xFF000000) >> (CST_BYTE_SIZE * 3));
    }
    else
    {
        LOG_ERR("cst_GetDERControlType(): sz_key not supported: %s", sz_key);
        return -EINVAL;
    }
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop int cst_GetCurrentRMS(SE_CurrentRMS_t *ptr_CurrentRMS, cst_ParameterType_t enm_type, const char *sz_key)
{
    // Retrieve the parameter from the translation map
    StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
    cst_GetParameterTypeString(enm_type, sz_type);
    StackScop ArduinoJson::JsonObject jo_parameter = jd_translationMap[sz_type][sz_key];

    // Retrieve the attributes for the given parameter
    StackScop cst_ParameterAttributes_t str_attributes;
    if (cst_GetAttributes(jo_parameter, &str_attributes) == CST_CALCULATION_REQUIRED)
    {
        // Read the registers required for the calculation
        StackScop uint16_t arr_u16_singleReadAddresses[] = { 
            (uint16_t)strtol(jo_parameter["p_addr"].as<const char*>(), nullptr, CST_HEX_BASE), 
            (uint16_t)strtol(jo_parameter["v_addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop size_t siz_singleReads = sizeof(arr_u16_singleReadAddresses) / sizeof(arr_u16_singleReadAddresses[0]);
        StackScop uint16_t arr_u16_singleReadValues[siz_singleReads] = {0};
        if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_singleReadAddresses, arr_u16_singleReadValues, 
                                    siz_singleReads) != EXIT_SUCCESS)
        {
            LOG_ERR("cst_GetCurrentRMS() Failed to read the single registers required");
            return -EXIT_FAILURE;
        }

        // Perform the calculation (P = IV -> I = P/V)
        StackScop uint16_t u16_rtgMaxW = (arr_u16_singleReadValues[0]);
        StackScop uint16_t u16_rtgNomV = arr_u16_singleReadValues[1];
        StackScop uint16_t u16_rtgMaxA = u16_rtgMaxW / u16_rtgNomV;
        ptr_CurrentRMS->value = u16_rtgMaxA;
    }
    else 
    {
        // @todo: Implement if no calculation is needed
    }
    // Set the multiplier
    cst_GetParameterMultiplier(enm_type, sz_key, &ptr_CurrentRMS->multiplier);
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop int cst_GetParameterValue(cst_ParameterType_t enm_type, const char *sz_key, uint16_t *arr_u16_readAddresses, 
                                    uint16_t *arr_u16_readValues, size_t siz_reads, int32_t *ptr_value)
{
    // Retrieve the parameter from the translation map
    StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
    cst_GetParameterTypeString(enm_type, sz_type);
    StackScop ArduinoJson::JsonObject jo_parameter = jd_translationMap[sz_type][sz_key];

    // Check if the type is an array (mirror meter readings)
    if (!jd_translationMap[sz_type].as<ArduinoJson::JsonArray>().isNull())
    {
        // Iterate through the array to find the parameter matching the description key
        for (ArduinoJson::JsonObject jo_object : jd_translationMap[sz_type].as<ArduinoJson::JsonArray>())
        {
            if (strcmp(sz_key, jo_object["desc"].as<const char*>()) == EXIT_SUCCESS)
            {
                jo_parameter = jo_object;
            }
        }
    }

    // Retrieve the attributes for the given parameter
    StackScop cst_ParameterAttributes_t str_attributes;
    StackScop int s16_getSt = cst_GetAttributes(jo_parameter, &str_attributes);
    if (s16_getSt == -EXIT_FAILURE)
    {
        // Parameter is not supported by the map. Set value accordingly
        *ptr_value = jo_parameter["val"].as<int32_t>();
        return -EINVAL;
    }
    else if (s16_getSt == CST_CALCULATION_REQUIRED)
    {
        // Perform the calculation
        return cst_GetCalculatedParameter(jo_parameter, sz_key, ptr_value);
    }

    // Find the read value from the array of previously read values
    StackScop uint16_t u16_rawValue = 0;
    for (uint8_t i = 0; i < siz_reads; i++)
    {
        if (arr_u16_readAddresses[i] == strtol((jo_parameter["addr"].as<const char*>()), nullptr, CST_HEX_BASE))
        {
            u16_rawValue = arr_u16_readValues[i];
            break;
        }
        if (i == siz_reads - 1)
        {
            LOG_ERR("cst_GetParameterValue() Failed to find the read value for the given parameter");
            return -EXIT_FAILURE;
        }
    }

    // Perform manipulations on read value
    StackScop uint8_t u8_bitLength = str_attributes.jv_bitLength.as<unsigned char>();
    StackScop uint32_t u32_baseMask     = 0xffffffff;
    StackScop uint8_t  u8_bitStartIndex = str_attributes.jv_bitStartIndex.as<unsigned char>();
    StackScop uint32_t u32_bitMask =
        (u32_baseMask >> (32 - u8_bitLength - u8_bitStartIndex)) & (u32_baseMask << u8_bitStartIndex);
    StackScop int32_t s32_finalValue = ((u16_rawValue & u32_bitMask) >> u8_bitStartIndex);
    if (str_attributes.jv_isSigned.as<bool>())
    {
        s32_finalValue = (int16_t)s32_finalValue;
    }
    if (str_attributes.jv_absoluteValue.as<bool>())
    {
        s32_finalValue = abs(s32_finalValue);
    }
    s32_finalValue *= str_attributes.jv_scalingFactor.as<float>();
    s32_finalValue += str_attributes.jv_offset.as<signed int>();
    if (str_attributes.jv_negate.as<bool>())
    {
        s32_finalValue = ~s32_finalValue;
    }

    *ptr_value = s32_finalValue;
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop int cst_GetParameterMultiplier(cst_ParameterType_t enm_type, const char *sz_key, 
                                        SE_PowerOfTenMultiplierType_t *ptr_multiplier)
{
    // Retrieve the parameter from the translation map
    StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
    cst_GetParameterTypeString(enm_type, sz_type);
    StackScop ArduinoJson::JsonObject jo_parameter = jd_translationMap[sz_type][sz_key];

    // Retrieve the multiplier for the given parameter
    if (jo_parameter["mult"].isNull())
    {
        LOG_ERR("cst_getParameterMultiplier() Failed to retrieve the multiplier for the given parameter");
        return -EXIT_FAILURE;
    }
    // Set the multiplier
    *ptr_multiplier = jo_parameter["mult"].as<SE_PowerOfTenMultiplierType_t>();
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop int cst_GetCalculatedParameter(ArduinoJson::JsonObject jo_parameter, const char *sz_key, int32_t *ptr_value)
{
    // Determine the calculation to perform based on the key name
    if (strcmp(sz_key, "rtgMaxAh") == EXIT_SUCCESS)
    // || strcmp(sz_key, "setMaxAh") == EXIT_SUCCESS)
    {
        // Read the data from the INV required for the calculation
        StackScop uint16_t arr_u16_readAddr[] =   { 
                                    (uint16_t)strtol(jo_parameter["btr_qty_addr"].as<const char*>(), nullptr, CST_HEX_BASE),
                                    (uint16_t)strtol(jo_parameter["li_cap_addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop size_t siz_readReq = sizeof(arr_u16_readAddr) / sizeof(arr_u16_readAddr[0]);
        StackScop uint16_t arr_u16_readVal[siz_readReq] = { 0 }; 
        if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_readAddr, 
                                arr_u16_readVal, siz_readReq) != EXIT_SUCCESS)
        {
            LOG_ERR("CST_GetCalculatedParameter() failed to read rtgMaxAh from the inverter");
        }

        *ptr_value = (arr_u16_readVal[0] * arr_u16_readVal[1]);
    }

    else if (strcmp(sz_key, "rtgMaxVar") == EXIT_SUCCESS || strcmp(sz_key, "setMaxVar") == EXIT_SUCCESS)
    {
        // Read the data from the INV required for the calculation
        StackScop uint16_t arr_u16_readAddr[] =   { 
                                    (uint16_t)strtol(jo_parameter["pwrat_addr"].as<const char*>(), nullptr, CST_HEX_BASE),
                                    (uint16_t)strtol(jo_parameter["maxva_addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop size_t siz_readReq = sizeof(arr_u16_readAddr) / sizeof(arr_u16_readAddr[0]);
        StackScop uint16_t arr_u16_readVal[siz_readReq] = { 0 }; 
        if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_readAddr, 
                                arr_u16_readVal, siz_readReq) != EXIT_SUCCESS)
        {
            LOG_ERR("CST_GetCalculatedParameter() failed to read rtgMaxVar/setMaxVar from the inverter");
        }

        StackScop float f_pwRatPer = (float)arr_u16_readVal[0] / 100;
        *ptr_value = (int32_t)((float)arr_u16_readVal[1] * f_pwRatPer);    
    }

    else if (strcmp(sz_key, "rtgMaxVarNeg") == EXIT_SUCCESS)
    {
        // Read the data from the INV required for the calculation
        StackScop uint16_t arr_u16_readAddr[] =   { 
                                    (uint16_t)strtol(jo_parameter["pwrat_addr"].as<const char*>(), nullptr, CST_HEX_BASE),
                                    (uint16_t)strtol(jo_parameter["maxva_addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop size_t siz_readReq = sizeof(arr_u16_readAddr) / sizeof(arr_u16_readAddr[0]);
        StackScop uint16_t arr_u16_readVal[siz_readReq] = { 0 }; 
        if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_readAddr, 
                                arr_u16_readVal, siz_readReq) != EXIT_SUCCESS)
        {
            LOG_ERR("CST_GetCalculatedParameter() failed to read rtgMaxVarNeg from the inverter");
        }

        StackScop float f_pwRatPer = (float)arr_u16_readVal[0] / 100;
        *ptr_value = ((int32_t)((float)arr_u16_readVal[1] * f_pwRatPer)) * -1;    
    }

    else if (strcmp(sz_key, "rtgMaxWh") == EXIT_SUCCESS)
    {
        // Read the data from the INV required for the calculation
        StackScop uint16_t arr_u16_readAddr[] =   { 
                                    (uint16_t)strtol(jo_parameter["btr_qty_addr"].as<const char*>(), nullptr, CST_HEX_BASE),
                                    (uint16_t)strtol(jo_parameter["li_cap_addr"].as<const char*>(), nullptr, CST_HEX_BASE),
                                    (uint16_t)strtol(jo_parameter["li_ch_v_addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop size_t siz_readReq = sizeof(arr_u16_readAddr) / sizeof(arr_u16_readAddr[0]);
        StackScop uint16_t arr_u16_readVal[siz_readReq] = { 0 }; 
        if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_readAddr, 
                                arr_u16_readVal, siz_readReq) != EXIT_SUCCESS)
        {
            LOG_ERR("CST_GetCalculatedParameter() failed to read rtgMaxWh from the inverter");
        }

        *ptr_value = ((arr_u16_readVal[0] * arr_u16_readVal[1] * arr_u16_readVal[2]));
    }
    else if ((strcmp(sz_key, "opModFixedPFAbsorbW") == EXIT_SUCCESS) || (strcmp(sz_key, "opModFixedPFInjectW") == EXIT_SUCCESS))
    {
        // Read the data from the INV required for the calculation
        StackScop uint16_t arr_u16_readAddr[] =   { 
                                    (uint16_t)strtol(jo_parameter["pfrd_addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop size_t siz_readReq = sizeof(arr_u16_readAddr) / sizeof(arr_u16_readAddr[0]);
        StackScop uint16_t arr_u16_readVal[siz_readReq] = { 0 }; 
        if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_readAddr, 
                                arr_u16_readVal, siz_readReq) != EXIT_SUCCESS)
        {
            LOG_ERR("CST_GetCalculatedParameter() failed to read PF from the inverter");
        }
        *ptr_value = (arr_u16_readVal[0]);
    }
    else if (strcmp(sz_key, "setMaxA") == EXIT_SUCCESS)
    {
        // // Read the data from the inverter required for the calculation
        // StackScop uint16_t arr_u16_readAddresses[] = {jo_parameter["dc_i"]};
        // StackScop size_t siz_reads = sizeof(arr_u16_readAddresses) / sizeof(arr_u16_readAddresses[0]);
        // StackScop uint16_t arr_u16_readValues[siz_reads] = {0};
        // if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_readAddresses, 
        //                             arr_u16_readValues, siz_reads) != EXIT_SUCCESS)
        // {
        //     LOG_ERR("cst_GetCalculatedParameter() Failed to read the single registers required");
        //     return -EXIT_FAILURE;
        // }
        // StackScop uint16_t u16_dcMaxI = arr_u16_readValues[0] / 10;

        // // Read the data from the PDU required for the calculation
        // StackScop uint32_t u32_pduParametersSize = sizeof(SCH_PduParameters_t);
        // StackScop uint8_t u8_pduParameters[u32_pduParametersSize];
        // StackScop SCH_PduParameters_t *ptr_pduParameters = (SCH_PduParameters_t *)u8_pduParameters;
        // StackScop int s16_rc = IAL_Data_GetBufferedData(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, CST_PDU_INDEX, u8_pduParameters, &u32_pduParametersSize);
        // if (s16_rc != EXIT_SUCCESS || u32_pduParametersSize != sizeof(SCH_PduParameters_t))
        // {
        //     LOG_ERR("Failed to get PDU Data for bytes: %d, with Error: %i", u32_pduParametersSize, s16_rc);
        //     return -EXIT_FAILURE;
        // }
        // StackScop uint8_t u8_brmQuantity = (uint8_t)ptr_pduParameters->signals.bms_qty;

        // // Perform the calculation: MaxA = (u16_dcMaxI*u16_dcNominalV)/(u16_acNominalV)
        // StackScop uint16_t u16_dcNominalV = (u8_brmQuantity * jo_parameter["brm_dv"].as<uint16_t>()) / 10;
        // StackScop uint16_t u16_acMaxV = jo_parameter["ac_v"].as<uint16_t>();
        // StackScop uint16_t u16_setMaxA = (u16_dcMaxI * u16_dcNominalV) / u16_acMaxV;
        // *ptr_value = u16_setMaxA;
    }
    else if (strcmp(sz_key, "setMaxChargeRateW") == EXIT_SUCCESS)
    {
        // Read the data from the INV required for the calculation
        StackScop uint16_t arr_u16_readAddr[] =   { 
                                    (uint16_t)strtol(jo_parameter["chrg_i_addr"].as<const char*>(), nullptr, CST_HEX_BASE),
                                    (uint16_t)strtol(jo_parameter["chrg_v_addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop size_t siz_readReq = sizeof(arr_u16_readAddr) / sizeof(arr_u16_readAddr[0]);
        StackScop uint16_t arr_u16_readVal[siz_readReq] = { 0 }; 
        if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_readAddr, 
                                arr_u16_readVal, siz_readReq) != EXIT_SUCCESS)
        {
            LOG_ERR("CST_GetCalculatedParameter() failed to read setMaxChargeRateW from the inverter");
        }

        *ptr_value = ((arr_u16_readVal[0] * arr_u16_readVal[1]));
    }
    else if (strcmp(sz_key, "setMaxDischargeRateW") == EXIT_SUCCESS)
    {
        // Read the data from the INV required for the calculation
        StackScop uint16_t arr_u16_readAddr[] =   { 
                                    (uint16_t)strtol(jo_parameter["dschrg_i_addr"].as<const char*>(), nullptr, CST_HEX_BASE),
                                    (uint16_t)strtol(jo_parameter["dschrg_v_addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop size_t siz_readReq = sizeof(arr_u16_readAddr) / sizeof(arr_u16_readAddr[0]);
        StackScop uint16_t arr_u16_readVal[siz_readReq] = { 0 }; 
        if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_readAddr, 
                                arr_u16_readVal, siz_readReq) != EXIT_SUCCESS)
        {
            LOG_ERR("CST_GetCalculatedParameter() failed to read setMaxDischargeRateW from the inverter");
        }

        *ptr_value = ((arr_u16_readVal[0] * arr_u16_readVal[1])) * -1;
    }
    else if (strcmp(sz_key, "setMaxW") == EXIT_SUCCESS)
    {
        // Read the data from the INV required for the calculation
        StackScop uint16_t arr_u16_readAddr[] =   { 
                                    (uint16_t)strtol(jo_parameter["v_addr"].as<const char*>(), nullptr, CST_HEX_BASE),
                                    (uint16_t)strtol(jo_parameter["dschrg_i_addr"].as<const char*>(), nullptr, CST_HEX_BASE),
                                    (uint16_t)strtol(jo_parameter["chrg_i_addr"].as<const char*>(), nullptr, CST_HEX_BASE)
        };
        StackScop size_t siz_readReq = sizeof(arr_u16_readAddr) / sizeof(arr_u16_readAddr[0]);
        StackScop uint16_t arr_u16_readVal[siz_readReq] = { 0 }; 
        if (IAL_Mbs_ReadRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_readAddr, 
                                arr_u16_readVal, siz_readReq) != EXIT_SUCCESS)
        {
            LOG_ERR("CST_GetCalculatedParameter() failed to read setMaxW from the inverter");
        }
        
        if(arr_u16_readVal[1] > arr_u16_readVal[2])
        {
            *ptr_value = arr_u16_readVal[0] * arr_u16_readVal[1] * -1;
        }
        else
        {
            *ptr_value = arr_u16_readVal[0] * arr_u16_readVal[2];
        }
    }
    else if (strcmp(sz_key, "stateOfChargeStatus") == EXIT_SUCCESS)
    {
    //     // Read the data from the PDU required for the calculation
    //     StackScop uint32_t u32_pduParametersSize = sizeof(SCH_PduParameters_t);
    //     StackScop uint8_t u8_pduParameters[u32_pduParametersSize];
    //     StackScop SCH_PduParameters_t *ptr_pduParameters = (SCH_PduParameters_t *)u8_pduParameters;
    //     StackScop int s16_rc = IAL_Data_GetBufferedData(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, CST_PDU_INDEX, u8_pduParameters, &u32_pduParametersSize);
    //     if (s16_rc != EXIT_SUCCESS || u32_pduParametersSize != sizeof(SCH_PduParameters_t))
    //     {
    //         LOG_ERR("Failed to get PDU Data for bytes: %d, with Error: %i", u32_pduParametersSize, s16_rc);
    //         return -EXIT_FAILURE;
    //     }
    //     StackScop uint16_t u16_stateOfCharge = (uint16_t)ptr_pduParameters->signals.pdu_soc * jo_parameter["sc"].as<float>();
    //     *ptr_value = u16_stateOfCharge;
    }
    else if (strcmp(sz_key, "storageModeStatus") == EXIT_SUCCESS)
    {
        // StackScop enum enm_storageModeStatus
        // {
        //     STORAGE_CHARGING = 0,
        //     STORAGE_DISCHARGING = 1,
        //     STORAGE_HOLDING = 2
        // };
        // // Read the data from the PDU required for the calculation
        // StackScop uint32_t u32_pduParametersSize = sizeof(SCH_PduParameters_t);
        // StackScop uint8_t u8_pduParameters[u32_pduParametersSize];
        // StackScop SCH_PduParameters_t *ptr_pduParameters = (SCH_PduParameters_t *)u8_pduParameters;
        // StackScop int s16_rc = IAL_Data_GetBufferedData(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, CST_PDU_INDEX, u8_pduParameters, &u32_pduParametersSize);
        // if (s16_rc != EXIT_SUCCESS || u32_pduParametersSize != sizeof(SCH_PduParameters_t))
        // {
        //     LOG_ERR("Failed to get PDU Data for bytes: %d, with Error: %i", u32_pduParametersSize, s16_rc);
        //     return -EXIT_FAILURE;
        // }
        // StackScop int32_t s32_pduP = ptr_pduParameters->signals.pdu_p;
        // if (s32_pduP > 0)
        // {
        //     *ptr_value = STORAGE_CHARGING; 
        // }
        // else if (s32_pduP < 0)
        // {
        //     *ptr_value = STORAGE_DISCHARGING; 
        // }
        // else
        // {
        //     *ptr_value = STORAGE_HOLDING; 
        // }
    }
    else if (!jo_parameter["val"].isNull())
    {
        // Populate the value with the value defined in the translation map
        *ptr_value = jo_parameter["val"].as<int32_t>();
    }
    else
    {
        LOG_ERR("cst_GetCalculatedParameter() Failed to find the calculation for the given key");
        return -EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// /**
//  * @brief   Sets the calculated parameter value for the given parameter
//  * @param   jo_parameter    JsonObject containing the parameter to retrieve the value for
//  * @param   sz_key          Key of the parameter to retrieve the value for
//  * @param   s32_value       Value to be written
//  * @returns Return EXIT_SUCCESS or -EXIT_FAILURE
// */
// FileScop int cst_SetCalculatedParameter(ArduinoJson::JsonObject jo_parameter, const char *sz_key, int32_t s32_value)
// {
//     // Determine the calculation to perform based on the key name
//     if (strcmp(sz_key, "setESRampTms") == EXIT_SUCCESS)
//     {
//         // Get initial input data required to perfrom control
//         StackScop ArduinoJson::JsonObject jo_rampTms = jd_translationMap["controls"]["rampTms"];

//         // Convert the time into a ramp rate (s -> %/s)
//         StackScop float f_rampTms_s = s32_value * 0.01;
//         StackScop float f_setGradW = (CST_MAX_PERCENT * jo_rampTms["setGradW"]["sc"].as<float>()) / f_rampTms_s;
//         StackScop uint16_t u16_setGradW =  (uint16_t)round(f_setGradW);
        
//         // Extract write values and addresses required to perform control 
//         StackScop uint16_t arr_u16_writeAddr[] =  { jo_rampTms["setGradW"]["addr"].as<uint16_t>() };
//         StackScop uint16_t arr_u16_writeVal[] =   { u16_setGradW };

//         StackScop size_t siz_request = sizeof(arr_u16_writeAddr) / sizeof(arr_u16_writeAddr[0]);

//         StackScop int16_t s16_WriteSt = cst_WriteControl(CST_SET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
//         if (s16_WriteSt != EXIT_SUCCESS)
//         {
//             LOG_ERR("cst_SetCalculatedParameter() failed to execute");
//             return -EXIT_FAILURE;
//         }
//         return EXIT_SUCCESS;
//     }
//     else
//     {
//         LOG_ERR("cst_SetCalculatedParameter() Failed to find the calculation for the given key %s",sz_key);
//         return -EXIT_FAILURE;
//     }
//     // return EXIT_SUCCESS;
// }

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// /**
//  * @brief   Resets the calculated parameter value for the given parameter
//  * @param   jo_parameter    JsonObject containing the parameter to retrieve the value for
//  * @param   sz_key          Key of the parameter to retrieve the value for
//  * @returns Return EXIT_SUCCESS or -EXIT_FAILURE
// */
// FileScop int cst_ResetCalculatedParameter(ArduinoJson::JsonObject jo_parameter, const char *sz_key)
// {
//     // Determine the default to write based on the key name
//     if (strcmp(sz_key, "setESRampTms") == EXIT_SUCCESS)
//     {

//         // Get initial input data required to perfrom control
//         StackScop ArduinoJson::JsonObject jo_rampTms = jd_translationMap["controls"]["rampTms"];

//         // Extract write values and addresses required to perform control 
//         StackScop uint16_t arr_u16_writeAddr[] =  { jo_rampTms["setGradW"]["addr"].as<uint16_t>() };
//         StackScop uint16_t arr_u16_writeVal[] =   { jo_rampTms["setGradW"]["def"].as<uint16_t>() };

//         StackScop size_t siz_request = sizeof(arr_u16_writeAddr) / sizeof(arr_u16_writeAddr[0]);

//         StackScop int16_t s16_WriteSt = cst_WriteControl(CST_RESET_REG, siz_request, arr_u16_writeAddr, arr_u16_writeVal);
//         if (s16_WriteSt != EXIT_SUCCESS)
//         {
//             LOG_ERR("cst_ResetCalculatedParameter() failed to execute");
//             return -EXIT_FAILURE;
//         }
//         return EXIT_SUCCESS;
//     }
//     else
//     {
//         LOG_ERR("cst_ResetCalculatedParameter() Failed to find the default value for the given key %s",sz_key);
//         return -EXIT_FAILURE;
//     }
// }

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop bool cst_IsModeEnabled(ArduinoJson::JsonObject jo_mode, const char *sz_key, uint16_t *arr_u16_readAddresses, 
                                uint16_t *arr_u16_readValues, size_t siz_reads)
{
    // Check if the mode calculation can be determined by the inverter
    if (jo_mode["addr"].isNull())
    {
        return jo_mode["val"].as<bool>();
    }

    // Find the read value from the array of previously read values
    StackScop uint16_t u16_rawValue = 0;
    for (uint8_t i = 0; i < siz_reads; i++)
    {
        if (arr_u16_readAddresses[i] == (uint16_t)strtol(jo_mode["addr"].as<const char*>(), nullptr, CST_HEX_BASE))
        {
            u16_rawValue = arr_u16_readValues[i];
            break;
        }
        if (i == siz_reads - 1)
        {
            LOG_ERR("cst_GetParameterValue() Failed to find the read value for the given parameter");
            return -EXIT_FAILURE;
        }
    }

    // Retrieve the attributes for the given parameter
    StackScop cst_ParameterAttributes_t str_attributes;
    StackScop int s16_getSt = cst_GetAttributes(jo_mode, &str_attributes);

    // Check if a direct value comparison is possible
    if(!(jo_mode["val"].isNull()))
    {
        // Determine if the mode is enabled by comparing to the predetermined value
        if (u16_rawValue == jo_mode["val"].as<uint16_t>())
        {
            return true;
        }
        else
        {
            return false;
        }
    }

    // Otherwise, this is a bitfield comparison
    else
    {
        // Perform manipulations on read value
        StackScop uint8_t u8_bitLength = str_attributes.jv_bitLength.as<unsigned char>();
        StackScop uint32_t u32_baseMask     = 0xffffffff;
        StackScop uint8_t  u8_bitStartIndex = str_attributes.jv_bitStartIndex.as<unsigned char>();
        StackScop uint32_t u32_bitMask =
            (u32_baseMask >> (32 - u8_bitLength - u8_bitStartIndex)) & (u32_baseMask << u8_bitStartIndex);
        StackScop int32_t s32_finalValue = ((u16_rawValue & u32_bitMask) >> u8_bitStartIndex);
        if (str_attributes.jv_isSigned.as<bool>())
        {
            s32_finalValue = (int16_t)s32_finalValue;
        }
        if (str_attributes.jv_absoluteValue.as<bool>())
        {
            s32_finalValue = abs(s32_finalValue);
        }
        s32_finalValue *= str_attributes.jv_scalingFactor.as<float>();
        s32_finalValue += str_attributes.jv_offset.as<signed int>();
        if (str_attributes.jv_negate.as<bool>())
        {
            s32_finalValue = ~s32_finalValue;
        }   

        // If non-zero final value, check for specific controls that need to be calculated
        // #TODO - is this the best way to do this?
        if(s32_finalValue)
        {
            StackScop int32_t s32_pfReading = 0;
            if((strcmp(sz_key, "opModFixedPFAbsorbW") == EXIT_SUCCESS))
            {
                cst_GetCalculatedParameter(jo_mode, sz_key, &s32_pfReading);
                if(s32_pfReading < 0)
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
            if((strcmp(sz_key, "opModFixedPFInjectW") == EXIT_SUCCESS))
            {
                cst_GetCalculatedParameter(jo_mode, sz_key, &s32_pfReading);
                if(s32_pfReading > 0)
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
        }


        return s32_finalValue;
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// FileScop int cst_FormatWriteValues(cst_ParameterType_t enm_type, const char *sz_key, int32_t s32_rawValue, 
//                                     uint16_t* arr_u16_writeAddr, uint16_t* arr_u16_writeVal, uint8_t u8_position)
// {
//     // Retrieve the parameter from the translation map
//     StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
//     cst_GetParameterTypeString(enm_type, sz_type);
//     StackScop ArduinoJson::JsonObject jo_parameter = jd_translationMap[sz_type][sz_key];

//     if (!jo_parameter["addr"].isNull())
//     {
//         arr_u16_writeAddr[u8_position] = jo_parameter["addr"].as<uint16_t>();
//         // Scale prior to writing if needed
//         if (!jo_parameter["sc"].isNull())
//         {
//             s32_rawValue /= jo_parameter["sc"].as<float>();
//         }
//         arr_u16_writeVal[u8_position] = (uint16_t)s32_rawValue;
//         return EXIT_SUCCESS;
//     }
//     else if (jo_parameter["calc"].as<bool>())
//     {
//         return cst_SetCalculatedParameter(jo_parameter, sz_key, s32_rawValue);
//     }
//     else
//     {
//         return -EINVAL;
//     }
// }

// FileScop int cst_FormatResetWriteValues(cst_ParameterType_t enm_type, const char *sz_key, uint16_t* arr_u16_writeAddr,
//                                     uint16_t* arr_u16_writeVal, uint8_t u8_position)
// {
//     // Retrieve the parameter from the translation map
//     StackScop char sz_type[CST_PARAMETER_TYPE_SIZE] = {""};
//     cst_GetParameterTypeString(enm_type, sz_type);
//     StackScop ArduinoJson::JsonObject jo_parameter = jd_translationMap[sz_type][sz_key];

//     if (!jo_parameter["addr"].isNull())
//     {
//         arr_u16_writeAddr[u8_position] = jo_parameter["addr"].as<uint16_t>();
//         arr_u16_writeVal[u8_position] = jo_parameter["def"].as<uint16_t>();
//         return EXIT_SUCCESS;
//     }
//     else if (jo_parameter["calc"].as<bool>())
//     {
//         return cst_ResetCalculatedParameter(jo_parameter, sz_key);
//     }
//     else
//     {
//         return -EINVAL;
//     }
// }

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GlobScop int cst_CurveBounds(ArduinoJson::JsonObject jo_curve, uint16_t *arr_u16_writeVal, size_t siz_request)
{

    // Pull the x and y boundaries from the mapping 
    StackScop ArduinoJson::JsonArray ja_xMin = jo_curve["x_min"];
    StackScop ArduinoJson::JsonArray ja_xMax = jo_curve["x_max"];
    StackScop ArduinoJson::JsonArray ja_yMin = jo_curve["y_min"];
    StackScop ArduinoJson::JsonArray ja_yMax = jo_curve["y_max"];

    size_t siz_iter = 0;
    size_t siz_xIx = 0;
    size_t siz_yIx = 0;

    while(siz_iter < siz_request)
    {

        // ~~~~~~~~~~~~~~~~~~~~~~~~~

        // Check the X values
        if ((int16_t)arr_u16_writeVal[siz_iter] < ja_xMin[siz_xIx].as<int16_t>())
        {
            arr_u16_writeVal[siz_iter] = ja_xMin[siz_xIx].as<int16_t>();
        }
        if ((int16_t)arr_u16_writeVal[siz_iter] > ja_xMax[siz_xIx].as<int16_t>())
        {
            arr_u16_writeVal[siz_iter] = ja_xMax[siz_xIx].as<int16_t>();
        }
        siz_iter++;
        siz_xIx++;

        // ~~~~~~~~~~~~~~~~~~~~~~~~~

        // Edge case for opModVoltVar curve for MegaRevo - negative values are not supported
        if(strcmp(jo_curve["desc"], "opModVoltVar") == 0 && (int16_t)arr_u16_writeVal[siz_iter] < 0)
        {
            arr_u16_writeVal[siz_iter] = -1*(int16_t)arr_u16_writeVal[siz_iter];
        }

        // Check the Y values
        if ((int16_t)arr_u16_writeVal[siz_iter] < ja_yMin[siz_yIx].as<int16_t>())
        {
            arr_u16_writeVal[siz_iter] = ja_yMin[siz_yIx].as<int16_t>();
        }
        if ((int16_t)arr_u16_writeVal[siz_iter] > ja_yMax[siz_yIx].as<int16_t>())
        {
            arr_u16_writeVal[siz_iter] = ja_yMax[siz_yIx].as<int16_t>();
        }
        siz_iter++;
        siz_yIx++;
    }

    return EXIT_SUCCESS;
}
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop int cst_WriteControl(cst_writeType_t cst_enum_writeType, size_t siz_request, 
                                    uint16_t* arr_u16_writeAddr, uint16_t* arr_u16_writeVal)
{

    bool bol_ExclusiveRegFound = false;
    std::vector <uint16_t> vec_ConflictReg;

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // If the write type is set, check for exclusive registers (send util_cmd_en-1), lock the registers, 
    // write the values, send settings modified event 
    if(cst_enum_writeType == CST_SET_REG)
    {

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        // Check for exclusive register in addr set being written
        StackScop ArduinoJson::JsonArray jo_conflictReg = jd_translationMap["conflict_registers"];
        if (!jo_conflictReg.isNull()) 
        {
            // Iterate through exclusive reg set
            for (ArduinoJson::JsonVariant value : jo_conflictReg) 
            {
                StackScop uint16_t u16_exclusiveRedAddr = value.as<uint16_t>(); 

                // Iterate through the write addr set
                for(int i = 0; i < siz_request; i++)
                {

                    // Check if exclusive register is found in write addr set
                    if(arr_u16_writeAddr[i] == u16_exclusiveRedAddr)
                    {
                        vec_ConflictReg.push_back(u16_exclusiveRedAddr);
                        bol_ExclusiveRegFound = true;
                    }
                }
            }
        } 

        // If the write request has exclusive registers
        if(bol_ExclusiveRegFound)
        {
            // Print out all the exclusive registers that are locked
            for(int i = 0; i < vec_ConflictReg.size(); i++)
            {
                LOG_INF("cst_WriteControl() locking exclusive register: %d", vec_ConflictReg[i]);
            }
        
            LOG_INF("cst_WriteControl() Sending util_cmd_en-1 to EMS");

            // // Send "util_cmd_en - 1"
            // StackScop int32_t arr_s32_status[1] = {0};
            // StackScop int16_t s16_WriteCmdSt = IAL_Data_SetEMSConfiguration(IAL_TARGET_COMP, IAL_sz_UNIT_TYPE_EMS, COD_u8_NULL_UNIT_ID, 
            //                                 1, CST_sz_key_utilCmdEN, CST_sz_val_utilCmdEnOFF, arr_s32_status);
            // if (s16_WriteCmdSt != EXIT_SUCCESS)
            // {
            //     LOG_ERR("cst_WriteControl() failed to write util_cmd_en-1 to EMS");
            //     return -EXIT_FAILURE;
            // }   
        }
       
        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        // // Lock the PCS registers
        // StackScop int32_t arr_s32_lockresults[siz_request] = { -EXIT_FAILURE };
        // StackScop int16_t s16_WriteLockSt = IAL_Mbs_LockRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_writeAddr,
        //                             arr_s32_lockresults, siz_request);
        // if (s16_WriteLockSt != EXIT_SUCCESS)
        // {
        //     LOG_ERR("cst_WriteControl() failed to write register lock to EMS");
        //     return -EXIT_FAILURE;
        // }

        // // Insert register addr keys into locked register collection
        // pthread_mutex_lock(&cst_mtx_lockedRegisters);   
        // for(int i = 0; i < siz_request; i++)
        // {
        //     cst_uset_lockedRegisters.insert(arr_u16_writeAddr[i]);
        // }
        // pthread_mutex_unlock(&cst_mtx_lockedRegisters); 

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        // Write the desired values to the register addresses 
        StackScop int32_t arr_s32_results[siz_request] = { -EXIT_FAILURE };
        StackScop int16_t s16_WriteValSt = IAL_Mbs_WriteRegisters_Sourced(IAL_TARGET_COMP, CST_MOD_ACRONYM, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, 
                                                            arr_u16_writeAddr, arr_u16_writeVal, arr_s32_results, siz_request);
        if (s16_WriteValSt != EXIT_SUCCESS)
        {
            LOG_ERR("CST_SetControl() failed to write IEEE 2030.5 control values to PCS");
            return -EXIT_FAILURE;
        }

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        // // Send settings modified event 
        // StackScop int16_t s16_WriteEventSt = IAL_Ctl_SettingsModified(CST_sz_unitTypesPCS, 1);
        // if (s16_WriteEventSt != EXIT_SUCCESS)
        // {
        //     LOG_ERR("CST_SetControl() failed to send settings updated event");
        //     return -EXIT_FAILURE;
        // }
    }

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    // If the write type is reset, write the default values, unlock the registers, send settings modified event 
    else if(cst_enum_writeType == CST_RESET_REG)
    {

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        // Write the default values to the register addresses 
        StackScop int32_t arr_s32_results[siz_request] = { -EXIT_FAILURE };
        StackScop int16_t s16_WriteValSt = IAL_Mbs_WriteRegisters_Sourced(IAL_TARGET_COMP, CST_MOD_ACRONYM, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, 
                                                            arr_u16_writeAddr, arr_u16_writeVal, arr_s32_results, siz_request);
        if (s16_WriteValSt != EXIT_SUCCESS)
        {
            LOG_ERR("CST_SetControl() failed to write IEEE 2030.5 default control values to PCS");
            return -EXIT_FAILURE;
        }

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        // // Unlock the PCS registers
        // StackScop int32_t arr_s32_unlockresults[siz_request] = { -EXIT_FAILURE };
        // StackScop int16_t s16_WriteLockSt = IAL_Mbs_UnlockRegisters(IAL_TARGET_COMP, COD_sz_UNIT_TYPE_INV, COD_u8_NULL_UNIT_ID, arr_u16_writeAddr,
        //                             arr_s32_unlockresults, siz_request);
        // if (s16_WriteLockSt != EXIT_SUCCESS)
        // {
        //     LOG_ERR("cst_WriteControl() failed to write register unlock to EMS");
        //     return -EXIT_FAILURE;
        // }

        // // Remove register addr keys from locked register collection
        // pthread_mutex_lock(&cst_mtx_lockedRegisters);   
        // for(int i = 0; i < siz_request; i++)
        // {
        //     cst_uset_lockedRegisters.erase(arr_u16_writeAddr[i]);
        // }
        // pthread_mutex_unlock(&cst_mtx_lockedRegisters); 

        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        // // Send settings modified event 
        // StackScop int16_t s16_WriteEventSt = IAL_Ctl_SettingsModified(CST_sz_unitTypesPCS, 1);
        // if (s16_WriteEventSt != EXIT_SUCCESS)
        // {
        //     LOG_ERR("CST_SetControl() failed to send settings updated event");
        //     return -EXIT_FAILURE;
        // }
    }

    return EXIT_SUCCESS;
}

FileScop float cst_interpolateY(float f_xNew, float f_x1, float f_y1, float f_x2, float f_y2) 
{
    // If the two incoming X values are the same, we are returning the middle y value between the two 
    if (f_x1 == f_x2) 
    {
        return (f_y1 + f_y2) / 2; 
    } 

    // If the two incoming x values are different, we are returning the interpolated y value.
    else 
    {
        // Regular piecewise linear interpolation
        return f_y1 + (f_xNew - f_x1) * (f_y2 - f_y1) / (f_x2 - f_x1);
    }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileScop std::vector<std::pair<float, float>> CST_piecewiseLinearInterpolation(const std::vector<float>& vec_f_xRaw, 
                                        const std::vector<float>& vec_f_yRaw, uint8_t u8_numPoints) 
{

    // Create a vector to store the interpolated points
    std::vector<std::pair<float, float>> vec_interpolatedPoints;
    
    // Determine the step size for x-axis interpolation (divide x space into equal steps)
    float f_xMin = vec_f_xRaw.front();
    float f_xMax = vec_f_xRaw.back();
    float f_step = (f_xMax - f_xMin) / (u8_numPoints - 1);

    // Loop over each new x value
    for (int i = 0; i < u8_numPoints; ++i) 
    {
        float f_xNew = f_xMin + i * f_step;  // Calculate the new x position

        // Find the segment where f_xNew fits (between two original x points)
        auto it = std::lower_bound(vec_f_xRaw.begin(), vec_f_xRaw.end(), f_xNew);
        if (it == vec_f_xRaw.end()) 
        {
            break;  // If f_xNew is beyond the last point, exit loop
        }

        int32_t s32_idx = std::distance(vec_f_xRaw.begin(), it);
        if (s32_idx == 0) 
        {
            s32_idx = 1;  // Make sure we have a valid previous point
        }

        // Get the points for interpolation
        float f_x1 = vec_f_xRaw[s32_idx - 1], f_x2 = vec_f_xRaw[s32_idx];
        float f_y1 = vec_f_yRaw[s32_idx - 1], f_y2 = vec_f_yRaw[s32_idx];

        // Use the cst_interpolateY function to handle regular and vertical interpolation
        float f_yNew = cst_interpolateY(f_xNew, f_x1, f_y1, f_x2, f_y2);

        // Store the new interpolated point
        vec_interpolatedPoints.emplace_back(f_xNew, f_yNew);
    }

    return vec_interpolatedPoints;
}


FileScop void cst_LogAddrVal(const uint16_t* arr_u16_addresses, const uint16_t* arr_u16_values, size_t size) 
{
    for (size_t i = 0; i < size; ++i) 
    {
        LOG_INF("Address: 0x%04X, Value: %d", arr_u16_addresses[i], (int16_t)arr_u16_values[i]);
    }
    return;
}

FileScop uint8_t cst_getUtilCmdState(void)
{
    pthread_mutex_lock(&cst_mtx_utilCmdState);
    uint8_t u8_state = CST_u8_previousStateUtilCmdEN;
    pthread_mutex_unlock(&cst_mtx_utilCmdState);
    return u8_state;
}
FileScop void cst_setUtilCmdState(uint8_t u8_state)
{
    pthread_mutex_lock(&cst_mtx_utilCmdState);
    CST_u8_previousStateUtilCmdEN = u8_state;
    pthread_mutex_unlock(&cst_mtx_utilCmdState);
}